/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

/*
  Emits a constant tone.
  
  Authors
  =======
  
    Copyright (C) 2021-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
  
    2021-03-29    Initial release.
  
  Description
  ===========
  
    This firmware configures the RADIO to emit a so-called constant tone in
    Bluetooth Low Energy (BLE) or Nordic Proprietary (NRF) 2 MBit/s or BLE or
    NRF 1 MBit/s mode for calibration or measurement purposes.
    
    What tone is emitted, the used antenna as well as the used RADIO frequency
    can be configured.
    
  RADIO Mode
  ==========
    
    The used RADIO mode can be set using the {RADIO_CONFIG_MODE} macro according
    to the following table:
      
      {0}   ->   BLE 2 MBits/s
      {1}   ->   BLE 1 MBits/s
      {2}   ->   NRF 2 MBits/s
      {3}   ->   NRF 1 MBits/s
    
  Frequency Deviation
  ===================
    
    The used frequency deviation, i.e. what kind of constant tone is emitted,
    can be set using the {RADIO_CONFIG_DEVIATION} macro according to the
    following table:
      
      {0}   ->   No deviation. Emit carrier.
      {1}   ->   Deviate to lower frequencies. Transmit 0s.
      {2}   ->   Deviate to higher frequencies. Transmit 1s.
    
  Antenna
  =======
    
    The used antenna can be set using the {RADIO_CONFIG_ANTENNA} macro according
    to the following table. Note, that this configuration has no effect for
    boards with just a single antenna:
      
      {0}   ->   No choice. Use standard antenna.
      {1}   ->   Use antenna 1.
      {2}   ->   Use antenna 2.
            ...  
      {8}   ->   Use antenna 8.
  
  Frequency
  =========
    
    The RADIO frequency can be chosen from 2402 MHz to 2481 MHz in steps of
    1 MHz by specifiying {RADIO_CONFIG_ANTENNA} to be the desired frequency in
    MHz minus 2400.
    
*/


#include <rss2i/clock.h>
#include <rss2i/measure.h>
#include <rss2i/battery.h>
#include <rss2i/led.h>
#include <rss2i/debug.h>
#include <rss2i/control.h>
#include <rss2i/icache.h>
#include <rss2i/radio.h>
#include <rss2i/power.h>
#include <rss2i/watchdog.h>
#include <rss2i/board.h>
#include <rss2i/mlc.h>
#include <rss2i/timer.h>
#include <rss2i/scheduler.h>
#include <rss2i/memory.h>
#include <rss2i/switch.h>

#include <rss2i/config/generated.radio.h>


_Static_assert(
	
	   RADIO_CONFIG_MODE == 0
	|| RADIO_CONFIG_MODE == 1
	|| RADIO_CONFIG_MODE == 2
	|| RADIO_CONFIG_MODE == 3,
	"Invalid RADIO MODE setting."
);

_Static_assert(
	
	   RADIO_CONFIG_DEVIATION == 0
	|| RADIO_CONFIG_DEVIATION == 1
	|| RADIO_CONFIG_DEVIATION == 2,
	"Invalid frequency DEVIATION setting."
);

_Static_assert(
	
	   RADIO_CONFIG_ANTENNA == 0
	|| RADIO_CONFIG_ANTENNA == 1
	|| RADIO_CONFIG_ANTENNA == 2
	|| RADIO_CONFIG_ANTENNA == 3
	|| RADIO_CONFIG_ANTENNA == 4
	|| RADIO_CONFIG_ANTENNA == 5
	|| RADIO_CONFIG_ANTENNA == 6
	|| RADIO_CONFIG_ANTENNA == 7
	|| RADIO_CONFIG_ANTENNA == 8,
	"Invalid ANTENNA setting."
);

_Static_assert(
	
	   RADIO_CONFIG_FREQUENCY <= 81
	&& RADIO_CONFIG_FREQUENCY >=  2
	&& (uint8_t) RADIO_CONFIG_FREQUENCY == RADIO_CONFIG_FREQUENCY,
	"Invalid RADIO FREQUENCY setting."
);


int main(){
	
	RadioMode mode;
	SwitchAntenna ant;
	uint8_t freq;
	RadioFrequencyDeviation dev;
	
	led_init();          /* Controls the onboard LEDs.                        */
	timer_init();        /* Provides some timers.                             */
	debug_init();        /* Debug and reporting system.                       */
	board_init();        /* Holds board-specific parameters.                  */
	power_init();        /* Enables the internal DC/DC regulator.             */
	clock_init();        /* Configures the High Frequency Clock (HFCLK).      */
/*	battery_init();    *//* Monitors the battery voltage.                     */
	mlc_init();          /* Monitors the main loop frequency.                 */
	control_init();      /* Handles incoming commands.                        */
	measure_init();      /* Enters control mode.                              */
	scheduler_init();    /* Schedules all periodic tasks.                     */
/*	icache_init();     *//* Initializes the nRF iCache technology.            */
	
	watchdog_init();     /* Enables the watchdog.                             */
	
	memory_init();       /* API for the external memory.                      */
	switch_init();       /* API for the external HF switch.                   */
	
	
	led_set(LED_YELLOW, LED_ON);
	led_set(LED_GREEN, LED_BLINK);
	
	/* Derive configuration. */
	
	switch(RADIO_CONFIG_MODE){
		
		default:
		/* fall through */
		case 0:
			mode = RADIO_MODE_BLE_2MBIT;
		break;
		case 1:
			mode = RADIO_MODE_BLE_1MBIT;
		break;
		case 2:
			mode = RADIO_MODE_NRF_2MBIT;
		break;
		case 3:
			mode = RADIO_MODE_NRF_1MBIT;
		break;
	}
	
	switch(RADIO_CONFIG_DEVIATION){
		
		default:
		/* fall through */
		case 0:
			dev = RADIO_FREQUENCY_DEVIATION_CENTER;
		break;
		case 1:
			dev = RADIO_FREQUENCY_DEVIATION_TRANS0;
		break;
		case 2:
			dev = RADIO_FREQUENCY_DEVIATION_TRANS1;
		break;
	}
	
	if(RADIO_CONFIG_ANTENNA){
		
		ant = SWITCH_ANTENNA_1 + RADIO_CONFIG_ANTENNA - 1;
	}
	else{
		
		ant = switch_get_standard_antenna();
	}
	
	freq = RADIO_CONFIG_FREQUENCY;
	
	/* Emit constant tone. */
	
	radio_emit_constant(freq, mode, dev, ant);
	
	while(1){
		
		scheduler();
	}
	
	return 0;
}

