/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

/*
  Thread safe (one producer, one consumer) circular buffers. Further
  documentation may be found in the corresponding source file.
  
  Authors
  =======
  
    Copyright (C) 2018-2019  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
  
    2018-07-17    Initial release.
    2020-02-08    Remove {empty} flag.
    2020-02-08    Add support for Direct Memory Access (DMA).
    2020-02-12    The buffer pointers itself have to be {volatile}.
*/

#include <stdint.h>
#include <stddef.h>
#include <stdbool.h>

typedef enum{
	
	BUFFER_ROW,
	BUFFER_STRING,
	
} BufferDataType;

typedef union{
/*
  Type for storing an array in memory with different interfaces to the user.
*/
	
	volatile char * volatile c;
	volatile uint8_t * volatile u;
	
} BufferDataPtr;

typedef struct{
/*
  Type describing a circular buffer of arrays. The function {buffer_init()}
  should be used to initialise it. The other functions {buffer_*()} should be
  used to handle it. Its members should never be accessed directly.
*/
	
	volatile size_t length;
	volatile size_t width;
	volatile BufferDataPtr dataPtr;
	volatile size_t * volatile size;
	volatile size_t writePointer;
	volatile size_t readPointer;
	volatile size_t lockPointer;
	
} Buffer;

void buffer_init_rows(Buffer * buffer, uint8_t * memory, size_t * size, size_t length, size_t width);
void buffer_init_strings(Buffer * buffer, char * memory, size_t * size, size_t length, size_t width);
void buffer_init(Buffer * buffer, BufferDataPtr dataPtr, size_t * size, size_t length, size_t width);
bool buffer_append_row(Buffer * buffer, uint8_t * row, size_t size);
bool buffer_append_string(Buffer * buffer, const char * string);
uint8_t * buffer_new_row_address(Buffer * buffer);
bool buffer_new_row_finish(Buffer * buffer, uint8_t * addr, size_t size);
BufferDataPtr buffer_get(Buffer * buffer, BufferDataType);
uint8_t * buffer_get_row(Buffer * buffer);
char * buffer_get_string(Buffer * buffer);
void buffer_unlock(Buffer * buffer);
size_t buffer_next_size(Buffer * buffer);
void buffer_empty(Buffer * buffer);
size_t buffer_vacant_entries(Buffer * buffer);

