/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

/*
  Process incoming commands. Further documentation may be found in the 
  corresponding source file.
  
  Authors
  =======
    
    Copyright (C) 2018-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
    
    2019-04-16    Initial public release.
    2019-05-29    Added support for iterable commands.
    2020-03-31    Adjusted {CMD_MAX_LENGTH} and {CMD_MAX_RESPONSE_LENGTH}.
    2020-03-31    Added commands 0x18 {CMD_RETURN_DEVICE_STATUS} as well as
                  0x19 {CMD_REPORT_ERROR_INFORMATION}.
    2020-08-03    Added {CMD_SET_CTE_CONFIGURATION}.
    2020-08-14    Added {CMD_*CTE_SAMPLE*} commands.
    2020-08-25    Added iterable version of {CMD_RETURN_CTE_SAMPLE_SHEET}.
    2021-03-23    Moved 0x16 to 0x17 and Added new 0x16 {CMD_SET_RADIO_MODE}.
    2021-03-23    Moved 0x18 to 0x19 and and 0x19 to 0x1a.
    2021-04-19    Renamed 0x1b to 0x1f.
    2021-04-20    Added 0x18 {CMD_SET_SAMPLE_CONFIGURATION}.
*/

#include <stdint.h>
#include <stdbool.h>
#include <stddef.h>


#define CMD_MAX_LENGTH 256
#define CMD_MAX_RESPONSE_LENGTH 256

typedef enum {
	
	/* 0x00 */ CMD_TELL_ME_WHO_YOU_ARE,
	/* 0x01 */ CMD_SET_BAUDRATE,
	/* 0x02 */ CMD_SET_PREFERENCES,
	/* 0x03 */ CMD_RETURN_MEASUREMENT_SESSION_DATA,
	/* 0x04 */ CMD_RETURN_CONTENT_OF_SPECIFIED_PAGE,
	/* 0x05 */ CMD_RETURN_CONTENT_OF_SPECIFIED_PAGES,
	/* 0x06 */ CMD_RETURN_CONTENT_OF_ABSOLUTE_PAGE,
	/* 0x07 */ CMD_RETURN_CONTENT_OF_ABSOLUTE_PAGES,
	/* 0x08 */ CMD_SET_LIVE_LOG_CONFIGURATION,
	/* 0x09 */ CMD_START_LIVE_LOG,
	/* 0x0a */ CMD_STOP_LIVE_LOG,
	/* 0x0b */ CMD_ENTER_CONTROL_MODE,
	/* 0x0c */ CMD_FORCE_CONTROL_MODE,
	/* 0x0d */ CMD_JUMP_ON_MEASUREMENT,
	/* 0x0e */ CMD_START_UP_MEASUREMENT,
	/* 0x0f */ CMD_RESERVED_0x0f,
	/* 0x10 */ CMD_SET_MEASUREMENT_CONFIGURATION,
	/* 0x11 */ CMD_SET_SYNCHRONIZATION_MAP,
	/* 0x12 */ CMD_SET_POOR_SYNC_THRESHOLDS,
	/* 0x13 */ CMD_SET_OUT_OF_SYNC_THRESHOLDS,
	/* 0x14 */ CMD_SET_REMOTE_RSS_INPUT_MAP,
	/* 0x15 */ CMD_SET_REMOTE_RSS_OUTPUT_MAP,
	/* 0x16 */ CMD_SET_RADIO_MODE,
	/* 0x17 */ CMD_SET_CTE_CONFIGURATION,
	/* 0x18 */ CMD_SET_SAMPLE_CONFIGURATION,
	/* 0x19 */ CMD_RETURN_DEVICE_STATUS,
	/* 0x1a */ CMD_REPORT_ERROR_INFORMATION,
	/* 0x1b */ CMD_COLLECT_SAMPLES,
	/* 0x1c */ CMD_RETURN_SAMPLE_STATUS,
	/* 0x1d */ CMD_CLEAR_SAMPLE_REQUEST,
	/* 0x1e */ CMD_RETURN_SAMPLE_SHEET,
	/* 0x1f */ CMD_RETURN_SAMPLE_SHEETS,

	CMD_ALWAYS_LAST_ITEM,

} CMDNames;

typedef struct{
	
	size_t length;           /* Command's length.                             */
	size_t responseLength;   /* Response's length.                            */
	bool targetFieldPresent; /* Whether a target field is present or not.     */
	bool responseExpected;   /* Whether a response is expected or not.        */
	bool isIterable;         /* Whether this command is iterable or not.      */
	size_t timeoutLength;    /* Number of bytes specified below. The missing  */
	                         /* bytes of the response may have arbitrary      */
	                         /* values.                                       */
	uint8_t * timeoutData;   /* First few bytes of the response in the case   */
	                         /* of a RADIO timeout.                           */
	size_t busyLength;       /* Number of bytes specified below. The missing  */
	                         /* bytes of the response may have arbitrary      */
	                         /* values.                                       */
	uint8_t * busyData;      /* First few bytes of the response in the case   */
	                         /* that a command should be forwarded over-the-  */
	                         /* air but the forwarding device is not in       */
	                         /* control mode yet.                             */
	
} CMD;

uint8_t cmd_is_iterable(uint8_t * cmd);
void cmd_iterate(uint8_t * cmd);
bool cmd_handle(uint8_t * cmd);
void cmd_timeout_response(uint8_t cmd, uint8_t * response);
void cmd_busy_response(uint8_t cmd, uint8_t * response);
void cmd_init(void);

