/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

/*
  Some settings regarding the dubug module.
  
  Authors
  =======
    
    Copyright (C) 2018-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
    
  Changelog
  =========
    
    2019-07-02    Initial public release.
    2019-10-31    Reduced the buffer length to 32 (Saving some RAM).
    2020-02-12    {REPETITION_DELAY} is now in ms.
    2020-02-13    Introduced {DEBUG_MLC_PRINT}.
    2020-06-08    Reduced the buffer length to 16 (Saving some RAM).
    2020-08-12    Enhance frame timing debug functionality.
    2020-08-13    Introduced {DEBUG_MEASUREMENT_FRAME} level {3}.
    2020-08-13    Enhance {DEBUG_MEASUREMENT_FRAME} level {3} functionality.
    2020-08-17    Reduced the buffer length to 8 (Saving some RAM).
    2021-03-05    Refine the {DEBUG_MEASUREMENT_FRAME} documentation.
    2021-03-10    Added {TOGGLE} Makefile switch.
    2021-03-25    Update doc with regard to PPI and GPIOTE toggling.
    2021-03-25    Update doc with regard to overlapping events.
*/


/*
   Some settings that affect the serial debug output.
*/

/* Pause between repeated error/warning messages in ms. */
#define DEBUG_REPETITION_DELAY   2000

/* Buffer dimensions */
#define DEBUG_BUFFER_LENGTH         8
#define DEBUG_BUFFER_WIDTH        110

/* Debug baudrate. This shouldn't be too high since this is a software UART. */
#define DEBUG_BAUDRATE           9600

/* Whether to print the main loop counter values over serial debug interface. */ 
#define DEBUG_PRINT_MLC          true


/*
  Switch for debugging measurement frames. The actual {DEBUG_MEASUREMENT_FRAME}
  level is specified via user generated config file.
    
    {0}: Disabled.
    {1}: {DEBUG_MEASUREMENT_FRAME_GPIO} pin gets toggled at the beginning of
         every frame. This is useful in order to measure the exact frame length
         with a scope, e.g. when trying to fit {MEASURE_*_*_NEGATIVE_TX_DELAY}
         and {MEASURE_*_*_SYNC_DELAY}.
    {2}: {DEBUG_MEASUREMENT_FRAME_GPIO} pin gets toggled multiple times during
         a frame. This allows precise debugging of the internal schedule
         management and the RADIO timings.
         
         GPIO behaviour when sending a measurement packet:
         
         
                    RADIO disabled (CTE sucessfully transmitted (if any)). --.
                                                                             |
                        Transmission successfull (actual packet). --.        |
                                                                    |        |
           .-- Frame's beginning. Starting early preparation stage. |        |
           |                                                        |        |
           |          .-- Initialize transmission.                  |        |
           |          |                                             |        |
           |          |                 .-- Ready to transmit       |        |
           |          |                |    (TX ramp-up finished).  |        |
           |          |                |                            |        |
           |          |                |   .-- Start transmission.  |        |
           |          |                |   |                        |        |
            _          ________________                                      
           | \/|      |                |   ||                       ||       ||
           |   |      |                |   ||                       ||       ||
          _|   |______|                |___||_______________________||_______||_
         
         
         
         GPIO behaviour when receiving/trying to receive a measurement packet:
         
         
                                                  Optional: RADIO disabled --.
                                                  (CTE successfully          |
                                                  received (if any)).        |
                                                                             |
                                                                             |
                                 Optional: Successfully received. --.        |
                                 The receiver now synchronizes      |        |
                                 himself to this event.             |        |
                                                                    |        |
           .-- Frame's beginning. Starting early preparation stage. |        |
           |                                                        |        |
           |          .-- Initialize reception.                     |        |
           |          |                                             |        |
           |          |                 .-- Ready to transmit.      |        |
           |          |                 |   (RX ramp-up finished)   |        |
            _          _________________                                      
           | \/|      |                 |                           ||       ||
           |   |      |                 |                           ||       ||
          _|   |______|                 |___________________________||_______||_
          
          
          
          Note, that due to the tight timings in {RADIO_MODE_BLE_2MBIT} and
          the CRC check after packet reception, the disable event may be delayed
          in a way such that it overlapse with the beginning of the next frame.
          In this case, the frame's beginning looks like this:
          
          
          
           .-- First toggle of the frame's beginning
           |   (triggered via PPI and GPIOTE).
           |
           |   .-- RADIO disabled (see above).
           |   |
           |   |   .-- Other toggles of the frame's beginning
           |   |   |   (triggered by software during the ISR).
           |   |   |
            ___  __
           |   ||  \/|
           |   ||    |
          _|   ||    |___...
         
         
          
    {3}: {DEBUG_MEASUREMENT_FRAME_GPIO} pin gets toggled multiple times during
         a frame as it is shown above. Additionally, during transmission/
         reception preperation and initialization, there are 4/3 togggles that
         indicate the status of the preparation and initialization phases. This
         allows an even more precise debugging of the internal RADIO timings.
         
         GPIO behaviour during transmission preperation and initialization
         phase:
         
         
           .-- Frame's beginning. Starting early preparation stage.
           |                            
           |                .-- Initialize transmission.               
           |                |                                          
           |                |                                          
           |                |                  Start transmission. --.
           |                |                                        |
           |                |             Ready to transmit -----.   |
           |                |             (TX ramp-up finished). |   |
            _                ____  _____  ___________  __________   
           | \/|      ||    |    ||     ||           ||          |   ||
           |   |      ||    |    ||     ||           ||          |   ||
          _|   |______||____|    ||     ||           ||          |___||______...
          
                      |          |      |            |
                      |          |      |            '-- Radio packet data and
                      |          |      |                remote RSS data
                      |          |      |                successfully
                      |          |      |                initialized.
                      |          |      |                  
                      |          |      '-- Ramp up RADIO TX.
                      |          |
                      |          '--Measurement frame configured (measure.c).
                      |             Started initialization of the RADIO
                      |             peripheral (radio.c).
                      |
                      '-- Frame and RADIO prepared.
         
         
         
         GPIO behaviour during reception preperation and initialization phase:
         
         
           .-- Frame's beginning. Starting early preparation stage.
           |                            
           |                .-- Initialize reception.                  
           |                |                                          
           |                |                                          
           |                |             Ready to receive ------.     
           |                |             (RX ramp-up finished). |     
            _                ____  _____  _______________________      
           | \/|      ||    |    ||     ||                       |     
           |   |      ||    |    ||     ||                       |     
          _|   |______||____|    ||     ||                       |___________...
          
                      |          |      |                  
                      |          |      '-- Ramp up RADIO RX.
                      |          |
                      |          '--Measurement frame configured (measure.c).
                      |             Started initialization of the RADIO
                      |             peripheral (radio.c).
                      |
                      '-- Frame and RADIO prepared.
         
*/
#define DEBUG_MEASUREMENT_FRAME_GPIO	GPIO6
#define DEBUG_MEASUREMENT_FRAME_PIN 	6

