/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

/*
  Some settings regarding the RSS measurements.
  
  Authors
  =======
    
    Copyright (C) 2018-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
    
  Changelog
  =========
    
    2019-07-02    Initial public release.
    2019-10-31    Reduced the buffer length to 16 (Saving some RAM).
    2020-02-13    Introduced {MEASURE_LOG_MISSED_FRAMES}.
    2020-06-08    Reduced the buffer length to 4 (Saving some RAM).
    2020-06-19    Store measurement results on buffer using DMA.
    2020-07-31    Outsource livelog functionality into own module.
    2020-08-03    Fit measurement schedule to variable CTE lengths.
    2020-08-13    Increase the buffer width in order to store IQ samples.
    2020-08-13    Make the measurement frame timings BLE 5.1 CTE ready.
    2020-08-13    Rearange measurement buffer and add {BUFFER_META_DATA_OFFSET}.
    2021-02-25    Added BLE 2MBit prefixes.
    2021-03-10    Fitted measurement timings to long CTE preparation time.
    2021-03-17    Added RADIO timings for different modes.
    2021-03-18    Added precise measurement buffer meta data offsets.
    2021-03-18    Enlarged the measure buffer meta data by four bytes.
    2021-03-18    Added compatibility mode timings.
    2021-03-18    Make the measurement frame timings BLE 5.1 CTE ready again.
    2021-03-24    Tuned the measurement frame timings and updated doc.
    2021-04-20    Added the {*_SAMPLE_STORAGE_TIME}s.
    2021-04-30    Enlarged the {*_SAMPLE_STORAGE_TIME}s.
    2021-05-04    Enlarge number of disruptive frames in sample storage mode.
    2021-05-04    Enlarged the {*_SAMPLE_STORAGE_TIME}s.
    
  Measurement frame timing and its nomenclature
  =============================================
  
  
    .-- Frame's begin: Disable the RADIO and prepare
    |   it for the upcoming initialization.
    |
    |         .-- Initialize the RADIO for the upcoming task.
    |         |
    |         |              .-- Trigger the transmission of a
    |         |              |   measurement packet. Due to processing
    |         |              |   time, this happens slightly before the
    |         |              |   actual transmission's start.
    |         |              |
    |         |              |                 .-- Synchronization delay.
    |         |              |                 |
    |         |              |               |---|
     _________ ___________ ____ _____________ __________ ____ __________
    |         |           |    |             |          |    |          |
    | Disable |  Ramp up  | GI | Measurement | Optional | GI | Optional |
    |  RADIO  |   RADIO   |    |   packet    |   CTE    |    | Padding  |
    |_________|___________|____|_____________|__________|____|__________|
    
                             |-|                 |      |----|----------|
                              |                  |         |       |
         Negative TX delay. --'                  |         |       |
                                                 |         |       |
     This is the synchronization's fixed point --'         |       |
     that is triggered by the CRCOK event. This            |       |
     event occures with a slight delay after               |       |
     the actual packet.                                    |       |
                                                           |       |
                  Guard Interval (GI) that compensates ----'       |
                  synchronization inaccuracies. Note,              |
                  that devices that are part of a CTE-             |
                  enabled measurement but do not                   |
                  perform an actual CTE just have an               |
                  extended post guard interval.                    |
                                                                   |
                               Additional padding time such that --'
                               the total, i.e. padded, frame
                               length is the same for all
                               RADIO modes in compatibility
                               mode. This padding is also used
                               in raw sample storage  mode in
                               order to reduce the sampling rate
                               to be able to store the immense
                               number of samples on the flash
                               memory.
  
*/

#include <rss2i/config/cte.h>


/*
  Frame timings. 
*/

#define MEASURE_X_2MBIT_FRAME_LENGTH                   200 /*      µs */
#define MEASURE_X_2MBIT_RADIO_RAMP_UP_TIME              48 /*      µs */
#define MEASURE_X_2MBIT_RADIO_DISABLE_TIME              22 /*      µs */
#define MEASURE_X_2MBIT_PACKET_LENGTH                  100 /*      µs */
#define MEASURE_X_2MBIT_GUARD_INTERVAL_LENGTH           15 /*      µs */
#define MEASURE_X_2MBIT_TX_NEGATIVE_DELAY               53 /* 1/16 µs */
#define MEASURE_X_2MBIT_COMPATIBILITY_MODE_DELAY       400 /*      µs */
#define MEASURE_X_2MBIT_SYNC_DELAY                      80 /* 1/16 µs */
#define MEASURE_X_2MBIT_SAMPLE_STORAGE_TIME          24800 /*      µs */

#define MEASURE_X_1MBIT_FRAME_LENGTH                   250 /*      µs */
#define MEASURE_X_1MBIT_RADIO_RAMP_UP_TIME              50 /*      µs */
#define MEASURE_X_1MBIT_RADIO_DISABLE_TIME              40 /*      µs */
#define MEASURE_X_1MBIT_PACKET_LENGTH                  120 /*      µs */
#define MEASURE_X_1MBIT_GUARD_INTERVAL_LENGTH           20 /*      µs */
#define MEASURE_X_1MBIT_TX_NEGATIVE_DELAY               55 /* 1/16 µs */
#define MEASURE_X_1MBIT_COMPATIBILITY_MODE_DELAY       350 /*      µs */
#define MEASURE_X_1MBIT_SYNC_DELAY                     150 /* 1/16 µs */
#define MEASURE_X_1MBIT_SAMPLE_STORAGE_TIME          24750 /*      µs */

#define MEASURE_IEEE_250KBIT_FRAME_LENGTH              600 /*      µs */
#define MEASURE_IEEE_250KBIT_RADIO_RAMP_UP_TIME         50 /*      µs */
#define MEASURE_IEEE_250KBIT_RADIO_DISABLE_TIME         40 /*      µs */
#define MEASURE_IEEE_250KBIT_PACKET_LENGTH             416 /*      µs */
#define MEASURE_IEEE_250KBIT_GUARD_INTERVAL_LENGTH      47 /*      µs */
#define MEASURE_IEEE_250KBIT_TX_NEGATIVE_DELAY        -200 /* 1/16 µs */
#define MEASURE_IEEE_250KBIT_COMPATIBILITY_MODE_DELAY    0 /*      µs */
#define MEASURE_IEEE_250KBIT_SYNC_DELAY                625 /* 1/16 µs */
#define MEASURE_IEEE_250KBIT_SAMPLE_STORAGE_TIME     24400 /*      µs */
/*
  Buffer dimensions for caching measurement results.
*/

#define MEASURE_BUFFER_LENGTH               4
/*
  Measure buffer content:
  
    - RADIO packet data (16 bytes)
    - IQ samples {CTE_MAX_NUMBER_OF_SAMPLES * 4} bytes
    - RSS data (2 bytes) (meta data)
    - Amount of RSS samples (1 byte) (meta data)
    - Amount of IQ samples(/raw samples) (1 byte) (meta data)
    - Used RADIO mode during packet (1 byte) (meta data)
    - Entry was written to? (1 byte) (meta data)
    - Raw sample type (1 byte) (meta data)
    - RESERVED for future extension (and alignment) (1 byte) (meta data)
*/
#define MEASURE_BUFFER_WIDTH               (16                            \
                                          + CTE_MAX_NUMBER_OF_SAMPLES * 4 \
                                          + 2                             \
                                          + 1                             \
                                          + 1                             \
                                          + 1                             \
                                          + 1                             \
                                          + 1                             \
                                          + 1)

/* Position of the radio packet inside the measure buffer. */
#define MEASURE_BUFFER_RADIO_PACKET_OFFSET  0

/* Position of the meta data inside the measure buffer. */
#define MEASURE_BUFFER_META_DATA_OFFSET    (16 + CTE_MAX_NUMBER_OF_SAMPLES * 4)
#define MEASURE_BUFFER_RSSAVERAGE_OFFSET   (MEASURE_BUFFER_META_DATA_OFFSET + 0)
#define MEASURE_BUFFER_RSSAMOUNT_OFFSET    (MEASURE_BUFFER_META_DATA_OFFSET + 2)
#define MEASURE_BUFFER_IQSAMOUNT_OFFSET    (MEASURE_BUFFER_META_DATA_OFFSET + 3)
#define MEASURE_BUFFER_RADIOMODE_OFFSET    (MEASURE_BUFFER_META_DATA_OFFSET + 4)
#define MEASURE_BUFFER_WRITTENTO_OFFSET    (MEASURE_BUFFER_META_DATA_OFFSET + 5)
#define MEASURE_BUFFER_SAMPLETYP_OFFSET    (MEASURE_BUFFER_META_DATA_OFFSET + 6)
#define MEASURE_BUFFER_RESERVED_OFFSET     (MEASURE_BUFFER_META_DATA_OFFSET + 7)


/* Default frequency for resynchonization. Used when this device */
/* has no passive part within a measurement configuration.       */
#define MEASURE_DEFAULT_FREQUENCY          17 /* + 2400 MHz */


/*
  Maximal number of frames within a measurement cycle. This value must not
  exceed {250} and should be at least {5}. Changing this value affects the valid
  ranges described in the user documentation but may save some RAM space.
*/
#define MEASURE_MAX_FRAME_COUNT           250

/*
  Approximate minimal number of disruptive frames that get sent when trying
  to force other divices into control mode. This is useful in order to avoid
  quitting the disrupting process to early e.g. when a remote device still
  sends measurement packets who wasn't received correctly due to temporary
  noise or other pertubations.
  
  Note, that changing this value affects the timings of the force-control-mode
  command.
*/
#define MEASURE_MIN_DISRUPT_FRAMES_X_2MBIT                           5000
#define MEASURE_MIN_DISRUPT_FRAMES_X_1MBIT                           4000
#define MEASURE_MIN_DISRUPT_FRAMES_IEEE_250KBIT                      2000
#define MEASURE_MIN_DISRUPT_FRAMES_X_2MBIT_COMPATIBILITY_MODE        2000
#define MEASURE_MIN_DISRUPT_FRAMES_X_1MBIT_COMPATIBILITY_MODE        2000
#define MEASURE_MIN_DISRUPT_FRAMES_IEEE_250KBIT_COMPATIBILITY_MODE   2000
#define MEASURE_MIN_DISRUPT_FRAMES_SAMPLE_STORAGE                     250

