/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

/*
  Some settings regarding the RADIO module.
  
  Authors
  =======
    
    Copyright (C) 2018-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
    
  Changelog
  =========
    
    2019-07-02    Initial public release.
    2019-09-26    Introduce {RADIO_RSSI_SAMPLE_DELAY}.
    2019-11-04    Adjusted the recommended value for {RADIO_RSSI_SAMPLE_DELAY}.
    2020-06-08    Reduced the buffer lengths to 2 (Saving some RAM).
    2020-06-19    Support DMA based measurment radio packet data RAM buffering.
    2020-08-13    Add {RADIO_MEASURE_CTE_SAMPLES_OFFSET}.
    2020-08-13    Rearange RADIO packet: Fit {RADIO_MEASURE_CTE_SAMPLES_OFFSET}.
    2021-02-25    Add 2MBIT RADIO mode prefix.
    2021-02-25    Add CRC defines.
    2021-02-26    Add {RADIO_PACKET_WIDTH}.
    2021-02-26    Add IEEE 802.15.4 defines.
    2021-02-28    Add BLE/NRF 1/2MBit defines.
    2021-02-28    Add {RADIO_MAX_PAYLOAD}.
    2021-03-23    Add {RADIO_MEASURE_*_*_SYNC_DELAY}.
    2021-03-23    Tuned compatibility mode settings.
*/

/*
  General settings.
*/

/* Size of the RADIO packet memory.                        */
/* This is used for OTA message transmission and reception */
/* as well as to transmission of radio packets.            */
#define RADIO_PACKET_WIDTH                      257

/* IEEE custom 802.15.4 Start Frame Delimiter (SFD) */
#define RADIO_IEEE_250KBIT_SFD                 0xcb

/* Maximal payload size. */
#define RADIO_MAX_PAYLOAD                       255

/*
  Some settings that affect the RSS measurements.
*/

/* Default RADIO TX power. */
#define RADIO_TX_POWER            3 /* 3 -> 0 dBm */

/* The delay between two successive RSSI conversions in 1/16 µs. The lower  */
/* this value the more RSSI samples are taken within one measurment packet. */
/* An extremly low value may lead to a system freeze or malfunction of      */
/* important processing tasks. You will be alerted in this case. However,   */
/* values below {6} are NOT recommended.                                    */
/*                                                                          */
/* Any change should be reviewed critically when operating in IEEE 802.15.4 */
/* RADIO mode due to the incorecct RSSI sampling errata described below.    */
#define RADIO_RSSI_SAMPLE_DELAY  6

/* Duration of the RSS measurement during packet reception in compatibility */
/* mode in number of bits.                                                  */
#define RADIO_MEASURE_DURATION_IEEE_250KBIT_COMPATIBILITY_MODE  17
#define RADIO_MEASURE_DURATION_X_1MBIT_COMPATIBILITY_MODE       75
/* Duration of the RSS measurement during packet reception in normal IEEE   */
/* 802.15.4 operation. The RSS samples are sometimes incorrect during       */
/* packet reception. This is a nasty workaround for this behaviour.         */
#define RADIO_MEASURE_DURATION_IEEE_250KBIT_ERRATA_WORKAROUND   50

/* Length of a measurement packet's payload in bytes.            */
/* This should be compatible with the measurement frame timings. */
#define RADIO_MEASURE_X_1MBIT_PAYLOAD_LENGTH       5
#define RADIO_MEASURE_X_2MBIT_PAYLOAD_LENGTH      14
#define RADIO_MEASURE_IEEE_250KBIT_PAYLOAD_LENGTH  7 /* 2 bytes CRC included! */

/* Where to find measurement meta data within the measurement radio packet. */
/* BLE/NRF 1MBit mode. */
#define RADIO_MEASURE_X_1MBIT_FRAME_DATA_OFFSET   0
#define RADIO_MEASURE_X_1MBIT_LENGTH_DATA_OFFSET  1 /* Required by BLE        */
#define RADIO_MEASURE_X_1MBIT_CYCLE_DATA_OFFSET   2
#define RADIO_MEASURE_X_1MBIT_REMOTE_DATA_OFFSET  5
#define RADIO_MEASURE_X_1MBIT_CTE_SAMPLES_OFFSET  8 /* Waste 1B for alignment */
/* BLE/NRF 2MBit mode. */
#define RADIO_MEASURE_X_2MBIT_FRAME_DATA_OFFSET   0
#define RADIO_MEASURE_X_2MBIT_LENGTH_DATA_OFFSET  1 /* Required by BLE        */
#define RADIO_MEASURE_X_2MBIT_CYCLE_DATA_OFFSET   2
#define RADIO_MEASURE_X_2MBIT_REMOTE_DATA_OFFSET  5
#define RADIO_MEASURE_X_2MBIT_CTE_SAMPLES_OFFSET 16 /* IQ sampling during CTE */
/* IEEE 250kBit mode. */
#define RADIO_MEASURE_IEEE_250KBIT_FRAME_DATA_OFFSET   2
#define RADIO_MEASURE_IEEE_250KBIT_LENGTH_DATA_OFFSET  0 /* Required.         */
#define RADIO_MEASURE_IEEE_250KBIT_CYCLE_DATA_OFFSET   3
#define RADIO_MEASURE_IEEE_250KBIT_STATUS_DATA_OFFSET  1


/*
  Some CRC defines
*/

/* According to BLE standard. */
#define RADIO_CRCPOLY_BLE                         UINT32_C(0x65b)
#define RADIO_CRCINIT_BLE                         UINT32_C(0x55555555)
#define RADIO_CRCLENG_BLE                         3

/* According to IEEE 802.15.4 standard. */
#define RADIO_CRCPOLY_IEEE                        UINT32_C(0x11021)
#define RADIO_CRCINIT_IEEE                        UINT32_C(0x00000)
#define RADIO_CRCLENG_IEEE                        2


/*
  Some settings that affect the OTA transmission/reception.
*/

/* Maximal OTA message size (PAYLOAD + S0) receivable by the RADIO      */
/* peripheral. Extra size for USB-CRC bytes that may be added later on. */
#define RADIO_OTA_RECV_BUFFER_WIDTH (256 + 4)

/* Maximal OTA message size (PAYLOAD + S0) the RADIO peripheral may send. */
/* Extra size for storing the targeted device inside the output buffer.   */
/* Extra size for the length field inside the RADIO packet.               */
#define RADIO_OTA_SEND_BUFFER_WIDTH (256 + 1 + 1)

/* Buffer dimensions. */
#define RADIO_OTA_RECV_BUFFER_LENGTH 2
#define RADIO_OTA_SEND_BUFFER_LENGTH 2

/* Affects the LED behaviour: The lower this value the */
/* faster the LED gets disabled after an OTA event.    */
#define RADIO_LED_TIMEOUT 60000

/* Default OTA frequency. */
#define RADIO_CONTROL_FREQUENCY 9 /* + 2400 [MHz] */

