/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

/*
  BLE 5.1 Constant Tone Extension (CTE) related functionality. Additional
  documentation may be found in the corresponding source file.
  
  Authors
  =======
  
    Copyright (C) 2020-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
  
    2020-07-31    Initial release.
    2020-07-31    Introduced {CTEAntenna*} enums and {*_is_valid()} methods.
    2020-08-03    Introduced {CTEAntennaPattern} and {CTEType} enums.
    2020-08-03    Added {cte_type_is_valid()} method.
    2020-08-03    Added {cte_length()} function.
    2020-08-03    Added {CTE_ANTENNA_PATTERN_DISABLED}.
    2020-08-13    Add {cte_radio_init_*()} functions to configure CTE.
    2020-08-14    Added {cte_is_sampling_configuration()}.
    2021-02-26    Added {cte_shutdown()}.
    2021-03-05    Support CTE preprocessing using {cte_prepare()}.
    2021-03-18    Made {dataPtr} volatile.
    2021-04-29    Added {CTE_2US_7PX1} {CTEType}.
*/

#ifndef RSS2I_INCLUDE_GUARD_CTE_H
#define RSS2I_INCLUDE_GUARD_CTE_H

#include <stdint.h>
#include <stdbool.h>

	/* The following should be consistent with the {cteinfo_hash()} function. */
	typedef enum {
		
		CTE_ANTENNA_PATTERN_INVALID_SWITCH = 0,
		
		CTE_ANTENNA_PATTERN_SWITCH_1TO8,
		
		CTE_ANTENNA_PATTERN_ALWAYS_LAST_SWITCH,
		
		CTE_ANTENNA_PATTERN_INVALID_CONST = 100,
		
		CTE_ANTENNA_PATTERN_CONST_1,
		CTE_ANTENNA_PATTERN_CONST_2,
		CTE_ANTENNA_PATTERN_CONST_3,
		CTE_ANTENNA_PATTERN_CONST_4,
		CTE_ANTENNA_PATTERN_CONST_5,
		CTE_ANTENNA_PATTERN_CONST_6,
		CTE_ANTENNA_PATTERN_CONST_7,
		CTE_ANTENNA_PATTERN_CONST_8,
		
		CTE_ANTENNA_PATTERN_ALWAYS_LAST_CONST,
		
		CTE_ANTENNA_PATTERN_DISABLED = 255
		
	} CTEAntennaPattern;

	/* The following should be consistent with the {cteinfo_hash()} function. */
	typedef enum {
		
		CTE_TYPE_NONE,
		
		CTE_TYPE_2US_7P1,
		CTE_TYPE_2US_7PX1,
		
		CTE_TYPE_ALWAYS_LAST_ITEM,
		
	} CTEType;


	uint8_t cte_num_of_antennas(void);
	bool cte_antenna_pattern_is_valid(CTEAntennaPattern pat, CTEType type);
	bool cte_type_is_valid(CTEType type);
	bool cte_is_sampling_configuration(CTEType type, CTEAntennaPattern ant);
	uint8_t cte_length(CTEType type);
	void cte_init(void);
	void cte_disable(void);
	void cte_shutdown(void);
	void cte_radio_init_tx(CTEType type, CTEAntennaPattern pat);
	void cte_radio_init_rx(CTEType type,
	                       CTEAntennaPattern pat,
	                       volatile uint8_t * dataPtr);
	void cte_unprepare(void);
	bool cte_prepare(CTEType type, CTEAntennaPattern pat);
	
#endif

