/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

/*
  Error reporting system with the possibility to send debug messages via UART.
  Further documentation may be found in the corresponding source file.
   
  Authors
  =======
  
    Copyright (C) 2018-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
  
    2018-09-27    Initial release.
    2018-10-30    Added CONTROL and OTA.
    2018-11-21    Added MEASURE.
    2018-12-21    Added WATCHDOG.
    2019-01-10    Added RADIO.
    2019-01-18    Added CMD.
    2019-03-11    Added BOARD.
    2019-04-11    Added BUFFER.
    2019-04-12    Rewritten. USB/CMD API added.
    2019-06-03    Added board rev. and type related debug vectors.
    2019-06-07    Added measure-invalid-indices related error vectors.
    2019-06-14    Added MLC.
    2019-06-24    Added board-specific battery threshold related debug vectors.
    2019-06-26    ADDED {debug_ok()}.
    2019-08-30    ADDED {debug_get_message()}.
    2019-10-38    ADDED UARTE state machine related error vectors.
    2020-02-08    ADDED {BUFFER_NOTHING_TO_FINISH}.
    2020-03-30    Renamed {DebugSource} to {DebugErrorCase} and rearrange order.
    2020-03-30    Replace {debug_ok()} with {debug_get_num_of_errors()}.
    2020-03-31    Added {DEBUG_MAX_ERROR_CASES}.
    2020-06-08    Added {DEBUG_BOARD_*_NOT_EMULATABLE}.
    2020-06-18    Added {DEBUG_MEASURE_RESULT_BUFFER_ADDR_NULL}.
    2020-07-03    Added {DEBUG_TIMER_LIMIT_EXCEEDED}.
    2020-07-06    Added {DEBUG_SCHEDULER_INVALID_TASK}.
    2020-07-06    Switched to C11's {_Static_assert} for static assertion.
    2020-07-06    Added {DEBUG_CMD_INVALID_WAIT_STATE}.
    2020-07-06    Added support for different {DEBUG_LEVEL}s.
    2020-07-07    Added {DEBUG_MEASURE_LIVELOG_PARALLEL_ACCESS}.
    2020-07-09    Added {DEBUG_POWER_BROWNOUT}.
    2020-07-31    Renamed error case to {DEBUG_LIVELOG_PARALLEL_ACCESS}.
    2020-08-03    Added {DEBUG_CTE_*} error cases.
    2020-08-13    Added {DEBUG_SWITCH_*} error cases.
    2021-02-25    Added {DEBUG_MEASURE_RADIO_PACKET_ADDR_NULL}.
    2021-03-05    Added {DEBUG_CTE_UNPREPARED}.
    2021-03-10    Added {DEBUG_CTE_ADDR_NULL}.
    2021-03-10    Added {TOGGLE} Makefile switch.
    2021-03-18    Added {DEBUG_MEASURE_INVALID_RADIO_MODE}.
    2021-03-25    Added {DEBUG_GPIOTE_OUT_OF_RANGE}.
    2021-03-25    Removed {*_SCHEDULER_INVALID_TASK} and {*_MEMORY_NO_SPACE_*}.
    2021-03-29    Added {DEBUG_RADIO_INVALID_CONST_CONF}.
    2021-04-01    Enhanced the maximal number of error cases significantly.
    2021-04-01    Added {DEBUG_CODER_*} error cases.
    2021-04-02    Added {DEBUG_MEMORY_INCONSISTENT_*} and {DEBUG_MEMORY_LOST_*}.
    2021-04-15    Rename {DEBUG_MEMORY_LOST_RESULTS} to {*_INVALID_STATE}.
    2021-04-15    Added {DEBUG_POSTPROC_*} error cases.
    2021-04-21    Added {DEBUG_CODER_INVALID_COMMAND_SIZE}.
*/

#include <stdbool.h>
#include <stddef.h>

#include <rss2i/led.h>

#include <rss2i/config/debug.h>
#include <rss2i/config/generated.debug.h>


#ifndef DEBUG_LEVEL
	#define DEBUG_MESSAGES_ENABLED false
	#define DEBUG_LEBEL            0
#endif

#if DEBUG_LEVEL != 0 && DEBUG_LEVEL != 1 && DEBUG_LEVEL != 2
	#error "Invalid DEBUG_LEVEL."
#endif

#if DEBUG_LEVEL >= 1
	#define DEBUG_MESSAGES_ENABLED true
#else
	#define DEBUG_MESSAGES_ENABLED false
#endif

#ifndef DEBUG_TOGGLE

	#define DEBUG_MEASUREMENT_FRAME            0
	
#elif DEBUG_TOGGLE >= 0 && DEBUG_TOGGLE <= 3

	#define DEBUG_MEASUREMENT_FRAME            DEBUG_TOGGLE
#else

	#error "Invalid DEBUG_TOGGLE level."

#endif


/* Maximal number of error cases. Note, that this limit is derived from the
 * limited length of OTA messages. Changes should be also applied in the docs.
 */
#define DEBUG_MAX_ERROR_CASES 256
#define DEBUG_MAX_ERROR_CASES_PER_PAGE 63

typedef enum {
/*
  List of at most {DEBUG_MAX_ERROR_CASES} predefined error cases.
 
  This should be consistent with
  ==============================

    - the data within the corresponding source file
    - the API documentation
*/
	
	DEBUG_DEBUG_BUFFER_OVERFLOW,
	DEBUG_DEBUG_PARALLEL_ACCESS,
	DEBUG_DEBUG_UNKNOWN_ERROR_CASE,
	DEBUG_CRYSTAL_OSCILLATOR_FAILED,
	DEBUG_BATTERY_LOW,
	DEBUG_USB_SEND_BUFFER_OVERFLOW,
	DEBUG_USB_RECV_BUFFER_OVERFLOW,
	DEBUG_NO_FREE_TIMERS,
	DEBUG_TIMER_LIMIT_EXCEEDED,
	DEBUG_MEMORY_BUFFER_OVERFLOW,
	DEBUG_MEMORY_INDEX_ERROR,
	DEBUG_MEMORY_CRC_FAILED,
	DEBUG_MEMORY_INCONSISTENT_SIZE,
	DEBUG_MEMORY_INVALID_STATE,
	DEBUG_CONTROL_USB_CRC_FAILED,
	DEBUG_CONTROL_UNKNOWN_CMD,
	DEBUG_CONTROL_INVALID_CMD,
	DEBUG_MEASURE_RESULT_BUFFER_OVERFLOW,
	DEBUG_MEASURE_RESULT_BUFFER_ADDR_NULL,
	DEBUG_MEASURE_RADIO_PACKET_ADDR_NULL,
	DEBUG_MEASURE_DISRUPT_NOTHING_TO_CHECK,
	DEBUG_MEASURE_NO_CONFIG,
	DEBUG_MEASURE_INVALID_FRAME_INDEX,
	DEBUG_MEASURE_INVALID_CYCLE_INDEX,
	DEBUG_MEASURE_INVALID_RADIO_MODE,
	DEBUG_LIVELOG_PARALLEL_ACCESS,
	DEBUG_WATCHDOG_TRIGGERED,
	DEBUG_RADIO_INVALID_PARAMETER,
	DEBUG_RADIO_TX_NOT_READY,
	DEBUG_RADIO_TX_UNPREPARED,
	DEBUG_RADIO_NOT_DISABLED,
	DEBUG_RADIO_UNEXPECTED_STATE,
	DEBUG_RADIO_OTA_MESSAGE_OVERSIZED,
	DEBUG_RADIO_OTA_SEND_BUFFER_OVERFLOW,
	DEBUG_RADIO_OTA_RECV_BUFFER_OVERFLOW,
	DEBUG_RADIO_INVALID_CONST_CONF,
	DEBUG_CMD_UNKNOWN_CMD,
	DEBUG_CMD_INVALID_CMD,
	DEBUG_CMD_ITERABLE_CALLED,
	DEBUG_CMD_INVALID_WAIT_STATE,
	DEBUG_BOARD_NOT_READ_YET,
	DEBUG_BOARD_ID_INVALID,
	DEBUG_BOARD_REV_INVALID,
	DEBUG_BOARD_TYPE_INVALID,
	DEBUG_BOARD_BATTERY_THRESH_INVALID,
	DEBUG_BOARD_TYPE_NOT_EMULATABLE,
	DEBUG_BOARD_REV_NOT_EMULATABLE,
	DEBUG_BUFFER_WIDTH_EXCEEDED,
	DEBUG_BUFFER_NOTHING_TO_FINISH,
	DEBUG_PPI_UNINITIALIZED,
	DEBUG_PPI_OUT_OF_RANGE,
	DEBUG_MLC_TOO_BUSY,
	DEBUG_UARTE_STATE_MACHINE_MISMATCH,
	DEBUG_POWER_BROWNOUT,
	DEBUG_CTE_UNEXPECTED_TYPE,
	DEBUG_CTE_UNEXPECTED_ANTENNA_PATTERN,
	DEBUG_CTE_INVALID_TYPE,
	DEBUG_CTE_INVALID_ANTENNA_PATTERN,
	DEBUG_CTE_UNPREPARED,
	DEBUG_CTE_ADDR_NULL,
	DEBUG_SWITCH_INVALID_BIT_NUMBER,
	DEBUG_SWITCH_INVALID_ANTENNA,
	DEBUG_GPIOTE_OUT_OF_RANGE,
	DEBUG_CODER_INVALID_FORMAT_SPECIFIER,
	DEBUG_CODER_INVALID_FORMAT_BYTE,
	DEBUG_CODER_TRUNCATED_VALID_RSS,
	DEBUG_CODER_INVALID_COMMAND_SIZE,
	DEBUG_POSTPROC_INVALID_STATE,
	DEBUG_POSTPROC_LOW_RSS_AMOUNT,
	DEBUG_POSTPROC_UNEXPECTED_SAMPLE_AMOUNT,
	
	DEBUG_ERROR_CASE_ALWAYS_LAST_ITEM

} DebugErrorCase;

_Static_assert(
	DEBUG_ERROR_CASE_ALWAYS_LAST_ITEM < DEBUG_MAX_ERROR_CASES,
	"{DEBUG_ERROR_CASE} has too many entries."
);


void debug_init(void);
bool debug_message(const char * message);
void debug_report(DebugErrorCase debugErrorCase);
void debug_trigger(void);
bool debug_info(const char * moduleName, const char * submoduleName, const char * message);
uint16_t debug_get_num_of_errors(void);
uint16_t debug_get_global_id(DebugErrorCase src);
uint16_t debug_get_local_id(DebugErrorCase src);
const char * debug_get_message(DebugErrorCase src);
size_t debug_get_num_of_error_cases(void);

/* Automatically fill in the RSS2I module name */
#define debug(X, Y) debug_info(RSS2I_MODULE, X, Y)

