/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

/*
  Measurement control. Further documentation may be found in the corresponding
  source file.
  
  Authors
  =======
    
    Copyright (C) 2018-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
  
    2019-04-12    Initial public release.
    2019-05-02    Added {measure_has_active_frame()}.
    2019-06-06    Added {measure_prepare()}. Changed the starting procedure.
    2019-06-14    Fitted to the changes within the radio module.
    2020-06-19    Store measurement results on buffer using DMA.
    2020-07-31    Outsource livelog functionality into own module.
    2020-08-03    Add CTE configuration.
    2020-08-13    Store RADIO packet meta data.
    2020-08-14    Added CTE sampling check functions (API for postproc module).
    2021-02-25    Added the BLE 2MBit RADIO mode prefix.
    2021-02-26    Removed {measure_ble_2mbit_rx_data_address()}.
    2021-03-17    Added {measure_*_rx_succeeded()} function for all RADIO modes.
    2021-03-18    Added {MeasureBufferSampleType}.
    2021-03-19    Added {measure_is_ieee_250kbit_ready()}.
    2021-03-23    Added {measure_set_radio_mode()}.
*/

#ifndef RSS2I_INCLUDE_GUARD_MEASURE_H
#define RSS2I_INCLUDE_GUARD_MEASURE_H

#include <stdint.h>
#include <stdbool.h>


typedef enum {
	
	MEASURE_BUFFER_SAMPLE_TYPE_NOTHING, /* There WON'T be any raw samples.    */
	MEASURE_BUFFER_SAMPLE_TYPE_IQ,      /* There MAY be IQ samples.           */
	                                    /* (Depends on CTE type and pattern.) */
	
	MEASURE_BUFFER_SAMPLE_TYPE_ALWAYS_LAST_ITEM

} MeasureBufferSampleType;


void measure_init(void);
void measure_trigger(void);


/* Methods for configuration (API for control/cmd) */

bool measure_is_ieee_250kbit_ready(void);

uint8_t measure_set_remote_rss_output_map(uint8_t outputCount,
                                          uint8_t * outputMap);

uint8_t measure_set_remote_rss_input_map(uint8_t * inputMap);

uint8_t measure_set_configuration(uint8_t frameCount,
                                  uint32_t cycleCount,
                                  uint8_t * frameFrequency);

uint8_t measure_set_cte_configuration(uint8_t cteType,
                                      uint8_t * cteConf,
                                      uint8_t * failPtr);

uint8_t measure_set_sample_configuration(uint8_t store,
                                         uint8_t * sample,
                                         uint8_t * failPtr);

uint8_t measure_set_radio_mode(uint8_t mode, uint8_t compatibility);

uint8_t measure_set_sync_map(uint8_t virtualDeviceCount, uint8_t * syncMap);
void measure_set_poor_sync_thresh(uint8_t * poorSyncThresh);
void measure_set_out_of_sync_thresh(uint8_t * outOfSyncThresh);

/* Methods for switching between measurement mode and control mode */
/* (API for control/cmd)                                           */

bool measure_is_configured(void);
bool measure_has_active_frame(void);
bool measure_running(void);
uint32_t measure_prepare(void);
void measure_start_up(void);
void measure_jump_on(void);
bool measure_disrupt(void);
bool measure_disrupted(void);
void measure_stop(void);

/* Methods for checking the (CTE) sampling configuration */
/* (API for postproc module).                            */

bool measure_is_rx_frame(uint8_t frame);
bool measure_is_cte_sampling_frame(uint8_t frame);
bool measure_is_raw_sampling_frame(uint8_t frame);
bool measure_is_sampling_frame(uint8_t frame);
bool measure_any_cte_sampling(void);
bool measure_any_raw_sampling(void);
bool measure_any_sampling(void);
bool measure_frame_in_range(uint8_t frame);

/* Methods for manipulation of the measurement schedule (API for radio) */

void measure_ieee_250kbit_rx_succeeded(uint16_t rAvg, uint8_t rAmnt);
void measure_nrf_2mbit_rx_succeeded(uint16_t rAvg, uint8_t rAmnt);
void measure_nrf_1mbit_rx_succeeded(uint16_t rAvg, uint8_t rAmnt);

void measure_ble_2mbit_rx_succeeded(uint16_t rAvg,
                                    uint8_t rAmnt,
                                    uint8_t iqAmnt);

void measure_ble_1mbit_rx_succeeded(uint16_t rAvg,
                                    uint8_t rAmnt,
                                    uint8_t iqAmnt);

void measure_rx_failed(void);

#endif /* RSS2I_INCLUDE_GUARD_MEASURE_H */

