/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

/*
  API for the RADIO peripheral. Further documentation may be found in the
  corresponding header file.
  
  Authors
  =======
    
    Copyright (C) 2018-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
    
    2019-07-02    Initial public release.
    2020-06-19    Support DMA based measurment radio packet data RAM buffering.
    2020-08-07    Added {CTEType} and {CTEAntennaPattern} arguments.
    2020-08-13    Added {radio_prepare_measure_*()} functions.
    2020-08-13    Enable antenna selection for {radio_init_emit_carrier()}.
    2020-08-17    Add {radio_get_iq_sample_address()}.
    2021-02-25    Remove unused {radio_init_measure_listen()}.
    2021-02-25    Add (BLE) 2MBIT RADIO mode prefixes and unify function names.
    2021-02-26    Add {packetDataAddr} argument within {radio_measure_*_rx()}.
    2021-02-28    Support IEEE 250kBit and BLE/NRF 1/2MBit modes.
    2021-03-17    Introduce {RadioMode}.
    2021-03-18    Amended include guard.
    2021-03-18    Made {packetDataAddr} volatile.
    2021-03-29    Use {SwitchAntenna} for {radio_emit_carrier()} configuration.
    2021-03-29    Added {RadioFrequencyDeviation}.
    2021-03-29    Added {radio_emit_constant()}.
*/

#ifndef RSS2I_INCLUDE_GUARD_RADIO_H
#define RSS2I_INCLUDE_GUARD_RADIO_H

	#include <stddef.h>
	#include <stdint.h>
	#include <stdbool.h>

	#include <rss2i/cte.h>
	#include <rss2i/switch.h>


	typedef enum {
		
		RADIO_MODE_BLE_2MBIT,
		RADIO_MODE_BLE_1MBIT,
		RADIO_MODE_NRF_2MBIT,
		RADIO_MODE_NRF_1MBIT,
		RADIO_MODE_IEEE_250KBIT,
		
		RADIO_MODE_ALWAYS_LAST_ITEM
		
	} RadioMode;
	
	typedef enum {
		
		RADIO_FREQUENCY_DEVIATION_CENTER,
		RADIO_FREQUENCY_DEVIATION_TRANS0,
		RADIO_FREQUENCY_DEVIATION_TRANS1,
		
	} RadioFrequencyDeviation;


	void radio_trigger(void);
	void radio_init(void);


	/* API for control.c/cmd.c. */

	void radio_ota_recv_flush(void);
	uint8_t * radio_ota_recv(void);
	bool radio_ota_send(uint8_t devID, uint8_t * data, uint16_t size);

	uint8_t radio_set_tx_power(uint8_t txPower);
	bool radio_set_tx_power_needs_control_mode(uint8_t txPower);


	/* API for measure.c. */

	void radio_check_disable(void);

	void radio_measure_ieee_250kbit_init_tx(uint8_t frequency,
	                                        bool disrupt,
	                                        uint8_t fI,
	                                        int32_t cI);

	void radio_measure_ieee_250kbit_prepare_tx(uint8_t frequency,
	                                           bool disrupt,
	                                           uint8_t fI,
	                                           int32_t cI);

	void radio_measure_ble_2mbit_init_tx(uint8_t frequency,
	                                     bool disrupt,
	                                     uint8_t fI,
	                                     int32_t cI,
	                                     uint16_t * rRSSV,
	                                     size_t rRSSC,
	                                     CTEType cteType,
	                                     CTEAntennaPattern cteAntPat);

	void radio_measure_ble_2mbit_prepare_tx(uint8_t frequency,
	                                        bool disrupt,
	                                        uint8_t fI,
	                                        int32_t cI,
	                                        uint16_t * rRSSV,
	                                        size_t rRSSC,
	                                        CTEType cteType,
	                                        CTEAntennaPattern cteAntPat);

	void radio_measure_ble_1mbit_init_tx(uint8_t frequency,
	                                     bool disrupt,
	                                     uint8_t fI,
	                                     int32_t cI,
	                                     uint16_t rRSSV,
	                                     CTEType cteType,
	                                     CTEAntennaPattern cteAntPat);

	void radio_measure_ble_1mbit_prepare_tx(uint8_t frequency,
	                                        bool disrupt,
	                                        uint8_t fI,
	                                        int32_t cI,
	                                        uint16_t rRSSV,
	                                        CTEType cteType,
	                                        CTEAntennaPattern cteAntPat);

	void radio_measure_nrf_2mbit_init_tx(uint8_t frequency,
	                                     bool disrupt,
	                                     uint8_t fI,
	                                     int32_t cI,
	                                     uint16_t * rRSSV,
	                                     size_t rRSSC);

	void radio_measure_nrf_2mbit_prepare_tx(uint8_t frequency,
	                                        bool disrupt,
	                                        uint8_t fI,
	                                        int32_t cI,
	                                        uint16_t * rRSSV,
	                                        size_t rRSSC);

	void radio_measure_nrf_1mbit_init_tx(uint8_t frequency,
	                                     bool disrupt,
	                                     uint8_t fI,
	                                     int32_t cI,
	                                     uint16_t rRSSV);

	void radio_measure_nrf_1mbit_prepare_tx(uint8_t frequency,
	                                        bool disrupt,
	                                        uint8_t fI,
	                                        int32_t cI,
	                                        uint16_t rRSSV);

	void radio_measure_start_tx(void);

	void radio_measure_ieee_250kbit_init_rx(uint8_t frequency,
	                                        bool compatibility,
	                                        volatile uint8_t * packetDataAddr);

	void radio_measure_ieee_250kbit_prepare_rx(uint8_t frequency,
	                                           bool compatibility,
	                                           volatile uint8_t * packetDataAddr);

	void radio_measure_ble_2mbit_init_rx(uint8_t frequency,
	                                     bool compatibility,
	                                     volatile uint8_t * packetDataAddr,
	                                     CTEType cteType,
	                                     CTEAntennaPattern cteAntPat);

	void radio_measure_ble_2mbit_prepare_rx(uint8_t frequency,
	                                        bool compatibility,
	                                        volatile uint8_t * packetDataAddr,
	                                        CTEType cteType,
	                                        CTEAntennaPattern cteAntPat);

	void radio_measure_ble_1mbit_init_rx(uint8_t frequency,
	                                     bool compatibility,
	                                     volatile uint8_t * packetDataAddr,
	                                     CTEType cteType,
	                                     CTEAntennaPattern cteAntPat);

	void radio_measure_ble_1mbit_prepare_rx(uint8_t frequency,
	                                        bool compatibility,
	                                        volatile uint8_t * packetDataAddr,
	                                        CTEType cteType,
	                                        CTEAntennaPattern cteAntPat);

	void radio_measure_nrf_2mbit_init_rx(uint8_t frequency,
	                                     bool compatibility,
	                                     volatile uint8_t * packetDataAddr);

	void radio_measure_nrf_2mbit_prepare_rx(uint8_t frequency,
	                                        bool compatibility,
	                                        volatile uint8_t * packetDataAddr);

	void radio_measure_nrf_1mbit_init_rx(uint8_t frequency,
	                                     bool compatibility,
	                                     volatile uint8_t * packetDataAddr);

	void radio_measure_nrf_1mbit_prepare_rx(uint8_t frequency,
	                                        bool compatibility,
	                                        volatile uint8_t * packetDataAddr);

	uint8_t radio_measure_ieee_250kbit_get_frame(volatile uint8_t * p);
	int32_t radio_measure_ieee_250kbit_get_cycle(volatile uint8_t * p);
	uint8_t radio_measure_ble_2mbit_get_frame(volatile uint8_t * p);
	int32_t radio_measure_ble_2mbit_get_cycle(volatile uint8_t * p);
	uint8_t radio_measure_ble_1mbit_get_frame(volatile uint8_t * p);
	int32_t radio_measure_ble_1mbit_get_cycle(volatile uint8_t * p);
	uint8_t radio_measure_nrf_2mbit_get_frame(volatile uint8_t * p);
	int32_t radio_measure_nrf_2mbit_get_cycle(volatile uint8_t * p);
	uint8_t radio_measure_nrf_1mbit_get_frame(volatile uint8_t * p);
	int32_t radio_measure_nrf_1mbit_get_cycle(volatile uint8_t * p);

	uint8_t radio_measure_ieee_250kbit_get_remote_battery_level(
		volatile uint8_t * p
	);
	uint8_t radio_measure_ieee_250kbit_get_remote_memory_load(
		volatile uint8_t * p
	);
	uint8_t radio_measure_ble_2mbit_get_remote_battery_level(
		volatile uint8_t * p
	);
	uint8_t radio_measure_ble_2mbit_get_remote_memory_load(
		volatile uint8_t * p
	);
	uint8_t radio_measure_ble_1mbit_get_remote_battery_level(
		volatile uint8_t * p
	);
	uint8_t radio_measure_ble_1mbit_get_remote_memory_load(
		volatile uint8_t * p
	);
	uint8_t radio_measure_nrf_2mbit_get_remote_battery_level(
		volatile uint8_t * p
	);
	uint8_t radio_measure_nrf_2mbit_get_remote_memory_load(
		volatile uint8_t * p
	);
	uint8_t radio_measure_nrf_1mbit_get_remote_battery_level(
		volatile uint8_t * p
	);
	uint8_t radio_measure_nrf_1mbit_get_remote_memory_load(
		volatile uint8_t * p
	);

	uint8_t radio_measure_ble_2mbit_get_remote_rss_integer_part(
		volatile uint8_t * p,
		uint8_t chan
	);

	uint8_t radio_measure_ble_2mbit_get_remote_rss_fractional_part(
		volatile uint8_t * p,
		uint8_t chan
	);

	uint8_t radio_measure_nrf_2mbit_get_remote_rss_integer_part(
		volatile uint8_t * p,
		uint8_t chan
	);

	uint8_t radio_measure_nrf_2mbit_get_remote_rss_fractional_part(
		volatile uint8_t * p,
		uint8_t chan
	);

	uint8_t radio_measure_ble_1mbit_get_remote_rss_integer_part(
		volatile uint8_t * p
	);
	uint8_t radio_measure_ble_1mbit_get_remote_rss_fractional_part(
		volatile uint8_t * p
	);
	uint8_t radio_measure_nrf_1mbit_get_remote_rss_integer_part(
		volatile uint8_t * p
	);
	uint8_t radio_measure_nrf_1mbit_get_remote_rss_fractional_part(
		volatile uint8_t * p
	);

	uint8_t * radio_measure_ble_2mbit_get_iq_sample_address(uint8_t * p);
	uint8_t * radio_measure_ble_1mbit_get_iq_sample_address(uint8_t * p);


	void radio_emit_carrier(uint8_t frequency, SwitchAntenna ant);
	
	void radio_emit_constant(
		uint8_t frequency,
		RadioMode mode,
		RadioFrequencyDeviation dev,
		SwitchAntenna ant
	);


	void radio_control_init_recv(void);
	void radio_control_init_send(void);

#endif /* RSS2I_INCLUDE_GUARD_RADIO_H */

