/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

/*
  Generates C code that contains a bytewise 32-bit-CRC lookup table.
  
  Authors
  =======
    
    Copyright (C) 2018-2019  Gerrit Maus (funk@maus.xyz)
    Copyright (C) 2018-2019  Mustafa Gemci (mustafagemci.ee@gmail.com)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
  
    2019-04-16    Initial public release.
  
  Description
  ===========
   
    A very good overview about CRC generation using lookup tables is given on
    one of Bastian Molkenthin's webpages
    (http://www.sunshine2k.de/articles/coding/crc/understanding_crc.html).
*/

#include <stdio.h>
#include <stdint.h>
#include <inttypes.h>

#include <rss2i/config/crc.h>


uint32_t shift_then_xor(uint32_t crc, uint8_t times){
/*
  Key step within CRC computation.
*/
	size_t i;
	
	for(i = 0; i < times; ++i){
		
		if(crc & (1u << 31)){
			
			crc <<= 1;
			crc ^= CRC_POLYNOMIAL;
		}
		else{
			
			crc <<= 1;
		}
	}
	
	return crc;
}

int main(void){
/*
  CRC lookup table and magick number generation.
*/
	
	uint16_t byte;
	
	printf("/*\n\n  This file has been generated by lookup/crc which is part of the RSS2I project.\n\n*/\n\n");
	printf("#include <stdint.h>\n\n");
	printf("const uint32_t crcMagicNumber = 0x%8.8" PRIx32 ";\n\n", shift_then_xor(0xffffffff, 32) ^ CRC_FINAL_XOR);
	printf("const uint32_t crcTable[] = {\n\n");
	
	for(byte = 0; byte < 256; ++byte){
		
		printf("\t0x%8.8" PRIx32 ",\n", shift_then_xor((uint32_t) byte << 24, 8));
	}
	
	printf("};\n\n");
	
	return 0;
}


