/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

/*
  Generates C code that contains LED fading lookup tables for three colors.
  
  Authors
  =======
    
    Copyright (C) 2018-2019  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
  
    2018-07-24    Initial release.
  
  Description
  ===========
   
    The human eye's brightness sensitivity depends non-linearly on the actual
    brightness. The generated lookup tables try to correct this non-linearity.
*/

#include <stdio.h>
#include <stdint.h>
#include <inttypes.h>
#include <math.h>

#include <rss2i/config/led.h>

uint16_t value(uint8_t index, uint8_t risingSteps, uint16_t topValue){
/*
  Calculates linear visual brightness values from 0 to {topValue} and back to
  0 again with {2 * risingSteps} steps.
  
  When {index} is zero it will return the first nonzero value within the
  calculated sequence. When {index} equals {risingSteps - 1} the top value
  gets returned. For the last {index} ({max - 1}) it will return zero.
*/
	
	if(index < risingSteps){
	
		return pow(topValue + 1, (index + 1) / (double) risingSteps) - 1;
	}
	else{
		
		return pow(topValue + 1, (2 * risingSteps - index - 1) / (double) risingSteps) - 1;
	}
}

int main(void){
	
	uint16_t index;
	uint8_t color;
	
	uint16_t topValue[3] = {PWM_TOP_GREEN, PWM_TOP_YELLOW, PWM_TOP_RED};
	char * colorName[3] = {"green", "yellow", "red"};
	
	printf("/*\n\n  This file has been generated by lookup/led which is part of the RSS2I project.\n\n*/\n\n");
	printf("#include <stdint.h>\n\n");
	
	for(color = 0; color < 3; ++color){
	
		printf("uint16_t pwm_seq_%s[] = {", colorName[color]);
	
		for(index = 0; index < 2 * PWM_CYCLES; ++index){
		
			printf("%" PRIu16, value(index, PWM_CYCLES, topValue[color]));
		
			if(index != 2 * PWM_CYCLES - 1){
			
				printf(", ");
			}
		}
	
		printf("};\n");
	}
	
	return 0;
}


