/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

/*
  Run a full memory test on an RSS2I device.
  
  Authors
  =======
  
    Copyright (C) 2018-2020  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
  
    2019-06-25    Initial public release.
    2019-07-12    Added triggering of the timer module.
    2020-07-03    Introduced the scheduler.
    2020-07-07    Added {scheduler_init()}.
    2020-07-10    Disabled nRF iCache technology explicitly.
  
  Description
  ===========
  
    Runs a full memory test on this RSS2I device after power up. This may take
    some time. The serial debug output gives useful status and progress
    information.
*/


#include <rss2i/clock.h>
#include <rss2i/measure.h>
#include <rss2i/battery.h>
#include <rss2i/led.h>
#include <rss2i/debug.h>
#include <rss2i/control.h>
#include <rss2i/icache.h>
#include <rss2i/power.h>
#include <rss2i/watchdog.h>
#include <rss2i/board.h>
#include <rss2i/mlc.h>
#include <rss2i/timer.h>
#include <rss2i/scheduler.h>
#include <rss2i/memory.h>


int main(){
	
	bool testRunning = true;
	MemoryTestStatus testStatus;
	
	led_init();          /* Controls the onboard LEDs.                        */
	timer_init();        /* Provides some timers.                             */
	debug_init();        /* Debug and reporting system.                       */
	board_init();        /* Holds board-specific parameters.                  */
	power_init();        /* Enables the internal DC/DC regulator.             */
	clock_init();        /* Configures the High Frequency Clock (HFCLK).      */
	battery_init();      /* Monitors the battery voltage.                     */
	mlc_init();          /* Monitors the main loop frequency.                 */
	control_init();      /* Handles incoming commands.                        */
	measure_init();      /* Enters control mode.                              */
	scheduler_init();    /* Schedules all periodic tasks.                     */
/*	icache_init();     *//* Initializes the nRF iCache technology.            */
	
	watchdog_init();     /* Enables the watchdog.                             */
	
	memory_init();       /* API for the external memory.                      */
	
	
	led_set(LED_YELLOW, LED_ON);
	led_set(LED_GREEN, LED_OFF);
	
	while(1){
		
		if(testRunning){
			
			/* Trigger the test routine until it returns a status. */
			if((testStatus = memory_full_test())){
				
				testRunning = false;
				
				if(testStatus == MEMORY_TEST_SUCCEEDED){
					
					led_set(LED_GREEN, LED_ON);
				}
				else{
					
					led_set(LED_RED, LED_ON);
				}
			}
		}
		
		scheduler();
	}
	
	return 0;
}

