/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

/*
  Wrapper for the coder module that is used by librss2i. This library gets
  shipped along with the RSS2I python package.
  
  Authors
  =======
    
    Copyright (C) 2021-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This program is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
    
  Changelog
  =========
    
    2021-04-21    Initial release.
*/

#include <stdint.h>
#include <stddef.h>
#include <stdbool.h>

#include <rss2i/coder.h>
#include <rss2i/debug.h>


bool rss2i_coder_errors(void){
/*
  Whether there occured errors during any call of the functions below.
*/
	
	if(debug_get_global_id(DEBUG_CODER_INVALID_FORMAT_SPECIFIER)
	|| debug_get_global_id(DEBUG_CODER_INVALID_FORMAT_BYTE)
	|| debug_get_global_id(DEBUG_CODER_TRUNCATED_VALID_RSS)
	|| debug_get_global_id(DEBUG_CODER_INVALID_COMMAND_SIZE)){
		
		return true;
	}
	else{
		
		return false;
	}
}

int rss2i_coder_error(void){
/*
  Returns the first ID out of the reported error cases. Returns {0} if
  {rss2i_coder_errors()} returns {false}.
*/
	
	if(debug_get_global_id(DEBUG_CODER_INVALID_FORMAT_SPECIFIER)){
		
		return DEBUG_CODER_INVALID_FORMAT_SPECIFIER;
	}
	else if(debug_get_global_id(DEBUG_CODER_INVALID_FORMAT_BYTE)){
		
		return DEBUG_CODER_INVALID_FORMAT_BYTE;
	}
	else if(debug_get_global_id(DEBUG_CODER_TRUNCATED_VALID_RSS)){
		
		return DEBUG_CODER_TRUNCATED_VALID_RSS;
	}
	else if(debug_get_global_id(DEBUG_CODER_INVALID_COMMAND_SIZE)){
		
		return DEBUG_CODER_INVALID_COMMAND_SIZE;
	}
	else{
		
		return 0;
	}
}

bool rss2i_coder_explicit_format_byte_present(int fmtSpec){
/*
  Whether the specified {fmtSpec} has to be explicitly present before the actual
  measurement result ({true}) or it is part (the first byte) of the actual
  measurement result ({false}).
*/
	
	return coder_explicit_format_byte_present(fmtSpec);
}

bool rss2i_coder_is_result(int fmtByte){
/*
  Whether the given {fmtByte} announces a measurement result or a command. A
  valid result {fmtByte} will always evaluate to {true}. A valid command
  {fmtByte} will always evaluate to {false}. However, the validity of the
  specified {fmtByte} has to be further checked in both cases.
*/
	
	return coder_is_result(fmtByte);
}

uint8_t rss2i_coder_command_size(int fmtByte, uint8_t maxSize){
/*
  Number of bytes of a command with the specified {fmtByte}. The number of
  bytes left on the current flash page (counted from the byte after the given
  {fmtByte}) is denoted by {maxSize}.
*/
	
	return coder_command_size(fmtByte, maxSize);
}

uint8_t rss2i_coder_result_size(int fmtSpec){
/*
  Number of bytes of a measurement result with the specified {fmtSpec}.
*/
	
	return coder_result_size(fmtSpec);
}

bool rss2i_coder_result_increments_expected_index(int fmtByte){
/*
  Whether a result with the specified {fmtByte} increments the expected index,
  i.e. the expected index of the proceeding results, or not.
*/
	
	return coder_result_increments_expected_index(fmtByte);
}

int rss2i_coder_postproc_format(int fmtSpec){
/*
  Given the {CoderFormatSpecifier} {fmtSpec} this function returns the
  corresponding {PostprocFormat}. An invalid format results in the return of
  {POSTPROC_FORMAT_INVALID}.
*/
	
	return coder_postproc_format(fmtSpec);
}

int rss2i_coder_format_specifier_from_byte(int fmtByte){
/*
  Given a {fmtByte} this function returns the corresponding format specifier.
  Returns {CODER_FORMAT_SPECIFIER_NONE} if the specified {fmtByte} is invalid.
*/
	
	return coder_format_specifier_from_byte(fmtByte);
}

uint8_t rss2i_coder_sheet_index_from_byte(int fmtByte){
/*
  Given the {fmtByte} this function returns the corresponding sheet index.
*/
	
	return coder_sheet_index_from_byte(fmtByte);
}

unsigned long int rss2i_coder_index(uint8_t * indx){
/*
  Calculates the index from the four bytes pointed to by {indx}.
*/
	
	return coder_index(indx);
}

unsigned long int rss2i_coder_next_index(
	unsigned long int currentIndex,
	int fmtByte,
	uint8_t * cmd
){
/*
  Returns the expected index of the measurement results that are stored after
  the command with the specified {fmtByte}. The (probably empty) content of the
  command should be available at the memory pointed to by {cmd}. The
  {currentIndex} denotes the index of the results prior to the command with
  the given {fmtByte}.
*/
	
	return coder_next_index(currentIndex, fmtByte, cmd);
}

