#
#  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
#

#
#  RSS2I project python package.
#
#  Authors
#  =======
#
#    Copyright (C) 2018-2021  Gerrit Maus (funk@maus.xyz)
#    Copyright (C) 2019-2019  Knut Niemann (post@knut-niemann.de)
#
#  License
#  =======
#
#    This program is free software: you can redistribute it and/or modify it
#    under the terms of the GNU General Public License as published by the Free
#    Software Foundation, either version 3 of the License, or (at your option)
#    any later version.
#
#    This program is distributed in the hope that it will be useful, but
#    WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
#    or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
#    more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program. If not, see <https://www.gnu.org/licenses/>.
#
#  Changelog
#  =========
#
#    2019-06-28    Initial public release.
#    2019-07-11    Added support for precompiled shared libraries.
#    2019-08-29    Introduced librss2i (instead of multiple shared libs).
#    2019-08-30    Added {version}, {versionString} and {debug_message()}.
#    2019-11-06    Fixed serial connection timeout handling.
#    2020-01-23    Do not modify mutable objects from inside the query method.
#    2020-03-12    Added the {laxTimeout} flag to the {rss2ipy.Device} class.
#    2020-03-31    Change terminology {DebugSource} -> {DebugErrorCase}.
#    2020-06-19    Increased lax {interByteTimeout} to 5 seconds.
#    2020-07-13    Increased lax {timeout} to 90 seconds.
#    2020-07-13    Remeber {laxTimeout} setting when changing the baudrate.
#    2020-08-24    Add those {cte_*} and {_cte_*} functions.
#    2020-08-28    Made the CTE compatibility check deactivateable.
#    2020-08-28    Fixed standard pattern if no antenna array is available.
#    2021-04-01    Add {debug_get_num_of_error_cases*} functions.
#    2021-04-23    Replace {page_decode} by {memory_decode}.
#    2021-04-29    Support {CTE_TYPE_2US_7PX1} {CTEType}.
#    2021-05-01    Fix raise of `DecodeError`.
#    2021-05-04    Fit timeout to larger disrupt times in sample storage mode.
#    2021-06-15    Add support for `T2US7PX1` CTE type.
#
#  Description
#  ===========
#    
#    This package implements the rss2ipy.Device class (see below) as well as
#    some other RSS2I device related functionality:
#    
#      - Decode flash memory contents.
#      - Decode debug message IDs into debug messages.
#      - Version information.
#      - Check board parameters for CTE compatibility
#


import serial
import ctypes
import pkg_resources
import time
import sys
import platform
import numpy as np
__all__ = ['config']




#-------------------------------------------------------------------------------
#  Possible Exceptions.
#-------------------------------------------------------------------------------


class SerialError(Exception):
# Cannot establish serial connection.
	pass
class SerialBusy(Exception):
# Serial device seems busy. Already in use elsewhere?
	pass
class SerialTimeout(Exception):
# Serial device does not respond as expected.
	pass
class LivelogTimeout(Exception):
# Could not read from livelog. Are there any packets?
	pass
class LibError(Exception):
# Cannot load shared library. Platform mismatch? File exists?
	pass
class CRCError(Exception):
# Communication error.
	pass
class BaudrateChangeError(Exception):
# Baudrate change unsuccessfull.
	pass
class RangeError(Exception):
# The given query contained non-byte values.
	pass
class IndexError(Exception):
# The requested index is invalid.
	pass
class DecodeError(Exception):
# The specified flash page has illegal contents.
	pass
class MissingPages(Exception):
# The measurement results are incomplete due to missing pages.
	pass




#-------------------------------------------------------------------------------
#  Get the shared library working the tryAndError way
#-------------------------------------------------------------------------------


libLoaded = False;
for libName in ['', '.windows.i386', '.windows.amd64', '.linux.amd64', '.linux.i386']:
	
	try:
		libcrcPath = pkg_resources.resource_filename(__name__, 'librss2i' + libName + '.so')
		lib = ctypes.CDLL(libcrcPath);
		libLoaded = True;
		break;
		
	except(OSError):
		pass;

if not libLoaded:
	raise LibError;



lib.rss2i_crc.argtypes = (ctypes.POINTER(ctypes.c_ubyte), ctypes.c_uint);
lib.rss2i_crc.restype = ctypes.c_ulong;

lib.rss2i_crc_check.argtype = ctypes.c_ulong;
lib.rss2i_crc_check.restype = ctypes.c_bool;


lib.rss2i_version_high_byte.argtype = ();
lib.rss2i_version_high_byte.restype = ctypes.c_ubyte;

lib.rss2i_version_mid_byte.argtype = ();
lib.rss2i_version_mid_byte.restype = ctypes.c_ubyte;

lib.rss2i_version_low_byte.argtype = ();
lib.rss2i_version_low_byte.restype = ctypes.c_ubyte;


lib.rss2i_debug_get_message.argtype = (ctypes.c_int);
lib.rss2i_debug_get_message.restype = ctypes.c_char_p;

lib.rss2i_debug_get_num_of_error_cases.argtype = ();
lib.rss2i_debug_get_num_of_error_cases.restype = ctypes.c_ulong;

lib.rss2i_debug_get_num_of_error_cases_per_page.argtype = ();
lib.rss2i_debug_get_num_of_error_cases_per_page.restype = ctypes.c_ubyte;


lib.rss2i_coder_errors.argtype = ();
lib.rss2i_coder_errors.restype = ctypes.c_bool;

lib.rss2i_coder_error.argtype = ();
lib.rss2i_coder_error.restype = ctypes.c_int;

lib.rss2i_coder_explicit_format_byte_present.argtype = (ctypes.c_int);
lib.rss2i_coder_explicit_format_byte_present.restype = ctypes.c_bool;

lib.rss2i_coder_is_result.argtype = (ctypes.c_int);
lib.rss2i_coder_is_result.restype = ctypes.c_bool;

lib.rss2i_coder_command_size.argtype = (ctypes.c_int, ctypes.c_ubyte);
lib.rss2i_coder_command_size.restype = ctypes.c_ubyte;

lib.rss2i_coder_result_size.argtype = (ctypes.c_int);
lib.rss2i_coder_result_size.restype = ctypes.c_ubyte;

lib.rss2i_coder_result_increments_expected_index.argtype = (ctypes.c_int);
lib.rss2i_coder_result_increments_expected_index.restype = ctypes.c_bool;

lib.rss2i_coder_format_specifier_from_byte.argtype = (ctypes.c_int);
lib.rss2i_coder_format_specifier_from_byte.restype = ctypes.c_int;

lib.rss2i_coder_sheet_index_from_byte.argtype = (ctypes.c_int);
lib.rss2i_coder_sheet_index_from_byte.restype = ctypes.c_ubyte;

lib.rss2i_coder_postproc_format.argtype = (ctypes.c_int);
lib.rss2i_coder_postproc_format.restype = ctypes.c_int;

lib.rss2i_coder_index.argtype = (ctypes.POINTER(ctypes.c_ubyte));
lib.rss2i_coder_index.restype = ctypes.c_ulong;

lib.rss2i_coder_next_index.argtype = (ctypes.c_ulong, ctypes.c_int, ctypes.POINTER(ctypes.c_ubyte));
lib.rss2i_coder_next_index.restype = ctypes.c_ulong;

lib.rss2i_cteinfo_type_is_valid.argtype = (ctypes.c_int);
lib.rss2i_cteinfo_type_is_valid.restype = ctypes.c_bool;

lib.rss2i_cteinfo_antenna_pattern_is_valid.argtype = (ctypes.c_ubyte, ctypes.c_ubyte, ctypes.c_int, ctypes.c_int);
lib.rss2i_cteinfo_antenna_pattern_is_valid.restype = ctypes.c_bool;

lib.rss2i_cteinfo_num_of_antennas.argtype = (ctypes.c_ubyte, ctypes.c_ubyte);
lib.rss2i_cteinfo_num_of_antennas.restype = ctypes.c_ubyte;



#-------------------------------------------------------------------------------
#  Version information.
#-------------------------------------------------------------------------------


version = [lib.rss2i_version_high_byte(), lib.rss2i_version_mid_byte(), lib.rss2i_version_low_byte()];
versionString = '{}.{}.{}'.format(*version);



#-------------------------------------------------------------------------------
#  Device class implementation.
#-------------------------------------------------------------------------------


class Device:
#
#  This class is an abstraction layer for the serial connection to an RSS2I
#  device. It provides methods for basic configuration and live logging. Beside
#  that, it implements an interface for transfering and receiving commands to an
#  RSS2I device. This API may be further used by other configuration utilities.
#
	
	def __init__(self, portName, baudrate=0, laxTimeout=False):
	#
	#  Creates (i.e. establishes a connection to) an RSS2I device.
	#
	#  Argument List
	#  =============
	#
	#    {portName}            Serial port ('/dev/ttyUSB0', 'COM0', ...).
	#    {baudrate}            Baudrate. See {self.open} for more informations.
	#                          Default is {0}.
	#    {laxTimeout}          Apply tolerant timeout settings e.g. for
	#                          downloading a lot of pages via a bad wireless
	#                          (i.e. OTA) connection. Defaults to {False}.
	#  
	#  May throw
	#  =========
	#  
	#    rss2ipy.LibError
	#    rss2ipy.SerialError
	#    
		
		self.portName = portName;
		self.livelogEnabled = False;
		self.laxTimeout = laxTimeout;
		self.open(baudrate, laxTimeout);
		self.lib = lib;
	
	def open(self, baudrate, laxTimeout):
	#
	#  Establishes a connection to the RSS2I device.
	#
	#  Argument List
	#  =============
	#
	#    {baudrate}            Baudrate according to the following table:
	#                          
	#                              0     115200 Baud/s
	#                              1     230400 Baud/s
	#                              2     250000 Baud/s
	#                              3     460800 Baud/s
	#                              4     921600 Baud/s
	#                              5    1000000 Baud/s
	#                          
	#                          Every other {baudrate} is treated a {0}.
	#    {laxTimeout}          Apply tolerant timeout settings e.g. for
	#                          downloading a lot of pages via a bad wireless
	#                          (i.e. OTA) connection.
	#  
	#  May throw
	#  =========
	#  
	#    rss2ipy.SerialError
	#    
	
		try:
		
			baudrateMap = [115200, 230400, 250000, 460800, 921600, 1000000];
			baudrate = int(baudrate);
			
			if baudrate < 0 or baudrate > 5:
				baudrate = 0;
			
			# The {inter_byte_timeout} setting needs to be fitted to the
			# {MEASURE_MIN_DISRUPT_FRAMES} preprocessor define within
			# 'include/rss2i/config/measure.h'.
			if laxTimeout:
				timeout = 90;
				interByteTimeout = 20;
			else:
				timeout = 3;
				interByteTimeout = 3;
			
			self.serialPort = serial.Serial(self.portName, baudrateMap[baudrate], timeout=timeout, inter_byte_timeout=interByteTimeout);
			self.baudrate = baudrate;
	
		except (FileNotFoundError, serial.serialutil.SerialException, OSError):
			
			raise SerialError;
		
	def crc(self, data):
	#
	#  Calculates the CRC sum over the given {data} (LSB first).
	#
	#  Return value
	#  ============
	#  
	#    Returns a list of four byte values representing the CRC sum over the
	#    given {data}. Those bytes are MSB first.
	#    
		
		crcSum = self.lib.rss2i_crc((ctypes.c_ubyte * len(data))(*data), len(data));
		crcSum = crcSum.to_bytes(4, byteorder='little');
		crcSum = list(crcSum);
		
		return crcSum[::-1];
	
	def crc_check(self, datas, iterations = 1):
	#
	#  Checks if the data and its appended CRC sum inside {datas} correspond to
	#  each other. If {iterations} is greater than one, it is assumed that
	#  {datas} consists of {iterations} euqally sized individual data sets that
	#  are stored consecutively one after the other.
	#
	#  Argument List
	#  =============
	#
	#    {datas}         Data and its (hopefully) corresponding CRC sum to be
	#                    checked.
	#    {iterations}    Number of consecutively stored data sets inside
	#                    {datas}. The length of the individual data sets is
	#                    assumed to be always the same and is hence calculated
	#                    automatically.
	#  
	#  Return value
	#  ============
	#  
	#    Returns {True} if all CRC sums correspond to their data. Returns
	#    {False} otherwise.
	#    
		
		status = True;
		length = int(len(datas) / iterations);
		
		for i in range(0, iterations):
			
			data = datas[i * length : (i + 1) * length];
			
			crcSum = self.lib.rss2i_crc((ctypes.c_ubyte * length)(*data), length);
			
			if not self.lib.rss2i_crc_check(crcSum):
				status = False;
				break;
			
		return status;
	
	def set_baudrate(self, baudrate, laxTimeout=None):
	#
	#  Changes the communication's baudrate. This implies querying the connected
	#  device as well as changing, i.e. re-establishing, the serial connection.
	#
	#  Argument List
	#  =============
	#
	#    {baudrate}            Desired baudrate according to the following table:
	#                    
	#                              0     115200 Baud/s
	#                              1     230400 Baud/s
	#                              2     250000 Baud/s
	#                              3     460800 Baud/s
	#                              4     921600 Baud/s
	#                              5    1000000 Baud/s
	#                
	#                          Every other {baudrate} is treated a {0}.
	#    {laxTimeout}          Apply tolerant timeout settings e.g. for
	#                          downloading a lot of pages via a bad wireless
	#                          (i.e. OTA) connection. Defaults to the same value
	#                          that was used during initial {Device} creation.
	#  
	#  May throw
	#  =========
	#  
	#    rss2ipy.SerialTimeout
	#    rss2ipy.SerialError
	#    rss2ipy.BaudrateChangeError
	#    rss2ipy.CRCError
	#    
		
		if laxTimeout is None:
			laxTimeout = self.laxTimeout;
		
		if self.livelogEnabled:
			self.stop_livelog();
		
		# Query the device to change the baudrate.
		
		instruction = [0xff, 0x01, baudrate] + self.crc([0x01, baudrate]) + [0xff] * 4;
				
		self.serialPort.write(instruction);
		self.serialPort.flush();
		
		# Re-establishing the serial connection.
		
		self.serialPort.close();
		self.open(baudrate, laxTimeout);
		self.serialPort.reset_input_buffer();
		
		# Check the device's answer.
		
		response = self.serialPort.read(5);
		
		if len(response) < 5:
			raise SerialTimeout;
		if not self.crc_check(response):
			raise CRCError;
		if response[0] != baudrate:
			self.baudrate = response[0];
			raise BaudrateChangeError;
		
	def query(self, instruction, responseLength, iterations = 0):
	#
	#  Queries the given {instruction} and waits for a response with the given
	#  {responseLength}. CRC gets applied automatically.
	#
	#  If {iterations} doesn't equal zero, the instruction is assumed to be a so
	#  called iterable command. See the docs for more information. The iteration
	#  byte gets appended automatically in such cases and _must_not_ be part of
	#  the specified {instruction}.
	#
	#  Argument List
	#  =============
	#
	#    {instruction}       Command to be queried _without_ a CRC sum,
	#                        _without_ start or fill up bytes ({0xff}) and
	#                        _without_ a possible iteration byte.
	#    {responseLength}    Expected response length _excluding_ the CRC sum.
	#                        In case of an iterable command, this should be the
	#                        length of a single iteration.
	#    {iterations}        Number of iterations. This should differ from zero
	#                        only, if the specified command is an iterable one.
	#  
	#
	#  May throw
	#  =========
	#
	#    rss2ipy.RangeError
	#    rss2ipy.SerialBusy
	#    rss2ipy.SerialTimeout
	#    rss2ipy.CRCError
	#
		
		if self.livelogEnabled:
			self.stop_livelog();
		
		if iterations > 0:
			# Lists are mutable objects! ;)
			# instruction += [iterations];
			instruction = instruction + [iterations];
			iterable = True;
		else:
			iterations = 1;
			iterable = False;
		
		instruction = [0xff] + instruction + self.crc(instruction) + [0xff] * 4;
				
		self.serialPort.reset_input_buffer();
		
		try:
			self.serialPort.write(instruction);
		except (ValueError):
			raise RangeError;
		
		self.serialPort.flush();
		
		if responseLength > 0:
			try:
				response = self.serialPort.read((responseLength + 4) * iterations);
			except(serial.serialutil.SerialException):
				raise SerialBusy;
		else:
			response = [];
		
		if responseLength > 0 and len(response) < (responseLength + 4) * iterations:
			raise SerialTimeout
		
		if response and not self.crc_check(response, iterations):
			raise CRCError;
		
		if responseLength > 0 and iterable:
		
			processedResponse = [];
			for i in range(0, iterations):
				processedResponse += [response[(responseLength + 4) * i:(responseLength + 4) * i + responseLength]];
		else:
			processedResponse = response[0:responseLength]
		
		return processedResponse;
	
	def start_livelog(self):
	#
	#  Starts the livelog. Extra implementation due to special timing
	#  requierements (see {self.stop_livelog()}).
	#  
	#  May throw
	#  =========
	#
	#    rss2ipy.RangeError
	#    rss2ipy.SerialBusy
	#    rss2ipy.SerialTimeout
	#    rss2ipy.CRCError
	#
		
		self.query([0x09], 1);
		self.livelogEnabled = True;
	
	def stop_livelog(self):
	#
	#  Stops the livelog. Extra implementation due to special timing
	#  requierements.
	#  
	#  May throw
	#  =========
	#
	#    rss2ipy.SerialTimeout
	#    rss2ipy.CRCError
	#
		
		instruction = [0xff, 0x0a] + self.crc([0x0a]) + [0xff] * 4;
				
		self.serialPort.write(instruction);
		self.serialPort.flush();
		
		# Wait until the livelog has been disabled.
		time.sleep(0.15);
		
		self.serialPort.reset_input_buffer();
		
		response = self.serialPort.read(5);
		
		if len(response) < 5:
			raise SerialTimeout;
		if not self.crc_check(response):
			raise CRCError;
		
		self.livelogEnabled = False;
	
	def read_livelog(self, numOfBytes):
	#
	#  Reads {numOfBytes} bytes from the livelog. The livelog gets enabled
	#  automatically, if necessary.
	#
	#  May throw
	#  =========
	#
	#    rss2ipy.RangeError
	#    rss2ipy.SerialBusy
	#    rss2ipy.SerialTimeout
	#    rss2ipy.CRCError
	#    rss2ipy.LivelogTimeout
	#
		
		if not self.livelogEnabled:
			self.start_livelog();
		
		response = self.serialPort.read(numOfBytes);
		
		if len(response) < numOfBytes:
			raise LivelogTimeout;
		
		return response;
		
	def close(self):
	#
	#  Resets the baudrate (if changed), Stops livelog (if started) and
	#  disconnects the serial port.
	#
	#  May throw
	#  =========
	#  
	#    rss2ipy.SerialTimeout
	#    rss2ipy.SerialError
	#    rss2ipy.BaudrateChangeError
	#    rss2ipy.CRCError
	#    
		
		if self.livelogEnabled:
			self.stop_livelog();
		
		if self.baudrate != 0:
			self.set_baudrate(0);
		
		self.serialPort.close();




#-------------------------------------------------------------------------------
#  Other functionality.
#-------------------------------------------------------------------------------

def rss_decode(result):
#
#  Decodes the two bytes given within `result` to a logarithmic RSS value with
#  dBm as unit.
#
#  Return Value
#  ============
#    
#    A float.
#
	
	return -(result[0] + result[1] / 128)

def sample_decode(sheet, fmt):
#
#  Decodes the given samples with the specified format `fmt`. The list `sheet`
#  should be a list of four sheets. Each sheet shall be a list with two entries.
#  The first entry specifies the sheet index. The second entry contains the
#  actual sheet data of, for reference, size `n`. All four sheets must have the
#  same size `n`.
#  
#  Return Value
#  ============
#  
#    A complex numpy array of `n` samples.
#
#  May Raise
#  =========
#    
#    rss2ipy.DecoderError
#    rss2ipy.MissingPages
#
	try:
		if any([len(sheet[i][1]) != len(sheet[i + 1][1]) for i in [0, 1, 2]]):
			raise DecodeError("All sheets must have the same size!")
	except DecodeError:
		raise
	except:
		raise MissingPages("There were less than four sheets per result.")
	
	numOfSamples = len(sheet[0][1])
	
	result = np.empty((numOfSamples), dtype=complex) * np.nan
	
	# Map the sheetIndices.
	sheetMap = [False] * 4
	for i in range(4):
		sheetMap[sheet[i][0]] = i
	
	if any([sM is False for sM in sheetMap]):
		raise MissingPages("At least one sheet occured twice.")
	
	
	if fmt == 0 or fmt == 1:
		
		# IQ samples
		
		for i in range(numOfSamples):
		
			bytesI = [sheet[sheetMap[0]][1][i], sheet[sheetMap[1]][1][i]]
			bytesQ = [sheet[sheetMap[2]][1][i], sheet[sheetMap[3]][1][i]]
			dataI = int.from_bytes(bytesI, byteorder="little", signed=True)
			dataQ = int.from_bytes(bytesQ, byteorder="little", signed=True)
			result[i] = dataQ + 1j * dataI
		
	else:
		
		# Unknown format
		
		raise DecodeError("Unrecognized format {}".format(fmt))
	
	return result

def memory_decode(memory, frameCount):
#
#  Decodes the given flash `memory` (a list of subsequent flash pages each
#  consisting of 252 bytes measurement data) assuming that the corresponding
#  measurement configuration consited of {frameCount} frames per cycle.
#  
#  Argument List
#  =============
#    
#    `memory`        List of subsequent flash pages. Each page is a list
#                    (or of compatible type e.g. byte array) that holds 252
#                    bytes of measurement data.
#    `frameCount`    Number of frames in the measurement configuration that was
#                    used during measurement.
#
#  Return Value
#  ============
#    
#    `rss`           A numpy array that holds the RSS measured during the
#                    individual frames for all measured cycle indices.
#                     
#                     [
#                      [cycle_0, frame_0_0, frame_0_1, ..., frame_0_n],
#                      ...,
#                      [cycle_m, frame_m_0, frame_m_1, ..., frame_m_n]
#                     ]
#                    
#                    where `m+1` denotes the number of cycles and `n+1` denotes
#                    the number of frames per cycle.
#    `samples`       A list of raw samples measured during the individual
#                    frames. May be `None` if there wasn't any raw sampling
#                    during measurement. Otherwise, each entry holds all raw
#                    samples measured during a specific frame. The first list
#                    entry holds the raw samples for frame 0, the last entry
#                    holds the raw samples for the last frame. Each entry is a
#                    numpy array of the form
#                     
#                     [
#                      [cycle_0, sample_0_0, sample_0_1, ..., sample_0_l],
#                      ...,
#                      [cycle_m, sample_m_0, sample_m_1, ..., sample_m_l]
#                     ]
#                    
#                    where `l+1` denotes the number of raw samples for the
#                    specific frame and `m+1` denotes the number of measured
#                    cycles. Note, that `l` may be different for different
#                    frames. Further note, that the returned samples may be
#                    of complex dtype.
#
#  May Throw
#  =========
#  
#    rss2ipy.DecodeError
#    rss2ipy.MissingPages
#    rss2ipy.IndexError
#
	
	def decode_cycle_and_frame(index):
	#
	#  Decodes the specified index into the corresponding cycle and frame.
	#
	#  Return Value
	#  ============
	#
	#    (cycle, frame)
	#
		
		cycle = index // frameCount
		frame = index % frameCount
		
		return cycle, frame
	
	def append_samples(smpls, sheet, cycle, fmt):
	#
	#  Appends the samples encoded in `sheet` to `smpls`.
	#
	#  May Raise
	#  =========
	#    
	#    rss2ipy.DecoderError
	#    rss2ipy.MissingPages
	#
		
		decodedSamples = sample_decode(sheet, fmt)
		
		# If there wasn't raised an exception it is safe to go on.
		size = len(sheet[0][1])
		smpls.append(
			np.empty((1, size + 1), dtype=complex) * np.nan
		)
		
		smpls[-1][0, 0] = cycle
		smpls[-1][0, 1:] = decodedSamples
	
	# Place for RSS and for raw samples for each frame.
	rss = []
	# Don't use `[[]] * frameCount` because this leads to `frameCount`
	# times the same reference to the same empty list.
	samples = [[] for i in range(frameCount)]
	
	# Place for the sheets as well as the format and the size.
	sheet = []
	fmt = -1
	
	oldIndex = -1
	currentCycle = -1
	currentFrame = -1
	
	for page in memory:
		
		# Read the index from the beginning of a page
		ptr = 0
		currentIndex = lib.rss2i_coder_index(
		               	ctypes.cast(
		               		(ctypes.c_ubyte * 4)(*page[ptr:ptr + 4]),
		               		ctypes.POINTER(ctypes.c_ubyte)
		               	)
		               )
		ptr = 4;
		
		while ptr < 252:
			
			if oldIndex != currentIndex:
				
				oldIndex = currentIndex
				oldCycle = currentCycle
				oldFrame = currentFrame
				currentCycle, currentFrame = decode_cycle_and_frame(
				                             	currentIndex
				                             )
				
				if oldFrame != currentFrame and oldFrame != -1:
					
					if len(sheet) != 0:
						
						# Save the samples for the past frame into `samples`.
						# When the shape of `sheet` isn't as expected an
						# exception is raised.
						append_samples(
							samples[oldFrame],
							sheet,
							oldCycle,
							fmt
						)
					
					sheet = []
					fmt = -1
				
				if oldCycle != currentCycle:
				
					# Append a new row to the RSS results.
					rss.append(
						np.empty((1, frameCount + 1), dtype=float) * np.nan
					)
					rss[-1][0, 0] = currentCycle
			
			if not lib.rss2i_coder_is_result(page[ptr]):
			
				# This is a command.
				
				size = lib.rss2i_coder_command_size(page[ptr], 252 - ptr)
				currentIndex = lib.rss2i_coder_next_index(
				               	currentIndex,
				               	page[ptr],
				               	ctypes.cast(
				               		(ctypes.c_ubyte * size)(
				               			*page[ptr + 1:ptr + size + 1]
				               		),
				               		ctypes.POINTER(ctypes.c_ubyte)
				               	)
				               )
				
				ptr += size + 1
			
			else:
				
				# This is a measurement result
				
				fmtByte = page[ptr]
				fmtSpec = lib.rss2i_coder_format_specifier_from_byte(fmtByte)
				size = lib.rss2i_coder_result_size(fmtSpec)
				
				if lib.rss2i_coder_explicit_format_byte_present(fmtSpec):
					
					# This is a bunch of raw samples
					
					fmtSize = 1
					fmt = lib.rss2i_coder_postproc_format(fmtSpec)
					sheetIndex = lib.rss2i_coder_sheet_index_from_byte(fmtByte)
					
					sheet.append(
						[
							sheetIndex,
							page[ptr + fmtSize: ptr + fmtSize + size]
						]
					)
				else:
					
					# This is a single rss value
					
					fmtSize = 0
					
					if size != 2:
						raise DecodeError("Unexpected RSS size of {}.".format(size))
					
					rss[-1][0, currentFrame + 1] = rss_decode(
					                               	page[ptr: ptr + size]
					                               )
				
				if lib.rss2i_coder_result_increments_expected_index(page[ptr]):
					currentIndex += 1;
				
				ptr += size + fmtSize;
	
	if len(sheet) != 0:
		
		# Save the samples for the past frame into `samples`.
		# When the shape of `sheet` isn't as expected an
		# exception is raised.
		append_samples(
			samples[currentFrame],
			sheet,
			currentCycle,
			fmt
		)
	
	# Check for librss2i errors
	if lib.rss2i_coder_errors():
		eC = lib.rss2i_coder_error()
		raise DecodeError(debug_message(eC))
	
	try:
		rss = np.concatenate(rss, axis=0)
		samples = [np.concatenate(s, axis=0) if len(s) != 0 else None for s in samples]
		
	except Exception as e:
		
		raise DecodeError(str(e))
		
	return rss, samples


def debug_message(ec):
#
#  Returns the debug message that corresponds to the specified error case
#  {ec}. (You may have a look into 'src/debug.c' for further details.)
#
#  May Throw
#  =========
#    
#    rss2ipy.IndexError
#
	
	try:
		return lib.rss2i_debug_get_message(ec).decode();
	except (AttributeError):
		raise IndexError('The requested error case does not exist.');


def debug_get_num_of_error_cases():
#
#  Returns the number of avaiable error cases.
#
	
	return lib.rss2i_debug_get_num_of_error_cases();


def debug_get_num_of_error_cases_per_page():
#
#  Returns the number of avaiable error cases per page.
#
	
	return lib.rss2i_debug_get_num_of_error_cases_per_page();
	

def _get_cte_type(typeString):
#
#  Returns the numeric CTE type. The case of {typeString} is ignored.
#
#  Valid Type Strings
#  ==================
#    
#    "DISABLED"
#    "T2US7P1"
#    "T2US7PX1"
#
#  May Throw
#  =========
#
#    rss2ipy.IndexError    if {typeString} is not valid
#
	
	tS = typeString.upper();
	
	if tS == "DISABLED":
		
		return 0;
	
	elif tS == "T2US7P1":
		
		return 1;
	
	elif tS == "T2US7PX1":
		
		return 2;
	
	else:
		
		raise IndexError('Invalid type string "{}".'.format(typeString));

def _get_cte_ant_pat(antPatString):
#
#  Returns the numeric CTE antenna pattern. The case of {antPatString} is
#  ignored.
#
#  Valid Antenna Pattern Strings
#  =============================
#    
#    "P1TO8"
#    "DISABLED"
#
#  May Throw
#  =========
#
#    rss2ipy.IndexError    if {antPatString} is not valid
#
	
	aPS = antPatString.upper();
	
	if aPS == "P1TO8":
		
		return 1;
	
	elif aPS == "DISABLED":
		
		return 255;
	
	else:
		
		raise IndexError('Invalid antenna pattern string "{}".'.format(antPatString));

def _get_cte_const(antennaNumber):
#
#  Return CTE configuration pattern for the specified {antennaNumber}.
#
	
	if antennaNumber < 1 or antennaNumber > 99:
		
		antennaNumber = 1;
	
	return 100 + antennaNumber;

def cte_type_is_valid(typeString):
#
#  Whether the specified CTE {typeString} is valid.
#

	try:
		
		tN = _get_cte_type(typeString);
		return lib.rss2i_cteinfo_type_is_valid(tN);
		
	except (IndexError):
		
		return False;
		
def cte_antenna_pattern_is_valid(boardRev, boardType, antPatString, typeString):
#
#  Whether the specified CTE antenna pattern is supported by an RSS2I device
#  with the specified {boardRev}ision and {boardType} and the given CTE type is
#  compatible with that choice. Both {boardRev} and {boardType} may be {None},
#  in which case no board compatibility check is performed and a CTE capable
#  device with a single antenna is assumed.
#
	
	# TODO: Remove magic numbers!
	
	if boardRev is None:
		boardRev = 6;
	
	if boardType is None:
		boardType = 2;
		
	
	try:
		
		tN = _get_cte_type(typeString);
		aPN = _get_cte_ant_pat(antPatString);
		
		if boardRev < 0 or boardRev > 255 or boardType < 0 or boardType > 255:
			
			return False;
			
		else:
		
			return lib.rss2i_cteinfo_antenna_pattern_is_valid(
				boardRev,
				boardType,
				aPN,
				tN
			);
		
	except (IndexError):
		
		return False;

def cte_const_antenna_is_valid(boardRev, boardType, antNumber, typeString):
#
#  Whether the specified const antenna is supported by an RSS2I device
#  with the specified {boardRev}ision and {boardType} and the given CTE type is
#  compatible with that choice. Both {boardRev} and {boardType} may be {None},
#  in which case no board compatibility check is performed and a CTE capable
#  device with an antenna array is assumed.
#
	
	# TODO: Remove magic numbers!
	
	if boardRev is None:
		boardRev = 6;
	
	if boardType is None:
		boardType = 2;
	
	
	try:
		
		tN = _get_cte_type(typeString);
		antPatNum = _get_cte_const(antNumber);
		
		if boardRev < 0 or boardRev > 255 or boardType < 0 or boardType > 255:
			
			return False;
			
		elif antNumber < 1 or antNumber > 99:
			
			return False;
			
		else:
		
			return lib.rss2i_cteinfo_antenna_pattern_is_valid(
				boardRev,
				boardType,
				antPatNum,
				tN
			);
		
	except (IndexError):
		
		return False;

def cte_num_of_antennas(boardRev, boardType):
#
#  Returns the number of antennas that may be used to transmit/receive a BLE 5.1
#  Constant Tone Extension (CTE) with an RSS2I device with the specified
#  {boardRev}ision and {boardType}. Note, that BLE 5.0 devices will always
#  return zero. Both {boardRev} and {boardType} may be {None}, in which case no
#  board compatibility check is performed and a CTE capable device with a single
#  antenna is assumed.
#
	
	# TODO: Remove magic numbers!
	
	if boardRev is None:
		boardRev = 6;
	
	if boardType is None:
		boardType = 1;
	
	
	if boardRev < 0 or boardRev > 255 or boardType < 0 or boardType > 255:
		
		return 0;
		
	else:
		
		return lib.rss2i_cteinfo_num_of_antennas(boardRev, boardType);

def cte_standard_pattern(boardRev, boardType, typeString):
#
#  Returns the standard numeric CTE antenna pattern to the specified CTE
#  {typeString} that is compatible with the specified {boardRev}ision and
#  {boardType}. The case of {typeString} is ignored. Both {boardRev} and
#  {boardType} may be {None}, in which case no board compatibility check is
#  performed and a CTE capable device with an antenna array is assumed.
#
#

	# TODO: Remove magic numbers!
	
	if boardRev is None:
		boardRev = 6;
	
	if boardType is None:
		boardType = 2;
	
	
	tS = typeString.upper();
	
	if tS == "T2US7P1" or tS == "T2US7PX1":
		
		if cte_num_of_antennas(boardRev, boardType) >= 8:
		
			return _get_cte_ant_pat("P1TO8");
			
		else:
			
			# Fall back to constant antenna (if supported) or disabled CTE.
			return cte_standard_antenna(boardRev, boardType);
	
	else:
		
		return _get_cte_ant_pat("DISABLED");


def cte_standard_antenna(boardRev, boardType):
#
#  Returns the standard numeric CTE antenna of an RSS2I device with the
#  specified {boardRev}ision and {boardType}. Both {boardRev} and {boardType}
#  may be {None}, in which case no board compatibility check is performed and
#  a CTE capable device with a single antenna is assumed.
#
		
	if cte_num_of_antennas(boardRev, boardType) == 8:
		
		return _get_cte_const(4);
		
	elif cte_num_of_antennas(boardRev, boardType) == 1:
		
		return _get_cte_const(1);
	
	else:
		
		return _get_cte_ant_pat("DISABLED");
	
