#
#  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
#

#
#  RSS2I project python package module.
#
#  Authors
#  =======
#
#    Copyright (C) 2020-2021  René Ahrens (rene.ahrens@uni-wuppertal.de)
#    Copyright (C) 2019-2021  Gerrit Maus (funk@maus.xyz)
#
#  License
#  =======
#
#    This program is free software: you can redistribute it and/or modify it
#    under the terms of the GNU General Public License as published by the Free
#    Software Foundation, either version 3 of the License, or (at your option)
#    any later version.
#
#    This program is distributed in the hope that it will be useful, but
#    WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
#    or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
#    more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program. If not, see <https://www.gnu.org/licenses/>.
#
#  Changelog
#  =========
#
#    2021-08-16    Initial public release.
#    2021-09-19    Added functions that convert config strings to a number code
#
#  Description
#  ===========
#    This module implements the rss2ipy.config namespace, in which functions
#    to help ease the configuration of the Rss2i-Boards.
#

import typing
from rss2ipy import lib


class CONST:
    """Constants for configuring the rss2i board"""
    SYNC_THR_MIN = 10
    SYNC_THR_MAX = 2550
    FREQ_MIN = 2402
    FREQ_MAX = 2481
    MEASURE_CYCLES_MAX = 16777215     # 2^24 - 1
    LATEST_BOARD_REV = 8              # latest board revision
    CTE_BOARD_REV = 5                 # the boards can do CTE and IEEE 802.15 since revision 5
    STANDARD_ANTENNA_8_BOARD_REV = 8  # with board revision 8, the standard antenna changed from 4 to 8


class Config:
    """Class Config

    Atrributes
    ----------
    code : list
        list of config number codes
    valid : bool
        expresses validity of config str
    reason : str
        reason for invalidity
    
    Functions
    ---------
    code()
        returns number code in a list
    valid()
        returns boolean value
    reason()
        returns string if valid was false
    set_code()
        set number code
    set_valid()
        set boolean value if config string was valid
    set_reason()
        set reason why a config string is false
    
    """

    def __init__(self, code: list, valid: bool, reason: str=None):
        self._code = code
        self._valid = valid
        self._reason = reason

    def __str__(self) -> str:
        return "code:{}\nvalid:{}\nreason:{}\n".format(
            self._code, self._valid, self._reason)

    def code(self):
        return self._code

    def valid(self):
        return self._valid

    def reason(self):
        return self._reason

    def set_code(self, code: int):
        self._code = code

    def set_valid(self, valid: bool):
        self._valid = valid

    def set_reason(self, reason: str):
        self._reason = reason



#-------------------------------------------------------------------------------
#  TX Power
#-------------------------------------------------------------------------------


def txpower(txpower: int) -> Config:
    """Returns code for the defined power target

    Parameters
    ----------
    txpower: int
        Power in dBm

    Returns
    -------
    Config
    """
    
    # possible power configurations
    dBmLookup = {
        4: 0x01,
        3: 0x02,
        0: 0x03,
        -4: 0x04,
        -8: 0x05,
        -12: 0x06,
        -16: 0x07,
        -20: 0x08,
        -40: 0x09,
    }

    # Value/out-of-bounds check
    if txpower not in dBmLookup:
        code = [0x00]
        valid = False
        reason = '{} dBm is not a possible power target'.format(txpower)
    else:
        code = [dBmLookup[txpower]]
        valid = True
        reason = None

    return Config(code=code, valid=valid, reason=reason)


def default_txpower() -> Config:
    """The default tx power is 0 dBm

    Parameters
    ----------
    None

    Returns
    -------
    Config
    """
    code = 0x03   	# txPowerDefault: 0 dBm
    return Config([code], True)



#-------------------------------------------------------------------------------
#  Cycles
#-------------------------------------------------------------------------------


def cycles(cycles: int) -> Config:
    """Returns code for the defined number of cycles

    Parameters
    ----------
    cycles: int
        number of cycles

    Returns
    -------
    Config
    """
    if cycles > CONST.MEASURE_CYCLES_MAX or cycles < 1:
        config = Config([0x00], False)
    else:
        config = Config(list(cycles.to_bytes(3, byteorder='big')[::-1]), True)

    return config


def default_cycles() -> Config:
    """default cycles

    Parameters
    ----------
    None

    Returns
    -------
    Config
    """
    code = list(CONST.MEASURE_CYCLES_MAX.to_bytes(3, byteorder='big')[::-1])
    return Config(code, True)



#-------------------------------------------------------------------------------
#  Frequency
#-------------------------------------------------------------------------------


def transmitter_frequency(frequency: int) -> Config:
    """Returns code for transmitter frequency

    Parameters
    ----------
    frequency: int
        transmitter frequency in MHz

    Returns
    -------
    Config
    """
    if not CONST.FREQ_MIN <= frequency <= CONST.FREQ_MAX:
        config = Config(0x00, False)
    else:
        config = Config([frequency - 2300], True)

    return config


def receiver_frequency(frequency: int) -> Config:
    """Returns code for receiver frequency

    Parameters
    ----------
    frequency: int
        receiver frequency in MHz

    Returns
    -------
    Config
    """
    if not CONST.FREQ_MIN <= frequency <= CONST.FREQ_MAX:
        config = Config(0x00, False)
    else:
        config = Config([frequency - 2400], True)

    return config


def default_frequency() -> Config:
    """default frequency

    Parameters
    ----------
    None

    Returns
    -------
    Config
    """
    return Config([0x00], True)



#-------------------------------------------------------------------------------
#  Radio Mode
#-------------------------------------------------------------------------------


def radio_mode(mode: str, *, boardRev: typing.Optional[int] = None) -> Config:
    """Checks radio mode string and returns the radio mode code.

    Parameters
    ----------
    mode: str
        A string that defines the desired radio mode
    boardRev: typing.Optional[int] = None
        The revision as a number from the rss2i board that is to be configured

    Returns
    -------
    Config
    """
    radioModeLookup = {
        'BLE 2MBIT': 0x00,
        'BLE 1MBIT': 0x01,
        'NRF 2MBIT': 0x02,
        'NRF 1MBIT': 0x03,
        'IEEE 802.15.4': 0x04
    }

    # mode string check
    if mode.upper() not in radioModeLookup:
        return Config([0x00], False)
    else:
        config = Config([radioModeLookup[mode.upper()]], True)

    # IEEE mode(0x04)/boardRev 'F'(0x05) check
    if boardRev:
        if boardRev < CONST.CTE_BOARD_REV:
            if config.code()[0] >= 0x04:   #IEEE 802.15.4
                config.set_valid(False)

    return config



def default_radio_mode(*, boardRev: typing.Optional[int] = None) -> Config:
    """default BLE 2 MBit/s normal radio mode

    Parameters
    ----------
    None

    Returns
    -------
    Config
    """
    return Config([0x00], True)


def default_compatibility_mode() -> Config:
    """Compatibility mode deactivated

    Parameters
    ----------
    None

    Returns
    -------
    Config
    """
    return Config([0x00], True)


def activated_compatibility_mode() -> Config:
    """Compatibility mode activated

    Parameters
    ----------
    None

    Returns
    -------
    Config
    """
    return Config([0x01], True)



#-------------------------------------------------------------------------------
#  CTE - Check Configuration Entries
#-------------------------------------------------------------------------------


def cte_type(type: str, *, mode: Config, compatibility: Config) -> Config:
    """
    Whether the specified CTE {type} is valid and returns the numeric CTE type. 
    The case of {typeString} is ignored.

    Valid Type Strings
    ------------------
    "DISABLED"
    "T2US7P1"
    "T2US7PX1"

    Parameters
    ----------
    type: str
        string that defines the CTE type
    mode: Config 
        radio mode Config object
    compatibility: Config
        compatibility Config object

    Returns
    -------
    Config
    """

    cteTypeLookup = {
        'DISABLED': 0x00,
        'T2US7P1': 0x01,
        'T2US7PX1': 0x02
    }

    # type string check
    if type.upper() not in cteTypeLookup:
        reason = 'Invalid type string "{}".'.format(type)
        valid = False
        code = [0x00]
        return Config(code=code, valid=valid, reason=reason)
    else:
        reason = None
        valid = True
        code = [cteTypeLookup[type.upper()]]

    try:
        valid = lib.rss2i_cteinfo_type_is_valid(code[0])
    except (IndexError):
        reason = 'lib IndexError'
        valid = False


    # mode Config check
    if code[0] > 0x00:
        if mode.code()[0] > 0x01:
            reason = 'radio mode not compatible with cte type '
            valid = False

        # compatibility Config check
        if compatibility.code()[0] > 0x00:
            reason = 'CTE not possible with compatibility '
            valid = False       


    return Config(code=code, valid=valid, reason=reason)


def cte_antenna(
    antenna: int, 
    *, 
    type: Config, 
    boardRev: typing.Optional[int] = None, 
    boardType: typing.Optional[int] = None
    ) -> Config:
    """
    Returns a config object from a given CTE antenna string. 
    The case of {antenna} is ignored.

    Parameters
    ----------
    antenna: Config 
        CTE antenna Config object

    type: Config
        CTE type Config object

    boardRev: typing.Optional[int] = None 
        The revision as a number from the rss2i board that is to be configured

    boardType: typing.Optional[int] = None
        The antenna type as a number from the rss2i board that is to be configured

    Returns
    -------
    Config
    """

    if antenna < 0 or antenna > 8:
        reason = "antenna not in range: 0..8"
        valid = False
        code = [0x00]
    else:
        reason = 'All Tests passed!'
        valid = True 
        code = [100 + antenna]


    if boardRev is None:
        boardRev = CONST.LATEST_BOARD_REV

    if boardType is None:
        boardType = 2

    try:
        tN = type.code()[0]
        antPatNum = code[0]

        if boardRev < 0 or boardRev > 255 or boardType < 0 or boardType > 255:
            valid = False
            reason = 'board rev oder type invalid'

        elif antenna < 1 or antenna > 99:
            valid = False
            reason = 'antenna invalid'

        else:
            valid = lib.rss2i_cteinfo_antenna_pattern_is_valid(
                boardRev,
                boardType,
                antPatNum,
                tN
                )
            reason = "Lib check: Board Rev. [{}] or Borad Type [{}] not capable ".format(
                boardRev,
                boardType
            )

    except (IndexError):	
        valid = False
        reason = 'Index Error'

    return Config(code=code, valid=valid, reason=reason)


def cte_pattern(
    pattern: str, 
    *, 
    type: Config, 
    boardRev: typing.Optional[int] = None, 
    boardType: typing.Optional[int] = None
    ) -> Config:
    """
    Returns a config object from a given CTE antenna pattern string. 
    The case of {pattern} is ignored.

    Parameters
    ----------
    pattern: Config 
        CTE pattern Config object

    type: Config
        CTE type Config object

    boardRev: typing.Optional[int] = None 
        The revision as a number from the rss2i board that is to be configured

    boardType: typing.Optional[int] = None
        The antenna type as a number from the rss2i board that is to be configured

    Returns
    -------
    Config
    """

    ctePatternLookup = {
        'DISABLED': 0xff,
        'P1TO8': 0x01
    }

    if pattern.upper() not in ctePatternLookup:
        reason = 'Invalid antenna pattern string "{}".'.format(pattern)
        valid = False
        code = [0x00]
    else:
        reason = "Everything is alright!"
        valid = True
        code = [ctePatternLookup[pattern.upper()]]

    if boardRev is None:
        boardRev = CONST.LATEST_BOARD_REV

    if boardType is None:
        boardType = 2
		
    try:     
        tN = type.code()[0]
        aPN = code[0]
        
        if boardRev < 0 or boardRev > 255 or boardType < 0 or boardType > 255:
            reason = "board Rev or board Type are not in a valid range " 
            valid = False
            
        else:
            reason = "Lib check: Board Rev. [{}] or Borad Type [{}] not capable ".format(
                boardRev,
                boardType
            )
            valid = lib.rss2i_cteinfo_antenna_pattern_is_valid(
                boardRev,
                boardType,
                aPN,
                tN
            )
        
    except (IndexError):
        reason = "Lib check: Board Rev to low... ?"
        valid = False

    return Config(code=code, valid=valid, reason=reason)


#-------------------------------------------------------------------------------
#  CTE - Induced Behaviour
#-------------------------------------------------------------------------------


def induced_cte_pattern(
    *, 
    type: Config, 
    antenna: Config, 
    boardRev: typing.Optional[int] = None, 
    boardType: typing.Optional[int] = None
    ) -> Config:
    """Returns Config object with CTE pattern byte code as 
    the current device is in receiving mode and another device
    sends a CTE on one antenna.

    Parameters
    ----------
    type: Config
        CTE type Config object

    antenna: Config 
        CTE antenna Config object

    boardRev: typing.Optional[int] = None 
        The revision as a number from the rss2i board that is to be configured

    boardType: typing.Optional[int] = None
        The antenna type as a number from the rss2i board that is to be configured

    Returns
    -------
    Config
    """
    reason = None
    
    if 101 <= antenna.code()[0] <= 108:
        valid = True
    else:
        valid = False
        reason = 'Expected constant antenna between 1-8!'

    # check if board type and board rev are valid
    if boardRev < 0 or boardRev > 255 or boardType < 0 or boardType > 255:   
        numOfCteAtennas = 0
        
    else:
        # retrive number of CTE antennas      
        numOfCteAtennas =  lib.rss2i_cteinfo_num_of_antennas(boardRev, boardType)


    if numOfCteAtennas == 8:
        # switching pattern: P1TO8
        code = [0x01]

    # should be for board revisions lower F
    elif numOfCteAtennas == 1:
        code = [101]

    return Config(code=code, valid=valid, reason=reason)


def induced_cte_antenna(
    *, 
    type: Config, 
    pattern: Config, 
    boardRev: typing.Optional[int] = None, 
    boardType: typing.Optional[int] = None
    ) -> Config:
    """Returns Config object with CTE antenna byte code as 
    the current device is in receiving mode and another device
    sends a switching pattern.

    Parameters
    ----------
    type: Config
        CTE type Config object

    pattern: Config 
        CTE pattern Config object

    boardRev: typing.Optional[int] = None 
        The revision as a number from the rss2i board that is to be configured

    boardType: typing.Optional[int] = None
        The antenna type as a number from the rss2i board that is to be configured

    Returns
    -------
    Config
    """

    if boardRev is None:
        boardRev = CONST.LATEST_BOARD_REV

    if boardType is None:
        boardType = 1

    # check if board type and board rev are valid
    if boardRev < 0 or boardRev > 255 or boardType < 0 or boardType > 255:   
        numOfCteAtennas = 0
        
    else:
        # retrive number of CTE antennas      
        numOfCteAtennas =  lib.rss2i_cteinfo_num_of_antennas(boardRev, boardType)

    # check for switching CTE pattern (behaviour)
    if pattern.code()[0] == 0x01: # P1TO8
        if numOfCteAtennas == 8:
            if boardRev < CONST.STANDARD_ANTENNA_8_BOARD_REV:
                # for board rev F to H
                code = [104]
            else:
                # since borad rev I
                code = [108]

        # should be for board revisions lower F
        elif numOfCteAtennas == 1:
            code = [101]

    # no CTE
    else:
        code = [0]  # antenna disabled

    return Config(code=code, valid=True)



#-------------------------------------------------------------------------------
#  CTE - Default Settings
#-------------------------------------------------------------------------------

 
def default_cte_type(
    *, 
    mode: Config, 
    compatibility: Config, 
    boardRev: typing.Optional[int] = None, 
    boardType: typing.Optional[int] = None
    ) -> Config:
    """Returns Config object with default cte type byte code

    Parameters
    ----------
    mode: Config
        radio mode Config object

    compatibility: Config 
        radio compatibility Config object

    boardRev: typing.Optional[int] = None 
        The revision as a number from the rss2i board that is to be configured

    boardType: typing.Optional[int] = None
        The antenna type as a number from the rss2i board that is to be configured

    Returns
    -------
    Config
    """

    return Config([0x00], True)


def default_cte_behaviour(
    *, 
    type: typing.Optional[Config] = None, 
    boardRev: typing.Optional[int] = None, 
    boardType: typing.Optional[int] = None
    ) -> Config:
    """Returns Config object with default cte behaviour(pattern/antenna) byte code

    Parameters
    ----------
    mode: Config
        radio mode Config object

    compatibility: Config 
        radio compatibility Config object

    boardRev: typing.Optional[int] = None 
        The revision as a number from the rss2i board that is to be configured

    boardType: typing.Optional[int] = None
        The antenna type as a number from the rss2i board that is to be configured

    Returns
    -------
    Config
    """
    # 255 CTE disabled, i. e. no active participation during
    # the CTE (if any). Use standard antenna during
    # the actual packet.
    return Config([0xff], True)



#-------------------------------------------------------------------------------
#  CTE - Sampling Configuration
#-------------------------------------------------------------------------------


def storage(store: str) -> Config:
    """Checks and decodes the storage string from a config file

    Parameters
    ----------
    store: str
        A string that describes the storage behaviour

    Returns
    -------
    Config
    """

    storeLookup = {
        'deactivated': 0x00,
        'compatible': 0x01,
        'activated': 0x02
    }

    # check and decode storage string
    if store.lower() not in storeLookup:
        config = Config([0x00], False)
    else:
        config = Config([storeLookup[store.lower()]], True)

    return config


def default_storage() -> Config:
    """Returns Config object with byte code that deactivates sample storage

    Parameters
    ----------
    None

    Returns
    -------
    Config
    """

    return Config([0x00], True)


def sampling(
    sample: str, 
    *, 
    frequency: Config, 
    mode: Config, 
    compatibility: Config, 
    type: Config, 
    behaviour: Config
    ) -> Config:
    """Checks and decodes the sample string from a config file

    Parameters
    ----------
    sample: str
        A string that describes the sample behaviour

    frequency: Config
        frequency Config object

    mode: Config
        radio mode Config object

    compatibility: Config 
        radio compatibility Config object

    type: Config
        CTE type Config object

    behaviour: Config
        CTE behaviour(pattern/antenna) Config object

    Returns
    -------
    Config
    """

    sampleLookup = {
        'none': 0x00,
        'iq': 0x01
    }

    # check and decode sample string
    if sample.lower() not in sampleLookup:
        valid = False
        code = [0x00]
        reason = 'not a valid sample behaviour'
    else:
        valid = True
        code = [sampleLookup[sample.lower()]]
        reason = None

    # check if CTE behaviour is present and the device is receiving in the given frame
    if not (2 <= frequency.code()[0] <= 81 and behaviour.code()[0] > 0):
        valid = False
        reason = 'there are no sample to collect'

    return Config(code=code, valid=valid, reason=reason)


def default_sampling(
    *, 
    frequency: Config, 
    mode: Config, 
    compatibility: Config, 
    type: Config, 
    behaviour: Config
    ) -> Config:
    """If CTE behaviour is present, the function returns a Config object
    that with a byte code to enable sampling.

    Parameters
    ----------
    frequency: Config
        frequency Config object

    mode: Config
        radio mode Config object

    compatibility: Config 
        radio compatibility Config object

    type: Config
        CTE type Config object

    behaviour: Config
        CTE behaviour(pattern/antenna) Config object

    Returns
    -------
    Config
    """

    # check if CTE behaviour is present and the device is receiving in the given frame
    if 2 <= frequency.code()[0] <= 81 and behaviour.code()[0] > 0:
        # enable sampling
        code = [0x01]
    else:
        # no CTE -> no sampling
        code = [0x00]

    return Config(code=code, valid=True)