#!/usr/bin/env python3

#
#  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
#

#
#  RSS2I command line tool for configuring and monitoring RSS2I devices and
#  reading data from them.
#
#  Authors
#  =======
#
#    Copyright (C) 2019-2020  Knut Niemann (post@knut-niemann.de)
#    Copyright (C) 2020-2021  René Ahrens (rene.ahrens@uni-wuppertal.de)
#    Copyright (C) 2019-2021  Gerrit Maus (funk@maus.xyz)
#
#  License
#  =======
#
#    This program is free software: you can redistribute it and/or modify it
#    under the terms of the GNU General Public License as published by the Free
#    Software Foundation, either version 3 of the License, or (at your option)
#    any later version.
#
#    This program is distributed in the hope that it will be useful, but
#    WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
#    or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
#    for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program. If not, see <https://www.gnu.org/licenses/>.
#
#  Changelog
#  =========
#
#    2019-09-17    Removed matplotlib and replaced it by pyqtgraph/pyqt5.
#    2019-07-01    Some subcommand (e.g. `start`) is required now.
#    2019-06-26    Made `output` switch optional for `start` command.
#    2020-02-12    Fixed/Added/Removed delays after "Jump on" API call.
#    2020-03-05    Fixed handling of Ctrl+C and closed monitor windows.
#    2020-03-06    Only search for valid USB UART bridge COM ports during
#                  the automatic detection (-a) of connected devices.
#    2020-03-19    #9: Check for the livelog start byte.
#    2020-03-19    #10: Catch livelog timeouts.
#    2020-05-20    (Try to) catch `BrokenPipeError`s.
#    2020-06-16    #14: Fix `KeyError` when running `getid` with unknown
#                  revision or board type. Add rev. G and the arr. antenna type.
#    2020-08-11    #9, #10: Unify data acquisition for debug and monitor command
#    2020-08-11    #12: Rename command aliases: 'debug' becomes 'livelog'.
#    2020-08-11    #13: Add a new debug command that is wothy its name.
#    2020-08-11    #15: Add '--csv' switch to enable comma separated output.
#    2020-08-21    #11: Changed the default RSS input map behaviour.
#    2021-03-19    Remove -c argument from livelog subcommand.
#    2021-03-19    Add local RSS sample amount to the default live log config.
#    2021-03-19    Support RSS2I board revision I.
#    2021-03-19    Support target field of the *Tell me who you are* query.
#    2021-03-19    Added the `-d` switch for the getid subcommand.
#    2021-03-25    Update and amend documentation.
#    2021-04-01    Add `page` field of `0x19` *Report error information*.
#    2021-04-01    Move the `0x18` and `0x19` cmds to `0x19` and `0x1a`.
#    2021-04-27    #18: Unify output prefix `-o` switch.
#    2021-04-29    #19: Support downloading of raw samples.
#    2021-04-30    Return a useful exit status.
#    2021-04-30    Fix `IndexError` during virtual memory standard configuration
#    2021-05-01    Fix `restore` subcmd behaviour when `-r` switch is given.
#    2021-05-02    Add `-n` switch to the `restore` subcommand.
#    2021-06-16    #26 Request lax timeouts for the stop cmd
#    2021-09-27    #25 Support radio mode, CTE and sampling configuration
#    2021-09-27    Complete compatibility with windows is restored
#
#  Description
#  ===========
#
#    There is no user documentation for the RSS2I command line tool yet. Until
#    it is available it may be helpful to run
#    ```
#    rss2i --help
#    ```
#    and
#    ```
#    rss2i <command> --help
#    ```
#    in order to get an overview about the tool.

import argparse
import enum
import functools
import inspect
import json
import math
import multiprocessing
import os
import queue
import re
import sys
import threading
import time
import textwrap
from datetime import datetime
from typing import Dict, List, TextIO, Tuple, Any, Optional
from collections import defaultdict

import numpy as np
import pandas as pd
import rss2ipy
import rss2ipy.config
import serial.tools.list_ports


# -----------------------------------------------------------------------------
# Debug -----------------------------------------------------------------------
def debug_caller(f):
    @functools.wraps(f)
    def inner_func(*args, **kwargs):
        caller = inspect.stack()[1]
        on_debug(f':: {caller[3]}() -> {f.__name__}() @ line #{caller.lineno}',
                 verbosity_threshold=3)
        for arg in args:
            on_debug(f'     with: {arg}', verbosity_threshold=3)
        for k, v in kwargs.items():
            on_debug(f'     with: {k}={v}', verbosity_threshold=3)
        return f(*args, **kwargs)

    return inner_func


# -----------------------------------------------------------------------------
# Global variables - connected devices ----------------------------------------
_connectedDevices: Dict[int, rss2ipy.Device] = {}
_openSerialAddresses: Dict[int, str] = {}
_proxyPreferences: Dict[int, List[int]] = {}

# Determines if and how many debug messages are to be printed
# Set via '-v'/'--verbose'...'-vvv' as launch argument
# 0 - None
# 1 - General info messages
# 2 - Debug messages
# 3 - In-depth debug messages: simultanious download is disabled to display
#     every action
_verbosityLevel = 0

_outputLock = multiprocessing.Value('i', 0)
_livelog_event = threading.Event()

_abort = False
_queue: queue.Queue = queue.Queue()

viewPortWidth = 10000
viewPortMemory = 2  # times the viewPortWidth
ppAmpOffset = -55
drawStep = 5

# print progress
_lastOutput = ''


# -----------------------------------------------------------------------------
# Error handling --------------------------------------------------------------


class Rss2iException(Exception):
    pass

class Rss2iError(Rss2iException):
    pass

class Rss2iWarning(Rss2iException, Warning):
    pass

# Parsing errors
class SyntaxError(Rss2iError, SyntaxError):
    def __init__(self, line: str, lineContent: str) -> None:
        super().__init__(line)
        self._lineContent = lineContent

    @property
    def lineContent(self):
        return self._lineContent

    def __str__(self) -> str:
        return "Syntax error while parsing the config file "\
               "@ line #{} [{}]".format(self.args[0], self._lineContent)

    def __repr__(self) -> str:
        return "SyntaxError({!r}, {!r})".format(self.args[0], self._lineContent)


class TxPowerInvalid(Rss2iError):
    def __init__(self, power: str, lineError: Exception=None) -> None:
        super().__init__(power)
        self._lineError =lineError

    def __str__(self) -> str:
        fs = "TxPower is not valid: {}".format(self.args[0])
        if self._lineError:
            return "{}\n\t{}".format(fs, self._lineError)
        else:
            return fs

    def __repr__(self) -> str:
        return "TxPowerInvalid({!r})".format(self.args[0])


class FrameMissing(Rss2iError):
    def __init__(self, missingFrame: list) -> None:
        super().__init__(missingFrame)
    
    def __str__(self) -> str:
        return "The parser could not find all frames it expected. Frame #{} "\
               "are missing".format(self.args[0])

    def __repr__(self) -> str:
        return "FrameMissing({!r})".format(self.args[0])


class FrameUndefined(Rss2iError):
    def __init__(self, frameIndex: str, line: str, lineContent: str) -> None:
        super().__init__(frameIndex)
        self._lineContent = lineContent
        self._line = line

    @property
    def line(self):
        return self._line

    @property
    def lineContent(self):
        return self._lineContent

    def __str__(self) -> str:
        return "Frame #{} is referenced but never defined @ line#{} "\
               "[{}]".format(self.args[0],
                             self._line,
                             self._lineContent)

    def __repr__(self) -> str:
        return "FrameUndefined({!r}, {!r}, {!r})".format(self.args[0],
                                                         self._line, 
                                                         self._lineContent)


class VDevMissing(Rss2iError):
    def __init__(self, missingVdevs: list) -> None:
        super().__init__(missingVdevs)

    def __str__(self) -> str:
        return "The virtual devices are explicitly defined, but not enough virtual "\
               "devices were declared for frame #{}".format(self.args[0])

    def __repr__(self) -> str:
        return "VDevMissing({!r})".format(self.args[0])


class ConfigFileNotFound(Rss2iError):
    def __init__(self, path: str) -> None:
        super().__init__(path)

    def __str__(self) -> str:
        return "No config file could be found at the given location [{}]. "\
               "Check if the path (given by \'-f\') leads to a exisitng "\
               "file".format(self.args[0])

    def __repr__(self) -> str:
        return "ConfigFileNotFound({!r})".format(self.args[0])


class FrameCountOutOfRange(Rss2iError):
    def __init__(self, device: int, frameCount: int) -> None:
        super().__init__(device)
        self._frameCount = frameCount
    
    @property
    def frameCount(self):
        return self._frameCount

    def __str__(self) -> str:
        return "The configuration of the device failed: The frame count is "\
               "set too high (Device, value): ({}, {})".format(self.args[0], self._frameCount)

    def __repr__(self) -> str:
        return "FrameCountOutOfRange({!r}, {!r})".format(self.args[0], self._frameCount)


class CycleCountOutOfRange(Rss2iError):
    def __init__(self, device: int, cycle: int) -> None:
        super().__init__(device)
        self._cycle = cycle

    @property
    def cycle(self):
        return self._cycle

    def __str__(self) -> str:
        return "The configuration of the device failed: The cycle count is set too "\
               "high (Device, value): ({}, {})".format(self.args[0], self._cycle)

    def __repr__(self) -> str:
        return "CycleCountOutOfRange({!r}, {!r})".format(self.args[0], self._cycle)


class FrequencyOutOfRange(Rss2iError):
    def __init__(self, device: int, frequency: int) -> None:
        super().__init__(device)
        self._frequency = frequency

    @property
    def frequency(self):
        return self._frequency

    def __str__(self) -> str:
        return "The configuration of the device failed: The frequency value is "\
               "out of bounds (Device, value): ({}, {})".format(self.args[0], self._frequency)

    def __repr__(self) -> str:
        return "FrequencyOutOfRange({!r}, {!r})".format(self.args[0], self._frequency)


class FrequencyInvalid(Rss2iError):
    def __init__(self, frequency: int, lineError: Exception=None) -> None:
        super().__init__(frequency)
        self._lineError =lineError

    def __str__(self) -> str:
        fs = "Frequency is not valid: {} MHz is not in range of {} MHz - {} MHz".format(
            self.args[0],
            rss2ipy.config.CONST.FREQ_MIN,
            rss2ipy.config.CONST.FREQ_MAX
        )
        if self._lineError:
            return "{}\n\t{}".format(fs, self._lineError)
        else:
            return fs

    def __repr__(self) -> str:
        return "FrequencyInvalid({!r})".format(self.args[0])


class VDevCountOutOfRange(Rss2iError):
    def __init__(self, device:int, vdevCount: int) -> None:
        super().__init__(device)
        self._vdevCount = vdevCount

    @property
    def vdevCount(self):
        return self._vdevCount

    def __str__(self) -> str:
        return "The configuration of the device failed: The virtual device count "\
               "is set too high (Device, value): ({}, {})".format(self.args[0], self._vdevCount)

    def __repr__(self) -> str:
        return "VDevCountOutOfRange({!r}, {!r})".format(self.args[0], self._vdevCount)


class VDevIdOutOfRange(Rss2iError):
    def __init__(self, device: int, vdevId: int) -> None:
        super().__init__(device)
        self._vdevId = vdevId

    @property
    def vdevId(self):
        return self._vdevId

    def __str__(self) -> str:
        return "The configuration of the device failed: The virtual device ID is "\
               "out of bounds (Device, value): ({}, {})".format(self.args[0], self._vdevId)

    def __repr__(self) -> str:
        return "VDevIdOutOfRange({!r}, {!r})".format(self.args[0], self._vdevId)


class VMemIdOutOfRange(Rss2iError):
    def __init__(self, device: int, vmemId: int) -> None:
        super().__init__(device)
        self._vmemId = vmemId

    @property
    def vmemId(self):
        return self._vmemId

    def __str__(self) -> str:
        return "The configuration of the device failed: The virtual memory ID is "\
               "set to high (Device, value): ({}, {})".format(self.args[0], self._vmemId)

    def __repr__(self) -> str:
        return "VMemIdOutOfRange({!r}, {!r})".format(self.args[0], self._vmemId)


class VMemMissing(Rss2iError):
    def __init__(self, missingVmems: list) -> None:
        super().__init__(missingVmems)

    def __str__(self) -> str:
        return "Not enough virtual memories declared for \
                frame #{}".format(self.args[0])

    def __repr__(self) -> str:
        return "VMemMissing({!r})".format(self.args[0])


class RadioModeInvalid(Rss2iError):
    def __init__(self, mode: str, lineError: Exception=None) -> None:
        super().__init__(mode)
        self._lineError = lineError

    def __str__(self) -> str:
        fs = "Radio mode is not supported: '{}'".format(self.args[0])
        if self._lineError:
            return "{}\n\t{}".format(fs, self._lineError)
        else:
            return fs

    def __repr__(self) -> str:
        return "RadioModeInvalid({!r})".format(self.args[0])


class CteTypeInvalid(Rss2iError):
    def __init__(self, cteType: str, lineError: Exception=None) -> None:
        super().__init__(cteType)
        self._lineError = lineError

    def __str__(self) -> str:
        fs = "CTE type is not supported: {}".format(self.args[0])
        if self._lineError:
            return "{}\n\t{}".format(fs, self._lineError)
        else:
            return fs

    def __repr__(self) -> str:
        return "CteTypeInvalid({!r})".format(self.args[0])


class CteConfigIncompatibleWithMeasureConfig(Rss2iError):
    def __init__(self, frame: int) -> None:
        super().__init__(frame)

    def __str__(self) -> str:
        return "The CTE configuration does not comply with the current measurement "\
            "configuration in frame #{}".format(self.args[0])

    def __repr__(self) -> str:
        return "CteConfigIncompatibleWithMeasureConfig({!r})".format(self.args[0])


class CteBehaviourInvalid(Rss2iError):
    def __init__(self, frame: int, cteType: str, device: int) -> None:
        super().__init__(frame)
        self._cteType = cteType
        self._device = device

    def __str__(self) -> str:
        return "The specified CTE behaviour is not supported by the device #{} or "\
            "invalid or incompatible with the stated CTE type [{}]. This error was "\
            "first reported within the configuration value of frame #{}".format(
                self._device, self._cteType, self.args[0])

    def __repr__(self) -> str:
        return "CteBehaviourInvalid({!r}, {!r}, {!r})".format(
            self.args[0], self._cteType, self._device)


class ConstantAntennaInvalid(Rss2iError):
    def __init__(self, constantAntenna: int, reason: str, lineError: Exception=None) -> None:
        super().__init__(constantAntenna)
        self._lineError = lineError
        self._reason = reason

    def __str__(self) -> str:
        fs = "Constant Antenna is not valid: '{}' {}".format(
            self.args[0], self._reason
        )
        if self._lineError:
            return "{}\n\t{}".format(fs, self._lineError)
        else:
            return fs

    def __repr__(self) -> str:
        return "ConstantAntennaInvalid({!r})".format(self.args[0])


class SwitchingPatternInvalid(Rss2iError):
    def __init__(self, switchingPattern: str, lineError: Exception=None) -> None:
        super().__init__(switchingPattern)
        self._lineError =lineError

    def __str__(self) -> str:
        fs = "Switching Pattern is not valid: '{}'".format(
            self.args[0],
        )
        if self._lineError:
            return "{}\n\t{}".format(fs, self._lineError)
        else:
            return fs

    def __repr__(self) -> str:
        return "SwitchingPatternInvalid({!r})".format(self.args[0])


class SyncThresholdInvalid(Rss2iError):
    def __init__(self, syncThreshold: str, lineError: Exception=None) -> None:
        super().__init__(syncThreshold)
        self._lineError =lineError

    def __str__(self) -> str:
        fs = "SyncThreshold is not valid: '{}' in not in range of {} - {}".format(
            self.args[0],
            rss2ipy.config.CONST.SYNC_THR_MIN, 
            rss2ipy.config.CONST.SYNC_THR_MAX,
        )
        if self._lineError:
            return "{}\n\t{}".format(fs, self._lineError)
        else:
            return fs

    def __repr__(self) -> str:
        return "SyncThresholdInvalid({!r})".format(self.args[0])


class CteIncompatibleWithRadio(Rss2iError):
    def __init__(self, radioMode: str) -> None:
        super().__init__(radioMode)
    
    def __str__(self) -> str:
        return "The currently configured radio mode '{}' is not supported with "\
            "a CTE".format(self.args[0])

    def __repr__(self) -> str:
        return "CteIncompatibleWithRadio({!r})".format(self.args[0])


class SampleConfigIncompatibleWithMeasureConfig(Rss2iError):
    def __init__(self, frame: int) -> None:
        super().__init__(frame)

    def __str__(self) -> str:
        return "The sample configuration does not comply with the current measurement "\
            "configuration in frame #{}".format(self.args[0])

    def __repr__(self) -> str:
        return "SampleConfigIncompatibleWithMeasureConfig({!r})".format(self.args[0])

class SampleTypeInvalid(Rss2iError):
    def __init__(self, frame: int) -> None:
        super().__init__(frame)

    def __str__(self) -> str:
        return "The sample type specified for frame #{} is not valid.".format(self.args[0])

    def __repr__(self) -> str:
        return "SampleTypeInvalid({!r})".format(self.args[0])


class SampleTypeIncompatibleWithConfig(Rss2iError):
    def __init__(self, frame: int, radioMode: str, cteType: str) -> None:
        super().__init__(frame)
        self._radioMode = radioMode
        self._cteType = cteType

    def __str__(self) -> str:
        return "The sample type specified for frame #{} is not supported by "\
            "the current radio mode [{}] or CTE [{}]".format(
                self.args[0], self._radioMode, self._cteType)

    def __repr__(self) -> str:
        return "SampleTypeInvalid({!r})".format(self.args[0])


class ConfigDecodeError(Rss2iError):
    pass


class LineError(Rss2iError):
    def __init__(self, 
        line: str, 
        lineContent: str=None) -> None:

        super().__init__(line)
        self._lineContent = lineContent

    @property
    def lineContent(self):
        return self._lineContent

    def __str__(self) -> str:
        return "The parsed value is invalid/out-of-bounds @ line #{} [{}]".format(
            self.args[0], 
            self._lineContent
        )
 
    def __repr__(self) -> str:
        return "LineError({!r}, {!r})".format(
            self.args[0], 
            self._lineContent
        )


# Device response errors
class ControlModeRequired(Rss2iError):
    def __init__(self, devID: int) -> None:
        super().__init__(devID)

    def __str__(self) -> str:
        return "The device #{} requires entering control mode to be accessed".format(self.args[0])

    def __repr__(self) -> str:
        return "ControlModeRequired({!r})".format(self.args[0])


class MissingMeasurementConfiguration(Rss2iError):
    def __init__(self, devID: int) -> None:
        super().__init__(devID)

    def __str__(self) -> str:
        return "The device #{} requires a configuration before it can be used".format(self.args[0])

    def __repr__(self) -> str:
        return "MissingMeasurementConfiguration({!r})".format(self.args[0])


class NoDeviceAbleToStartMeasurement(Rss2iError):
    def __str__(self) -> str:
        return "No connected device is able to start the measurement, "\
                "as no device has an active frame in the current configuration"


class FlashMemoryBusy(Rss2iError):
    def __init__(self, devID: int) -> None:
        super().__init__(devID)

    def __str__(self) -> str:
        return "The flash memory of device #{} could not be read. The memory "\
               "controller reports high load".format(self.args[0])

    def __repr__(self) -> str:
        return "FlashMemoryBusy({!r})".format(self.args[0])


class FlashMemoryCorruption(Rss2iError):
    def __init__(self, devID: int) -> None:
        super().__init__(devID)

    def __str__(self) -> str:
        return "The flash memory of device #{} could not be read. The memory "\
                  "controller reports corrupt memory/access of not-yet-written-to "\
               "memory ranges".format(self.args[0])

    def __repr__(self) -> str:
        return "FlashMemoryCorruption({!r})".format(self.args[0])


class NoConnectedDeviceAbleToAbort(Rss2iError):
    def __str__(self) -> str:
        return "The measurement could not be stopped, as no connected device has an "\
               "active part in the measurement to propagate the stop signal"


class DownloadTimeout(Rss2iError):
    def __init__(self, devID: int) -> None:
        super().__init__(devID)

    def __str__(self) -> str:
        return "A serial timeout while downloading from device #{} has ocurred".format(self.args[0])

    def __repr__(self) -> str:
        return "DownloadTimeout({!r})".format(self.args[0])


# Device access not possible errors
class ProxyArgumentSyntaxError(Rss2iError):
    def __init__(self, proxy: str) -> None:
        super().__init__(proxy)

    def __str__(self) -> str:
        return "The proxy argument is not parseable [{}]. Check for the correct "\
               "formatting: \'-x target_device:proxy_device_1,..,proxy_device_n\' "\
               "ex: \'-x 1:7,8,9\'".format(self.args[0])

    def __repr__(self) -> str:
        return "ProxyArgumentSyntaxError({!r})".format(self.args[0])


class NoDevicesFound(Rss2iError):
    def __init__(self, device: str) -> None:
        super().__init__(device)

    def __str__(self) -> str:
        return "No connected device found with #{}".format(self.args[0])

    def __repr__(self) -> str:
        return "NoDevicesFound({!r})".format(self.args[0])


# Critical workflow errors
class ConfigurationFailed(Rss2iError):
    def __init__(self, device: int) -> None:
        super().__init__(device)

    def __str__(self) -> str:
        return "The configuration of device #{} has not been successful".format(self.args[0])

    def __repr__(self) -> str:
        return "ConfigurationFailed({!r})".format(self.args[0])


class MonitorSetupFailed(Rss2iError):
    def __init__(self, message: str) -> None:
        super().__init__(message)

    def __str__(self) -> str:
        return "The monitor window setup failed {}".format(self.args[0])

    def __repr__(self) -> str:
        return "MonitorSetupFailed({!r})".format(self.args[0])


# Warnings
class DevicesStillMeasuring(Rss2iWarning):
    def __str__(self) -> str:
        return "The measurement could not be stopped completely."\
               "There are still devices sending measurement packets"


class SyncThresholdHigh(Rss2iWarning):
    def __init__(self, defined: int, required: int) -> None:
        super().__init__(defined)
        self._required = required

    @property
    def required(self):
        return self._required

    def __str__(self) -> str:
        return "There are more virtual devices for a synchronisation required, than "\
               "defined [defined: {}, required: {}]".format(self.args[0], self._required)

    def __repr__(self) -> str:
        return "SyncThresholdHigh({!r}, {!r})".format(self.args[0], self._required)


class SyncThresholdRounding(Rss2iWarning):
    def __init__(self, cycles: int, line: str, lineContent:str) -> None:
        super().__init__(cycles)
        self._lineContent = lineContent
        self._line = line

    @property
    def line(self):
        return self._line

    @property
    def lineContent(self):
        return self._lineContent

    def __str__(self) -> str:
        return "The number of given cycles {} will be rounded, after "\
               "being devided by 10 (see docs) {} {}".format(self.args[0],
                                                             self._line, 
                                                             self._lineContent)

    def __repr__(self) -> str:
        return "SyncThresholdRounding({!r}, {!r}, {!r})".format(self.args[0],
                                                                self._line, 
                                                                self._lineContent)


class OverTheAirTimeout(Rss2iWarning):
    def __init__(self, devID: int) -> None:
        super().__init__(devID)

    def __str__(self) -> str:
        return "A timeout ocurred when trying to send data over-the-air "\
               "to device #{}".format(self.args[0])

    def __repr__(self) -> str:
        return "OverTheAirTimeout({!r})".format(self.args[0])


class DeviceWasAlreadyMeasuring(Rss2iWarning):
    def __init__(self, devID: int) -> None:
        super().__init__(devID)

    def __str__(self) -> str:
        return "The first response from device #{} may have been lost due to "\
               "bad connection quality".format(self.args[0])

    def __repr__(self) -> str:
        return "DeviceWasAlreadyMeasuring({!r})".format(self.args[0])


# -----------------------------------------------------------------------------
# Utility and helper methods --------------------------------------------------
def print_(outStream: TextIO, message: str, newLine: bool = True,
           usePrefix: bool = True, formatIndent: bool = False,
           formatIndentWidth: int = 80) -> None:
    """Outputs 'message' to 'outStream' with the caller module as prefix

    Parameters
    ----------
    outStream : TestIO
        Output Stream to be printed to

    message : str
        Object to be printed to stderr

    newLine : bool, optional
        If set prints a new line after the message, by default True

    usePrefix : bool, optional
        if set uses the caller module as output prefix, by default True

    formatIndent: bool, optional
        if set formats the output within 'formatIndentWidth' (default 80)
        chars to wrap and indent; no effect, if 'usePrefix' is not set,
        by default False

    formatIndentWidth: int, optional
        format width, by default 80

    Returns
    -------
    None
    """
    message = str(message)

    prefix = ''
    if usePrefix:
        caller = os.path.splitext(os.path.basename(inspect.stack()[1].filename))[0]
        prefix = '{}: '.format(caller)

        if formatIndent:
            indentCount = ceil(len(prefix) / 8.0)
            indent = ''.join(['\t'] * indentCount)
            prefixIndentCount = (indentCount - int(len(prefix) / 8.0))
            prefix += ''.join(['\t'] * prefixIndentCount)
            message = '\n{}'.format(indent).join(
                wrap(message, formatIndentWidth - 8 * prefixIndentCount)
            )

    output = '{}{}'.format(prefix, message)

    outStream.write(output)

    if newLine:
        outStream.write('\n')

    outStream.flush()


def print_err(message: str, newLine: bool = True, usePrefix: bool = True,
              formatIndent: bool = False, formatIndentWidth: int = 80) -> None:
    """Outputs 'message' to stderr with the caller module as prefix

    Parameters
    ----------
    message : str
        Object to be printed to stderr

    newLine : bool, optional
        If set prints a new line after the message, by default True

    usePrefix : bool, optional
        if set uses the caller module as output prefix, by default True

    formatIndent: bool, optional
        if set formats the output within 'formatIndentWidth' (default 80)
        chars to wrap and indent; no effect, if 'usePrefix' is not set,
        by default False

    formatIndentWidth: int, optional
        format width, by default 80

    Returns
    -------
    None
    """
    print_(
        sys.stderr, 
        message, 
        newLine, 
        usePrefix, 
        formatIndent,
        formatIndentWidth
    )


def print_out(message: str, newLine: bool = True, usePrefix: bool = True,
              formatIndent: bool = False, formatIndentWidth: int = 80) -> None:
    """Outputs 'message' to stdout with the caller module as prefix

    Parameters
    ----------
    message : str
        Object to be printed to stdout

    newLine : bool, optional
        If set prints a new line after the message, by default True

    usePrefix : bool, optional
        if set uses the caller module as output prefix, by default True

    formatIndent: bool, optional
        if set formats the output within 'formatIndentWidth' (default 80)
        chars to wrap and indent; no effect, if 'usePrefix' is not set,
        by default False

    formatIndentWidth: int, optional
        format width, by default 80

    Returns
    -------
    None
    """
    print_(
        sys.stdout, 
        message, 
        newLine, 
        usePrefix, 
        formatIndent,
        formatIndentWidth
    )


def on_error(category: Rss2iError) -> None:
    """Outputs a string message to stderr

    Parameters
    ----------
    category : Rss2iError
        Outputs the string version of the given object.
        Output message. Will be prefixed by [Error] and timestamp
    """
    caller = inspect.stack()[1]
    # Only print a verbose error message, when the verbosity level is > 1
    verbosePrefix = ''
    if _verbosityLevel > 1:
        verbosePrefix = f'{caller.function}@{caller.lineno}: {repr(category)}: '

    timestamp = datetime.now().strftime('%H:%M:%S')
    print_err(f'\033[91m[Error] [{timestamp}] {verbosePrefix}{category}\033[00m')


def on_warn(category: Rss2iWarning) -> None:
    """Outputs a string message to stderr

    Parameters
    ----------
    category : Rss2iWarning
        Outputs the string version of the given object. 
        Will be prefixed by [WARNING] and timestamp
    """
    timestamp = datetime.now().strftime('%H:%M:%S')
    print_err(f'\033[93m[WARNING] [{timestamp}] {category}\033[00m')


def on_info(message: str) -> None:
    """Outputs a string message to stderr if at least -v is set (default)

    Parameters
    ----------
    message : str
        Output message. Will be prefixed by [INFO] and timestamp
    """
    if _verbosityLevel > 0:
        timestamp = datetime.now().strftime('%H:%M:%S')
        green = '\033[92m{}\033[00m'
        info_message = '[INFO] [{}] {}'.format(timestamp, message)
        print_err(green.format(info_message))


def on_debug(message: str, verbosity_threshold: int = 2) -> None:
    """Outputs a string message to stderr if at least -vv is set (default)

    Parameters
    ----------
    message : str
        Output message. Will be prefixed by [DEBUG] and timestamp

    verbosity_threshold : int, optional
        The threshold deciding whether the message will be outputted,
        by default 2
    """
    if _verbosityLevel >= verbosity_threshold:
        timestamp = datetime.now().strftime('%H:%M:%S')
        print_err(f'\033[96m[DEBUG] [{timestamp}] {message}\033[00m')


def print_progress(iteration: int, total: int, prefix: str = '',
                   suffix: str = '', barLength: int = 40, lineOffset: int = 0,
                   finishWithNewLine: bool = True, useMemory: bool = True,
                   startTime: datetime = None) -> None:
    """Outputs a progressbar

    The progressbar is calculated by a current value and a total value.

    Parameters
    ----------
    iteration : int
        Current value

    total : int
        Total/Maximum value

    prefix : str, optional
        Outputs before bar, by default ''

    suffix : str, optional
        Outputs after bar, by default ''

    barLength : int, optional
        Length of the progesss bar, by default 40

    lineOffset : int, optional
        Vertical, downward line offset, by default 0

    finishWithNewLine : bool, optional
        Outputs new line, when current reaches total, by default True

    useMemory : bool, optional
        Only outputs, if changes since last call, by default True

    startTime: datetime, optional
        Start time, by default None

    Returns
    -------
    None
    """
    progress = iteration / float(total)
    percents = '{: >3d}'.format(int(100 * (progress)))
    filledLength = int(round(barLength * progress))
    if percents == '100':
        bar = '=' * (filledLength)
    else:
        bar = '=' * (filledLength - 1) + (
            '>' if filledLength else '') + ' ' * (barLength - filledLength)

    time = ''
    if startTime:
        time_diff = datetime.now() - startTime
        time = 'elp:{: >9} eta:{: >9} '.format(
            _format_time(*_unpack_time(int(time_diff.total_seconds()))),
            _format_time(*_unpack_time(
                int(((1 - progress) / progress) * time_diff.total_seconds()))
            )
            if progress > 0.1 else '-')

    output = '\r{} |{}|{}% {}{} '.format(
        prefix,
        bar, 
        percents, 
        time, 
        suffix
    )

    if useMemory:
        global _lastOutput
        if _lastOutput == output:
            return
        else:
            _lastOutput = output

    for i in range(lineOffset):
        print_err('', usePrefix=False)

    print_err(
        output, 
        newLine=False,
        usePrefix=False
    )

    for i in range(lineOffset):
        print_err('\033[F', newLine=False, usePrefix=False)
        # FIXME: Ansi escape code must be enabled in powershell
        # https://stackoverflow.com/a/51681675
        # pwsh Set-ItemProperty HKCU:\Console VirtualTerminalLevel -Type DWORD 1

    if iteration == total and finishWithNewLine:
        print_err('', usePrefix=False)


# -----------------------------------------------------------------------------
# Config/Parse classes --------------------------------------------------------
class RegexPatterns:
    """Regex Patterns that are used to extract information from config files"""

    EmptyLinePattern = r'^\s*$'
    EmptyLineRegex = re.compile(EmptyLinePattern)

    FramePattern = (
        r'^\s*frame\s*(?P<frame>\d{1,})\s*:\s*dev\s*(?P<device>[\d]{1,})\s*'
        r'@\s*(?P<frequency>[\d]{1,})\s*MHz\s*(with\s*(?P<behaviour1>[\S]{1,})\s*'
        r' (?P<behaviour2>[\S]{1,})\s* (?P<pattern>[\S]{1,})\s*)?$'
    )
    FrameRegex = re.compile(FramePattern, re.IGNORECASE)

    ModePattern = (
        r'^\s*mode\s*:\s*(compatible\s*)?(?P<mode>[\S]{1,})\s*(?P<spec>[\S]{1,})\s*'
    )
    ModeRegex = re.compile(ModePattern, re.IGNORECASE)

    CompatPattern = (
        r'^\s*mode\s*:\s*compatible\s*(?P<mode>[\S]{1,})\s*(?P<spec>[\S]{1,})\s*'
    )
    CompatRegex = re.compile(CompatPattern, re.IGNORECASE)

    CteTypePattern = r'^\s*CTE\s*:\s*type\s*(?P<type>[\S]{1,})\s*$'
    CteTypeRegex = re.compile(CteTypePattern, re.IGNORECASE)

    CtePattern = (
        r'^\s*CTE\s*:\s*frame\s*(?P<frame>\d{1,})\s*->\s*(?P<mode>[\S]{1,})\s*'
        r'(with\s*(?P<behaviour1>[\S]{1,})\s* (?P<behaviour2>[\S]{1,})\s*'
        r' (?P<pattern>[\S]{1,})\s*)?$'
    )
    CteRegex = re.compile(CtePattern, re.IGNORECASE)

    SamplingPattern = (
		r'^\s*sampling\s*:\s*frame\s*(?P<frame>\d{1,})\s*->\s*'
		r'(?P<sampling>[\S]{1,})\s*$'
	)
    SamplingRegex = re.compile(SamplingPattern, re.IGNORECASE)

    StoragePattern = r'^\s*storage\s*:\s*(?P<storage>[\S]{1,})\s*$'
    StorageRegex = re.compile(StoragePattern, re.IGNORECASE)

    CyclesPattern = r'^\s*cycles\s*:\s*(?P<cycles>\d{1,})\s*$'
    CyclesRegex = re.compile(CyclesPattern, re.IGNORECASE)

    DeviceContextPattern = r'^\s*devs\s*{\s*(\d+)(\s*,\s*\d+\s*)*}\s*:\s*$'
    DeviceContextRegex = re.compile(DeviceContextPattern, re.IGNORECASE)

    TXPowerPattern = r'^\s*txpower\s*:\s*(?P<power>-?\d+)\s*dBm\s*$'
    TXPowerRegex = re.compile(TXPowerPattern, re.IGNORECASE)

    SyncMapPattern = (r'^\s*sync\s*:\s*frame\s*(?P<frame>\d+)\s*->\s*vdev\s*'
                      r'(?P<vdev>\d+)\s*$')
    SyncMapRegex = re.compile(SyncMapPattern, re.IGNORECASE)

    SyncThrPoorPattern = (r'^\s*sync\s*:\s*(?P<cycles>\d+)\s*cycles\s*\?\s*'
                          r'(?P<vdevs>\d+)\s*vdevs\s*$')
    SyncThrPoorRegex = re.compile(SyncThrPoorPattern, re.IGNORECASE)

    SyncThrOOSPattern = (r'^\s*sync\s*:\s*(?P<cycles>\d+)\s*cycles\s*\!\s*'
                         r'(?P<vdevs>\d+)\s*vdevs\s*$')
    SyncThrOOSRegex = re.compile(SyncThrOOSPattern, re.IGNORECASE)

    MemMapOutPattern = (r'^\s*out\s*:\s*frame\s*(?P<frame>\d+)\s*->\s*vmems\s*'
                        r'(?P<offset>\d+)\s*:\s*(?P<length>\d+)\s*$')
    MemMapOutRegex = re.compile(MemMapOutPattern, re.IGNORECASE)

    MemMapInPattern = (r'^\s*in\s*:\s*frame\s*(?P<frame>\d+)\s*->\s*vmem\s*'
                       r'(?P<offset>\d+)\s*$')
    MemMapInRegex = re.compile(MemMapInPattern, re.IGNORECASE)


class Line:
    def __init__(self, number, content):
        self.number = number
        self.content = content

    def get_number(self):
        return self.number

    def get_line(self):
        return self.content


class Radio(Line):
    def __init__(self, mode: str, compat: bool, lineNumber=0, lineContent='') -> None:
        Line.__init__(self, lineNumber, lineContent)
        self.mode = mode
        self.compat = compat

    def __str__(self) -> str:
        if self.compat:
            return str.format('compatible {}',self.mode)
        else:
            return self.mode


class Frame(Line):
    def __init__(self, index:int, device:int, frequency:int,
        cteBehaviour: str, pattern:str, lineNumber=0, lineContent='') -> None:
        Line.__init__(self, lineNumber, lineContent)
        self.index = index
        self.device = device
        self.frequency = frequency
        self.cteBehaviour = cteBehaviour
        if self.cteBehaviour:
            self.cteBehaviour = self.cteBehaviour.lower()
        self.pattern = pattern

    def __str__(self) -> str:
        if self.cteBehaviour == 'constant':
            return str.format('Frame {:0>2d} -> Device {:02d} @ {} Mhz with {} Antenna {}',
                self.index, self.device, self.frequency, self.cteBehaviour.title(), self.pattern)
        elif self.cteBehaviour == 'switching':
            return str.format('Frame {:0>2d} -> Device {:02d} @ {} Mhz with Antenna {} {}',
                self.index, self.device, self.frequency, self.cteBehaviour.title(), self.pattern.upper())
        else:
            return str.format('Frame {:0>2d} -> Device {:02d} @ {} Mhz',
                          self.index, self.device, self.frequency)


class Cte(Line):
    def __init__(self, index: int, mode: str, cteBehaviour: str,
        pattern: str, lineNumber=0, lineContent='') -> None:
        Line.__init__(self, lineNumber, lineContent)
        self.index = index
        self.mode = mode
        self.cteBehaviour = cteBehaviour
        self.pattern = pattern

    def __str__(self) -> str:
        if self.cteBehaviour.lower() == 'constant':
            return str.format('Frame {:0>2d} -> {} with {} Antenna {}',
                        self.index, self.mode, self.cteBehaviour.title(), self.pattern)
        elif self.cteBehaviour.lower() == 'switching':
            return str.format('Frame {:0>2d} -> {} with Antenna {} {}',
                        self.index, self.mode, self.cteBehaviour.title(), self.pattern.upper())


class Sampling(Line):
    def __init__(self, index: int, sampleBehaviour: str, lineNumber=0, lineContent='') -> None:
        Line.__init__(self, lineNumber, lineContent)
        self.index = index
        self.sampleBehaviour = sampleBehaviour

    def __str__(self) -> str:
        return str.format('Frame {:0>2d} −> {}',
                    self.index, self.sampleBehaviour)


class SyncMap(Line):
    def __init__(self, frameIndex, vdev, lineNumber=0, lineContent='') -> None:
        Line.__init__(self, lineNumber, lineContent)
        self.frameIndex = frameIndex
        self.vdev = vdev

    def __str__(self) -> str:
        return str.format('Frame {:0>2d} -> VDevice {:0>2d}', self.frameIndex,
                          self.vdev)


class SyncThreshold(Line):
    def __init__(self, cycles, vdevs, lineNumber=0, lineContent='') -> None:
        Line.__init__(self, lineNumber, lineContent)
        self.cycles = cycles
        self.vdevs = vdevs

    def __str__(self) -> str:
        return str.format('In {} cycles expecting {} VDevices', self.cycles,
                          self.vdevs)


class MemoryMap:
    def __init__(self, frameIndex: int, offset: int, length: int,
                 lineNumber: int = 0, lineContent: str = '') -> None:
        Line.__init__(self, lineNumber, lineContent)
        self.frameIndex = frameIndex
        self.offset = offset
        self.length = length

    def __str__(self) -> str:
        return str.format('Frame {:0>2d} @ {:0>2d} for {}', self.frameIndex,
                          self.offset, self.length)


class DeviceConfiguration:
    def __init__(self, 
        ID:int=0, 
        cycles:int=0, 
        power:int=0, 
        frames:list=[],
        ctes: list=[],
        samplings: list=[],
        storage:str='',
        syncMap:list =[],
        thrPoor:list=[], 
        thrOOS:list=[], 
        memMapOut:list=[], 
        memMapIn:list=[],
        radio:object=None,
        cteType:str=''
        ) -> None:
        self.ID = ID
        self.cycles = cycles
        self.power = power
        self.frames = frames
        self.ctes = ctes
        self.samplings = samplings
        self.storage = storage
        self.syncMap = syncMap
        self.thrPoor = thrPoor
        self.thrOOS = thrOOS
        self.memMapOut = memMapOut
        self.memMapIn = memMapIn
        self.radio = radio
        self.cteType = cteType

    def __str__(self) -> str:
        text = str.format(
            'Device ID:\t{:0>2d}\n'
            'RadioMode:\t{}\n'
            'Cycles   :\t{}\n'
            'TxPower  :\t{} dBm\n'
            'Frames   :\t{}\n'
            'CTE Type :\t{}\n'
            "CTE's    :\t{}\n"
            'Samplings:\t{}\n'
            'Storage  :\t{}\n'
            'Vdevices :\t{}\n'
            'Poor thr.:\t{}\n'
            'OOS thr. :\t{}\n'
            'MemOutMap:\t{}\n'
            'MemInMap :\t{}\n', 
            self.ID,
            self.radio,
            self.cycles, 
            self.power,
            '' if self.frames is None else '\n\t\t'.join(
                [str(f) for f in self.frames]
            ),
            self.cteType,
            '' if self.ctes is None else '\n\t\t'.join(
                [str(c) for c in self.ctes]
            ),
            '' if self.samplings is None else '\n\t\t'.join(
                [str(s) for s in self.samplings]
            ),
            self.storage,
            '' if self.syncMap is None else '\n\t\t'.join(
                [str(m) for m in self.syncMap]
            ),
            '' if self.thrPoor is None else '\n\t\t'.join(
                [str(t) for t in self.thrPoor]
            ),
            '' if self.thrOOS is None else '\n\t\t'.join(
                [str(t) for t in self.thrOOS]
            ),
            '' if self.memMapOut is None else '\n\t\t'.join(
                [str(m) for m in self.memMapOut]
            ), 
            '' if self.memMapIn is None else '\n\t\t'.join(
                [str(m) for m in self.memMapIn]
            )
        )
        return text


# -----------------------------------------------------------------------------
# Command handling ------------------------------------------------------------
# Cmd Download
def download(args):
    argDevices = args.device

    if args.file:
        config = _parse(args.file, argDevices)

    records = {
        str(devID): {
            'PageCount': None,
            'Address': None
        }
        for devID in argDevices
    }
    if hasattr(args, 'resume') and args.resume:
        with open(args.resume, 'r', encoding='utf-8') as f:
            records = json.loads(f.read())
        
        on_info('Running download in resume mode')

    # Split devices into local devices and over-the-air devices.
    localDevices = [ID for ID in argDevices if ID in _connectedDevices]
    remoteDevices = [ID for ID in argDevices if ID not in _connectedDevices]

    # ----------------------------------------------------------------
    # Parallel local download
    # Only parallelises the download of locally connected devices, if so
    # specified by '-s'/'-simultaneous' command and if more than one device is
    # locally connected.
    if args.simultaneous and len(localDevices) > 1:
        # Creates a tuple with the for a download required parameters.
        # The tuple as a single argument is a limitaion of 'Pool.map()'.
        localArgs = [
            (
                config[devID], 
                devID, 
                args.output,
                args.frameCount if hasattr(args, 'framecount') else None,
                _openSerialAddresses[devID], 
                i,
                args.pagecount if args.pagecount else 16,
                records[str(devID)]['PageCount'], 
                records[str(devID)]['Address']
            )
            for i, devID in enumerate(localDevices)
        ]

        # Parallel execution
        on_info('Closing device connections for simultaneous download')
        _close_device_connections()

        lock = multiprocessing.Value('i', 0)
        with multiprocessing.Pool(processes=len(localArgs),
                                  initializer=_init_shared_lock,
                                  initargs=(lock, )) as p:
            results = p.map(_p_download, localArgs)

        _handle_results(results)

        if remoteDevices:
            on_info('Re-opening device connections to download remote data')
            _establish_device_connections(args)

    # ----------------------------------------------------------------
    # Sequential local download
    # Runs if not explicitly simultaneous or there is only one locally
    # connected device.
    else:
        localArgs = [
            (
                config[devID], 
                devID, 
                args.output,
                args.frameCount if hasattr(args, 'framecount') else None, 
                None, 
                0,
                args.pagecount if args.pagecount else 255,
                records[str(devID)]['PageCount'], 
                records[str(devID)]['Address'],
                None, 
                args.csv    
            )
            for i, devID in enumerate(localDevices)
        ]

        for localArg in localArgs:
            result = _download(*localArg)

            _handle_results([result])

    # ----------------------------------------------------------------
    # Remote download (only sequentially)
    remoteArgs = [
        (
            config[devID], 
            devID, 
            args.output,
            args.frameCount if hasattr(args, 'framecount') else None, 
            None, 
            0,
            args.pagecount if args.pagecount else 255,
            records[str(devID)]['PageCount'], 
            records[str(devID)]['Address'],
            None, 
            args.csv
        )
        for i, devID in enumerate(remoteDevices)
    ]

    for remoteArg in remoteArgs:
        result = _download(*remoteArg)

        _handle_results([result])


def _handle_results(results) -> None:
    for result in results:
        success, devID, errorArgs = result
        if success:
            print_err('', newLine=True, usePrefix=False)
        else:
            print_err('Error{} while downloading from device {}'.format(
                's' if len(errorArgs) > 1 else '', result[1]))

        for errorArg in errorArgs:
            errorCode, args, kwargs = errorArg[0], errorArg[1], errorArg[2]
            on_error(errorCode, *args, **kwargs)


@debug_caller
def get_id(args) -> None:
    """Command <getid>.

    Prints the revision, antenna type and firmware of specifed devices or connected devices

    Parameters
    ----------
    args.device
        Only get info from specifed device(s), other connected devices are neglected

    Returns
    -------
    None
    """
    
    revisions = defaultdict(
                    lambda: "Rev. ?",
                    {
                        0x00: 'Rev. A',
                        0x01: 'Rev. B',
                        0x02: 'Rev. C',
                        0x03: 'Rev. D',
                        0x04: 'Rev. E',
                        0x05: 'Rev. F',
                        0x06: 'Rev. G',
                        0x07: 'Rev. H',
                        0x08: 'Rev. I',
                    }
                )
    antenna_type = defaultdict(
                       lambda: "unknown",
                       {
                           0x00: 'Onb.',
                           0x01: 'Ext.',
                           0x02: 'Arr.',
                       }
                   )
    
    fs = 'rss2i: getid: Dev. {:0>2d}: {}, Ant. {}, Firmware v{}.{}.{} @ {}'
    
    if hasattr(args, 'device') and args.device:
        devicesToShow = args.device
    else:
        devicesToShow = _connectedDevices
    
    print_err("getid: Listing connected devices:")
    
    for devID in _connectedDevices:
        
        _, response, _ = _query(
                             devID,
                             [0x00, devID],
                             10,
                             device=_connectedDevices[devID],
                             handleDefaultErrors=True
                         )
        print_out(
            fs.format(
                response[0],
                revisions[response[1]],
                antenna_type[response[2]],
                response[3],
                response[4],
                response[5],
                _openSerialAddresses[devID]
            )
        )
    
    if hasattr(args, 'device') and args.device:
    
        fs = 'rss2i: getid: Dev. {:0>2d}: {}, Ant. {}, Firmware v{}.{}.{}'
        
        print_err("getid: Listing specified devices:")
        
        for devID in args.device:
        
            status, response, _ = _query(
                                      devID,
                                      [0x00, devID],
                                      10,
                                      handleDefaultErrors=True
                             )
            
            if status:
                break
            
            print_out(
                fs.format(
                    response[0],
                    revisions[response[1]],
                    antenna_type[response[2]],
                    response[3],
                    response[4],
                    response[5]
                )
            )
    
# Cmd Ping
@debug_caller
def ping(args) -> None:
    for devID in args.device:
        print_err('Pinging device {}: '.format(devID), newLine=False)
        for i in range(5):
            time.sleep(0.5)
            print_err('*', newLine=False, usePrefix=False)
            dummyData = [0x02, devID, 0x01] + [0x00] * 7

            # 'Ping'
            for _ in range(10):
                resp_id, response, _ = _query(
                    devID, 
                    dummyData, 
                    9,
                    handleDefaultErrors=False,
                    suppressOTANotification=True
                )
                time.sleep(0.025)

        # new line
        print_err('', usePrefix=False)


# Cmd SetConfig
@debug_caller
def config(args) -> None:
    """Configures given devices with given config file

    Normally this config function calls _parse() to decode the config file and gets 
    DeviceConfiguration objects form _parse(), but optionally it can also accept
    DeviceConfiguration objects directly.

    Parameters
    ----------
    args.file : str
        A string that is the path to the config file

    args.device : list
        A list with integer id of deevices that are supposed to be configured

    or

    args : dict
        Dictionary that consists of DeviceConfiguration objects

    Returns
    -------
    None
    """
    
    if isinstance(args, dict) and all(
        isinstance(e, DeviceConfiguration) for e in args.values()):
        configs = args
    else:
        configs = _parse(args.file, args.device)

    boardinfos = {}

    for dev in configs:
        # Get BoardRev and BoardType
        status, response, _ = _query(
                                    dev,
                                    [0x00, dev],
                                    10,
                                    handleDefaultErrors=True,
                                    suppressOTANotification=True
                            )
        
        if status:
            raise Rss2iError("Retriving BoardRev and BoardType failed!")

        boardRev = response[1]
        boardType = response[2]

        boardinfos[dev] = (boardRev, boardType)

    for configKey in configs:
        try:
            config = configs[configKey]
            boardRev, boardType = boardinfos[configKey]

            # Preferences (txPower)---------------------------------------
            queryData = [0x00 for _ in range(10)]
            queryData[0] = 0x02
            queryData[1] = config.ID
            queryData[2] = rss2ipy.config.txpower(config.power).code()[0]
            status, _, _ = _query(
                devID=config.ID, 
                queryData=queryData, 
                responseLength=9
            )


            # Measurement configuration ----------------------------------
            # 256 query length
            queryData = [0x00 for _ in range(256)]
            queryData[0] = 0x10
            queryData[1] = config.ID
            queryData[2] = len(config.frames)
            # Convert cycles into bytes and reversing msb->lsb
            cyclesCode = rss2ipy.config.cycles(config.cycles).code()
            queryData[3] = cyclesCode[0]
            queryData[4] = cyclesCode[1]
            queryData[5] = cyclesCode[2]
            # Set frames
            frameData = []
            frequencyConfigs = []
            for f in config.frames:
                if config.ID == f.device:
                    frequencyConfig = rss2ipy.config.transmitter_frequency(f.frequency)
                else:
                    frequencyConfig = rss2ipy.config.receiver_frequency(f.frequency)

                frameData.append(frequencyConfig.code()[0])
                frequencyConfigs.append(frequencyConfig)
            
            queryData = [
                frameData[i - 6] if 5 < i < 5 + len(frameData) + 1 else queryData[i]
                for i in range(256)
            ]

            status, _, _ = _query(
                devID=config.ID, 
                queryData=queryData, 
                responseLength=0x01,
                suppressOTANotification=True
            )
            if status == 3:
                raise FrameCountOutOfRange(
                    device=config.ID, 
                    frameCount=len(config.frames)
                )
            elif status == 4:
                raise CycleCountOutOfRange(
                    device=config.ID, 
                    cycle=config.cycles
                )
            elif status >= 5:
                freq = queryData[status + 1]
                freq += 2300 if freq > 101 else 2400
                raise FrequencyOutOfRange(
                    device=config.ID,
                    frequency=freq
                )


            # Set Radio Mode ---------------------------------------------
            # 6 query length
            radioModeConfig = rss2ipy.config.radio_mode(
                mode=config.radio.mode,
                boardRev=boardRev
            )

            if radioModeConfig.valid():
                queryData = [0x00 for _ in range(6)]
                queryData[0] = 0x16
                queryData[1] = config.ID
                queryData[2] = radioModeConfig.code()[0]


                if config.radio.compat:
                    compatConfig = rss2ipy.config.activated_compatibility_mode()
                    queryData[3] = compatConfig.code()[0]
                else:
                    compatConfig = rss2ipy.config.default_compatibility_mode()
                    queryData[3] = compatConfig.code()[0]

                status, _, _ = _query(
                    devID=config.ID, 
                    queryData=queryData, 
                    responseLength=0x01, 
                    suppressOTANotification=True
                )
                if (status > 2):
                    raise RadioModeInvalid(mode=config.radio.mode)
            else:
                raise RadioModeInvalid(mode=config.radio.mode)


            # Set CTE configuration ----------------------------------
            # 253 query length
            queryData = [0x00 for _ in range(253)]
            queryData[0] = 0x17
            queryData[1] = config.ID
            # CTE type config object
            typeConfig = rss2ipy.config.cte_type(
                type=config.cteType,
                mode=radioModeConfig,
                compatibility=compatConfig
            )
            queryData[2] = typeConfig.code()[0]

            transmit = [cte for cte in config.ctes if cte.mode == 'transmit']
            receive = [cte for cte in config.ctes if cte.mode == 'receive']
            
            c = 0
            if transmit:
                for frame in config.frames:
                    cte = transmit[c]
                    if cte.index == frame.index:
                        frame.cteBehaviour = cte.cteBehaviour
                        frame.pattern = cte.pattern
                        if c < len(transmit) - 1:
                            c += 1
            
            transmitterCteConfigs = []
            for frame in config.frames:
                transboardRev, transboardType = boardinfos[frame.device]
                if frame.cteBehaviour == 'switching':
                    cteConfig = rss2ipy.config.cte_pattern(
                        pattern=frame.pattern,
                        type=typeConfig,
                        boardRev=transboardRev,
                        boardType=transboardType
                    )
                elif frame.cteBehaviour == 'constant':
                    cteConfig = rss2ipy.config.cte_antenna(
                        antenna=int(frame.pattern),
                        type=typeConfig,
                        boardRev=transboardRev,
                        boardType=transboardType
                    )
                else:
                    cteConfig = rss2ipy.config.default_cte_behaviour()

                if cteConfig.valid():
                    transmitterCteConfigs.append(
                        (
                        frame.device, 
                        frame.index,
                        frame.cteBehaviour,
                        cteConfig
                        )
                    )
                else:
                    raise ConfigDecodeError(cteConfig.reason())

            for device, frame, behaviour, transmitterConfig in transmitterCteConfigs:
                # current device is transmitting
                if config.ID == device: 
                    code = transmitterConfig.code()[0]

                # current device is receiving -> induced behaviour
                else:
                    if behaviour == 'switching':
                        code = rss2ipy.config.induced_cte_antenna(
                            type=typeConfig,
                            pattern=transmitterConfig,
                            boardRev=boardRev,
                            boardType=boardType
                        ).code()[0]
                    elif behaviour == 'constant':
                        code = rss2ipy.config.induced_cte_pattern(
                            type=typeConfig,
                            antenna=transmitterConfig,
                            boardRev=boardRev,
                            boardType=boardType
                        ).code()[0]
                    else:
                        code = rss2ipy.config.default_cte_behaviour().code()[0]

                    if receive:
                        for cte in receive:
                            if cte.index == frame:
                                if cte.cteBehaviour == 'switching':
                                    cteConfig = rss2ipy.config.cte_pattern(
                                        pattern=cte.pattern,
                                        type=typeConfig,
                                        boardRev=boardRev,
                                        boardType=boardType
                                    )
                                elif cte.cteBehaviour == 'constant':
                                    cteConfig = rss2ipy.config.cte_antenna(
                                        antenna=int(cte.pattern),
                                        type=typeConfig,
                                        boardRev=boardRev,
                                        boardType=boardType
                                    )
                                if cteConfig.valid():
                                    code = cteConfig.code()[0]
                                else:
                                    raise ConfigDecodeError(cteConfig.reason())
                                

                queryData[3 + frame] = code

            
            status, data, _ = _query(
                devID=config.ID, 
                queryData=queryData, 
                responseLength=0x02,
                suppressOTANotification=True
            )
            ptr = data[0]

            if status == 3:
                raise CteTypeInvalid(cteType=config.cteType)
            elif status == 4:
                raise CteConfigIncompatibleWithMeasureConfig(frame=ptr)        
            elif status == 5:
                raise CteBehaviourInvalid(
                    frame=ptr,
                    cteType=config.cteType,
                    device=config.ID
                )
            elif status >= 6:
                raise CteIncompatibleWithRadio(radioMode=config.radio.mode)


            # Set sample configuration ----------------------------------
            # 253 query length
            queryData = [0x00 for _ in range(253)]
            queryData[0] = 0x18
            queryData[1] = config.ID

            storageConfig = rss2ipy.config.storage(config.storage)
            if storageConfig.valid():
                queryData[2] = storageConfig.code()[0]

            if not (len(frequencyConfigs) == len(transmitterCteConfigs)):
                raise Rss2iError("Something went wrong! number of measurment frames and cte frames don't match")
            
            for i in range(len(frequencyConfigs)):
                _, frame, _, behaviourConfig = transmitterCteConfigs[i]
                frequencyConfig = frequencyConfigs[i]
                
                code = rss2ipy.config.default_sampling(
                    frequency=frequencyConfig,
                    mode=radioModeConfig,
                    compatibility=compatConfig,
                    type=typeConfig,
                    behaviour=behaviourConfig
                ).code()[0]
                
                queryData[3 + frame] = code


                for sample in config.samplings:
                    _, _, _, behaviourConfig = transmitterCteConfigs[sample.index]
                    sampleConfig = rss2ipy.config.sampling(
                        sample=sample.sampleBehaviour,
                        frequency=frequencyConfigs[sample.index],
                        mode=radioModeConfig,
                        compatibility=compatConfig,
                        type=typeConfig,
                        behaviour=behaviourConfig
                    )
                    if sampleConfig.valid():
                        queryData[3 + sample.index] = sampleConfig.code()[0]
                    else:
                        raise ConfigDecodeError(sampleConfig.reason())


            status, data, _ = _query(
                devID=config.ID, 
                queryData=queryData, 
                responseLength=0x02,
                suppressOTANotification=True
            )
            ptr = data[0]

            if status == 3:
                raise SampleConfigIncompatibleWithMeasureConfig(frame=ptr) 
            elif status == 4:
                raise SampleTypeInvalid(frame=ptr)
            elif status >= 5:
                raise SampleTypeIncompatibleWithConfig(
                    frame=ptr,
                    radioMode=config.radio.mode,
                    cteType=config.cteType
                )

            # Synchronisation map ----------------------------------------
            # 253 query length
            queryData = [0x00 for _ in range(253)]
            queryData[0] = 0x11
            queryData[1] = config.ID
            queryData[2] = max([m.vdev for m in config.syncMap]) + 1
            syncData = [m.vdev for m in config.syncMap]
            queryData = [
                syncData[i - 3] if 2 < i < 3 + len(syncData) else queryData[i]
                for i in range(253)
            ]

            status, _, _ = _query(
                devID=config.ID, 
                queryData=queryData, 
                responseLength=1,
                suppressOTANotification=True
            )
            if status == 3:
                raise VDevCountOutOfRange(
                    device=config.ID, 
                    vdevCount=max([m.vdev for m in config.syncMap])
                )
            elif status > 3:
                raise VDevIdOutOfRange(
                    device=config.ID, 
                    vdevId=queryData[status - 1]
                )


            # Poor synchronisation threshold -----------------------------
            # 252 query length
            queryData = [0x00 for _ in range(252)]
            queryData[0] = 0x12
            queryData[1] = config.ID
            for thr in config.thrPoor:
                queryData[thr.vdevs + 1] = round(thr.cycles / 10)

            status, _, _ = _query(
                devID=config.ID, 
                queryData=queryData, 
                responseLength=1,
                suppressOTANotification=True
            )
            # No extra error codes


            # Out of synchronisation threshold ---------------------------
            # 252 query length
            queryData = [0x00 for _ in range(252)]
            queryData[0] = 0x13
            queryData[1] = config.ID
            for thr in config.thrOOS:
                queryData[thr.vdevs + 1] = round(thr.cycles / 10)

            status, _, _ = _query(
                devID=config.ID, 
                queryData=queryData, 
                responseLength=1,
                suppressOTANotification=True
            )


            # Remote input map -------------------------------------------
            # 252 query length
            queryData = [0x00 for _ in range(252)]
            queryData[0] = 0x14
            queryData[1] = config.ID
            for m in config.memMapIn:
                queryData[m.frameIndex + 2] = m.offset

            status, _, _ = _query(
                devID=config.ID, 
                queryData=queryData, 
                responseLength=1,
                suppressOTANotification=True
            )
            
            if status > 2:
                raise VMemIdOutOfRange(
                    device=config.ID, 
                    vmemId=queryData[status - 1]
                )


            # Remote output map ------------------------------------------
            # 253 query length
            queryData = [0x00 for _ in range(253)]
            queryData[0] = 0x15
            queryData[1] = config.ID
            # Length of memory range
            queryData[2] = max([m.length for m in config.memMapOut], default=0)
            for m in config.memMapOut:
                queryData[m.frameIndex + 3] = m.offset

            status, _, _ = _query(
                devID=config.ID, 
                queryData=queryData, 
                responseLength=1,
                suppressOTANotification=True
            )

            print_err('Configuration of device {} successful'.format(config.ID))

        except Rss2iException as fail:
            on_error(fail)
            raise ConfigurationFailed(device=config.ID)


# Cmd Record
@debug_caller
def record(args) -> None:
    records = {}
    for devID in args.device:
        status, response, _ = _query(devID, [0x03, devID], 7)
        records[devID] = {}
        records[devID]['PageCount'] = (response[2] << 16) + (
            response[1] << 8) + response[0]
        records[devID]['Address'] = (response[5] << 16) + (
            response[4] << 8) + response[3]

    output_path = _output(args.output) + 'rec.json'
    with open(output_path, 'w', encoding='utf-8') as f:
        f.write(json.dumps(records, indent=4))


@debug_caller
def restore(args) -> None:
    """
    Restores the measurement results from the on-board flash memory. In contrast
    to the `download` subcommand, the session data is not read from the devices
    but has to be specified on the command line or has to be specified within a
    so-called record file, see `record` subcommand for the later case. This may
    be useful if the session data isn't available on the device for some reason.

    Parameters
    ----------
    args.device : list
        
    args.resume

    args.file

    args.output

    args.pagecount

    args.gap

    Returns
    -------
    None
    """
    
    for devID in [int(d) for d in args.device]:
        if args.resume:
            with open(args.resume, 'r', encoding='utf-8') as f:
                records = json.loads(f.read())
            address = records[str(devID)]['Address']
            amount = records[str(devID)]['PageCount']
            on_info(
                "Start restoring {} pages of device {} from addr {}.".format(
                    amount,
                    devID,
                    address
                )
            )
        else:
            address = int(args.address, 16)
            amount = 0
            on_info(
                "Start restoring flash pages of device {} from addr {}.".format(
                    devID,
                    address
                )
            )
        
        result = _download(
            _parse(args.file, [devID])[devID],
            devID,
            args.output,
            None,
            None,
            0,
            args.pagecount if args.pagecount else 255,
            amount,
            address,
            gap=args.gap
        )
        
        _handle_results([result])


# Cmd Parse
@debug_caller
def parse(args) -> None:
    """Calls _parse() and prints the returned DeviceConfiguration objects

    Parameters
    ----------
    args.device : list
        A list with integer id of devices that are supposed to be parsed

    args.file : str
        A string that is the path to the config file

    Returns
    -------
    None
    """
    configs = _parse(args.file, args.device)

    print_out('\n'+'\n'.join([str(configs[c]) for c in configs]))


# Cmd Start
@debug_caller
def start_measurement(args) -> None:
    startAddresses = {}
    argDevices = args.device

    # Set ota devices as 'jump on'
    for devID in [d for d in argDevices if d not in _connectedDevices]:
        status, address, multipleRetries = _query(devID, [0x0D, devID], 4)
        
        if status == 3:
            raise MissingMeasurementConfiguration(devID=devID)
        if status == 4 and multipleRetries:
            on_warn(DeviceWasAlreadyMeasuring(devID=devID))

        address = (address[2] << 16) + (address[1] << 8) + address[0]
        print_err(
            'Device {} is ready (Memory address: {:06X})'.format(
                devID, 
                address
            )
        )
        startAddresses[devID] = '{:0>6X}'.format(address)
    # Set local devices as 'jump on'
    for devID in [d for d in argDevices if d in _connectedDevices]:
        status, address, _ = _query(devID, [0x0D, devID], 4)
        
        if status == 3:
            raise MissingMeasurementConfiguration(devID=devID)

        address = (address[2] << 16) + (address[1] << 8) + address[0]
        print_err(
            'Device {} is ready (Memory address: {:06X})'.format(
                devID, 
                address
            )
        )
        startAddresses[devID] = '{:0>6X}'.format(address)

    print_err('Probing for connected device with active frame...',
              newLine=False)
    connectedDevID = 0
    
    # Wait until the "Jump on" is finished (see API doc).
    time.sleep(0.1);
    
    # Probing for kick off device
    for devID, dev in {
        k: v
        for k, v in _connectedDevices.items()
        if k in argDevices
    }.items():

        # Probe for ability to kick off the measurement
        # Enter control mode
        status, _, _ = _query(devID, [0x0B], 1, device=dev)

        # Start
        status, address, _ = _query(devID, [0x0E, devID], 4, device=dev)
        if status == 0:
            connectedDevID = devID
            break
        elif status == 3:
            raise MissingMeasurementConfiguration(devID=devID)
        else:
            # status == 4: Impossible with the current measurement configuration
            # Jump on again
            status, _, _ = _query(devID, [0x0D, devID], 4, device=dev)
            
            # Wait until the "Jump on" is finished (see API doc).
            time.sleep(0.1);
            continue
    
    
    if connectedDevID == 0:
        print_err('failed', usePrefix=False)
        raise NoDeviceAbleToStartMeasurement()
    else:
        address = (address[2] << 16) + (address[1] << 8) + address[0]
        print_err('ok', usePrefix=False)
        print_err(
            'Device {} has kicked off the measurement (Memory address: {:06X})'
            .format(connectedDevID, address)
        )
        startAddresses[connectedDevID] = '{:0>6X}'.format(address)

        #file output
        output_path = _output(args.output) + 'mem_addr.json'

        on_info('Saving start addresses @ {}'.format(output_path))

        with open(output_path, 'w', encoding='utf-8') as f:
            f.write(json.dumps(startAddresses, indent=4))


# Cmd Abort
@debug_caller
def abort(args, retry=True) -> None:
    abortedSuccessfully = False
    
    for devID in _connectedDevices:
        print_err(
            'Stopping measurement through device #{} as proxy...'.format(devID), 
            newLine=False
        )

        status, _, _ = _query(devID, [0x0C], 1, handleDefaultErrors=False)
        if status == 0:
            print_err('ok', usePrefix=False)
            abortedSuccessfully = True
            break
        elif status == 2:
            print_err('ok', usePrefix=False)
            on_warn(DevicesStillMeasuring())
            abortedSuccessfully = True
            break
        elif status == 1:
            print_err('failed', usePrefix=False)

            # If given a config file, try again after a reconfiguration
            if not args.file or not retry:
                continue

            print_err('Reconfiguring with given config file')

            config(_parse(args.file, [devID]))
            # Jump on and wait for sync
            _query(devID, [0x0D, devID], 4)
            time.sleep(1)
            # Stop anew
            abort(args, retry=False)
            return

    if not abortedSuccessfully:
        raise NoConnectedDeviceAbleToAbort()


@debug_caller
def monitor(args) -> None:
    if args.device[0] not in _connectedDevices:
        raise NoDevicesFound(device=args.device)

    deviceID = args.device[0]
    device = _connectedDevices[deviceID]
    # Flatten to keep compatibility with _int_list_parser_helper().
    # This allows -n 1,2,3 as well as -n 1 -n 2 -n 4
    frames = [frame for sublist in args.frames for frame in sublist]

    # 0: frame, 1-3: cycle msb...lsb, 4: rec. int, 5: rec. frac
    configData = [0x00, 0x03, 0x02, 0x01] + list(range(0x04, 0x16))

    # Setup variables and pyqt-window --------------------------------
    # Try import of display libs
    try:
        from PyQt5 import QtWidgets, QtGui
        from PyQt5.QtCore import pyqtSignal, pyqtSlot
        from pyqtgraph import GraphicsWindow, setConfigOption, mkPen
        from pyqtgraph.Qt import QtCore
    except ImportError as err:
        raise MonitorSetupFailed(message=err.msg)

    # Define window
    class MonitorWindow(GraphicsWindow):
        dataUpdateEvent = pyqtSignal(object)
        keyPressedEvent = pyqtSignal(object)

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

            self.setWindowTitle('monitor')
            self.resize(1200, 600)

            # Calculate subplotgrid
            root = len(frames)**(1 / 2)
            rootCeiling = math.ceil(root)
            nrows = rootCeiling
            ncols = rootCeiling
            if rootCeiling**2 - len(frames) >= rootCeiling:
                nrows += -1

            self.plot = {
                frame: self.addPlot(row=int(i / ncols), col=i % ncols)
                for i, frame in enumerate(frames)
            }

            for plot in self.plot.values():
                # plot.showGrid(x=True, y=True)
                plot.disableAutoRange()
                xAxis = plot.getAxis('bottom')

            # Set curves
            self.curves = {
                frameIndex: self.plot[frameIndex].plot(
                    (range(viewPortMemory * -viewPortWidth - 2, -1)),
                    [ppAmpOffset] * (viewPortMemory * viewPortWidth + 1),
                    pen=mkPen(color='b', width=2)
                )
                for frameIndex in frames
            }

            self.dataUpdateEvent.connect(self.update_data)
            self.keyPressedEvent.connect(self.on_key_press)
            self.threadKill = threading.Event()
            self.threadPause = threading.Event()
            self.threadPause.set()

            # Start data collection
            self.livelog = threading.Thread(
                target=_livelog,
                args=(deviceID, device, configData, self.threadKill)
            )
            self.livelog.start()

            # Start data update
            self.updater = threading.Thread(
                target=_read_queue, args=(
                    self.dataUpdateEvent.emit,
                    self.threadKill, 
                    self.threadPause
                )
            )
            self.updater.start()

        def closeEvent(self, close_event):
            self.threadKill.set()
            self.threadPause.set()
            self.livelog.join()

        @pyqtSlot(object)
        def update_data(self, data: Dict[int, Dict[str, List[float]]]):

            for frame, item in data.items():
                if frame not in self.curves:
                    continue

                xs = item['cycle']
                ys = item['value']
                oldXs, oldYs = self.curves[frame].getData()

                keepLen = len(oldXs) - len(xs)
                newLen = len(xs)
                newXs: np.ndarray = np.empty_like(oldXs, dtype=float)
                newXs[:keepLen] = oldXs[newLen:]
                newXs[keepLen:] = xs
                newYs: np.ndarray = np.empty_like(oldYs, dtype=float)
                newYs[:keepLen] = oldYs[newLen:]
                newYs[keepLen:] = ys

                self.plot[frame].setXRange(newXs[-1] - viewPortWidth,
                                           newXs[-1], update=False)
                self.plot[frame].setYRange(-70, -15, update=False)

                self.curves[frame].setData(newXs, newYs)

        def keyPressEvent(self, ev):
            self.scene().keyPressEvent(ev)
            self.keyPressedEvent.emit(ev)

        @pyqtSlot(object)
        def on_key_press(self, keyEventData):
            if keyEventData.key() == 80:
                if self.threadPause.is_set():
                    self.setWindowTitle('[PAUSED] - monitor')
                    self.threadPause.clear()
                else:
                    self.setWindowTitle('monitor')
                    self.threadPause.set()

    # Launch window --------------------------------------------------
    setConfigOption('background', 'w')
    setConfigOption('foreground', 'k')
    app = QtWidgets.QApplication([__file__])
    w = MonitorWindow()
    w.show()

    sys.exit(app.exec_())

    global _abort
    _abort = True


@debug_caller
def debug(args) -> None:
    """Outputs debug info with given device id into stderr
    
    Lists the output of the `0x19` *Return device status* and
    `0x1a` *Report error information* commands for every specified device.

    Parameters
    ----------
    args.device : list
        A list with integer id of devices from wich to get debug information

    Returns
    -------
    None
    """
    
    for devID in args.device:
        status, response, _ = _query(devID, [0x19, devID], 12)

        if status == 0:
            numberOfErrors = response[0] + 256 * response[1]
            print_err(
                'debug: dev. {}: Number of errors:{}'.format(
                    devID,
                    str(numberOfErrors).rjust(7)
                )
            )
            print_err(
                'debug: dev. {}: Main loop cycles:{}'.format(
                    devID,
                    str(response[2] + 256 * response[3]).rjust(7)
                )
            )
            print_err(
                'debug: dev. {}: Raw battery value:{}'.format(
                    devID,
                    str(response[4] + 256 * response[5]).rjust(6)
                )
            )
            print_err(
                'debug: dev. {}: Battery level / %:{}'.format(
                    devID,
                    str(response[6]).rjust(6)
                )
            )
            print_err(
                'debug: dev. {}: Free memory / %:{}'.format(
                    devID,
                    str(100 - response[7]).rjust(8)
                )
            )

            if numberOfErrors > 0:
                print_err('debug: dev. {}: Occured error cases:'.format(devID))
                
                overallStatus = 0;
                numOfErrorCases = rss2ipy.debug_get_num_of_error_cases()
                pages = math.ceil(
                            numOfErrorCases
                            / rss2ipy.debug_get_num_of_error_cases_per_page()
                        )
                errorIDs = []
                for page in range(0, pages):
                
                    status, res, _ = _query(
                        devID,
                        [0x1a, devID, page],
                        255,
                        suppressOTANotification=True
                    )
                    errorIDs.extend(res[2:])
                    overallStatus += status

                if overallStatus == 0:
                    print_err('')
                    for ec in range(0, numOfErrorCases):
                    
                        specificID = errorIDs[0 + 4 * ec] + 256 * errorIDs[1 + 4 * ec]
                        globalID = errorIDs[2 + 4 * ec] + 256 * errorIDs[3 + 4 * ec]
                        
                        if specificID > 0:
                            print_err(
                                "	[0x{:02x} 0x{:04x}/0x{:04x}] {}".format(
                                    ec,
                                    specificID,
                                    globalID,
                                    rss2ipy.debug_message(ec)
                                )
                            )

                    print_err('')
            else:
                print_err('debug: dev. {}: No errors occured.'.format(devID))


@debug_caller
def log(args) -> None:
    if args.device[0] not in _connectedDevices:
        raise NoDevicesFound(device=args.device)

    devID = args.device[0]
    device = _connectedDevices[devID]

    # preset: cycle + all remote values + all battery/memory + RSS sample amount
    configData = [0x00, 0x03, 0x02, 0x01] + list(range(0x04, 0x17))

    # Start reading the livelog
    threadKill = threading.Event()
    thread = threading.Thread(
        target=_livelog,
        args=(devID, device, configData, threadKill))
    thread.start()

    outputFormat = (
        'c:{:0>8d} f:{:0>3d} l:{:0>6.2f} a:{:0>3d} '
        'r1:-{:0>5.2f} r2:-{:0>5.2f} r3:-{:0>5.2f} '
        'r4:-{:0>5.2f} r5:-{:0>5.2f} r6:-{:0>5.2f} '
        'b:{:0>2.0f} m:{:0>2.0f} rb:{:0>2.0f} rm:{:0>2.0f}'
    )
    while not _abort:
        try:
            ans = _queue.get()
            frame = ans[0]
            cycle = (ans[1] << 16) + (ans[2] << 8) + ans[3]
            rssValues = [
                (ans[4] + ans[5] / (1 << 7)),
                (ans[6] + ans[7] / (1 << 7)),
                (ans[8] + ans[9] / (1 << 7)),
                (ans[10] + ans[11] / (1 << 7)),
                (ans[12] + ans[13] / (1 << 7)),
                (ans[14] + ans[15] / (1 << 7)),
                (ans[16] + ans[17] / (1 << 7))
            ]
            batteryMemory = [
                100 - ans[18], 
                100 - ans[19], 
                ans[20], 
                ans[21]
            ]
            amount = ans[22]
            output = outputFormat.format(
                cycle,
                frame,
                rssValues[0],
                amount,
                rssValues[1],
                rssValues[2],
                rssValues[3],
                rssValues[4],
                rssValues[5],
                rssValues[6],
                batteryMemory[2],
                batteryMemory[0],
                batteryMemory[3],
                batteryMemory[1],
            )
            print_out(output)
        except (KeyboardInterrupt, IOError, BrokenPipeError):
            _livelog_event.set()
            print_err('Signal to abort')
            thread.join()
            break


@debug_caller
def measure(args) -> None:
    on_info('Configure devices #{}'.format(args.device))
    config(args)

    on_info('Start measurement')
    start_measurement(args)

    # Handle the connection interrupt
    if args.interrupt and _connectedDevices.get(args.interrupt, False):
        on_info('Disconnect device #{}'.format(args.interrupt))
        device = _connectedDevices[args.interrupt]
        device.close()
        del _connectedDevices[args.interrupt]
        args.device = [ID for ID in args.device if not ID == args.interrupt]

    # Wait given delay before stopping the measurement
    print_err('------')
    print_err('Delay:')
    for i in range(1, args.tdelay * 10 + 1):
        hr, mn, sc = _unpack_time(int(i / 10))
        print_progress(
            i, 
            args.tdelay * 10,
            'Waiting {}s before stopping:'.format(args.tdelay),
            '{: >9}'.format(_format_time(hr, mn, sc)), 
            barLength=30,
        )
        time.sleep(0.1)

    on_info('Stop measurement')
    abort(args)

    on_info('Begin download')
    download(args)


# -----------------------------------------------------------------------------
# Internal functions ----------------------------------------------------------
def _unpack_time(seconds: int) -> Tuple[int, int, int]:
    """converts seconds into hours, minutes and seconds

    Parameters
    ----------
    seconds : int
        Time in seconds

    Returns
    -------
    hr, mn, sc : Tuple[int, int, int]
    """
    hr = int(seconds / 60 / 60)
    mn = int(seconds / 60)
    sc = seconds - (hr * 60 * 60 + mn * 60)

    return hr, mn, sc


def _format_time(houres: int, minutes: int, seconds: int) -> str:
    """formats given hours, minutes and seconds into string in hh:mm:ss format

    Parameters
    ----------
    houres : int
    minutes : int
    seconds : int

    Returns
    -------
    tdFormatted : str
    """
    tdFormatted = ''
    if houres > 0:
        tdFormatted += '{:0>2}h'.format(houres)
    if minutes > 0:
        tdFormatted += '{:0>2}m'.format(minutes)
    tdFormatted += '{:0>2}s'.format(seconds)

    return tdFormatted


@debug_caller
def _establish_device_connections(args: argparse.Namespace) -> None:
    """
    Establishes connections to the specified (if `args.port` is provided) or
    to all connected RSS2I devices (if it isn't).

    Argument List
    =============
    
    `args`                     All specified command line arguments.

    Modified Global Variables
    =========================

    `_connectedDevices`        Dictionary of connected devices.
    `_openSerialAddresses`     Dictionary of those serial ports the devices are
                                connected to.
      `_proxyPreferences`        Dictionary of so-called proxies.
    """
    
    def _is_timeout_lax(option):
        """Whether the serial timeout should be lax or not."""
        
        funcList = [download, abort, restore, measure]
        isTimeoutLax = option in funcList
        return isTimeoutLax
    
    
    if args.port:
        
        try:
            device = rss2ipy.Device(
                args.port,
                laxTimeout=_is_timeout_lax(args.func)
            )
            status, data, _ = _query(
                '?',
                [0x00, 0xff],
                10,
                device=device,
                handleDefaultErrors=True
            )
            devID = data[0];
            _connectedDevices[devID] = device
            _openSerialAddresses[devID] = args.port
            
        except (
            rss2ipy.SerialError, 
            rss2ipy.SerialTimeout, 
            rss2ipy.CRCError
        ):
            pass
        
    else:
        on_info('Detecting connected devices... ')
        
        deviceBlacklist = [r'/dev/ttyS[\d]+', r'COM1(^\d|$)']
        deviceBlacklistRegex = [re.compile(b) for b in deviceBlacklist]
        
        ports = [
            p.device for p in
                serial.tools.list_ports.comports()
                if not any(r.match(p.device) for r in deviceBlacklistRegex)
                    and p.vid == 0x0403 # FTDI
                    and p.pid == 0x6015 # UART USB Bridge
        ]
        
        for p in ports:
            
            try:
                
                device = rss2ipy.Device(
                    p,
                    laxTimeout=_is_timeout_lax(args.func)
                )
                status, data, _ = _query(
                    '?',
                    [0x00, 0xff],
                    10,
                    device=device,
                    handleDefaultErrors=True
                )
                devID = data[0]
                _connectedDevices[devID] = device
                _openSerialAddresses[devID] = p
                
            except (
                rss2ipy.SerialError,
                rss2ipy.SerialTimeout,
                rss2ipy.CRCError
            ):
                pass
    
    # Output short list of found devices.
    print_err([x for x in _connectedDevices.keys()], usePrefix=False)
    
    if len(_connectedDevices) == 0:
        
        # No devices found.
        if args.port:
            info = [args.port]
        else:
            info = ['automatic device detection']
        
        raise NoDevicesFound(device=info)
    else:
        on_info(f'Connected device IDs: {list(_connectedDevices.keys())}')

    # Parsing proxy preferences
    if args.proxy:
    
        for proxy in   args.proxy if not type(args.proxy) == str \
                  else [args.proxy]:
            
            # Determine the targeted device from the specified proxy rule,
            # i.e. TARGET:proxy1[,proxy2[...]]
            
            firstSplit = proxy.split(':')
            
            if not len(firstSplit) == 2:
                raise ProxyArgumentSyntaxError(proxy=proxy)
                
            target = firstSplit[0]
            
            try:
                targetInt = int(target)
            except ValueError:
                raise ProxyArgumentSyntaxError(proxy=proxy)

            # Determine the proxy devices from the specified proxy rule,
            # i.e. target:PROXY1[,PROXY2[...]]
            
            secondSplit = firstSplit[1].split(',')
            
            if len(secondSplit) == 0:
                raise ProxyArgumentSyntaxError(proxy=proxy)
                
            try:
                proxiesInt = [int(p) for p in secondSplit]
            except ValueError:
                raise ProxyArgumentSyntaxError(proxy=proxy)
            
            _proxyPreferences[targetInt] = proxiesInt


@debug_caller
def _close_device_connections() -> None:
    if len(_connectedDevices) == 0:
        return

    on_info('Disconnecting devices...')

    for devID in _connectedDevices:
        _connectedDevices[devID].close()
    _connectedDevices.clear()
    
    on_info('Disconnecting devices ok')


def _init_shared_lock(lock: multiprocessing.Value)  -> None:
    global _outputLock
    _outputLock = lock


def _output(prefix: str = None) -> str:
    if prefix:
        path = re.split(r'/|\\', prefix)[:-1]
        if path:
            os.makedirs(os.path.join(*path), exist_ok=True)
    else:
        prefix = datetime.now().strftime('%Y-%m-%d_%H-%M-%S_')
    return prefix


def _p_download(args):
    return _download(*args)


@debug_caller
def _download(config, devID, argOutput, argFrameCount, port, lineOffset,
              argPageCount, argAmount, argAddress, gap=None, 
              csv=False) -> Tuple[bool, int, List[Tuple[str, bool, Any]]]:#str: ErrorCode-> old Error System

    # Init -----------------------------------------------------------
    # Prepare error handling, when running in parallel processes
    global on_error
    errorFunc = on_error
    errorList = []
    numWarnings = 0
    numErrors = 0

    def error_catcher(*args, **kwargs):
        nonlocal numWarnings
        nonlocal numErrors
        nonlocal errorFunc
        nonlocal errorList
        nonlocal devID

        
        
        errorCode = next(filter(lambda x: isinstance(x, ErrorCode), args))
        if str(errorCode).startswith('ERR_'):
            numErrors += 1
        else:
            numWarnings += 1

        errorList.append((
            errorCode, 
            filter(lambda x: not isinstance(x, ErrorCode), args), 
            kwargs
        ))

        if str(errorCode).startswith('ERR_'):
            if __name__ == '__main__':
                errorFunc(errorCode, *args, **kwargs)
            else:
                return (False, devID, errorList)

    on_error = error_catcher

    try:
        # Parameter handling -----------------------------------------
        # Open conncetion, if simultaneous download
        device = None
        if port and isinstance(port, str):
            device = rss2ipy.Device(port, laxTimeout=True)

        frameCount = 0
        if argFrameCount:
            frameCount = argFrameCount
        else:
            frameCount = len(config.frames)

        # gap not None -> restore
        restoring = False
        if gap:
            restoring = True
            amount = 2**16 - 1
            address = argAddress

        # Address calculation ----------------------------------------
        # Number of last measurement's written flash pages -> amount
        elif argAmount or restoring:
            amount = argAmount
            address = argAddress
        else:
            status, response, _ = _query(
                devID, 
                [0x03, devID], 
                7,
                device=device
            )
            if status > 0:
                return (False, devID, errorList)

            amount = (response[2] << 16) + (response[1] << 8) + response[0]
            address = (response[5] << 16) + (response[4] << 8) + response[3]

        # Progress output setup --------------------------------------
        queriedKB = 0
        kBPerSecond = 0
        totalDataKB = amount * 253 / 1024
        eta = 0

        start = datetime.now()
        memory = []
        k = 0

        # Restore setup 2 / old data work-around ---------------------
        # Restore variables
        restore_last_cycle = 0  # comparison with next cycle + gap
        restore_expected_error_occurred = False
        
        while k < amount:
            # Data transfer from device ------------------------------
            pageCount = argPageCount if amount - k >= argPageCount else amount - k

            # Convert the absAddress (k) to a bytes array as int (lsb->msb)
            absAddress = (address + k).to_bytes(3, 'big')[::-1]
            absAddress = [absAddress[i] for i in range(len(absAddress))]
            
            try:
                queryResults, _ = _query(
                    devID, 
                    [0x07, devID] + absAddress,
                    253, 
                    device=device,
                    suppressOTANotification=True,
                    iterations=pageCount,
                    max_baudrate=True
                )
            except rss2ipy.SerialTimeout:
                raise DownloadTimeout(devID=devID)
                #FIXME: why is here a return statement?
                return (False, devID, errorList)
            
            # Error handling -----------------------------------------
            # Device errors
            
            for j, queryResult in enumerate(queryResults):
                status, response = queryResult[0], queryResult[1:]
                
                if status == 1:
                    # OTA timeout and all retries are exhausted
                    raise OverTheAirTimeout(devID=devID)
                if status == 3:
                    raise FlashMemoryBusy(devID=devID)
                if status == 4:
                    # handle 'expected' error when restoring
                    if restoring:
                        restore_expected_error_occurred = True
                        relAddress = absAddress[0] + (absAddress[1] << 8) + (
                            absAddress[2] << 16) + j
                        print_err('', usePrefix=False)
                        print_err(
                            'restore: Expected error while restoring @{}|0x{:X}'
                            .format(relAddress, relAddress), newLine=False)
                        break
                    else:
                        raise FlashMemoryCorruption(devID=devID)
                
                
                # Data parsing ---------------------------------------
                # Append single Page to memory list
                memory.append(response)

                # Exit middle loop
                if restore_expected_error_occurred:
                    break

                # Progress update ------------------------------------
                timespan = datetime.now() - start
                elapsedHours, elapsedMinutes, elapsedSeconds = _unpack_time(
                    timespan.seconds)

                timespanFormatted = _format_time(
                    elapsedHours, 
                    elapsedMinutes,
                    elapsedSeconds
                )

                queriedKB += 253 / 1024
                queriedMB = queriedKB / 1024
                if queriedMB > 1:
                    queriedDataFormatted = '{: >4.1f}MB'.format(queriedMB)
                else:
                    queriedDataFormatted = '{: >4.0f}kB'.format(queriedKB)

                if timespan.seconds > 0:
                    kBPerSecond = queriedKB / timespan.seconds
                kBPerSecondFormatted = '{: >5.1f}kB/s'.format(kBPerSecond)

                etaFormatted = ''
                if kBPerSecond > 0:
                    eta = totalDataKB / kBPerSecond - timespan.seconds
                    if eta < 0:
                        eta = 0
                    etaHr, etaMn, etaSc = _unpack_time(int(eta))
                    etaFormatted = _format_time(etaHr, etaMn, etaSc)
                    etaFormatted = 'eta {}'.format(etaFormatted)

                outputCur = k + pageCount
                outputMax = amount

            # Exit outer loop
            if restore_expected_error_occurred:
                break

            k += pageCount

            # Update
            with _outputLock.get_lock():
                print_progress(
                    outputCur, 
                    outputMax, 
                    'dev {: >3}'.format(devID),
                    '{} {} {: >9} {: >13} W:{} E:{}'.format(
                        queriedDataFormatted, 
                        kBPerSecondFormatted,
                        timespanFormatted,
                        etaFormatted,
                        numWarnings,
                        numErrors
                    ), 
                    15, 
                    lineOffset, 
                    False
                )

        #Memory decode
        rss, samples = rss2ipy.memory_decode(memory, frameCount)
        
        # Only create files, if any data was read
        if not rss.any():
            return (False, devID, errorList)

        # RSS data processing
        # NaN get replaced with similar Value to its neighbor
        # If a column is only NaN then it is replaced with '*'
        df_rss = pd.DataFrame(rss)
        df_rss[0] = df_rss[0].astype(int)
        df_rss = df_rss.replace(np.nan,"*")

        # Write RSS data to file ---------------------------------------
        # Format:
        # # cycles\tframe  -  device#1\t...\tframe  -  device#n
        #   cycle#\t	value\t...\t   value
        tableHeader = [
            '{: >11}'.format('frame {}'.format(i))
            for i in range(frameCount)
        ]
        tableHeader.insert(0, '#  cycles')

        if config:
            tableHeaderConfig = [
                '{: >11}'.format('device {}'.format(i))
                for i in [f.device for f in config.frames]
            ]
            tableHeaderConfig.insert(0, '')
            tableHeader = list(map(' |'.join, zip(tableHeader, tableHeaderConfig)))
            tableHeader[0] = tableHeader[0][:-1]
            
        df_rss.columns = tableHeader

        #file output
        output_path = _output(argOutput)

        # Check if .csv output is desired
        if csv:
            sep, ext = ',', 'csv'
        else:
            sep, ext = '\t', 'tsv'

        df_rss.to_csv(
            '{}RSS_{:0>3d}.{}'.format(
                output_path, 
                devID, 
                ext
            ),
            sep = sep, 
            index=False
        )		

        # CTE samples processing 
        if any(isinstance(n, np.ndarray) for n in samples):
            # Write sample data to file ---------------------------------------
            # Format:
            # # cycles\tsample#1\t...\tsample#n
            #   cycle#\t	value\t...\t   value
            frames = [i for i in range(len(samples)) if type(samples[i]) == np.ndarray]

            for frame in frames:
                df_samples = pd.DataFrame(samples[frame])
                df_samples[0] = df_samples[0].abs().astype(int)
                tableHeader = [
                    '{: >11}'.format('sample {}'.format(i))
                    for i in range(df_samples.shape[1]-1)
                ]
                tableHeader.insert(0, '#  cycle')
                df_samples.columns = tableHeader

                df_samples.to_csv(
                    '{}samples_{:0>3d}_{:0>3d}.{}'.format(
                        output_path, 
                        frame, 
                        devID, 
                        ext
                    ),
                    sep = sep, 
                    index=False
                )


        return (True, devID, errorList)

    finally:
        on_error = errorFunc

        if device:
            device.close()


@debug_caller
def _livelog(deviceID: int, device: rss2ipy.Device,
                   configData: List[int], kill: threading.Event) -> None:
    # Configure livelog
    queryData = [0x08, len(configData)
                 ] + configData + [0x00] * (24 - len(configData))
    status, _, _ = _query(deviceID, queryData, 1, device=device,
                          max_baudrate=True)

    bufferLen = len(configData) + 1
    # bufferScale must be > 1 to check for 0xFF, and large enoght to
    # lower the read_livelog() frequency and to reduce the likelyhood of
    # falling out of sync.
    bufferScale = 50
    while not (_abort or kill.is_set()):
        # Loop to handle livelog timeouts occurring by not receiving packets
        # for a while
        dataReceived = False
        while not dataReceived:
            if _livelog_event.is_set() or kill.is_set():
                device.stop_livelog()
                return

            try:
                response = bytearray(
                    device.read_livelog(bufferLen * bufferScale))
                dataReceived = True
            except rss2ipy.LivelogTimeout:
                print_err('No livelog data available. Are there any received '
                        'packets?')
                
        # Find the 0xFF signalling the start of a new livelog packet.
        # The livelog configuration is setup in a way that in a aingle
        # response there are 'MSB', 'middle byte' and 'LSB' giving the cycle
        # count after the first '0xFF', which heralds the start of the
        # response.
        # Since the 'middle byte' and 'LSB' can also be 0xFF, it must be
        # checked, that there is no 0xFF 3/4 places to the left or
        # bufferLen - 3/4 to the right.
        # It further is to be assumed, that there is no issue with checking
        # the next response to the right, since the bufferScale > 1!
        indexIsResponseStart = False
        while not indexIsResponseStart:
            firstIndexOf0xFF = response.index(0xFF)
            # Check, whether the index has the expected next 0xFF one
            # bufferLen+1 to the right
            next0xFFAsExpected = response[firstIndexOf0xFF + bufferLen] == 0xFF
            if next0xFFAsExpected:
                # Now check for the absence of a response start bufferLen-3/4 to
                # the right
                indexIsResponseStart = (
                    response[firstIndexOf0xFF + bufferLen - 3] != 0xFF
                    and response[firstIndexOf0xFF + bufferLen - 4] != 0xFF)

            response = response[firstIndexOf0xFF + 1:]

        response.insert(0, 0xFF)

        for i in range(bufferScale-1):
            values = []
            for k in range(1, len(configData) + 1):
                values.append(response[k + i * bufferLen])
            _queue.put(values)

            if kill.is_set():
                device.stop_livelog()
                break


@debug_caller
def _read_queue(callback, kill: threading.Event,
                pause: threading.Event) -> None:
    def default_factory():
        return {'cycle': [], 'value': []}

    localDrawStep = drawStep
    counter = 0
    while not kill.is_set():
        pause.wait()
        data = defaultdict(default_factory)

        qSize = _queue.qsize()
        if qSize > drawStep * 10:
            localDrawStep = drawStep * 10
        else:
            localDrawStep = drawStep

        while counter < localDrawStep and not kill.is_set():
            try:
                ans = _queue.get(timeout=2)
            except queue.Empty:
                continue

            frame = ans[0]
            cycle = (ans[1] << 16) + (ans[2] << 8) + ans[3]
            value = (ans[4] + ans[5] / (1 << 7))
            value *= -1
            data[frame]['cycle'].append(cycle)
            data[frame]['value'].append(value)
            counter += 1


        # If get() blocked for too long check for existing shutdown
        if kill.is_set():
            return
        callback(data)

        counter = 0


@debug_caller
def _query(devID, queryData, responseLength, device=None, exitOnWarn=False,
           suppressOTANotification=False, maxRetries=5,
           handleDefaultErrors=True, iterations=0, max_baudrate=False):
    if not device:
        # Use direct conncetion to USB-connceted device, if available
        if devID in _connectedDevices:
            device = _connectedDevices[devID]
        else:
            usedProxyPreference = ''

            # Use proxy if preference is set
            if devID in _proxyPreferences:
                possibleProxies = [
                    pID for pID in _proxyPreferences[devID]
                    if pID in _connectedDevices
                ]
                if len(possibleProxies) > 0:
                    proxyID = possibleProxies[0]
                    device = _connectedDevices[proxyID]
                    usedProxyPreference = ' (pref)'
            else:
                # any device for over-the-air communication
                if not device:
                    proxyID = next(iter(_connectedDevices))
                    device = _connectedDevices[proxyID]

            if not suppressOTANotification:
                print_err('Device {}{} --over-the-air--> device {}'.format(
                    proxyID, usedProxyPreference, devID))

    retries = 0
    retriesExhausted = False
    while True:
        responses = []
        
        # very verbose debug output
        if 3 > _verbosityLevel > 1:
            on_debug(
                'query:0x{}'.format(
                    ''.join('{:02X}'.format(x) for x in queryData)
                    )
            )

        if max_baudrate and not device.baudrate == 5:
            device.set_baudrate(5)

        if iterations:
            responses = device.query(queryData, responseLength,
                                     iterations=iterations)
        else:
            queryResult = device.query(queryData, responseLength)
            responses = [[queryResult[i] for i in range(len(queryResult))]]
        
        # TODO: The number of retries and the amount of sleep should be
        # configurable via some --switch.
        if 3 <= retries < 5:
            time.sleep(0.5)

        # Only repeat, if OTA timeout and retries not exhausted
        # TODO: Readability
        # TODO: In case of a really bad radio connection, it may be
        # significantly faster to request single missing pages instead of
        # requesting the whole bunch of pages.
        if any(r[0] == 1 for r in responses):
            if retries < maxRetries:
                retries += 1
                continue
            else:
                retriesExhausted = True
        
        
        # Successful query or retries exhausted
        break
    
    if handleDefaultErrors:
        for response in responses:
            if response[0] == 1:
                on_warn(OverTheAirTimeout(devID=devID))
            elif response[0] == 2:
                raise ControlModeRequired(devID=devID)

    if iterations:
        return (responses, retriesExhausted)
    else:
        # TODO: Consistent response
        status = responses[0][0]
        data = responses[0][1:]
        return (status, data, retriesExhausted)


@debug_caller
def _parse(file: str,
           configPerspective: List[int],
           boardRev: Optional[int] = None, 
           boardType: Optional[int] = None
    ) -> Dict[int, DeviceConfiguration]:
    """Uses config file to generate DeviceConfiguration objects

    This function matches the lines of the given config file with string patterns 
    in the RegexPatterns class. The information obtained is put into corresponding
    objects and these are then put into DeviceConfiguration objects. Finally, these
    objects are put into a dictionary with their associated device id as a key. 

    Parameters
    ----------
    file : str
        A string that is the path to the config file

    configPerspective : list
        A list with integer id of devices that are supposed to be configured

    Returns
    -------
    configs : dict
        Dictionary that consists of DeviceConfiguration objects and device id as key
    """

    if not os.path.exists(file) or not os.path.isfile(file):
        raise ConfigFileNotFound(path=file)

    # ----------------------------------------------------------------
    # Config elements ------------------------------------------------
    # frames: context : (index, frame)
    frames = {'global': []}

    # cycles: context : (amount, explicitly_set)
    cyclesDefault = rss2ipy.config.CONST.MEASURE_CYCLES_MAX
    cycles = {'global': (cyclesDefault, False)}

    # Transmission power
    txPowerDefault = 0
    txPower = {'global': (txPowerDefault, False)}

    # Radio Mode (protocol)
    modeDefault = 'BLE 2MBit'
    radio = {'global': Radio(mode=modeDefault, compat=False)}

    # CTE Settings
    cteTypeDefault = 'disabled'
    cteType = {'global': (cteTypeDefault, False)}

    # CTE's
    ctes = {'global': []}

    # Samplings
    samplings = {'global': []}

    # Storage setting for sampling
    storageDefault = 'deactivated'
    storage = {'global': (storageDefault, False)}

    # Settings for frame -> vdev maps
    syncMaps = {'global': []}

    # Synchronisation thresholds
    syncThrPoor = {'global': []}
    syncThrOOS = {'global': []}

    # Memory mappings
    memMapOut = {'global': []}
    memMapIn = {'global': []}

    # States
    # Contextsensitive options, frames cycles
    deviceContext = ['global']

    # Standart CTE and Radio config
    radioModeConfig = rss2ipy.config.default_radio_mode()
    compatConfig = rss2ipy.config.default_compatibility_mode()
    typeConfig = rss2ipy.config.default_cte_type(
        mode=radioModeConfig, 
        compatibility=compatConfig
    )

    # ----------------------------------------------------------------
    # Parsing --------------------------------------------------------
    # TODO: Value/out-of-bounds check right after matching
    with open(file) as f:
        lines = [line.rstrip('\n') for line in f]
    for i in range(len(lines)):
        line = lines[i]

        # Check for comments and empty lines
        if '#' in line:
            line = line[0:line.find('#')]
        if RegexPatterns.EmptyLineRegex.match(line):
            continue

        # Set context
        deviceContextMatch = RegexPatterns.DeviceContextRegex.match(line)
        if deviceContextMatch:
            deviceContext = []
            match = deviceContextMatch.string
            for d in [s.strip() for s in re.findall(r'[\d]+', match)]:
                deviceContext.append(d)

                # Add context to each collection
                frames[d] = []
                cycles[d] = (cyclesDefault, False)
                txPower[d] = (txPowerDefault, False)
                syncMaps[d] = []
                syncThrPoor[d] = []
                syncThrOOS[d] = []
                memMapOut[d] = []
                memMapIn[d] = []
            continue

        frameMatch = RegexPatterns.FrameRegex.match(line)
        if frameMatch:
            matchDict = frameMatch.groupdict()
            
            if matchDict['behaviour1'] == 'antenna':
                cteBehaviour = matchDict['behaviour2']
            elif matchDict['behaviour2'] == 'antenna':
                cteBehaviour = matchDict['behaviour1']
            else:
                cteBehaviour = None

            if cteBehaviour:
                if not ((cteBehaviour.lower() == 'constant') or (cteBehaviour.lower() == 'switching')):
                    raise SyntaxError(
                        line=i + 1, 
                        lineContent=line
                    )

            frame = Frame(
                index=int(matchDict['frame']),
                device=int(matchDict['device']),
                frequency=int(matchDict['frequency']),
                cteBehaviour=cteBehaviour,
                pattern=matchDict['pattern'],
                lineNumber=i + 1,
                lineContent=line
            )
            for context in deviceContext:
                frames[context].append(frame)
            continue

        modeMatch = RegexPatterns.ModeRegex.match(line)
        if modeMatch:
            m=modeMatch.groupdict()['mode']
            s=modeMatch.groupdict()['spec']
            radioMode = m+' '+s

            radioCompat = False
            if RegexPatterns.CompatRegex.match(modeMatch.string):
                radioCompat = True
            
            # Radio Mode Check
            radioModeConfig = rss2ipy.config.radio_mode(radioMode)
            if not radioModeConfig.valid():
                raise RadioModeInvalid(
                    mode=radioMode,
                    lineError=LineError(
                        line=i + 1, 
                        lineContent=line
                    )
                )
            # Compatibility Config genrate
            if radioCompat:
                compatConfig = rss2ipy.config.activated_compatibility_mode()
            else:
                compatConfig = rss2ipy.config.default_compatibility_mode()


            for context in deviceContext:
                radio[context] = Radio(
                    mode=radioMode,
                    compat=radioCompat
                )

            # Standart CTE for explict radio and compat config
            typeConfigRadio = rss2ipy.config.default_cte_type(
                mode=radioModeConfig, 
                compatibility=compatConfig
            )
            if typeConfigRadio.code() > typeConfig.code():
                typeConfig = typeConfigRadio
            continue
        cteTypeMatch = RegexPatterns.CteTypeRegex.match(line)
        if cteTypeMatch:
            typeStr = cteTypeMatch.groupdict()['type']

            # cte type check
            typeConfig = rss2ipy.config.cte_type(
                type=typeStr,
                mode=radioModeConfig,
                compatibility=compatConfig
            )
            if not typeConfig.valid():
                raise CteTypeInvalid(
                    cteType=typeConfig.reason(),
                    lineError=LineError(
                        line=i + 1, 
                        lineContent=line
                    )
                )
            
            for context in deviceContext:
                cteType[context] = (typeStr, True)
            continue

        cteMatch = RegexPatterns.CteRegex.match(line)
        if cteMatch:
            matchDict = cteMatch.groupdict()
            cteMode = matchDict['mode']

            if not ((cteMode.lower() == 'receive') or (cteMode.lower() == 'transmit')):
                raise SyntaxError(
                    line=i + 1, 
                    lineContent=line
                )
            
            if matchDict['behaviour1'] == 'antenna':
                cteBehaviour = matchDict['behaviour2']
            elif matchDict['behaviour2'] == 'antenna':
                cteBehaviour = matchDict['behaviour1']

            if not ((cteBehaviour.lower() == 'constant') or (cteBehaviour.lower() == 'switching')):
                raise SyntaxError(
                    line=i + 1, 
                    lineContent=line
                )

            cte = Cte(
                index=int(matchDict['frame']),
                mode=cteMode,
                cteBehaviour=cteBehaviour,
                pattern=matchDict['pattern'],
                lineNumber=i + 1,
                lineContent=line
            )
            for context in deviceContext:
                ctes[context].append(cte)
            continue

        samplingMatch = RegexPatterns.SamplingRegex.match(line)
        if samplingMatch:
            sampling = Sampling(
                index=int(samplingMatch.groupdict()['frame']),
                sampleBehaviour=samplingMatch.groupdict()['sampling'],
                lineNumber=i + 1,
                lineContent=line
            )
            for context in deviceContext:
                samplings[context].append(sampling)
            continue

        storageMatch = RegexPatterns.StorageRegex.match(line)
        if storageMatch:
            s = storageMatch.groupdict()['storage']
            for context in deviceContext:
                storage[context] = (s, True)
            continue

        cyclesMatch = RegexPatterns.CyclesRegex.match(line)
        if cyclesMatch:
            cyclesSpec = int(cyclesMatch.groupdict()['cycles'])
            for context in deviceContext:
                cycles[context] = (cyclesSpec, True)
            continue

        txPowerMatch = RegexPatterns.TXPowerRegex.match(line)
        if txPowerMatch:
            power = int(txPowerMatch.groupdict()['power'])

            # Value/out-of-bounds check
            txpowerConfig = rss2ipy.config.txpower(power)
            if not txpowerConfig.valid():
                raise TxPowerInvalid(
                    power=txpowerConfig.reason(),
                    lineError=LineError(
                        line=i + 1, 
                        lineContent=line
                    )
                )

            for context in deviceContext:
                txPower[context] = (power, True)
            continue

        syncMapMatch = RegexPatterns.SyncMapRegex.match(line)
        if syncMapMatch:
            sMap = SyncMap(
                frameIndex=int(syncMapMatch.groupdict()['frame']),
                vdev=int(syncMapMatch.groupdict()['vdev']),
                lineNumber=i + 1,
                lineContent=line
            )
            for context in deviceContext:
                syncMaps[context].append(sMap)
            continue

        syncThrPoorMatch = RegexPatterns.SyncThrPoorRegex.match(line)
        if syncThrPoorMatch:
            thrPoor = SyncThreshold(
                cycles=int(syncThrPoorMatch.groupdict()['cycles']),
                vdevs=int(syncThrPoorMatch.groupdict()['vdevs']),
                lineNumber=i + 1,
                lineContent=line
            )
            for context in deviceContext:
                syncThrPoor[context].append(thrPoor)
            continue

        syncThrOOSMatch = RegexPatterns.SyncThrOOSRegex.match(line)
        if syncThrOOSMatch:
            thrOOS = SyncThreshold(
                cycles=int(syncThrOOSMatch.groupdict()['cycles']),
                vdevs=int(syncThrOOSMatch.groupdict()['vdevs']),
                lineNumber=i + 1,
                lineContent=line
            )
            for context in deviceContext:
                syncThrOOS[context].append(thrOOS)
            continue

        memMapOutMatch = RegexPatterns.MemMapOutRegex.match(line)
        if memMapOutMatch:
            mMap = MemoryMap(
                frameIndex=int(memMapOutMatch.groupdict()['frame']),
                offset=int(memMapOutMatch.groupdict()['offset']),
                length=int(memMapOutMatch.groupdict()['length']),
                lineNumber=i + 1,
                lineContent=line
            )
            for context in deviceContext:
                memMapOut[context].append(mMap)
            continue

        memMapInMatch = RegexPatterns.MemMapInRegex.match(line)
        if memMapInMatch:
            mMap = MemoryMap(
                frameIndex=int(memMapInMatch.groupdict()['frame']),
                offset=int(memMapInMatch.groupdict()['offset']),
                length=1,
                lineNumber=i + 1,
                lineContent=line
            )
            for context in deviceContext:
                memMapIn[context].append(mMap)
            continue

        # Throw syntax error, if the line could not be matched
        raise SyntaxError(line=str(i+1), lineContent=line)

    # ----------------------------------------------------------------
    # Check config validity ---------------------------------------
    if not configPerspective:
        # No devices given => using all devices specified when declaring frames
        configPerspective = list(
            {f.device
             for flist in frames.values()
             for f in flist})
    contexts = ['global'] + [str(i) for i in configPerspective]

    # Infos ----------------------------------------------------------
    # Info about defaulting 'global' cycle value to max
    if not cycles['global'][1] and not any(
        [cycles[c][1] for c in cycles if not c == 'global']):
        on_info(
            'The global amount of cycles for the measurement is not specified. '
            'Defaulting to the maximal possible cycle count of 16777215'
        )
        
    # Info about defaulting not set device cycle values to 'global'
    for context in [c for c in contexts if not c == 'global']:
        if context not in cycles or not cycles[context][1]:
            on_info(
                'The amount of cycles for device #{} was not explicitly specified. ' 
                'Defaulting to global: {}'.format(
                    context,
                    cycles['global'][0]
                )
            )

    # Check for correctly set txpower values
    if not txPower['global'][1] and not any(
        [txPower[tx][1] for tx in txPower if not tx == 'global']):
        on_info(
            'The global TXPower setting was not explicitly specified. '
            'Defaulting to 0dBm'
        )
        txPower['global'] = (0, True)
    for context in [c for c in contexts if not c == 'global']:
        if context not in txPower or not txPower[context][1]:
            on_info(
                'The TXPower setting for device #{} was not explicitly '
                'specified. Defaulting to global: {}dBm'.format(
                    context,
                    txPower['global'][0]
                )
            )
            txPower[context] = txPower['global']

    # Default mapping of vdevs
    if len(syncMaps['global']) == 0:
        for context in syncMaps:
            if len(syncMaps[context]) == 0:
                on_info(
                    'There was no explicit virtual devices configuation for device '
                    '#{}. Defaulting to one virtual device for each physical device'
                    .format(context)
                )

    # Default mapping of vmems
    if len(memMapIn['global']) == 0:
        for context in memMapIn:
            if len(memMapIn[context]) == 0:
                on_info(
                    'There was no explicit virtual memory configuation for device '
                    '#{}. Defaulting to one virtual memory per measurement frame'
                    .format(context)
                )

    # Warnings -------------------------------------------------------
    # Check for more vdevs in threshold rules than defined
    differentVDevs = []
    for context in syncMaps:
        for sMap in syncMaps[context]:
            if sMap.vdev not in differentVDevs:
                differentVDevs.append(sMap.vdev)
    thrRules = [syncThrPoor, syncThrOOS]
    for rules in thrRules:
        # If default config is to be used later on:
        #     set 'differentVDevs' to different physical devices
        if len(differentVDevs) == 0:
            differentVDevs = list(
                {frame.device
                 for flist in frames.values()
                 for frame in flist})
        for context in rules:
            for rule in rules[context]:
                if rule.vdevs > len(differentVDevs):
                    on_warn(
                        SyncThresholdHigh(
                            defined=len(differentVDevs), 
                            required=rule.vdevs
                        )
                    )

    # Errors
    # Check for missing frames
    missing = []
    allFrames = {}
    for context in frames:
        for frame in frames[context]:
            allFrames[frame.index] = frame
    for i in range(len(allFrames)):
        if i not in allFrames:
            missing.append(i)

    if len(missing) > 0:
        raise FrameMissing(missingFrame=missing)

    # Check for missing virtual devices
    allSyncMaps = [
        m.frameIndex for contextMaps in syncMaps.values() for m in contextMaps
    ]
    missingVdevs = [fID for fID in allFrames if fID not in allSyncMaps]
    if len(allSyncMaps) > 0 and len(missingVdevs) > 0:
        raise VDevMissing(missingVdevs=missingVdevs)
    
    # Check for missing virtual memory defintions
    allMemMapIn = [
        m.frameIndex for contextMaps in memMapIn.values() for m in contextMaps
    ]
    missingVmems = [fID for fID in allFrames if fID not in allMemMapIn]
    if len(allMemMapIn) > 0 and len(missingVmems) > 0:
        raise VMemMissing(missingVmems=missingVmems)

    # Check for undefined frames
    # i.e. used frames, that are not defined
    undefined = {}
    referencedFrameIndices = {}
    variables = [syncMaps, memMapOut, memMapOut]
    for var in variables:
        for context in var:
            for obj in var[context]:
                referencedFrameIndices[obj.frameIndex] = obj

    for frameIndex in referencedFrameIndices:
        if frameIndex not in allFrames:
            undefined[frameIndex] = referencedFrameIndices[frameIndex]
            raise FrameUndefined(
                frameIndex=frameIndex,
                line=str(referencedFrameIndices[frameIndex].get_number()),
                lineContent=referencedFrameIndices[frameIndex].get_line()
            )

    # ----------------------------------------------------------------
    # Parsed values sanity check -------------------------------------
    # Frequency and convinience CTE

    for context in frames:
        for frame in frames[context]:
            if not rss2ipy.config.CONST.FREQ_MIN <= frame.frequency <= rss2ipy.config.CONST.FREQ_MAX:
                raise FrequencyInvalid(
                    frequency=frame.frequency,
                    lineError=LineError(
                        line=frame.get_number(), 
                        lineContent=frame.get_line()
                    )
                )

            if frame.cteBehaviour:
                if frame.cteBehaviour.lower() == 'constant':
                    antennaConfig = rss2ipy.config.cte_antenna(
                        antenna=int(frame.pattern), 
                        type=typeConfig,
                        boardRev=boardRev, 
                        boardType=boardType
                    )
                    if not antennaConfig.valid():
                        raise ConstantAntennaInvalid(
                            constantAntenna=frame.pattern,
                            reason=antennaConfig.reason(),
                            lineError=LineError(
                                line=frame.get_number(), 
                                lineContent=frame.get_line()
                            )
                        )

                if frame.cteBehaviour.lower() == 'switching':
                    if not rss2ipy.config.cte_pattern( 
                        pattern=frame.pattern, 
                        type=typeConfig,
                        boardRev=boardRev, 
                        boardType=boardType
                    ).valid():
                        raise SwitchingPatternInvalid(
                            switchingPattern=frame.pattern,
                            lineError=LineError(
                                line=frame.get_number(), 
                                lineContent=frame.get_line()
                            )
                        )

    # CTE Config with CTE: ....
    for context in ctes:
        for cte in ctes[context]:
            if cte.cteBehaviour.lower() == 'constant':
                if not rss2ipy.config.cte_antenna(
                    antenna=int(cte.pattern), 
                    type=typeConfig,
                    boardRev=boardRev, 
                    boardType=boardType
                ).valid():
                    raise ConstantAntennaInvalid(
                        constantAntenna=cte.pattern,
                        lineError=LineError(
                            line=cte.get_number(), 
                            lineContent=cte.get_line()
                        )
                    )

            if cte.cteBehaviour.lower() == 'switching':
                if not rss2ipy.config.cte_pattern( 
                    pattern=cte.pattern, 
                    type=typeConfig,
                    boardRev=boardRev, 
                    boardType=boardType
                ).valid():
                    raise SwitchingPatternInvalid(
                        switchingPattern=cte.pattern,
                        lineError=LineError(
                            line=cte.get_number(), 
                            lineContent=cte.get_line()
                        )
                    )
            

    # Sync threshold cycles
    thresholds = [syncThrPoor, syncThrOOS]
    for threshold in thresholds:
        for context in threshold:
            for thr in threshold[context]:
                if not rss2ipy.config.CONST.SYNC_THR_MIN <= thr.cycles <= rss2ipy.config.CONST.SYNC_THR_MAX:
                    raise SyncThresholdInvalid(
                        syncThreshold=thr.cycles,
                        lineError=LineError(
                            line=thr.get_number(), 
                            lineContent=thr.get_line()
                        )
                    )
                elif not thr.cycles % 10 == 0:
                    on_warn(
                        SyncThresholdRounding(
                            cycles=thr.cycles, 
                            line=thr.get_number(),
                            lineContent=thr.get_line()
                        )
                    )


    # ----------------------------------------------------------------
    # Prepare device specific configuarion -------------------------
    configs = {}
    for device in configPerspective:
        config = DeviceConfiguration(ID=int(device))

        # Handle only global an matching contexts
        contexts = ['global'] + [
            str(cPer) for cPer in configPerspective if config.ID == cPer
        ]

        # Always start with 'global', device specific settings overwrite
        # the initially set global settings
        for context in contexts:
            # Cycles
            if context in cycles:
                config.cycles = cycles[context][0]

            # Transmission power
            if context in txPower:
                config.power = txPower[context][0]

            # Radio Mode
            if context in radio:
                config.radio = radio[context]

            # CTE type
            if context in cteType:
                config.cteType = cteType[context][0]

            # Sampling storage
            if context in storage:
                config.storage = storage[context][0]

            # Keep all already set frames, that are not to be overwritten
            # + add new frames, and frames replacing the filtered frames
            config.frames = [
                cf for cf in config.frames
                if not any(cf.index == f.index
                           for f in frames.get(context, []))
            ] + frames.get(context, [])
            config.frames.sort(key=lambda x: x.index, reverse=False)

            #
            # CTE
            config.ctes = [
                cc for cc in config.ctes
                if not any(cc.index == c.index
                           for c in ctes.get(context, []))
            ] + ctes.get(context, [])
            config.ctes.sort(key=lambda x: x.index, reverse=False)

            #
            # Sampling
            config.samplings = [
                cs for cs in config.samplings
                if not any(cs.index == s.index
                           for s in samplings.get(context, []))
            ] + samplings.get(context, [])
            config.samplings.sort(key=lambda x: x.index, reverse=False)

            # Sync map; filtering: see frames
            config.syncMap = [
                cm for cm in config.syncMap
                if not any(cm.frameIndex == m.frameIndex
                           for m in syncMaps.get(context, []))
            ] + syncMaps.get(context, [])
            # Set default Sync Map
            if not any(config.syncMap):
                vdevMap = {}
                i = 0
                for frame in config.frames:
                    if frame.device not in vdevMap:
                        vdevMap[frame.device] = i
                        i += 1
                config.syncMap = [
                    SyncMap(f.index, vdevMap[f.device]) for f in config.frames
                ]
            config.syncMap.sort(key=lambda x: x.frameIndex, reverse=False)

            # Sync threshold poor; filtering: see frames
            config.thrPoor = [
                cthr for cthr in config.thrPoor
                if not any(cthr.vdevs == thr.vdevs
                           for thr in syncThrPoor.get(context, []))
            ] + syncThrPoor.get(context, [])

            # Sync threshold out-of-sync; filtering: see frames
            config.thrOOS = [
                cthr for cthr in config.thrOOS
                if not any(cthr.vdevs == thr.vdevs
                           for thr in syncThrOOS.get(context, []))
            ] + syncThrOOS.get(context, [])

            # Output maps; filtering: see frames
            config.memMapOut = [
                cm for cm in config.memMapOut
                if not any(cm.frameIndex == m.frameIndex
                           for m in memMapOut.get(context, []))
            ] + memMapOut.get(context, [])

            # Input maps; filtering: see frames
            config.memMapIn = [
                cm for cm in config.memMapIn
                if not any(cm.frameIndex == m.frameIndex
                           for m in memMapIn.get(context, []))
            ] + memMapIn.get(context, [])

            # Set default Memory input Map
            if not any(config.memMapIn):
                vmemMapIn = {}
                i = 0
                for frame in config.frames:
                    vmemMapIn[frame.index] = i
                    i += 1
                config.memMapIn = [
                    MemoryMap(f.index, vmemMapIn[f.index],1) for f in config.frames
                ]
            config.memMapIn.sort(key=lambda x: x.frameIndex, reverse=False)

        configs[config.ID] = config

    return configs


# -----------------------------------------------------------------------------
# Cmdline parser --------------------------------------------------------------
# Declaration
def _int_list_parser_helper(arg: str) -> List[int]:
    return [int(x) for x in arg.split(',')]


parser = argparse.ArgumentParser(
    description='command line interface for rss2i'
)
port_group = parser.add_mutually_exclusive_group(required=True)
port_group.add_argument(
    '-p', 
    '--port', 
    help='connected serial port'
)
port_group.add_argument(
    '-a', 
    '--auto', 
    action='store_true',
    help='detect connected devices automatically'
)
parser.add_argument(
    '-x', 
    '--proxy', 
    action='append', 
    required=False,
    help='specifies, which devices should be preferentially '
         'used to establish over-the-air connections. ex: \'-x 1:7,8,9\''
)
parser.add_argument(
    '-v', 
    '--verbose', 
    required=False, 
    action='count',
    default=0,
    help='enables verbose output. (i.e. -v for general info, '
         '-vv for debug messages and '
         '-vvv for in-depth debug messages/debug trace)'
)
subparsers = parser.add_subparsers(dest='subcommand', help='sub-commands')
subparsers.required = True

# parse config parser
config_parser = subparsers.add_parser(
    'parse', 
    help='parses specified config file and outputs the parsed info'
)
config_parser.add_argument(
    '-f', 
    '--file', 
    type=str, 
    required=True,
    help='path to config file'
)
config_parser.add_argument(
    '-d', 
    '--device', 
    type=_int_list_parser_helper, 
    action='append',
    required=False, 
    help='deviceIDs for which to parse the configuation'
)
config_parser.set_defaults(func=parse)

# getid parser
getid_parser = subparsers.add_parser(
    'getid', 
    aliases=['whois'],
    help='returns connected device\'s IDs'
)
getid_parser.add_argument(
    '-d',
    '--device',
    type=_int_list_parser_helper,
    action='append',
    required=False,
    help='deviceIDs for which to display the configuation'
)
getid_parser.set_defaults(func=get_id)

# ping parser
ping_parser = subparsers.add_parser(
    'ping', 
    help='shows the specified device by flashing the LED'
)
ping_parser.add_argument(
    '-d', 
    '--device', 
    type=_int_list_parser_helper,
    action='append', 
    required=False,
    help='deviceIDs to show/ping'
)
ping_parser.set_defaults(func=ping)

# config parser
config_parser = subparsers.add_parser(
    'config', 
    help='sets the configuration (frames, synchronisation map, '
    'poor/out-of-sync thresholds, remote rssi out- and input) '
    'by config file'
)
config_parser.add_argument(
    '-f', 
    '--file', 
    type=str, 
    required=True,
    help='path to config file'
)
config_parser.add_argument(
    '-d', 
    '--device', 
    type=_int_list_parser_helper,
    action='append', required=True,
    help='deviceIDs for which to set the configuation'
)
config_parser.set_defaults(func=config)

# download parser
download_parser = subparsers.add_parser(
    'download', 
    help='downloads last measurement\'s data'
)
download_parser.add_argument(
    '-d', 
    '--device', 
    type=_int_list_parser_helper, 
    action='append',
    required=True, 
    help='deviceIDs for which to set the configuation'
)
count_group = download_parser.add_mutually_exclusive_group(required=True)
count_group.add_argument(
    '-c', 
    '--frameCount', 
    type=int,
    help='frame count per cycle'
)
count_group.add_argument(
    '-f', 
    '--file', 
    type=str, 
    help='path to config file'
)
download_parser.add_argument(
    '-o', 
    '--output', 
    type=str, 
    required=False,
    help='output pattern: path/to/file -> here file is the prefix for the data file.\n\
          path/to/file/ -> here file is the subdirectory in wich the data file is stored.'
)
download_parser.add_argument(
    '--csv', 
    required=False, 
    action='store_true',
    help='output .csv files with download'
)
download_parser.add_argument(
    '-s', 
    '--simultaneous', 
    action='store_true', 
    required=False,
    help='enables simultaneous download of data, if possible'
)
download_parser.add_argument(
    '-n', 
    '--pagecount', 
    type=int, 
    required=False,
    help='the amount of pages, which will be queried at once'
)
download_parser.add_argument(
    '-r', 
    '--resume', 
    required=False, 
    type=str,
    help='path to a file, created by the \'record\'-command'
)
download_parser.set_defaults(func=download)

# record parser to enable a later download
record_parser = subparsers.add_parser(
    'record', 
    help='saves device ID and measurement memory address to a file'
)
record_parser.add_argument(
    '-o', 
    '--output', 
    required=True, 
    type=str,
    help='path to file'
)
record_parser.add_argument(
    '-d', 
    '--device', 
    type=_int_list_parser_helper,
    action='append', required=True,
    help='deviceIDs for which to set the configuation'
)
record_parser.set_defaults(func=record)

# `restore` subcommand.
restoreParser = subparsers.add_parser(
    'restore',
    help='Download measurement data from the specified devices '
        + 'but do not rely on the session data reported by the '
        + 'devices (if any). Instead, the session data has to '
        + 'be specified on command line.'
)
addressGroup = restoreParser.add_mutually_exclusive_group(required=True)
addressGroup.add_argument(
    '-r',
    '--resume',
    required=False,
    type=str,
    help='Path to session data file created by the \'record\' subcommand.'
)
addressGroup.add_argument(
    '-x',
    '--address',
    required=False,
    type=str,
    help='Hexadecimal measurement memory start address.'
)
restoreParser.add_argument(
    '-f',
    '--file',
    type=str,
    help='Path to config file.'
)
restoreParser.add_argument(
    '-o',
    '--output',
    type=str,
    required=False,
    help='Output prefix of the measurement data files. '
       + 'Non-existent directories are created automatically.'
)
restoreParser.add_argument(
    '-g',
    '--gap',
    required=False,
    type=int,
    help='Maximal gap between subsequent cycle numbers.'
)
restoreParser.add_argument(
    '-d',
    '--device',
    type=_int_list_parser_helper,
    action='append',
    required=True,
    help='DeviceID(s) for which to download measurement data.'
)
restoreParser.add_argument(
    '-n',
    '--pagecount',
    type=int,
    required=False,
    help='Amount of pages that are queried at once.'
)
restoreParser.set_defaults(func=restore)

# measurement parser
start_parser = subparsers.add_parser(
    'start', 
    aliases=['kick-off', 'begin'],
    help='starts measurement for all given devices; to set '
         'the measurement configuation, use \'config\''
)
start_parser.add_argument(
    '-d', 
    '--device', 
    type=_int_list_parser_helper,
    action='append', 
    required=True,
    help='deviceIDs for which to start the measurement'
)
start_parser.add_argument(
    '-o', 
    '--output', 
    type=str, 
    required=False,
    help='path to output folder, where the start addresses will be saved'
)
start_parser.set_defaults(func=start_measurement)

# abort parser
stop_parser = subparsers.add_parser(
    'stop', 
    aliases=['abort'],
    help='forces control mode, ending all measurements '
)
stop_parser.add_argument(
    '-f', 
    '--file', 
    required=False, 
    type=str,
    help='config file to reconfigure the device in case of a missing config'
)
stop_parser.set_defaults(func=abort)

# monitor/livelog parser
monitor_parser = subparsers.add_parser(
    'monitor',
    help='provides live data about the currently running measuement'
)
monitor_parser.add_argument(
    '-d', 
    '--device', 
    type=_int_list_parser_helper, 
    action='append',
    required=True, 
    help='single deviceID for which to start the livelog'
)
monitor_parser.add_argument(
    '-n', 
    '--frames', 
    type=_int_list_parser_helper, 
    action='append',
    required=True,
    help='received frames which should be plotted in the monitor window'
)
monitor_parser.set_defaults(func=monitor)

# log by livelog parser
log_parser = subparsers.add_parser(
    'log', 
    aliases=['livelog'],
    help='outputs livelog with preset configuration into stderr'
)
log_parser.add_argument(
    '-d', 
    '--device', 
    type=_int_list_parser_helper,
    required=True, action='append',
    help='Device ID for which to start the debug log'
)
log_parser.set_defaults(func=log)

# debug
debug_parser = subparsers.add_parser(
    'debug', 
    help='Outputs debug info with given argument into stderr'
)
debug_parser.add_argument(
    '-d', 
    '--device', 
    type=_int_list_parser_helper,
    required=True, action='append',
    help="Device IDs for which to display status and error info."
)
debug_parser.set_defaults(func=debug)

# measure parser (config, start, stop, download)
measure_parser = subparsers.add_parser(
    'measure', 
    aliases=['join', 'cssd', 'oneshot'],
    help='combines \'config\', \'start\', \'stop\', \'download\' '
         'and executes the commands with a given delay after \'start\''
)
measure_parser.add_argument(
    '-f', 
    '--file', 
    type=str, 
    required=True,
    help='path to config file'
)
measure_parser.add_argument(
    '-t', 
    '--tdelay', 
    type=int, 
    required=True,
    help='delay in seconds before \'stop\'ing the measurement'
)
measure_parser.add_argument(
    '-o', 
    '--output', 
    type=str, 
    required=False,
    help='output pattern'
)
measure_parser.add_argument(
    '-s', 
    '--simultaneous', 
    action='store_true', 
    required=False,
    help='enables simultaneous download of data, if possible'
)
measure_parser.add_argument(
    '-n', 
    '--pagecount', 
    type=int, 
    required=False,
    help='the amount of pages, which will be queried at once'
)
measure_parser.add_argument(
    '-i', 
    '--interrupt', 
    type=int, 
    required=False,
    help='disconnects the connection to given device, while waiting'
)
measure_parser.add_argument(
    '-d', 
    '--device', 
    type=_int_list_parser_helper,
    action='append', required=True,
    help='deviceIDs for which to set the configuation'
)
measure_parser.add_argument(
    '--csv', 
    required=False, 
    action='store_true',
    help='output .csv files with download'
)
measure_parser.set_defaults(func=measure)


# -----------------------------------------------------------------------------
# Main ------------------------------------------------------------------------
def main():
    try:
        # If fixed_cmd is ['DEBUG', *args], the given args will be parsed by
        # argparse instead of the command line
        fixed_cmd = []

        if len(sys.argv) == 1 and fixed_cmd and fixed_cmd[0] == 'DEBUG':
            fixed_cmd = fixed_cmd[1:]
            print_err('Executing with hardcoded debug arguments')
            print_err('\'{}\''.format(' '.join(fixed_cmd)), formatIndent=True)
            args = parser.parse_args(fixed_cmd)
        else:
            args = parser.parse_args()

        # Flatten devices list
        if hasattr(args, 'device') and args.device:
            args.device = [dev for sublist in args.device for dev in sublist]

        global _verbosityLevel
        _verbosityLevel = args.verbose

        if _verbosityLevel > 2:
            on_debug('Program was launched with:')
            for key, value in vars(args).items():
                # Skip func attribute, as this is not a launch argument
                if key == 'func':
                    continue
                on_debug(f'{key}={value}')

            if hasattr(args, 'simultaneous') and args.simultaneous:
                args.simultaneous = False
                on_debug('Simultaneous download of device connected '
                          'in parallel is disabled, when running the program '
                          f'with a debug level of {_verbosityLevel}')

        # Allow for missing devices, when parsing/testing a config file
        if args.subcommand == 'parse':
            pass
        else:
            _establish_device_connections(args)
            if args.verbose:
                get_id(args)

        args.func(args)


    except KeyboardInterrupt:
        print_err('', usePrefix=False)
        print_err('', usePrefix=False)
        
    except SystemExit:
        pass

    except Rss2iException as error:
        on_error(error)

    finally:
        global _abort
        _abort = True
        _close_device_connections()


if __name__ == '__main__':
    main()