/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

#define RSS2I_MODULE "battery"

/*
  Battery monitoring.
  
  Authors
  =======
  
    Copyright (C) 2018-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
  
    2019-06-24    Initial public release.
    2021-03-27    Use {NVIC_RSS2I_*_PRIORITY} macros.
  
  Description
  ===========
    
    After initialization using {battery_init()}, the battery volatage gets
    measured periodically. When it deceeds a critical threshold (see board.c),
    the debug module gets informed.
    The latest ADC value as well as the latest normalized battery level can be
    obtained by calling {battery_get_value()} and {battery_get_level()},
    respectively. Note, that for the moment, the latter will return {4}, if the
    battery is nearly empty, and {100}, if not (TODO).
*/


#include <stdint.h>
#include <stdbool.h>

#include <libopencm3/nrf5/saadc.h>
#include <libopencm3/cm3/nvic.h>

#include <rss2i/debug.h>
#include <rss2i/timer.h>
#include <rss2i/board.h>
#include <rss2i/nvic.h>

#include <rss2i/config/battery.h>

static volatile uint16_t batterySAADCValue;
static volatile uint16_t batteryLastSAADCValue = 0;
static volatile uint8_t batteryLevel = 0;

uint8_t battery_get_level(void){
/*
  Returns the current battery level (from {0} to {100}).
  TODO Note, that for the moment, it will return {4}, if the battery is nearly
  empty, and {100}, if not.
*/

	return batteryLevel;
}

uint16_t battery_get_value(void){
/*
  Returns the last measurement value of the SAADC (API for cmd).
*/
	
	return batteryLastSAADCValue;
}

static void battery_start(void){
	
	SAADC_RESULT.PTR = (uint32_t) &batterySAADCValue;
	SAADC_RESULT.MAXCNT = 1;
	SAADC_TASKS_START = 1;
}

void battery_init(void){
/*
  Initializes this module. May be called multiple times.
*/
	
	static bool initialized = false;
	
	if(!initialized){
	
		debug_init();
		
		nvic_enable_irq(NVIC_SAADC_IRQ);
		nvic_set_priority(NVIC_SAADC_IRQ, NVIC_RSS2I_LOW_PRIORITY);
	
		SAADC_ENABLE = SAADC_ENABLE_ENABLE_ENABLED;
	
		SAADC_CH[0].PSELP = SAADC_CH_PSELP_PSELP_ANALOGINPUT7;
	
		SAADC_CH[0].CONFIG = SAADC_CH_CONFIG_RESP_BYPASS
		                   | SAADC_CH_CONFIG_RESN_BYPASS
		                   | SAADC_CH_CONFIG_GAIN_GAIN1_6
		                   | SAADC_CH_CONFIG_REFSEL_INTERNAL
		                   | SAADC_CH_CONFIG_TACQ_10US
		                   | SAADC_CH_CONFIG_MODE_SE
		                   | SAADC_CH_CONFIG_BURST_DISABLED;
		                   
		SAADC_RESOLUTION = SAADC_RESOLUTION_VAL_10BIT;
		SAADC_OVERSAMPLE = SAADC_OVERSAMPLE_OVERSAMPLE_BYPASS;
		SAADC_RESULT.MAXCNT = 1;
	
		SAADC_INTENSET = SAADC_INTENSET_STARTED_SET
		               | SAADC_INTENSET_END_SET;
	
		SAADC_TASKS_CALIBRATEOFFSET = 1;
	
		timer_call(&battery_start, 1000 * 10 / BATTERY_FREQUENCY, TIMER_REPEAT);
		
		initialized = true;
	}
}


void saadc_isr(void){
	
	static uint16_t threshCount = 0;
	static bool reported = false;
	static uint16_t calCount = 0;
	
	if(SAADC_EVENTS_STARTED){
		
		SAADC_EVENTS_STARTED = 0;
		SAADC_TASKS_SAMPLE = 1;
		while(SAADC_EVENTS_STARTED);
	}
	else if(SAADC_EVENTS_END){
	
		SAADC_EVENTS_END = 0;
		
		/* Save current value. */
		batteryLastSAADCValue = batterySAADCValue;
		
		if(batterySAADCValue <= board_battery_threshold() && !reported && ++threshCount == BATTERY_ITERATIONS){
			
			debug_report(DEBUG_BATTERY_LOW);
			reported = true;
		}
		else if(batterySAADCValue >= board_battery_threshold() + BATTERY_THRESHOLD_DIFF){
			
			threshCount = 0;
		}
		
		 /* TODO This is just a temporary solution... */
		batteryLevel = (reported) ? 4 : 100;
		
		/* Auto-re-calibration needed? */
		
		if(calCount++ == BATTERY_CALIBRATION_PERIOD){
			
			calCount = 0;
			SAADC_TASKS_CALIBRATEOFFSET = 1;
		}
		
		/* while(SAADC_EVENTS_END); */
	}
}


