/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

#define RSS2I_MODULE "ble"

/*
  Holds BLE 5.0 Uncoded PHY compatible Access Addresses.
   
  Authors
  =======
  
    Copyright (C) 2018-2019  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
  
    2019-05-27    Initial release.
  
  Description
  ===========

    Each RADIO packet (of measurement or control type) gets equipped with an
    Access Address that depends on the packet's type as well as on the targeted
    RSS2I device.
    Since the Access Address has no influence on the packet's CRC sum, the set
    of all used Access Addresses must have an adequate minimal Hamming Distance
    in order to minimize the possibility of misaddressing. Meanwhile all used
    Access Addresses should meet the requirements of the BLE 5.0 Uncoded PHY.
*/

#include <stdint.h>

uint32_t ble_gen_control_access_address(uint8_t devID){
/*
  Generates a 32-bit access address that corresponds to the given {devID}
  uniquely while complying to the BLE 5.0 Uncoded PHY's Access Address
  requirements.
*/
	
	static const uint32_t address[256] = {
	
		0x44d76390UL, 0x76dbef52UL, 0x9f728789UL, 0x68526570UL,
		0x4bbbec37UL, 0xa72727ceUL, 0x8c0dba65UL, 0x10ed28e2UL,
		0xd98309ecUL, 0xf4917dd5UL, 0xa5f0898fUL, 0xab2ca9d6UL,
		0x4e8f2191UL, 0xbc516d3cUL, 0xa83f1987UL, 0x6ab3db5cUL,
		0x587b4438UL, 0xe4daed62UL, 0x562bd04eUL, 0x176d14e4UL,
		0x4f534bf6UL, 0x25b09aafUL, 0xccd9dab1UL, 0x6414bd4aUL,
		0xc96a763fUL, 0x237e7efdUL, 0xc6887c37UL, 0xa0bd7925UL,
		0x5d583904UL, 0x3b3ae0cfUL, 0x76d7c3f7UL, 0x645b72e1UL,
		0xd1e1839bUL, 0x2048385aUL, 0x56b8faeaUL, 0x682ed493UL,
		0x55b061e0UL, 0x6f0e5fbaUL, 0xaf3071cfUL, 0x48693d75UL,
		0xa8c78413UL, 0xcdc38bcbUL, 0x76104aa5UL, 0x70b0a290UL,
		0xc63b0c26UL, 0xac0d0eddUL, 0x87452da8UL, 0x38a40f4eUL,
		0x4ee9d5eaUL, 0x4b3c6181UL, 0x77bb82d6UL, 0x502fc24fUL,
		0x9c26f35bUL, 0x4b92a4e2UL, 0x4b8719d2UL, 0x11a31a35UL,
		0x47a81315UL, 0x9bd306c6UL, 0x324c9347UL, 0xb8446099UL,
		0xf6e70dc5UL, 0x1349f315UL, 0xb690dcf6UL, 0xecde9638UL,
		0x584e0b9dUL, 0x65618b70UL, 0xd43029caUL, 0x86740bb8UL,
		0xa9364eccUL, 0x7bccbd1cUL, 0x412b6b53UL, 0xde08353fUL,
		0x97099531UL, 0x220d89dbUL, 0xd821463eUL, 0x50ad848bUL,
		0x5d63e7a8UL, 0x21d949bcUL, 0x1aea74bfUL, 0x23198536UL,
		0x8e08c427UL, 0xc6163a8eUL, 0x519e3067UL, 0x14d90984UL,
		0x6abb4107UL, 0x75db8821UL, 0xb712a704UL, 0xd9910ef2UL,
		0xac7b4322UL, 0x418fd4d5UL, 0xddc62ce2UL, 0xb84666faUL,
		0x61b2ecbbUL, 0x9466e992UL, 0x6c248a81UL, 0x16906c8cUL,
		0x335bcb3bUL, 0xd7a693bfUL, 0x49af7bb2UL, 0xdba5c9e0UL,
		0xdbc09451UL, 0x8df7bc4bUL, 0x6d6f5f6fUL, 0x8f23d0d5UL,
		0x65f5e8d9UL, 0x482c086fUL, 0x30c94b52UL, 0x87283485UL,
		0x55784ca4UL, 0x34e6f9d6UL, 0xcd83c156UL, 0x50fd96e5UL,
		0x662b7766UL, 0xac954166UL, 0x41ee48b0UL, 0xcfdde61cUL,
		0x2bb96e0cUL, 0x9b99d9f4UL, 0xd28f77ddUL, 0xeaaeeebaUL,
		0xc7874da8UL, 0xba85f4cbUL, 0xce206962UL, 0x8671e65cUL,
		0x98e47c87UL, 0x2a0b5df9UL, 0x16d0b40fUL, 0x39ba1eb7UL,
		0x4dc889b1UL, 0xe56a5d68UL, 0x7ad88c25UL, 0xd4aebeefUL,
		0x31db7440UL, 0x26eeef41UL, 0x2b15fa0dUL, 0xda8186c6UL,
		0x228111ceUL, 0xd5a5cf2eUL, 0x2b6bdea0UL, 0x52bd0fd9UL,
		0x4793e2ddUL, 0x9f5c1605UL, 0x85ef4bb0UL, 0x27a962d1UL,
		0x1391a38aUL, 0x1a6d44f5UL, 0xd608a331UL, 0xcaf447d9UL,
		0x8dae1d60UL, 0x94d5c78bUL, 0xdd8e42d7UL, 0x74afb362UL,
		0x6c564914UL, 0xe60ab7daUL, 0x8f053121UL, 0xb39371ccUL,
		0x95a20b72UL, 0x8874f128UL, 0xe9a435d4UL, 0x4e1f5692UL,
		0x10af4fcaUL, 0xa5444b2bUL, 0x2b79d3e6UL, 0x58accd17UL,
		0xe56b4aa2UL, 0xad9cd1dcUL, 0x353d5a06UL, 0x5aa8c623UL,
		0x2bcb90d5UL, 0x52163c35UL, 0x53bbdb1aUL, 0x61f66aadUL,
		0x3b2f78b0UL, 0x2cea84f0UL, 0x2d253b9bUL, 0x99478b16UL,
		0x6b3e86bbUL, 0xc94e934aUL, 0x228c7a4bUL, 0x8b6adf09UL,
		0x90dd3a21UL, 0x5d33b4eeUL, 0x18538dbfUL, 0xcd056466UL,
		0xe61e2722UL, 0x081dcecdUL, 0x932f7773UL, 0x8b729728UL,
		0xd998bed5UL, 0xedde23b7UL, 0x5164953fUL, 0x917bc40cUL,
		0x12ea50edUL, 0x9bdf73ccUL, 0x9e309c9aUL, 0xc9a75020UL,
		0x5aa7e79bUL, 0x52d4c0d8UL, 0x2236c795UL, 0xaedeb2e0UL,
		0x4c4d7223UL, 0x858bf748UL, 0xd478714cUL, 0xdf5f4946UL,
		0xd9556319UL, 0x6e0ece31UL, 0x74f033bbUL, 0x6bd3690dUL,
		0x437455e6UL, 0x71112ebcUL, 0xcbcb187dUL, 0x6732879eUL,
		0x4cb1d25cUL, 0x2d22a7d7UL, 0x4704d336UL, 0xd6eb6f20UL,
		0xdc484264UL, 0x4e65bb47UL, 0x57aabb5aUL, 0xafd48e5dUL,
		0x9cbb3b29UL, 0xb5bc8386UL, 0xc89b5d0dUL, 0x6cdc11b0UL,
		0x460c1474UL, 0x23e0d620UL, 0xd85cdc08UL, 0x89c958d2UL,
		0x71e48712UL, 0xbc16c5e5UL, 0xc8d40862UL, 0xa79550dcUL,
		0x62db1dc6UL, 0x209a0c3aUL, 0xc5d4b551UL, 0x865ac42bUL,
		0x2cbbc91bUL, 0x9a71a0f1UL, 0xf4826331UL, 0x9cfd8470UL,
		0x62827dbaUL, 0x23d3f458UL, 0x72d99a74UL, 0x9394395aUL,
		0xd51a72bbUL, 0x3661239aUL, 0xe51cc2cfUL, 0x3b551b49UL,
		0x399ea063UL, 0x49ec1cfaUL, 0x46d984c2UL, 0x768558d8UL,
		0xf4c22e17UL, 0x9628c2b3UL, 0x0be635cbUL, 0x2944ad8fUL,
	};
	
	return address[devID];
}

uint32_t ble_gen_measure_access_address(void){
/*
  Returns a BLE 5.0 Uncoded PHY compatible Access Address that should get
  used within normal measurement packets.
*/
	return 0x130517cbUL;
}

uint32_t ble_gen_disrupt_access_address(void){
/*
  Returns a BLE 5.0 Uncoded PHY compatible Access Address that should get
  used within disruptive measurement packets.
*/
	return 0x69ee7986UL;
}
