/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

#define RSS2I_MODULE "board"

/*
  Set and get board-specific parameters to/from the non-volatile User
  Information Configuration Registers (UICR).
  
  Authors
  =======
    
    Copyright (C) 2018-2020  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
    
    2019-04-11    Initial release.
    2019-05-29    Added "Emulating..." debug output.
    2019-06-03    Added {BOARD_PARAMS_REV} and {BOARD_PARAMS_TYPE}.
    2019-06-24    Added {BOARD_BATTERY_THRESH}.
    2020-06-04    Introduced {BOARD_TYPE_MAX} and added support for type 2.
    2020-06-08    Check for invalid config combinations.
    2020-07-30    Remove most magic numbers.
    2020-07-30    Added {board_cte_antennas()}.
    2020-07-30    Distinguish more clearly between board and chip revision.
    2020-07-31    Moved {board_cte_antennas()} to the new cte module.
    2020-08-05    Remove misplaced comma within call to {board_hello_world()}.
  
  Description
  ===========
    
    Initialization of this module should be done by calling {board_init()}. This
    reads all supported board-specific parameters from the non-volatile User
    Information Configuration Registers (UICR).
    Each of them may be accessed by any part of the firmware by calling the
    corresponding getter function. Setting the whole bunch of parameters is done
    by calling {board_store_params()}. The supported parameters are:
     
      {id}             board ID (a.k.a. RSS2I device ID) in the range
                       [1,{BOARD_ID_ALWAYS_LAST_ITEM})
      {rev}            board revision (hardware revision) in the range
                       [0,{BOARD_REV_ALWAYS_LAST_ITEM})
      {type}           board type in the range [0,{BOARD_TYPE_ALWAYS_LAST_ITEM})
      {batteryThresh}  board-specific ADC battery warning threshold in the range
                       [0, 1024).
*/

#include <stdbool.h>
#include <stdint.h>

#include <rss2i/board.h>
#include <rss2i/rev.h>
#include <rss2i/print.h>
#include <rss2i/debug.h>

#include <rss2i/config/board.h>

#include <libopencm3/nrf5/nvmc.h>
#include <libopencm3/nrf5/uicr.h>

enum BoardParams {
	
	BOARD_PARAMS_ID,
	BOARD_PARAMS_REV,
	BOARD_PARAMS_TYPE,
	BOARD_PARAMS_BATTERY_THRESH,
	
	BOARD_PARAMS_ALWAYS_LAST_ITEM
};

static uint32_t boardParams[BOARD_PARAMS_ALWAYS_LAST_ITEM] = {};
static bool boardInitialized = false;


uint8_t board_id(void){
/*
  Returns the board ID.
*/
	
	#ifdef BOARD_ID           /* Hard coded...        */
	
		return BOARD_ID;
	#else                     /* From UICR memory... */
	
		if(!boardInitialized){
		
			debug_report(DEBUG_BOARD_NOT_READ_YET);
			return 0;
		}
		else if(boardParams[BOARD_PARAMS_ID] < 1
		     || boardParams[BOARD_PARAMS_ID] >= BOARD_ID_ALWAYS_LAST_ITEM){
		
			debug_report(DEBUG_BOARD_ID_INVALID);
			return 0;
		}
		else{
			
			return boardParams[BOARD_PARAMS_ID];
		}
	#endif
}

uint8_t board_type(void){
/*
  Returns the board (antenna) type.
*/
	
	#ifdef BOARD_TYPE         /* Hard coded...        */
	
		if(BOARD_TYPE >= BOARD_TYPE_ARR && board_rev() < BOARD_REV_F){
		
			debug_report(DEBUG_BOARD_TYPE_NOT_EMULATABLE);
			return 0;
		}
		
		return BOARD_TYPE;
	#else                     /* From UICR memory... */
	
		if(!boardInitialized){
		
			debug_report(DEBUG_BOARD_NOT_READ_YET);
			return 0;
		}
		else if(boardParams[BOARD_PARAMS_TYPE] >= BOARD_TYPE_ALWAYS_LAST_ITEM){
		
			debug_report(DEBUG_BOARD_TYPE_INVALID);
			return 0;
		}
		else if(boardParams[BOARD_PARAMS_TYPE] >= BOARD_TYPE_ARR
		     && board_rev() < BOARD_REV_F){
		
			debug_report(DEBUG_BOARD_TYPE_NOT_EMULATABLE);
			return 0;
		}
		else{
			
			return boardParams[BOARD_PARAMS_TYPE];
		}
	#endif
}

uint8_t board_rev(void){
/*
  Returns the board's hardware revision.
*/
	
	#ifdef BOARD_REV          /* Hard coded...        */
	
		/* We do not have to check against invalid chip revisions          */
		/* since this is checked already during compile time in this case. */
		
		if(
			/* A cannot emulate >= B. */
			(
				   boardInitialized
				&& boardParams[BOARD_PARAMS_REV] == BOARD_REV_A
				&& BOARD_REV > BOARD_REV_A
			)
		||
			/* >=B cannot emulate A. */
			(
				   boardInitialized
				&& boardParams[BOARD_PARAMS_REV] > BOARD_REV_A
				&& boardParams[BOARD_PARAMS_REV] < BOARD_REV_ALWAYS_LAST_ITEM
				&& BOARD_REV == BOARD_REV_A
			)
		){
			
			debug_report(DEBUG_BOARD_REV_NOT_EMULATABLE);
			return 0;
		}
			
		return BOARD_REV;
	#else                     /* From UICR memory... */
	
		if(!boardInitialized){
		
			debug_report(DEBUG_BOARD_NOT_READ_YET);
			return 0;
		}
		else if(boardParams[BOARD_PARAMS_REV] >= BOARD_REV_ALWAYS_LAST_ITEM){
		
			debug_report(DEBUG_BOARD_REV_INVALID);
			return 0;
		}
		else if(!rev_is_compatible_with(boardParams[BOARD_PARAMS_REV])){
		
			debug_report(DEBUG_BOARD_REV_NOT_EMULATABLE);
			return 0;
		}
		else{
			
			return boardParams[BOARD_PARAMS_REV];
		}
	#endif
}

uint16_t board_battery_threshold(void){
/*
  Returns the board-specific ADC battery warning threshold.
*/
	
	#ifdef BOARD_BATTERY_THRESHOLD      /* Hard coded...        */
	
		return BOARD_BATTERY_THRESHOLD;
	#else                               /* From UICR memory... */
	
		if(!boardInitialized){
		
			debug_report(DEBUG_BOARD_NOT_READ_YET);
			return 0;
		}
		else if(boardParams[BOARD_PARAMS_BATTERY_THRESH] > 1023){
		
			debug_report(DEBUG_BOARD_BATTERY_THRESH_INVALID);
			return 0;
		}
		else{
			
			return boardParams[BOARD_PARAMS_BATTERY_THRESH];
		}
	#endif
}

bool board_store_params(uint8_t id,
                        uint8_t rev,
                        uint8_t type,
                        uint16_t batteryThresh){
/*
  Stores the specified parameters inside the non-volatile flash memory.
  The supported parameters are:
     
      {id}             board ID (a.k.a. RSS2I device ID) in the range
                       [1,{BOARD_ID_ALWAYS_LAST_ITEM})
      {rev}            board revision (hardware revision) in the range
                       [0,{BOARD_REV_ALWAYS_LAST_ITEM})
      {type}           board type in the range [0,{BOARD_TYPE_ALWAYS_LAST_ITEM})
      {batteryThresh}  board-specific ADC battery warning threshold in the range
                       [0, 1024).
  
  Returns {true} in case of success (i.e. all supplied parameters are in-range
  and this module is initialized).
  
  Note that the arguments are not checked against invalid configuration
  combinations.
*/
	
	uint8_t i;
	
	if(boardInitialized && id >= 1 && id < BOARD_ID_ALWAYS_LAST_ITEM
	                    && type < BOARD_TYPE_ALWAYS_LAST_ITEM
	                    && rev < BOARD_REV_ALWAYS_LAST_ITEM){
		
		if(boardParams[BOARD_PARAMS_ID]   != id
		|| boardParams[BOARD_PARAMS_REV]  != rev
		|| boardParams[BOARD_PARAMS_TYPE] != type
		|| boardParams[BOARD_PARAMS_BATTERY_THRESH] != batteryThresh){
			
			/* At least one board-specific parameter should be changed. */
			
			boardParams[BOARD_PARAMS_ID] = id;
			boardParams[BOARD_PARAMS_REV] = rev;
			boardParams[BOARD_PARAMS_TYPE] = type;
			boardParams[BOARD_PARAMS_BATTERY_THRESH] = batteryThresh;
		
			NVMC_CONFIG = NVMC_CONFIG_WEN_Een;
			while(!NVMC_READY);
			NVMC_ERASEUICR = NVMC_ERASEUICR_ERASEUICR_Erase;
			while(!NVMC_READY);
			NVMC_CONFIG = NVMC_CONFIG_WEN_Wen;
			while(!NVMC_READY);
			
			/* Disabled NFC functionality for Rev. A boards in order to */
			/* enable UART over USB communication capabilities.         */
			
			if(rev == 0){
				
				UICR_NFCPINS = 0;
			}
			
			/* Store board parameters. */
			
			for(i = 0; i < BOARD_PARAMS_ALWAYS_LAST_ITEM && i < 32; ++i){
			
				UICR_CUSTOMER[i] = boardParams[i];
				while(!NVMC_READY);
			}
			
			NVMC_CONFIG = NVMC_CONFIG_WEN_Ren;
			while(!NVMC_READY);
			
			debug("store", "Successfully stored new board parameters.");
			
			return true;
		}
		else{
			
			debug("store", "Specified board parameters were already configured.");
			return true;
		}
	}
	else{
		
		debug("store", "Specifying board parameters FAILED.");
		return false;
	}
}

static void board_hello_world(uint8_t id,
                              uint8_t rev,
                              uint8_t type,
                              uint16_t batteryThresh){
/*
  Prints board info onto debug interface.
*/
	
	static char msg[] = "  ID:                 000\r\n"\
	                    "  Revision:            x\r\n" \
	                    "  Type:               uvw\r\n"\
	                    "  Battery threshold:  000\r\n";
	
	print_udec(id, msg, 22, 3);
	msg[50] = rev + 'A';
	
	if(type == 0){
	
		msg[75] = 'o';
		msg[76] = 'n';
		msg[77] = 'b';
	}
	else if(type == 1){
	
		msg[75] = 'e';
		msg[76] = 'x';
		msg[77] = 't';
	}
	else if(type == 2){
	
		msg[75] = 'a';
		msg[76] = 'r';
		msg[77] = 'r';
	}
	else{
	
		msg[75] = '?';
		msg[76] = '?';
		msg[77] = '?';
	}
	
	print_udec(batteryThresh, msg, 102, 3);
	
	debug_message(msg);
}

void board_init(void){
	
	uint8_t i;
	
	while(!NVMC_READY);
	
	for(i = 0; i < BOARD_PARAMS_ALWAYS_LAST_ITEM && i < 32; ++i){
	
		boardParams[i] = UICR_CUSTOMER[i];
	}
	
	debug_init();
	boardInitialized = true;
	
	debug("init", "Some information about this RSS2I device:\r\n");
	board_hello_world(boardParams[BOARD_PARAMS_ID],
	                  boardParams[BOARD_PARAMS_REV],
	                  boardParams[BOARD_PARAMS_TYPE],
	                  boardParams[BOARD_PARAMS_BATTERY_THRESH]);
	
	#if defined BOARD_ID \
	 || defined BOARD_REV \
	 || defined BOARD_TYPE \
	 || defined BOARD_BATTERY_THRESH
		
		debug_message("Emulating the following RSS2I device:\r\n");
		board_hello_world(
			#ifdef BOARD_ID
				/* Make sure to call {board_id()} in order to run checks. */
				(board_id(), BOARD_ID),
			#else
				board_id(),
			#endif
		
			#ifdef BOARD_REV
				/* Make sure to call {board_rev()} in order to run checks. */
				(board_rev(), BOARD_REV),
			#else
				board_rev(),
			#endif
		
			#if defined BOARD_TYPE
				/* Make sure to call {board_type()} in order to run checks. */
				(board_type(), BOARD_TYPE),
			#else
				board_type(),
			#endif
			
			#if defined BOARD_BATTERY_THRESH
				/* Make sure to call {board_battery_threshold()} */
				/* in order to run checks.                       */
				(board_battery_threshold(), BOARD_BATTERY_THRESH)
			#else
				board_battery_threshold()
			#endif
		);
	#endif
	
}

