/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

#define RSS2I_MODULE "buffer"

/*
  Thread safe (one producer, one consumer) circular buffers.
  
  Authors
  =======
  
    Copyright (C) 2018-2019  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
  
    2018-07-17    Initial release.
    2019-06-14    Report if content exceeds the buffer dimensions.
    2020-02-04    Add a terminating null character to a truncated string.
    2020-02-08    Remove race condition(s).
    2020-02-08    Add support for Direct Memory Access (DMA).
  
  Description
  ===========

    Create and use circular buffers for arrays of {char} or {uint8_t}. Writing
    to a buffer copies the content into the buffer's memory. However, a pending
    buffer entry can be read directly from the buffer's memory, which is
    protected against overwriting until the next entry gets unlocked explicitly.
    
    The buffer management doesn't allocate any memory. The user has to provide
    a sufficient amount of memory as well as the actual {Buffer} object. This
    allows the use of an arbitrary number of buffers without the need of
    dynamic memory allocation.
*/

#include <stdint.h>
#include <stdbool.h>
#include <stddef.h>

#include <rss2i/buffer.h>
#include <rss2i/debug.h>


void buffer_init_rows(Buffer * buffer, uint8_t * memory, size_t * size, size_t length, size_t width){
/*
  Initializes a specified buffer of data rows (i.e. {uint8_t} arrays).
  
  Argument list
  =============
  
    {buffer}           Buffer to be initialized.
    {memory}           Pointer to a section in memory with a minimal size of
                       {length * width}.
    {size}             {size_t} array with a minimal size of {length}.
    {length}           Number of buffer entries. A length of at least two
                       entries is required.
    {width}            Number of bytes within each buffer entry. A width of at
                       least one byte is required.
  
  Description
  ===========
    
    This is just a wrapper arround {buffer_init()} with the aim to simplify
    initialization of {uint8_t} array buffer.
*/
	
	BufferDataPtr dataPtr;
	dataPtr.u = memory;
	
	buffer_init(buffer, dataPtr, size, length, width);
}

void buffer_init_strings(Buffer * buffer, char * memory, size_t * size, size_t length, size_t width){
/*
  Initializes a specified buffer of strings.
  
  Argument list
  =============
  
    {buffer}           Buffer to be initialized.
    {memory}           Pointer to a section in memory with a minimal size of
                       {length * width}.
    {size}             {size_t} array with a minimal size of {length}.
    {length}           Number of buffer entries. A length of at least two
                       entries is required.
    {width}            Number of bytes within each buffer entry. A width of at
                       least one byte is required.
  
  Description
  ===========
    
    This is just a wrapper arround {buffer_init()} with the aim to simplify
    initialization of a {char} array buffer.
*/
	
	BufferDataPtr dataPtr;
	dataPtr.c = memory;
	
	buffer_init(buffer, dataPtr, size, length, width);
}

void buffer_init(Buffer * buffer, BufferDataPtr dataPtr, size_t * size, size_t length, size_t width){
/*
  Initializes a specified buffer.
  
  Argument list
  =============
  
    {buffer}           Buffer to be initialized.
    {dataPtr}          {BufferDataPtr} pointing to a section in memory with a
                       minimal size of {length * width}.
    {size}             {size_t} array with a minimal size of {length}.
    {length}           Number of buffer entries. A length of at least two
                       entries is required.
    {width}            Number of bytes within each buffer entry. A width of at
                       least one byte is required.
  
  Description
  ===========
    
    This function doesn't allocate any memory. The user has to ensure that the
    memory pointed to by {dataPtr} and {size} may be written without painful
    side effects. It is up to the user whether the specified memory holds
    {char}s or {uint8_t}s. A standard compliant way to access {char} and
    {uint8_t} arrays is to use the {*string*()} and {*row*()} functions
    respectively. If the user uses a compiler extension that allows cross-access
    of unions it may be safe to mix both data types.
    
    Please note that this buffer is guaranteed to be free of race conditions
    only if there is just one producer (who is writing into the buffer) and one
    consumer (who is reading from the buffer) at the same time. If there are two
    or more producers the user has to ensure that those do not interrupt each
    other. The same holds in the case of two or more consumers.
    
    Using a buffer length of 2^n, for positive n, may result in slightly faster
    operation depending on compiler optimization.
*/
	
	buffer->dataPtr = dataPtr;
	buffer->size = size;
	buffer->length = length;
	buffer->width = width;
	
	buffer->readPointer = 0;
	buffer->writePointer = 0;
	buffer->lockPointer = 0;
}

bool buffer_append_row(Buffer * buffer, uint8_t * row, size_t size){
/*
  Appends a data row (i.e. an {uint8_t} array) to the specified buffer. The size
  of the specified row has to be given explicitly. If necessary, it will be
  truncated to the buffer's width.
  
  Argument list
  =============
  
    {buffer}           Buffer to be appended to.
    {row}              {uint8_t} array containing the data row to be appended.
    {size}             Number of bytes to be stored.

  Return value
  ============
  
    Boolean {true} if the specified data row has been appended successfully.
    Returns {false} if an overflow would happen when appending the data.
*/	
	
	size_t i;
	
	size_t writePtr = buffer->writePointer % buffer->length;
	size_t lockPtr = buffer->lockPointer % buffer->length;
	
	/* If the {writePointer} differs from the {lockPointer} but both point to */
	/* the same entry (i.e. they are equal mod {length}), the buffer is full. */
	if(writePtr == lockPtr && buffer->writePointer != buffer->lockPointer){
	
		/* There would be a buffer overflow */
		return false;
	}
	else{
		
		/* Store data */
		
		if(size > buffer->width){
			
			size = buffer->width;
			debug_report(DEBUG_BUFFER_WIDTH_EXCEEDED);
		}
		
		for(i = 0; i < size; ++i){
			
			((buffer->dataPtr).u)[writePtr * buffer->width + i] = row[i];
		}
		
		(buffer->size)[writePtr] = size;
		
		/* Append data */
		
		++buffer->writePointer;
		
		return true;
	}
}

bool buffer_append_string(Buffer * buffer, const char * string){
/*
  Appends a string (i.e. a {char} array), including the terminating null
  character, to the specified buffer. The size of the string gets detected
  automatically. If necessary, the string gets truncated to the buffer's width.
  
  Argument list
  =============
  
    {buffer}           Buffer to be appended to.
    {string}           {char} array containing the string to be appended.

  Return value
  ============
  
    Boolean {true} if the specified string has been appended successfully.
    Returns {false} if there would have been a buffer overflow when appending
    the string to the buffer (i.e. the buffer is full).
*/	
	size_t i;
	
	size_t writePtr = buffer->writePointer % buffer->length;
	size_t lockPtr = buffer->lockPointer % buffer->length;
	
	/* If the {writePointer} differs from the {lockPointer} but both point to */
	/* the same entry (i.e. they are equal mod {length}), the buffer is full. */
	if(writePtr == lockPtr && buffer->writePointer != buffer->lockPointer){
		
		/* There would be a buffer overflow */
		return false;
	}
	else{
		
		/* Store data */
		
		i = 0;
		while(i < buffer->width && (((buffer->dataPtr).c)[writePtr * buffer->width + i] = string[i])){
			
			++i;
		}
		
		if(i == buffer->width){
			
			/* Truncate string. */
			((buffer->dataPtr).c)[writePtr * buffer->width + i - 1] = '\0';
			debug_report(DEBUG_BUFFER_WIDTH_EXCEEDED);
		}
		else{
			
			++i;
		}
		
		(buffer->size)[writePtr] = i;
		
		/* Append data */
		
		++buffer->writePointer;
		
		return true;
	}
}

uint8_t * buffer_new_row_address(Buffer * buffer){
/*
  Returns the address of the next free buffer entry. The user may write to this
  memory and afterwards call {buffer_new_row_finish()} in order to append the
  written row to the buffer.
  Note, that the user is responsible for not exceeding the buffer's width. Also
  note, that data that was written to the memory, which the return value points
  to, is not stored on the buffer until {buffer_new_row_finish()} gets called. A
  call to {buffer_append_row()} instead of {buffer_new_row_finish()} overwrites
  this already written data.
  
  Argument list
  =============
  
    {buffer}           Buffer to append to.

  Return value
  ============
  
    Address of the next free buffer entry. Returns {NULL} if there isn't any.
*/
	
	size_t writePtr = buffer->writePointer % buffer->length;
	size_t lockPtr = buffer->lockPointer % buffer->length;
	
	/* If the {writePointer} differs from the {lockPointer} but both point to */
	/* the same entry (i.e. they are equal mod {length}), the buffer is full. */
	if(writePtr == lockPtr && buffer->writePointer != buffer->lockPointer){
		
		/* There would be a buffer overflow */
		return NULL;
	}
	else{
		
		return (uint8_t *) &(((buffer->dataPtr).u)[writePtr * buffer->width]);
	}
}

bool buffer_new_row_finish(Buffer * buffer, uint8_t * addr, size_t size){
/*
  Appends the data, which was already written to the address of the next free
  buffer entry, to the buffer (see {buffer_new_row_address()}). Although the
  specified {size} is checked to comply with the buffer's width and is reduced
  if not, the user is responsible to take care of the memory limits and entry
  boundaries.
  While it is OK to overwrite already written data (i.e. not calling this
  function after writing to the address returned by {buffer_new_row_address()}),
  there should be no buffer activity (i.e. function calls) between the call of
  {buffer_new_row_address()}) and the call of this function. Doing so may result
  in a nonsense buffer entry.
  
  Argument list
  =============
  
    {buffer}           Buffer to be appended to.
    {addr}             Memory address where the row data was written to. This is
                       used for checking for buffer activities only.
    {size}             Size of the previously written data.
  
  Return value
  ============
    
    Returns {false} if there were buffer activities between the call of
    {buffer_new_row_address()}) and the call of this function. No entry gets
    appended to the buffer in this case. Note, that this function cannot
    reliably detect those activities in all cases. Returns {true} otherwise.
*/
	
	size_t writePtr = buffer->writePointer % buffer->length;
	size_t lockPtr = buffer->lockPointer % buffer->length;
	
	/* If the {writePointer} differs from the {lockPointer} but both point to */
	/* the same entry (i.e. they are equal mod {length}), the buffer is full. */
	if((writePtr == lockPtr && buffer->writePointer != buffer->lockPointer)
	|| &(((buffer->dataPtr).u)[writePtr * buffer->width]) != addr){
		
		/* The buffer would overflow OR the specified {addr}ess differs */
		/* from that one of the next free buffer entry. Both means that */
		/* there were activities on this buffer after the previous call */
		/* to {buffer_new_row_address()}.                               */
		debug_report(DEBUG_BUFFER_NOTHING_TO_FINISH);
		return false;
	}
	else{
	
		if(size > buffer->width){
			
			size = buffer->width;
			/* What could the user have done?! */
			debug_report(DEBUG_BUFFER_WIDTH_EXCEEDED);
		}
		
		(buffer->size)[writePtr] = size;
		
		/* Append data */
		
		++buffer->writePointer;
		
		return true;
	}
}

BufferDataPtr buffer_get(Buffer * buffer, BufferDataType type){
/*
  Reads the first pending entry of the specified circular buffer.
  
  Argument list
  =============
  
    {buffer}           Buffer to be read from.
    {type}             Type to be returned.

  Return value
  ============
  
    A {BufferDataPtr} containing a {uint8_t} or {char} array (depending on the
    type specified). Please note that this memory lies inside the circular
    buffer and therefore may be overwritten at some later point. Nevertheless it
    is guaranteed that this part of memory will hold its data at least until the
    next read request to the specified buffer. You may call {buffer_unlock()} in
    order to unlock it earlier. Note, that requesting the size of the next
    pending entry using {buffer_next_size()} is treated as read request in this
    context.
*/
	
	BufferDataPtr dataPtr;
	size_t lockPtr;
	
	buffer_unlock(buffer);
	
	/* If the {lockPointer} equals the {writePointer}, the buffer is empty. */
	if(buffer->lockPointer == buffer->writePointer){
	
		if(type == BUFFER_ROW){
			
			dataPtr.u = NULL;
		}
		else{
			
			dataPtr.c = NULL;
		}
	}
	else{
		
		lockPtr = buffer->lockPointer % buffer->length;
		
		++buffer->readPointer;
		
		if(type == BUFFER_ROW){
			
			dataPtr.u = (buffer->dataPtr).u + lockPtr * buffer->width;
		}
		else{
			
			dataPtr.c = (buffer->dataPtr).c + lockPtr * buffer->width;
		}
	}
	
	return dataPtr;
}

uint8_t * buffer_get_row(Buffer * buffer){
/*
  Reads the first pending data row of the specified circular buffer.
  
  Argument list
  =============
  
    {buffer}           Buffer to be read from.

  Return value
  ============
  
    An array of {uint8_t} containing the requested data. Please note that this
    memory lies inside the circular buffer and therefore may be overwritten at
    some later point. Nevertheless it is guaranteed that this part of memory
    will hold its data at least until the next read request to the specified
    buffer. You may call {buffer_unlock()} in order to unlock it earlier. Note,
    that requesting the size of the next pending entry using
    {buffer_next_size()} is treated as read request in this context.
*/
  	
	BufferDataPtr dataPtr = buffer_get(buffer, BUFFER_ROW);
	
	return (uint8_t *) dataPtr.u;
}

char * buffer_get_string(Buffer * buffer){
/*
  Reads the first pending string of the specified circular buffer.
  
  Argument list
  =============
  
    {buffer}           Buffer to be read from.

  Return value
  ============
  
    An array of {char} containing the requested string. Please note that this
    memory lies inside the circular buffer and therefore may be overwritten at
    some later point. Nevertheless it is guaranteed that this part of memory
    will hold its data at least until the next read request to the specified
    buffer. You may call {buffer_unlock()} in order to unlock it earlier. Note,
    that requesting the size of the next pending entry using
    {buffer_next_size()} is treated as read request in this context.
*/
  
	BufferDataPtr dataPtr = buffer_get(buffer, BUFFER_STRING);
	
	return (char *) dataPtr.c;
}

void buffer_unlock(Buffer * buffer){
/*
  Unlocks the previously read entry of the specified buffer (if any).
  
  Argument list
  =============
  
    {buffer}           Buffer.
*/

	if(buffer->lockPointer != buffer->readPointer){
		
		/* unlock/release previous position in the buffer */
		buffer->lockPointer = buffer->readPointer;
	}
}

void buffer_empty(Buffer * buffer){
/*
  Empties the specified buffer.
  
  Argument list
  =============
  
    {buffer}           Buffer to be emptied.
*/
	
	buffer->readPointer = 0;
	buffer->writePointer = 0;
	buffer->lockPointer = 0;
}


size_t buffer_next_size(Buffer * buffer){
/*
  Returns the size of the first pending entry of the specified circular buffer.
  
  Argument list
  =============
  
    {buffer}           Buffer to be read from.

  Return value
  ============
  
    The size of the first pending entry of the specified circular buffer. This
    size is stored inside the buffer and therefore doesn't need to be calculated
    when calling this function. Please note that the size of a string appended
    to this buffer via {buffer_append_string()} gets computed including the
    terminating null character (i.e. NOT like {strlen()} does).
    
    If there is no pending entry, {0} will be returned.
*/
	
	size_t readPtr = buffer->readPointer % buffer->length;
	
	buffer_unlock(buffer);
	
	return (buffer->lockPointer == buffer->writePointer) ? 0 : (buffer->size)[readPtr];
}

size_t buffer_vacant_entries(Buffer * buffer){
/*
  Returns the number of vacant entries within the specified circular buffer.
  
  Argument list
  =============
  
    {buffer}           Buffer to be read from.
    
*/
	
	size_t writePtr = buffer->writePointer % buffer->length;
	size_t lockPtr = buffer->lockPointer % buffer->length;
	
	if(buffer->writePointer == buffer->lockPointer){
		
		return buffer->length;
	}
	else if(writePtr > lockPtr){
		
		return lockPtr + buffer->length - writePtr;
	}
	else{
		
		return lockPtr - writePtr;
	}
}


