/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

#define RSS2I_MODULE "clock"

/*
  Selects the external crystal oscillator as source for the high frequency
  clock (HFCLK).
  
  Authors
  =======
    
    Copyright (C) 2018-2019  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
    
    2019-04-12    Initial release.
    2019-07-12    Added triggering of the timer module.
    
   Description
   ===========
     
     Calling {clock_init()} does all the magick.
*/

#include <stdbool.h>

#include <libopencm3/nrf5/clock.h>

#include <rss2i/clock.h>
#include <rss2i/timer.h>
#include <rss2i/debug.h>


void clock_init(void){
/*
  Selects the external crystal oscillator as source for the high frequency
  clock.
  
  This function waits until the crystal is guaranteed to be working with a
  stable frequency. If this doesn't happen within a maximum ramp-up time of
  2 ms, the debug system gets alerted.
*/
	
	volatile bool timeout = false;
	
	debug_init();
	
	timer_flag(&timeout, 2, TIMER_NO_REPEAT);
	
	CLOCK_EVENTS_HFCLKSTARTED = 0;
	CLOCK_TASKS_HFCLKSTART = 1;
	
	while(!CLOCK_EVENTS_HFCLKSTARTED && !timeout){
		
		/* This is a bit hacky: Since this function is usually called before */
		/* the main loop starts triggering all modules (including the timer) */
		/* periodically, we have to do it on ower own.                       */
		timer_trigger();
	}
	
	if(!CLOCK_EVENTS_HFCLKSTARTED){
		
		debug_report(DEBUG_CRYSTAL_OSCILLATOR_FAILED);
	}
	else{
		
		debug("init", "Crystal oscillator ramped up successfully.");
	}
}
