/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

#define RSS2I_MODULE "cmd"

/*
  Process incoming commands.
  
  Authors
  =======
    
    Copyright (C) 2018-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
    
    2019-04-16    Initial public release.
    2019-05-02    Added error code {0x04} within {CMD_START_UP_MEASUREMENT}.
    2019-05-29    Added support for iterable commands.
    2019-10-31    Fitted to the changed baudrate change procedure (UARTE).
    2019-11-04    Split page reading into multiple tasks for the sake of lower
                  blocking times. Took the memory controller load into account
                  before requesting a page read. This partially fixed #6.
    2020-03-31    Added commands 0x18 {CMD_RETURN_DEVICE_STATUS} as well as
                  0x19 {CMD_REPORT_ERROR_INFORMATION}.
    2020-07-06    Split a USB respond into two tasks for lower blocking times.
    2020-07-30    Add CTE related information to 0x00 {TELL_ME_WHO_YOU_ARE} cmd.
    2020-07-31    Made some changes related to the new cte module.
    2020-07-31    Made some changes related to the new livelog module.
    2020-08-03    Added {CMD_SET_CTE_CONFIGURATION}.
    2020-08-14    Added {CMD_*CTE_SAMPLE*} commands.
    2020-08-25    Added iterable version of {CMD_RETURN_CTE_SAMPLE_SHEET}.
    2021-03-19    Added a target field to the {CMD_TELL_ME_WHO_YOU_ARE}.
    2021-03-23    Moved 0x16 to 0x17 and Added new 0x16 {CMD_SET_RADIO_MODE}.
    2021-04-01    Enhanced the maximal number of error cases significantly.
    2021-03-23    Moved 0x18 to 0x19 and and 0x19 to 0x1a.
    2021-04-19    Renamed 0x1b to 0x1f.
    2021-04-20    Added 0x18 {CMD_SET_SAMPLE_CONFIGURATION}.
    2021-05-04    Fit {remainingTrials} to enlarged frame lengths.
  
  Description
  ===========
    
    ALL the command magick is done here.
*/

#include <stdint.h>
#include <stdbool.h>
#include <stddef.h>

#include <rss2i/cmd.h>
#include <rss2i/usb.h>
#include <rss2i/debug.h>
#include <rss2i/crc.h>
#include <rss2i/measure.h>
#include <rss2i/livelog.h>
#include <rss2i/radio.h>
#include <rss2i/uarte.h>
#include <rss2i/memory.h>
#include <rss2i/battery.h>
#include <rss2i/mlc.h>
#include <rss2i/timer.h>
#include <rss2i/control.h>
#include <rss2i/board.h>
#include <rss2i/cte.h>
#include <rss2i/version.h>
#include <rss2i/print.h>
#include <rss2i/postproc.h>

#include <rss2i/config/cmd.h>


const CMD cmds[CMD_ALWAYS_LAST_ITEM] = {
  /*                                                                                                                              */
  /* CLen      Command length (excl. start byte, CRC bytes, fill-up bytes).                                                       */
  /* RLen      Response length (excl. CRC bytes).                                                                                 */
  /* TaFld?    Whether there is a target field or not.                                                                            */
  /* Resp?     Whether there should be a response or not.                                                                         */
  /* Iter?     Whether this command is iterable or not.                                                                           */
  /* TOLen     Length of the OTA timeout data.                                                                                    */
  /* TODat     Content of the OTA timeout response. The rest of the response will have arbitrary values.                          */
  /* BsyLen    Length of the "control mode recquired" data.                                                                       */
  /* BsyDat    Content of the "control mode required" response. The rest of the response will have arbitrary values.              */
  /*                                                                                                                              */
  /*                                                                                                                              */
  /*                                            |      CLen RLen  TaFld? Resp?  Iter?  TOLen    TODat      BsyLen     BsyDat      */
  /* -------------------------------------------+-------------------------------------------------------------------------------- */
  /* 0x00: Tell me who you are.                 | */   {  2,  10,  true,  true, false, 1, (uint8_t[1]) {1}, 1, (uint8_t[1]) {2}},
  /* 0x01: Set Baudrate.                        | */   {  2,   1, false,  true, false, 0,             NULL, 0,             NULL},
  /* 0x02: Set preferences.                     | */   { 10,   9,  true,  true, false, 1, (uint8_t[1]) {1}, 1, (uint8_t[1]) {2}},
  /* 0x03: Return measurement session data.     | */   {  2,   7,  true,  true, false, 1, (uint8_t[1]) {1}, 1, (uint8_t[1]) {2}},
  /* 0x04: Return content of specified page.    | */   {  5, 253,  true,  true, false, 1, (uint8_t[1]) {1}, 1, (uint8_t[1]) {2}},
  /* 0x05: Return content of specified pages.   | */   {  6, 253,  true,  true,  true, 1, (uint8_t[1]) {1}, 1, (uint8_t[1]) {2}}, /* ITERABLE */
  /* 0x06: Return content of absolute page.     | */   {  5, 253,  true,  true, false, 1, (uint8_t[1]) {1}, 1, (uint8_t[1]) {2}},
  /* 0x07: Return content of absolute pages.    | */   {  6, 253,  true,  true,  true, 1, (uint8_t[1]) {1}, 1, (uint8_t[1]) {2}}, /* ITERABLE */
  /* 0x08: Set live log configuration.          | */   { 26,   1, false,  true, false, 0,             NULL, 0,             NULL},
  /* 0x09: Start live log.                      | */   {  1,   1, false,  true, false, 0,             NULL, 0,             NULL},
  /* 0x0a: Stop live log.                       | */   {  1,   1, false,  true, false, 0,             NULL, 0,             NULL},
  /* 0x0b: Enter control mode.                  | */   {  1,   1, false,  true, false, 0,             NULL, 0,             NULL},
  /* 0x0c: Force control mode.                  | */   {  1,   1, false,  true, false, 0,             NULL, 0,             NULL},
  /* 0x0d: Jump on measurement.                 | */   {  2,   4,  true,  true, false, 1, (uint8_t[1]) {1}, 1, (uint8_t[1]) {2}},
  /* 0x0e: Start up measurement.                | */   {  2,   4,  true,  true, false, 1, (uint8_t[1]) {1}, 1, (uint8_t[1]) {2}},
  /* 0x0f: Reserved.                            | */   {  0,   0, false, false, false, 0,             NULL, 0,             NULL},
  /* 0x10: Set measurement configuration.       | */   {256,   1,  true,  true, false, 1, (uint8_t[1]) {1}, 1, (uint8_t[1]) {2}},
  /* 0x11: Set synchronization map.             | */   {253,   1,  true,  true, false, 1, (uint8_t[1]) {1}, 1, (uint8_t[1]) {2}},
  /* 0x12: Set poor-sync thresholds.            | */   {252,   1,  true,  true, false, 1, (uint8_t[1]) {1}, 1, (uint8_t[1]) {2}},
  /* 0x13: Set out-of-sync thresholds.          | */   {252,   1,  true,  true, false, 1, (uint8_t[1]) {1}, 1, (uint8_t[1]) {2}},
  /* 0x14: Set remote RSS input map.            | */   {252,   1,  true,  true, false, 1, (uint8_t[1]) {1}, 1, (uint8_t[1]) {2}},
  /* 0x15: Set remote RSS output map.           | */   {253,   1,  true,  true, false, 1, (uint8_t[1]) {1}, 1, (uint8_t[1]) {2}},
  /* 0x16: Set radio mode.                      | */   {  6,   1,  true,  true, false, 1, (uint8_t[1]) {1}, 1, (uint8_t[1]) {2}},
  /* 0x17: Set CTE configuration.               | */   {253,   2,  true,  true, false, 1, (uint8_t[1]) {1}, 1, (uint8_t[1]) {2}},
  /* 0x18: Set sample configuration.            | */   {253,   2,  true,  true, false, 1, (uint8_t[1]) {1}, 1, (uint8_t[1]) {2}},
  /* 0x19: Return device status.                | */   {  2,  12,  true,  true, false, 1, (uint8_t[1]) {1}, 1, (uint8_t[1]) {2}},
  /* 0x1a: Report error information.            | */   {  3, 255,  true,  true, false, 1, (uint8_t[1]) {1}, 1, (uint8_t[1]) {2}},
  /* 0x1b: Collect samples.                     | */   {  2,   1, false,  true, false, 0,             NULL, 0,             NULL},
  /* 0x1c: Return sample status.                | */   {  1,   1, false,  true, false, 0,             NULL, 0,             NULL},
  /* 0x1d: Clear sample request.                | */   {  1,   1, false,  true, false, 0,             NULL, 0,             NULL},
  /* 0x1e: Return sample sheet.                 | */   {  2, 253, false,  true, false, 0,             NULL, 0,             NULL},
  /* 0x1f: Return sample sheets.                | */   {  3, 253, false,  true,  true, 0,             NULL, 0,             NULL}, /* ITERABLE */
};


uint8_t cmd_is_iterable(uint8_t * cmd){
/*
  Returns {0} if the specified {cmd} is not iterable.
  
  If the given {cmd} is iterable, this function returns the number of iterations
  that was specified inside that command. In this case the memory pointed to by
  {cmd} gets modified in order to equal its non-iterable counterpart.
*/

	uint8_t cmdIndex = cmd[0];
	
	if(cmdIndex == CMD_RETURN_CONTENT_OF_SPECIFIED_PAGES){
		
		cmd[0] = CMD_RETURN_CONTENT_OF_SPECIFIED_PAGE;
		
		return cmd[cmds[cmdIndex].length - 1];
	}
	else if(cmdIndex == CMD_RETURN_CONTENT_OF_ABSOLUTE_PAGES){
		
		cmd[0] = CMD_RETURN_CONTENT_OF_ABSOLUTE_PAGE;
		
		return cmd[cmds[cmdIndex].length - 1];
	}
	else if(cmdIndex == CMD_RETURN_SAMPLE_SHEETS){
		
		cmd[0] = CMD_RETURN_SAMPLE_SHEET;
		
		return cmd[cmds[cmdIndex].length - 1];
	}
	else{
		
		return 0;
	}
}

void cmd_iterate(uint8_t * cmd){
/*
  Given the non-iterable counterpart ({cmd}) of an iterable command, this
  function increments the relevant index inside the memory pointed to by {cmd}.
  
  Does nothing if the specified {cmd} has no iterable counterpart.
*/

	uint32_t index;
	
	if(cmd[0] == CMD_RETURN_CONTENT_OF_SPECIFIED_PAGE
	|| cmd[0] == CMD_RETURN_CONTENT_OF_ABSOLUTE_PAGE){
		
		index = (uint32_t) cmd[2] <<  0
		      | (uint32_t) cmd[3] <<  8
		      | (uint32_t) cmd[4] << 16;
		
		++index;
		
		cmd[2] = index >>  0;
		cmd[3] = index >>  8;
		cmd[4] = index >> 16;
	}
	else if(cmd[0] == CMD_RETURN_SAMPLE_SHEET){
		
		cmd[1] += 1;
	}
}

bool cmd_handle(uint8_t * cmd){
/*
  Handles a command.
  
  Argument List
  =============
  
    {cmd}    Address of the command to be handled. After specified once it may
             be {NULL} during future calls to this function until the initially
             given command got handled completely (see below). Nevertheless the
             memory pointed to by {cmd} should meanwhile retain its content.
  
  Return Value
  ============
  
    Returns {true} when the handling of the (previously) specified command is
    still in process and this function should hence get called at least one more
    time. Returns {false} when the (previously) specified command got handled
    completely.
*/
	
	
	typedef enum {
		WAIT_FOR_NOTHING,
		WAIT_FOR_MEMORY_BUSY,
		WAIT_FOR_MEMORY_REQUEST,
		WAIT_STATE_1,
		WAIT_STATE_2,
		WAIT_STATE_PREPARE_RESPOND,
		
	} WaitState;

	static uint8_t response[CMD_MAX_RESPONSE_LENGTH + 2]; /* +2: Refer to {radio_ota_send()}. */
	static uint8_t currentCMDIndex;
	static CMD currentCMD;
	static uint8_t currentTarget;
	static volatile bool timeout;
	static uint8_t remainingTrials;
	static bool ongoingProcess = false;
	static WaitState waitState;
	static uint8_t * page;
	static uint32_t address;
	static TimerID timer;
	static uint16_t rejected = 0;
	static char dbgMsg[] = "Memory request rejected 0000 times.";
	
	bool responseAvailable = true;
	
	uint32_t cycleCount;
	uint8_t frameCount;
	uint32_t pageCount;
	uint32_t firstPage;
	uint8_t status;
	uint8_t failPtr;
	uint16_t errors;
	uint16_t mlc;
	uint16_t raw;
	uint8_t bat;
	uint8_t mem;
	uint16_t eid;
	size_t i;
	
	if(!ongoingProcess && !cmd){
		
		/* The first processing call of this function must specify a command! */
		debug_report(DEBUG_CMD_INVALID_CMD);
	}
	else if(waitState != WAIT_STATE_PREPARE_RESPOND){
		
		if(!ongoingProcess){
			
			currentCMDIndex = cmd[0];
			currentCMD = cmds[currentCMDIndex];
			currentTarget = cmd[1];
		}
	
		switch(currentCMDIndex){
		
			  /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
			case CMD_TELL_ME_WHO_YOU_ARE:
			  /* ========================                                     */
			
				response[2] = 0;
				response[3] = board_id();
				response[4] = board_rev();
				response[5] = board_type();
				response[6] = version_high_byte();
				response[7] = version_mid_byte();
				response[8] = version_low_byte();
				response[9] = cte_num_of_antennas();
				response[10] = measure_is_ieee_250kbit_ready() ? 1 : 0;
				response[11] = 0;
		
			break;
			  /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
			case CMD_SET_BAUDRATE:
			  /* =================                                            */
				
				if(ongoingProcess){
				
					if(response[2] == UARTE_BAUDRATE_CHANGE_ONGOING){
					
						if((response[2] = uarte_change_baudrate(UARTE_BAUDRATE_CHANGE_ONGOING)) != UARTE_BAUDRATE_CHANGE_ONGOING){
							
							/* Wait before responding success. */
							timeout = false;
							timer_flag(&timeout, CMD_BAUDRATE_CHANGE_DELAY, TIMER_NO_REPEAT);
						}
					}
					else{
					
						if(timeout){
							ongoingProcess = false;
						}
					}
				}
				else{
					
					response[2] = uarte_change_baudrate(cmd[1]);
					ongoingProcess = true;
				}
			
			break;
			  /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
			case CMD_SET_PREFERENCES:
			  /* ====================                                         */
				
				if((
				    control_set_ota_tries_needs_control_mode(cmd[2])
				 || radio_set_tx_power_needs_control_mode(cmd[3])
				   )
				 && measure_running()){
					
					cmd_busy_response(currentCMDIndex, &response[2]);
				}
				else{
					
					response[2] = 0;
					response[3] = radio_set_tx_power(cmd[2]);
					response[4] = control_set_ota_tries(cmd[3]);
					response[5] = 0;
					response[6] = 0;
					response[7] = 0;
					response[8] = 0;
					response[9] = 0;
					response[10] = 0;
				}
			
			break;
			  /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
			case CMD_RETURN_MEASUREMENT_SESSION_DATA:
			  /* ====================================                         */
				
				response[2] = 0;
				
				pageCount = memory_get_page_count();
				firstPage = memory_get_start_address();
				
				response[3] = pageCount >>  0;
				response[4] = pageCount >>  8;
				response[5] = pageCount >> 16;
				response[6] = firstPage >>  0;
				response[7] = firstPage >>  8;
				response[8] = firstPage >> 16;
			
			break;
			  /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
			case CMD_RETURN_CONTENT_OF_SPECIFIED_PAGE:
			  /* =====================================                        */
				
				if(ongoingProcess){
					
					switch(waitState){
						
						case WAIT_FOR_MEMORY_BUSY:
							
							if(memory_ready_to_request()){
								
								timer_clear(timer);
								
								if(rejected){
								
									print_udec(rejected, dbgMsg, 24, 4);
									debug("read page", dbgMsg);
								}
								
								if(memory_request_offset(address)){
									/* Success. */
									waitState = WAIT_FOR_MEMORY_REQUEST;
								}
								else{
									/* Error. */
									response[2] = 3;
									ongoingProcess = false;
									waitState = WAIT_FOR_NOTHING;
								}
							}
							else if(timeout){
								
								/* Memory controller is too busy. Error */
								response[2] = 3;
								ongoingProcess = false;
								waitState = WAIT_FOR_NOTHING;
							}
							else{
								
								++rejected;
							}
						break;
						case WAIT_FOR_MEMORY_REQUEST:
					
							if(memory_request_ready()){
						
								if((page = memory_read())){
									/* Success. */
									response[2] = 0;
									waitState = WAIT_STATE_1;
								}
								else{
									/* Error. */
									response[2] = 4;
									ongoingProcess = false;
									waitState = WAIT_FOR_NOTHING;
								}
							}
						break;
						/* Process the read page and response to the command */
						/* within two additional processing tasks.           */
						case WAIT_STATE_1:
							
							for(i = 0; i < 252; ++i){
								response[3 + i] = page[i];
							}
							
							waitState = WAIT_STATE_2;
						break;
						case WAIT_STATE_2:
							
							ongoingProcess = false;
							waitState = WAIT_FOR_NOTHING;
						break;
						default:
							debug_report(DEBUG_CMD_INVALID_WAIT_STATE);
					}
				}
				else{
				
					address = ((uint32_t) cmd[2] <<  0)
					        | ((uint32_t) cmd[3] <<  8)
					        | ((uint32_t) cmd[4] << 16);
					
					timeout = false;
					rejected = 0;
					timer = timer_flag(&timeout, 300, TIMER_NO_REPEAT);
					waitState = WAIT_FOR_MEMORY_BUSY;
					ongoingProcess = true;
				}
		
			break;
			  /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
			case CMD_RETURN_CONTENT_OF_SPECIFIED_PAGES:
			  /* ======================================                       */
				
				debug_report(DEBUG_CMD_ITERABLE_CALLED);
		
			break;
			  /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
			case CMD_RETURN_CONTENT_OF_ABSOLUTE_PAGE:
			  /* ====================================                         */
			  
				if(ongoingProcess){
					
					switch(waitState){
						
						case WAIT_FOR_MEMORY_BUSY:
							
							if(memory_ready_to_request()){
								
								timer_clear(timer);
								
								if(rejected){
								
									print_udec(rejected, dbgMsg, 24, 4);
									debug("read page", dbgMsg);
								}
								
								if(memory_request_absolute(address)){
									/* Success. */
									waitState = WAIT_FOR_MEMORY_REQUEST;
								}
								else{
									/* Error. */
									response[2] = 3;
									ongoingProcess = false;
									waitState = WAIT_FOR_NOTHING;
								}
							}
							else if(timeout){
								
								/* Memory controller is too busy. Error */
								response[2] = 3;
								ongoingProcess = false;
								waitState = WAIT_FOR_NOTHING;
							}
							else{
								
								++rejected;
							}
						break;
						case WAIT_FOR_MEMORY_REQUEST:
					
							if(memory_request_ready()){
						
								if((page = memory_read())){
									/* Success. */
									response[2] = 0;
									waitState = WAIT_STATE_1;
								}
								else{
									/* Error. */
									response[2] = 4;
									ongoingProcess = false;
									waitState = WAIT_FOR_NOTHING;
								}
							}
						break;
						/* Process the read page and response to the command */
						/* within two additional processing tasks.           */
						case WAIT_STATE_1:
							
							for(i = 0; i < 252; ++i){
								response[3 + i] = page[i];
							}
							
							waitState = WAIT_STATE_2;
						break;
						case WAIT_STATE_2:
							
							ongoingProcess = false;
							waitState = WAIT_FOR_NOTHING;
						break;
						default:
							debug_report(DEBUG_CMD_INVALID_WAIT_STATE);
					}
				}
				else{
				
					address = ((uint32_t) cmd[2] <<  0)
					        | ((uint32_t) cmd[3] <<  8)
					        | ((uint32_t) cmd[4] << 16);
					
					timeout = false;
					rejected = 0;
					timer = timer_flag(&timeout, 300, TIMER_NO_REPEAT);
					waitState = WAIT_FOR_MEMORY_BUSY;
					ongoingProcess = true;
				}
		
			break;
			  /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
			case CMD_RETURN_CONTENT_OF_ABSOLUTE_PAGES:
			  /* =====================================                        */
				
				debug_report(DEBUG_CMD_ITERABLE_CALLED);
		
			break;
			  /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
			case CMD_SET_LIVE_LOG_CONFIGURATION:
			  /* ===============================                              */
		
				if(cmd[1] >= currentCMD.length - 2){
			
					response[2] = 1;
				}
				else{
			
					response[2] = livelog_config(cmd[1], &cmd[2]);
				}
		
			break;
			  /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
			case CMD_START_LIVE_LOG:
			  /* ===================                                          */
		
				timer_call(&livelog_enable, 100, TIMER_NO_REPEAT);
				response[2] = 0;
	
			break;
			  /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
			case CMD_STOP_LIVE_LOG:
			  /* ==================                                           */
				
				if(ongoingProcess){
					if(timeout){
					
						ongoingProcess = false;
						response[2] = 0;
					}
				}
				else{
				
					livelog_disable();
					timeout = false;
					timer_flag(&timeout, 200, TIMER_NO_REPEAT);
					ongoingProcess = true;
				}
	
			break;
			  /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
			case CMD_ENTER_CONTROL_MODE:
			  /* =======================                                      */
		
				response[2] = 0;
				measure_stop();
			
			break;
			  /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
			case CMD_FORCE_CONTROL_MODE:
			  /* =======================                                      */
				
				if(ongoingProcess){
					if(timeout){
						if(measure_disrupted()){
					
							response[2] = 0;
							ongoingProcess = false;
						}
						else{
						
							if(remainingTrials){
								
								--remainingTrials;
								timeout = false;
								timer_flag(&timeout, 100, TIMER_NO_REPEAT);
							}
							else{
							
								response[2] = 2;
								ongoingProcess = false;
							}
						}
					}
				}
				else{
					
					if(measure_disrupt()){
					
						remainingTrials = 150;
						timeout = false;
						timer_flag(&timeout, 100, TIMER_NO_REPEAT);
						ongoingProcess = true;
					}
					else{
						
						response[2] = 1;
					}
				}
				
			break;
			  /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
			case CMD_JUMP_ON_MEASUREMENT:
			  /* ========================                                     */
				
				if(measure_is_configured()){
					
					address = measure_prepare();
					
					if(address == 0xfffffffful){
						
						response[2] = 4;
					}
					else{
						
						response[2] = 0;
						response[3] = address >>  0;
						response[4] = address >>  8;
						response[5] = address >> 16;
						
						timer_call(&measure_jump_on, 100, TIMER_NO_REPEAT);
					}
				}
				else{
					
					response[2] = 3;
				}
	
			break;
			  /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
			case CMD_START_UP_MEASUREMENT:
			  /* =========================                                    */
				
				if(measure_is_configured()){
					
					if(measure_has_active_frame()){
				
						address = measure_prepare();
					
						if(address == 0xfffffffful){
						
							response[2] = 5;
						}
						else{
						
							response[2] = 0;
							response[3] = address >>  0;
							response[4] = address >>  8;
							response[5] = address >> 16;
						
							timer_call(&measure_start_up, 100, TIMER_NO_REPEAT);
						}
					}
					else{
					
						response[2] = 4;
					}
				}
				else{
					
					response[2] = 3;
				}
	
			break;
			  /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
			case CMD_SET_MEASUREMENT_CONFIGURATION:
			  /* ==================================                           */
		
				if(measure_running()){
					
					cmd_busy_response(currentCMDIndex, &response[2]);
				}
				else{
					
					frameCount = cmd[2];
					cycleCount = (uint32_t) cmd[3] <<  0
					           | (uint32_t) cmd[4] <<  8
					           | (uint32_t) cmd[5] << 16;
					
					status = measure_set_configuration(frameCount, cycleCount, &cmd[6]);
					response[2] = status ? status + 2 : 0;
				}
	
			break;
			  /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
			case CMD_SET_SYNCHRONIZATION_MAP:
			  /* ============================                                 */
		
				if(measure_running()){
					
					cmd_busy_response(currentCMDIndex, &response[2]);
				}
				else{
					
					status = measure_set_sync_map(cmd[2], &cmd[3]);
					response[2] = status ? status + 2 : 0;
				}
	
			break;
			  /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
			case CMD_SET_POOR_SYNC_THRESHOLDS:
			  /* =============================                                */
		
				if(measure_running()){
					
					cmd_busy_response(currentCMDIndex, &response[2]);
				}
				else{
					
					measure_set_poor_sync_thresh(&cmd[2]);
					response[2] = 0;
				}
	
			break;
			  /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
			case CMD_SET_OUT_OF_SYNC_THRESHOLDS:
			  /* ===============================                              */
		
				if(measure_running()){
					
					cmd_busy_response(currentCMDIndex, &response[2]);
				}
				else{
					
					measure_set_out_of_sync_thresh(&cmd[2]);
					response[2] = 0;
				}
	
			break;
			  /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
			case CMD_SET_REMOTE_RSS_INPUT_MAP:
			  /* =============================                                */
		
				if(measure_running()){
					
					cmd_busy_response(currentCMDIndex, &response[2]);
				}
				else{
					
					status = measure_set_remote_rss_input_map(&cmd[2]);
					response[2] = status ? status + 2 : 0;
				}
	
			break;
			  /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
			case CMD_SET_REMOTE_RSS_OUTPUT_MAP:
			  /* ==============================                               */
				
				if(measure_running()){
					
					cmd_busy_response(currentCMDIndex, &response[2]);
				}
				else{
					
					status = measure_set_remote_rss_output_map(cmd[2], &cmd[3]);
					response[2] = status ? status + 2 : 0;
				}
				
			break;
			  /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
			case CMD_SET_RADIO_MODE:
			  /* ===================                                          */
				
				if(measure_running()){
					
					cmd_busy_response(currentCMDIndex, &response[2]);
				}
				else{
					
					status = measure_set_radio_mode(
					         	cmd[2],
					         	cmd[3]
					         );
					response[2] = status ? status + 2 : 0;
				}
				
			break;
			  /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
			case CMD_SET_CTE_CONFIGURATION:
			  /* ==========================                                   */
				
				if(measure_running()){
					
					cmd_busy_response(currentCMDIndex, &response[2]);
				}
				else{
					
					failPtr = 255;
					status = measure_set_cte_configuration(
					         	cmd[2],
					         	&cmd[3],
					         	&failPtr
					         );
					response[2] = status ? status + 2 : 0;
					response[3] = failPtr;
				}
				
			break;
			  /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
			case CMD_SET_SAMPLE_CONFIGURATION:
			  /* =============================                                */
				
				if(measure_running()){
					
					cmd_busy_response(currentCMDIndex, &response[2]);
				}
				else{
					
					failPtr = 255;
					status = measure_set_sample_configuration(
					         	cmd[2],
					         	&cmd[3],
					         	&failPtr
					         );
					response[2] = status ? status + 2 : 0;
					response[3] = failPtr;
				}
				
			break;
			  /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
			case CMD_RETURN_DEVICE_STATUS:
			  /* =========================                                    */
		
				errors = debug_get_num_of_errors();
				mlc = mlc_get();
				raw = battery_get_value();
				bat = battery_get_level();
				mem = memory_get_load();
				
				response[2] = 0;
				response[3] = (uint8_t) (errors >> 0);
				response[4] = (uint8_t) (errors >> 8);
				response[5] = (uint8_t) (mlc >> 0);
				response[6] = (uint8_t) (mlc >> 8);
				response[7] = (uint8_t) (raw >> 0);
				response[8] = (uint8_t) (raw >> 8);
				response[9] = bat;
				response[10] = mem;
	
			break;
			  /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
			case CMD_REPORT_ERROR_INFORMATION:
			  /* =============================                                */
		
				response[2] = 0;
				
				errors = debug_get_num_of_errors();
				
				response[3] = (uint8_t) (errors >> 0);
				response[4] = (uint8_t) (errors >> 8);
				
				for(
					i = 0;
					   DEBUG_MAX_ERROR_CASES_PER_PAGE*cmd[2] + i < DEBUG_MAX_ERROR_CASES
					&& 5 + i * 4 + 3 < CMD_MAX_RESPONSE_LENGTH + 2;
					++i
				){
					
					eid = debug_get_local_id(
					      	DEBUG_MAX_ERROR_CASES_PER_PAGE*cmd[2] + i
					      );
					
					response[5 + i * 4 + 0] = (uint8_t) (eid >> 0);
					response[5 + i * 4 + 1] = (uint8_t) (eid >> 8);
					
					eid = debug_get_global_id(
					      	DEBUG_MAX_ERROR_CASES_PER_PAGE*cmd[2] + i
					      );
					
					response[5 + i * 4 + 2] = (uint8_t) (eid >> 0);
					response[5 + i * 4 + 3] = (uint8_t) (eid >> 8);
				}
				
			break;
			  /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
			case CMD_COLLECT_SAMPLES:
			  /* ====================                                         */
		
				response[2] = postproc_sample_request(cmd[1]);
				
			break;
			  /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
			case CMD_RETURN_SAMPLE_STATUS:
			  /* =========================                                    */
		
				response[2] = postproc_sample_status();
				
			break;
			  /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
			case CMD_CLEAR_SAMPLE_REQUEST:
			  /* =========================                                    */
		
				response[2] = 0;
				postproc_sample_clear_request();
				
			break;
			  /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
			case CMD_RETURN_SAMPLE_SHEET:
			  /* ========================                                     */
		
				response[2] = postproc_sample_sheet(
				              	cmd[1],
				              	&response[3],
				              	&response[4],
				              	&response[5]
				              );
				
			break;
			  /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
			case CMD_RETURN_SAMPLE_SHEETS:
			  /* =========================                                    */
				
				debug_report(DEBUG_CMD_ITERABLE_CALLED);
		
			break;
			default:
		
				debug_report(DEBUG_CMD_UNKNOWN_CMD);
				responseAvailable = false;
		}
	}
	
	if(ongoingProcess && waitState != WAIT_STATE_PREPARE_RESPOND){
		
		/* There is something left to do. */
		return true;
	}
	else{
	
		if(currentCMD.responseExpected && responseAvailable){

			if(currentCMD.targetFieldPresent && currentTarget > 0){
			
				/* This command was received over-the-air. */
				radio_ota_send(currentTarget, response, currentCMD.responseLength);
			}
			else{
			
				/* This command was received via USB. */
				
				if(waitState != WAIT_STATE_PREPARE_RESPOND){
					
					/* Prepare respond. */
					control_respond_prepare(&response[2], currentCMD.responseLength);
					waitState = WAIT_STATE_PREPARE_RESPOND;
					ongoingProcess = true;
				}
				else{
					
					/* Respond, finally. */
					control_respond_finish(&response[2], currentCMD.responseLength);
					waitState = WAIT_FOR_NOTHING;
					ongoingProcess = false;;
				}
			}
		}
		
		return (waitState == WAIT_STATE_PREPARE_RESPOND) ? true : false;
	}
}

void cmd_timeout_response(uint8_t cmd, uint8_t * response){
/*
   Generates the timeout response for the specified command {cmd} and writes
   it to the specified address {response}.
*/
	
	size_t i;
	
	if(cmd < CMD_ALWAYS_LAST_ITEM){
		
		for(i = 0; i < cmds[cmd].timeoutLength; ++i){
			
			response[i] = cmds[cmd].timeoutData[i];
		}
	}
	else{
		
		debug_report(DEBUG_CONTROL_UNKNOWN_CMD);
	}
}

void cmd_busy_response(uint8_t cmd, uint8_t * response){
/*
   Generates the busy ("control mode required") response for the specified
   command {cmd} and writes it to the specified address {response}.
*/
	
	size_t i;
	
	if(cmd < CMD_ALWAYS_LAST_ITEM){
		
		for(i = 0; i < cmds[cmd].busyLength; ++i){
			
			response[i] = cmds[cmd].busyData[i];
		}
	}
	else{
		
		debug_report(DEBUG_CONTROL_UNKNOWN_CMD);
	}
}

void cmd_init(void){
	
	memory_init();
}

