/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

#define RSS2I_MODULE "coder"

/*
  Holds some routines for encoding and decoding flash pages.
  
  Authors
  =======
    
    Copyright (C) 2021-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
    
    2021-04-01    Initial release.
    2021-04-21    Replace {fmtSpec} with {fmtByte} in {*_increments_*_index()}.
    2021-04-21    Added decoding functionality.
    2021-04-29    Support {CTE_2US_7PX1} {CTEType}.
    
  Description
  ===========
    
    Measurement results, i.e. RSS values as well as IQ samples, need to be
    stored on the external on-board flash memory. This modules provides some
    basic decoding and encoding helper functions. The actual encoding takes
    place within the memory module. The decoding is performed by the user.
    
  Memory Format
  =============
    
    A flash page consists of 256 bytes. The last 4 bytes are reserved for a CRC
    sum. CRC generation and check are performed automatically by the memory
    module during writing and reading, respectively, of flash pages. Thus, 252
    bytes are usable for storing measurement results.
    
    A measurement value is indexed using a single value, the so-called index.
    For a frame f within a cylce c during a measurement with n frames per cycle,
    this index i is calculated by
      
      i = c * n + f.
      
    An index value is stored within 4 consecutive bytes with the Most
    Significant Byte (MSB) first.
    
    The first 4 bytes of a flash page hold such an index value. We say, this
    index is the "current index". Within the following bytes, there may be
    stored several measurement results that correspond to the current index.
    
    The number of bytes a measurement result consists of as well as the type of
    measurement result is encoded into the value of the first byte of the
    measurement result, the so-called format byte, according to the following
    table:
      
       Format Byte | Description
      -------------+----------------------------------------------------------
         < 128     | This measurement result is an RSS value that belongs to 
                   | the current index. This format byte itself holds its
                   | integer part R_i and the following byte holds its
                   | fractional part R_f. Assuming R to be the negative RSS
                   | value in dBm/128, it holds that
                   |   
                   |   
                   |   R_f = R % 128
                   |   
                   | and
                   |
                   |   R_i = R // 128
                   |   
                   | where // denotes the C-type integer division. After
                   | encoding the RSS value into the mentioned 2 bytes, the
                   | current index is incremented by 1, i.e. measurement
                   | results that are stored in the following bytes belong
                   | to the next index without the need of re-specifying the
                   | current index.
      -------------+----------------------------------------------------------
          128      | For some reason, there is no measurement value
                   | available that belongs to the current index. The
                   | current index is incremented by 1, i.e. measurement
                   | results that are stored in the following bytes belong
                   | to the next index without the need of re-specifying the
                   | current index.
      -------------+----------------------------------------------------------
          129      | The current index needs to be re-specified for some
                   | reason. The following 4 bytes contain the new index.
                   | This index is the new "current index", i.e. measurement
                   | results that are stored after the mentioned 5 bytes
                   | (this format byte and the following 4 index bytes)
                   | belong to this new index.
      -------------+----------------------------------------------------------
          130      | End of page. The following bytes may contain garbage.
      -------------+----------------------------------------------------------
        128        | The following l bytes contain 32 bit samples. The 
      + 10*(x + 1) | format of these samples, their meaning as well as the
      + y          | number l of bytes solely depend on the so-called format
                   | x. 
                   | 
                   | One measurement result, i.e. the following l bytes,
                   | only contains a specific group of 8 bits of these 32
                   | bit samples according to the following table:
                   |   
                   |    y | Byte out of the 32 bit sample
                   |   ---+-------------------------------
                   |    0 | LSB
                   |   ---+-------------------------------
                   |     ...
                   |   ---+-------------------------------
                   |    3 | MSB
                   |   ---+-------------------------------
                   |   
                   | In other words: In order to store l 32 bit samples, it
                   | is necessary to store 4 measurement results with the
                   | same format specifier x and 4 different y values,
                   | so-called sheet indices, one after the other.
                   |
                   | NOTE, that the current index is not incremented and
                   | indeed stays the same after encoding the l + 1 bytes
                   | mentioned above (l sample bytes, 1 format byte), i.e.
                   | the next measurement result belongs to the same index.
                   |
                   | The available format specifiers can be referenced from
                   | the following table:
                   |   
                   |    x | Format
                   |   ---+------------------------------------------
                   |    0 | 224 IQ samples recoreded during a CTE.
                   |      | The first 16 samples are recorded during
                   |      | the reference period. The next 13*16
                   |      | IQ samples are recorded in the following
                   |      | manner: 16 IQ samples during sampling
                   |      | slot 1, 16 IQ samples during switching
                   |      | slot 1, 16 IQ samples during sampling
                   |      | slot 2, 16 IQ samples during switching
                   |      | slot 2, ..., 16 IQ samples during
                   |      | sampling slot 7.
                   |   ---+------------------------------------------
                   |    1 | 240 IQ samples recoreded during a CTE.
                   |      | The first 32 samples are recorded during
                   |      | the reference period. The next 13*16
                   |      | IQ samples are recorded in the following
                   |      | manner: 16 IQ samples during sampling
                   |      | slot 1, 16 IQ samples during switching
                   |      | slot 1, 16 IQ samples during sampling
                   |      | slot 2, 16 IQ samples during switching
                   |      | slot 2, ..., 16 IQ samples during
                   |      | sampling slot 7.
                   |   ---+------------------------------------------
                   |   
      -------------+----------------------------------------------------------
*/


#include <stdint.h>
#include <stdbool.h>

#include <rss2i/coder.h>
#include <rss2i/debug.h>

#include <rss2i/config/cte.h>


/* Check consistency between {CoderFormatByte} and {CoderFormatSpecifier}. */
_Static_assert(
	   CODER_FORMAT_BYTE_SAMPLE_FORMAT_0_SHEET_0
	   	== CODER_FORMAT_SPECIFIER_SAMPLE_FORMAT_0 + 0
	&& CODER_FORMAT_BYTE_SAMPLE_FORMAT_0_SHEET_1
	   	== CODER_FORMAT_SPECIFIER_SAMPLE_FORMAT_0 + 1
	&& CODER_FORMAT_BYTE_SAMPLE_FORMAT_0_SHEET_2
	   	== CODER_FORMAT_SPECIFIER_SAMPLE_FORMAT_0 + 2
	&& CODER_FORMAT_BYTE_SAMPLE_FORMAT_0_SHEET_3
	   	== CODER_FORMAT_SPECIFIER_SAMPLE_FORMAT_0 + 3
	
	&& CODER_FORMAT_BYTE_SAMPLE_FORMAT_1_SHEET_0
	   	== CODER_FORMAT_SPECIFIER_SAMPLE_FORMAT_1 + 0
	&& CODER_FORMAT_BYTE_SAMPLE_FORMAT_1_SHEET_1
	   	== CODER_FORMAT_SPECIFIER_SAMPLE_FORMAT_1 + 1
	&& CODER_FORMAT_BYTE_SAMPLE_FORMAT_1_SHEET_2
	   	== CODER_FORMAT_SPECIFIER_SAMPLE_FORMAT_1 + 2
	&& CODER_FORMAT_BYTE_SAMPLE_FORMAT_1_SHEET_3
	   	== CODER_FORMAT_SPECIFIER_SAMPLE_FORMAT_1 + 3,
	"An inconsistency between CoderFormatByte and CoderFormatSpecifier arised."
);

/* Check position of the command {CoderFormatByte}s. */
_Static_assert(
	   CODER_FORMAT_BYTE_COMMAND_NEXT_INDEX
	 > CODER_FORMAT_BYTE_VALID_RSS_127
	&& CODER_FORMAT_BYTE_COMMAND_NEXT_INDEX
	 < CODER_FORMAT_BYTE_SAMPLE_FORMAT_0_SHEET_0
	&& CODER_FORMAT_BYTE_COMMAND_NEW_INDEX
	 > CODER_FORMAT_BYTE_VALID_RSS_127
	&& CODER_FORMAT_BYTE_COMMAND_NEW_INDEX
	 < CODER_FORMAT_BYTE_SAMPLE_FORMAT_0_SHEET_0
	&& CODER_FORMAT_BYTE_COMMAND_END_OF_PAGE
	 > CODER_FORMAT_BYTE_VALID_RSS_127
	&& CODER_FORMAT_BYTE_COMMAND_END_OF_PAGE
	 < CODER_FORMAT_BYTE_SAMPLE_FORMAT_0_SHEET_0
	&& CODER_FORMAT_BYTE_COMMAND_RESERVED_FIRST
	 > CODER_FORMAT_BYTE_VALID_RSS_127
	&& CODER_FORMAT_BYTE_COMMAND_RESERVED_FIRST
	 < CODER_FORMAT_BYTE_SAMPLE_FORMAT_0_SHEET_0
	&& CODER_FORMAT_BYTE_COMMAND_RESERVED_LAST
	 > CODER_FORMAT_BYTE_VALID_RSS_127
	&& CODER_FORMAT_BYTE_COMMAND_RESERVED_LAST
	 < CODER_FORMAT_BYTE_SAMPLE_FORMAT_0_SHEET_0,
	"The position of the CoderFormatByte commands may be unexpected."
);

/* Check sample format specifier as it is promised in the API doc. */
_Static_assert(
	   (CODER_FORMAT_SPECIFIER_SAMPLE_FORMAT_0 - 128) % 10 == 0
	&& (CODER_FORMAT_SPECIFIER_SAMPLE_FORMAT_0 - 128) / 10 - 1 == 0
	
	&& (CODER_FORMAT_SPECIFIER_SAMPLE_FORMAT_1 - 128) % 10 == 0
	&& (CODER_FORMAT_SPECIFIER_SAMPLE_FORMAT_1 - 128) / 10 - 1 == 1,
	"The flash decoding algorithms may not work as expected."
);

/* Check for consistency between {PostprocFormat} and {CoderFormatSpecifier}. */
_Static_assert(
	   POSTPROC_FORMAT_INVALID < 0
	&& POSTPROC_FORMAT_0 == 0
	&& POSTPROC_FORMAT_1 == 1,
	"PostprocFormat and CoderFormatSpecifier may be inconsistent."
);


inline uint8_t coder_truncate_valid_rss(uint8_t rss){
/*
  Truncates the specified {rss} integer value (if needed) such that it becomes
  a valid RSS format byte. If a truncation was needed, this will be reported
  to the debug system.
*/
	
	if(rss > CODER_FORMAT_SPECIFIER_VALID_RSS){
		
		debug_report(DEBUG_CODER_TRUNCATED_VALID_RSS);
		return CODER_FORMAT_SPECIFIER_VALID_RSS;
	}
	else{
		
		return rss;
	}
}

bool coder_explicit_format_byte_present(CoderFormatSpecifier fmtSpec){
/*
  Whether the specified {fmtSpec} has to be explicitly present before the actual
  measurement result ({true}) or it is part (the first byte) of the actual
  measurement result ({false}).
*/
	
	return (fmtSpec > CODER_FORMAT_SPECIFIER_VALID_RSS)
	     ? true
	     : false;
}

inline CoderFormatByte coder_format_byte(
	CoderFormatSpecifier fmtSpec,
	uint8_t sheetIndex
){
/*
  Given a {fmtSpec} and a {sheetIndex} this function returns the corresponding
  format byte.
*/
	
	if(fmtSpec < CODER_FORMAT_SPECIFIER_NONE
	&& fmtSpec >= CODER_FORMAT_SPECIFIER_SAMPLE_FORMAT_0){
		
		return fmtSpec + sheetIndex;
	}
	else{
		
		debug_report(DEBUG_CODER_INVALID_FORMAT_SPECIFIER);
		return fmtSpec;
	}
}

bool coder_is_result(CoderFormatByte fmtByte){
/*
  Whether the given {fmtByte} announces a measurement result or a command. A
  valid result {fmtByte} will always evaluate to {true}. A valid command
  {fmtByte} will always evaluate to {false}. However, the validity of the
  specified {fmtByte} has to be further checked in both cases.
*/
	
	if(fmtByte > CODER_FORMAT_BYTE_VALID_RSS_127
	&& fmtByte < CODER_FORMAT_BYTE_SAMPLE_FORMAT_0_SHEET_0){
		
		return false;
	}
	else{
		
		return true;
	}
}

uint8_t coder_result_size(CoderFormatSpecifier fmtSpec){
/*
  Number of bytes of a measurement result with the specified {fmtSpec}.
*/
	
	switch(fmtSpec){
		case CODER_FORMAT_SPECIFIER_VALID_RSS:
		
			return 2;
			
		break;
		case CODER_FORMAT_SPECIFIER_SAMPLE_FORMAT_0:
		
			return CTE_2US_7P1_NUMBER_OF_SAMPLES;
			
		break;
		case CODER_FORMAT_SPECIFIER_SAMPLE_FORMAT_1:
		
			return CTE_2US_7PX1_NUMBER_OF_SAMPLES;
			
		break;
		default:
		
			debug_report(DEBUG_CODER_INVALID_FORMAT_SPECIFIER);
			return 0;
	}
}

uint8_t coder_command_size(CoderFormatByte fmtByte, uint8_t maxSize){
/*
  Number of bytes of a command with the specified {fmtByte}. The number of
  bytes left on the current flash page (counted from the byte after the given
  {fmtByte}) is denoted by {maxSize}.
*/
	
	switch(fmtByte){
		case CODER_FORMAT_BYTE_COMMAND_NEXT_INDEX:
		
			return 0;
			
		break;
		case CODER_FORMAT_BYTE_COMMAND_NEW_INDEX:
		
			return 4;
			
		break;
		case CODER_FORMAT_BYTE_COMMAND_END_OF_PAGE:
		
			return maxSize;
			
		break;
		default:
		
			debug_report(DEBUG_CODER_INVALID_FORMAT_BYTE);
			return 0;
	}
}

bool coder_result_increments_expected_index(CoderFormatByte fmtByte){
/*
  Whether a result with the specified {fmtByte} increments the expected index,
  i.e. the expected index of the proceeding results, or not.
*/
	
	/* Currently, only a valid RSS result increments the expected index. */
	return (fmtByte > CODER_FORMAT_BYTE_VALID_RSS_127)
	     ? false
	     : true;
}

CoderFormatSpecifier coder_format_specifier(PostprocFormat fmt){
/*
  Given the {PostprocFormat} {fmt} this function returns the corresponding
  {CoderFormatSpecifier}. An invalid format index results in the return of
  {CODER_FORMAT_SPECIFIER_NONE}.
*/
	
	if(fmt >= 0 && fmt < 12){
		
		return CODER_FORMAT_SPECIFIER_SAMPLE_FORMAT_0 + fmt*10;
	}
	else{
		
		return CODER_FORMAT_SPECIFIER_NONE;
	}
}

PostprocFormat coder_postproc_format(CoderFormatSpecifier fmtSpec){
/*
  Given the {CoderFormatSpecifier} {fmtSpec} this function returns the
  corresponding {PostprocFormat}. An invalid format results in the return of
  {POSTPROC_FORMAT_INVALID}.
*/
	
	if(fmtSpec >= CODER_FORMAT_SPECIFIER_SAMPLE_FORMAT_0
	&& fmtSpec < CODER_FORMAT_SPECIFIER_NONE){
		
		return (fmtSpec - CODER_FORMAT_SPECIFIER_SAMPLE_FORMAT_0) / 10;
	}
	else{
		
		return POSTPROC_FORMAT_INVALID;
	}
}

CoderFormatSpecifier coder_format_specifier_from_byte(CoderFormatByte fmtByte){
/*
  Given a {fmtByte} this function returns the corresponding format specifier.
  Returns {CODER_FORMAT_SPECIFIER_NONE} if the specified {fmtByte} is invalid.
*/
	
	if(fmtByte <= CODER_FORMAT_BYTE_VALID_RSS_127){
		
		return CODER_FORMAT_SPECIFIER_VALID_RSS;
	}
	else if(fmtByte >= CODER_FORMAT_BYTE_SAMPLE_FORMAT_0_SHEET_0){
		
		/* Yes, we do not check for not-yet-implemented formats. */
		
		if((fmtByte - CODER_FORMAT_BYTE_SAMPLE_FORMAT_0_SHEET_0) % 10 < 4){
			
			return CODER_FORMAT_SPECIFIER_SAMPLE_FORMAT_0
			     + (fmtByte - CODER_FORMAT_BYTE_SAMPLE_FORMAT_0_SHEET_0)
			     / 10
			     * 10;
		}
		else{
			
			/* Invalid sheet index. */
			
			debug_report(DEBUG_CODER_INVALID_FORMAT_BYTE);
			return CODER_FORMAT_SPECIFIER_NONE;
		}
	}
	else{
		
		debug_report(DEBUG_CODER_INVALID_FORMAT_BYTE);
		return CODER_FORMAT_SPECIFIER_NONE;
	}
}

uint8_t coder_sheet_index_from_byte(CoderFormatByte fmtByte){
/*
  Given the {fmtByte} this function returns the corresponding sheet index.
*/
	
	uint8_t sheetIndex;
	
	if(fmtByte <= CODER_FORMAT_BYTE_VALID_RSS_127){
		
		return 0;
	}
	else if(fmtByte >= CODER_FORMAT_BYTE_SAMPLE_FORMAT_0_SHEET_0){
		
		/* Yes, we do not check for not-yet-implemented formats. */
		sheetIndex = (fmtByte - CODER_FORMAT_BYTE_SAMPLE_FORMAT_0_SHEET_0) % 10;
		
		if(sheetIndex < 4){
			
			return sheetIndex;
		}
		else{
			
			/* Invalid sheet index. */
			
			debug_report(DEBUG_CODER_INVALID_FORMAT_BYTE);
			return 0;
		}
	}
	else{
		
		debug_report(DEBUG_CODER_INVALID_FORMAT_BYTE);
		return 0;
	}
}

uint32_t coder_index(uint8_t * indx){
/*
  Calculates the index from the four bytes pointed to by {indx}.
*/
	
	/* The index is stored MSB first. */
	return (((uint32_t) indx[0]) << 24)
	     + (((uint32_t) indx[1]) << 16)
	     + (((uint32_t) indx[2]) <<  8)
	     + (((uint32_t) indx[3]) <<  0);
}

uint32_t coder_next_index(
	uint32_t currentIndex,
	CoderFormatByte fmtByte,
	uint8_t * cmd
){
/*
  Returns the expected index of the measurement results that are stored after
  the command with the specified {fmtByte}. The (probably empty) content of the
  command should be available at the memory pointed to by {cmd}. The
  {currentIndex} denotes the index of the results prior to the command with
  the given {fmtByte}.
*/
	
	switch(fmtByte){
		case CODER_FORMAT_BYTE_COMMAND_NEXT_INDEX:
		
			return currentIndex + 1;
			
		break;
		case CODER_FORMAT_BYTE_COMMAND_NEW_INDEX:
		
			return coder_index(cmd);
			
		break;
		case CODER_FORMAT_BYTE_COMMAND_END_OF_PAGE:
			
			/* There is no subsequent result on this page. */
			return 0;
			
		break;
		default:
			
			/* TODO : What about results? They may also change the index? */
			debug_report(DEBUG_CODER_INVALID_FORMAT_BYTE);
			return 0;
	}
}

