/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

#define RSS2I_MODULE "control"

/*
  Handle and process incoming commands.
  
  Authors
  =======
    
    Copyright (C) 2018-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
    
    2019-04-16    Initial public release.
    2019-05-29    Added support for iterable commands.
    2019-07-12    Made the USB command processing more reliable.
    2019-10-31    Made the debug messages static.
    2020-03-31    Fitted to the changes within the command module header.
    2020-07-06    Split {control_respond()} into {*_prepare()} and {*_finish()}.
    2020-08-25    Restore target field only if there actually is one.
    2021-03-19    Support target {0xff}.
  
  Description
  ===========
    
    After initialization using {control_init()} the control module should get
    triggered "frequently" using {control_trigger()}. Latter's documentation
    contains some information on how the command handling and processing works.
*/

#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>

#include <rss2i/usb.h>
#include <rss2i/debug.h>
#include <rss2i/radio.h>
#include <rss2i/crc.h>
#include <rss2i/measure.h>
#include <rss2i/uarte.h>
#include <rss2i/memory.h>
#include <rss2i/timer.h>
#include <rss2i/control.h>
#include <rss2i/cmd.h>
#include <rss2i/board.h>
#include <rss2i/print.h>

#include <rss2i/config/usb.h>
#include <rss2i/config/control.h>


extern const CMD cmds[];

static uint8_t controlOTATries = CONTROL_OTA_TRIES;

bool control_set_ota_tries_needs_control_mode(uint8_t tries){
/*
  Whether the corresponding call to {control_set_ota_tries()} needs control mode
  or not.
*/
	return tries;
}

uint8_t control_set_ota_tries(uint8_t tries){
/*
  Sets the value of {controlOTATries} to {tries}. Specifiying {0} won't change
  the current setting. Returns the designated {controlOTATries}.
*/
	
	if(tries){
		
		controlOTATries = tries;
	}
	
	return controlOTATries;
}

void control_respond_prepare(uint8_t * response, size_t size){
/*
 Prepares the {size} bytes of the specified response message for being send via
 USB (i.e. adding a CRC sum). Please make sure {response} to point to at least
 {size + 4} bytes.
*/
	
	uint32_t crcSum;
	
	crcSum = crc(response, size);
	
	response[size + 0] = crcSum >> 24;
	response[size + 1] = crcSum >> 16;
	response[size + 2] = crcSum >>  8;
	response[size + 3] = crcSum >>  0;
}


void control_respond_finish(uint8_t * response, size_t size){
/*
 Responds the given {response} via USB. The CRC should be already added e.g. by
 calling {control_respond_prepare()} prior to this function. However, {size}
 should refer to the size of the actual response, i.e. should not include the
 four CRC bytes.
*/
	
	usb_send(response, size + 4);
}

void control_respond(uint8_t * response, size_t size){
/*
 Responds the given {response} via USB. The CRC sum is added automatically.
 Therefore, please make sure {response} to point to at least {size + 4} bytes.
*/
	
	control_respond_prepare(response, size);
	control_respond_finish(response, size);
}

static bool control_handle_usb(uint8_t * cmd){
/*
  Handles a command (that was received via USB).
  
  Description
  ===========
  
    The command may get processed directly or may be forwarded to another device
    over-the-air (timeout management inclusive).
    In order to avoid long blocking delays, it may be necessary to call this
    function multiple times in order to handle a command completely.
  
  Argument List
  =============
  
    {cmd}    Address of the command to be handled. After specified once it may
             be {NULL} during future calls to this function until the initially
             given command got handled completely (see below). Nevertheless the
             memory pointed to by {cmd} should meanwhile retain its content.
  
  Return Value
  ============
  
    Returns {true} when the handling of the (previously) specified command is
    still in process and this function should hence get called at least one more
    time. Returns {false} when the (previously) specified command got handled
    completely.
*/
	
	typedef enum {
		
		NONE,
		WAIT_FOR_OTA,
		WAIT_FOR_CMD,
		
	} DelayReason;
	
	static DelayReason delayReason = NONE;
	static volatile bool timeout;
	static TimerID timer;
	static uint8_t * currentCMDPtr;
	static CMD currentCMD;
	static uint8_t currentCMDIndex;
	static uint8_t target;
	static uint8_t tries;
	static uint8_t iterations = 0;
	static uint8_t tmp;
	
	/* Do not put this onto the stack. */
	static char debugMsg[] = "OTA request timed out 0x00 times.";
	
	uint8_t * response;
	uint8_t alternativeResponse[CMD_MAX_RESPONSE_LENGTH + 4];
	
	
	if(delayReason == WAIT_FOR_OTA){
		
		if((response = radio_ota_recv())){
			
			/* Received a response over-the-air. */
			
			control_respond(&response[0], currentCMD.responseLength);
			
			timer_clear(timer);
			
			if(tries){
			
				print_uhex(tries, debugMsg, 24, 2);
				debug("usb handle", debugMsg);
			}
			
			delayReason = NONE;
		}
		else if(timeout){
			
			if(++tries >= controlOTATries){
				
				/* Reached maximum number of retries. */
				
				cmd_timeout_response(currentCMDIndex, alternativeResponse);
				control_respond(alternativeResponse, currentCMD.responseLength);
			
				delayReason = NONE;
				
				debug("usb handle", "OTA request finally timed out.");
			}
			else{
				
				/* Retry. */
				
				radio_ota_send(target, currentCMDPtr, currentCMD.length);
				timeout = false;
				timer = timer_flag(&timeout, CONTROL_OTA_TIMEOUT, TIMER_NO_REPEAT);
			}
		}
	}
	else if(delayReason == WAIT_FOR_CMD){
		
		if(cmd_handle(NULL) == false){
			
			delayReason = NONE;
		}
	}
	else if(!cmd && !iterations){
		
		debug_report(DEBUG_CONTROL_INVALID_CMD);
	}
	else{
		
		/* Make sure to have enough entries available on the USB output */
		/* buffer for this iteration (and the last one).                */
		if(!iterations || usb_vacant_entries() > 3){
		
			if(iterations){
			
				/* This command should be iterated one more time. */
				
				if(currentCMD.targetFieldPresent){
				
					/* Restore target field */
					/* (See below).         */
					currentCMDPtr[3] = tmp;
				}
				
				cmd_iterate(&currentCMDPtr[2]);
			}
			else{
			
				/* Ready to process a new command. */
			
				iterations = cmd_is_iterable(&cmd[2]);
			
				currentCMDPtr = cmd;
				currentCMDIndex = currentCMDPtr[2];
				currentCMD = cmds[currentCMDIndex];
			
				/* Since the target field is sometimes overwritten, a backup */
				/* is made for restoring within possible iterations.         */
				tmp = currentCMDPtr[3];
			}
			
			if(iterations){ --iterations; }
		
			if(currentCMD.targetFieldPresent){
			
				/* This command MAY be addressed to another device. */
			
				if(currentCMDPtr[3] == board_id() || currentCMDPtr[3] == 0xff){
				
					/* This one is for us. */
				
					currentCMDPtr[3] = 0;
					if(cmd_handle(&currentCMDPtr[2])){
					
						delayReason = WAIT_FOR_CMD;
					}
				}
				else{
			
					/* This one is for another device. */
				
					if(measure_running()){
				
						/* Forwarding commands over-the-air needs control mode active */
						cmd_busy_response(currentCMDIndex, alternativeResponse);
						control_respond(alternativeResponse, currentCMD.responseLength);
					}
					else{
			
						target = currentCMDPtr[3];
						currentCMDPtr[3] = board_id();
			
						if(currentCMD.responseExpected){
				
							radio_ota_recv_flush();
						}
			
						radio_ota_send(target, currentCMDPtr, currentCMD.length);
			
						if(currentCMD.responseExpected){
				
							timeout = false;
							timer = timer_flag(&timeout, CONTROL_OTA_TIMEOUT, TIMER_NO_REPEAT);
							tries = 0;
							delayReason = WAIT_FOR_OTA;
						}
					}
				}
			}
			else{
		
				/* This one is for us. */
			
				if(cmd_handle(&currentCMDPtr[2])){
				
					delayReason = WAIT_FOR_CMD;
				}
			}
		}
	}
	
	if(delayReason){
		
		/* We are still busy. */
		return true;
	}
	else{
		
		/* We finished the handling of a(n iteration of a) command… */
		if(iterations){
			
			/* …but we have to iterate it at least one more time. */
			return true;
		}
		else{
			
			/* …and this command was either not iterable or */
			/* was iterated completely.                     */
			return false;
		}
	}
}

static bool control_process_usb(uint8_t * instr){
/*
  Given a data row from the USB input buffer ({instr}) this function assembles
  the received input, searches for the beginning of a command, applies a cyclic
  redundancy check and, finally, forwards the detected command to
  {control_handle_usb()}.
  Returns {true} if the detected command's handling (i.e. the forwarding
  mentioned above) needs some more calls in order to finish. Returns {false}
  otherwise, i.e. this function is ready to process new data rows.
*/

	size_t i;
	static bool cmdDetected = false;
	static volatile bool timeout = false;
	static TimerID timer;
	static bool ignoreStartByte = false;
	/* +2: See to {radio_ota_send()}.
	 * +4: Space for the four CRC bytes.
	 */
	static uint8_t cmd[CMD_MAX_LENGTH + 2 + 4]; 
	static uint16_t cmdPtr = 0;
	
	bool ongoingProcess = false;
	CMD currentCMD;
	
	/* Reset command detection after some timeout. */
	
	timer_clear(timer);
	
	if(timeout){
	
		ignoreStartByte = false;
		cmdDetected = false;
		timeout = false;
	}
	
	timer = timer_flag(&timeout, CONTROL_USB_RECV_TIMEOUT, TIMER_NO_REPEAT);
	
	/* Process USB buffer entry. */
	
	for(i = 0; i < USB_RECV_BUFFER_WIDTH; ++i){
		
		if(instr[i] == 0xff && !ignoreStartByte){
		
			/* Maybe this is the start of a new command. */
		
			cmdDetected = true;
			cmdPtr = 2;
		}
		else{
		
			if(cmdDetected){
			
				cmd[cmdPtr++] = instr[i];
	
				if(cmd[2] < CMD_ALWAYS_LAST_ITEM){
					
					/* Seems to be a valid command. */
					
					currentCMD = cmds[cmd[2]];
					ignoreStartByte = true;
					
					if(currentCMD.length + 4 + 2 == cmdPtr){
						
						/* This should be the detected command's end. */
						
						if(crc_check(crc(&cmd[2], currentCMD.length + 4))){
						
							ongoingProcess = control_handle_usb(cmd);
						}
						else{
					
							debug_report(DEBUG_CONTROL_USB_CRC_FAILED);
						}
						
						/* Reset state machine. */
						cmdDetected = false;
						ignoreStartByte = false;
					}
				}
				else{
					
					cmdDetected = false;
					ignoreStartByte = false;
					debug_report(DEBUG_CONTROL_UNKNOWN_CMD);
				}
			}
		}
	}
	
	return ongoingProcess;
}

void control_trigger(void){
/*
  This function should be called "frequently".
  
  Some Details
  ============
  
    A command that comes over-the-air gets handled directly, i.e. gets forwarded
    to {cmd_handle()}. Commands that get received via the UART-over-USB
    interface need to get processed before (i.e. assemble the actual command
    that may be spread over the USB input buffer, detecting the command's start,
    applying redundancy check, ...).
    In order to avoid long blocking delays (command processing may need some
    time, command OTA forwarding may time out several times, ...) most handling
    and processing tasks are split into small steps that may get done during
    subsequent calls of the respective procedures. Whether such a procedure
    needs some more calls in order to finish a previously specified task or not
    can be seen from its return value.
*/
	
	uint8_t * instr;
	
	static bool ongoingUSB = false;
	static bool ongoingOTA = false;
	
	if(ongoingUSB){
		
		/* Ongoing process. Call USB handle directly. */
		ongoingUSB = control_handle_usb(NULL);
	}
	else if(ongoingOTA){
		
		/* Ongoing process. Call CMD handle directly. */
		ongoingOTA = cmd_handle(NULL);
	}
	else if((instr = usb_recv())){
		
		/* Check USB buffer for new commands. */
		ongoingUSB = control_process_usb(instr);
	}
	else if((instr = radio_ota_recv())){
		
		/* Received a command over-the-air. */
		ongoingOTA = cmd_handle(instr);
	}
}

void control_init(void){
/*
  Initializes this module. Should be called only once.
*/
	usb_init();
	cmd_init();
	radio_init();
}


