/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

#define RSS2I_MODULE "crc"

/*
  Calculation of 32 bit CRC sums using a fixed generator polynomial.
  
  Authors
  =======
    
    Copyright (C) 2018-2021  Gerrit Maus (funk@maus.xyz)
    Copyright (C) 2018-2018  Mustafa Gemci (mustafagemci.ee@gmail.com)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
    
    2018-09-25    Initial release.
    2019-04-15    Initial public release.
    2021-04-02    Added {crc_append_zeros()}.
    
  Description
  ===========
    
    Byte-wise calculation
    ---------------------
    
     It is possible to do byte-wise CRC calculation. Therefore, the user has to
     specify a memory address storing the (temporary) CRC sum. That value gets
     updated everytime a new byte comes in and needs to be finalized after the
     last byte.
   
    External Resources
    ------------------
    
     The {extern} variable {crcTable} holds the 32 bit CRC sums for all (i.e.
     256) possible bytes. The generation of this lookup table depends on the
     chosen generator polynomial.
*/


#include <stdint.h>
#include <stddef.h>
#include <stdbool.h>

#include <rss2i/crc.h>

#include <rss2i/config/crc.h>


extern const uint32_t crcTable[];
extern const uint32_t crcMagicNumber;


void crc_reset(uint32_t * crc){
/*
  Initializes byte-wise CRC calculation. The (temporary) result of the crc
  calculation is stored at the specified memory address.
*/
	
	*crc = CRC_START_VALUE;
}

void crc_append(uint32_t * crc, uint8_t byte){
/*
  Adds a {byte} to the CRC sum at the specified memory address.
*/

	uint8_t index = (*crc ^ ((uint32_t) byte << 24)) >> 24;
	*crc = (*crc << 8) ^ crcTable[index];
}

void crc_append_zeros(uint32_t * crc, size_t amount){
/*
  Adds {amount} zero bytes to the CRC sum at the specified memory address.
  This is faster than repeatedly calling {crc_append(crc, 0)}.
*/
	
	size_t i;
	uint8_t index;
	
	for(i = 0; i < amount; ++i){
	
		index = (*crc) >> 24;
		*crc = (*crc << 8) ^ crcTable[index];
	}
}

void crc_finish(uint32_t * crc){
/*
  Finalizes byte-wise CRC calculation.
*/
	
	*crc ^= CRC_FINAL_XOR;
}

bool crc_check(uint32_t crc){
/*
  The CRC checksum calculated for some data including an appended CRC checksum
  looks always the same, independent of the actual data prior to the original
  checksum. This specific value is called the magick number and depends solely
  on the CRC parameters.
  
  Returns {true} if the given CRC sum equals that magick sumber. Returns {false}
  if not.
*/
	
	return (crc == crcMagicNumber);
}

uint32_t crc(uint8_t * byte, size_t size){
/*
  Calculates the CRC sum of the {size} bytes within the array {byte} (Least
  significant byte first).
*/
	uint32_t crc;
	size_t i;
	
	crc_reset(&crc);
	
	for(i = 0; i < size; ++i){
		
		crc_append(&crc, byte[i]);
	}
	
	crc_finish(&crc);
	
	return crc;
}

