/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

#define RSS2I_MODULE "cte"

/*
  BLE 5.1 Constant Tone Extension (CTE) related functionality.
  
  Authors
  =======
  
    Copyright (C) 2020-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
  
    2020-07-31    Initial release.
    2020-07-31    Added {*_is_valid()} methods.
    2020-08-03    Replaced {*_is_valid()} with {cte_antenna_pattern_is_valid()}.
    2020-08-03    Added {cte_type_is_valid()} method.
    2020-08-03    Added {cte_length()} function.
    2020-08-03    Added {CTE_ANTENNA_PATTERN_DISABLED}.
    2020-08-03    Changed the behaviour of {cte_type_is_valid()}.
    2020-08-13    Add {cte_radio_init()} and related functions to configure CTE.
    2020-08-14    Added {cte_is_sampling_configuration()}.
    2020-08-21    Changed oversampling configuration.
    2020-08-24    Outsourced some CTE compatibility infos to the cteinfo module.
    2021-02-17    ALWAYS use the standard pattern for the standard antenna.
    2021-02-26    Added {cte_shutdown()}.
    2021-02-26    Fix unused parameter warnings for REV prior to F.
    2021-03-04    Fixed {TSWITCHSPACING}.
    2021-03-05    Support CTE preprocessing using {cte_prepare()}.
    2021-03-10    Configure {TSAMPLEOFFSET} and {TSWITCHOFFSET}.
    2021-03-10    Always use the specified constant antenna during packet.
    2021-03-10    Check for {NULL} pointer.
    2021-03-18    Made {dataPtr} volatile.
    2021-03-24    Retrieve static data during init. This saves some config time.
    2021-04-14    Fix wrong antenna assignment.
    2021-04-01    Introduced static assert for {CTE_MAX_NUMBER_OF_SAMPLES}.
    2021-04-29    Support {CTE_2US_7PX1} {CTEType}.
*/

#include <stdint.h>
#include <stdbool.h>
#include <stddef.h>

#include <libopencm3/nrf5/gpio.h>
#include <libopencm3/nrf5/radio.h>

#include <rss2i/cte.h>
#include <rss2i/rev.h>
#include <rss2i/board.h>
#include <rss2i/debug.h>
#include <rss2i/switch.h>
#include <rss2i/cteinfo.h>
#include <rss2i/hook.h>

#include <rss2i/config/cte.h>


/* Available antennas. */
typedef enum {
	
	CTE_ANTENNA_STANDARD,
	CTE_ANTENNA_1,
	CTE_ANTENNA_2,
	CTE_ANTENNA_3,
	CTE_ANTENNA_4,
	CTE_ANTENNA_5,
	CTE_ANTENNA_6,
	CTE_ANTENNA_7,
	CTE_ANTENNA_8,
	
	CTE_ANTENNA_ALWAYS_LAST_ITEM
	
} CTEAntenna;

/* Whether we are in TX or RX mode. */
typedef enum {
	
	CTE_MODE_TX,
	CTE_MODE_RX,
	
	CTE_MODE_ALWAYS_LAST_ITEM
	
} CTEMode;

/* Values for preprocessed settings. */
typedef enum {
	
	CTE_PREPROC_UNPREPARED,
	
	CTE_PREPROC_DFEOPMODE_DISABLED,
	CTE_PREPROC_DFEOPMODE_AOA,
	CTE_PREPROC_DFEOPMODE_AOD,
	
	CTE_PREPROC_ALWAYS_LAST_ITEM
	
} CTEPreproc;


_Static_assert(
	   CTE_2US_7P1_NUMBER_OF_SAMPLES <= CTE_MAX_NUMBER_OF_SAMPLES
	&& CTE_2US_7PX1_NUMBER_OF_SAMPLES <= CTE_MAX_NUMBER_OF_SAMPLES,
	"False maximal number of CTE samples."
);


/* Store preprocessed information for {CTE_MODE_TX} and {CTE_MODE_RX} mode. */
volatile CTEPreproc ctePreproc[CTE_MODE_ALWAYS_LAST_ITEM]
                              [CTEINFO_MAX_HASH_VALUE + 1];


/* Number of pins that control the external HF switch. */
static uint8_t cteNumOfPins;
/* Static GPIO patterns for controling the HF switch. */
static uint32_t cteSwitchGPIOPattern[CTE_ANTENNA_ALWAYS_LAST_ITEM];


void cte_init(void){
/*
  Initializes the CTE module:
    
    - Acquires the GPIO pins that should be controlled by the CTE module.
    - Apply some static CTE configuration.
*/
	
	/* Acquire GPIOs. */
	switch_init();
	
	/* Static CTE configurations. */
	
	RADIO_DFEPACKET.MAXCNT = CTE_MAX_NUMBER_OF_SAMPLES;
	RADIO_DFECTRL2 |= RADIO_DFECTRL2_TSWITCHOFFSET_VAL(16); /* TODO */
	RADIO_DFECTRL2 |= RADIO_DFECTRL2_TSAMPLEOFFSET_VAL(-16); /* TODO */
	
	/* Retrieve static information. */
	
	/* Get the number of pins that control the external HF switch. */
	cteNumOfPins = switch_get_number_of_used_pins();
	
	/* Get the GPIO patterns for controlling the switch. */
	cteSwitchGPIOPattern[CTE_ANTENNA_STANDARD] = switch_get_standard_pattern();
	cteSwitchGPIOPattern[CTE_ANTENNA_1] = switch_get_pattern(SWITCH_ANTENNA_1);
	cteSwitchGPIOPattern[CTE_ANTENNA_2] = switch_get_pattern(SWITCH_ANTENNA_2);
	cteSwitchGPIOPattern[CTE_ANTENNA_3] = switch_get_pattern(SWITCH_ANTENNA_3);
	cteSwitchGPIOPattern[CTE_ANTENNA_4] = switch_get_pattern(SWITCH_ANTENNA_4);
	cteSwitchGPIOPattern[CTE_ANTENNA_5] = switch_get_pattern(SWITCH_ANTENNA_5);
	cteSwitchGPIOPattern[CTE_ANTENNA_6] = switch_get_pattern(SWITCH_ANTENNA_6);
	cteSwitchGPIOPattern[CTE_ANTENNA_7] = switch_get_pattern(SWITCH_ANTENNA_7);
	cteSwitchGPIOPattern[CTE_ANTENNA_8] = switch_get_pattern(SWITCH_ANTENNA_8);
	
	
	cte_unprepare();
}

static bool cte_validate(CTEType type, CTEAntennaPattern pat){
/*
  Whether the specified CTE {type} and CTE antenna {pat}tern are valid (for this
  RSS2I board) and are compatible with each other. In error case, i.e. if this
  function returns {false}, that error is reported via the debug module.
*/
	
	if(cte_type_is_valid(type)){
	
		if(cte_antenna_pattern_is_valid(pat, type)){
		
			return true;
		}
		else{
			
			debug_report(DEBUG_CTE_INVALID_ANTENNA_PATTERN);
			return false;
		}
	}
	else{
		
		debug_report(DEBUG_CTE_INVALID_TYPE);
		return false;
	}
}

void cte_unprepare(void){
/*
  Destroys all saved information made by {cte_prepare()}.
*/
	
	size_t i;
	
	for(i = 0; i <= CTEINFO_MAX_HASH_VALUE; ++i){
		
		ctePreproc[CTE_MODE_TX][i] = CTE_PREPROC_UNPREPARED;
		ctePreproc[CTE_MODE_RX][i] = CTE_PREPROC_UNPREPARED;
	}
}

bool cte_prepare(CTEType type, CTEAntennaPattern pat){
/*
  Preprocesses some information regarding the given CTE {type} and antenna
  {pat}tern.
  
  Return value
  ============
  
    Returns false if the specified {type} and {pat}tern are
    incompatible with each other or invalid (with regard to this board). In
    error case, i.e. if this function returns {false}, that error is reported
    via the debug module. Returns {true} otherwise.
*/
	
	size_t hash = cteinfo_hash(type, pat);
	
	if(cte_validate(type, pat)){
		
		/* Preprocess the DFEOPMODE since this lasts a while. */
		
		/* {CTE_MODE_RX}. */
		
		if(pat > CTE_ANTENNA_PATTERN_INVALID_SWITCH
		&& pat < CTE_ANTENNA_PATTERN_ALWAYS_LAST_SWITCH){
			
			/* AoA. */
			ctePreproc[CTE_MODE_RX][hash] = CTE_PREPROC_DFEOPMODE_AOA;
		}
		else if(type != CTE_TYPE_NONE
		     && pat > CTE_ANTENNA_PATTERN_INVALID_CONST
		     && pat < CTE_ANTENNA_PATTERN_ALWAYS_LAST_CONST){
			
			/* AoD. */
			ctePreproc[CTE_MODE_RX][hash] = CTE_PREPROC_DFEOPMODE_AOD;
		}
		else{
			
			/* No CTE. */
			ctePreproc[CTE_MODE_RX][hash] = CTE_PREPROC_DFEOPMODE_DISABLED;
		}
		
		/* {CTE_MODE_TX}. */
		
		if(type != CTE_TYPE_NONE
		&& pat > CTE_ANTENNA_PATTERN_INVALID_CONST
		&& pat < CTE_ANTENNA_PATTERN_ALWAYS_LAST_CONST){
			
			/* AoA. */
			ctePreproc[CTE_MODE_TX][hash] = CTE_PREPROC_DFEOPMODE_AOA;
		}
		else if(pat > CTE_ANTENNA_PATTERN_INVALID_SWITCH
		     && pat < CTE_ANTENNA_PATTERN_ALWAYS_LAST_SWITCH){
			
			/* AoD. */
			ctePreproc[CTE_MODE_TX][hash] = CTE_PREPROC_DFEOPMODE_AOD;
		}
		else{
			
			/* No CTE. */
			ctePreproc[CTE_MODE_TX][hash] = CTE_PREPROC_DFEOPMODE_DISABLED;
		}
		
		return true;
	}
	else{
		
		ctePreproc[CTE_MODE_TX][hash] = CTE_PREPROC_UNPREPARED;
		ctePreproc[CTE_MODE_RX][hash] = CTE_PREPROC_UNPREPARED;
		return false;
	}
}

static void cte_set_conf(CTEType type){
/*
  General CTE configurations.
*/
	
	RADIO_DFECTRL1 &= ~(RADIO_DFECTRL1_NUMBEROF8US_MASK
	                  | RADIO_DFECTRL1_TSWITCHSPACING_MASK
	                  | RADIO_DFECTRL1_TSAMPLESPACINGREF_MASK
	                  | RADIO_DFECTRL1_SAMPLETYPE_MASK
	                  | RADIO_DFECTRL1_TSAMPLESPACING_MASK);
	
	switch(type){
		
		case CTE_TYPE_NONE:
			
			RADIO_DFECTRL1 |= RADIO_DFECTRL1_NUMBEROF8US_VAL(0);
		
		break;
		case CTE_TYPE_2US_7P1:
			
			/* 16 samples during 8 µs reference interval. */
			/* 16 samples during 2 µs sampling slot.      */
			RADIO_DFECTRL1 |= RADIO_DFECTRL1_NUMBEROF8US_VAL(5)
			                | RADIO_DFECTRL1_TSWITCHSPACING_4US
			                | RADIO_DFECTRL1_TSAMPLESPACINGREF_500NS
			                | RADIO_DFECTRL1_SAMPLETYPE_IQ
			                | RADIO_DFECTRL1_TSAMPLESPACING_125NS;
		
		break;
		case CTE_TYPE_2US_7PX1:
			
			/* 32 samples during 8 µs reference interval. */
			/* 16 samples during 2 µs sampling slot.      */
			RADIO_DFECTRL1 |= RADIO_DFECTRL1_NUMBEROF8US_VAL(5)
			                | RADIO_DFECTRL1_TSWITCHSPACING_4US
			                | RADIO_DFECTRL1_TSAMPLESPACINGREF_250NS
			                | RADIO_DFECTRL1_SAMPLETYPE_IQ
			                | RADIO_DFECTRL1_TSAMPLESPACING_125NS;
		
		break;
		default:
			
			debug_report(DEBUG_CTE_UNEXPECTED_TYPE);
	}
}

static void cte_set_sampling(volatile uint8_t * dataPtr){
/*
  CTE IQ sampling related configurations.
*/
	
	#if REVISION >= REVISION_F
		
		RADIO_DFEPACKET.PTR = (uint32_t) dataPtr;
	
	#else
		
		HOOK_UNUSED(dataPtr);
		
	#endif
}


static void cte_set_antenna_const(CTEAntennaPattern pat){
/*
  Enables constant antenna according to the specified antenna {pat}tern.
*/
	
	SwitchAntenna ant;
	uint8_t i;
	
	for(i = 0; i < cteNumOfPins; ++i){
		
		/* Disconnect all GPIO pins from the RADIO peripheral. */
		
		/* This is a workaround for undocumented behaviour!    */
		/* The GPIO pins shouldn't be acquired by the RADIO    */
		/* peripheral in AoA TX or AoD RX mode at all.         */
		/* However, this is apparently the case.               */
		
		RADIO_PSEL.DFEGPIO[i] = (1ul << 31);
	}
	
	if(pat > CTE_ANTENNA_PATTERN_INVALID_CONST
	&& pat < CTE_ANTENNA_PATTERN_ALWAYS_LAST_CONST){
		
		/* Make a {SwitchAntenna} from the given {CTEAntennaPattern}. */
		ant = pat - CTE_ANTENNA_PATTERN_CONST_1 + SWITCH_ANTENNA_1;
		
		switch_set(switch_get_pattern(ant));
	}
	else{
		
		/* Standard antenna                                           */
		/* e.g. in the case of {pat == CTE_ANTENNA_PATTERN_DISABLED}. */
		switch_set(switch_get_standard_pattern());
	}
}

static void cte_set_antenna_switch(CTEAntennaPattern pat){
/*
  Enables antenna switching according to the specified antenna {pat}tern.
*/
	
	#if REVISION >= REVISION_F
	
		uint8_t i;
		
		for(i = 0; i < cteNumOfPins; ++i){
			
			/* TODO: The information retrieved by the function call could */
			/*       be made static, too. This would save a few µs.       */
			RADIO_PSEL.DFEGPIO[i] = switch_get_pin_number_of_bit(i);
		}
		
		RADIO_CLEARPATTERN = 1;
		
		switch(pat){
			
			case CTE_ANTENNA_PATTERN_SWITCH_1TO8:
				
				RADIO_SWITCHPATTERN = cteSwitchGPIOPattern[CTE_ANTENNA_STANDARD];
				RADIO_SWITCHPATTERN = cteSwitchGPIOPattern[CTE_ANTENNA_1];
				RADIO_SWITCHPATTERN = cteSwitchGPIOPattern[CTE_ANTENNA_2];
				RADIO_SWITCHPATTERN = cteSwitchGPIOPattern[CTE_ANTENNA_3];
				RADIO_SWITCHPATTERN = cteSwitchGPIOPattern[CTE_ANTENNA_4];
				RADIO_SWITCHPATTERN = cteSwitchGPIOPattern[CTE_ANTENNA_5];
				RADIO_SWITCHPATTERN = cteSwitchGPIOPattern[CTE_ANTENNA_6];
				RADIO_SWITCHPATTERN = cteSwitchGPIOPattern[CTE_ANTENNA_7];
				RADIO_SWITCHPATTERN = cteSwitchGPIOPattern[CTE_ANTENNA_8];
				
			break;
			default:
			
				debug_report(DEBUG_CTE_UNEXPECTED_ANTENNA_PATTERN);
		}
	
	#else
		
		HOOK_UNUSED(pat);
		
	#endif
}

static void cte_set_switching(CTEAntennaPattern pat, CTEPreproc preproc){
/*
  CTE antenna switching related configurations.
  
  Arguments
  =========
    
    pat        {CTEAntennaPattern} to be configured.
    preproc    {CTEPreproc} value thats corresponds to the {CTEAntennaPattern}
               {pat} (specified) and the {CTEType} (not needed here).
*/
	
	/* Derive operation mode. */
	
	#if REVISION >= REVISION_F
		
		switch(preproc){
		
			case CTE_PREPROC_DFEOPMODE_AOA:
			
				RADIO_DFEMODE = RADIO_DFEMODE_DFEOPMODE_AOA;
				
			break;
			case CTE_PREPROC_DFEOPMODE_AOD:
			
				RADIO_DFEMODE = RADIO_DFEMODE_DFEOPMODE_AOD;
				
			break;
			default:
			
				RADIO_DFEMODE = RADIO_DFEMODE_DFEOPMODE_DISABLED;
		}
	
	#else
		
		HOOK_UNUSED(preproc);
		
	#endif
	
	/* Set antenna switching (if any). */

	if(pat > CTE_ANTENNA_PATTERN_INVALID_SWITCH
	&& pat < CTE_ANTENNA_PATTERN_ALWAYS_LAST_SWITCH){
		
		/* Switch antennas. */
		
		cte_set_antenna_switch(pat);
	}
	else{
		
		cte_set_antenna_const(pat);
	}
}

void cte_disable(void){
/*
  Disables the CTE, i.e. deactivates all Direction Finding features (if any)
  inside the RADIO peripheral, and configures the board to use its standard
  antenna.
*/
	#if REVISION >= REVISION_F
	
		uint8_t i;
		
		RADIO_DFEMODE = RADIO_DFEMODE_DFEOPMODE_DISABLED;
			
		for(i = 0; i < cteNumOfPins; ++i){
			
			/* Disconnect all GPIO pins from the RADIO peripheral. */
			RADIO_PSEL.DFEGPIO[i] = (1ul << 31);
		}
		
		switch_set(switch_get_standard_pattern());
		
	#endif
}

void cte_shutdown(void){
/*
  Disables the CTE feature completely. Locks the HF switch in its standard
  configuration, i.e. all configuration pins are driven LOW.
  
  This can be undone only via call to {cte_init()}.
*/
	
	#if REVISION >= REVISION_F
	
		cte_disable();
		switch_lock();
		
	#endif
}

static void cte_radio_init(CTEType type,
                           CTEAntennaPattern pat,
                           CTEMode mode,
                           volatile uint8_t * dataPtr){
/*
  Configures and initializes the BLE 5.1 Constant Tone Extension (CTE) according
  to the given CTE {type} and CTE antenna {pat}tern during a received
  ({mode == CTE_MODE_RX}) or transmitted ({mode == CTE_MODE_TX}) packet.
  
  {dataPtr} should point to a place in memory where the IQ samples, recorded
  during CTE reception, may be stored. May be {NULL} if {mode == CTE_MODE_TX}.
*/
	
	/* Retrieve some preprocessed information. */
	CTEPreproc preproc = ctePreproc[mode][cteinfo_hash(type, pat)];
	
	if(preproc != CTE_PREPROC_UNPREPARED){
	
		cte_set_conf(type);
		
		if(mode == CTE_MODE_RX){
			
			if(dataPtr){
			
				cte_set_sampling(dataPtr);
			}
			else{
				
				/* This really should not happen.         */
				debug_report(DEBUG_CTE_ADDR_NULL);
				cte_disable();
			}
		}
		
		cte_set_switching(pat, preproc);
	}
	else{
		
		debug_report(DEBUG_CTE_UNPREPARED);
		cte_disable();
	}
}

void cte_radio_init_tx(CTEType type, CTEAntennaPattern pat){
/*
  Configures and initializes the BLE 5.1 Constant Tone Extension (CTE) according
  to the given CTE {type} and CTE antenna {pat}tern during a transmitted packet.
  
  Wrapper function for {cte_radio_init()}.
*/
	
	cte_radio_init(type, pat, CTE_MODE_TX, NULL);
}

void cte_radio_init_rx(CTEType type,
                       CTEAntennaPattern pat,
                       volatile uint8_t * dataPtr){
/*
  Configures and initializes the BLE 5.1 Constant Tone Extension (CTE) according
  to the given CTE {type} and CTE antenna {pat}tern during a received packet.
  
  Wrapper function for {cte_radio_init()}.
*/
	
	cte_radio_init(type, pat, CTE_MODE_RX, dataPtr);
}

uint8_t cte_length(CTEType type){
/*
  Returns the duration of the CTE (including guard period) according to the
  specified CTE {type} in number of µs.
  
  Returns {0} if the specified CTE {type} is invalid.
*/
	
	if(cte_type_is_valid(type)){
		
		switch(type){
			
			case CTE_TYPE_2US_7P1:
				
				return 40;
				
			break;
			case CTE_TYPE_2US_7PX1:
				
				return 40;
				
			break;
			default:
				
				debug_report(DEBUG_CTE_UNEXPECTED_TYPE);
				return 0;
		}
	}
	else{
	
		return 0;
	}
}

bool cte_type_is_valid(CTEType type){
/*
  Whether the specified CTE {type} is valid.
*/
	
	return cteinfo_type_is_valid(type);
}

bool cte_is_sampling_configuration(CTEType type, CTEAntennaPattern ant){
/*
  Whether there will be CTE samplings during/after packet reception with the
  specified CTE configuration.
*/
	
	if(type != CTE_TYPE_NONE && ant != CTE_ANTENNA_PATTERN_DISABLED){
		
		return true;
	}
	else{
		
		return false;
	}
}

bool cte_antenna_pattern_is_valid(CTEAntennaPattern ant, CTEType type){
/*
  Whether the specified CTE {ant}enna pattern is supported by the actual RSS2I
  board revision and type and the given {type} is compatible with that choice.
*/
	
	return cteinfo_antenna_pattern_is_valid(
		board_rev(),
		board_type(),
		ant,
		type
	);
}

uint8_t cte_num_of_antennas(void){
/*
  Returns the number of antennas that may be used to transmit/receive a BLE 5.1
  Constant Tone Extension (CTE). Note, that BLE 5.0 devices will always return
  zero.
*/
	
	return cteinfo_num_of_antennas(board_rev(), board_type());
}

