/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

#define RSS2I_MODULE "cteinfo"

/*
  BLE 5.1 Constant Tone Extension (CTE) related compatibility informations.
  
  Authors
  =======
  
    Copyright (C) 2020-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
  
    2020-08-24    Outsourced from the cte module.
    2021-03-05    Added {cteinfo_hash()}.
    2021-03-11    Do not return negative {cteinfo_hash}es anymore.
    2021-03-11    Constant antenna is allowed only for active CTEs.
    2021-04-29    Support {CTE_2US_7PX1} {CTEType}.
*/


#include <stdbool.h>
#include <stdint.h>
#include <stddef.h>

#include <rss2i/cte.h>
#include <rss2i/board.h>


bool cteinfo_type_is_valid(CTEType type){
/*
  Whether the specified CTE {type} is valid.
*/
	
	if(type >= CTE_TYPE_ALWAYS_LAST_ITEM){
		
		/* Invalid CTE type. */
		return false;
	}
	else{
		
		return true;
	}
}

bool cteinfo_antenna_pattern_is_valid(uint8_t boardRev,
                                      uint8_t boardType,
                                      CTEAntennaPattern ant,
                                      CTEType type){
/*
  Whether the specified CTE {ant}enna pattern is supported by an RSS2I device
  with the specified {boardRev}ision and {boardType} and the given CTE {type} is
  compatible with that choice.
*/
	
	if(
		ant <= CTE_ANTENNA_PATTERN_INVALID_SWITCH
		|| (
		   	   ant >= CTE_ANTENNA_PATTERN_ALWAYS_LAST_SWITCH
		   	&& ant <= CTE_ANTENNA_PATTERN_INVALID_CONST
		   )
		|| (
		   	   ant >= CTE_ANTENNA_PATTERN_ALWAYS_LAST_CONST
		   	&& ant < CTE_ANTENNA_PATTERN_DISABLED
		   )
	){
		
		/* Invalid antenna pattern. */
		return false;
	}
	else if(!cteinfo_type_is_valid(type)){
		
		/* Invalid CTE type. */
		return false;
	}
	else{
		
		if(boardType == BOARD_TYPE_ARR
		&& ant >= CTE_ANTENNA_PATTERN_CONST_1
		&& ant <= CTE_ANTENNA_PATTERN_CONST_8
		&& type != CTE_TYPE_NONE){
			
			/* In-range const antenna and arbitrary (active) CTE type. */
			return true;
		}
		else if(boardRev >= BOARD_REV_F
		     && ant == CTE_ANTENNA_PATTERN_CONST_1
		     && type != CTE_TYPE_NONE){
			
			/* In-range const antenna and arbitrary (active) CTE type. */
			return true;
		}
		else if(boardType == BOARD_TYPE_ARR
		     && ant == CTE_ANTENNA_PATTERN_SWITCH_1TO8
		     && (type == CTE_TYPE_2US_7P1
		      || type == CTE_TYPE_2US_7PX1)){
			
			/* Linear antenna array switch and compatible CTE type. */
			return true;
		}
		else if(ant == CTE_ANTENNA_PATTERN_DISABLED){
			
			/* CTE disabled, i.e. no _active_ participation.      */
			/* Standard antenna during the actual packet is used. */
			return true;
		}
		else{
			
			return false;
		}
	}
}

uint8_t cteinfo_num_of_antennas(uint8_t boardRev, uint8_t boardType){
/*
  Returns the number of antennas that may be used to transmit/receive a BLE 5.1
  Constant Tone Extension (CTE) with an RSS2I device with the specified
  {boardRev}ision and {boardType}. Note, that BLE 5.0 devices will always return
  zero.
*/
	
	if(boardRev <= BOARD_REV_E){
		
		/* BLE 5.0 device. */
		return 0;
	}
	else{
		
		if(boardType == BOARD_TYPE_ARR){
			
			return 8;
		}
		else{
			
			return 1;
		}
	}
}


size_t cteinfo_hash(CTEType type, CTEAntennaPattern pat){
/*
  Maps the (valid) combination of {type} and {pat} to a unique ID. The ID may
  not be unique in the case of an invalid {type} or {pat}.
*/
		
	if(type < CTE_TYPE_ALWAYS_LAST_ITEM){
	
		/* Process those cases first that need longest preparation time */
		/* during frame/RADIO preparation during measurement.           */
		
		if(pat < CTE_ANTENNA_PATTERN_ALWAYS_LAST_SWITCH){
			
			/* Valid switching pattern or                          */
			/* {CTE_ANTENNA_PATTERM_INVALID_SWITCH}.               */
			
			/* We do not check for the latter case explicitly      */
			/* in order not to slow down hash generation for       */
			/* the switching patterns, because the CTE preparation */
			/* in this case needs long time anyway.                */
			
			/* Note, that the uniquness is promised only for valid */
			/* types and valid patterns.                           */
			
			/* Also note, that this might be dangerous to do if    */
			/* {CTE_ANTENNA_PATTERN_INVALID_SWITCH} is not zero.   */
			/* In this case we may return a negative value.        */
			/* However, the corresponding unit test should catch   */
			/* this anyway.                                        */
			
			return (pat - CTE_ANTENNA_PATTERN_INVALID_SWITCH)
			     * CTE_TYPE_ALWAYS_LAST_ITEM
			     + type;
		}
		else if(pat < CTE_ANTENNA_PATTERN_ALWAYS_LAST_CONST
		     && pat > CTE_ANTENNA_PATTERN_INVALID_CONST){
		     
			/* Valid const pattern.                                */
			
			/* Here, we check both boundaries in order not to      */
			/* return a negative value. The corresponding unit     */
			/* test would complain otherwise.                      */
			
			return (pat - CTE_ANTENNA_PATTERN_INVALID_CONST
			            + CTE_ANTENNA_PATTERN_ALWAYS_LAST_SWITCH
			            - CTE_ANTENNA_PATTERN_INVALID_SWITCH)
			     * CTE_TYPE_ALWAYS_LAST_ITEM
			     + type;
		}
		else if(pat == CTE_ANTENNA_PATTERN_DISABLED){
			
			return (CTE_ANTENNA_PATTERN_ALWAYS_LAST_SWITCH
			      - CTE_ANTENNA_PATTERN_INVALID_SWITCH
			      + CTE_ANTENNA_PATTERN_ALWAYS_LAST_CONST
			      - CTE_ANTENNA_PATTERN_INVALID_CONST)
			     * CTE_TYPE_ALWAYS_LAST_ITEM
			     + type;
		}
		else{
		
			/* Invalid pattern. */
			
			return (CTE_ANTENNA_PATTERN_ALWAYS_LAST_SWITCH
			      - CTE_ANTENNA_PATTERN_INVALID_SWITCH
			      + CTE_ANTENNA_PATTERN_ALWAYS_LAST_CONST
			      - CTE_ANTENNA_PATTERN_INVALID_CONST)
			     * CTE_TYPE_ALWAYS_LAST_ITEM
			     + CTE_TYPE_ALWAYS_LAST_ITEM;
		}
	}
	else{
		
		/* Invalid type. */
		
		return (CTE_ANTENNA_PATTERN_ALWAYS_LAST_SWITCH
		      - CTE_ANTENNA_PATTERN_INVALID_SWITCH
		      + CTE_ANTENNA_PATTERN_ALWAYS_LAST_CONST
		      - CTE_ANTENNA_PATTERN_INVALID_CONST)
		     * CTE_TYPE_ALWAYS_LAST_ITEM
		     + CTE_TYPE_ALWAYS_LAST_ITEM;
	}
}

