/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

#define RSS2I_MODULE "debug"

/*
  Error reporting system with the possibility to send debug messages via UART.
  
  Authors
  =======
  
    Copyright (C) 2018-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
  
    2018-09-27    Initial release.
    2018-10-30    Added CONTROL and OTA.
    2018-11-21    Added MEASURE.
    2018-12-21    Added WATCHDOG.
    2019-01-10    Added RADIO.
    2019-01-18    Added CMD.
    2019-03-11    Added BOARD.
    2019-04-11    Added BUFFER.
    2019-04-12    Rewritten. USB/CMD API added.
    2019-06-03    Added board revision and type related error vectors.
    2019-06-07    Added measure-invalid-indices related error vectors.
    2019-06-14    Added MLC.
    2019-06-24    Added board-specific battery threshold related debug vectors.
    2019-06-26    ADDED {debug_ok()}.
    2019-08-30    ADDED {debug_get_message()}.
    2019-10-38    ADDED UARTE state machine related error vectors.
    2020-02-08    ADDED {BUFFER_NOTHING_TO_FINISH}.
    2020-02-12    The repetition delay is now controlled via timer.
    2020-02-13    Adjusted message of {DEBUG_MLC_TOO_BUSY}.
    2020-03-30    Renamed {DebugSource} to {DebugErrorCase} and rearrange order.
    2020-03-30    Replace {debug_ok()} with {debug_get_num_of_errors()}.
    2020-06-08    Added {DEBUG_BOARD_*_NOT_EMULATABLE}.
    2020-06-18    Added {DEBUG_MEASURE_RESULT_BUFFER_ADDR_NULL}.
    2020-07-03    Added {DEBUG_TIMER_LIMIT_EXCEEDED}.
    2020-07-06    Added {DEBUG_SCHEDULER_INVALID_TASK}.
    2020-07-06    Adjusted message of {DEBUG_MLC_TOO_BUSY}.
    2020-07-06    Added {DEBUG_CMD_INVALID_WAIT_STATE}.
    2020-07-06    Added support for different {DEBUG_LEVEL}s.
    2020-07-07    Switched to DMA based debug message buffering.
    2020-07-07    Adjusted message of {DEBUG_MLC_TOO_BUSY}.
    2020-07-07    Added {DEBUG_MEASURE_LIVELOG_PARALLEL_ACCESS}.
    2020-07-07    Fixed the message for the {DEBUG_MLC_TOO_BUSY} error case.
    2020-07-09    Added {DEBUG_POWER_BROWNOUT}.
    2020-07-31    Renamed error case to {DEBUG_LIVELOG_PARALLEL_ACCESS}.
    2020-08-03    Added {DEBUG_CTE_*} error cases.
    2020-08-13    Added {DEBUG_SWITCH_*} error cases.
    2021-02-25    Added {DEBUG_MEASURE_RADIO_PACKET_ADDR_NULL}.
    2021-03-05    Added {DEBUG_CTE_UNPREPARED}.
    2021-03-10    Added {DEBUG_CTE_ADDR_NULL}.
    2021-03-18    Added {DEBUG_MEASURE_INVALID_RADIO_MODE}.
    2021-03-25    Added {DEBUG_GPIOTE_OUT_OF_RANGE}.
    2021-03-25    Removed {*_SCHEDULER_INVALID_TASK} and {*_MEMORY_NO_SPACE_*}.
    2021-03-29    Added {DEBUG_RADIO_INVALID_CONST_CONF}.
    2021-04-01    Enhanced the maximal number of error cases significantly.
    2021-04-01    Added {DEBUG_CODER_*} error cases.
    2021-04-02    Added {DEBUG_MEMORY_INCONSISTENT_*} and {DEBUG_MEMORY_LOST_*}.
    2021-04-02    Make {debugLEDPattern} {const}.
    2021-04-15    Rename {DEBUG_MEMORY_LOST_RESULTS} to {*_INVALID_STATE}.
    2021-04-15    Added {DEBUG_POSTPROC_*} error cases.
    2021-04-21    Added {DEBUG_CODER_INVALID_COMMAND_SIZE}.
  
  Description
  ===========

    There are two reporting possibilities: The first one is to print a debug
    message via UART directly. Those messages don't get repeated. The second
    possibility is to report an error/give some warning from a predefined list
    of error/warning cases. Each of those error/warning cases become (once
    activated) reported periodically via UART. Depending on the reported case,
    the RSS2I battery indicator LED changes its status.
*/

#include <stdint.h>
#include <stdbool.h>
#include <stddef.h>

#include <rss2i/buffer.h>
#include <rss2i/suart.h>
#include <rss2i/debug.h>
#include <rss2i/led.h>
#include <rss2i/print.h>
#include <rss2i/timer.h>


static const char * debugMessage[DEBUG_ERROR_CASE_ALWAYS_LAST_ITEM] = {
/*
  Debug messages for the predefined error cases.
 
  This should be consistent with
  ==============================

    - the list of predefined error cases within the corresponding header file
    - the API documentation
*/

	"Buffer overflow while trying to send a debug message.",
	"Could not send a debug message due to parallel access.",
	"Unknown error case.",
	"Crystal oscillator didn't ramp up successfully.",
	"Battery voltage is low.",
	"Buffer overflow while trying to send via USB.",
	"Buffer overflow while receiving an instruction via USB.",
	"There are no additional timers available.",
	"A timer request exceeded the upper time bound.",
	"Buffer overflow during communication with FLASH device.",
	"There are two or more measurement results within the same frame and the same cycle.",
	"FLASH memory may be corrupt.",
	"The size of a measurement result does not match the given format specifier.",
	"An invalid state of the memory write results state machine occured.",
	"Received a corrupt instruction via USB.",
	"Received an unknown instruction via USB.",
	"Expected command to handle but no one was given.",
	"Buffer overflow while trying to store measurement results.",
	"Unexpected entry address of the measurement result buffer.",
	"Unexpected RADIO packet address.",
	"Should check for disruption but this is impossible.",
	"There isn't any measurement configuration to start with.",
	"Won't update to invalid frame index.",
	"Won't update to invalid cycle index.",
	"Invalid RADIO mode.",
	"Parallel access while sending the livelog?",
	"A watchdog timeout caused the last reset.",
	"Invalid measurement packet RX frequency given.",
	"RADIO-TX-Ramp-Up hasn't finished when it was expected to have.",
	"RADIO-TX-Start was requested but hasn't been prepared yet.",
	"The RADIO hasn't been disabled when it was expected to have.",
	"RADIO state machine error.",
	"OTA message was oversized and has been truncated.",
	"Buffer overflow while trying to send OTA.",
	"Buffer overflow while receiving OTA data.",
	"Invalid RADIO mode or frequency deviation.",
	"Received an unknown command.",
	"Expected command to process but no one was given.",
	"Processing of an iterable command was requested.",
	"Invalid wait state during command handling.",
	"Board specific parameter was requested but had not been read yet from UICR.",
	"Board ID was requested but had not been stored yet inside UICR.",
	"Board revision was requested but had not been stored yet inside UICR.",
	"Board type was requested but had not been stored yet inside UICR.",
	"Battery threshold was requested but had not been stored yet inside UICR.",
	"The stored board type cannot be emulated with the actual hardware revision.",
	"The stored revision cannot be emulated with the actual hardware revision.",
	"Given data exceeded the buffer's width.",
	"There is no buffer entry to finish.",
	"Tried to use an uninitialized inter peripheral shortcut.",
	"The specified inter peripheral shortcut is unknown or out-of-range.",
	"Less than 1.5 complete asynchronous processing cycles per measurement frame.",
	"The UARTE state machine may be in an unexpected state.",
	"A pending brownout was detected. Expect malfunction!",
	"An unexpected CTE type occured.",
	"An unexpected CTE antenna pattern occured.",
	"An invalid CTE type was specified.",
	"An invalid CTE antenna type was specified.",
	"The CTE was unprepared.",
	"The specified CTE IQ sample address is invalid.",
	"An invalid antenna switch pattern bit number was specified.",
	"An invalid HF switch antenna was specified.",
	"The specified GPIOTE channel is unknown or out-of-range.",
	"An invalid format specifier occured.",
	"An invalid format byte occured.",
	"An RSS integer value had to be truncated to a valid RSS format byte.",
	"An invalid coder command size occured.",
	"An invalid state occured within the postprocess state machine.",
	"The number of raw RSS samples seems too low.",
	"Unexpected number of samples to postprocess."
};

static const LEDPattern debugLEDPattern[DEBUG_ERROR_CASE_ALWAYS_LAST_ITEM] = {
	
	LED_ON,    /* DEBUG_BUFFER_OVERFLOW             */
	LED_ON,    /* DEBUG_PARALLEL_ACCESS             */
	LED_ON,    /* DEBUG_UNKNOWN_ERROR_CASE          */
	LED_ON,    /* CRYSTAL_OSCILLATOR_FAILED         */
	LED_BLINK, /* BATTERY_LOW                       */
	LED_ON,    /* USB_BUFFER_OVERFLOW               */
	LED_ON,    /* USB_RECV_BUFFER_OVERFLOW          */
	LED_ON,    /* NO_FREE_TIMERS                    */
	LED_ON,    /* TIMER_LIMIT_EXCEEDED              */
	LED_ON,    /* MEMORY_BUFFER_OVERFLOW            */
	LED_ON,    /* MEMORY_INDEX_ERROR                */
	LED_ON,    /* MEMORY_CRC_FAILED                 */
	LED_ON,    /* MEMORY_INCONSISTENT_SIZE          */
	LED_ON,    /* MEMORY_INVALID_STATE              */
	LED_ON,    /* CONTROL_USB_CRC_FAILED            */
	LED_ON,    /* CONTROL_UNKNOWN_CMD               */
	LED_ON,    /* CONTROL_INVALID_CMD               */
	LED_ON,    /* MEASURE_RESULT_BUFFER_OVERFLOW    */
	LED_ON,    /* MEASURE_RESULT_BUFFER_ADDR_NULL   */
	LED_ON,    /* MEASURE_RADIO_PACKET_ADDR_NULL    */
	LED_ON,    /* MEASURE_DISRUPT_NOTHING_TO_CHECK  */
	LED_ON,    /* MEASURE_NO_CONFIG                 */
	LED_ON,    /* MEASURE_INVALID_FRAME_INDEX       */
	LED_ON,    /* MEASURE_INVALID_CYCLE_INDEX       */
	LED_ON,    /* MEASURE_INVALID_RADIO_MODE        */
	LED_ON,    /* LIVELOG_PARALLEL_ACCESS           */
	LED_ON,    /* WATCHDOG_TRIGGERED                */
	LED_ON,    /* RADIO_INVALID_PARAMETER           */
	LED_ON,    /* RADIO_INVALID_PARAMETERS          */
	LED_ON,    /* RADIO_TX_NOT_READY                */
	LED_ON,    /* RADIO_TX_UNPREPARED               */
	LED_ON,    /* RADIO_NOT_DISABLED                */
	LED_ON,    /* RADIO_UNEXPECTED_STATE            */
	LED_ON,    /* RADIO_OTA_MESSAGE_OVERSIZED       */
	LED_ON,    /* RADIO_OTA_SEND_BUFFER_OVERFLOW    */
	LED_ON,    /* RADIO_OTA_RECV_BUFFER_OVERFLOW    */
	LED_ON,    /* RADIO_INVALID_CONST_CONF          */
	LED_ON,    /* CMD_UNKNOWN_CMD                   */
	LED_ON,    /* CMD_INVALID_CMD                   */
	LED_ON,    /* CMD_ITERABLE_CALLED               */
	LED_ON,    /* CMD_INVALID_WAIT_STATE            */
	LED_ON,    /* BOARD_NOT_READ_YET                */
	LED_ON,    /* BOARD_ID_INVALID                  */
	LED_ON,    /* BOARD_REV_INVALID                 */
	LED_ON,    /* BOARD_TYPE_INVALID                */
	LED_ON,    /* BOARD_BATTERY_THRESH_INVALID      */
	LED_ON,    /* BOARD_TYPE_NOT_EMULATABLE         */
	LED_ON,    /* BOARD_REV_NOT_EMULATABLE          */
	LED_ON,    /* BUFFER_WIDTH_EXCEEDED             */
	LED_ON,    /* BUFFER_NOTHING_TO_FINISH          */
	LED_ON,    /* PPI_UNINITIALIZED                 */
	LED_ON,    /* MLC_TOO_BUSY                      */
	LED_ON,    /* UARTE_STATE_MACHINE_MISMATCH      */
	LED_ON,    /* POWER_BROWNOUT                    */
	LED_ON,    /* CTE_UNEXPECTED_TYPE               */
	LED_ON,    /* CTE_UNEXPECTED_ANTENNA_PATTERN    */
	LED_ON,    /* CTE_INVALID_TYPE                  */
	LED_ON,    /* CTE_INVALID_ANTENNA_PATTERN       */
	LED_ON,    /* CTE_UNPREPARED                    */
	LED_ON,    /* CTE_ADDR_NULL                     */
	LED_ON,    /* SWITCH_INVALID_BIT_NUMBER         */
	LED_ON,    /* SWITCH_INVALID_ANTENNA            */
	LED_ON,    /* GPIOTE_OUT_OF_RANGE               */
	LED_ON,    /* CODER_INVALID_FORMAT_SPECIFIER    */
	LED_ON,    /* CODER_INVALID_FORMAT_BYTE         */
	LED_ON,    /* CODER_TRUNCATED_VALID_RSS         */
	LED_ON,    /* CODER_INVALID_COMMAND_SIZE        */
	LED_ON,    /* POSTPROC_INVALID_STATE            */
	LED_ON,    /* POSTPROC_LOW_RSS_AMOUNT           */
	LED_ON,    /* POSTPROC_UNEXPECTED_SAMPLE_AMOUNT */
};

static uint16_t debugGlobalID[DEBUG_ERROR_CASE_ALWAYS_LAST_ITEM] = {0};
static uint16_t debugLocalID[DEBUG_ERROR_CASE_ALWAYS_LAST_ITEM] = {0};
static uint16_t debugLastID = 0;

static volatile bool debugNextTurn = false;


void debug_init(void){
/*
  Initializes this module. May be called multiple times.
*/
	
	static bool initialized = false;
	static char data[DEBUG_BUFFER_LENGTH][DEBUG_BUFFER_WIDTH];
	static size_t size[DEBUG_BUFFER_LENGTH];
	
	if(!initialized){
	
		suart_init((16000000ul + DEBUG_BAUDRATE / 2) / DEBUG_BAUDRATE,
			       &data[0][0],
			       &size[0],
			       DEBUG_BUFFER_LENGTH,
			       DEBUG_BUFFER_WIDTH);
	
		led_init();
		
		/* Trigger the repetion of warnings/errors periodically. */
		timer_flag(&debugNextTurn, DEBUG_REPETITION_DELAY, TIMER_REPEAT);
		
		initialized = true;
	}
}

static bool debug_print(const char * s1, const char * s2, const char * s3,
                        const char * s4, const char * s5, const char * s6){
/*
  Prints the concatenated strings onto serial debug interface.
*/
	
	/* Avoid ISR interruptions (see buffer.c docs). */
	static volatile bool locked = false;
	
	char * tmpString;
	
	size_t i;
	size_t j;
	
	if(!locked){
		
		locked = true;
		
		if(!(tmpString = suart_send_request())){
		
			debug_report(DEBUG_DEBUG_BUFFER_OVERFLOW);
			locked = false;
			return false;
		}
		
		i = 0;
		j = 0;
		while(i < DEBUG_BUFFER_WIDTH - 1 && s1[j]){ tmpString[i] = s1[j]; ++i; ++j; }
		j = 0;
		while(i < DEBUG_BUFFER_WIDTH - 1 && s2[j]){ tmpString[i] = s2[j]; ++i; ++j; }
		j = 0;
		while(i < DEBUG_BUFFER_WIDTH - 1 && s3[j]){ tmpString[i] = s3[j]; ++i; ++j; }
		j = 0;
		while(i < DEBUG_BUFFER_WIDTH - 1 && s4[j]){ tmpString[i] = s4[j]; ++i; ++j; }
		j = 0;
		while(i < DEBUG_BUFFER_WIDTH - 1 && s5[j]){ tmpString[i] = s5[j]; ++i; ++j; }
		j = 0;
		while(i < DEBUG_BUFFER_WIDTH - 1 && s6[j]){ tmpString[i] = s6[j]; ++i; ++j; }
		
		tmpString[i] = '\0';
		
		if(suart_send_finish(tmpString, i + 1)){
			
			locked = false;
			return true;
		}
		else{
			
			debug_report(DEBUG_DEBUG_PARALLEL_ACCESS);
			locked = false;
			return false;
		}
	}
	else{
		
		debug_report(DEBUG_DEBUG_PARALLEL_ACCESS);
		return false;
	}
}

bool debug_info(const char * moduleName,
                const char * submoduleName,
                const char * message){
/*
  Sends the specified debug info via serial debug port.
  Returns {false} when the debug output buffer is full.
  Returns {true} otherwise.
*/
	
	if(DEBUG_MESSAGES_ENABLED){
	
		return debug_print(moduleName, ": ", submoduleName, ": ", message, "\r\n");
	}
	else{
		
		return true;
	}
}

bool debug_message(const char * message){
/*
  Sends the specified debug message via serial debug port.
  Returns {false} when the debug output buffer is full.
  Returns {true} otherwise.
*/
	
	if(DEBUG_MESSAGES_ENABLED){
	
		return debug_print(message, "\r\n", "", "", "", "");
	}
	else{
		
		return true;
	}
}

void debug_report(DebugErrorCase errorCase){
/*
  Reports a known error case via the serial debug port.
*/
	
	if(errorCase < DEBUG_ERROR_CASE_ALWAYS_LAST_ITEM){
	
		/* Assign an ID to this report in order to remeber it */
	
		debugGlobalID[errorCase] = ++debugLastID;   /* Overall ID        */
		++debugLocalID[errorCase];                  /* Error-specific ID */
	
		/* Indicate error status with onboard LED */
		if(debugLEDPattern[errorCase] != LED_OFF){
			led_set(LED_RED, debugLEDPattern[errorCase]);
		}
	}
	else{
		
		debug_report(DEBUG_DEBUG_UNKNOWN_ERROR_CASE);
	}
}

void debug_trigger(void){
/*
  (Re)prints debug messages (one per error case (if any) per call).
  This function should be called "frequently".
*/
	
	static uint8_t casePtr = 0;
	
	/* Don't put the string onto the stack. */
	static char idString[] = "[0x00 0x0000/0x0000] ";
	
	if(debugNextTurn){
	
		if(!suart_busy()){
		
			if(debugGlobalID[casePtr]){
			
				print_uhex(casePtr, idString, 3, 2);
				print_uhex(debugLocalID[casePtr], idString, 8, 4);
				print_uhex(debugGlobalID[casePtr], idString, 15, 4);
				
				debug_print(idString, debugMessage[casePtr], "\r\n", "", "", "");
			}
		
			if(++casePtr == DEBUG_ERROR_CASE_ALWAYS_LAST_ITEM){
				
				debugNextTurn = false;
				casePtr = 0;
			}
		}
	}
}

uint16_t debug_get_num_of_errors(void){
/*
  Returns the number of errors occured. (API for cmd)
*/
	
	return debugLastID;
}

uint16_t debug_get_global_id(DebugErrorCase src){
/*
  Returns the global error ID of the specified error case. (API for cmd)
*/
	if(src >= DEBUG_ERROR_CASE_ALWAYS_LAST_ITEM){
		
		return 0;
	}
	else{
		
		return debugGlobalID[src];
	}
}

uint16_t debug_get_local_id(DebugErrorCase src){
/*
  Returns the local error ID of the specified error case. (API for cmd)
*/
	if(src >= DEBUG_ERROR_CASE_ALWAYS_LAST_ITEM){
		
		return 0;
	}
	else{
		
		return debugLocalID[src];
	}
}

const char * debug_get_message(DebugErrorCase src){
/*
  Returns the debug message for the specified error case. (API for librss2i)
*/
	if(src >= DEBUG_ERROR_CASE_ALWAYS_LAST_ITEM){
		
		return NULL;
	}
	else{
		
		return debugMessage[src];
	}
}

size_t debug_get_num_of_error_cases(void){
/*
  Returns the number of available error cases.
*/
	
	return DEBUG_ERROR_CASE_ALWAYS_LAST_ITEM;
}

