/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

#define RSS2I_MODULE "errata"

/*
  Some errata workarounds.
   
  Authors
  =======
  
    Copyright (C) 2021-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
  
    2021-02-28    Initial release.
  
  Description
  ===========

    Provides some functionality to workaround known bugs in the nRF52832 or
    nRF52811 BLE SoC.
  
  Errata
  ======
    
    Errata [232] FICR: IEEE 802.15.4 RADIO disabling time is too long
    -----------------------------------------------------------------
     
     Without this workaround the RADIO needs up to 64 µs to disable when
     configured in IEEE 802.15.4 RADIO mode.
     
     This module provides a workaround that can be applied using
     {errata_232_apply_workaround()} before configuring the RADIO in IEEE
     802.15.4 mode. This workaround must be reverted prior to configuring the
     RADIO in other modes by using {errata_232_revert_workaround()}.
*/

#include <stdint.h>
#include <stdbool.h>

volatile uint32_t errata232TMP1;
volatile uint32_t errata232TMP2;


/*
  Errata [232] FICR: IEEE 802.15.4 RADIO disabling time is too long.
*/

static void errata_232_init(void){
/*
  Stores boot configuration of two affected registers for later restore.
*/
	
	errata232TMP1 = *(uint32_t volatile *) UINT32_C(0x40001778);
	errata232TMP2 = *(uint32_t volatile *) UINT32_C(0x4000172c);
}

void errata_232_apply_workaround(void){
/*
  Applies the proposed workaround.
*/
	
	*(uint32_t volatile *) UINT32_C(0x40001778)
		= (*(uint32_t volatile *) UINT32_C(0x40001778) & UINT32_C(0xffffff00))
		| UINT32_C(0x80000010);
	*(uint32_t volatile *) UINT32_C(0x4000172c)
		= (*(uint32_t volatile *) UINT32_C(0x4000172c) & UINT32_C(0xffffff00))
		| UINT32_C(0x8000004d);
}

void errata_232_revert_workaround(void){
/*
  Reverts the proposed workaround.
*/
	
	*(uint32_t volatile *) UINT32_C(0x40001778) = errata232TMP1;
	*(uint32_t volatile *) UINT32_C(0x4000172c) = errata232TMP2;
}

void errata_init(void){
/*
  Initializes the errara module ad applies some errata workarounds and/or
  prepares such workarounds.
*/
	
	static bool initialized = false;
	
	if(!initialized){
	
		/* Errata [232] FICR: IEEE 802.15.4 RADIO disabling time is too long. */
		errata_232_init();
		
		initialized = true;
	}
}

