/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

#define RSS2I_MODULE "led"

/*
  Controls the onboard LEDs.
  
  Authors
  =======
  
    Copyright (C) 2018-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
  
    2018-07-24    Initial release.
    2020-06-10    Fitted to revision F (missing PWM and TIMER peripherals).
    2020-08-12    Use {REVISION_*} macros instead of magic numbers.
    2020-08-12    Disable {GPIO6} if measurement frame debug mode is enabled.
    2021-03-27    Use {NVIC_RSS2I_*_PRIORITY} macros.
    2021-04-29    Include RSS2I nvic module needed for revisions prior to F.
  
  Description
  ===========

    The three onboard LEDs ({LED_RED}, {LED_GREEN} and {LED_BLUE}) may be off
    ({LED_OFF}) or on ({LED_ON}) or may blink ({LED_BLINK}). The module has to
    get initialized using {led_init()}.
  
  Some Details
  ============
    
    With revisions A to E the LED management is completely done by this module.
    A timer peripheral schedules the PWM sequences for each of the three LEDs.
    
    With revision F (and higher) the LED module just triggers the blinking
    sequence. The LED management is done by the PWM module which utilzes just
    a single PWM peripheral. The missing TIMER peripheral gets replaced by
    a timer from the timer module. However, the API remains the same!
    
*/

#include <stdint.h>
#include <stdbool.h>

#include <libopencm3/cm3/nvic.h>
#include <libopencm3/nrf5/gpio.h>
#include <libopencm3/nrf5/timer.h>

#include <rss2i/led.h>
#include <rss2i/pwm.h>
#include <rss2i/timer.h>
#include <rss2i/debug.h>
#include <rss2i/rev.h>
#include <rss2i/nvic.h>

#include <rss2i/config/led.h>


#if REVISION < REVISION_F

	static volatile LEDPattern ledPattern[3];
	
#else

	static volatile TimerID ledTimer;
	
#endif


void led_init(void){
/*
  Initializes this module. May be called multiple times.
*/
	static bool initialized = false;
	
	if(!initialized){
		
		#if DEBUG_MEASUREMENT_FRAME               \
		 && DEBUG_MEASUREMENT_FRAME_GPIO != GPIO6 \
		 && DEBUG_MEASUREMENT_FRAME_GPIO != GPIO7 \
		 && DEBUG_MEASUREMENT_FRAME_GPIO != GPIO8
		
			#error "Failed to enable measurement frame debug GPIO."
			
		#endif
		
		/* Red LED */
		GPIO_PIN_CNF[6] |= GPIO_PIN_CNF_DRIVE_H0S1;
		/* Yellow LED */
		GPIO_PIN_CNF[7] |= GPIO_PIN_CNF_DRIVE_H0S1;
		/* Green LED */
		GPIO_PIN_CNF[8] |= GPIO_PIN_CNF_DRIVE_H0S1;
	
		GPIO_DIRSET = GPIO6 | GPIO7 | GPIO8;
		GPIO_OUTSET = GPIO6 | GPIO7 | GPIO8;
		
		#if REVISION < REVISION_F
		
			ledPattern[LED_GREEN] = LED_OFF;
			ledPattern[LED_YELLOW] = LED_OFF;
			ledPattern[LED_RED] = LED_OFF;
	
			nvic_enable_irq(NVIC_TIMER4_IRQ);
			nvic_set_priority(NVIC_TIMER4_IRQ, NVIC_RSS2I_LOW_PRIORITY);
		
			TIMER_INTENSET(TIMER4) = TIMER_INTENSET_COMPARE0_SET
			                       | TIMER_INTENSET_COMPARE1_SET
			                       | TIMER_INTENSET_COMPARE2_SET;
			                       
			TIMER_SHORTS(TIMER4) = TIMER_SHORTS_COMPARE3_CLEAR_ENABLED;
			TIMER_PRESCALER(TIMER4) = 4;
			TIMER_BITMODE(TIMER4) = TIMER_BITMODE_BITMODE_32Bit;
		
			/* Definition of the ((maybe) overlapping) time slots for all three LEDs. */
			TIMER_CC(TIMER4)[0] = (LED_SLOT_PAUSE / 2 + LED_SLOT_OFFSET * 0) * 1000ul;
			TIMER_CC(TIMER4)[1] = (LED_SLOT_PAUSE / 2 + LED_SLOT_OFFSET * 1) * 1000ul;
			TIMER_CC(TIMER4)[2] = (LED_SLOT_PAUSE / 2 + LED_SLOT_OFFSET * 2) * 1000ul;
			TIMER_CC(TIMER4)[3] = (LED_SLOT_PAUSE / 1 + LED_SLOT_OFFSET * 2) * 1000ul;
		
			pwm_init();
		
			TIMER_TASKS_START(TIMER4) = 1;
			
		#else
		
			ledTimer = timer_call(pwm_start, LED_SLOT_PAUSE / 1 + LED_SLOT_OFFSET * 2, TIMER_REPEAT);
			pwm_init();
			
		#endif
		
		initialized = true;
	}
}

void led_trigger(void){
/*
  Triggers the start of a blink sequence (all three LEDs and a pause before).
*/
	#if REVISION < REVISION_F
	
		TIMER_TASKS_CLEAR(TIMER4) = 1;
		
	#else
	
		timer_reset(ledTimer);
		pwm_start();
		
	#endif
}

void led_set(LED led, LEDPattern pattern){
/*
  Changes the state of the specified {led} into the specified {pattern}.
*/
	#if REVISION < REVISION_F
	
		ledPattern[led] = pattern;
		
		if(pattern == LED_ON){
			
			pwm_start_const(led);
		}
		else if(pattern == LED_OFF){
			
			pwm_off(led);
		}
		
	#else
	
		pwm_set(led, pattern);
		
	#endif
}

#if REVISION < REVISION_F
	
	void timer4_isr(void){
	/*
	  Starts the blinking sequences for each of the three LEDs.
	*/
		
		if(TIMER_EVENTS_COMPARE(TIMER4)[0]){
			
			TIMER_EVENTS_COMPARE(TIMER4)[0] = 0;
			
			if(ledPattern[LED_GREEN] == LED_BLINK){
			
				pwm_start_seq_green();
			}
			
			while(TIMER_EVENTS_COMPARE(TIMER4)[0]);
		}
		else if(TIMER_EVENTS_COMPARE(TIMER4)[1]){
			
			TIMER_EVENTS_COMPARE(TIMER4)[1] = 0;
			
			if(ledPattern[LED_YELLOW] == LED_BLINK){
			
				pwm_start_seq_yellow();
			}
			
			while(TIMER_EVENTS_COMPARE(TIMER4)[1]);
		}
		else if(TIMER_EVENTS_COMPARE(TIMER4)[2]){
			
			TIMER_EVENTS_COMPARE(TIMER4)[2] = 0;
			
			if(ledPattern[LED_RED] == LED_BLINK){
			
				pwm_start_seq_red();
			}
			
			while(TIMER_EVENTS_COMPARE(TIMER4)[2]);
		}
	}
	
#endif

