/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

#define RSS2I_MODULE "livelog"

/*
  Live logging of received measurement packets via USB.
  
  Authors
  =======
  
    Copyright (C) 2018-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
  
    2020-07-31    Outsourced from the measure module.
    2021-02-25    Added BLE 2MBit prefixes.
    2021-03-18    Added {radioMode} argument in {livelog_log()}.
    2021-03-19    Added support for {LIVELOG_ENTRY_LOCAL_RSS_SAMPLE_AMOUNT}.
  
  Description
  ===========

    An entry of the so called live log is generated everytime a valid
    measurement packet was received during measurement mode. This entry
    contains informations about the received packet as well as so called
    remote (meta) data that was sent within that packet. The assembled entry
    is sent via USB.
*/

#include <stdint.h>
#include <stddef.h>
#include <stdbool.h>

#include <rss2i/radio.h>
#include <rss2i/memory.h>
#include <rss2i/battery.h>
#include <rss2i/debug.h>
#include <rss2i/usb.h>
#include <rss2i/random.h>
#include <rss2i/hook.h>

#include <rss2i/config/livelog.h>
#include <rss2i/config/usb.h>


typedef enum {
	
	LIVELOG_ENTRY_FRAME,
	LIVELOG_ENTRY_CYCLE_LSB,
	LIVELOG_ENTRY_CYCLE_ISB,
	LIVELOG_ENTRY_CYCLE_MSB,
	LIVELOG_ENTRY_LOCAL_RSS_INTGR,
	LIVELOG_ENTRY_LOCAL_RSS_FRACT,
	LIVELOG_ENTRY_REMOTE_RSS_0_INTGR,
	LIVELOG_ENTRY_REMOTE_RSS_0_FRACT,
	LIVELOG_ENTRY_REMOTE_RSS_1_INTGR,
	LIVELOG_ENTRY_REMOTE_RSS_1_FRACT,
	LIVELOG_ENTRY_REMOTE_RSS_2_INTGR,
	LIVELOG_ENTRY_REMOTE_RSS_2_FRACT,
	LIVELOG_ENTRY_REMOTE_RSS_3_INTGR,
	LIVELOG_ENTRY_REMOTE_RSS_3_FRACT,
	LIVELOG_ENTRY_REMOTE_RSS_4_INTGR,
	LIVELOG_ENTRY_REMOTE_RSS_4_FRACT,
	LIVELOG_ENTRY_REMOTE_RSS_5_INTGR,
	LIVELOG_ENTRY_REMOTE_RSS_5_FRACT,
	LIVELOG_ENTRY_LOCAL_MEMORY_LOAD,
	LIVELOG_ENTRY_REMOTE_MEMORY_LOAD,
	LIVELOG_ENTRY_LOCAL_BATTERY_LEVEL,
	LIVELOG_ENTRY_REMOTE_BATTERY_LEVEL,
	LIVELOG_ENTRY_LOCAL_RSS_SAMPLE_AMOUNT,
	
	LIVELOG_ENTRY_ALWAYS_LAST_ITEM
	
} LivelogEntry;


_Static_assert(
	LIVELOG_ENTRY_ALWAYS_LAST_ITEM + 1 <= USB_SEND_BUFFER_WIDTH,
	"The USB livelog won't fit into the USB send buffer."
);


static bool livelogEnabled = false;
static uint8_t livelogLength = 3;
static uint8_t livelogContent[LIVELOG_ENTRY_ALWAYS_LAST_ITEM] = {
	
	LIVELOG_ENTRY_FRAME,
	LIVELOG_ENTRY_LOCAL_RSS_INTGR,
	LIVELOG_ENTRY_LOCAL_RSS_FRACT,
};


void livelog_enable(void){
/*
  Enables live logging of measurement results.
*/

	livelogEnabled = true;
}

void livelog_disable(void){
/*
  Disables live logging of measurement results.
*/
	
	livelogEnabled = false;
}

uint8_t livelog_config(uint8_t length, uint8_t * content){
/*
  Configures the live log to have a length of {length} with contents
  specified within {content}.
  
  Return value
  ============
    
    1      if {length} is greater than {LIVELOG_ENTRY_ALWAYS_LAST_ITEM},
    {i}    if {content[i - 2]} is not lower than {LIVELOG_ENTRY_ALWAYS_LAST_ITEM},
    0      else.
    
    The live log configuration gets changed in the last case only.
*/
	
	size_t i;
	
	if(length > LIVELOG_ENTRY_ALWAYS_LAST_ITEM){
		
		return 1;
	}
	else{
	
		for(i = 0; i < length; ++i){
			
			if(content[i] >= LIVELOG_ENTRY_ALWAYS_LAST_ITEM){
				
				return i + 2;
			}
		}
		
		for(i = 0; i < length; ++i){
			
			livelogContent[i] = content[i];
		}
		
		livelogLength = length;
		
		return 0;
	}
}

static uint8_t livelog_ieee_250kbit_get_remote_rss_integer_part(

	volatile uint8_t * radioPtr,
	uint8_t channel
){
/*
  Serves the request for the integer part of the {channel}'s remote RSS in the
  case of an IEEE 250kBit radio packet pointed to by {radioPtr}.
*/
	
	HOOK_UNUSED(radioPtr);
	HOOK_UNUSED(channel);
	
	return 0;
}

static uint8_t livelog_ieee_250kbit_get_remote_rss_fractional_part(

	volatile uint8_t * radioPtr,
	uint8_t channel
){
/*
  Serves the request for the fractional part of the {channel}'s remote RSS in
  the case of an IEEE 250kBit radio packet pointed to by {radioPtr}.
*/
	
	HOOK_UNUSED(radioPtr);
	HOOK_UNUSED(channel);
	
	return 0;
}

static uint8_t livelog_ble_1mbit_get_remote_rss_integer_part(

	volatile uint8_t * radioPtr,
	uint8_t channel
){
/*
  Handles the request for the integer part of the {channel}'s remote RSS in the
  case of an BLE 1MBit radio packet pointed to by {radioPtr}.
*/
	
	if(channel > 0){
		
		return 0;
	}
	else{
		
		return radio_measure_ble_1mbit_get_remote_rss_integer_part(
			radioPtr
		);
	}
}

static uint8_t livelog_ble_1mbit_get_remote_rss_fractional_part(

	volatile uint8_t * radioPtr,
	uint8_t channel
){
/*
  Handles the request for the fractional part of the {channel}'s remote RSS in
  the case of an BLE 1MBit radio packet pointed to by {radioPtr}.
*/
	
	if(channel > 0){
		
		return 0;
	}
	else{
		
		return radio_measure_ble_1mbit_get_remote_rss_fractional_part(
			radioPtr
		);
	}
}

static uint8_t livelog_nrf_1mbit_get_remote_rss_integer_part(

	volatile uint8_t * radioPtr,
	uint8_t channel
){
/*
  Handles the request for the integer part of the {channel}'s remote RSS in the
  case of an NRF 1MBit radio packet pointed to by {radioPtr}.
*/
	
	if(channel > 0){
		
		return 0;
	}
	else{
		
		return radio_measure_nrf_1mbit_get_remote_rss_integer_part(
			radioPtr
		);
	}
}

static uint8_t livelog_nrf_1mbit_get_remote_rss_fractional_part(

	volatile uint8_t * radioPtr,
	uint8_t channel
){
/*
  Handles the request for the fractional part of the {channel}'s remote RSS in
  the case of an NRF 1MBit radio packet pointed to by {radioPtr}.
*/
	
	if(channel > 0){
		
		return 0;
	}
	else{
		
		return radio_measure_nrf_1mbit_get_remote_rss_fractional_part(
			radioPtr
		);
	}
}

void livelog_log(uint8_t frame,
                 int32_t cycle,
                 uint8_t sampleAmount,
                 uint8_t localRSSIntegerPart,
                 uint8_t localRSSFractionalPart,
                 RadioMode radioMode,
                 uint8_t * radioPacketPtr,
                 uint8_t frameCount){
/*
  Generates a live log entry from the given data (if enabled, see
  {livelog_enable()} and tries to send this live log entry via USB.
  
  Argument list
  =============
  
    {frame}                     Frame index of the received measurement packet.
    {cycle}                     Cycle index of the received measurement packet.
    {localRSSIntegerPart}       Integer part of the measured RSS.
    {localRSSFractionalPart}    Fractional part of the measured RSS.
    {radioMode}                 The used RADIO mode during packet reception.
    {radioPacketPtr}            Pointer to received radio packet. This radio
                                packet is used in order to read various (meta)
                                informations about the received packet from it.
    {frameCount}                Number of frames within the current
                                measurement configuration.
*/
	
	uint8_t * data;
	size_t i;
	
	uint8_t (* get_remote_rss_integer_part)(volatile uint8_t * p, uint8_t c);
	uint8_t (* get_remote_rss_fractional_part)(volatile uint8_t * p, uint8_t c);
	uint8_t (* get_remote_battery_level)(volatile uint8_t * p);
	uint8_t (* get_remote_memory_load)(volatile uint8_t * p);
	
	static int32_t activelyLoggedCycle = 0;
	static uint8_t activelyLoggedFrame = 0;
	static size_t missedFrames = 0;
	
	if(livelogEnabled){
	
		if(cycle > activelyLoggedCycle || frame >= activelyLoggedFrame){
			
			/* We have logging permission. */
			missedFrames = 0;
			
			if(usb_vacant_entries() && (data = usb_send_request())){
				
				/* Got buffer memory to write on. */
				
				data[0] = 0xff;
				
				if(livelogLength > LIVELOG_ENTRY_ALWAYS_LAST_ITEM){
					
					livelogLength = LIVELOG_ENTRY_ALWAYS_LAST_ITEM;
				}
				
				/* Dispatch data getters according to {radioMode}. */
				
				switch(radioMode){
					default:
						
						debug_report(DEBUG_MEASURE_INVALID_RADIO_MODE);
					
					/* fall through */
					case RADIO_MODE_BLE_2MBIT:
						
						get_remote_rss_integer_part = &radio_measure_ble_2mbit_get_remote_rss_integer_part;
						get_remote_rss_fractional_part = &radio_measure_ble_2mbit_get_remote_rss_fractional_part;
						get_remote_battery_level = &radio_measure_ble_2mbit_get_remote_battery_level;
						get_remote_memory_load = &radio_measure_ble_2mbit_get_remote_memory_load;
						
					break;
					case RADIO_MODE_BLE_1MBIT:
						
						get_remote_rss_integer_part = &livelog_ble_1mbit_get_remote_rss_integer_part;
						get_remote_rss_fractional_part = &livelog_ble_1mbit_get_remote_rss_fractional_part;
						get_remote_battery_level = &radio_measure_ble_1mbit_get_remote_battery_level;
						get_remote_memory_load = &radio_measure_ble_1mbit_get_remote_memory_load;
						
					break;
					case RADIO_MODE_NRF_2MBIT:
						
						get_remote_rss_integer_part = &radio_measure_nrf_2mbit_get_remote_rss_integer_part;
						get_remote_rss_fractional_part = &radio_measure_nrf_2mbit_get_remote_rss_fractional_part;
						get_remote_battery_level = &radio_measure_nrf_2mbit_get_remote_battery_level;
						get_remote_memory_load = &radio_measure_nrf_2mbit_get_remote_memory_load;
						
					break;
					case RADIO_MODE_NRF_1MBIT:
						
						get_remote_rss_integer_part = &livelog_nrf_1mbit_get_remote_rss_integer_part;
						get_remote_rss_fractional_part = &livelog_nrf_1mbit_get_remote_rss_fractional_part;
						get_remote_battery_level = &radio_measure_nrf_1mbit_get_remote_battery_level;
						get_remote_memory_load = &radio_measure_nrf_1mbit_get_remote_memory_load;
						
					break;
					case RADIO_MODE_IEEE_250KBIT:
						
						get_remote_rss_integer_part = &livelog_ieee_250kbit_get_remote_rss_integer_part;
						get_remote_rss_fractional_part = &livelog_ieee_250kbit_get_remote_rss_fractional_part;
						get_remote_battery_level = &radio_measure_ieee_250kbit_get_remote_battery_level;
						get_remote_memory_load = &radio_measure_ieee_250kbit_get_remote_memory_load;
						
					break;
				}
				
				/* Assemble live log. */
				
				for(i = 0; i < livelogLength; ++i){
				
					switch(livelogContent[i]){
					
						case LIVELOG_ENTRY_FRAME:
						
							data[i + 1] = frame;
						break;
						case LIVELOG_ENTRY_CYCLE_LSB:
						
							data[i + 1] = cycle >>  0;
						break;
						case LIVELOG_ENTRY_CYCLE_ISB:
						
							data[i + 1] = cycle >>  8;
						break;
						case LIVELOG_ENTRY_CYCLE_MSB:
						
							data[i + 1] = cycle >> 16;
						break;
						case LIVELOG_ENTRY_LOCAL_RSS_INTGR:
						
							data[i + 1] = localRSSIntegerPart;
						break;
						case LIVELOG_ENTRY_LOCAL_RSS_FRACT:
						
							data[i + 1] = localRSSFractionalPart;
						break;
						case LIVELOG_ENTRY_REMOTE_RSS_0_INTGR:
						
							data[i + 1] = get_remote_rss_integer_part(
								radioPacketPtr, 0
							);
						break;
						case LIVELOG_ENTRY_REMOTE_RSS_0_FRACT:
						
							data[i + 1] = get_remote_rss_fractional_part(
								radioPacketPtr, 0
							);
						break;
						case LIVELOG_ENTRY_REMOTE_RSS_1_INTGR:
						
							data[i + 1] = get_remote_rss_integer_part(
								radioPacketPtr, 1
							);
						break;
						case LIVELOG_ENTRY_REMOTE_RSS_1_FRACT:
						
							data[i + 1] = get_remote_rss_fractional_part(
								radioPacketPtr, 1
							);
						break;
						case LIVELOG_ENTRY_REMOTE_RSS_2_INTGR:
						
							data[i + 1] = get_remote_rss_integer_part(
								radioPacketPtr, 2
							);
						break;
						case LIVELOG_ENTRY_REMOTE_RSS_2_FRACT:
						
							data[i + 1] = get_remote_rss_fractional_part(
								radioPacketPtr, 2
							);
						break;
						case LIVELOG_ENTRY_REMOTE_RSS_3_INTGR:
						
							data[i + 1] = get_remote_rss_integer_part(
								radioPacketPtr, 3
							);
						break;
						case LIVELOG_ENTRY_REMOTE_RSS_3_FRACT:
						
							data[i + 1] = get_remote_rss_fractional_part(
								radioPacketPtr, 3
							);
						break;
						case LIVELOG_ENTRY_REMOTE_RSS_4_INTGR:
						
							data[i + 1] = get_remote_rss_integer_part(
								radioPacketPtr, 4
							);
						break;
						case LIVELOG_ENTRY_REMOTE_RSS_4_FRACT:
						
							data[i + 1] = get_remote_rss_fractional_part(
								radioPacketPtr, 4
							);
						break;
						case LIVELOG_ENTRY_REMOTE_RSS_5_INTGR:
						
							data[i + 1] = get_remote_rss_integer_part(
								radioPacketPtr, 5
							);
						break;
						case LIVELOG_ENTRY_REMOTE_RSS_5_FRACT:
						
							data[i + 1] = get_remote_rss_fractional_part(
								radioPacketPtr, 5
							);
						break;
						case LIVELOG_ENTRY_LOCAL_MEMORY_LOAD:
						
							data[i + 1] = memory_get_load();
						break;
						case LIVELOG_ENTRY_REMOTE_MEMORY_LOAD:
						
							data[i + 1] = get_remote_memory_load(
								radioPacketPtr
							);
						break;
						case LIVELOG_ENTRY_LOCAL_BATTERY_LEVEL:
						
							data[i + 1] = battery_get_level();
						break;
						case LIVELOG_ENTRY_REMOTE_BATTERY_LEVEL:
						
							data[i + 1] = get_remote_battery_level(
								radioPacketPtr
							);
						break;
						case LIVELOG_ENTRY_LOCAL_RSS_SAMPLE_AMOUNT:
						
							data[i + 1] = sampleAmount;
						break;
						default:
						
							data[i + 1] = 0;
					}
				}
				
				if(!usb_send_finish(data, livelogLength + 1)){
					
					debug_report(DEBUG_LIVELOG_PARALLEL_ACCESS);
				}
			}
			else{
				/* No space on USB output buffer. */
				
				/* Try again within the next cycle. Make sure to restart  */
				/* logging at a random frame. Otherwise, the first frames */
				/* have a higher chance to get logged than other ones.    */
				activelyLoggedCycle = cycle + 1;
				activelyLoggedFrame = random_byte() % frameCount;
			}
		}
		else{
			
			if(++missedFrames == LIVELOG_MISSED_FRAMES){
				
				/* Log control seems not to work properly. */
				missedFrames = 0;
				
				/* Restart with the next cycle. Make sure to restart      */
				/* logging at a random frame. Otherwise, the first frames */
				/* have a higher chance to get logged than other ones.    */
				activelyLoggedCycle = cycle + 1;
				activelyLoggedFrame = random_byte() % frameCount;
			}
		}
	}
}

