/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

#define RSS2I_MODULE "measure"

/*
  Measurement control.
  
  Authors
  =======
  
    Copyright (C) 2018-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
  
    2018-07-31    Initial release.
    2019-05-02    Added {measure_has_active_frame()}.
    2019-06-06    Added {measure_prepare()}. Changed the starting procedure.
    2019-06-14    Fitted to the changes inside the radio module.
    2019-10-04    Fixed the configuration of the syncMap.
    2020-02-12    Fixed the live log behaviour.
    2020-06-09    Use {TIMER1} instead of {TIMER3}.
    2020-06-19    Store measurement results on buffer using DMA.
    2020-07-07    Store livelog data on USB output buffer using DMA.
    2020-07-13    Unlock the memory when stopping the measurement.
    2020-07-31    Outsource livelog functionality into own module.
    2020-08-03    Add CTE configuration.
    2020-08-03    Fit measurement schedule to variable CTE lengths.
    2020-08-05    Set the {CTEANtennaPattern}s when the CTE gets disabled.
    2020-08-07    Forward {CTEType} and {CTEAntennaPattern} to the radio module.
    2020-08-13    Split frame preparation/initilization into two parts.
    2020-08-13    Assert that the IQ sample memory pointer is word-aligned.
    2020-08-13    Assert that each measurement buffer entry is word-aligned.
    2020-08-13    Remove magic numbers and store RADIO packet meta data.
    2020-08-14    Added CTE sampling check functions (API for postproc module).
    2020-08-14    Fix {MEASURE_BUFFER_*_OFFSET} issues and call {postprocess()}.
    2020-08-25    Start measurement schedule directly by the RADIO when resynced
    2020-08-27    Fix #16: Need to stop twice after synchronization loss.
    2021-02-25    Added the BLE 2MBit RADIO mode prefix.
    2021-02-26    Specify the {measureBuffer} radio packet address along with
                  the call to the {radio_measure_*_rx()} functions.
    2021-03-05    Prepare CTE prior to measurement.
    2021-03-19    Added initial support for different RADIO modes.
    2021-03-19    Added support for {LIVELOG_ENTRY_LOCAL_RSS_SAMPLE_AMOUNT}.
    2021-03-19    Added {measure_is_ieee_250kbit_ready()}.
    2021-03-22    Changed behaviour of {measure_set_cte_configuration()}.
    2021-03-23    Added {measure_set_radio_mode()} and changed related stuff.
    2021-03-24    Fix behaviour of {measure_set_cte_configuration()}.
    2021-03-25    Use GPIOTE peripheral for debug toggling.
    2021-03-25    Add some static asserts regarding the {CTE*} types.
    2021-03-27    Use {NVIC_RSS2I_*_PRIORITY} macros.
    2021-03-31    Outsourced storing measurement results to postproc module.
    2021-04-14    Support multi-call {postprocess()}.
    2021-04-19    Renamed {postproc_cte_sample*()} to {postproc_sample*()}.
    2021-04-20    Add support for raw sample storage.
    2021-04-29    Finish flash memory write process when stopping.
    2021-04-29    Support {CTE_2US_7PX1} {CTEType}.
    2021-05-01    Fix LED synchronization for long sample storage frames.
    2021-05-04    Ensure to have enough disruptive frames in sample storage mode
    2021-05-04    Change sync threshold encoding.
    2021-06-15    Remove {MEASURE_SILENT} and related functionality.
  
  Description
  ===========

    This module controls the whole measurement process including the measurement
    schedule management. It has to get initialized using {measure_init()}, which
    automatically enters control mode after initialization.
    You may have a look onto the corresponding config file in order to get an
    overview about the timings of a measurement frame as well as about the used
    nomenclature.
*/

#include <stdint.h>
#include <stddef.h>
#include <stdbool.h>

#include <libopencm3/nrf5/timer.h>
#include <libopencm3/cm3/nvic.h>
#include <libopencm3/nrf5/gpio.h>
#include <libopencm3/nrf5/gpiote.h>

#include <rss2i/radio.h>
#include <rss2i/measure.h>
#include <rss2i/cte.h>
#include <rss2i/livelog.h>
#include <rss2i/led.h>
#include <rss2i/debug.h>
#include <rss2i/memory.h>
#include <rss2i/buffer.h>
#include <rss2i/timer.h>
#include <rss2i/postproc.h>
#include <rss2i/gpiote.h>
#include <rss2i/ppi.h>
#include <rss2i/nvic.h>
#include <rss2i/hook.h>

#include <rss2i/config/led.h>
#include <rss2i/config/measure.h>
#include <rss2i/config/radio.h>
#include <rss2i/config/debug.h>


/* TODO Outsource this into own module under test/ */
/* TODO Use {_Static_assert()} instead of macros. */

#if MEASURE_MAX_FRAME_COUNT > 250 || MEASURE_MAX_FRAME_COUNT < 5
	#error "Invalid setting for MEASURE_MAX_FRAME_COUNT"
#endif


#if MEASURE_X_2MBIT_FRAME_LENGTH != MEASURE_X_2MBIT_PACKET_LENGTH \
                                  + MEASURE_X_2MBIT_GUARD_INTERVAL_LENGTH * 2 \
                                  + MEASURE_X_2MBIT_RADIO_RAMP_UP_TIME \
                                  + MEASURE_X_2MBIT_RADIO_DISABLE_TIME
	#error "2MBit Measurement frame timing is inconsistent."
#endif

#if MEASURE_X_1MBIT_FRAME_LENGTH != MEASURE_X_1MBIT_PACKET_LENGTH \
                                  + MEASURE_X_1MBIT_GUARD_INTERVAL_LENGTH * 2 \
                                  + MEASURE_X_1MBIT_RADIO_RAMP_UP_TIME \
                                  + MEASURE_X_1MBIT_RADIO_DISABLE_TIME
	#error "1MBit Measurement frame timing is inconsistent."
#endif

#if MEASURE_IEEE_250KBIT_FRAME_LENGTH != MEASURE_IEEE_250KBIT_PACKET_LENGTH \
                                       + MEASURE_IEEE_250KBIT_GUARD_INTERVAL_LENGTH * 2 \
                                       + MEASURE_IEEE_250KBIT_RADIO_RAMP_UP_TIME \
                                       + MEASURE_IEEE_250KBIT_RADIO_DISABLE_TIME
	#error "IEEE 250kBit measurement frame timing is inconsistent."
#endif


#if 45 > MEASURE_X_2MBIT_RADIO_RAMP_UP_TIME \
       + MEASURE_X_2MBIT_GUARD_INTERVAL_LENGTH \
       - MEASURE_X_2MBIT_TX_NEGATIVE_DELAY / 16 \
 || 45 > MEASURE_X_2MBIT_RADIO_RAMP_UP_TIME
	#error "Actual 2MBit RADIO ramp up time may be too short."
#endif

#if 45 > MEASURE_X_1MBIT_RADIO_RAMP_UP_TIME \
       + MEASURE_X_1MBIT_GUARD_INTERVAL_LENGTH \
       - MEASURE_X_1MBIT_TX_NEGATIVE_DELAY / 16 \
 || 45 > MEASURE_X_1MBIT_RADIO_RAMP_UP_TIME
	#error "Actual 1MBit RADIO ramp up time may be too short."
#endif

#if 45 > MEASURE_IEEE_250KBIT_RADIO_RAMP_UP_TIME \
       + MEASURE_IEEE_250KBIT_GUARD_INTERVAL_LENGTH \
       - MEASURE_IEEE_250KBIT_TX_NEGATIVE_DELAY / 16 \
 || 45 > MEASURE_IEEE_250KBIT_RADIO_RAMP_UP_TIME
	#error "Actual IEEE 250kBit RADIO ramp up time may be too short."
#endif


#if (2 + 4 + 1 + 1 + RADIO_MEASURE_X_2MBIT_PAYLOAD_LENGTH + 3) * 8 / 2 \
 != MEASURE_X_2MBIT_PACKET_LENGTH
	#error "2MBit payload length doesn't fit into packet length."
#endif

#if (1 + 4 + 1 + 1 + RADIO_MEASURE_X_1MBIT_PAYLOAD_LENGTH + 3) * 8 / 1 \
 != MEASURE_X_1MBIT_PACKET_LENGTH
	#error "1MBit payload length doesn't fit into packet length."
#endif

#if (4 + 1 + 1 + RADIO_MEASURE_IEEE_250KBIT_PAYLOAD_LENGTH) * 8 * 4 \
 != MEASURE_IEEE_250KBIT_PACKET_LENGTH
	#error "IEEE 250kBit payload length doesn't fit into packet length."
#endif


/* Make sure the packets from different RADIO modes */
/* are aligned in compatibility mode.               */
_Static_assert(
	   (     MEASURE_X_2MBIT_FRAME_LENGTH
	       + MEASURE_X_2MBIT_COMPATIBILITY_MODE_DELAY
	    ==   MEASURE_X_1MBIT_FRAME_LENGTH
	       + MEASURE_X_1MBIT_COMPATIBILITY_MODE_DELAY
	   )
	&& (     MEASURE_X_2MBIT_FRAME_LENGTH
	       + MEASURE_X_2MBIT_COMPATIBILITY_MODE_DELAY
	    ==   MEASURE_IEEE_250KBIT_FRAME_LENGTH
	       + MEASURE_IEEE_250KBIT_COMPATIBILITY_MODE_DELAY
	   ),
	"There are compatibility mode incompatibilities."
);

/* Make sure the packets from different RADIO modes */
/* are aligned in raw sample storage mode.          */
_Static_assert(
	   (     MEASURE_X_2MBIT_FRAME_LENGTH
	       + MEASURE_X_2MBIT_SAMPLE_STORAGE_TIME
	    ==   MEASURE_X_1MBIT_FRAME_LENGTH
	       + MEASURE_X_1MBIT_SAMPLE_STORAGE_TIME
	   )
	&& (     MEASURE_X_2MBIT_FRAME_LENGTH
	       + MEASURE_X_2MBIT_SAMPLE_STORAGE_TIME
	    ==   MEASURE_IEEE_250KBIT_FRAME_LENGTH
	       + MEASURE_IEEE_250KBIT_SAMPLE_STORAGE_TIME
	   ),
	"There are raw sample storage mode incompatibilities."
);

/* Make sure the IQ sample address is word-aligned. */
_Static_assert(

	 (RADIO_MEASURE_X_2MBIT_CTE_SAMPLES_OFFSET
	+ MEASURE_BUFFER_RADIO_PACKET_OFFSET) % 4 == 0,
	"IQ sample memory address has to be word-aligned."
);

/* Make sure each buffer entry is word-aligned. */
_Static_assert(

	MEASURE_BUFFER_WIDTH % 4 == 0,
	"Each measurement buffer entry has to be word-aligned."
);

/* Make sure the {MeasureBufferSampleType} fits into one byte. */
_Static_assert(

	MEASURE_BUFFER_SAMPLE_TYPE_ALWAYS_LAST_ITEM <= 256,
	"The MeasureBufferSampleType won't fit into a single byte."
);

/* Make sure the {CTEType} fits into one byte. */
_Static_assert(

	CTE_TYPE_ALWAYS_LAST_ITEM <= 256,
	"The CTEType won't fit into a single byte."
);

/* Make sure the {CTEAntennaPattern} fits into one byte. */
_Static_assert(

	CTE_ANTENNA_PATTERN_DISABLED < 256,
	"The CTEAntennaPattern won't fit into a single byte."
);

/* (Try to) check that {CTE_ANTENNA_PATTERN_DISABLED} is the last item. */
_Static_assert(
	   CTE_ANTENNA_PATTERN_DISABLED > CTE_ANTENNA_PATTERN_ALWAYS_LAST_CONST
	&& CTE_ANTENNA_PATTERN_DISABLED > CTE_ANTENNA_PATTERN_ALWAYS_LAST_SWITCH,
	"Some static asserts may be useless."
);

/* Make sure the {RadioMode} fits into one byte. */
_Static_assert(

	RADIO_MODE_ALWAYS_LAST_ITEM <= 256,
	"The RadioMode won't fit into a single byte."
);

/* Check for consistency. */
_Static_assert(
	(1 << (DEBUG_MEASUREMENT_FRAME_PIN)) == DEBUG_MEASUREMENT_FRAME_GPIO,
	"DEBUG_MEASUREMENT_FRAME inconsistency."
);

/* Check that all IQ samples fit into memory. */
_Static_assert(
	   CTE_TYPE_ALWAYS_LAST_ITEM == 3
	&& 16 + 6*16 + 7*16 <= CTE_MAX_NUMBER_OF_SAMPLES
	&& 32 + 6*16 + 7*16 <= CTE_MAX_NUMBER_OF_SAMPLES,
	"The total amount of IQ samples might not fit into memory."
);

/* Ensure that there are enough disruptive frames in sample storage mode. */
_Static_assert(
	MEASURE_MIN_DISRUPT_FRAMES_SAMPLE_STORAGE >= MEASURE_MAX_FRAME_COUNT,
	"There may be less disruptive frames than frames per cycle."
);




typedef enum {
	
	MEASURE_DISABLED,
	MEASURE_TX,
	MEASURE_RX,
	MEASURE_LISTEN,
	
} MeasureState;

typedef enum {
	
	MEASURE_ACTIVE,
	MEASURE_DISRUPTIVE,
	
} MeasureTXMode;

typedef enum {
	
	MEASURE_POOR_SYNC,
	MEASURE_OUT_OF_SYNC,
	MEASURE_SYNCHRONIZED,
	
} MeasureSyncState;

typedef enum {

	MEASURE_JUMP_ON,
	MEASURE_START_UP,
	
} MeasureStartType;

typedef enum {
	
	MEASURE_SAMPLE_STORAGE_NEVER_SHORT_FRAMES,
	MEASURE_SAMPLE_STORAGE_NEVER_LARGE_FRAMES,
	MEASURE_SAMPLE_STORAGE_ALWAYS_LARGE_FRAMES

} MeasureSampleStorage;

_Static_assert(
	MEASURE_SAMPLE_STORAGE_NEVER_SHORT_FRAMES == 0,
	"MEASURE_SAMPLE_STORAGE_NEVER_SHORT_FRAMES should cast to false."
);


static volatile MeasureTXMode measureTXMode = MEASURE_ACTIVE;

static volatile uint8_t measureFrameIndex = 255;
static volatile int32_t measureCycleIndex = -1;

static volatile bool measureRunning = false;

static Buffer measureBuffer;
/* Stores the currently written {measureBuffer} entry. */
static volatile uint8_t * measureBufferRowAddr;
/* Stores the exact address of the currently written RADIO    */
/* packet within the currently written {measureBuffer} entry, */
/* i.e usually {measureBufferRowAddr} + some offset.          */
static volatile uint8_t * measureBufferRadioAddr;

/* Signal a measurement stop. */
static volatile bool measureWasStopped = false;

static volatile MeasureState measureState = MEASURE_DISABLED;
static MeasureSyncState measureSyncState = MEASURE_SYNCHRONIZED;

/* Number of frames an out-of-sync state is ignored. */
static volatile uint16_t measureForgivingFrames;
static volatile uint16_t measurePrecomputedForgivingFrames;

/* Frame length without CTE. */
static volatile uint16_t measurePrecomputedFrameLength;

static int32_t measureSyncCycle[MEASURE_MAX_FRAME_COUNT];
static int32_t measureSyncLastCheckedCycle = -1;
static uint8_t measureSyncMap[MEASURE_MAX_FRAME_COUNT];
static size_t measureSyncVirtualDeviceCount;
static uint8_t measureMissedCyclesOutOfSyncThresh[MEASURE_MAX_FRAME_COUNT];
static uint8_t measureMissedCyclesPoorSyncThresh[MEASURE_MAX_FRAME_COUNT];

static int8_t measureFrameFrequency[MEASURE_MAX_FRAME_COUNT];
static uint8_t measureFrameCount;
static int32_t measureCycleCount;

static CTEType measureCTEType = CTE_TYPE_NONE;
static bool measureAnyCTESampling = false;
static CTEAntennaPattern measureCTEAntennaPattern[MEASURE_MAX_FRAME_COUNT];
static uint8_t measureCTELength = 0;

static RadioMode measureRadioMode = RADIO_MODE_BLE_2MBIT;
static bool measureCompatibilityMode = false;

/* Whether to reduce the sampling rate, i.e. the number of frames per */
/* second, by enlarging the frame length with the purpose to be able  */
/* to store an immense number of raw samplings on the flash memory.   */
static MeasureSampleStorage measureSampleStorage
	= MEASURE_SAMPLE_STORAGE_NEVER_SHORT_FRAMES;
/* The raw sample type of each frame. */
static MeasureBufferSampleType measureSampleType[MEASURE_MAX_FRAME_COUNT];
static bool measureAnyRawSampling = false;

static uint8_t measureRemoteRSSInputMap[MEASURE_MAX_FRAME_COUNT];
static uint8_t measureRemoteRSSOutputCount;
static volatile uint16_t measureRemoteRSS[MEASURE_MAX_FRAME_COUNT];
static uint8_t measureRemoteRSSOutputMap[MEASURE_MAX_FRAME_COUNT];


bool measure_is_ieee_250kbit_ready(void){
/*
  Whether this device is ready to perform measurements according to the IEEE
  802.15.4 ({RADIO_MODE_IEEE_250KBIT}) standard or not.
*/
	
	#if REVISION >= REVISION_F
		return true;
	#else
		return false;
	#endif
}

uint8_t measure_set_remote_rss_output_map(uint8_t outputCount,
                                          uint8_t * outputMap){
/*
  Configures the remote sharing of RSS values.
  
  Some Details
  ============
    
    When this device sends a measurement packet within frame {f} it shares those
    previously measured RSS values stored at the virtual memory locations
    {measureRemoteRSSOutputMap[f]}, {measureRemoteRSSOutputMap[f]} + 1, ...,
    {measureRemoteRSSOutputMap[f] + measureRemoteRSSOutputCount - 1}.
  
  Argument List
  =============
    
    {outputCount}   Number of RSS values that should be shared with other
                    devices within each measurement packet in the range [0, 6].
    {outputMap}     Array with a length of {MEASURE_MAX_FRAME_COUNT} entries
                    representing the desired RSS output map (see above). The
                    entries should be  in the range
                    [0, {MEASURE_MAX_FRAME_COUNT - outputCount}].
  
  Return Value
  ============
    
    Returns {1} if {outputCount} is out-of-range. Returns {i + 2} if the {i}th
    entry of {outputMap} is out-of-range. Returns {0} else. The remote RSS
    related settings get changed in that last case only.
*/

	uint8_t i;
	
	if(outputCount > 6){
		
		return 1;
	}
	
	for(i = 0; i < MEASURE_MAX_FRAME_COUNT; ++i){
		
		if(outputMap[i] > MEASURE_MAX_FRAME_COUNT - outputCount){
			
			return i + 2;
		}
	}
	
	measureRemoteRSSOutputCount = outputCount;
	
	for(i = 0; i < MEASURE_MAX_FRAME_COUNT; ++i){
		
		measureRemoteRSSOutputMap[i] = outputMap[i];
	}
	
	return 0;
}

uint8_t measure_set_remote_rss_input_map(uint8_t * inputMap){
/*
  Sets the {measureRemoteRSSInputMap} i.e. the function which maps a
  {measureFrameIndex} to a virtual memory location where an RSS value (measured
  during that frame) may be stored.
  
  Argument list
  =============
  
    {inputMap}   Array with a length of {MEASURE_MAX_FRAME_COUNT} entries
                 representing the desired remote RSS input map (see above).
                 The entries should be in the range
                 [0, {MEASURE_MAX_FRAME_COUNT - 1}].
                           
  Return value
  ============
    
    Returns {i + 1} if there is an out-of-range error within the {i}th entry.
    Returns {0} else. The RSS input map gets changed in that last case only.
*/
	
	uint8_t i;
	
	for(i = 0; i < MEASURE_MAX_FRAME_COUNT; ++i){
		
		if(inputMap[i] >= MEASURE_MAX_FRAME_COUNT){
			
			return i + 1;
		}
	}
	
	for(i = 0; i < MEASURE_MAX_FRAME_COUNT; ++i){
		
		measureRemoteRSSInputMap[i] = inputMap[i];
	}
	
	return 0;
}

void measure_set_poor_sync_thresh(uint8_t * poorSyncThresh){
/*
  Sets the {measureMissedCyclesPoorSyncThresh}olds.
  
  Some Details
  ============
  
    If there are less than {j + 1} synchronization events (i.e. received
    measurement packets) from different (virtual) devices within the last
    {value[j]} cycles, the device is muted temporarily (i.e. is not _sending_
    measurement packets anymore). The value of {value[j]} is equal to
                  
      poorSyncThresh[j]
    
    if {poorSyncThresh[j] <= 100} or
      
      100 + (poorSyncThresh[j] - 100) * 10
    
    if not.
  
  Argument list
  =============
  
    {poorSyncThresh}          Values to be set. Please note that independent
                              of the current {measureSyncVirtualDeviceCount}
                              all {MEASURE_MAX_FRAME_COUNT} entries are
                              processed.
*/
	
	uint8_t i;
	
	for(i = 0; i < MEASURE_MAX_FRAME_COUNT; ++i){
		
		measureMissedCyclesPoorSyncThresh[i] = poorSyncThresh[i];
	}
}

void measure_set_out_of_sync_thresh(uint8_t * outOfSyncThresh){
/*
  Sets the {measureMissedCyclesOutOfSyncThresh}olds.
  
  Some Details
  ============
  
    If there are less than {j + 1} synchronization events (i.e. received
    measurement packtes) from different (virtual) devices within the last
    {value[j]} cycles, the device leaves the measurement schedule and tries to
    resync with the beginning of the next measurement cycle. The value of
    {value[j]} is equal to
                  
      outOfSyncThresh[j]
    
    if {outOfSyncThresh[j] <= 100} or
      
      100 + (outOfSyncThresh[j] - 100) * 10
    
    if not.
  
  Argument list
  =============
  
    {outOfSyncThresh}         Values to be set. Please note that independent
                              of the current {measureSyncVirtualDeviceCount}
                              all {MEASURE_MAX_FRAME_COUNT} entries are
                              processed.
*/
	
	uint8_t i;
	
	for(i = 0; i < MEASURE_MAX_FRAME_COUNT; ++i){
		
		measureMissedCyclesOutOfSyncThresh[i] = outOfSyncThresh[i];
	}
}

uint8_t measure_set_sync_map(uint8_t virtualDeviceCount, uint8_t * syncMap){
/*
  Sets the synchronization map i.e. the function which maps
  {measureFrameIndex}es to a list of (virtual) devices.
  
  Some Details
  ============
  
    This is usually used to assign a received measurement packet to a physical
    device but may be further used to group a bunch of physical devices into a
    virtual device. The information about received packets per virtual device
    then gets used within the out-of-sync detection algorithm.
  
  Argument list
  =============
  
    {virtualDeviceCount}   Number of (virtual) devices that should be taken into
                           account when checking for synchronization in the
                           range [1, {MEASURE_MAX_FRAME_COUNT}].
    {syncMap}              Array with a length of {MEASURE_MAX_FRAME_COUNT}
                           entries representing the synchronization map.
                           The value {d} of the {i}th entry specifies the
                           virtual device index the {i}th frame should get
                           mapped to. Hence {d} should be in the range
                           [0, {virtualDeviceCount - 1}]. Please note that
                           independent of the current measurement configuration
                           (specifically the current value of
                           {measureFrameCount}) all entries of {syncMap}
                           are processed.
                           
  Return value
  ============
    
    Returns {1} if {virtualDeviceCount} is out-of-range. Returns {i} if
    {syncMap[i - 2]} is out-of-range. Else the requested change gets applied and
    a {0} gets returned.
*/
	
	uint8_t i;
	
	if(virtualDeviceCount < 1 || virtualDeviceCount > MEASURE_MAX_FRAME_COUNT){
		
		return 1;
	}
	
	for(i = 0; i < MEASURE_MAX_FRAME_COUNT; ++i){
		
		if(syncMap[i] >= virtualDeviceCount){
			
			return i + 2;
		}
	}
	
	measureSyncVirtualDeviceCount = virtualDeviceCount;
	
	for(i = 0; i < MEASURE_MAX_FRAME_COUNT; ++i){
		
		measureSyncMap[i] = syncMap[i];
	}
	
	return 0;
}

uint8_t measure_set_configuration(uint8_t frameCount,
                                  uint32_t cycleCount,
                                  uint8_t * frameFrequency){
/*
  Sets the measurement configuration (i.e. alters the variables
  {measureFrameCount}, {measureCycleCount} and {measureFrameFrequency}).
  
  Argument list
  =============
  
    {frameCount}           Number of frames the new measurement configuration
                           should consist of in the range
                           [1, {MEASURE_MAX_FRAME_COUNT}]
    {cycleCount}           Number of cycles the new measurement configuration
                           should consist of in the range [1, 2^24]
    {frameFrequency}       Array with a minimal length of {frameCount}
                           specifiying the frequency of the corresponding frame.
                           A value {f} in the range [2, 80] induces the device
                           to listen to other devices at a center frequency of
                           {2400 + f} MHz while a value {f} in the range
                           [102, 180] induces the device to send measurement
                           packets at a center frequency of {2400 + f - 100} MHz
                           within the corresponding frame.
  
  Return value
  ============
    
    Returns {1} if {frameCount} is out-of-range. Returns {2} if {cylceCount}
    is out-of-range. Returns {i} if the {i-3}th frequency is not within the
    documented ranges. Else the requested change gets applied and a {0} gets
    returned.
*/
	uint8_t i;
	uint8_t fF;
	
	if(frameCount < 1 || frameCount > MEASURE_MAX_FRAME_COUNT){ return 1; }
	if(cycleCount < 1 || cycleCount > ((uint32_t) 1 << 24)){ return 2; }
	
	for(i = 0; i < frameCount; ++i){
		
		fF = frameFrequency[i];
		
		if(fF < 2 || (fF > 81 && fF < 102) || fF > 181){
			
			return i + 3;
		}
	}
	
	measure_set_sample_configuration(0, NULL, NULL);
	measure_set_cte_configuration(CTE_TYPE_NONE, NULL, NULL);
	measure_set_radio_mode(RADIO_MODE_BLE_2MBIT, false);
	
	measureFrameCount = frameCount;
	measureCycleCount = cycleCount;
	
	for(i = 0; i < frameCount; ++i){
		
		fF = frameFrequency[i];
		
		measureFrameFrequency[i] = (fF > 100) ? fF - 100 : (int8_t) fF * -1;
	}
	
	return 0;
}

bool measure_is_rx_frame(uint8_t frame){
/*
  Whether the specified {frame} is a receiving frame in the current
  configuration or not.
*/
	
	if(frame < measureFrameCount && measureFrameFrequency[frame] < 0){
		
		return true;
	}
	else{
		
		return false;
	}
}

static bool measure_could_be_cte_sampling_frame(uint8_t frame){
/*
  Whether the specified frame is configured to be able to perform a CTE
  sampling. (This includes an out-of-range check).
*/
	
	if(measure_is_rx_frame(frame)
	&& cte_is_sampling_configuration(
	   	measureCTEAntennaPattern[frame],
	   	measureCTEType
	   )
	){
		
		return true;
	}
	else{
		
		return false;
	}
}

bool measure_is_cte_sampling_frame(uint8_t frame){
/*
  Whether a CTE sampling takes place during the specified {frame}
*/
	
	if(measure_could_be_cte_sampling_frame(frame)
	&& measureSampleType[frame] == MEASURE_BUFFER_SAMPLE_TYPE_IQ
	){
		
		return true;
	}
	else{
		
		return false;
	}
}

bool measure_is_raw_sampling_frame(uint8_t frame){
/*
  Whether a sampling of raw RSS samples takes place during the specified {frame}
  (This includes an out-of-range check).
*/
	
	/* TODO once support for raw RSS samples is added. */
	HOOK_UNUSED(frame);
	return false;
}

bool measure_is_sampling_frame(uint8_t frame){
/*
  Whether any sampling takes place during the specified {frame}
  (This includes an out-of-range check).
*/
	
	if(measure_is_cte_sampling_frame(frame)
	|| measure_is_raw_sampling_frame(frame)){
		
		return true;
	}
	else{
		
		return false;
	}
}

static void measure_set_x_2mbit_timings(void){
/*
  Applies the 2MBit schedule timings to the measurement schedule manager.
  
  Refer to 'rss2i/config/measure.h' in order to get an overview about the
  schedule timings.
*/
	
	uint16_t compatibilityDelay;
	
	if(measureSampleStorage){
		
		/* In sample storage mode, every frame has the same length regardless */
		/* of the CTE type and radio mode.                                    */
		compatibilityDelay = MEASURE_X_2MBIT_SAMPLE_STORAGE_TIME
		                   - measureCTELength;
	}
	else if(measureCompatibilityMode){
		
		compatibilityDelay = MEASURE_X_2MBIT_COMPATIBILITY_MODE_DELAY;
	}
	else{
		
		compatibilityDelay = 0;
	}
	
	/* Frame's start: Disable the RADIO peripheral. */
	TIMER_CC(TIMER1)[0] = 16 * (measureCTELength
	                          + compatibilityDelay
	                          + MEASURE_X_2MBIT_GUARD_INTERVAL_LENGTH)
	                    - MEASURE_X_2MBIT_SYNC_DELAY;
	
	/* Prepare the RADIO peripheral. */
	TIMER_CC(TIMER1)[1] = 16 * (measureCTELength
	                          + compatibilityDelay
	                          + MEASURE_X_2MBIT_GUARD_INTERVAL_LENGTH
	                          + MEASURE_X_2MBIT_RADIO_DISABLE_TIME)
	                    - MEASURE_X_2MBIT_SYNC_DELAY;
	
	 /* Packet's start: Send packet (if frame is an active one) */
	TIMER_CC(TIMER1)[2] = 16 * (measureCTELength
	                          + compatibilityDelay
	                          + MEASURE_X_2MBIT_GUARD_INTERVAL_LENGTH
	                          + MEASURE_X_2MBIT_RADIO_DISABLE_TIME
	                          + MEASURE_X_2MBIT_RADIO_RAMP_UP_TIME
	                          + MEASURE_X_2MBIT_GUARD_INTERVAL_LENGTH)
	                    - MEASURE_X_2MBIT_SYNC_DELAY
	                    - MEASURE_X_2MBIT_TX_NEGATIVE_DELAY;
	
	/* Packet's end: Reset schedule.                 */
	TIMER_CC(TIMER1)[3] = 16 * (measureCTELength
	                          + compatibilityDelay
	                          + MEASURE_X_2MBIT_GUARD_INTERVAL_LENGTH
	                          + MEASURE_X_2MBIT_RADIO_DISABLE_TIME
	                          + MEASURE_X_2MBIT_RADIO_RAMP_UP_TIME
	                          + MEASURE_X_2MBIT_GUARD_INTERVAL_LENGTH
	                          + MEASURE_X_2MBIT_PACKET_LENGTH)
	                    - MEASURE_X_2MBIT_SYNC_DELAY
	                    + MEASURE_X_2MBIT_SYNC_DELAY;
}

static void measure_set_ble_2mbit_timings(void){
/*
  Applies the BLE 2MBit schedule timings to the measurement schedule manager.
*/
	
	measure_set_x_2mbit_timings();
}

static void measure_set_nrf_2mbit_timings(void){
/*
  Applies the NRF 2MBit schedule timings to the measurement schedule manager.
*/
	
	measureCTELength = 0;
	measure_set_x_2mbit_timings();
}

static void measure_set_x_1mbit_timings(void){
/*
  Applies the 1MBit schedule timings to the measurement schedule manager.
  
  Refer to 'rss2i/config/measure.h' in order to get an overview about the
  schedule timings.
*/
	
	uint16_t compatibilityDelay;
	
	if(measureSampleStorage){
		
		/* In sample storage mode, every frame has the same length regardless */
		/* of the CTE type and radio mode.                                    */
		compatibilityDelay = MEASURE_X_1MBIT_SAMPLE_STORAGE_TIME
		                   - measureCTELength;
	}
	else if(measureCompatibilityMode){
		
		compatibilityDelay = MEASURE_X_1MBIT_COMPATIBILITY_MODE_DELAY;
	}
	else{
		
		compatibilityDelay = 0;
	}
	
	/* Frame's start: Disable the RADIO peripheral. */
	TIMER_CC(TIMER1)[0] = 16 * (measureCTELength
	                          + compatibilityDelay
	                          + MEASURE_X_1MBIT_GUARD_INTERVAL_LENGTH)
	                    - MEASURE_X_1MBIT_SYNC_DELAY;
	
	/* Prepare the RADIO peripheral. */
	TIMER_CC(TIMER1)[1] = 16 * (measureCTELength
	                          + compatibilityDelay
	                          + MEASURE_X_1MBIT_GUARD_INTERVAL_LENGTH
	                          + MEASURE_X_1MBIT_RADIO_DISABLE_TIME)
	                    - MEASURE_X_1MBIT_SYNC_DELAY;
	 /* Packet's start: Send packet (if frame is an active one) */
	TIMER_CC(TIMER1)[2] = 16 * (measureCTELength
	                          + compatibilityDelay
	                          + MEASURE_X_1MBIT_GUARD_INTERVAL_LENGTH
	                          + MEASURE_X_1MBIT_RADIO_DISABLE_TIME
	                          + MEASURE_X_1MBIT_RADIO_RAMP_UP_TIME
	                          + MEASURE_X_1MBIT_GUARD_INTERVAL_LENGTH)
	                    - MEASURE_X_1MBIT_SYNC_DELAY
	                    - MEASURE_X_1MBIT_TX_NEGATIVE_DELAY;
	
	/* Packet's end: Reset schedule.                 */
	TIMER_CC(TIMER1)[3] = 16 * (measureCTELength
	                          + compatibilityDelay
	                          + MEASURE_X_1MBIT_GUARD_INTERVAL_LENGTH
	                          + MEASURE_X_1MBIT_RADIO_DISABLE_TIME
	                          + MEASURE_X_1MBIT_RADIO_RAMP_UP_TIME
	                          + MEASURE_X_1MBIT_GUARD_INTERVAL_LENGTH
	                          + MEASURE_X_1MBIT_PACKET_LENGTH)
	                    - MEASURE_X_1MBIT_SYNC_DELAY
	                    + MEASURE_X_1MBIT_SYNC_DELAY;
}

static void measure_set_ble_1mbit_timings(void){
/*
  Applies the BLE 1MBit schedule timings to the measurement schedule manager.
*/
	
	measure_set_x_1mbit_timings();
}

static void measure_set_nrf_1mbit_timings(void){
/*
  Applies the NRF 1MBit schedule timings to the measurement schedule manager.
*/
	
	measureCTELength = 0;
	measure_set_x_1mbit_timings();
}

static void measure_set_ieee_250kbit_timings(void){
/*
  Applies the IEEE 250kBit timings within the measurement schedule manager.
  Refer to 'rss2i/config/measure.h' in order to get an overview about the
  schedule timings.
*/
	
	uint16_t compatibilityDelay;
	
	measureCTELength = 0;
	
	if(measureSampleStorage){
		
		compatibilityDelay = MEASURE_IEEE_250KBIT_SAMPLE_STORAGE_TIME;
	}
	else if(measureCompatibilityMode){
		
		compatibilityDelay = MEASURE_IEEE_250KBIT_COMPATIBILITY_MODE_DELAY;
	}
	else{
		
		compatibilityDelay = 0;
	}
	
	/* Frame's start: Disable the RADIO peripheral. */
	TIMER_CC(TIMER1)[0] = 16 * (compatibilityDelay
	                          + MEASURE_IEEE_250KBIT_GUARD_INTERVAL_LENGTH)
	                    - MEASURE_IEEE_250KBIT_SYNC_DELAY;
	
	/* Prepare the RADIO peripheral. */
	TIMER_CC(TIMER1)[1] = 16 * (compatibilityDelay
	                          + MEASURE_IEEE_250KBIT_GUARD_INTERVAL_LENGTH
	                          + MEASURE_IEEE_250KBIT_RADIO_DISABLE_TIME)
	                    - MEASURE_IEEE_250KBIT_SYNC_DELAY;
	
	 /* Packet's start: Send packet (if frame is an active one) */
	TIMER_CC(TIMER1)[2] = 16 * (compatibilityDelay
	                          + MEASURE_IEEE_250KBIT_GUARD_INTERVAL_LENGTH
	                          + MEASURE_IEEE_250KBIT_RADIO_DISABLE_TIME
	                          + MEASURE_IEEE_250KBIT_RADIO_RAMP_UP_TIME
	                          + MEASURE_IEEE_250KBIT_GUARD_INTERVAL_LENGTH)
	                    - MEASURE_IEEE_250KBIT_SYNC_DELAY
	                    - MEASURE_IEEE_250KBIT_TX_NEGATIVE_DELAY;
	
	/* Packet's end: Reset schedule.                 */
	TIMER_CC(TIMER1)[3] = 16 * (compatibilityDelay
	                          + MEASURE_IEEE_250KBIT_GUARD_INTERVAL_LENGTH
	                          + MEASURE_IEEE_250KBIT_RADIO_DISABLE_TIME
	                          + MEASURE_IEEE_250KBIT_RADIO_RAMP_UP_TIME
	                          + MEASURE_IEEE_250KBIT_GUARD_INTERVAL_LENGTH
	                          + MEASURE_IEEE_250KBIT_PACKET_LENGTH)
	                    - MEASURE_IEEE_250KBIT_SYNC_DELAY
	                    + MEASURE_IEEE_250KBIT_SYNC_DELAY;
}

static void measure_set_timings(void){
/*
  Sets the timings of the measurement scheduler according to the current
  {measureRadioMode}, {measureCompatibilityMode}, {measureSampleStorage} and
  {measureCTELength} setting.
*/
	
	switch(measureRadioMode){
		case RADIO_MODE_BLE_2MBIT:
			measure_set_ble_2mbit_timings();
		break;
		case RADIO_MODE_BLE_1MBIT:
			measure_set_ble_1mbit_timings();
		break;
		case RADIO_MODE_NRF_2MBIT:
			measure_set_nrf_2mbit_timings();
		break;
		case RADIO_MODE_NRF_1MBIT:
			measure_set_nrf_1mbit_timings();
		break;
		case RADIO_MODE_IEEE_250KBIT:
			measure_set_ieee_250kbit_timings();
		break;
		default:
			measure_set_ble_2mbit_timings();
			debug_report(DEBUG_MEASURE_INVALID_RADIO_MODE);
	}
}

static void measure_precompute_forgiving_frames(void){
/*
  Precomputes the number of forgiving frames. This solely depends on the current
  {measureRadioMode}, {measureCompatibilityMode} and {measureSampleStorage}
  setting.
*/
	
	if(measureSampleStorage){
		
		measurePrecomputedForgivingFrames
			= MEASURE_MIN_DISRUPT_FRAMES_SAMPLE_STORAGE;
	}
	else{
	
		switch(measureRadioMode){
		
			default:
			
				debug_report(DEBUG_MEASURE_INVALID_RADIO_MODE);
				
			/* fall through */
			case RADIO_MODE_BLE_2MBIT:
			case RADIO_MODE_NRF_2MBIT:
			
				measurePrecomputedForgivingFrames
					= (measureCompatibilityMode)
					? MEASURE_MIN_DISRUPT_FRAMES_X_2MBIT_COMPATIBILITY_MODE
					: MEASURE_MIN_DISRUPT_FRAMES_X_2MBIT;
			
			break;
			case RADIO_MODE_BLE_1MBIT:
			case RADIO_MODE_NRF_1MBIT:
			
				measurePrecomputedForgivingFrames
					= (measureCompatibilityMode)
					? MEASURE_MIN_DISRUPT_FRAMES_X_1MBIT_COMPATIBILITY_MODE
					: MEASURE_MIN_DISRUPT_FRAMES_X_1MBIT;
				
			break;
			case RADIO_MODE_IEEE_250KBIT:
				
				measurePrecomputedForgivingFrames
					= (measureCompatibilityMode)
					? MEASURE_MIN_DISRUPT_FRAMES_IEEE_250KBIT_COMPATIBILITY_MODE
					: MEASURE_MIN_DISRUPT_FRAMES_IEEE_250KBIT;
				
			break;
		}
	}
}

static void measure_precompute_frame_length(void){
/*
  Precomputes the frame length (without CTE). This solely depends on the current
  {measureRadioMode}, {measureCompatibilityMode}, {measureSampleStorage} and
  {measureCTELength} setting.
*/
	
	uint8_t compatibilityDelayCoeff;
	uint8_t sampleStorageTimeCoeff;
	
	if(measureSampleStorage){
		
		sampleStorageTimeCoeff = 1;
		compatibilityDelayCoeff = 0;
	}
	else if(measureCompatibilityMode){
		
		sampleStorageTimeCoeff = 0;
		compatibilityDelayCoeff = 1;
	}
	else{
	
		sampleStorageTimeCoeff = 0;
		compatibilityDelayCoeff = 0;
	}
		
	switch(measureRadioMode){
	
		default:
		
			debug_report(DEBUG_MEASURE_INVALID_RADIO_MODE);
			
		/* fall through */	
		case RADIO_MODE_BLE_2MBIT:
		case RADIO_MODE_NRF_2MBIT:
			
			measurePrecomputedFrameLength = MEASURE_X_2MBIT_FRAME_LENGTH
			                              + MEASURE_X_2MBIT_COMPATIBILITY_MODE_DELAY
			                              * compatibilityDelayCoeff
			                              + MEASURE_X_2MBIT_SAMPLE_STORAGE_TIME
			                              * sampleStorageTimeCoeff
			                              - measureCTELength
			                              * sampleStorageTimeCoeff;
		
		break;
		case RADIO_MODE_BLE_1MBIT:
		case RADIO_MODE_NRF_1MBIT:
		
			measurePrecomputedFrameLength = MEASURE_X_1MBIT_FRAME_LENGTH
			                              + MEASURE_X_1MBIT_COMPATIBILITY_MODE_DELAY
			                              * compatibilityDelayCoeff
			                              + MEASURE_X_1MBIT_SAMPLE_STORAGE_TIME
			                              * sampleStorageTimeCoeff
			                              - measureCTELength
			                              * sampleStorageTimeCoeff;
			
		break;
		case RADIO_MODE_IEEE_250KBIT:
			
			measurePrecomputedFrameLength = MEASURE_IEEE_250KBIT_FRAME_LENGTH
			                              + MEASURE_IEEE_250KBIT_COMPATIBILITY_MODE_DELAY
			                              * compatibilityDelayCoeff
			                              + MEASURE_IEEE_250KBIT_SAMPLE_STORAGE_TIME
			                              * sampleStorageTimeCoeff
			                              - measureCTELength
			                              * sampleStorageTimeCoeff;
			
		break;
	}
}

static void measure_precompute_any_cte_sampling(void){
/*
  Precomputes whether there is any CTE sampling in the current configuration.
*/
	
	uint8_t i;
	
	for(i = 0; i < measureFrameCount && i < MEASURE_MAX_FRAME_COUNT; ++i){
		
		if(measure_is_cte_sampling_frame(i)){
			
			measureAnyCTESampling = true;
		}
	}
}

uint8_t measure_set_cte_configuration(uint8_t cteType,
                                      uint8_t * cteConf,
                                      uint8_t * failPtr){
/*
  Sets the Constant Tone Extension (CTE) configuration (i.e. alters the
  variables {measureCTELength}, {measureCTEAntennaPattern}, {measureCTEType} and
  {measureAnyCTESampling}).
  Note, that the specified {cteConf}iguration must comply with the current
  measurement configuration. It is hence recommended to call
  {measure_set_configuration()} prior to the call of this function.
  
  Argument list
  =============
  
    {cteType}    Type of CTE. This is equal for every frame of the current
                 measurement configuration. The given type has to be a valid
                 {CTEType} as defined by the cte module. An exception concerns
                 the value {0}: Specifying {0} will disable the CTE feature for
                 the current measurement configuration.
    {cteConf}    Array of antenna patterns that shall be used within the
                 corresponding frames. An entry has to be a valid
                 {CTEAntennaPattern} as defined by the cte module and must be
                 supported by the actual RSS2I revision and type. The specified
                 antenna patterns have to be compatible with the specified
                 {cteType}. Exceptions to this concern the values {0} and {255}:
                 {0} shall be specified if the corresponding frame is not part
                 of the current measurement configuration. The value {255}
                 indicates that there shall be no CTE in the corresponding frame
                 at all.
                 If {cteType} is {0}, the memory pointed to by {cteConf} is not
                 read from and hence may be {NULL}.
    {failPtr}    Where to write the index of that frame, for which the first
                 error occured during parsing of the given configuration.
                 The memory pointed to by {failPtr} is not modified in success
                 case.
  
  Return value
  ============
    
    Returns {0} in success case. In error case, i.e. if the specified
    configuration is not a valid one, the CTE configuration is not changed, the
    first faulty frame index is reported within {failPtr} and one of the
    follownig error codes is returned:
      
      {1}    Invalid CTE type ({failPtr} is not modified).
      {2}    Configuration does not comply with the current measurement
             configuraiton.
      {3}    CTE Antenna Pattern not supported or invalid or incompatible
             with the stated CTE type.
      {4}    The current {measureRadioMode} does not support CTEs.
*/
	
	uint8_t i;
	
	/* Reset sample configuration. */
	measure_set_sample_configuration(0, NULL, NULL);
	
	/* Reset CTE preparation. */
	cte_unprepare();
	cte_prepare(CTE_TYPE_NONE, CTE_ANTENNA_PATTERN_DISABLED);
	
	if(cteType == 0){
		
		measureCTEType = CTE_TYPE_NONE;
		measureCTELength = 0;
		measure_set_timings();
		measure_precompute_frame_length();
		measureAnyCTESampling = false;
		
		for(i = 0; i < MEASURE_MAX_FRAME_COUNT; ++i){
		
			measureCTEAntennaPattern[i] = CTE_ANTENNA_PATTERN_DISABLED;
		}
		
		return 0;
	}
	else{
		
		if((measureRadioMode != RADIO_MODE_BLE_2MBIT
		 && measureRadioMode != RADIO_MODE_BLE_1MBIT)
		|| measureCompatibilityMode != false){
			
			/* CTE is only supported for BLE normal mode. */
			return 4;
		}
		
		if(!cte_type_is_valid(cteType)){
			
			return 1;
		}
		
		for(i = 0; i < MEASURE_MAX_FRAME_COUNT; ++i){
			
			if((i < measureFrameCount && cteConf[i] == 0)
			|| (i >= measureFrameCount && cteConf[i] != 0)){
				
				*failPtr = i;
				return 2;
			}
			
			if(i < measureFrameCount){
			
				if(!cte_antenna_pattern_is_valid(cteConf[i], cteType)
				&& cteConf[i] != CTE_ANTENNA_PATTERN_DISABLED){
					
					*failPtr = i;
					return 3;
				}
			}
		}
		
		measureAnyCTESampling = false;
		measureCTEType = cteType;
		measureCTELength = cte_length(cteType);
		measure_set_timings();
		measure_precompute_frame_length();
		
		for(i = 0; i < measureFrameCount && i < MEASURE_MAX_FRAME_COUNT; ++i){
			
			measureCTEAntennaPattern[i] = cteConf[i];
			
			/* Prepare CTE. */
			if(!cte_prepare(cteType, cteConf[i])){
				
				/* This should never happen since the validity of the */
				/* {cteType} and {cteConf[i]} was already checked!    */
				
				/* Error gets reported by {cte_prepare()}. */
			}
		}
		
		measure_precompute_any_cte_sampling();
		
		return 0;
	}
}

uint8_t measure_set_radio_mode(uint8_t mode, uint8_t compatibility){
/*
  Configures the RADIO and compatibility mode. Note, that the CTE configuration
  is resetted when calling this function but may be reconfigured by calling
  {measure_set_cte_configuration()}. Note, that the sampling configuration
  is resetted when calling this function but may be reconfigured by calling
  {measure_set_sample_configuration()}.
  
  Argument List
  =============
    
    {mode}             RADIO mode to be set according to the following table:
                         
                         {0}  +->  {RADIO_MODE_BLE_2MBIT}
                         {1}  +->  {RADIO_MODE_BLE_1MBIT}
                         {2}  +->  {RADIO_MODE_NRF_2MBIT}
                         {3}  +->  {RADIO_MODE_NRF_1MBIT}
                         {4}  +->  {RADIO_MODE_IEEE_250KBIT}
                         
    {compatibility}    Whether to set compatibility mode (any value except {0})
                       or not ({0}).
  
  Return Value
  ============
    
    Returns {1} if the specified RADIO {mode} is not supported. Returns {0}
    otherwise. The RADIO mode settings are changed in the latter case only.
*/
	
	if(mode < RADIO_MODE_ALWAYS_LAST_ITEM){
		
		measureRadioMode = mode;
		measureCompatibilityMode = compatibility ? true : false;
		measure_set_cte_configuration(CTE_TYPE_NONE, NULL, NULL);
		measure_set_sample_configuration(0, NULL, NULL);
		
		/* Yes, this is redundant! */
		measure_set_timings();
		measure_precompute_forgiving_frames();
		/* Yes, this is redundant, too! */
		measure_precompute_frame_length();
		
		return 0;
	}
	else{
		
		/* RADIO {mode} not supported. */
		
		return 1;
	}
}

uint8_t measure_set_sample_configuration(
	uint8_t store,
	uint8_t * sample,
	uint8_t * failPtr
){
/*
  Configures the raw sampling behaviour as well as the raw sampling storage.
  
  Argument List
  =============
    
    {store}     Whether to store the collected samples on the onboard flash
                memory.
                         
                  {0}  +->  Do not store any collected samples on the flash
                            memory and keep normal frame length.
                  {1}  +->  Do not store any collected samples on the flash
                            memory but enlarge the frame length.
                  {2}  +->  Store all collected samples on the flash memory and
                            necessarily enlarge the frame length.
                        
                Every other value is treated as {0}.
    {sample}    Sample configuration for each frame (see
                {MeasureBufferSampleType} documentation). {NULL} is treated like
                a pointer to a memory full of zeros.
    {failPtr}   Where to write the index of that frame, for which the first
                error occured during parsing of the given configuration.
                The memory pointed to by {failPtr} is not modified in success
                case.
  
  Return Value
  ============
    
    Returns {0} in success case. In error case, i.e. if the specified
    configuration is not a valid one, the sample configuration is not changed,
    the first faulty frame index is reported within {failPtr} and one of the
    follownig error codes is returned:
      
      {1}    Configuration does not comply with the current measurement
             configuration.
      {2}    Invalid sample type specified.
      {3}    The chosen sample type is not supported by the current
             radio mode or CTE configuration.
*/
	
	uint8_t i;
	uint8_t smpl;
	
	for(i = 0; i < MEASURE_MAX_FRAME_COUNT; ++i){
		
		smpl = sample
		     ? sample[i]
		     : 0;
		
		if(smpl && i >= measureFrameCount){
			
			*failPtr = i;
			return 1;
		}
		
		if(smpl >= MEASURE_BUFFER_SAMPLE_TYPE_ALWAYS_LAST_ITEM){
			
			*failPtr = i;
			return 2;
		}
		
		if(smpl == MEASURE_BUFFER_SAMPLE_TYPE_IQ
		&& !measure_could_be_cte_sampling_frame(i)){
			
			*failPtr = i;
			return 3;
		}
	}
	
	if(store != MEASURE_SAMPLE_STORAGE_NEVER_LARGE_FRAMES
	&& store != MEASURE_SAMPLE_STORAGE_ALWAYS_LARGE_FRAMES){
		
		store = MEASURE_SAMPLE_STORAGE_NEVER_SHORT_FRAMES;
	}
	
	measureSampleStorage = store;
	
	for(i = 0; i < measureFrameCount && i < MEASURE_MAX_FRAME_COUNT; ++i){
		
		smpl = sample
		     ? sample[i]
		     : 0;
		
		measureSampleType[i] = smpl;
	}
	
	/* TODO once adding support for raw RSS samplings. */
	/* measureAnyRawSampling */
	
	measure_set_timings();
	measure_precompute_forgiving_frames();
	measure_precompute_any_cte_sampling();
	measure_precompute_frame_length();
	
	return 0;
}

bool measure_running(void){
/*
  Whether the measurement is running (i.e. the device is in measurement mode)
  or not.
*/

	return measureRunning;
}

bool measure_any_cte_sampling(void){
/*
  Whether there is any CTE sampling in the current configuration.
*/

	return measureAnyCTESampling;
}

bool measure_any_raw_sampling(void){
/*
  Whether there is any sampling of raw RSS in the current configuration.
*/
	
	return measureAnyRawSampling;
}

bool measure_any_sampling(void){
/*
  Whether there is any sampling in the current configuration.
*/
	
	if(measure_any_cte_sampling()
	|| measure_any_raw_sampling()){
		
		return true;
	}
	else{
		
		return false;
	}
}

bool measure_frame_in_range(uint8_t frame){
/*
  Whether the given {frame} index is in range within the current measurement
  configuration.
*/
	
	if(frame < measureFrameCount){
		
		return true;
	}
	else{
		
		return false;
	}
}

static void measure_enter_control_mode(void){
/*
  Leaves measurement mode and enters control mode.
*/
	measureRunning = false;
	led_set(LED_GREEN, LED_ON);
	led_set(LED_YELLOW, LED_OFF);
	radio_control_init_recv();
}

static uint8_t measure_get_first_rx_frequency(void){
/*
  Returns the frequency of the first receiving frame within the current
  measurement configuration. If there isn't any such frame, it returns
  {MEASURE_DEFAULT_FREQUENCY}.
*/

	uint8_t i;
	
	for(i = 0; i < measureFrameCount && i < MEASURE_MAX_FRAME_COUNT; ++i){
		
		if(measureFrameFrequency[i] < 0){
			
			return -measureFrameFrequency[i];
		}
	}
	
	return MEASURE_DEFAULT_FREQUENCY;
}

static uint8_t measure_get_first_tx_frequency(void){
/*
  Returns the frequency of the first active (i.e. transmitting) frame within the
  current measurement configuration. If there isn't any such frame, it returns
  {0}.
*/

	uint8_t i;
	
	for(i = 0; i < measureFrameCount && i < MEASURE_MAX_FRAME_COUNT; ++i){
		
		if(measureFrameFrequency[i] > 0){
			
			return measureFrameFrequency[i];
		}
	}
	
	return 0;
}

static void measure_update_indices(uint8_t frameIndex, int32_t cycleIndex){
/*
  Updates the frame and cycle index and applies some plausibility tests.
*/
	
	if((frameIndex < measureFrameCount
	 || (measureFrameCount == 0 && frameIndex == 0))
	&& cycleIndex < measureCycleCount){
		
		measureFrameIndex = frameIndex;
		measureCycleIndex = cycleIndex;
	}
	else if(frameIndex >= measureFrameCount){
		
		debug_report(DEBUG_MEASURE_INVALID_FRAME_INDEX);
	}
	else if(cycleIndex >= measureCycleCount){
		
		debug_report(DEBUG_MEASURE_INVALID_CYCLE_INDEX);
	}
}

static void measure_clean(void){
/*
  Resets the measurement session (i.e. out-of-sync detection, time schedule,
  result buffer, RADIO peripheral, measurement indices).
  Does not reset memory.
*/
	
	uint8_t i;
	
	measureState = MEASURE_DISABLED;
	
	/* Simulate synchronization events for all virtual devices in order */
	/* not to quit the freshly started measurement schedule to early.  */
	
	for(i = 0; i < measureSyncVirtualDeviceCount; ++i){
		
		measureSyncCycle[i] = 0;
	}
	
	measureSyncLastCheckedCycle = -1;
	measure_update_indices(measureFrameCount ? measureFrameCount - 1 : 0, -1);
	
	buffer_empty(&measureBuffer);
	
	TIMER_TASKS_CLEAR(TIMER1) = 1;
	
	measureSyncState = MEASURE_POOR_SYNC;
	measureForgivingFrames = 0;
}

void measure_init(void){
/*
  Initializes the measurement unit and enters control mode after that.
*/

	uint8_t i;
	
	static uint8_t measureBufferData[MEASURE_BUFFER_LENGTH]
	                                [MEASURE_BUFFER_WIDTH];
	static size_t measureBufferSize[MEASURE_BUFFER_LENGTH];
	
	/* Initialize schedule timer */
	
	nvic_enable_irq(NVIC_TIMER1_IRQ);
	nvic_set_priority(NVIC_TIMER1_IRQ, NVIC_RSS2I_HIGH_PRIORITY);
	
	TIMER_INTENSET(TIMER1) = TIMER_INTENSET_COMPARE0_SET
	                       | TIMER_INTENSET_COMPARE1_SET
	                       | TIMER_INTENSET_COMPARE2_SET;
	                       
	TIMER_SHORTS(TIMER1) = TIMER_SHORTS_COMPARE3_CLEAR_ENABLED;
	
	TIMER_PRESCALER(TIMER1) = 0;
	TIMER_BITMODE(TIMER1) = TIMER_BITMODE_BITMODE_32Bit;
	
	/* Initialize default schedule */
	
	measureCTELength = 0;
	measure_set_timings();
	measure_precompute_forgiving_frames();
	measure_precompute_frame_length();
	
	/* Initialize buffer for measurement results */
	
	buffer_init_rows(&measureBuffer,
	                 &measureBufferData[0][0],
	                 measureBufferSize,
	                 MEASURE_BUFFER_LENGTH,
	                 MEASURE_BUFFER_WIDTH);
	
	/* Initialize dependencies. */
	
	memory_init();
	debug_init();
	radio_init();
	
	/* Set up the pin toggling in debug mode. */
	
	#if DEBUG_MEASUREMENT_FRAME > 0
	
		gpiote_init(GPIOTE_CHANNEL_DEBUG_PIN);
		ppi_init(PPI_TIMER1_COMPARE0_GPIOTE_OUT_CHANNEL_DEBUG_PIN);
		ppi_enable(PPI_TIMER1_COMPARE0_GPIOTE_OUT_CHANNEL_DEBUG_PIN);
		
		#if DEBUG_MEASUREMENT_FRAME > 1
		
			ppi_init(PPI_TIMER1_COMPARE1_GPIOTE_SET_CHANNEL_DEBUG_PIN);
			ppi_enable(PPI_TIMER1_COMPARE1_GPIOTE_SET_CHANNEL_DEBUG_PIN);
		
		#endif
	#endif
	
	/* Set default measurement configuration */
	
	for(i = 0; i < MEASURE_MAX_FRAME_COUNT; ++i){
		
		measureFrameFrequency[i] = -17; /* Just some nice number */
	}
	
	/* Clean up */
	
	measure_clean();
	measure_enter_control_mode();
}

static uint8_t * measure_buffer_address(void){
/*
  Provides a place to store the received RADIO packet and possible raw
  samples during a measurement packet as well as measurement meta data. This
  function also registers this address within {measureBufferRowAddr} for later
  processing.
  
  Returns {NULL} if a buffer overflow occured.
*/
	
	uint8_t * rowAddr = buffer_new_row_address(&measureBuffer);
	
	if(!rowAddr){
	
		debug_report(DEBUG_MEASURE_RESULT_BUFFER_OVERFLOW);
		return NULL;
	}
	else{
		
		/* Remeber the current buffer address for later processing. */
		measureBufferRowAddr = rowAddr;
		/* Mark this entry as not-yet-written-to. */
		rowAddr[MEASURE_BUFFER_WRITTENTO_OFFSET] = 0;
		return rowAddr;
	}
}

static uint8_t * measure_ble_2mbit_rx_data_address(uint8_t frame){
/*
  Provides a place to store the received RADIO packet and possible raw
  samples during a BLE 2MBit measurement packet.
  
  Returns {NULL} if a buffer overflow occured.
*/
	
	uint8_t * rowAddr =  measure_buffer_address();
	
	if(rowAddr){
	
		/* Store the used RADIO mode. */
		rowAddr[MEASURE_BUFFER_RADIOMODE_OFFSET] = RADIO_MODE_BLE_2MBIT;
		/* Store the raw sample type. */
		rowAddr[MEASURE_BUFFER_SAMPLETYP_OFFSET] = measureSampleType[frame];
		
		/* We need some place for the measured RSS. */
		return rowAddr + MEASURE_BUFFER_RADIO_PACKET_OFFSET;
	}
	else{
		
		return NULL;
	}
}

static uint8_t * measure_ble_1mbit_rx_data_address(uint8_t frame){
/*
  Provides a place to store the received RADIO packet and possible raw
  samples during a BLE 1MBit measurement packet.
  
  Returns {NULL} if a buffer overflow occured.
*/
	
	uint8_t * rowAddr =  measure_buffer_address();
	
	if(rowAddr){
		
		/* Store the used RADIO mode. */
		rowAddr[MEASURE_BUFFER_RADIOMODE_OFFSET] = RADIO_MODE_BLE_1MBIT;
		/* Store the raw sample type. */
		rowAddr[MEASURE_BUFFER_SAMPLETYP_OFFSET] = measureSampleType[frame];
		
		/* We need some place for the measured RSS. */
		return rowAddr + MEASURE_BUFFER_RADIO_PACKET_OFFSET;
	}
	else{
		
		return NULL;
	}
}

static uint8_t * measure_ble_2mbit_listen_data_address(void){
/*
  Provides a place to store the received RADIO packet and possible raw
  samples during a BLE 2MBit measurement packet in listeing mode.
  
  Returns {NULL} if a buffer overflow occured.
*/
	
	uint8_t * rowAddr =  measure_buffer_address();
	
	if(rowAddr){
	
		/* Store the used RADIO mode. */
		rowAddr[MEASURE_BUFFER_RADIOMODE_OFFSET] = RADIO_MODE_BLE_2MBIT;
		/* Store the raw sample type. */
		/* We do not know whether there will be raw samples available. */
		rowAddr[MEASURE_BUFFER_SAMPLETYP_OFFSET] = MEASURE_BUFFER_SAMPLE_TYPE_NOTHING;
		
		/* We need some place for the measured RSS. */
		return rowAddr + MEASURE_BUFFER_RADIO_PACKET_OFFSET;
	}
	else{
		
		return NULL;
	}
}

static uint8_t * measure_ble_1mbit_listen_data_address(void){
/*
  Provides a place to store the received RADIO packet and possible raw
  samples during a BLE 1MBit measurement packet in listening mode.
  
  Returns {NULL} if a buffer overflow occured.
*/
	
	uint8_t * rowAddr =  measure_buffer_address();
	
	if(rowAddr){
	
		/* Store the used RADIO mode. */
		rowAddr[MEASURE_BUFFER_RADIOMODE_OFFSET] = RADIO_MODE_BLE_1MBIT;
		/* Store the raw sample type. */
		/* We do not know whether there will be raw samples available. */
		rowAddr[MEASURE_BUFFER_SAMPLETYP_OFFSET] = MEASURE_BUFFER_SAMPLE_TYPE_NOTHING;
		
		/* We need some place for the measured RSS. */
		return rowAddr + MEASURE_BUFFER_RADIO_PACKET_OFFSET;
	}
	else{
		
		return NULL;
	}
}

static uint8_t * measure_nrf_2mbit_rx_data_address(uint8_t frame){
/*
  Provides a place to store the received RADIO packet and possible raw
  samples during an NRF 2MBit measurement packet.
  
  Returns {NULL} if a buffer overflow occured.
*/
	
	uint8_t * rowAddr =  measure_buffer_address();
	
	if(rowAddr){
	
		/* Store the used RADIO mode. */
		rowAddr[MEASURE_BUFFER_RADIOMODE_OFFSET] = RADIO_MODE_NRF_2MBIT;
		/* Store the raw sample type. */
		rowAddr[MEASURE_BUFFER_SAMPLETYP_OFFSET] = measureSampleType[frame];
		
		/* We need some place for the measured RSS. */
		return rowAddr + MEASURE_BUFFER_RADIO_PACKET_OFFSET;
	}
	else{
		
		return NULL;
	}
}

static uint8_t * measure_nrf_1mbit_rx_data_address(uint8_t frame){
/*
  Provides a place to store the received RADIO packet and possible raw
  samples during an NRF 1MBit measurement packet.
  
  Returns {NULL} if a buffer overflow occured.
*/
	
	uint8_t * rowAddr =  measure_buffer_address();
	
	if(rowAddr){
	
		/* Store the used RADIO mode. */
		rowAddr[MEASURE_BUFFER_RADIOMODE_OFFSET] = RADIO_MODE_NRF_1MBIT;
		/* Store the raw sample type. */
		rowAddr[MEASURE_BUFFER_SAMPLETYP_OFFSET] = measureSampleType[frame];
		
		/* We need some place for the measured RSS. */
		return rowAddr + MEASURE_BUFFER_RADIO_PACKET_OFFSET;
	}
	else{
		
		return NULL;
	}
}

static uint8_t * measure_nrf_2mbit_listen_data_address(void){
/*
  Provides a place to store the received RADIO packet and possible raw
  samples during an NRF 2MBit measurement packet in listeing mode.
  
  Returns {NULL} if a buffer overflow occured.
*/
	
	uint8_t * rowAddr =  measure_buffer_address();
	
	if(rowAddr){
	
		/* Store the used RADIO mode. */
		rowAddr[MEASURE_BUFFER_RADIOMODE_OFFSET] = RADIO_MODE_NRF_2MBIT;
		/* Store the raw sample type. */
		/* We do not know whether there will be raw samples available. */
		rowAddr[MEASURE_BUFFER_SAMPLETYP_OFFSET] = MEASURE_BUFFER_SAMPLE_TYPE_NOTHING;
		
		/* We need some place for the measured RSS. */
		return rowAddr + MEASURE_BUFFER_RADIO_PACKET_OFFSET;
	}
	else{
		
		return NULL;
	}
}

static uint8_t * measure_nrf_1mbit_listen_data_address(void){
/*
  Provides a place to store the received RADIO packet and possible raw samples
  during an NRF 1MBit measurement packet in listening mode.
  
  Returns {NULL} if a buffer overflow occured.
*/
	
	uint8_t * rowAddr =  measure_buffer_address();
	
	if(rowAddr){
	
		/* Store the used RADIO mode. */
		rowAddr[MEASURE_BUFFER_RADIOMODE_OFFSET] = RADIO_MODE_NRF_1MBIT;
		/* Store the raw sample type. */
		/* We do not know whether there will be raw samples available. */
		rowAddr[MEASURE_BUFFER_SAMPLETYP_OFFSET] = MEASURE_BUFFER_SAMPLE_TYPE_NOTHING;
		
		/* We need some place for the measured RSS. */
		return rowAddr + MEASURE_BUFFER_RADIO_PACKET_OFFSET;
	}
	else{
		
		return NULL;
	}
}

static uint8_t * measure_ieee_250kbit_rx_data_address(uint8_t frame){
/*
  Provides a place to store the received RADIO packet and possible raw
  samples during an IEEE 250kBit measurement packet.
  
  Returns {NULL} if a buffer overflow occured.
*/
	
	uint8_t * rowAddr =  measure_buffer_address();
	
	if(rowAddr){
		
		/* Store the used RADIO mode. */
		rowAddr[MEASURE_BUFFER_RADIOMODE_OFFSET] = RADIO_MODE_IEEE_250KBIT;
		/* Store the raw sample type. */
		rowAddr[MEASURE_BUFFER_SAMPLETYP_OFFSET] = measureSampleType[frame];
		
		/* We need some place for the measured RSS. */
		return rowAddr + MEASURE_BUFFER_RADIO_PACKET_OFFSET;
	}
	else{
		
		return NULL;
	}
}

static uint8_t * measure_ieee_250kbit_listen_data_address(void){
/*
  Provides a place to store the received RADIO packet and possible raw samples
  during an IEEE 250kBit measurement packet in listeing mode.
  
  Returns {NULL} if a buffer overflow occured.
*/
	
	uint8_t * rowAddr =  measure_buffer_address();
	
	if(rowAddr){
		
		/* Store the used RADIO mode. */
		rowAddr[MEASURE_BUFFER_RADIOMODE_OFFSET] = RADIO_MODE_IEEE_250KBIT;
		/* Store the raw sample type. */
		/* We do not know whether there will be raw samples available. */
		rowAddr[MEASURE_BUFFER_SAMPLETYP_OFFSET] = MEASURE_BUFFER_SAMPLE_TYPE_NOTHING;
		
		/* We need some place for the measured RSS. */
		return rowAddr + MEASURE_BUFFER_RADIO_PACKET_OFFSET;
	}
	else{
		
		return NULL;
	}
}

void measure_listen(void){
/*
  Leaves the measurement schedule and prepares the RADIO in order to try to
  resync.
*/
	
	uint8_t * radioAddr;

	if(measureState != MEASURE_LISTEN){
		
		TIMER_TASKS_STOP(TIMER1) = 1;
		TIMER_TASKS_CLEAR(TIMER1) = 1;
		
		measureState = MEASURE_LISTEN;
		
		led_set(LED_YELLOW, LED_OFF);
		debug("schedule manager", "Trying to resync");
	}
	
	switch(measureRadioMode){
	
		case RADIO_MODE_BLE_2MBIT:
		
			radioAddr = measure_ble_2mbit_listen_data_address();
			
			radio_measure_ble_2mbit_prepare_rx(
				measure_get_first_rx_frequency(),
				measureCompatibilityMode,
				radioAddr,
				CTE_TYPE_NONE,
				CTE_ANTENNA_PATTERN_DISABLED
			);
			radio_measure_ble_2mbit_init_rx(
				measure_get_first_rx_frequency(),
				measureCompatibilityMode,
				radioAddr,
				CTE_TYPE_NONE,
				CTE_ANTENNA_PATTERN_DISABLED
			);
			
		break;
		case RADIO_MODE_BLE_1MBIT:
		
			radioAddr = measure_ble_1mbit_listen_data_address();
			
			radio_measure_ble_1mbit_prepare_rx(
				measure_get_first_rx_frequency(),
				measureCompatibilityMode,
				radioAddr,
				CTE_TYPE_NONE,
				CTE_ANTENNA_PATTERN_DISABLED
			);
			radio_measure_ble_1mbit_init_rx(
				measure_get_first_rx_frequency(),
				measureCompatibilityMode,
				radioAddr,
				CTE_TYPE_NONE,
				CTE_ANTENNA_PATTERN_DISABLED
			);
			
		break;
		case RADIO_MODE_NRF_2MBIT:
		
			radioAddr = measure_nrf_2mbit_listen_data_address();
			
			radio_measure_nrf_2mbit_prepare_rx(
				measure_get_first_rx_frequency(),
				measureCompatibilityMode,
				radioAddr
			);
			radio_measure_nrf_2mbit_init_rx(
				measure_get_first_rx_frequency(),
				measureCompatibilityMode,
				radioAddr
			);
			
		break;
		case RADIO_MODE_NRF_1MBIT:
		
			radioAddr = measure_nrf_1mbit_listen_data_address();
			
			radio_measure_nrf_1mbit_prepare_rx(
				measure_get_first_rx_frequency(),
				measureCompatibilityMode,
				radioAddr
			);
			radio_measure_nrf_1mbit_init_rx(
				measure_get_first_rx_frequency(),
				measureCompatibilityMode,
				radioAddr
			);
			
		break;
		case RADIO_MODE_IEEE_250KBIT:
			
			radioAddr = measure_ieee_250kbit_listen_data_address();
			
			radio_measure_ieee_250kbit_prepare_rx(
				measure_get_first_rx_frequency(),
				measureCompatibilityMode,
				radioAddr
			);
			radio_measure_ieee_250kbit_init_rx(
				measure_get_first_rx_frequency(),
				measureCompatibilityMode,
				radioAddr
			);
			
		break;
		default:
		
			radioAddr = measure_ble_2mbit_listen_data_address();
			
			radio_measure_ble_2mbit_prepare_rx(
				measure_get_first_rx_frequency(),
				measureCompatibilityMode,
				radioAddr,
				CTE_TYPE_NONE,
				CTE_ANTENNA_PATTERN_DISABLED
			);
			radio_measure_ble_2mbit_init_rx(
				measure_get_first_rx_frequency(),
				measureCompatibilityMode,
				radioAddr,
				CTE_TYPE_NONE,
				CTE_ANTENNA_PATTERN_DISABLED
			);
			
			debug_report(DEBUG_MEASURE_INVALID_RADIO_MODE);
	}
}

static void measure_resume(bool startup){
/*
  (Re-)enters the measurement schedule. If {startup} is set, the schedule
  manager is initiated, i.e. the corresponding timers are resetted and started.
  Note, that in case of resynchronization, i.e. in the case of an earlier
  synchronization loss and a subsequent successfull resynchronization event, the
  timers should be started directly by the radio module in order to minimize the
  delay between the event and the scheduler reset, i.e. {startup} should be
  {false}.
*/
	
	if(startup){
		
		/* Startup the measurement schedule manager. */
		
		radio_check_disable();
		
		measureState = MEASURE_DISABLED;
		
		TIMER_TASKS_CLEAR(TIMER1) = 1;
		TIMER_TASKS_START(TIMER1) = 1;
	}
	
	led_set(LED_YELLOW, LED_BLINK);
}


bool measure_is_configured(void){
/*
  Whether the device is configured for measurement or not.
  
  This should be checked before trying to start up/jump on a measurement. This
  was not built into {measure_start()} in order to allow situations where the
  measurement gets started with some delay (after which the cmd module cannot
  process the return status anymore).
*/

	return (measureFrameCount > 0) ? true : false;
}

bool measure_has_active_frame(void){
/*
  Whether the device is configured with an active (i.e. sending) frame.
  
  This should be checked before trying to start up a measurement. This was
  not built into {measure_start_up()} in order to allow situations where the
  measurement gets started with some delay (after which the cmd module cannot
  process the return status anymore).
*/

	return measure_get_first_tx_frequency();
}

uint32_t measure_prepare(void){
/*
  Prepares a new measurement session (resets memory), if there is no measurement
  session running.
  In that case, it returns the memory's absolute start adress. If there is
  already an ongoing measuement session, it returns {0xffffffff}.
*/
	
	if(!measure_running()){
	
		memory_reset();
		return memory_get_start_address();
	}
	else{
		
		return 0xfffffffful;
	}
}

static void measure_start(MeasureStartType startType){
/*
  Starts or jumps on a measurement session, if there is no ongoing one. This is
  the only entry point into measurement mode.
*/
	
	if(!measure_running()){
		
		measure_clean();
		measureTXMode = MEASURE_ACTIVE;
		led_set(LED_GREEN, LED_BLINK);
		measureRunning = true;

		if(startType == MEASURE_START_UP){
	
			measure_resume(true);
		}
		else{
	
			measure_listen();
		}
		
		postproc_sample_clear_request();
	}
}

void measure_start_up(void){
/*
  Starts up a measurement session. Should be preceded by a call to
  {measure_prepare()}.
*/

	measure_start(MEASURE_START_UP);
}

void measure_jump_on(void){
/*
  Jumps on/waits for an ongoing measurement session. Should be preceded by a
  call to {measure_prepare()}.
*/
	
	measure_start(MEASURE_JUMP_ON);
}

static uint8_t measure_vdev_active(uint8_t vdev, int32_t cycles){
/*
  Checks whether the virtual device {vdev} was heared from in the past {cycles}
  cycles, or not.
*/
	
	return (measureCycleIndex - measureSyncCycle[vdev] <= cycles) ? 1 : 0;
}

bool measure_disrupt(void){
/*
  Tries to put all devices within the radio range into control mode by sending
  so called disruptive frames.
  Returns {false} if the measurement configuration doesn't support this
  purpose. Returns {true} otherwise.
*/
	
	if(measure_has_active_frame()){
	
		measureTXMode = MEASURE_DISRUPTIVE;
		measureForgivingFrames = measurePrecomputedForgivingFrames;
		
		if(measureRunning){
			
			if(measureState == MEASURE_LISTEN){
				
				/* Prepare clean measurement settings before resuming. */
				measure_clean();
				
				/* Re-set {measureForgivingFrames} because it */
				/* gets resetted within {measure_clean()}.   */
				measureForgivingFrames = measurePrecomputedForgivingFrames;
				
				measure_resume(true);
			}
		}
		else{
		
			measureRunning = true;
			led_set(LED_GREEN, LED_BLINK);
			measure_resume(true);
		}
		
		return true;
	}
	else{
		
		/* It is not possible to force other devices into control mode with */
		/* this (or even no) measurement configuration.                     */
		
		return false;
	}
}

bool measure_disrupted(void){
/*
  Checks if the measurement schedule was already disrupted. In this case it
  returns {true} and enters control mode.
*/
	
	uint8_t i;
	int32_t activeDevices = 0;
	uint8_t frameCount = measureFrameCount;
	
	if(frameCount){
	
		for(i = 0; i < measureSyncVirtualDeviceCount || i == 0; ++i){
			
			/* Check for active devices within the last */
			/* ~{MEASURE_MIN_DISRUPT_FRAMES} frames.    */
			activeDevices += measure_vdev_active(
			                 	i,
			                 	measurePrecomputedForgivingFrames / frameCount
			                 );
		}
		
		/* If there are devices sending measurement packets, the following */
		/* two conditions are true. Especially if there is not any         */
		/* synchronization rule specified (yet).                           */
		
		if(activeDevices && measureState != MEASURE_LISTEN){
	
			return false;
		}
		else{
			
			measure_stop();
			return true;
		}
	}
	else{
		
		debug_report(DEBUG_MEASURE_DISRUPT_NOTHING_TO_CHECK);
		return false;
	}
}

void measure_stop(void){
/*
  Stops the measurement. This is the only entry point from measurement mode into
  control mode.
*/

	TIMER_TASKS_STOP(TIMER1) = 1;
	measure_clean();
	debug("stop", "Entering control mode.");
	/* Trigger post-measurement processes. */
	measureWasStopped = true;
	measure_enter_control_mode();
}

void measure_trigger(void){
/*
  Triggers processing of measurement results. This should be done "frequently"
  (at least twice a frame) otherwise a buffer overflow may occur.
   
  Things that get done within this function:
    
    - Saving measurement results within flash memory.
    - Logging results via serial over USB interface.
    - Preparing results for remote sharing
    - Checking for missing synchronization (a.k.a out-of-sync check)
   
*/
	
	typedef enum {
		
		STATE_READY,
		STATE_WAITING_FOR_POSTPROCESS
		
	} State;
	
	static State state = STATE_READY;
	
	uint8_t * result;
	
	RadioMode radioMode;
	
	uint8_t outOfSyncThreshCount;
	uint8_t poorSyncThreshCount;
	MeasureSyncState syncState;
	
	size_t i, j;
	
	uint8_t frame;
	int32_t cycle;
	
	uint16_t rssAvg;
	bool store;
	
	bool ready = false;
	
	uint16_t thresh;
	
	if(state == STATE_READY){
	
		/* Fetch waiting measurement buffer entries but process  */
		/* only those ones that actually has been written to.    */
		/* Proceed only if the previous postprocessing finished. */
		while(state == STATE_READY
		   && (result = buffer_get_row(&measureBuffer))
		   && result[MEASURE_BUFFER_WRITTENTO_OFFSET]){
			
			radioMode = result[MEASURE_BUFFER_RADIOMODE_OFFSET];
			
			switch(radioMode){
			
				default:
				
					debug_report(DEBUG_MEASURE_INVALID_RADIO_MODE);
				
				/* fall through */
				case RADIO_MODE_BLE_2MBIT:
			
					frame = radio_measure_ble_2mbit_get_frame(
					        	result + MEASURE_BUFFER_RADIO_PACKET_OFFSET
					        );
					cycle = radio_measure_ble_2mbit_get_cycle(
					        	result + MEASURE_BUFFER_RADIO_PACKET_OFFSET
					        );
					
				break;
				case RADIO_MODE_BLE_1MBIT:
			
					frame = radio_measure_ble_1mbit_get_frame(
					        	result + MEASURE_BUFFER_RADIO_PACKET_OFFSET
					        );
					cycle = radio_measure_ble_1mbit_get_cycle(
					        	result + MEASURE_BUFFER_RADIO_PACKET_OFFSET
					        );
					
				break;
				case RADIO_MODE_NRF_2MBIT:
			
					frame = radio_measure_nrf_2mbit_get_frame(
					        	result + MEASURE_BUFFER_RADIO_PACKET_OFFSET
					        );
					cycle = radio_measure_nrf_2mbit_get_cycle(
					        	result + MEASURE_BUFFER_RADIO_PACKET_OFFSET
					        );
					
				break;
				case RADIO_MODE_NRF_1MBIT:
			
					frame = radio_measure_nrf_1mbit_get_frame(
					        	result + MEASURE_BUFFER_RADIO_PACKET_OFFSET
					        );
					cycle = radio_measure_nrf_1mbit_get_cycle(
					        	result + MEASURE_BUFFER_RADIO_PACKET_OFFSET
					        );
					
				break;
				case RADIO_MODE_IEEE_250KBIT:
				
					frame = radio_measure_ieee_250kbit_get_frame(
					        	result + MEASURE_BUFFER_RADIO_PACKET_OFFSET
					        );
					cycle = radio_measure_ieee_250kbit_get_cycle(
					        	result + MEASURE_BUFFER_RADIO_PACKET_OFFSET
					        );
					
				break;
			}
			
			rssAvg = result[MEASURE_BUFFER_RSSAVERAGE_OFFSET + 0] * 128
				   + result[MEASURE_BUFFER_RSSAVERAGE_OFFSET + 1];
			
			
			/* Measurement data post processing. */
			
			store = (measureSampleType[frame]
			      != MEASURE_BUFFER_SAMPLE_TYPE_NOTHING
			      && measureSampleStorage
			      == MEASURE_SAMPLE_STORAGE_ALWAYS_LARGE_FRAMES)
			      ? true
			      : false;
			
			ready = postprocess(
			        	frame,
			        	cycle,
			        	measureFrameCount,
			        	measureCTEType,
			        	measureCTEAntennaPattern[frame],
			        	result[MEASURE_BUFFER_RSSAVERAGE_OFFSET + 0],
			        	result[MEASURE_BUFFER_RSSAVERAGE_OFFSET + 1],
			        	result[MEASURE_BUFFER_RSSAMOUNT_OFFSET],
			        	result[MEASURE_BUFFER_IQSAMOUNT_OFFSET],
			        	radioMode,
			        	result[MEASURE_BUFFER_SAMPLETYP_OFFSET],
			        	store,
			        	result + MEASURE_BUFFER_RADIO_PACKET_OFFSET
			        );
			
			if(!ready){
				
				state = STATE_WAITING_FOR_POSTPROCESS;
			}
			else{
				
				state = STATE_READY;
			}
			
			/* Remember this successfull synchronization. */
			measureSyncCycle[measureSyncMap[frame]] = cycle;
			
			/* Save result for remote sharing. */
			measureRemoteRSS[measureRemoteRSSInputMap[frame]] = rssAvg;
			
			/* Log result via USB. */
			livelog_log(
				frame,
				cycle,
				result[MEASURE_BUFFER_RSSAMOUNT_OFFSET],
				result[MEASURE_BUFFER_RSSAVERAGE_OFFSET + 0],
				result[MEASURE_BUFFER_RSSAVERAGE_OFFSET + 1],
				radioMode,
				result + MEASURE_BUFFER_RADIO_PACKET_OFFSET,
				measureFrameCount
			);
		}
		
		/* Finish flash memory write process when measurement got stopped. */
		
		if(measureWasStopped){
			
			memory_finish();
			
			measureWasStopped = false;
		}
		
		/* Check for synchronization. */
		
		if(measureSyncLastCheckedCycle != measureCycleIndex){
			
			syncState = MEASURE_SYNCHRONIZED;
			
			for(j = 0; j < measureSyncVirtualDeviceCount; ++j){
				
				/*
				   If there are less than {j + 1} synchronization events (i.e.
				   received measurement packtes) from different devices within
				   the last {measureMissedCyclesPoorSyncThresh[j] * 10} cycles
				   or within the last {measureMissedCyclesOutOfSyncThresh[j]*10}
				   cycles, the device is muted temporarily (i.e. is not
				   _sending_ measurement packets  anymore) or, resp., leaves the
				   measurement schedule and tries to resync with the beginning
				   of the next measurement cycle.
				*/
				
				if(measureMissedCyclesOutOfSyncThresh[j]){
				
					outOfSyncThreshCount = 0;
				
					for(i = 0; i < measureSyncVirtualDeviceCount; ++i){
						
						if(measureMissedCyclesOutOfSyncThresh[j] > 100){
							thresh = 100 + (measureMissedCyclesOutOfSyncThresh[j] - 100) * 10;
						}
						else{
							thresh = measureMissedCyclesOutOfSyncThresh[j];
						}
						
						outOfSyncThreshCount += measure_vdev_active(
							i, thresh 
						);
					}
				
					if(outOfSyncThreshCount < j + 1){
					
						/* TODO: Debug message containing - rule number - cycle information - ... */
						
						/* Leaving measurement schedule... */
						syncState = MEASURE_OUT_OF_SYNC;
						break;
					}
				}
				
				if(measureMissedCyclesPoorSyncThresh[j]){
					
					poorSyncThreshCount = 0;
					
					for(i = 0; i < measureSyncVirtualDeviceCount; ++i){
				
						if(measureMissedCyclesPoorSyncThresh[j] > 100){
							thresh = 100 + (measureMissedCyclesPoorSyncThresh[j] - 100) * 10;
						}
						else{
							thresh = measureMissedCyclesPoorSyncThresh[j];
						}
						
						poorSyncThreshCount += measure_vdev_active(
							i, thresh
						);
					}
					
					if(poorSyncThreshCount < j + 1){
						
						/* TODO: Debug message containing - rule number - cycle information - ... */
						
						/* Muting... */
						syncState = MEASURE_POOR_SYNC;
						break;
					}
				}
			}
			
			if(syncState != measureSyncState){
				
				if(syncState == MEASURE_POOR_SYNC){
					
					debug("async sync check", "Muted.");
				}
				
				if(syncState == MEASURE_OUT_OF_SYNC){
					
					debug("async sync check", "Leaving (resync) suggested.");
				}
				
				if(syncState == MEASURE_SYNCHRONIZED){
					
					debug("async sync check", "Successfully resynced.");
				}
				
				measureSyncState = syncState;
			}
			
			measureSyncLastCheckedCycle = measureCycleIndex;
		}
	}
	else{
		
		/* The postproc module needs some additional calls to finish the */
		/* postprocessing initiated earlier.                             */
		
		/* Measurement data post processing. */
		ready = postprocess(
		        	0,
		        	0,
		        	0,
		        	CTE_TYPE_NONE,
		        	CTE_ANTENNA_PATTERN_DISABLED,
		        	0,
		        	0,
		        	0,
		        	0,
		        	RADIO_MODE_BLE_2MBIT,
		        	MEASURE_BUFFER_SAMPLE_TYPE_NOTHING,
		        	false,
		        	NULL
		        );
		
		if(!ready){
			
			state = STATE_WAITING_FOR_POSTPROCESS;
		}
		else{
			
			state = STATE_READY;
		}
	}
}

static void measure_prepare_frame(void){
/*
   Gets called by the schedule manager at the beginning of a frame. Maintains
   frame and cylce indexing and prepares the RADIO peripheral to be initialized
   later on.
*/
	
	bool disrupt;
	
	if(measureFrameCount){
	
		/* Maintain frame and cycle indices. */
		
		if(++measureFrameIndex == measureFrameCount){
		
			measureFrameIndex = 0;
		
			if(++measureCycleIndex == measureCycleCount){
			
				measure_stop();
			}
		}
	
		/* Prepare RADIO peripheral. */
		
		/* Make sure we haven't quit measurement mode. */
		if(measure_running()){
			
			if(measureFrameFrequency[measureFrameIndex] > 0){
				
				/* Send a measurement packet during this frame */
				
				disrupt = (measureTXMode == MEASURE_DISRUPTIVE) ? true : false;
				
				switch(measureRadioMode){
					
					default:
						
						debug_report(DEBUG_MEASURE_INVALID_RADIO_MODE);
					
					/* fall through */
					case RADIO_MODE_BLE_2MBIT:
					
						radio_measure_ble_2mbit_prepare_tx(
							measureFrameFrequency[measureFrameIndex],
							disrupt,
							measureFrameIndex, measureCycleIndex,
							(uint16_t *) &measureRemoteRSS[
								measureRemoteRSSOutputMap[measureFrameIndex]
							],
							measureRemoteRSSOutputCount,
							(CTEType) measureCTEType,
							(CTEAntennaPattern) measureCTEAntennaPattern[measureFrameIndex]
						);
						
					break;
					case RADIO_MODE_BLE_1MBIT:
					
						radio_measure_ble_1mbit_prepare_tx(
							measureFrameFrequency[measureFrameIndex],
							disrupt,
							measureFrameIndex, measureCycleIndex,
							measureRemoteRSSOutputCount
								? measureRemoteRSS[
								  	measureRemoteRSSOutputMap[measureFrameIndex]
								  ]
								: 0,
							(CTEType) measureCTEType,
							(CTEAntennaPattern) measureCTEAntennaPattern[measureFrameIndex]
						);
						
					break;
					case RADIO_MODE_NRF_2MBIT:
					
						radio_measure_nrf_2mbit_prepare_tx(
							measureFrameFrequency[measureFrameIndex],
							disrupt,
							measureFrameIndex, measureCycleIndex,
							(uint16_t *) &measureRemoteRSS[
								measureRemoteRSSOutputMap[measureFrameIndex]
							],
							measureRemoteRSSOutputCount
						);
						
					break;
					case RADIO_MODE_NRF_1MBIT:
					
						radio_measure_nrf_1mbit_prepare_tx(
							measureFrameFrequency[measureFrameIndex],
							disrupt,
							measureFrameIndex, measureCycleIndex,
							measureRemoteRSSOutputCount
								? measureRemoteRSS[
								  	measureRemoteRSSOutputMap[measureFrameIndex]
								  ]
								: 0
						);
						
					break;
					case RADIO_MODE_IEEE_250KBIT:
					
						radio_measure_ieee_250kbit_prepare_tx(
							measureFrameFrequency[measureFrameIndex],
							disrupt,
							measureFrameIndex, measureCycleIndex
						);
					break;
				}
			}
			else{
				
				/* Try to receive a measurement packet during this frame */
				
				switch(measureRadioMode){
					
					default:
						
						debug_report(DEBUG_MEASURE_INVALID_RADIO_MODE);
					
					/* fall through */
					case RADIO_MODE_BLE_2MBIT:
					
						/* Register a place for storing the radioPacket. */
						/* Gets reused in {measure_start_frame()}.       */
						measureBufferRadioAddr
							= measure_ble_2mbit_rx_data_address(
							  	measureFrameIndex
							  );
						
						/* TODO once adding support for raw RSS sampling. */
						radio_measure_ble_2mbit_prepare_rx(
							-measureFrameFrequency[measureFrameIndex],
							measureCompatibilityMode,
							measureBufferRadioAddr,
							measureCTEType,
							measureCTEAntennaPattern[measureFrameIndex]
						);
					
					break;
					case RADIO_MODE_BLE_1MBIT:
					
						/* Register a place for storing the radioPacket. */
						/* Gets reused in {measure_start_frame()}.       */
						measureBufferRadioAddr
							= measure_ble_1mbit_rx_data_address(
							  	measureFrameIndex
							  );
						
						/* TODO once adding support for raw RSS sampling. */
						radio_measure_ble_1mbit_prepare_rx(
							-measureFrameFrequency[measureFrameIndex],
							measureCompatibilityMode,
							measureBufferRadioAddr,
							measureCTEType,
							measureCTEAntennaPattern[measureFrameIndex]
						);
					
					break;
					case RADIO_MODE_NRF_2MBIT:
					
						/* Register a place for storing the radioPacket. */
						/* Gets reused in {measure_start_frame()}.       */
						measureBufferRadioAddr
							= measure_nrf_2mbit_rx_data_address(
							  	measureFrameIndex
							  );
						
						/* TODO once adding support for raw RSS sampling. */
						radio_measure_nrf_2mbit_prepare_rx(
							-measureFrameFrequency[measureFrameIndex],
							measureCompatibilityMode,
							measureBufferRadioAddr
						);
					
					break;
					case RADIO_MODE_NRF_1MBIT:
					
						/* Register a place for storing the radioPacket. */
						/* Gets reused in {measure_start_frame()}.       */
						measureBufferRadioAddr
							= measure_nrf_1mbit_rx_data_address(
							  	measureFrameIndex
							  );
						
						/* TODO once adding support for raw RSS sampling. */
						radio_measure_nrf_1mbit_prepare_rx(
							-measureFrameFrequency[measureFrameIndex],
							measureCompatibilityMode,
							measureBufferRadioAddr
						);
					
					break;
					case RADIO_MODE_IEEE_250KBIT:
						
						/* Register a place for storing the radioPacket. */
						/* Gets reused in {measure_start_frame()}.       */
						measureBufferRadioAddr
							= measure_ieee_250kbit_rx_data_address(
							  	measureFrameIndex
							  );
						
						/* TODO once adding support for raw RSS sampling. */
						radio_measure_ieee_250kbit_prepare_rx(
							-measureFrameFrequency[measureFrameIndex],
							measureCompatibilityMode,
							measureBufferRadioAddr
						);
					
					break;
				}
			}
		}
	}
	else{
		
		debug_report(DEBUG_MEASURE_NO_CONFIG);
	}
}

static void measure_start_frame(void){
/*
   Gets called by the schedule manager in the early phase of a frame.
   Initializes RADIO peripheral and synchronizes the onboard LEDs.
*/
	
	bool disrupt;
	
	if(measureFrameCount){
	
		/* Initialize RADIO peripheral. */
		
		if(measureFrameFrequency[measureFrameIndex] > 0){
		
			if((measureTXMode == MEASURE_ACTIVE
			 && measureSyncState == MEASURE_SYNCHRONIZED)
			|| measureTXMode == MEASURE_DISRUPTIVE){
				
				/* Send a measurement packet during this frame */
				
				disrupt = (measureTXMode == MEASURE_DISRUPTIVE) ? true : false;
				measureState = MEASURE_TX;
				
				switch(measureRadioMode){
					
					default:
						
						debug_report(DEBUG_MEASURE_INVALID_RADIO_MODE);
					
					/* fall through */
					case RADIO_MODE_BLE_2MBIT:
					
						radio_measure_ble_2mbit_init_tx(
							measureFrameFrequency[measureFrameIndex],
							disrupt,
							measureFrameIndex, measureCycleIndex,
							(uint16_t *) &measureRemoteRSS[
								measureRemoteRSSOutputMap[measureFrameIndex]
							],
							measureRemoteRSSOutputCount,
							measureCTEType,
							measureCTEAntennaPattern[measureFrameIndex]
						);
						
					break;
					case RADIO_MODE_BLE_1MBIT:
					
						radio_measure_ble_1mbit_init_tx(
							measureFrameFrequency[measureFrameIndex],
							disrupt,
							measureFrameIndex, measureCycleIndex,
							measureRemoteRSSOutputCount
								? measureRemoteRSS[
								  	measureRemoteRSSOutputMap[measureFrameIndex]
								  ]
								: 0,
							measureCTEType,
							measureCTEAntennaPattern[measureFrameIndex]
						);
						
					break;
					case RADIO_MODE_NRF_2MBIT:
					
						radio_measure_nrf_2mbit_init_tx(
							measureFrameFrequency[measureFrameIndex],
							disrupt,
							measureFrameIndex, measureCycleIndex,
							(uint16_t *) &measureRemoteRSS[
								measureRemoteRSSOutputMap[measureFrameIndex]
							],
							measureRemoteRSSOutputCount
						);
						
					break;
					case RADIO_MODE_NRF_1MBIT:
					
						radio_measure_nrf_1mbit_init_tx(
							measureFrameFrequency[measureFrameIndex],
							disrupt,
							measureFrameIndex, measureCycleIndex,
							measureRemoteRSSOutputCount
								? measureRemoteRSS[
								  	measureRemoteRSSOutputMap[measureFrameIndex]
								  ]
								: 0
						);
						
					break;
					case RADIO_MODE_IEEE_250KBIT:
					
						radio_measure_ieee_250kbit_init_tx(
							measureFrameFrequency[measureFrameIndex],
							disrupt,
							measureFrameIndex, measureCycleIndex
						);
					break;
				}
			}
		}
		else{
			
			/* Try to receive a measurement packet during this frame */
			
			if(measureBufferRadioAddr){
				
				measureState = MEASURE_RX;
				
				switch(measureRadioMode){
					
					default:
						
						debug_report(DEBUG_MEASURE_INVALID_RADIO_MODE);
					
					/* fall through */
					case RADIO_MODE_BLE_2MBIT:
						
						/* TODO once adding support for raw RSS sampling. */
						radio_measure_ble_2mbit_init_rx(
							-measureFrameFrequency[measureFrameIndex],
							measureCompatibilityMode,
							measureBufferRadioAddr,
							measureCTEType,
							measureCTEAntennaPattern[measureFrameIndex]
						);
					
					break;
					case RADIO_MODE_BLE_1MBIT:
						
						/* TODO once adding support for raw RSS sampling. */
						radio_measure_ble_1mbit_init_rx(
							-measureFrameFrequency[measureFrameIndex],
							measureCompatibilityMode,
							measureBufferRadioAddr,
							measureCTEType,
							measureCTEAntennaPattern[measureFrameIndex]
						);
					
					break;
					case RADIO_MODE_NRF_2MBIT:
					
						/* TODO once adding support for raw RSS sampling. */
						radio_measure_nrf_2mbit_init_rx(
							-measureFrameFrequency[measureFrameIndex],
							measureCompatibilityMode,
							measureBufferRadioAddr
						);
					
					break;
					case RADIO_MODE_NRF_1MBIT:
						
						/* TODO once adding support for raw RSS sampling. */
						radio_measure_nrf_1mbit_init_rx(
							-measureFrameFrequency[measureFrameIndex],
							measureCompatibilityMode,
							measureBufferRadioAddr
						);
					
					break;
					case RADIO_MODE_IEEE_250KBIT:
						
						/* TODO once adding support for raw RSS sampling. */
						radio_measure_ieee_250kbit_init_rx(
							-measureFrameFrequency[measureFrameIndex],
							measureCompatibilityMode,
							measureBufferRadioAddr
						);
					
					break;
				}
			}
			else{
				
				/* This should never happen!          */
				/* (The frame should not be initiated */
				/* before it was prepared)            */
				debug_report(DEBUG_MEASURE_RADIO_PACKET_ADDR_NULL);
			}
		}
			
		
		/* Trigger LEDs */
		
		if((measureFrameIndex + measureFrameCount*measureCycleIndex)
		 % ((((int32_t) (LED_SLOT_PAUSE + 2 * LED_SLOT_OFFSET))
		   * 1000
		   / (measurePrecomputedFrameLength + measureCTELength))
		  + 1) == 0){
	
			led_trigger();
		}
	}
	else{
		
		debug_report(DEBUG_MEASURE_NO_CONFIG);
	}
}

static void measure_rx_succeeded(uint16_t rAvg,
                                 uint8_t rAmnt,
                                 uint8_t iqAmnt,
                                 uint8_t (* get_frame)(volatile uint8_t * radioAddr),
                                 int32_t (* get_cycle)(volatile uint8_t * radioAddr)){
/*
  Buffers the radio packet and the given meta data and updates frame and cycle
  indices.
  
  Argument Overview
  =================
    
    {rAvg}      Average Received Signal Strength (RSS) during packet reception.
    {rAmnt}     Amount of taken RSS samples during packet reception.
    {iqAmnt}    Amount of taken IQ samples during packet reception.
    {get_frame} Function for getting the frame index out of a received packet.
    {get_cycle} Function for getting the cycle index out of a received packet.
*/
	
	if(measureState == MEASURE_LISTEN){
		
		measure_resume(false);
	}
	
	measureState = MEASURE_DISABLED;
	
	if(measureBufferRowAddr){
	
		measure_update_indices(
			get_frame(
				  (uint8_t *) measureBufferRowAddr
				+ MEASURE_BUFFER_RADIO_PACKET_OFFSET
			),
			get_cycle(
				  (uint8_t *) measureBufferRowAddr
				+ MEASURE_BUFFER_RADIO_PACKET_OFFSET
			)
		);
		
		measureBufferRowAddr[MEASURE_BUFFER_RSSAVERAGE_OFFSET + 0] = rAvg / 128;
		measureBufferRowAddr[MEASURE_BUFFER_RSSAVERAGE_OFFSET + 1] = rAvg % 128;
		measureBufferRowAddr[MEASURE_BUFFER_RSSAMOUNT_OFFSET] = rAmnt;
		measureBufferRowAddr[MEASURE_BUFFER_IQSAMOUNT_OFFSET] = iqAmnt;
		measureBufferRowAddr[MEASURE_BUFFER_WRITTENTO_OFFSET] = 1;
		
		buffer_new_row_finish(
			&measureBuffer,
			(uint8_t *) measureBufferRowAddr,
			MEASURE_BUFFER_WIDTH
		);
		measureBufferRowAddr = NULL;
		measureBufferRadioAddr = NULL;
	}
	else{
		
		/* This should never happen!                  */
		/* (The measurement task should not ever have */
		/* been initiated within the radio module.)   */
		debug_report(DEBUG_MEASURE_RESULT_BUFFER_ADDR_NULL);
	}
}

void measure_ieee_250kbit_rx_succeeded(uint16_t rAvg, uint8_t rAmnt){
/*
  Should be called by the RADIO whenever a valid IEEE 250kBit measurement packet
  was received. Buffers the radio packet and the given meta data and updates
  frame and cycle indices.
  
  Argument Overview
  =================
    
    {rAvg}      Average Received Signal Strength (RSS) during packet reception.
    {rAmnt}     Amount of taken RSS samples during packet reception.
*/
	
	measure_rx_succeeded(
		rAvg,
		rAmnt,
		0,
		&radio_measure_ieee_250kbit_get_frame,
		&radio_measure_ieee_250kbit_get_cycle
	);
}

void measure_ble_2mbit_rx_succeeded(uint16_t rAvg,
                                    uint8_t rAmnt,
                                    uint8_t iqAmnt){
/*
  Should be called by the RADIO whenever a valid BLE 2MBit measurement packet
  was received. Buffers the radio packet and the given meta data and updates
  frame and cycle indices.
  
  Argument Overview
  =================
    
    {rAvg}      Average Received Signal Strength (RSS) during packet reception.
    {rAmnt}     Amount of taken RSS samples during packet reception.
    {iqAmnt}    Amount of taken IQ samples during packet reception.
*/
	
	
	measure_rx_succeeded(
		rAvg,
		rAmnt,
		iqAmnt,
		&radio_measure_ble_2mbit_get_frame,
		&radio_measure_ble_2mbit_get_cycle
	);
}

void measure_ble_1mbit_rx_succeeded(uint16_t rAvg,
                                    uint8_t rAmnt,
                                    uint8_t iqAmnt){
/*
  Should be called by the RADIO whenever a valid BLE 1MBit measurement packet
  was received. Buffers the radio packet and the given meta data and updates
  frame and cycle indices.
  
  Argument Overview
  =================
    
    {rAvg}      Average Received Signal Strength (RSS) during packet reception.
    {rAmnt}     Amount of taken RSS samples during packet reception.
    {iqAmnt}    Amount of taken IQ samples during packet reception.
*/
	
	
	measure_rx_succeeded(
		rAvg,
		rAmnt,
		iqAmnt,
		&radio_measure_ble_1mbit_get_frame,
		&radio_measure_ble_1mbit_get_cycle
	);
}

void measure_nrf_2mbit_rx_succeeded(uint16_t rAvg,
                                    uint8_t rAmnt){
/*
  Should be called by the RADIO whenever a valid NRF 2MBit measurement packet
  was received. Buffers the radio packet and the given meta data and updates
  frame and cycle indices.
  
  Argument Overview
  =================
    
    {rAvg}      Average Received Signal Strength (RSS) during packet reception.
    {rAmnt}     Amount of taken RSS samples during packet reception.
*/
	
	
	measure_rx_succeeded(
		rAvg,
		rAmnt,
		0,
		&radio_measure_nrf_2mbit_get_frame,
		&radio_measure_nrf_2mbit_get_cycle
	);
}

void measure_nrf_1mbit_rx_succeeded(uint16_t rAvg,
                                    uint8_t rAmnt){
/*
  Should be called by the RADIO whenever a valid NRF 1MBit measurement packet
  was received. Buffers the radio packet and the given meta data and updates
  frame and cycle indices.
  
  Argument Overview
  =================
    
    {rAvg}      Average Received Signal Strength (RSS) during packet reception.
    {rAmnt}     Amount of taken RSS samples during packet reception.
*/
	
	
	measure_rx_succeeded(
		rAvg,
		rAmnt,
		0,
		&radio_measure_ble_1mbit_get_frame,
		&radio_measure_ble_1mbit_get_cycle
	);
}

void measure_rx_failed(void){
/*
  Should be called by the RADIO whenever a corrupt packet was received
  during listening for measurement packets.
*/
	
	if(measureState == MEASURE_LISTEN){
		
		/* Retry! */
		measure_listen();
	}
}

void timer1_isr(void){
/*
   This is the measurement schedule manager. There are four important dates
   within a schedule period:
     
     "Frame's begin":  The previous frame's guard interval is over. Prepare the
                       next frame by disabling the RADIO peripheral.
     
     "Prepare RADIO":  Initialize (i.e. ramp up) the RADIO peripheral for the
                       upcoming task.
     "Transmit":       If the current frame is an active one, trigger the
                       transmission of a measurement packet.
     "Packet's end":   The nominal packet reception time is over. Start the
                       post guard interval by resetting the schedule timer.
                       A reset may be triggered earlier or may even be triggered
                       a second time from the RADIO peripheral, when a valid
                       measurement packet was received or a measurement packet
                       has been transmitted successfully.
*/
	
	if(TIMER_EVENTS_COMPARE(TIMER1)[0]){
	/*
	  Disable the RADIO peripheral.
	*/
	
		#if DEBUG_MEASUREMENT_FRAME > 0
		
			/* GPIO debug pin is toggled via PPI and GPIOTE! */
			
			#if DEBUG_MEASUREMENT_FRAME > 1
			
				GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
				GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
				GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
				
			#endif
		#endif
	
		TIMER_EVENTS_COMPARE(TIMER1)[0] = 0;
		
		radio_check_disable();
		measure_prepare_frame();
		
		while(TIMER_EVENTS_COMPARE(TIMER1)[0]);
	}
	else if(TIMER_EVENTS_COMPARE(TIMER1)[1]){
	/*
	  Prepare (i.e. ramp up) the RADIO peripheral.
	*/
		
		#if DEBUG_MEASUREMENT_FRAME > 1
		
			/* GPIO debug pin is set to high via PPI and GPIOTE! */
			
		#endif
		
		TIMER_EVENTS_COMPARE(TIMER1)[1] = 0;
		
		if(measure_running()){
		
			if(measureSyncState != MEASURE_OUT_OF_SYNC){
				
				measure_start_frame();
			}
			else if(measureForgivingFrames){
				
				/* Ignore synchronization loss. */
				--measureForgivingFrames;
				measure_start_frame();
			}
			else{
			
				measureSyncState = MEASURE_POOR_SYNC;
				measure_listen();
			}
		}
		
		/* Make sure the above to take at least four CPU cycles. */
		/* while(TIMER_EVENTS_COMPARE(TIMER1)[1]);               */
	}
	else if(TIMER_EVENTS_COMPARE(TIMER1)[2]){
	/*
	  Transmit a measurement packet, if the current frame is an active one.
	*/
		
		TIMER_EVENTS_COMPARE(TIMER1)[2] = 0;
		
		if(measureState == MEASURE_TX
		&& measureSyncState == MEASURE_SYNCHRONIZED){
		
			radio_measure_start_tx();
		}
		
		while(TIMER_EVENTS_COMPARE(TIMER1)[2]);
	}
}

