/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

#define RSS2I_MODULE "memory"

/*
  API for controlling the external flash memory.
  
  Authors
  =======
    
    Copyright (C) 2018-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
    
    2019-06-28    Initial public release.
    2019-10-31    Added {memory_ready_to_request()}. This is part of bugfix #6.
    2019-11-04    Fix {memoryPageCount} behaviour when the memory is full.
    2020-03-24    Fix {memory_reset()} behaviour. Measurement data from a new
                  measurement session is now stored within a new 32 kiB block.
    2020-07-08    Remove magic numbers and add some static asserts.
    2020-07-08    Switched to DMA-based flash page managing.
    2020-07-10    Improved error handling of {memory_program_next_page()}.
    2020-07-13    Added {memory_*lock()} and {memory_measure_write_results*()}.
    2020-07-14    Removed some more magic numbers and stabilized memory tests.
    2021-04-15    Support raw sample storage.
    2021-04-21    Use {fmtByte} instead of {fmtSpec} to get index increment.
    2021-04-29    Make it possible to finish a partially written flash page.
    
  Description
  ===========
    
    Addresses, Pages, Blocks and Chips
    ----------------------------------
     
     Every single byte inside the two external flash chips is addressable.
     However, with respect to performance, the external flash memory is written
     page-wise, i.e. 256 bytes at a time.
     
     128 of such pages form a 32 kiB block of memory. An 8 MB flash chip
     consists of 256 such 32 kiB blocks. Hence there are 512 32 kiB blocks
     available on both flash chips in total.
     
     Some illustrations regarding the address of a byte inside the memory:
     
     
                          page index
                 32-kiB   within a
                 block     32-kiB-
                 index     block
              .----'----..---'--.
              |         ||      |
            0b0000 1111 0000 1111 0000 1111
              |                 | |       |
              '--------.--------' '---.---'
                   page index         |
                                      |
                                 bytes within
                                    a page
                                      |    
            chip ID                   |    
              .                   .---'---.
              |                   |       |
            0b0000 1111 0000 1111 0000 1111
               |                |
               '--------.-------'
                    page index
                     within a
                       chip
    
    
    Initialization
    --------------
     
     In order to uniformly distribute the write/erase cycles over the whole
     memory, the write process starts at a randomly chosen 32 kiB block. This
     is done, beside other things, by {memory_init()}.
    
    Writing measurement values
    --------------------------
     
     Using {memory_measure_write_results()} one may store to bytes (measurement
     values), a bunch of raw CTE IQ samples and a corresponding four byte index
     inside the memory. Everything else, i.e. the page buffering, adding
     redundancy, erasing blocks, etc., gets abstracted.
    
    Reading
    -------
     
     Before reading a page from the memory make sure that the memory controller
     is ready to process a read request by calling {memory_ready_to_request()}.
     
     Reading a page from the memory then happens by requesting a page using
     {memory_request_offset()} or {memory_request_absolute()}. The former
     requests a page whose address is numbered with respect to the randomly
     chosen start block (see 'Initialization' above). The latter request a page
     with an absolute address like illustrated above.
     
     After requesting a page it is necessary to wait until
     {memory_request_ready()} signals that the request was handled completely.
     The page then may be read by a call to {memory_read()}.
   
    Testing
    -------
    
     The {memory_full_test()} may be used to validate the integrity of the whole
     memory. The {memory_fast_test()} may be used to assure that both flash
     chips are addressable and functional.
*/

#include <stdint.h>
#include <stdbool.h>
#include <stddef.h>

#include <rss2i/spim.h>
#include <rss2i/memory.h>
#include <rss2i/random.h>
#include <rss2i/debug.h>
#include <rss2i/crc.h>
#include <rss2i/print.h>
#include <rss2i/coder.h>

#include <rss2i/config/memory.h>
#include <rss2i/config/spim.h>

#if MEMORY_RESERVED_WRITE_ENTRIES > SPIM_BUFFER_LENGTH - 1
	#error "SPIM buffer is completely reserved for writing data."
#endif


#define MEMORY_ERASE_CHIP_LENGTH             5
#define MEMORY_ERASE_BLOCK32_LENGTH          8
#define MEMORY_PROGRAM_PAGE_CONTENT_LENGTH 256
#define MEMORY_PROGRAM_PAGE_CONTENT_OFFSET   8
#define MEMORY_READ_PAGE_CONTENT_LENGTH    256
#define MEMORY_READ_PAGE_CONTROL_LENGTH      8

#define MEMORY_PROGRAM_PAGE_LENGTH       ( MEMORY_PROGRAM_PAGE_CONTENT_LENGTH \
                                         + MEMORY_PROGRAM_PAGE_CONTENT_OFFSET )
#define MEMORY_READ_PAGE_LENGTH          ( MEMORY_READ_PAGE_CONTENT_LENGTH \
                                         + MEMORY_READ_PAGE_CONTROL_LENGTH )

_Static_assert(
	   MEMORY_ERASE_CHIP_LENGTH    <= SPIM_BUFFER_WIDTH
	&& MEMORY_ERASE_BLOCK32_LENGTH <= SPIM_BUFFER_WIDTH
	&& MEMORY_PROGRAM_PAGE_LENGTH  <= SPIM_BUFFER_WIDTH
	&& MEMORY_READ_PAGE_LENGTH     <= SPIM_BUFFER_WIDTH,
	"At least one SPIM query does not fit onto the SPIM buffer."
);

#define MEMORY_PROGRAM_PAGE_DATA_LENGTH    252
#define MEMORY_PROGRAM_PAGE_CRC_OFFSET     260
#define MEMORY_PROGRAM_PAGE_CRC_LENGTH       4
#define MEMORY_READ_PAGE_CONTENT_OFFSET      4

#define MEMORY_NUM_OF_PAGES              (1ul << 16)
#define MEMORY_NUM_OF_CHIPS                  2


typedef enum {
	
	MEMORY_TEST_STATE_PREPARE,
	MEMORY_TEST_STATE_ERASE_1,
	MEMORY_TEST_STATE_ERASE_2,
	MEMORY_TEST_STATE_WRITE,
	MEMORY_TEST_STATE_REQUEST,
	MEMORY_TEST_STATE_READ,
	
} MemoryTestState;


static uint16_t memoryFirstPage;
static volatile uint32_t memoryPageCount;
static volatile uint16_t memoryCurrentPage;
static volatile bool memoryLocked = false;


static bool memory_spim_check_wip(uint8_t * data){
/*
   Given the received data during a RDSR (ReaD Status Register) SPIM operation
   it checks if the WIP (Write In Progress) bit is set.
*/
	
	return (data[1] & (1 << 0)) ? true : false;
}

static bool memory_erase_chip(uint16_t address){
/*
  Erases the specified chip. If {address} isn't aligned with the beginning of a
  chip, it gets rounded down to the next valid chip start address.
  
  Returns {true} in case of success. Returns {false} if the SPIM peripheral is
  too busy to process the generated query.
*/
	
	uint8_t data[MEMORY_ERASE_CHIP_LENGTH] = {
	
		0,                                  /* Reserved for future extensions */
		(address & (1ul << 15)) ? 1 : 0,    /* Chip ID                        */
		0,                                  /* Reception disabled             */
		1,                                  /* WEL needed                     */
		0xc7,                               /* CER SPIM instruction           */
	};
	
	if(!spim_query(data, MEMORY_ERASE_CHIP_LENGTH)){
		
		debug_report(DEBUG_MEMORY_BUFFER_OVERFLOW);
		return false;
	}
	else{
		
		return true;
	}
}

static bool memory_erase_block32(uint16_t address){
/*
  Erases a 32-kiB-block of memory beginning with the given page {address}.
  If {address} isn't aligned with the beginning of a 32-kiB-block, it gets
  rounded down to the next valid page address.
  
  Returns {true} in case of success. Returns {false} if the SPIM peripheral is
  too busy to process the generated query.
*/
	
	uint8_t data[MEMORY_ERASE_BLOCK32_LENGTH] = {
	
		0,                                  /* Reserved for future extensions */
		(address & (1ul << 15)) ? 1 : 0,    /* Chip ID                        */
		0,                                  /* Reception disabled             */
		1,                                  /* WEL needed                     */
		0x52,                               /* BER32K SPIM instruction        */
		address >> 8,                       /* Address byte 2                 */
		address & 0x80,                     /* Address byte 1                 */
		0                                   /* Address byte 0                 */
	};
	
	if(!spim_query(data, MEMORY_ERASE_BLOCK32_LENGTH)){
		
		debug_report(DEBUG_MEMORY_BUFFER_OVERFLOW);
		return false;
	}
	else{
		
		return true;
	}
}

static uint8_t * memory_program_page_request(void){
/*
  Returns a RAM address where the next page to be programmed can be stored
  before it is finally programmed by calling {memory_program_page_finish()}.
  Returns {NULL} if the SPIM peripheral is too busy to process the necessary
  query.
  
  First {MEMORY_PROGRAM_PAGE_CONTENT_OFFSET} Bytes
  ================================================
  
    Note, that the first {MEMORY_PROGRAM_PAGE_CONTENT_OFFSET} bytes pointed to
    by the return value are reserved for internal use and hence shall not be
    written to by the user.
    
  Memory Write Warning
  ====================
    
    The user has to ensure that the total length, i.e. including the first
    {MEMORY_PROGRAM_PAGE_CONTENT_OFFSET} bytes, of the written data complies
    with the reserved memory per buffer entry as it is specified by
    {SPIM_BUFFER_WIDTH}.
*/
	
	uint8_t * addr;
	
	if((addr = spim_query_request())){
		
		return addr;
	}
	else{
		
		debug_report(DEBUG_MEMORY_BUFFER_OVERFLOW);
		return NULL;
	}
}

static bool memory_program_page_finish(uint16_t address, uint8_t * addr){
/*
  Finishes the DMA-based page programming procedure that was initiated by
  calling {memory_program_page_request()}.
  
  Some Notes
  ==========
  
    This function does not perform any copying tasks. The specified {addr} is
    used for validation purposes only.
  
  Argument List
  =============
    
    address    Page address to be programmed.
    addr       Pointer to the already written RAM as it was returned by the
               previous call to {memory_program_page_request()}.
  
  Return Value
  ============
    
    Returns {false} when illegal buffer activities, i.e. buffer activities
    between the call of {memory_program_page_request()} and the call to this
    function, have been detetcted. Returns {true} otherwise.
*/
	
	addr[0] = 0;                               /* Reserved            */
	addr[1] = (address & (1 << 15)) ? 1 : 0;   /* Chip ID             */
	addr[2] = 0;                               /* Reception disabled  */
	addr[3] = 1;                               /* WEL needed          */
	addr[4] = 0x02;                            /* PP SPIM instruction */
	addr[5] = (address >> 8) & 0x7F;           /* Address byte 2      */
	addr[6] = address & 0xFF;                  /* Address byte 1      */
	addr[7] = 0;                               /* Address byte 0      */
	
	return spim_query_finish(addr, MEMORY_PROGRAM_PAGE_LENGTH);
}

static bool memory_program_page(uint16_t address, uint8_t * data){
/*
  Programs the specified page with {MEMORY_PROGRAM_PAGE_CONTENT_LENGTH} bytes
  pointed to by {&data[MEMORY_PROGRAM_PAGE_CONTENT_OFFSET]}. The first
  {MEMORY_PROGRAM_PAGE_CONTENT_OFFSET} bytes pointed to by {data} may contain
  arbitrary values but will be overwritten by calling this function.
  
  Returns {true} in case of success. Returns {false} if the SPIM peripheral is
  too busy to process the generated query.
*/
	
	data[0] = 0;                               /* Reserved            */
	data[1] = (address & (1 << 15)) ? 1 : 0;   /* Chip ID             */
	data[2] = 0;                               /* Reception disabled  */
	data[3] = 1;                               /* WEL needed          */
	data[4] = 0x02;                            /* PP SPIM instruction */
	data[5] = (address >> 8) & 0x7F;           /* Address byte 2      */
	data[6] = address & 0xFF;                  /* Address byte 1      */
	data[7] = 0;                               /* Address byte 0      */
	
	if(!spim_query(data, MEMORY_PROGRAM_PAGE_LENGTH)){
		
		debug_report(DEBUG_MEMORY_BUFFER_OVERFLOW);
		return false;
	}
	else{
		
		return true;
	}
}

static uint8_t * memory_program_next_page_request(void){
/*
  Returns a RAM address where the next page to be programmed can be stored
  before it is finally programmed by {memory_program_next_page_finish()}. If an
  erase is necessary before progamming the page, it will be performed
  automatically.
  
  Returns {NULL} if the SPIM peripheral is too busy to process the necessary
  query or if the memory is already full.
  
  First {MEMORY_PROGRAM_PAGE_CONTENT_OFFSET} Bytes
  ================================================
  
    Note, that the first {MEMORY_PROGRAM_PAGE_CONTENT_OFFSET} bytes pointed to
    by the return value are reserved for internal use and hence shall not be
    written to by the user.
    
  Memory Write Warning
  ====================
    
    The user has to ensure that the total length, i.e. including the first
    {MEMORY_PROGRAM_PAGE_CONTENT_OFFSET} bytes, of the written data complies
    with the reserved memory per buffer entry as it is specified by
    {SPIM_BUFFER_WIDTH}.
*/
	
	if(memoryPageCount < MEMORY_NUM_OF_PAGES){
		
		if((memoryCurrentPage & 0x7f) == 0){
		
			/* The current page is the first page within a 32-kiB-block. */
			
			memory_erase_block32(memoryCurrentPage);
		}
		
		return memory_program_page_request();
	}
	else{
		
		/* Memory full. */
		return NULL;
	}
}

static bool memory_program_next_page_finish(uint8_t * addr){
/*
  Finishes the DMA-based programming procedure of the next free flash page that
  was initiated by calling {memory_program_next_page_request()}.
  
  Some Notes
  ==========
  
    This function does not perform any copying tasks. The specified {addr} is
    used for validation purposes only.
  
  Argument List
  =============
    
    addr       Pointer to the already written RAM as it was returned by the
               previous call to {memory_program_next_page_request()}.
  
  Return Value
  ============
    
    Returns {false} when illegal buffer activities, i.e. buffer activities
    between the call of {memory_program_next_page_request()} and the call to
    this function, have been detetcted. Returns {true} otherwise.
*/
	
	if(memory_program_page_finish(memoryCurrentPage, addr)){
		
		++memoryCurrentPage;
		++memoryPageCount;
		
		return true;
	}
	else{
		
		return false;
	}
}

uint8_t * memory_read(void){
/*
  Reads the page (i.e. {MEMORY_READ_PAGE_CONTENT_LENGTH} bytes) that was
  requested with the last call to {memory_request_address()},
  {memory_request_absolute()} or {memory_request_offset()}. The memory pointed
  to by the return value is guaranteed to hold its value until the next read
  request. Please make sure the request to be terminated before calling this
  function.
   
  Returns {NULL} if the CRC check failed. Otherwise the return value points
  to the requested page.
*/
	
	uint8_t * recvData = spim_recv();
	
	if(!crc_check(crc(
	              	(uint8_t *) &recvData[MEMORY_READ_PAGE_CONTENT_OFFSET],
	              	MEMORY_READ_PAGE_CONTENT_LENGTH
	              ))){
		
		debug_report(DEBUG_MEMORY_CRC_FAILED);
		return NULL;
	}
	else{
		
		return (uint8_t *) &recvData[MEMORY_READ_PAGE_CONTENT_OFFSET];
	}
}

bool memory_request_ready(void){
/*
  Returns {true} if the last read request (queried by
  {memory_request_address()}, {memory_request_absolute()} or
  {memory_request_offset()}) has terminated. Returns {false} if the last read
  request is still in process.
*/
	
	return spim_recv_ready();
}

bool memory_ready_to_request(void){
/*
  Returns {false} if the memory controller is too busy to process the request.
  Returns {true} otherwise.
*/
	
	if(spim_vacant_entries() > MEMORY_RESERVED_WRITE_ENTRIES && !memoryLocked){
		
		return true;
	}
	else{
		
		return false;
	}
}

static bool memory_request_address(uint16_t address){
/*
  Requests the page (i.e. {MEMORY_READ_PAGE_CONTENT_LENGTH} bytes) with the
  specified {address} from the FLASH memory.
  
  Returns {false} if the SPIM peripheral is too busy to process this request.
  Returns {true} otherwise. In the latter case, the status of the request can
  be polled by calling {memory_request_ready()}.
*/
	
	uint8_t query[MEMORY_READ_PAGE_LENGTH];
	
	query[0] = 0;                               /* Reserved              */
	query[1] = (address & (1 << 15)) ? 1 : 0;   /* Chip ID               */
	query[2] = 1;                               /* Enable reception      */
	query[3] = 0;                               /* WEL isn't needed      */
	query[4] = 0x03;                            /* NORD SPIM instruction */
	query[5] = (address >> 8) & 0x7F;           /* Address byte 2        */
	query[6] = address & 0xFF;                  /* Address byte 1        */
	query[7] = 0;                               /* Address byte 0        */

	if(!spim_query(query, MEMORY_READ_PAGE_LENGTH)){
		
		debug_report(DEBUG_MEMORY_BUFFER_OVERFLOW);
		return false;
	}
	else{
	
		return true;
	}
	
}

bool memory_request_absolute(uint32_t address){
/*
  Requests the page (i.e. {MEMORY_READ_PAGE_CONTENT_LENGTH} bytes) from the
  FLASH memory with the specified absolute {address}. Any mapping to real
  memory addresses is done automatically.
  
  Returns {false} if the SPIM peripheral is too busy to process this request.
  Returns {true} otherwise. In the latter case, the status of the request can
  be polled by calling {memory_request_ready()}.
*/
	
	if(!memoryLocked){
		
		/* Mapping is done by implicitly casting to {uint16_t}. */
		return memory_request_address(address);
	}
	else{
		
		return false;
	}

}

bool memory_request_offset(uint32_t offset){
/*
  Requests the specified page (i.e. {MEMORY_READ_PAGE_CONTENT_LENGTH} bytes)
  from the FLASH memory where {offset} specifies the page number counted from
  the first written page during this session (i.e. since the last call of
  {memory_init()}).
  
  Returns {false} if the SPIM peripheral is too busy to process this request.
  Returns {true} otherwise. In the latter case, the status of the request can
  be polled by calling {memory_request_ready()}.
*/
	
	if(!memoryLocked){
		
		/* Mapping is done by implicitly casting to {uint16_t}. */
		return memory_request_address(memoryFirstPage + offset);
	}
	else{
		
		return false;
	}
}

void memory_unlock(void){
/*
  Unlocks the memory for read requests.
*/
	
	memoryLocked = false;
}

void memory_lock(void){
/*
  Locks the memory for read requests.
*/
	
	memoryLocked = true;
}

uint32_t memory_get_page_count(void){
/*
  Returns the number of FLASH pages written during this session (i.e. since
  the last call to {memory_init()}).
*/
	
	return memoryPageCount;
}

uint32_t memory_get_start_address(void){
/*
  Returns the address of the first page that was written during this session
  (i.e. since the last call to {memory_init()}).
*/
	
	return memoryFirstPage;
}

uint8_t memory_get_load(void){
/*
  Returns the percentage of memory used during this session (i.e. since the
  last call to {memory_init()}).
*/
	
	return memoryPageCount * 100 / MEMORY_NUM_OF_PAGES;
}

MemoryTestStatus memory_full_test(void){
/*
  Runs a full memory test. Should be triggered until it stops returning
  {MEMORY_STATUS_ONGOING}.
  
  Finally returns {MEMORY_STATUS_SUCCEEDED} or {MEMORY_STATUS_FAILED}, if the
  memory test succeeded or failed, respectively.
*/
	
	static MemoryTestState testState = MEMORY_TEST_STATE_PREPARE;
	static uint32_t testPage = 0;
	static uint8_t progress = 0;
	static bool ok = false;
	
	/* Don't put those strings onto the stack. */
	static char progressMsg[] = "000% of flash memory verified.";
	static char failureMsg[] = "Page 0xghijk was corrupt.";
	
	uint32_t crcSum;
	uint8_t data[MEMORY_PROGRAM_PAGE_LENGTH];
	
	switch(testState){
		
		case MEMORY_TEST_STATE_PREPARE:
		
			debug("full test", "Full memory validation started.");
			testState = MEMORY_TEST_STATE_ERASE_1;
		
		break;
		case MEMORY_TEST_STATE_ERASE_1:
		
			if(memory_erase_chip((uint16_t) 0 << 15)){
				
				debug("full test", "Erase of first flash chip queued...");
				testState = MEMORY_TEST_STATE_ERASE_2;
			}
			
		break;
		case MEMORY_TEST_STATE_ERASE_2:
			
			if(memory_erase_chip((uint16_t) 1 << 15)){
			
				debug("full test", "Erase of second flash chip queued...");
				
				/* Reset status and progress variables. */
				ok = true;
				testPage = 0;
				progress = 255;
				testState = MEMORY_TEST_STATE_WRITE;
			}
			
		break;
		case MEMORY_TEST_STATE_WRITE:
		
			if(testPage == MEMORY_NUM_OF_PAGES){
				
				/* Tested all pages successfully! */
				
				testState = MEMORY_TEST_STATE_PREPARE;
				
				if(ok){
					
					debug("full test", "Memory validation succeeded.");
					return MEMORY_TEST_SUCCEEDED;
				}
				else{
					
					debug("full test", "Memory validation failed partially.");
					return MEMORY_TEST_FAILED;
				}
			}
			else{
				
				/* We have to test at least one more page... */
				
				if(spim_vacant_entries() > 1){
				
					random_page(
						&data[MEMORY_PROGRAM_PAGE_CONTENT_OFFSET],
						MEMORY_PROGRAM_PAGE_DATA_LENGTH
					);
					
					crcSum = crc(
					         	&data[MEMORY_PROGRAM_PAGE_CONTENT_OFFSET],
					         	MEMORY_PROGRAM_PAGE_DATA_LENGTH
					         );
					
					data[MEMORY_PROGRAM_PAGE_CRC_OFFSET + 0] = crcSum >> 24;
					data[MEMORY_PROGRAM_PAGE_CRC_OFFSET + 1] = crcSum >> 16;
					data[MEMORY_PROGRAM_PAGE_CRC_OFFSET + 2] = crcSum >>  8;
					data[MEMORY_PROGRAM_PAGE_CRC_OFFSET + 3] = crcSum >>  0;
					
					/* Program page. */
					if(memory_program_page(testPage, data)){
						
						testState = MEMORY_TEST_STATE_REQUEST;
					}
				}
			}
			
		break;
		case MEMORY_TEST_STATE_REQUEST:
			
			if(memory_request_absolute(testPage)){
			
				testState = MEMORY_TEST_STATE_READ;
			}
			
		break;
		case MEMORY_TEST_STATE_READ:
			
			if(memory_request_ready()){
				
				if(!memory_read()){
					
					/* CRC fail during page read. */
					
					print_uhex(testPage, failureMsg, 7, 5);
					debug("full test", failureMsg);
					ok = false;
				}
				
				/* Next page. */
				++testPage;
				
				if(testPage * 100 / MEMORY_NUM_OF_PAGES != progress){
			
					progress = testPage * 100 / MEMORY_NUM_OF_PAGES;
					print_udec(progress, progressMsg, 0, 3);
					debug("full test", progressMsg);
				}
				
				testState = MEMORY_TEST_STATE_WRITE;
			}
		
		break;
	}
	
	return MEMORY_TEST_ONGOING;
}

MemoryTestStatus memory_fast_test(void){
/*
  Runs a fast memory test, i.e. writes and reads a single flash page from every
  flash chip. Should be triggered until it stops returning
  {MEMORY_STATUS_ONGOING}.
  
  Finally returns {MEMORY_STATUS_SUCCEEDED} or {MEMORY_STATUS_FAILED}, if the
  memory test succeeded or failed, respectively.
*/
	
	static MemoryTestState testState = MEMORY_TEST_STATE_PREPARE;
	static uint8_t testChip;
	static uint16_t testAddress[MEMORY_NUM_OF_CHIPS];
	
	static bool ok = false;
	
	/* Don't put those strings onto the stack. */
	static char failureMsg[] = "Accessing flash chip x failed.";
	
	uint8_t i;
	uint32_t crcSum;
	uint8_t data[MEMORY_PROGRAM_PAGE_LENGTH];
	
	switch(testState){
		
		case MEMORY_TEST_STATE_PREPARE:
		
			debug("fast test", "Fast memory validation started.");
			
			/* Choose a page on the flash chips randomly. */
			
			for(i = 0; i < MEMORY_NUM_OF_CHIPS; ++i){
			
				testAddress[i] = ((uint16_t) random_byte() << 7)  /* Page */
				               | ((uint16_t) 0 << 15);            /* Chip */
			}
			
			testState = MEMORY_TEST_STATE_ERASE_1;
			
		break;
		case MEMORY_TEST_STATE_ERASE_1:
			
			if(memory_erase_block32(testAddress[0])){
				
				testState = MEMORY_TEST_STATE_ERASE_2;
			}
		
		break;
		case MEMORY_TEST_STATE_ERASE_2:
				
			if(memory_erase_block32(testAddress[1])){
				
				/* Reset status and progress variables. */
				ok = true;
				testChip = 0;
				
				testState = MEMORY_TEST_STATE_WRITE;
			}
			
		break;
		case MEMORY_TEST_STATE_WRITE:
		
			if(testChip == MEMORY_NUM_OF_CHIPS){
				
				/* Sucessfully tested both flash chips. */
				
				testState = MEMORY_TEST_STATE_PREPARE;
				
				if(ok){
					
					debug("fast test", "Fast memory validation succeeded.");
					return MEMORY_TEST_SUCCEEDED;
				}
				else{
					
					debug("fast test", "Fast memory validation failed partially.");
					return MEMORY_TEST_FAILED;
				}
			}
			else{
				
				/* There is a least one more chip to be tested. */
				
				if(spim_vacant_entries() > 1){
				
					random_page(
						&data[MEMORY_PROGRAM_PAGE_CONTENT_OFFSET],
						MEMORY_PROGRAM_PAGE_DATA_LENGTH
					);
					
					crcSum = crc(
					         	&data[MEMORY_PROGRAM_PAGE_CONTENT_OFFSET],
					         	MEMORY_PROGRAM_PAGE_DATA_LENGTH
					         );
					
					data[MEMORY_PROGRAM_PAGE_CRC_OFFSET + 0] = crcSum >> 24;
					data[MEMORY_PROGRAM_PAGE_CRC_OFFSET + 1] = crcSum >> 16;
					data[MEMORY_PROGRAM_PAGE_CRC_OFFSET + 2] = crcSum >>  8;
					data[MEMORY_PROGRAM_PAGE_CRC_OFFSET + 3] = crcSum >>  0;
					
					/* Program that page. */
					
					if(memory_program_page(testAddress[testChip], data)){
						
						testState = MEMORY_TEST_STATE_REQUEST;
					}
				}
			}
			
		break;
		case MEMORY_TEST_STATE_REQUEST:
			
			if(memory_request_absolute(testAddress[testChip])){
			
				testState = MEMORY_TEST_STATE_READ;
			}
			
		break;
		case MEMORY_TEST_STATE_READ:
			
			if(memory_request_ready()){
				
				if(!memory_read()){
					
					/* CRC error during page read. */
					
					print_uhex(testChip, failureMsg, 21, 1);
					debug("fast test", failureMsg);
					ok = false;
				}
				
				/* Next chip. */
				++testChip;
				
				testState = MEMORY_TEST_STATE_WRITE;
			}
		
		break;
	}
	
	return MEMORY_TEST_ONGOING;
}

static inline void memory_measure_write_results_on_page_insert_index(
	uint16_t * ptr,
	uint8_t * data,
	uint32_t * crc,
	uint32_t index,
	uint32_t * expectedIndex
){
/*
  Inserts the specified {index} on the specified page pointed to by {data} at
  position pointed to by {ptr} and updates the specified {crc} sum. The {index}
  is the {*expectedIndex} of the measurement results that are stored in the
  following bytes on this page. The {*expectedIndex} gets updated accordingly.
*/
	
	data[(*ptr)++] = index >> 24;
	data[(*ptr)++] = index >> 16;
	data[(*ptr)++] = index >>  8;
	data[(*ptr)++] = index >>  0;
	
	crc_append(crc, index >> 24);
	crc_append(crc, index >> 16);
	crc_append(crc, index >>  8);
	crc_append(crc, index >>  0);
	
	*expectedIndex = index;
}

static void memory_measure_write_results_on_page_init(
	uint16_t * ptr,
	uint8_t * data,
	uint32_t * crc,
	uint32_t index,
	uint32_t * expectedIndex
){
/*
  Initializes a page pointed to by {data} and its corresponding CRC sum pointed
  to by {crc} with the specified {index}. This {index} is the {*expectedIndex}
  of the measurement results that are stored in the following bytes on this
  page. The {*expectedIndex} gets updated accordingly.
  
  Argument List
  =============
    
    {ptr}            Page write pointer.
    {data}           The page to be written on.
    {crc}            Temporary CRC sum.
    {index}          Index to be stored on the beginning of the page.
    {expectedIndex}  Pointer to the expected index, see above.
*/
	
	crc_reset(crc);
	memory_measure_write_results_on_page_insert_index(
		ptr,
		data,
		crc,
		index,
		expectedIndex
	);
}

static inline void memory_measure_write_results_on_page_fill_missing(
	uint16_t * ptr,
	uint8_t * data,
	uint32_t * crc,
	uint32_t * expectedIndex
){
/*
  Inserts a {CODER_FORMAT_BYTE_COMMAND_NEXT_INDEX} on the specified page pointed
  to by {data} at position pointed to by {ptr} and updates the specified {crc}
  sum as well as the {*expectedIndex}.
*/
	
	data[(*ptr)++] = CODER_FORMAT_BYTE_COMMAND_NEXT_INDEX;
	crc_append(crc, CODER_FORMAT_BYTE_COMMAND_NEXT_INDEX);
	++(*expectedIndex);
}

static bool memory_measure_write_results_on_page_append(

	uint16_t * ptr,
	uint8_t * data,
	uint32_t * crc,
	uint32_t index,
	uint32_t * expectedIndex,
	CoderFormatSpecifier fmtSpec,
	uint8_t sheetIndex,
	uint8_t * result,
	uint8_t size
){
/*
  Writes {size} bytes from {result} with the specified {index} as well as the
  corresponding {fmtSpec} onto a page pointed to by {data} and updates the
  corresponding CRC sum pointed to by {crc}.
  
  The {sheetIndex} specifies the bytes from {result} that are written onto the
  flash page. Only the {size} bytes {result[4*i + sheetIndex]}
  for {i} = 0, 1, ..., {size} - 1 are considered.
  
  Index Storage
  =============
    
    Consecutive measurement results may be stored in the memory without
    re-specifying the index everytime. When there is a gap between the {index}
    of the current result and the previous one, it may though be necessary to
    specify the index again. In order to do so, this function needs to know
    the size of this gap, i.e. the difference between the current {index} and
    the consecutively {expectedIndex}.
    
    Depending on the {fmtSpec}, the {expectedIndex} is set to {index + 1} or
    {index} after the specified measurement {result} has been written onto the
    flash page.
  
  Argument List
  =============
    
    {ptr}            Page write pointer.
    {data}           The page to be written on.
    {crc}            Temporary CRC sum.
    {index}          Index of the measurement results.
    {expectedIndex}  Expected index, see above.
    {fmt}            The format specifier of the measurement result.
    {sheetIndex}     Which bytes from {result} to write onto the page:
                       
                       {result[4*i + sheetIndex]}
                       
                     for {i} = 0, 1, ..., {size} - 1.
    {result}         Data to be written. Should point to a portion of memory
                     with a size of a least {size * 4}.
    {size}           Number of bytes from {result} to be written.
  
  
  Return Value
  ============
  
    Returns {true} if the measurement results were written on the page
    successfully. Returns {false} if the page has not enough free space for the
    measurement result to be written.
*/
	
	uint8_t fmtByteSize;
	CoderFormatByte fmtByte;
	uint8_t i;
	int32_t diff = (int32_t) index - (int32_t) *expectedIndex;
	
	if(diff){
		
		/* There is a discrepancy between the    */
		/* specified index and the expected one. */
		
		if(diff >= 1 && diff <= 4){
			
			/* Fill up to 4 missing indices with       */
			/* {CODER_FORMAT_BYTE_COMMAND_NEXT_INDEX}. */
			
			while(MEMORY_PROGRAM_PAGE_CRC_OFFSET - *ptr >= 1 && diff){
			
				memory_measure_write_results_on_page_fill_missing(
					ptr,
					data,
					crc,
					expectedIndex
				);
				--diff;
			}
		}
		else{
			
			/* We have to re-specify the index completely. */
			
			if(MEMORY_PROGRAM_PAGE_CRC_OFFSET - *ptr >= 5){
					
				data[(*ptr)++] = CODER_FORMAT_BYTE_COMMAND_NEW_INDEX;
				crc_append(crc, CODER_FORMAT_BYTE_COMMAND_NEW_INDEX);
				memory_measure_write_results_on_page_insert_index(
					ptr,
					data,
					crc,
					index,
					expectedIndex
				);
				diff = 0;
			}
		}
	}
	
	fmtByteSize = coder_explicit_format_byte_present(fmtSpec)
	            ? 1
	            : 0;
	
	/* Check whether the specified size matches */
	/* the specified format specifier.          */
	if(size != coder_result_size(fmtSpec)){
		
		debug_report(DEBUG_MEMORY_INCONSISTENT_SIZE);
	}
	
	if(!diff && MEMORY_PROGRAM_PAGE_CRC_OFFSET - *ptr >= size + fmtByteSize){
		
		/* We have enough space left on this page */
		
		if(fmtByteSize){
			
			/* We have an extra format byte prior */
			/* to the actual measurement result.  */
			
			fmtByte = coder_format_byte(fmtSpec, sheetIndex);
			data[(*ptr)++] = fmtByte;
			crc_append(crc, fmtByte);
		}
		else{
			
			/* The first result ({result[sheetIndex]}) IS the format byte. */
			/* Make sure this format byte is valid.                        */
			result[sheetIndex] = coder_truncate_valid_rss(result[sheetIndex]);
			fmtByte = result[sheetIndex];
		}
		
		/* Save the specified bytes from {result} onto flash page. */
		for(i = 0; i < size; ++i){
		
			data[(*ptr)++] = result[4*i + sheetIndex];
			crc_append(crc, result[4*i + sheetIndex]);
		}
		
		/* Update expected index. */
		*expectedIndex = coder_result_increments_expected_index(fmtByte)
		               ? index + 1
		               : index;
		
		return true;
	}
	else{
		
		/* We do NOT have enough space left on the page. */
		
		return false;
	}
}

static void memory_measure_write_results_on_page_finish(uint16_t * ptr,
                                                        uint8_t * data,
                                                        uint32_t * crc){
/*
  Finishes the page pointed to by {data}.
  
  Argument List
  =============
    
    ptr      Page write pointer.
    data     The page to be written on.
    crc      Temporary CRC sum.
*/
	
	if(MEMORY_PROGRAM_PAGE_CRC_OFFSET - *ptr > 0){
		
		data[(*ptr)++] = CODER_FORMAT_BYTE_COMMAND_END_OF_PAGE;
		crc_append(crc, CODER_FORMAT_BYTE_COMMAND_END_OF_PAGE);
	}
	
	crc_append_zeros(crc, MEMORY_PROGRAM_PAGE_CRC_OFFSET - *ptr);
	while(MEMORY_PROGRAM_PAGE_CRC_OFFSET - *ptr > 0){
		
		data[(*ptr)++] = 0;
	}
	
	crc_finish(crc);
	
	data[MEMORY_PROGRAM_PAGE_CRC_OFFSET + 0] = (*crc) >> 24;
	data[MEMORY_PROGRAM_PAGE_CRC_OFFSET + 1] = (*crc) >> 16;
	data[MEMORY_PROGRAM_PAGE_CRC_OFFSET + 2] = (*crc) >>  8;
	data[MEMORY_PROGRAM_PAGE_CRC_OFFSET + 3] = (*crc) >>  0;
}

bool memory_measure_write_results(
	uint32_t index,
	uint8_t rssIntegerPart,
	uint8_t rssFractionalPart,
	CoderFormatSpecifier fmtSpec,
	uint8_t * samples,
	uint8_t size,
	bool reset,
	bool finish
){
/*
  Stores the given measurement results with the given {index} onto the flash
  memory. Page management, error management as well as redundancy checks are
  done/applied automatically.
  
  Measurement Result
  ==================
    
    A measurement result consist of two bytes {rssIntegerPart} and
    {rssFractionalPart} as well as of {size} words, i.e. {size*4} bytes,
    pointed to by {samples}. The format of these {samples} is specified by
    {fmtSpec}.
    
  Argument List
  =============
    
    {index}                Index of the measurement value.
    {rssIntegerPart}       Integer part of the average RSS.
    {rssFractionalPart}    Fractional part of the average RSS.
    {fmtSpec}              The {CoderFormatSpecifier} of the given {samples}.
                           May be {CODER_FORMAT_SPECIFIER_NONE}. In this case
                           the {samples} are ignored.
    {samples}              {size*4} samples to be stored.
    {size}                 Number of sample words.
    {reset}                Whether to reset the internal state machine as well
                           as memory locking. All other arguments are ignored
                           in this case.
    {finish}               Whether to finish the currently written flash page,
                           i.e. finalizing it. All other arguments, except
                           reset, see below, are ignored in this case.
    
  Repeated Calls
  ==============
  
    It may be necessary to call this function multiple times in order to finish
    the write process. In order to initiate the storage of the results, all
    arguments must be specified as it is listed below. If the write process
    has not finished after return, i.e. if this function returns {false}, see
    below, this function has to be called at least one more time. For such
    additional calls (so-called follow-up calls) it is not necessary to provide
    all listed arguments again. Indeed they may have arbitrary values. It is,
    however, recommended to specify zeros or equivalent values.
    
    However, the {samples}, whose address was specified during the initial
    call of this function, must stay in memory for the whole time of writing.
    
    When this function (finally) returns {true}, the write process finished.
    The next write process may now be initiated as it is described above.
  
  Indices
  =======
    
    Subsequent (init) calls of this function should always increase the
    specified index. An error is reported via the debug module otherwise.
  
  Reset and Finish
  ================
    
    The internal state machine may be resetted by specifiying arbitrary
    arguments expect {reset}. The currently written page may be finished, i.e.
    finalized, by specifying arbitrary arguments expect {finish}. If both
    {reset} and {finish} are given {reset} wins.
  
  Memory Locking
  ==============
    
    For the time of the write process, i.e. between the init call and the
    final follow-up call, the memory is locked, i.e. no reading tasks can be
    initiated or requested. When the internal state machine is resetted, the
    memory lock is removed, too.
  
  Return Value
  ============
    
    It may be necessary to call this function multiple times in order to finish
    the write process, see above. The return value indicates whether this
    function has to be called at least one more time ({false}) or whether the
    writing process finished during the current call ({true}).
*/
	
	typedef enum {
		
		STATE_IDLE          = 0,
		STATE_WRITE_SHEET_0 = 1,
		STATE_WRITE_SHEET_1 = 2,
		STATE_WRITE_SHEET_2 = 3,
		STATE_WRITE_SHEET_3 = 4,
		STATE_WRITE_RSS     = 5,
	
	} State;
	
	_Static_assert(
		   STATE_WRITE_SHEET_0 == 1
		&& STATE_WRITE_SHEET_1 == STATE_WRITE_SHEET_0 + 1
		&& STATE_WRITE_SHEET_2 == STATE_WRITE_SHEET_1 + 1
		&& STATE_WRITE_SHEET_3 == STATE_WRITE_SHEET_2 + 1,
		"State values may be not as expected."
	);
	
	/* State machine. */
	static State state = STATE_IDLE;
	
	/* Page management. */
	static uint32_t nextIndex = 0;
	static uint16_t ptr = MEMORY_PROGRAM_PAGE_CONTENT_OFFSET;
	static uint32_t crc;
	static uint8_t * data = NULL;
	
	/* Remembering given data for subsequent (follow-up) calls. */
	static uint32_t cachedIndex;
	static uint8_t cachedRSS[5]; /* Place for two bytes on sheet index 0. */
	static CoderFormatSpecifier cachedFmtSpec;
	static uint8_t * cachedSamples;
	static uint8_t cachedSize;
	
	/* Whether the result could be written on a page or not (page full). */
	bool wrote = false;
	
	if(reset){
		
		/* Reset internal state machine. */
		
		nextIndex = 0;
		ptr = MEMORY_PROGRAM_PAGE_CONTENT_OFFSET;
		state = STATE_IDLE;
		memoryLocked = false;
		
		return true;
	}
	else{
		
		/* Store measurement results. */
		
		if(state == STATE_IDLE && !finish){
			
			/* Remember the given data for subsequent (follow-up) calls. */
			
			if(rssIntegerPart == 0 || (samples == NULL && size != 0)){
				
				/* This seems to be a follow-up call of a multi-call process */
				/* but the state machine expects the entry call.             */
				
				debug_report(DEBUG_MEMORY_INVALID_STATE);
				
				/* Set some default values. */
				cachedIndex = 0;
				cachedRSS[0] = 0;
				cachedRSS[4] = 0;
				cachedFmtSpec = CODER_FORMAT_SPECIFIER_NONE;
				cachedSamples = NULL;
				cachedSize = 0;
			}
			else{
				
				cachedIndex = index;
				cachedRSS[0] = rssIntegerPart;
				cachedRSS[4] = rssFractionalPart;
				cachedFmtSpec = fmtSpec;
				cachedSamples = samples;
				cachedSize = size;
				
				if(cachedIndex < nextIndex){
					
					/* This should never happen. */
					debug_report(DEBUG_MEMORY_INDEX_ERROR);
				}
			}
			
			state = STATE_WRITE_SHEET_0;
		}
		
		if(ptr == MEMORY_PROGRAM_PAGE_CONTENT_OFFSET){
			
			/* This is the beginning of a new page */
			
			/* Disallow parallel read requests. */
			memoryLocked = true;
			
			data = memory_program_next_page_request();
			
			if(data){
			
				memory_measure_write_results_on_page_init(
					&ptr, data, &crc, cachedIndex, &nextIndex
				);
			}
		}
		
		if(data){
			
			if(!finish){
				if(state < STATE_WRITE_RSS){
					
					/* Write sheet 0, 1, 2 or 3 of the measurement   */
					/* results (samples) onto the flash memory page. */
					
					if(cachedFmtSpec != CODER_FORMAT_SPECIFIER_NONE){
						
						/* Store a single page. */
						
						/* This updates {nextIndex} automatically. */
						wrote = memory_measure_write_results_on_page_append(
							    	&ptr,
							    	data,
							    	&crc,
							    	cachedIndex,
							    	&nextIndex,
							    	cachedFmtSpec,
							    	state - 1,
							    	cachedSamples,
							    	cachedSize
							    );
						
						if(wrote){
							state += 1;
						}
					}
					else{
						
						/* No samples given. Continue with   */
						/* storing the RSS at the next call. */
						
						state = STATE_WRITE_RSS;
						/* Not writing at all is some kind of success. */
						wrote = true;
					}
				}
				else{
					
					/* Store RSS values onto flash memory page. */
					
					/* This updates {nextIndex} automatically. */
					wrote = memory_measure_write_results_on_page_append(
						    	&ptr,
						    	data,
						    	&crc,
						    	cachedIndex,
						    	&nextIndex,
						    	CODER_FORMAT_SPECIFIER_VALID_RSS,
						    	0, /* Just LSBs from {cachedRSS} are used. */
						    	cachedRSS,
						    	2
						    );
					
					if(wrote){
						
						/* Now, we may process another measurement result.  */
						/* at the next call to this function, i.e. the next */
						/* call of this function should be an entry call.   */
						state = STATE_IDLE;
					}
				}
			}
			
			if(!wrote){
				
				/* No space left on the current flash page.     */
				/* (Or page {finish}ing was requested.)         */
				/* Write flash page onto flash memory and       */
				/* continue with the next step at the next call.*/
				
				/* Write flash page into flash memory. */
				memory_measure_write_results_on_page_finish(&ptr, data, &crc);
				memory_program_next_page_finish(data);
				
				/* Reset page pointer and unlock memory. */
				ptr = MEMORY_PROGRAM_PAGE_CONTENT_OFFSET;
				memoryLocked = false;
			}
			
			if(state == STATE_IDLE){
				
				/* Completly saved all {samples} and the {RSS*} values. */
				return true;
			}
			else{
				
				/* There is still something to be stored on flash memory. */
				return false;
			}
		}
		else{
			
			if(memory_get_load() < 100){
			
				/* Retry. */
				return false;
			}
			else{
				
				/* Memory is full. Retrying is useless. */
				return true;
			}
		}
	}
}

void memory_finish(void){
/*
  Finishes, i.e. finalizes, the currently writte flash page.
*/
	
	memory_measure_write_results(
		0,
		0, 0,
		CODER_FORMAT_SPECIFIER_NONE, NULL, 0,
		false,
		true
	);
}

void memory_reset(void){
/*
  Resets the memory to a blank session. This function does NOT reset the
  circular buffer's start page randomly but instead writes new measurement
  values at the beginning of the next 32 kiB block right after the
  the memory locations of the previous session.
*/
	
	/* Reset the page that may be partially written. */
	memory_measure_write_results(
		0,
		0, 0,
		CODER_FORMAT_SPECIFIER_NONE, NULL, 0,
		true,
		false
	);
	
	/* Reset session. */
	
	/* Next 32 kiB block. */
	memoryCurrentPage = (memoryCurrentPage + 0x7f) & 0xffffff80;
	
	memoryFirstPage = memoryCurrentPage;
	memoryPageCount = 0;
	
	debug("reset", "Reset memory session.");
}

void memory_init(void){
/*
  Initializes this module. May be called multiple times.
*/

	static uint8_t spimWIPQuery[] = {0x05, 0x00};
	static uint8_t spimWELQuery[] = {0x06};
	static size_t  spimWIPSize    = 2;
	static size_t  spimWELSize    = 1;
	static bool initialized = false;
	
	if(!initialized){
	
		spim_init(spimWELQuery, spimWELSize, spimWIPQuery, spimWIPSize, &memory_spim_check_wip);
	
		memoryFirstPage = (random_byte() << 8) | (random_byte() & 0x80);
		memoryCurrentPage = memoryFirstPage;
		
		memory_reset();
	
		initialized = true;
	}
}

