/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

#define RSS2I_MODULE "mlc"

/*
  Main loop counter: Monitors the frequency of the main loop.
  
  Authors
  =======
  
    Copyright (C) 2018-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
  
    2019-06-14    Initial release.
    2019-10-31    Introduced {MLC_THRESH} and increased this threshold to {100}.
    2020-02-12    Print current mlc value onto serial debug output.
    2020-03-31    Added {mlc_get()}.
    2020-07-06    Added support for different {DEBUG_LEVEL}s.
    2021-02-25    Added BLE 2MBit prefixes.
  
  Description
  ===========

    Since there is a lot of asynchronous stuff to do, it is necessary that the
    main loop triggers all processing/handling tasks with an adequate frequency.
    After initialization using {mlc_init()} this module should be triggered
    ({mlc_trigger()}) once per main loop cycle. The rest (monitoring, reporting,
    warning) is done automatically.
    
    After introduction of the scheduler module, the mlc module has become a bit
    redundant.
*/


#include <stdbool.h>
#include <stdint.h>

#include <rss2i/timer.h>
#include <rss2i/print.h>
#include <rss2i/debug.h>

#include <rss2i/config/debug.h>
#include <rss2i/config/mlc.h>
#include <rss2i/config/measure.h>


static volatile uint32_t mlcCounter = 0;
static volatile uint32_t mlc;
static volatile bool mlcUpdated = false;


static void mlc_update(void){
/*
  Makes a backup of {mlcCounter} and sets the {mlcUpdated} flag.
*/
	
	mlc = mlcCounter;
	mlcCounter = 0;
	mlcUpdated = true;
}

uint16_t mlc_get(void){
/*
  Return the current number of main loop cycles per measurement frame.
  (API for cmd)
*/
	
	return mlc * 10
	     / ((uint32_t) MLC_PERIOD * 1000
	     / MEASURE_X_2MBIT_FRAME_LENGTH);
}

void mlc_init(void){
/*
  Initializes this module. May be called multiple times.
*/
	
	static bool initialized = false;
	
	if(!initialized){
		
		debug_init();
		
		timer_call(mlc_update, MLC_PERIOD, TIMER_REPEAT);
		
		initialized = true;
	}
}

void mlc_trigger(void){
	
	uint16_t mlcPMF;
	
	#if DEBUG_LEVEL >= 2
		static char m[] = "  .0 async processing loops per measurement frame.";
	#endif
	
	if(mlcUpdated){
		
		mlcPMF = mlc * 10
		       / ((uint32_t) MLC_PERIOD * 1000 / MEASURE_X_2MBIT_FRAME_LENGTH);
		
		#if DEBUG_LEVEL >= 2
			print_udec(mlcPMF / 10, m, 0, 2);
			m[3] = '0' + (mlcPMF % 10);
			debug("updated", m);
		#endif
		
		if(mlcPMF < MLC_THRESH){
			
			/* Less than {MLC_THRESH} main loop cycles per measurement frame. */
			debug_report(DEBUG_MLC_TOO_BUSY);
		}
		
		mlcUpdated = false;
	}
	
	++mlcCounter;
}

