/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

#define RSS2I_MODULE "postproc"

/*
  Measurement data post processing.
  
  Authors
  =======
  
    Copyright (C) 2020-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
  
    2020-08-14    Initial release.
    2020-08-24    Introduce raw IQ sample requests.
    2020-08-25    Wrap around {sheetIndex} if necessary.
    2021-02-25    Added BLE 2MBit prefixes.
    2021-03-18    Added {radioMode} and {sampleType} args to {postprocess()}.
    2021-03-31    Outsourced storing measurement results to postproc module.
    2021-04-14    Support multi-call {memory_measure_write_results()}.
    2021-04-15    Added the {store} argument to {postprocess()}.
    2021-04-19    Renamed {postproc_cte_sample*()} to {postproc_sample*()}.
    2021-04-29    Fit {memory_wirte_results()} call to changes within memory.c.
    2021-04-29    Support {CTE_2US_7PX1} {CTEType}.
    
  Description
  ===========
  
    This is the place where all the measurement data post processing tasks are
    bundled. The measurement module should call the {postprocess()} function
    whenever measurement data is available. This function will then perform:
      
      - Store RSS data into the flash memory.
      - Store raw sample data into the flash memory (if any).
      - Prepare raw samples to be sent via USB (if any and if requested).
      - Some consistency checks.
*/

#include <stdint.h>
#include <stdbool.h>
#include <stddef.h>

#include <rss2i/measure.h>
#include <rss2i/cte.h>
#include <rss2i/debug.h>
#include <rss2i/radio.h>
#include <rss2i/hook.h>
#include <rss2i/memory.h>
#include <rss2i/coder.h>
#include <rss2i/postproc.h>

#include <rss2i/config/cte.h>


#define POSTPROC_MAX_NUMBER_OF_32_BIT_SAMPLES 248


/* Assert that the sample amount fits into a single byte and that the */
/* maximal number of CTE IQ samples fit into the postproc buffer.     */
_Static_assert(
	   POSTPROC_MAX_NUMBER_OF_32_BIT_SAMPLES >= CTE_MAX_NUMBER_OF_SAMPLES
	&& POSTPROC_MAX_NUMBER_OF_32_BIT_SAMPLES <= 256,
	"There may be too less buffer memory for the postproc module."
);


typedef enum {
	
	POSTPROC_SAMPLE_STATUS_IDLE,
	POSTPROC_SAMPLE_STATUS_ONGOING,
	POSTPROC_SAMPLE_STATUS_READY
	
} PostprocSampleStatus;


PostprocSampleStatus postprocSampleStat = POSTPROC_SAMPLE_STATUS_IDLE;
uint8_t postprocSampleFrame;
PostprocFormat postprocSampleFormat;
uint8_t postprocSampleAmount;
uint8_t postprocSample[POSTPROC_MAX_NUMBER_OF_32_BIT_SAMPLES * 4];


uint8_t postproc_sample_request(uint8_t frame){
/*
  Requests to store raw samples data during the next occurence of the
  specified {frame}. If {frame} equals {255}, the raw samples will be collected
  as soon as possible during an arbitrary frame.
  
  Return Value
  ============
  
    {0}    OK. Request accepted.
    {1}    Failed. There already is an ongoing sample request.
    {2}    Failed. No ongoing measurement.
    {3}    Failed. No samples available due to the current configuration.
    {4}    Failed. The specified {frame} is out-of-range.
    {5}    Failed. The specified {frame} is no receiving frame.
    {6}    Failed. There is no sampling during the specified {frame}.
*/
	
	if(postprocSampleStat != POSTPROC_SAMPLE_STATUS_IDLE){
		
		return 1;
	}
	else if(!measure_running()){
		
		return 2;
	}
	else if(!measure_any_sampling()){
		
		return 3;
	}
	else if(frame != 255 && !measure_frame_in_range(frame)){
		
		return 4;
	}
	else if(frame != 255 && !measure_is_rx_frame(frame)){
		
		return 5;
	}
	else if(frame != 255 && !measure_is_sampling_frame(frame)){
		
		return 6;
	}
	else{
		
		postprocSampleFrame = frame;
		postprocSampleStat = POSTPROC_SAMPLE_STATUS_ONGOING;
		
		return 0;
	}
}

void postproc_sample_clear_request(void){
/*
  Clears the ongoing (or even ready) sample request (if any).
*/
	
	postprocSampleStat = POSTPROC_SAMPLE_STATUS_IDLE;
}

uint8_t postproc_sample_status(void){
/*
  Returns the status of the current sample request.
  
  Return Value
  ============
    
    {0}    Sample ready to be read.
    {1}    Sample request still ongoing.
    {2}    No sample requested yet.
    {3}    Unknown error.
*/
	
	switch(postprocSampleStat){
		
		case POSTPROC_SAMPLE_STATUS_IDLE:
		
			return 2;
		
		break;
		case POSTPROC_SAMPLE_STATUS_ONGOING:
			
			return 1;
		
		break;
		case POSTPROC_SAMPLE_STATUS_READY:
			
			return 0;
		
		break;
		default:
			
			return 3;
	}
}

uint8_t postproc_sample_sheet(uint8_t sheetIndex,
                              uint8_t * framePtr,
                              uint8_t * formatPtr,
                              uint8_t * sheet){
/*
  Returns the recently sampled IQ data.
  
  Argument Overview
  =================
    
    {sheetIndex}    The IQ data is organized into so called sheets. There may be
                    up to four of such sheets available, see below.
    {framePtr}      The IQ samples were recorded during the frame whose index
                    is pointed to by {framePtr}.
    {formatPtr}     The structure of those sheets depends on the returned format
                    to which {formatPtr} points to after return.
    {sheet}         The actual IQ data sample sheet address to which the data is
                    written.
  
  Sheet Format
  ============
    
    General
    -------
    
     A bunch of raw samples contains N data words, i.e. 4N bytes. The given
     {sheetIndex} specifies which byte of each data word should be written to
     the given {sheet} address. The first sheet ({sheetIndex == 0}) corresponds
     to the LSB bytes, the last sheet ({sheetIndex == 3}) belongs to the MSB
     byte of the samples.
     
     In other words: All sheets contain N bytes. The first byte corresponds to
     the first sample taken, the second byte corresponds to the second sample
     taken and so on.... The first (LSB) byte of a four-byte sample (data word)
     is available in the first ({sheetIndex == 0}) sheet. The last (MSB) byte is
     available in the last ({sheetIndex == 3}) sheet.
     
     The format, i.e. the meaning of the individual data words, solely depends
     on the format specifier which {formatPtr} points to after return. See the
     documentation of the coder module for more information about formats and
     sheet indices.
    
    Format 0
    --------
    
      After a successfull sample request, there will be
      {CTE_2US_7P1_NUMBER_OF_SAMPLES} samples (data words) available each of
      which is a four byte value with the following structure:
    
         ________________________________ ________________________________
        |                                |                                |
        |31                            16|15                             0|
        |________________________________|________________________________|
        |                                |                                |
        |    Q data (two's complement)   |    I data (two's complement)   |
        |   in the range [-4096, 4095]   |   in the range [-4096, 4095]   |
        |________________________________|________________________________|
    
    
  Return Value
  ============
    
    {0}    OK.
    {1}    Sample request still ongoing.
    {2}    No sample requested yet.
    {3}    Unknown error.
*/
	
	uint8_t stat;
	uint8_t i;
	
	stat = postproc_sample_status();
	sheetIndex %= 4;
	
	if(stat == 0){
		
		*framePtr = postprocSampleFrame;
		*formatPtr = postprocSampleFormat;
		
		for(i = 0; i < postprocSampleAmount; ++i){
			
			sheet[i] = postprocSample[4*i + sheetIndex];
		}
		
		return 0;
	}
	else{
		
		return stat;
	}
}

bool postprocess(uint8_t frame,
                 int32_t cycle,
                 uint8_t frameCount,
                 CTEType cteType,
                 CTEAntennaPattern cteAntPat,
                 uint8_t rssAvgIntegerPart,
                 uint8_t rssAvgFractionalPart,
                 uint8_t rssAmount,
                 uint8_t sampleAmount,
                 RadioMode radioMode,
                 MeasureBufferSampleType sampleType,
                 bool storeSamples,
                 uint8_t * radioPacket){
/*
  This function should be called whenever measurement data is available. It will
  then perform:
    
    - Store RSS data into the flash memory.
    - Store raw sample data into the flash memory (if any).
    - Prepare raw samples to be sent via USB (if any and if requested).
    - TODO Some consistency checks.
    
  Repeated Calls
  ==============
  
    It may be necessary to call this function multiple times in order to finish
    the post processing. In order to initiate a post processing task, all
    arguments must be specified as it is listed below. If the post processing
    has not finished after return, i.e. if this function returns {false}, see
    below, this function has to be called at least one more time. For such
    additional calls (so-called follow-up calls) it is not necessary to provide
    all listed arguments again. Indeed they may have arbitrary values. It is,
    however, recommended to specify zeros or equivalent values.
    
    However, the radio packet, whose address was specified during the initial
    call of this function, must stay in memory for the whole time of post
    processing.
    
    When this function (finally) returns {true}, the post processing finished.
    The next post processing task may now be initiated as it is described above.
  
  Argument Overview
  =================
    
    {frame}                   Measurement value's frame index.
    {cycle}                   Measurement value's cycle index.
    {frameCount}              Number of measurement frames per cycle.
    {cteType}                 {CTEType} of the current measurement.
    {cteAntPat}               Corresponding {frame}'s {CTEAntennaPattern}.
    {rssAvgIntegerPart}       Integer part of the measured RSS
    {rssAvgFractionalPart}    Fractional part of the measured RSS
    {rssAmount}               Number of averaged raw RSS values.
    {sampleAmount}            Number of raw samples available for post
                              processing.
    {radioMode}               Used radioMode during packet reception.
    {sampleType}              Type of possible samples inside the radio packet.
    {storeSamples}            Whether to store raw samples on the flash or not.
    {radioPacket}             Address of the radio packet during which the
                              measurement value was taken.
    
  Return Value
  ============
    
    It may be necessary to call this function multiple times in order to finish
    the post processing, see above. The return value indicates whether this
    function has to be called at least one more time ({false}) or whether the
    post processing finished during the current call ({true}).
*/
	
	typedef enum {
		
		STATE_READY,
		STATE_WAITING_FOR_MEMORY
	
	} State;
	
	static State state = STATE_READY;
	
	size_t i;
	/* The (CTE or raw RSS) samples are stored on a byte buffer      */
	/* even if they may contain word (32 bit) data (CTE IQ samples). */
	/* The sample amount always refers to the number of actual       */
	/* samples and not to the number of bytes.                       */
	uint8_t * samples;
	uint8_t expectedAmount;
	PostprocFormat fmt;
	CoderFormatSpecifier fmtSpec;
	bool ready = false;
	
	
	HOOK_UNUSED(rssAmount);
	HOOK_UNUSED(radioMode);
	HOOK_UNUSED(cteAntPat);
	
	
	if(state == STATE_READY){
	
		if(frameCount == 0
		|| rssAvgIntegerPart == 0
		|| rssAmount == 0
		|| radioPacket == NULL){
			
			/* This seems to be a follow-up call of a multi-call process */
			/* but the state machine expects the entry call.             */
			
			debug_report(DEBUG_POSTPROC_INVALID_STATE);
			
			/* Set some default values. */
			frameCount = 1;
			cteType = CTE_TYPE_NONE;
			cteAntPat = CTE_ANTENNA_PATTERN_DISABLED;
			radioMode = RADIO_MODE_BLE_2MBIT;
			sampleType = MEASURE_BUFFER_SAMPLE_TYPE_NOTHING;
			samples = NULL;
			sampleAmount = 0;
			fmt = POSTPROC_FORMAT_INVALID;
		}
		else{
			
			/* Derive the format of the samples. */
		
			if(sampleType == MEASURE_BUFFER_SAMPLE_TYPE_IQ
			&& (cteType == CTE_TYPE_2US_7P1
			 || cteType == CTE_TYPE_2US_7PX1)
			&& measure_is_cte_sampling_frame(frame)){
				
				switch(radioMode){
					
					case RADIO_MODE_BLE_2MBIT:
					
						samples = radio_measure_ble_2mbit_get_iq_sample_address(
								  	radioPacket
								  );
						
					break;
					case RADIO_MODE_BLE_1MBIT:
					
						samples = radio_measure_ble_1mbit_get_iq_sample_address(
								  	radioPacket
								  );
						
					break;
					default:
					
						samples = NULL;
						/* Prevent out-of-bound reads/writes. */
						sampleAmount = 0;
				}
				
				if(cteType == CTE_TYPE_2US_7P1){
				
					fmt = POSTPROC_FORMAT_0;
					expectedAmount = CTE_2US_7P1_NUMBER_OF_SAMPLES;
				}
				else if(cteType == CTE_TYPE_2US_7PX1){
				
					fmt = POSTPROC_FORMAT_1;
					expectedAmount = CTE_2US_7PX1_NUMBER_OF_SAMPLES;
				}
				else{
					
					fmt = POSTPROC_FORMAT_INVALID;
					expectedAmount = 0;
				}
			}
			else{
				
				fmt = POSTPROC_FORMAT_INVALID;
				expectedAmount = 0;
				samples = NULL;
				sampleAmount = 0;
			}
			
			if(expectedAmount != sampleAmount){
				
				debug_report(DEBUG_POSTPROC_UNEXPECTED_SAMPLE_AMOUNT);
				
				/* Prevent out-of-bound reads/writes. */
				sampleAmount = 0;
			}
		}
		
		/* Retrieve the corresponding format specifier for the flash memory. */
		
		if(storeSamples){
		
			fmtSpec = coder_format_specifier(fmt);
		}
		else{
			
			/* Do not store samples on flash memory. */
			fmtSpec = CODER_FORMAT_SPECIFIER_NONE;
		}
		
		/*
		  Store RSS data.
		*/
		
		ready = memory_measure_write_results(
		        	(uint32_t) cycle * frameCount + frame,
		        	rssAvgIntegerPart,
		        	rssAvgFractionalPart,
		        	fmtSpec,
		        	samples,
		        	sampleAmount,
		        	false,
		        	false
		        );
		
		if(!ready){
			
			state = STATE_WAITING_FOR_MEMORY;
		}
		else{
			
			state = STATE_READY;
		}
	
		/*
		  Prepare raw samples to be sent via USB (if any and if requested).
		*/
		
		if(postprocSampleStat == POSTPROC_SAMPLE_STATUS_ONGOING
		&& fmt != POSTPROC_FORMAT_INVALID){
			
			if(frame == postprocSampleFrame
			|| postprocSampleFrame == 255){
				
				/* Remeber frame in the latter case. */
				postprocSampleFrame = frame;
				
				for(
					i = 0;
					   i < sampleAmount * 4
					&& i < POSTPROC_MAX_NUMBER_OF_32_BIT_SAMPLES * 4;
					++i
				){
					
					postprocSample[i] = samples[i];
				}
				
				postprocSampleStat = POSTPROC_SAMPLE_STATUS_READY;
				postprocSampleFormat = fmt;
				postprocSampleAmount = sampleAmount;
			}
		}
	}
	else{
		
		/* The memory module needs some additional calls to finish the */
		/* storage of measurement results requested earlier.           */
		
		ready = memory_measure_write_results(
		        	0,
		        	0,
		        	0,
		        	CODER_FORMAT_SPECIFIER_NONE,
		        	NULL,
		        	0,
		        	false,
		        	false
		        );
		
		if(!ready){
			
			state = STATE_WAITING_FOR_MEMORY;
		}
		else{
			
			state = STATE_READY;
		}
	}
	
	return ready;
}

