/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

#define RSS2I_MODULE "power"

/*
  Initialization of the on-chip DC/DC regulator and the brownout detection.
  
  Authors
  =======
  
    Copyright (C) 2018-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
  
    2018-10-16    Initial release.
    2020-07-09    Added brownout detection functionality.
    2020-08-25    Added HF switch protection in brownout case.
    2021-02-26    Call {cte_shutdown()} instead of {cte_disable()}.
    2021-03-27    Specify interrupt priority explicitly.
  
*/

#include <libopencm3/cm3/nvic.h>
#include <libopencm3/nrf5/power.h>

#include <rss2i/debug.h>
#include <rss2i/switch.h>
#include <rss2i/cte.h>
#include <rss2i/nvic.h>

void power_init(void){
/*
  Initializes/Enables the DC/DC regulator and the brownout (a.k.a. powerfail)
  detection.
*/
	
	POWER_DCDCEN = POWER_DCDCEN_DCDCEN_ENABLED;
	
	debug_init();
	debug("init", "Enabled the DC/DC regulator successfully.");
	
	POWER_INTENSET = POWER_INTENSET_POFWARN_SET;
	
	nvic_enable_irq(NVIC_POWER_CLOCK_IRQ);
	nvic_set_priority(NVIC_POWER_CLOCK_IRQ, NVIC_RSS2I_LOW_PRIORITY);
	
	POWER_POFCON = POWER_POFCON_POF_ENABLED
	             | POWER_POFCON_THRESHOLD_V28;
}

void power_clock_isr(void){
/*
  Catch a brownout (a.k.a. powerfail).
*/
	
	if(POWER_EVENTS_POFWARN){
		
		POWER_EVENTS_POFWARN = 0;
		
		debug_report(DEBUG_POWER_BROWNOUT);
		
		/* Protect the external HF switch. */
		cte_shutdown();
		switch_lock();
		
		
		while(POWER_EVENTS_POFWARN);
	}
}

