/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

#define RSS2I_MODULE "ppi"

/*
  API for enabling and disabling some predefined inter peripheral shortcuts.
  
  Authors
  =======
    
    Copyright (C) 2018-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
    
    2019-06-13    Initial release.
    2019-09-26    Add {PPI_TIMER2_COMPARE0_RADIO_RSSISTART}.
    2020-06-09    Use {TIMER1} instead of {TIMER3}.
    2020-08-07    Replaced {RADIO_DISABLED} by {RADIO_END}.
    2020-08-25    Added {PPI_RADIO_CRCOK_TIMER1_START}.
    2021-03-01    Added {PPI_RADIO_FRAMESTART_RADIO_RSSISTART}.
    2021-03-25    Added {PPI_*_GPIOTE_*_CHANNEL_DEBUG_PIN}.
    2021-03-25    Removed {PPI_RADIO_END_TIMER1_CLEAR}.
    2021-03-25    Made the test for {PPI_CHANNEL_COUNT} static.
  
  Description
  ===========
    
    This is an abstraction layer for the PPI peripheral that allows enabling and
    disabling of some predefined inter peripheral shortcuts. A complete list of
    available inter peripheral shortcuts can be obtained from the corresponding
    header file.
    
    Each inter peripheral shoutcut has to be initialized via the {ppi_init()}
    function before it can be enabled or disabled by the {ppi_enable()} or
    {ppi_disable()} functions, respectively.
*/

#include <stdbool.h>
#include <stdint.h>

#include <libopencm3/nrf5/ppi.h>
#include <libopencm3/nrf5/radio.h>
#include <libopencm3/nrf5/timer.h>
#include <libopencm3/nrf5/gpiote.h>

#include <rss2i/ppi.h>
#include <rss2i/debug.h>
#include <rss2i/rev.h>
#include <rss2i/gpiote.h>


/* PPI configurations up from channel 20 are predefined by hardware. */
#define PPI_CHANNEL_COUNT 20

/* Check for the maximal number of PPIs. */
_Static_assert(
	PPI_TYPE_ALWAYS_LAST_ITEM <= PPI_CHANNEL_COUNT,
	"The maximal number of PPI channels was exceeded."
);


static bool ppiInitialized[PPI_TYPE_ALWAYS_LAST_ITEM];


void ppi_init(PPIType ppiType){
/*
  Initialize the specified inter peripheral shortcut.
*/
	
	if(ppiType < PPI_TYPE_ALWAYS_LAST_ITEM){
		
		ppiInitialized[ppiType] = true;
		
		switch(ppiType){
			
			case PPI_RADIO_CRCOK_TIMER1_CLEAR:
				
				PPI_CH[ppiType].EEP = (uint32_t) &(RADIO_EVENTS_CRCOK);
				PPI_CH[ppiType].TEP = (uint32_t) &(TIMER_TASKS_CLEAR(TIMER1));
				
			break;
			case PPI_RADIO_CRCOK_TIMER1_START:
			
				PPI_CH[ppiType].EEP = (uint32_t) &(RADIO_EVENTS_CRCOK);
				PPI_CH[ppiType].TEP = (uint32_t) &(TIMER_TASKS_START(TIMER1));
			
			break;
			case PPI_TIMER2_COMPARE0_RADIO_RSSISTART:
			
				PPI_CH[ppiType].EEP = (uint32_t) &(TIMER_EVENTS_COMPARE(TIMER2)[
				                                   	0
				                                   ]);
				PPI_CH[ppiType].TEP = (uint32_t) &(RADIO_TASKS_RSSISTART);
			
			break;
		#if REVISION >= REVISION_F
			case PPI_RADIO_FRAMESTART_RADIO_RSSISTART:
			
				PPI_CH[ppiType].EEP = (uint32_t) &(RADIO_EVENTS_FRAMESTART);
				PPI_CH[ppiType].TEP = (uint32_t) &(RADIO_TASKS_RSSISTART);
		#endif
			break;
			case PPI_TIMER1_COMPARE0_GPIOTE_OUT_CHANNEL_DEBUG_PIN:
				
				PPI_CH[ppiType].EEP = (uint32_t) &(TIMER_EVENTS_COMPARE(TIMER1)[
				                                   	0
				                                   ]);
				PPI_CH[ppiType].TEP = (uint32_t) &(GPIOTE_TASKS_OUT[
				                                   	GPIOTE_CHANNEL_DEBUG_PIN
				                                   ]);
			break;
			case PPI_TIMER1_COMPARE1_GPIOTE_SET_CHANNEL_DEBUG_PIN:
				
				PPI_CH[ppiType].EEP = (uint32_t) &(TIMER_EVENTS_COMPARE(TIMER1)[
				                                   	1
				                                   ]);
				PPI_CH[ppiType].TEP = (uint32_t) &(GPIOTE_TASKS_SET[
				                                   	GPIOTE_CHANNEL_DEBUG_PIN
				                                   ]);
			break;
			case PPI_RADIO_READY_GPIOTE_CLR_CHANNEL_DEBUG_PIN:
				
				PPI_CH[ppiType].EEP = (uint32_t) &(RADIO_EVENTS_READY);
				PPI_CH[ppiType].TEP = (uint32_t) &(GPIOTE_TASKS_CLR[
				                                   	GPIOTE_CHANNEL_DEBUG_PIN
				                                   ]);
			break;
			default:
				
				debug_report(DEBUG_PPI_OUT_OF_RANGE);
		}
	}
	else{
		
		debug_report(DEBUG_PPI_OUT_OF_RANGE);
	}
}

void ppi_enable(PPIType ppiType){
/*
  Enable the specified inter peripheral shortcut.
*/
	
	if(ppiType < PPI_TYPE_ALWAYS_LAST_ITEM){
		
		if(ppiInitialized[ppiType]){
		
			PPI_CHENSET = (uint32_t) 1 << ppiType;
		}
		else{
		
			debug_report(DEBUG_PPI_UNINITIALIZED);
		}
	}
}

void ppi_disable(PPIType ppiType){
/*
  Disable the specified inter peripheral shortcut.
*/
	
	if(ppiType < PPI_TYPE_ALWAYS_LAST_ITEM){
	
		if(ppiInitialized[ppiType]){
		
			PPI_CHENCLR = (uint32_t) 1 << ppiType;
		}
		else{
		
			debug_report(DEBUG_PPI_UNINITIALIZED);
		}
	}
}

