/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

#define RSS2I_MODULE "print"

/*
  Small and incomprehensive printing library.
  
  Authors
  =======
  
    Copyright (C) 2018-2019  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
  
    2019-06-21    Initial public release.
*/

#include <stdint.h>
#include <stddef.h>


void print_uoct(uint32_t d, char * dest, size_t pos, size_t len){
/*
  Represents the given unsigned integer as a fixed width octal string with
  leading zeros.
  
  Argument List
  =============
  
    {d}       Unsigned integer to print.
    {dest}    Destination string.
    {pos}     Position inside the given {dest}.
    {len}     Length of the fixed width representation.
    
*/
	
	size_t i;
	
	for(i = 0; i < len; ++i){
		
		dest[pos + len - 1 - i] = ((i < 11) ? ((d >> 3 * i) & 07) : 0) + '0';
	}
}

void print_uhex(uint32_t d, char * dest, size_t pos, size_t len){
/*
  Represents the given unsigned integer as a fixed width hexadecimal string with
  leading zeros.
  
  Argument List
  =============
  
    {d}       Unsigned integer to print.
    {dest}    Destination string.
    {pos}     Position inside the given {dest}.
    {len}     Length of the fixed width representation.
    
*/
	
	size_t i;
	uint8_t digit;
	
	for(i = 0; i < len; ++i){
		
		digit = (i < 8) ? ((d >> (4 * i)) & 0xf) : 0;
		dest[pos + len - 1 - i] = (digit > 9) ? (digit - 10 + 'A') : (digit + '0');
	}
}

void print_udec(uint32_t d, char * dest, size_t pos, size_t len){
/*
  Represents the given unsigned integer as a fixed width decimal string with
  leading zeros.
  
  Argument List
  =============
  
    {d}       Unsigned integer to print.
    {dest}    Destination string.
    {pos}     Position inside the given {dest}.
    {len}     Length of the fixed width representation.
    
*/
	
	size_t i;
	uint8_t digit;
	uint64_t power = 1;
	
	for(i = 0; i < len; ++i){
		
		digit = (d / power) % 10;
		
		if(power < ((uint64_t) 1 << 32)){
			power *= 10;
		}
		
		dest[pos + len - 1 - i] = digit + '0';
	}
}
