/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

#define RSS2I_MODULE "pwm"

/*
  PWM module that controls the onboard LEDs using the PWM peripheral(s).
   
  Authors
  =======
  
    Copyright (C) 2018-2020  Gerrit Maus (funk@maus.xyz)
    Copyright (C) 2018-2018  Luong Le (novercy@live.com)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
  
    2018-10-16    Initial release.
    2020-06-10    Fitted to revision F (missing PWM peripherals).
    2020-08-12    Use {REVISION_*} macros instead of magic numbers.
    2020-08-12    Disable {GPIO6} if measurement frame debug mode is enabled.
    
  Description
  ===========
    
    There are three supported states: "off", "const" and "sequence". The first
    one disables the LED. The second one dims the LED to a constant value. The
    last one plays back a sequence of {2 * PWM_CYCLES} brightness values given
    in the {extern} {pwm_const_*} arrays.
    
    Starting from revision F the LED management was taken over from the LED
    module in order to be able to implement a (constantly) dimmed blinking
    sequence with just a single PWM peripheral. This change coincides with a
    slower response behaviour for {LED_OFF} and {LED_ON} states when one or more
    LEDs are in the {LED_BLINK} state.
    
    This module has to get initialized using {pwm_init()}.
*/


#include <stdint.h>

#include <libopencm3/nrf5/pwm.h>
#include <libopencm3/nrf5/gpio.h>

#include <rss2i/led.h>
#include <rss2i/pwm.h>
#include <rss2i/board.h>
#include <rss2i/debug.h>
#include <rss2i/rev.h>

#include <rss2i/config/led.h>


#if REVISION < REVISION_F

	extern uint16_t pwm_seq_green[];
	extern uint16_t pwm_seq_yellow[];
	extern uint16_t pwm_seq_red[];
	
	uint16_t pwm_const_green[] = {PWM_TOP_GREEN};
	uint16_t pwm_const_yellow[] = {PWM_TOP_YELLOW};
	uint16_t pwm_const_red[] = {PWM_TOP_RED};
	
#else

	/* This depends on hardware! */
	#define PWM_CHANNELS 4
	
	/* This will probably work with most C compilers on this platform. */
	/* However, when padding is applied, this approach may fail.       */
	
/*	typedef struct {*/
/*		uint16_t green;*/
/*		uint16_t yellow;*/
/*		uint16_t red;*/
/*		uint16_t unusedPWMChannel;*/
/*	} PWMData;*/
/*	*/
/*	PWMData pwmData[PWM_STEPS];*/

	static volatile uint16_t pwmData[PWM_STEPS][PWM_CHANNELS];
	static uint16_t pwmConst[] = {PWM_TOP_GREEN, PWM_TOP_YELLOW, PWM_TOP_RED};
	
	static volatile LEDPattern pwmLEDPattern[3] = {LED_OFF, LED_OFF, LED_OFF};
	
#endif



void pwm_init(void){
/*
  Initializes this module. Should be called only once.
*/
	
	PWM_ENABLE(PWM0) = PWM_ENABLE_ENABLE_ENABLED;
	PWM_COUNTERTOP(PWM0) = 4095;
	PWM_PRESCALER(PWM0) = PWM_PRESCALER_PRESCALER_DIV_1;
	PWM_MODE(PWM0) = PWM_MODE_UPDOWN_UPANDDOWN;
	
	
	#if REVISION < REVISION_F
	
		PWM_DECODER(PWM0) = PWM_DECODER_LOAD_COMMON
		                  | PWM_DECODER_MODE_REFRESHCOUNT;
		
		PWM_SEQ(PWM0)[0].REFRESH = ((uint32_t) LED_SLOT_DURATION) * 4096 * 2 * 2 / 16000 / PWM_CYCLES;
		
		PWM_PSEL(PWM0).OUT[0] = 8;
		
		PWM_ENABLE(PWM1) = PWM_ENABLE_ENABLE_ENABLED;
		PWM_COUNTERTOP(PWM1) = 4095;
		PWM_PRESCALER(PWM1) = PWM_PRESCALER_PRESCALER_DIV_1;
		PWM_MODE(PWM1) = PWM_MODE_UPDOWN_UPANDDOWN;
		
		PWM_DECODER(PWM1) = PWM_DECODER_LOAD_COMMON
		                  | PWM_DECODER_MODE_REFRESHCOUNT;
		
		PWM_SEQ(PWM1)[0].REFRESH = ((uint32_t) LED_SLOT_DURATION) * 4096 * 2 * 2 / 16000 / PWM_CYCLES;
		
		PWM_PSEL(PWM1).OUT[0] = 7;
		
		#if !DEBUG_MEASUREMENT_FRAME
			
			PWM_ENABLE(PWM2) = PWM_ENABLE_ENABLE_ENABLED;
			PWM_COUNTERTOP(PWM2) = 4095;
			PWM_PRESCALER(PWM2) = PWM_PRESCALER_PRESCALER_DIV_1;
			PWM_MODE(PWM2) = PWM_MODE_UPDOWN_UPANDDOWN;
			
			PWM_DECODER(PWM2) = PWM_DECODER_LOAD_COMMON
			                  | PWM_DECODER_MODE_REFRESHCOUNT;
			
			PWM_SEQ(PWM2)[0].REFRESH = ((uint32_t) LED_SLOT_DURATION) * 4096 * 2 * 2 / 16000 / PWM_CYCLES;
			
			PWM_PSEL(PWM2).OUT[0] = 6;
			
		#else
			#if DEBUG_MEASUREMENT_FRAME_GPIO != GPIO6
				#error "Failed to protect measurement frame debug GPIO."
			#endif
		#endif
		
	#else
	
		PWM_SEQ(PWM0)[0].REFRESH = ((uint32_t) PWM_SLOT_DURATION) * 16000 / 4096 / 2;
		
		pwm_set(LED_GREEN, LED_OFF);
		pwm_set(LED_YELLOW, LED_OFF);
		pwm_set(LED_RED, LED_OFF);
		
		PWM_DECODER(PWM0) = PWM_DECODER_LOAD_INDIVIDUAL
		                  | PWM_DECODER_MODE_REFRESHCOUNT;
		
		PWM_SEQ(PWM0)[0].PTR = (uint32_t) &pwmData[0][0];
		PWM_SEQ(PWM0)[0].CNT = PWM_CHANNELS * PWM_STEPS;
		
		PWM_PSEL(PWM0).OUT[0] = 8; /* Green  */
		PWM_PSEL(PWM0).OUT[1] = 7; /* Yellow */
		#if !DEBUG_MEASUREMENT_FRAME
			
			PWM_PSEL(PWM0).OUT[2] = 6; /* Red    */
			
		#else
			#if DEBUG_MEASUREMENT_FRAME_GPIO != GPIO6
				#error "Failed to protect measurement frame debug GPIO."
			#endif
		#endif
		
		PWM_TASKS_SEQSTART(PWM0)[0] = 1;
	#endif
}

#if REVISION < REVISION_F

	void pwm_start_seq_green(void){
	/*
	  Starts the fading sequence for the green LED.
	*/
		PWM_SEQ(PWM0)[0].PTR = (uint32_t) pwm_seq_green;
		PWM_SEQ(PWM0)[0].CNT = 2 * PWM_CYCLES;
		PWM_TASKS_SEQSTART(PWM0)[0] = 1;
	}

	void pwm_start_seq_yellow(void){
	/*
	  Starts the fading sequence for the yellow LED.
	*/

		PWM_SEQ(PWM1)[0].PTR = (uint32_t) pwm_seq_yellow;
		PWM_SEQ(PWM1)[0].CNT = 2 * PWM_CYCLES;
		PWM_TASKS_SEQSTART(PWM1)[0] = 1;
	}

	void pwm_start_seq_red(void){
	/*
	  Starts the fading sequence for the red LED.
	*/

		PWM_SEQ(PWM2)[0].PTR = (uint32_t) pwm_seq_red;
		PWM_SEQ(PWM2)[0].CNT = 2 * PWM_CYCLES;
		PWM_TASKS_SEQSTART(PWM2)[0] = 1;
	}

	void pwm_start_const(LED led){
	/*
	  Dimms the specified {led} to a constant brightness.
	*/
		
		if(led == LED_GREEN){
			
			PWM_SEQ(PWM0)[0].PTR = (uint32_t) pwm_const_green;
			PWM_SEQ(PWM0)[0].CNT = 1;
			PWM_TASKS_SEQSTART(PWM0)[0] = 1;
		}
		else if(led == LED_YELLOW){
		
			PWM_SEQ(PWM1)[0].PTR = (uint32_t) pwm_const_yellow;
			PWM_SEQ(PWM1)[0].CNT = 1;
			PWM_TASKS_SEQSTART(PWM1)[0] = 1;
		}
		else if(led == LED_RED){
		
			PWM_SEQ(PWM2)[0].PTR = (uint32_t) pwm_const_red;
			PWM_SEQ(PWM2)[0].CNT = 1;
			PWM_TASKS_SEQSTART(PWM2)[0] = 1;
		}
	}

	void pwm_off(LED led){
	/*
	  Disables the specified {led}.
	*/
		
		if(led == LED_GREEN){
			
			PWM_TASKS_STOP(PWM0) = 1;
		}
		else if(led == LED_YELLOW){
		
			PWM_TASKS_STOP(PWM1) = 1;
		}
		else if(led == LED_RED){
		
			PWM_TASKS_STOP(PWM2) = 1;
		}
	}
	
#else
	
	void pwm_set(LED led, LEDPattern pattern){
	/*
	  Set the state of the specified {led} according to {pattern}.
	*/
		
		uint8_t i;
		
		pwmLEDPattern[led] = pattern;
		
		if(pattern == LED_OFF){
			
			for(i = 0; i < PWM_STEPS; ++i){
				
				pwmData[i][led] = 0;
			}
		}
		else if(pattern == LED_ON){
			
			for(i = 0; i < PWM_STEPS; ++i){
				
				pwmData[i][led] = pwmConst[led];
			}
		}
		if(pattern == LED_BLINK){
			
			for(i = 0; i < PWM_STEPS; ++i){
				
				pwmData[i][led] = (i == PWM_SLOTS_BEFORE + led) ? pwmConst[led] : 0;
			}
		}
		
		if(pwmLEDPattern[LED_GREEN]  != LED_BLINK
		&& pwmLEDPattern[LED_YELLOW] != LED_BLINK
		&& pwmLEDPattern[LED_RED]    != LED_BLINK){
			
			/* There is no blinking LED. It is hence safe to */
			/* apply the new settings right now instead of   */
			/* waiting for the ongoing sequence to end.      */
			
			pwm_start();
		}
	}
	
	void pwm_start(void){
	/*
	  Enters the first step of the PWM sequence.
	*/
	
		PWM_TASKS_SEQSTART(PWM0)[0] = 1;
	}
	
#endif

