/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

#define RSS2I_MODULE "radio"

/*
  API for the RADIO peripheral.
  
  Authors
  =======
    
    Copyright (C) 2018-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
    
  Changelog
  =========
    
    2019-07-02    Initial public release.
    2019-08-29    Round measured RSS values properly.
    2019-09-26    Enhance the number of RSSI measurements per packet.
    2020-06-09    Use {TIMER1} instead of {TIMER3}.
    2020-06-19    Support DMA based measurement radio packet data RAM buffering.
    2020-07-30    Use {PHYEND} instead of {END} event starting from revision F.
    2020-08-04    Replace undefinded {BOARD_REV_*} enum with {REVISION_*} macro.
    2020-08-07    Added {CTEType} and {CTEAntennaPattern} arguments.
    2020-08-07    Fit measurement schedule to variable length CTEs.
    2020-08-12    Enhance frame timing debug functionality.
    2020-08-13    Introduced {DEBUG_MEASUREMENT_FRAME} level {3}.
    2020-08-13    Introduced {radio_prepare_measure_*()} functions.
    2020-08-13    Enhance {DEBUG_MEASUREMENT_FRAME} level {3} functionality.
    2020-08-13    Configure CTE during RADIO RX/TX preparation/initialization.
    2020-08-13    Enable antenna selection for {radio_init_emit_carrier()}.
    2020-08-13    Remove magic numbers within {radio_init_measure_tx()}.
    2020-08-13    Forward RADIO packet meta data.
    2020-08-17    Add {radio_get_iq_sample_address()}.
    2020-08-25    Start measurement schedule once successfully resynced.
    2021-02-25    Add (BLE) 2MBIT RADIO mode prefixes and unify function names.
    2021-02-26    Receive the radio packet address along with the call to the
                  {radio_measure_*_rx()} functions instead of asking for it.
    2021-02-28    Support IEEE 250kBit and BLE/NRF 1/2MBit modes.
    2021-03-18    Made {packetDataAddr} volatile.
    2021-03-25    Use GPIOTE peripheral for debug toggling.
    2021-03-25    Do not manipulate the measurement scheduler during TX anymore.
    2021-03-25    Fix {*_1mbit_get_remote_rss_fractional_part()} conversion bug.
    2021-03-27    Use {NVIC_RSS2I_*_PRIORITY} macros.
    2021-03-29    Use {SwitchAntenna} for {radio_emit_carrier()} configuration.
    2021-03-29    Made the {radio_config_*()} functions inline.
    2021-03-29    Added {radio_emit_constant()} function.
  
  Description
  ===========
    
    This API offers an interface to some functionality that can be used for
    over-the-air (OTA) messaging and for measurement tasks: Listening to or
    sending of OTA messages, listening/receiving of measurement packets,
    transmitting of measurement packets or transmitting a constant carrier (for
    testing purposes).
    
    Entering _most_ of those states, described above, can be requested at any
    time: If necessary, this module shuts down the ongoing RADIO task, before it
    continues to prepare the desired state. Note, that due to the precise timing
    requirements for a measurement packet's transmission, its preparation (using
    {radio_measure_*_tx()}) excepts this rule. It needs a manual shut down
    (using {radio_check_disable()}) about 5 µs earlier.
    
    Preparation of measurement tasks is split into two different functions for
    both the reception and the transmission of measurement packets. In both
    cases there are two functions {radio_measure_*_prepare_*()} as well as
    {radio_measure_*_init_*()}. The former should be called in an early stage
    of the preparation process. The latter should be called later on. Both
    function calls shall provide the same arguments. The RADIO module decides
    which settings may be applied in the individual stages.
*/

#include <stdint.h>
#include <stddef.h>
#include <stdbool.h>

#include <libopencm3/nrf5/gpio.h>
#include <libopencm3/nrf5/gpiote.h>
#include <libopencm3/nrf5/radio.h>
#include <libopencm3/nrf5/timer.h>
#include <libopencm3/cm3/nvic.h>

#include <rss2i/radio.h>
#include <rss2i/measure.h>
#include <rss2i/led.h>
#include <rss2i/board.h>
#include <rss2i/rev.h>
#include <rss2i/ble.h>
#include <rss2i/buffer.h>
#include <rss2i/debug.h>
#include <rss2i/battery.h>
#include <rss2i/memory.h>
#include <rss2i/ppi.h>
#include <rss2i/cte.h>
#include <rss2i/hook.h>
#include <rss2i/errata.h>
#include <rss2i/gpiote.h>
#include <rss2i/nvic.h>
#include <rss2i/switch.h>

#include <rss2i/config/radio.h>
#include <rss2i/config/debug.h>


/* Make sure that all radio packets fit into {radioPacketData[]}. */
_Static_assert(
	   /* OTA send: Targeted device isn't part of the radio packet.          */
	   RADIO_OTA_SEND_BUFFER_WIDTH - 1                    <= RADIO_PACKET_WIDTH
	   /* OTA recv: USB-CRC bytes is not part of the radio packet.           */
	&& RADIO_OTA_RECV_BUFFER_WIDTH - 4                    <= RADIO_PACKET_WIDTH
	   /* X 2MBit: Additional length and S0 bytes                            */
	&& RADIO_MEASURE_X_2MBIT_PAYLOAD_LENGTH + 2           <= RADIO_PACKET_WIDTH
	   /* X 1MBit: Additional length and S0 bytes                            */
	&& RADIO_MEASURE_X_1MBIT_PAYLOAD_LENGTH + 2           <= RADIO_PACKET_WIDTH
	   /* IEEE 250kBit: Additional length byte                               */
	&& RADIO_MEASURE_IEEE_250KBIT_PAYLOAD_LENGTH + 1      <= RADIO_PACKET_WIDTH
	   /* Max Payload: Additional length and S0 bytes                        */
	&& RADIO_MAX_PAYLOAD + 2                              <= RADIO_PACKET_WIDTH,
	
	"Not all RADIO packets fit into the RADIO packet memory."
);


/*
  There is no {RADIO_EVENTS_PHYEND} or {RADIO_SHORTS_PHYEND_DISABLE}
  for revisions prior to F.
*/
#if REVISION <= REVISION_E
	#define RADIO_EVENTS_DISPATCH_PHYEND RADIO_EVENTS_END
	#define RADIO_SHORTS_DISPATCH_PHYEND_DISABLE_ENABLED \
	        RADIO_SHORTS_END_DISABLE_ENABLED
#else
	#define RADIO_EVENTS_DISPATCH_PHYEND RADIO_EVENTS_PHYEND
	#define RADIO_SHORTS_DISPATCH_PHYEND_DISABLE_ENABLED \
	        RADIO_SHORTS_PHYEND_DISABLE_ENABLED
#endif


typedef enum {
	
	RADIO_BLE_MODE_UNDECIDED, /* Symbolic use only. */
	RADIO_BLE_MODE_BLE,
	RADIO_BLE_MODE_NRF

} RadioBLEMode;

typedef enum {
	
	RADIO_DISABLED,
	
	RADIO_MEASURE_BLE_2MBIT_TX,
	RADIO_MEASURE_NRF_2MBIT_TX,
	RADIO_PREPARE_X_2MBIT_RX,
	RADIO_MEASURE_NRF_2MBIT_RX,
	RADIO_MEASURE_BLE_2MBIT_RX,
	
	RADIO_MEASURE_BLE_1MBIT_TX,
	RADIO_MEASURE_NRF_1MBIT_TX,
	RADIO_PREPARE_X_1MBIT_RX,
	RADIO_MEASURE_BLE_1MBIT_RX,
	RADIO_MEASURE_NRF_1MBIT_RX,
	
	RADIO_MEASURE_IEEE_250KBIT_TX,
	RADIO_PREPARE_IEEE_250KBIT_RX,
	RADIO_MEASURE_IEEE_250KBIT_RX,
	
	RADIO_PREPARE_SEND,
	RADIO_CONTROL_SEND,
	
	RADIO_PREPARE_RECV,
	RADIO_CONTROL_RECV,
	
	RADIO_PREPARE_CONSTANT,
	RADIO_EMIT_CONSTANT,

} RadioState;

static volatile RadioState radioState = RADIO_DISABLED;

static volatile uint8_t radioPacketData[RADIO_PACKET_WIDTH];
static volatile uint32_t radioRSSAvg;
static volatile uint16_t radioRSSCnt;

static uint8_t radioBatteryLevel = 100;
static uint8_t radioMemoryLoad = 0;

static volatile uint16_t radioLEDTimeout;
static volatile bool radioControlSendLedOn;

Buffer radioRecvBuffer;
Buffer radioSendBuffer;

volatile bool radioSendBusy = false;

static uint8_t radioTXPower = RADIO_TX_POWER;


bool radio_set_tx_power_needs_control_mode(uint8_t txPower){
/*
  Whether the corresponding call to {radio_set_tx_power()} needs control mode
  or not.
*/
	
	return (txPower > 0 && txPower < 10) ? true : false;
}

uint8_t radio_set_tx_power(uint8_t txPower){
/*
  Sets the RADIO's transmit power.
  
  Description
  ===========
    
    The transmit power is set according to the following table. If the specified
    {txPower} is out-of-range, the transmit power doesn't get changed.
  
          {txPower} | actual transmit power 
         -----------+-----------------------
              1     |      + 4 dBm
              2     |      + 3 dBm
              3     |        0 dBm
              4     |      - 4 dBm
              5     |      - 8 dBm
              6     |      -12 dBm
              7     |      -16 dBm
              8     |      -20 dBm
              9     |      -40 dBm
  
  Return Value
  ============
    
    Returns the designated transmit power according the the table above.
*/
	
	switch(txPower){
		
		case 1: RADIO_TXPOWER = RADIO_TXPOWER_TXPOWER_Pos4dBm; break;
		case 2: RADIO_TXPOWER = RADIO_TXPOWER_TXPOWER_Pos3dBm; break;
		case 3: RADIO_TXPOWER = RADIO_TXPOWER_TXPOWER_0dBm; break;
		case 4: RADIO_TXPOWER = RADIO_TXPOWER_TXPOWER_Neg4dBm; break;
		case 5: RADIO_TXPOWER = RADIO_TXPOWER_TXPOWER_Neg8dBm; break;
		case 6: RADIO_TXPOWER = RADIO_TXPOWER_TXPOWER_Neg12dBm; break;
		case 7: RADIO_TXPOWER = RADIO_TXPOWER_TXPOWER_Neg16dBm; break;
		case 8: RADIO_TXPOWER = RADIO_TXPOWER_TXPOWER_Neg20dBm; break;
		case 9: RADIO_TXPOWER = RADIO_TXPOWER_TXPOWER_Neg40dBm; break;
		default:
			
			txPower = radioTXPower;
	}
	
	radioTXPower = txPower;
	return txPower;
}

void radio_ota_recv_flush(void){
/*
  Empties the RADIO (OTA) input buffer.
*/
	
	buffer_empty(&radioRecvBuffer);
}

uint8_t * radio_ota_recv(void){
/*
  Returns the address of a received OTA message (if any). The memory pointed
  to by the return value is guaranteed to retain its value until the next call
  of this function.
*/

	return buffer_get_row(&radioRecvBuffer);
}

bool radio_ota_send(uint8_t devID, uint8_t * data, uint16_t size){
/*
  Sends {size} bytes starting from {&data[2]} over-the-air to device {devID}.
  The first two bytes may contain arbitrary values but will be overwritten by
  calling this function.
  
  Argument list
  =============
  
    {devID}       RSS2I device ID of the device to be addressed. {0} addresses
                  every device within the RADIO's range.
    {data}        Data to be sent starting from {&data[2]}. The first two bytes
                  may contain arbitrary values but will be overwritten by
                  calling this function.
    {size}        Number of bytes to be sent. The maximal allowed {size} is
                  {256}. Higher values get truncated to {256} automatically.
                 
  Buffer entry content vs. packet layout
  ======================================
  
  
          Target's       User data                             User data       
         device ID     (first byte)    PAYLOAD length    (second byte, third …) 
   
             |               |               |                    |
       ______.______   ______._____   _______._______   __________.__________
      ´             ` ´            ` ´               ` ´                     `
       ______________ ______________ _________________ _______________________
      |              |              |                 |                       |
      | bffrEntry[0] | bffrEntry[1] |  bffrEntry[2]   | bffrEntries[3, 4, …]  |
      |______________|______________|_________________|_______________________|
      
                      ______________ _________________ _______________________
                     |              |                 |                       |
                     | S0 (1 byte)  | LENGTH (1 byte) | PAYLOAD (<=255 bytes) |
                     |______________|_________________|_______________________|
          
                     `------.------´ `-------.-------´ `----------.----------´
                            |                |                    |
                          Packet       PAYLOAD length        Packet content
                         content                         (second byte, third …)
                       (first byte)
      
  Return value
  ============
  
    Returns {false} if the RADIO peripheral is too busy to send. Returns {true}
    otherwise.
*/
	
	if(size > 256){
	
		size = 256;
		debug_report(DEBUG_RADIO_OTA_MESSAGE_OVERSIZED);
	}
	
	data[0] = devID;
	data[1] = data[2];
	data[2] = size - 1;
	
	if(buffer_append_row(&radioSendBuffer, data, size + 2)){
		
		/* Start OTA transmission if necessary. */
		
		if(!radioSendBusy){
			
			radio_control_init_send();
		}
		
		/* Restore original data. */
		data[2] = data[1];
		
		return true;
	}
	else{
		
		debug_report(DEBUG_RADIO_OTA_SEND_BUFFER_OVERFLOW);
		
		/* Restore original data. */
		data[2] = data[1];
		
		return false;
	}
}

static void radio_set_control_access_addresses(uint32_t devID){
/*
  Sets the BLE logic access addresses that get used during control mode:
    
    0: Listen to (unique device ID)
    1: Additionally listen to/Send to (specified by {devID})
*/
	
	uint32_t ownAddress = ble_gen_control_access_address(board_id());
	uint32_t givenAddress = ble_gen_control_access_address(devID);
	
	
	RADIO_PREFIX0 = 0x12340000
	              | ((ownAddress >> 24) & 0xff)
	              | (((givenAddress >> 24) & 0xff) << 8);

	RADIO_BASE0 = ownAddress << 8;
	RADIO_BASE1 = givenAddress << 8;
}

static void radio_set_measure_access_addresses(void){
/*
  Sets the BLE logic access addresses that get used during measurement mode:
    
    0: Listen to/Send with (measurement packet)
    1: Listen to/Send with (disruptive packet)
*/
	
	uint32_t measureAddress = ble_gen_measure_access_address();
	uint32_t disruptAddress = ble_gen_disrupt_access_address();
	
	
	RADIO_PREFIX0 = 0x12340000
	              | ((measureAddress >> 24) & 0xff)
	              | (((disruptAddress >> 24) & 0xff) << 8);

	RADIO_BASE0 = measureAddress << 8;
	RADIO_BASE1 = disruptAddress << 8;
}

static inline void radio_config_frequency_deviation_trans1(void){
/*
  Configures the RADIO to transmit a constant series of 1s over the air while
  in TXIDLE state. Furthermore, this enables the fast-ramp-up RADIO mode.
*/
	
	RADIO_MODECNF0 = RADIO_MODECNF0_RU_FAST
	               | RADIO_MODECNF0_DTX_B1;
}

static inline void radio_config_frequency_deviation_trans0(void){
/*
  Configures the RADIO to transmit a constant series of 0s over the air while
  in TXIDLE state. Furthermore, this enables the fast-ramp-up RADIO mode.
*/
	
	RADIO_MODECNF0 = RADIO_MODECNF0_RU_FAST
	               | RADIO_MODECNF0_DTX_B0;
}

static inline void radio_config_frequency_deviation_center(void){
/*
  Configures the RADIO to emit the undeviated carrier while in TXIDLE state.
  Furthermore, this enables the fast-ramp-up RADIO mode.
*/
	
	RADIO_MODECNF0 = RADIO_MODECNF0_RU_FAST
	               | RADIO_MODECNF0_DTX_CENTER;
}

static inline void radio_config_x_x(void){
/*
  Configures the RADIO to comply with the common parts of the BLE 5.x 1MBit and
  2MBit link layers. This configuration is used by the BLE 1MBit and 2MBit RADIO
  modes but also used by the NRF 1MBit and 2MBit RADIO modes. The initial value
  of the whitening register is configured by {radio_init()}.
*/
	
	/* CRC configuration. */
	
	RADIO_CRCCNF = RADIO_CRCCNF_LEN_VAL(RADIO_CRCLENG_BLE)
	             | RADIO_CRCCNF_SKIPADDR_SKIP;

	RADIO_CRCPOLY = RADIO_CRCPOLY_BLE;
	RADIO_CRCINIT = RADIO_CRCINIT_BLE;
	
	/* Configure packet structure. */
	
	RADIO_PCNF1 = RADIO_PCNF1_MAXLEN_VAL(RADIO_MAX_PAYLOAD)
	            | RADIO_PCNF1_STATLEN_VAL(0)
	            | RADIO_PCNF1_BALEN_VAL(3)
	            | RADIO_PCNF1_ENDIAN_LITTLE
	            | RADIO_PCNF1_WHITEEN_ENABLED;
}

static inline void radio_config_x_2mbit(void){
/*
  Configures the RADIO to comply with the BLE 5.x 2MBit link layer. This
  configuration is used by the BLE 2MBit RADIO mode but also used by the NRF 
  2MBit RADIO mode. The initial value of the whitening register is configured
  by {radio_init()}.
*/
	
	/* Configure packet structure. */
	
	RADIO_PCNF0 = RADIO_PCNF0_LFLEN_VAL(8)
	            | RADIO_PCNF0_S0LEN_VAL(1)
	            | RADIO_PCNF0_S1LEN_VAL(0)
	            | RADIO_PCNF0_S1INCL_AUTOMATIC
	            | RADIO_PCNF0_PLEN_16BIT
	            | RADIO_PCNF0_CRCINC_EXCLUDE
	            | RADIO_PCNF0_TERMLEN_VAL(0);
}

static inline void radio_config_x_1mbit(void){
/*
  Configures the RADIO to comply with the BLE 5.x 1MBit link layer. This
  configuration is used by the BLE 1MBit RADIO mode but also used by the NRF 
  1MBit RADIO mode. The initial value of the whitening register is configured
  by {radio_init()}.
*/
	
	/* Configure packet structure. */
	
	RADIO_PCNF0 = RADIO_PCNF0_LFLEN_VAL(8)
	            | RADIO_PCNF0_S0LEN_VAL(1)
	            | RADIO_PCNF0_S1LEN_VAL(0)
	            | RADIO_PCNF0_S1INCL_AUTOMATIC
	            | RADIO_PCNF0_PLEN_8BIT
	            | RADIO_PCNF0_CRCINC_EXCLUDE
	            | RADIO_PCNF0_TERMLEN_VAL(0);
}

static inline void radio_config_ble_2mbit(void){
/*
  Configures the RADIO to comply with the BLE 5.x 2MBit link layer. The initial
  value of the whitening register is configured by {radio_init()}.
*/
	
	/* Revert workaround for some IEEE 802.15.4 related errata.     */
	/* (See corresponding doc for more information on this errata.) */
	errata_232_revert_workaround();
	
	RADIO_MODE = RADIO_MODE_MODE_BLE_2MBIT;
	
	radio_config_x_x();
	radio_config_x_2mbit();
	radio_config_frequency_deviation_center();
}

static inline void radio_config_ble_1mbit(void){
/*
  Configures the RADIO to comply with the BLE 5.x 1MBit link layer. The initial
  value of the whitening register is configured by {radio_init()}.
*/
	
	/* Revert workaround for some IEEE 802.15.4 related errata.     */
	/* (See corresponding doc for more information on this errata.) */
	errata_232_revert_workaround();
	
	RADIO_MODE = RADIO_MODE_MODE_BLE_1MBIT;
	
	radio_config_x_x();
	radio_config_x_1mbit();
	radio_config_frequency_deviation_center();
}

static inline void radio_config_nrf_2mbit(void){
/*
  Configures the RADIO to comply with the Nordic proprietary 2MBit link layer.
  The initial value of the whitening register is configured by {radio_init()}.
*/
	
	/* Revert workaround for some IEEE 802.15.4 related errata.     */
	/* (See corresponding doc for more information on this errata.) */
	errata_232_revert_workaround();
	
	RADIO_MODE = RADIO_MODE_MODE_NRF_2MBIT;
	
	radio_config_x_x();
	radio_config_x_2mbit();
	radio_config_frequency_deviation_center();
}

static inline void radio_config_nrf_1mbit(void){
/*
  Configures the RADIO to comply with the Nordic proprietary 1MBit link layer.
  The initial value of the whitening register is configured by {radio_init()}.
*/
	
	/* Revert workaround for some IEEE 802.15.4 related errata.     */
	/* (See corresponding doc for more information on this errata.) */
	errata_232_revert_workaround();
	
	RADIO_MODE = RADIO_MODE_MODE_NRF_1MBIT;
	
	radio_config_x_x();
	radio_config_x_1mbit();
	radio_config_frequency_deviation_center();
}

static inline void radio_config_ieee_250kbit(void){
/*
  Configures the RADIO to comply with the IEEE 802.15.4 link layer.
*/
	
	#if REVISION >= REVISION_F
	
		/* Apply workaround for some IEEE 802.15.4 related errata.      */
		/* (See corresponding doc for more information on this errata.) */
		errata_232_apply_workaround();
		
		RADIO_MODE = RADIO_MODE_MODE_IEEE802154_250KBIT;
		
		/* CRC configuration. */
		
		RADIO_CRCCNF = RADIO_CRCCNF_LEN_VAL(RADIO_CRCLENG_IEEE)
		             | RADIO_CRCCNF_SKIPADDR_IEEE802154;
		
		RADIO_CRCPOLY = RADIO_CRCPOLY_IEEE;
		RADIO_CRCINIT = RADIO_CRCINIT_IEEE;
		
		/* Configure packet structure. */
		
		RADIO_PCNF0 = RADIO_PCNF0_LFLEN_VAL(8)
		            | RADIO_PCNF0_S0LEN_VAL(0)
		            | RADIO_PCNF0_S1LEN_VAL(0)
		            | RADIO_PCNF0_S1INCL_AUTOMATIC
		            | RADIO_PCNF0_PLEN_32BITZERO
		            | RADIO_PCNF0_CRCINC_INCLUDE
		            | RADIO_PCNF0_TERMLEN_VAL(0);
		
		RADIO_PCNF1 = RADIO_PCNF1_MAXLEN_VAL(RADIO_MAX_PAYLOAD)
		            | RADIO_PCNF1_STATLEN_VAL(0)
		            | RADIO_PCNF1_BALEN_VAL(0)
		            | RADIO_PCNF1_ENDIAN_LITTLE
		            | RADIO_PCNF1_WHITEEN_DISABLED;
		
		radio_config_frequency_deviation_center();
	
	#else
		
		;
		
	#endif
}

static inline void radio_disable_all_interrupts(void){
/*
  Disables all interrupt sources of the RADIO peripheral.
*/
	
	RADIO_INTENCLR = RADIO_INTENCLR_READY_CLEAR
	               | RADIO_INTENCLR_ADDRESS_CLEAR
	               | RADIO_INTENCLR_PAYLOAD_CLEAR
	               | RADIO_INTENCLR_END_CLEAR
	               | RADIO_INTENCLR_DISABLED_CLEAR
	#if REVISION >= REVISION_F
	               | RADIO_INTENCLR_FRAMESTART_CLEAR
	               | RADIO_INTENCLR_EDEND_CLEAR
	               | RADIO_INTENCLR_EDSTOPPED_CLEAR
	               | RADIO_INTENCLR_CCAIDLE_CLEAR
	               | RADIO_INTENCLR_CCABUSY_CLEAR
	               | RADIO_INTENCLR_CCASTOPPED_CLEAR
	               | RADIO_INTENCLR_RATEBOOST_CLEAR
	               | RADIO_INTENCLR_TXREADY_CLEAR
	               | RADIO_INTENCLR_RXREADY_CLEAR
	               | RADIO_INTENCLR_MHRMATCH_CLEAR
	               | RADIO_INTENCLR_PHYEND_CLEAR
	               | RADIO_INTENCLR_CTEPRESENT_CLEAR
	#endif
	               | RADIO_INTENCLR_DEVMATCH_CLEAR
	               | RADIO_INTENCLR_DEVMISS_CLEAR
	               | RADIO_INTENCLR_RSSIEND_CLEAR
	               | RADIO_INTENCLR_BCMATCH_CLEAR
	               | RADIO_INTENCLR_CRCOK_CLEAR
	               | RADIO_INTENCLR_CRCERROR_CLEAR;
}

static inline void radio_disable_all_critical_ppis(void){
/*
  Disables all critical PPIs.
*/
	
	#if REVISION >= REVISION_F
		ppi_disable(PPI_RADIO_FRAMESTART_RADIO_RSSISTART);
	#endif
	ppi_disable(PPI_RADIO_CRCOK_TIMER1_START);
	ppi_disable(PPI_RADIO_CRCOK_TIMER1_CLEAR);
}

void radio_init(void){
/*
  Initializes the RADIO peripheral and OTA command forwarding/reception. May
  be called multiple times.
  
  General Packet Layout
  =====================
       
       _____________ _________________ _______________________
      |             |                 |                       |
      | S0 (1 byte) | LENGTH (1 byte) | PAYLOAD (<=255 bytes) |
      |_____________|_________________|_______________________|
      
      `-----.------´ `-------.-------´ `----------.----------´
            |                |                    |
         Packet       PAYLOAD length        Packet content
        content                         (second byte, third ...)
      (first byte)
     
*/
	
	static uint8_t recvData[RADIO_OTA_RECV_BUFFER_LENGTH]
	                       [RADIO_OTA_RECV_BUFFER_WIDTH];
	static uint8_t sendData[RADIO_OTA_SEND_BUFFER_LENGTH]
	                       [RADIO_OTA_SEND_BUFFER_WIDTH];
	static size_t recvSize[RADIO_OTA_RECV_BUFFER_LENGTH];
	static size_t sendSize[RADIO_OTA_RECV_BUFFER_LENGTH];
	static bool initialized = false;
	
	if(!initialized){
	
		/* Buffer initialization. */
		
		buffer_init_rows(&radioRecvBuffer,
		                 &recvData[0][0],
		                 recvSize,
		                 RADIO_OTA_RECV_BUFFER_LENGTH,
		                 RADIO_OTA_RECV_BUFFER_WIDTH);
		buffer_init_rows(&radioSendBuffer,
		                 &sendData[0][0],
		                 sendSize,
		                 RADIO_OTA_SEND_BUFFER_LENGTH,
		                 RADIO_OTA_SEND_BUFFER_WIDTH);
		
		/* Static RADIO settings. */
		
		radio_set_tx_power(RADIO_TX_POWER);
		
		RADIO_RXADDRESSES = RADIO_RXADDRESSES_ADDR0_ENABLED
		                  | RADIO_RXADDRESSES_ADDR1_ENABLED;
		
		RADIO_SFD = RADIO_IEEE_250KBIT_SFD;
		RADIO_DATAWHITEIV = 37;
		
		/* Workaround for some IEEE 802.15.4 related errata. */
		/* (See corresponding doc for more information.)     */
		errata_init();
	
		/* Initialize synchronization of the measurement scheduler. */
		
		ppi_init(PPI_RADIO_CRCOK_TIMER1_CLEAR);
		ppi_init(PPI_RADIO_CRCOK_TIMER1_START);
	
		/* Measurement timing (and hence RADIO timing) is critical. */
		
		nvic_enable_irq(NVIC_RADIO_IRQ);
		nvic_set_priority(NVIC_RADIO_IRQ, NVIC_RSS2I_HIGH_PRIORITY);
		
		/* Set up the TIMER that is used for generating */
		/* the delay between two RSSI conversions.      */
		
		TIMER_PRESCALER(TIMER2) = 0;
		TIMER_BITMODE(TIMER2) = TIMER_BITMODE_BITMODE_32Bit;
		TIMER_CC(TIMER2)[0] = RADIO_RSSI_SAMPLE_DELAY;
		TIMER_SHORTS(TIMER2) = TIMER_SHORTS_COMPARE0_CLEAR_ENABLED
		                     | TIMER_SHORTS_COMPARE0_STOP_ENABLED;
		
	#if REVISION >= REVISION_F
		ppi_init(PPI_RADIO_FRAMESTART_RADIO_RSSISTART);
	#endif
		ppi_init(PPI_TIMER2_COMPARE0_RADIO_RSSISTART);
		ppi_enable(PPI_TIMER2_COMPARE0_RADIO_RSSISTART);
		
		/* Set up the pin toggling in debug mode. */
		
		#if DEBUG_MEASUREMENT_FRAME > 1
		
			gpiote_init(GPIOTE_CHANNEL_DEBUG_PIN);
			ppi_init(PPI_RADIO_READY_GPIOTE_CLR_CHANNEL_DEBUG_PIN);
			ppi_enable(PPI_RADIO_READY_GPIOTE_CLR_CHANNEL_DEBUG_PIN);
		
		#endif
		
		
		cte_init();
		
		
		initialized = true;
	}
}

void radio_measure_ieee_250kbit_prepare_tx(uint8_t frequency,
                                           bool disrupt,
                                           uint8_t fI,
                                           int32_t cI){
/*
  Prepares the radio for transmitting an IEEE 250kBit measurement packet.
  
  Argument List
  =============
    
    {frequency}          Send the measurement packet with a center frequency of
                         {2400 + frequency} MHz.
    {disrupt}            Wheter to disrupt (force to enter control mode) other
                         devices or not.
    {fI}                 Index of the measurement packet's frame.
    {cI}                 Index of the measurement packet's cycle.
  
  Description
  ===========
  
    In contrast to {radio_measure_ieee_250kbit_init_tx()}, this is an early
    preparation stage where only those settings are applied that do not need the
    RADIO to be disabled (completely).
*/
	
#if REVISION >= REVISION_F

	HOOK_UNUSED(frequency);
	HOOK_UNUSED(disrupt);
	HOOK_UNUSED(fI);
	HOOK_UNUSED(cI);
	
	
	cte_disable();
	radio_config_ieee_250kbit();
	
	
	#if DEBUG_MEASUREMENT_FRAME >= 3
	
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
		
	#endif
	
#else
	
	HOOK_UNUSED(frequency);
	HOOK_UNUSED(disrupt);
	HOOK_UNUSED(fI);
	HOOK_UNUSED(cI);
	
	/* Avoid unsued function warnings. */
	radio_config_ieee_250kbit();
	
#endif

}

void radio_measure_ieee_250kbit_init_tx(uint8_t frequency,
                                        bool disrupt,
                                        uint8_t fI,
                                        int32_t cI){
/*
  Prepares the radio for transmitting an IEEE 250kBit measurement packet.
  
  Argument List
  =============
    
    {frequency}          Send the measurement packet with a center frequency of
                         {2400 + frequency} MHz.
    {disrupt}            Wheter to disrupt (force to enter control mode) other
                         devices or not.
    {fI}                 Index of the measurement packet's frame.
    {cI}                 Index of the measurement packet's cycle.
                         
  Packet Layout
  =============
       _________________ ______________________________________
      |                 |                                      |
      | LENGTH (1 byte) |           PAYLOAD (7 bytes)          |
      |_________________|______________________________________|
      
      `-------.--------´ `-------------------.----------------´
              |                              |
       PAYLOAD length         STATUS (1 byte), FRAME (1 byte),
                              CYCLE (3 bytes) and CRC (2 bytes)
    
  PAYLOAD Layout
  ==============
       _________________ ________________ _________________ _______________
      |                 |                |                 |               |
      | STATUS (1 byte) | FRAME (1 byte) | CYCLE (2 bytes) | CRC (2 bytes) |
      |_________________|________________|_________________|_______________|
    
    
  STATUS Layout
  ==============
       __________________ ____________________________________
      |                  |                                    |
      | Stop Bit (1 bit) | Battery and Memory Status (7 bits) |
      |__________________|____________________________________|
      
*/

#if REVISION >= REVISION_F

	uint8_t status;
	uint8_t stop;
	uint8_t batMem;
	
	/* Convenience definition */
	volatile uint8_t * rPD = radioPacketData;
	
	#if DEBUG_MEASUREMENT_FRAME >= 3
	
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
		
	#endif
	
	if(RADIO_STATE != RADIO_STATE_STATE_DISABLED){
		
		debug_report(DEBUG_RADIO_NOT_DISABLED);
	}
	else{
		
		radio_disable_all_interrupts();
		radio_disable_all_critical_ppis();
		
		rPD[RADIO_MEASURE_IEEE_250KBIT_LENGTH_DATA_OFFSET]
			= RADIO_MEASURE_IEEE_250KBIT_PAYLOAD_LENGTH;
		
		RADIO_PACKETPTR = (uint32_t) rPD;
		RADIO_FREQUENCY = frequency;
		
		#if DEBUG_MEASUREMENT_FRAME >= 2
		
			RADIO_INTENSET = RADIO_INTENSET_END_SET
			               | RADIO_INTENSET_DISABLED_SET;
			
		#endif
		
		RADIO_SHORTS = RADIO_SHORTS_DISPATCH_PHYEND_DISABLE_ENABLED;
		
		RADIO_EVENTS_DISPATCH_PHYEND = 0;
		RADIO_EVENTS_END = 0;
		RADIO_EVENTS_READY = 0;
		RADIO_EVENTS_DISABLED = 0;
		
		#if DEBUG_MEASUREMENT_FRAME >= 3

			GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
			GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
			
		#endif
		
		/* Ramp up RADIO TX. */
		
		radioState = RADIO_MEASURE_IEEE_250KBIT_TX;
		RADIO_TASKS_TXEN = 1;
		
		/* Prepare STOP/Disrupt bit. */
		stop = disrupt ? 1 : 0;
		
		/* Prepare battery charge and memory load values. */
		batMem = radioBatteryLevel / 10 * 11 + radioMemoryLoad / 10;
		
		/* Prepare status byte. */
		status = stop | (batMem << 1);
		
		/* Prepare radio packet. */
		
		rPD[RADIO_MEASURE_IEEE_250KBIT_STATUS_DATA_OFFSET]    = status;
		rPD[RADIO_MEASURE_IEEE_250KBIT_FRAME_DATA_OFFSET]     = fI;
		rPD[RADIO_MEASURE_IEEE_250KBIT_CYCLE_DATA_OFFSET + 0] = ((uint32_t) cI) >>  0;
		rPD[RADIO_MEASURE_IEEE_250KBIT_CYCLE_DATA_OFFSET + 1] = ((uint32_t) cI) >>  8;
		rPD[RADIO_MEASURE_IEEE_250KBIT_CYCLE_DATA_OFFSET + 2] = ((uint32_t) cI) >> 16;
		
		#if DEBUG_MEASUREMENT_FRAME >= 3
		
			GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
			GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
			
		#endif
	}
	
#else

	HOOK_UNUSED(frequency);
	HOOK_UNUSED(disrupt);
	HOOK_UNUSED(fI);
	HOOK_UNUSED(cI);
	
	/* Avoid unsued function warnings. */
	radio_config_ieee_250kbit();
	
#endif

}

void radio_measure_ble_2mbit_prepare_tx(uint8_t frequency,
                                        bool disrupt,
                                        uint8_t fI,
                                        int32_t cI,
                                        uint16_t * rRSSV,
                                        size_t rRSSC,
                                        CTEType cteType,
                                        CTEAntennaPattern cteAntPat){
/*
  Prepares the radio for transmitting a BLE 2MBit measurement packet.
  
  Argument List
  =============
    
    {frequency}          Send the measurement packet with a center frequency of
                         {2400 + frequency} MHz.
    {disrupt}            Wheter to disrupt (force to enter control mode) other
                         devices or not.
    {fI}                 Index of the measurement packet's frame.
    {cI}                 Index of the measurement packet's cycle.
    {rRSSV}              Address of RSS values to be sent along with the
                         measurement packet (a.k.a. Remote RSS).
                         Each RSS value is interpreted to be given as an ab-
                         solute (i.e. positive) value in multiples of 1/128 dBm.
    {rRSSC}              Number of RSS values to be sent along with the
                         measurement packet (a.k.a. Remote RSS).
    {cteType}            Type of the CTE (if any).
    {cteAntPat}          CTE antenna behaviour (if any).
  
  Description
  ===========
  
    In contrast to {radio_measure_ble_2mbit_init_tx()}, this is an early
    preparation stage where only those settings are applied that do not need the
    RADIO to be disabled (completely).
*/
	
	HOOK_UNUSED(frequency);
	HOOK_UNUSED(disrupt);
	HOOK_UNUSED(fI);
	HOOK_UNUSED(cI);
	HOOK_UNUSED(rRSSV);
	HOOK_UNUSED(rRSSC);

	/* Configure CTE. */
	cte_radio_init_tx(cteType, cteAntPat);
	/* Configure RADIO. */
	radio_config_ble_2mbit();
	
	#if DEBUG_MEASUREMENT_FRAME >= 3
	
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
		
	#endif
}

static void radio_measure_x_2mbit_init_tx(uint8_t frequency,
                                          bool disrupt,
                                          uint8_t fI,
                                          int32_t cI,
                                          uint16_t * rRSSV,
                                          size_t rRSSC,
                                          RadioBLEMode mode){
/*
  Prepares the radio for transmitting a 2MBit measurement packet. CTE
  configuration is applied in the corresponding {*_prepare_tx()} function.
  
  Argument List
  =============
    
    {frequency}          Send the measurement packet with a center frequency of
                         {2400 + frequency} MHz.
    {disrupt}            Wheter to disrupt (force to enter control mode) other
                         devices or not.
    {fI}                 Index of the measurement packet's frame.
    {cI}                 Index of the measurement packet's cycle.
    {rRSSV}              Address of RSS values to be sent along with the
                         measurement packet (a.k.a. Remote RSS).
                         Each RSS value is interpreted to be given as an ab-
                         solute (i.e. positive) value in multiples of 1/128 dBm.
    {rRSSC}              Number of RSS values to be sent along with the
                         measurement packet (a.k.a. Remote RSS).
    {mode}               Whether to ramp up Nordic proprietary or BLE mode.
                         
  Packet Layout
  ============= 
       _____________ _________________ _______________________
      |             |                 |                       |
      | S0 (1 byte) | LENGTH (1 byte) |  PAYLOAD (14 bytes)   |
      |_____________|_________________|_______________________|
      
      `-----.------´ `-------.-------´ `----------.----------´
            |                |                    |
          frame       PAYLOAD length       cycle (3 bytes) and
                                       remote RSS, battery level and
                                        memory load (11 bytes)
    
  PAYLOAD Layout
  ==============
       _________________ ________________________ 
      |                 |                        |
      | cycle (3 bytes) | remote data (11 bytes) |
      |_________________|________________________|
  
  Remote Data Layout
  ==================
   
       byte 0  byte 1   ...   byte 4  byte 5  byte 6   ...   byte 9  byte 10
       ______  ______         ______  ______  ______         ______  ______
      |      ||      || ... ||      ||      ||      || ... ||      ||      |
      |______||______||     ||______||______||______||     ||______||______|
       _____  _____  __     ___  _____  ____  ____  __       ____  ____  __
      |_____||_____||__ ... ___||_____||____||____||__ ... ||____||____||__|
         |      |                  |      |     |             |      |    |
         |      |                  |      |     |             |      |    |
   remote RSS 0 |                  |   {batMem} |     remote RSS 4   | {batMem}
   integer part |                  |    6 LSb   |    fractional part |  4 MSb
     (7 bits)   |                  |   (6 bits) |       (6 bits)     | (4 bits)
                |                  |            |                    |
          remote RSS 1      remote RSS 5        |              remote RSS 5
          integer part      integer part   remote RSS 0       fractional part
            (7 bits)         (7 bits)     fractional part        (6 bits)
                                             (6 bits)
*/
	
	size_t i;
	uint16_t batMem; /* battery and memory data       */
	uint64_t iBytes; /* remote data (integer part)    */
	uint64_t fBytes; /* remote data (fractional part) */
	
	/* Convenience definition */
	volatile uint8_t * rPD = radioPacketData;
	
	#if DEBUG_MEASUREMENT_FRAME >= 3
	
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
		
	#endif
	
	if(RADIO_STATE != RADIO_STATE_STATE_DISABLED){
		
		debug_report(DEBUG_RADIO_NOT_DISABLED);
	}
	else{
		
		radio_disable_all_interrupts();
		radio_disable_all_critical_ppis();
		
		rPD[RADIO_MEASURE_X_2MBIT_LENGTH_DATA_OFFSET]
			= RADIO_MEASURE_X_2MBIT_PAYLOAD_LENGTH;
		
		RADIO_PACKETPTR = (uint32_t) rPD;
		RADIO_FREQUENCY = frequency;
		
		radio_set_measure_access_addresses();
		RADIO_TXADDRESS = disrupt ? 1 : 0;
		
		#if DEBUG_MEASUREMENT_FRAME >= 2
		
			RADIO_INTENSET = RADIO_INTENSET_END_SET
			               | RADIO_INTENSET_DISABLED_SET;
			
		#endif
		
		RADIO_SHORTS = RADIO_SHORTS_DISPATCH_PHYEND_DISABLE_ENABLED;
		
		RADIO_EVENTS_DISPATCH_PHYEND = 0;
		RADIO_EVENTS_END = 0;
		RADIO_EVENTS_READY = 0;
		RADIO_EVENTS_DISABLED = 0;
		
		#if DEBUG_MEASUREMENT_FRAME >= 3

			GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
			GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
			
		#endif
		
		/* Ramp up RADIO TX. */
		
		radioState = (mode == RADIO_BLE_MODE_BLE)
		           ? RADIO_MEASURE_BLE_2MBIT_TX
		           : RADIO_MEASURE_NRF_2MBIT_TX;
		RADIO_TASKS_TXEN = 1;
		
		
		/* Setup remote data. */
		
		/* This is a bit risky but also necessary to reduce the RADIO ramp up */
		/* time: Configure the packet data after triggering the {TXEN} task.  */
		
		rPD[RADIO_MEASURE_X_2MBIT_FRAME_DATA_OFFSET]     = fI;
		rPD[RADIO_MEASURE_X_2MBIT_CYCLE_DATA_OFFSET + 0] = ((uint32_t) cI) >>  0;
		rPD[RADIO_MEASURE_X_2MBIT_CYCLE_DATA_OFFSET + 1] = ((uint32_t) cI) >>  8;
		rPD[RADIO_MEASURE_X_2MBIT_CYCLE_DATA_OFFSET + 2] = ((uint32_t) cI) >> 16;

		/* Prepare remote RSS values. */

		iBytes = 0;
		for(i = 0; i < 6 && i < rRSSC; ++i){
	
			iBytes |= ((uint64_t) ((rRSSV[i] / 128) & 0x7f)) << (7 * i);
		}

		fBytes = 0;
		for(i = 0; i < 6 && i < rRSSC; ++i){
	
			fBytes |= ((uint64_t) (((rRSSV[i] % 128) & 0x7e) >> 1)) << (6 * i);
		}

		/* Prepare battery charge and memory load values. */

		batMem = radioBatteryLevel / 5 * 51 + radioMemoryLoad / 2;

		/* Prepare packet data */

		for(i = 0; i < 6 && i < RADIO_MEASURE_X_2MBIT_PAYLOAD_LENGTH - 3; ++i){
	
			rPD[RADIO_MEASURE_X_2MBIT_REMOTE_DATA_OFFSET + i] = (iBytes >> (8 * i))
			                                                  & 0xff;
		}

		rPD[RADIO_MEASURE_X_2MBIT_REMOTE_DATA_OFFSET + 6 - 1] |= (batMem & 0x3f) << 2;

		for(i = 0; i < 5 && i < RADIO_MEASURE_X_2MBIT_PAYLOAD_LENGTH - 3 - 6; ++i){
	
			rPD[RADIO_MEASURE_X_2MBIT_REMOTE_DATA_OFFSET + 6 + i] = (fBytes >> (8 * i))
			                                                      & 0xff;
		}

		rPD[RADIO_MEASURE_X_2MBIT_REMOTE_DATA_OFFSET + 6 + 5 - 1] |= ((batMem >> 6)
		                                                            & 0x0f) << 4;
		
		#if DEBUG_MEASUREMENT_FRAME >= 3
		
			GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
			GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
			
		#endif
	}
}

void radio_measure_ble_2mbit_init_tx(uint8_t frequency,
                                     bool disrupt,
                                     uint8_t fI,
                                     int32_t cI,
                                     uint16_t * rRSSV,
                                     size_t rRSSC,
                                     CTEType cteType,
                                     CTEAntennaPattern cteAntPat){
/*
  Prepares the radio for transmitting a BLE 2MBit measurement packet.
  
  Argument List
  =============
    
    {frequency}          Send the measurement packet with a center frequency of
                         {2400 + frequency} MHz.
    {disrupt}            Wheter to disrupt (force to enter control mode) other
                         devices or not.
    {fI}                 Index of the measurement packet's frame.
    {cI}                 Index of the measurement packet's cycle.
    {rRSSV}              Address of RSS values to be sent along with the
                         measurement packet (a.k.a. Remote RSS).
                         Each RSS value is interpreted to be given as an ab-
                         solute (i.e. positive) value in multiples of 1/128 dBm.
    {rRSSC}              Number of RSS values to be sent along with the
                         measurement packet (a.k.a. Remote RSS).
    {cteType}            Type of the CTE (if any).
    {cteAntPat}          CTE antenna behaviour (if any).
                         
*/
	
	/* CTE config is done within {radio_measure_ble_2mbit_prepare_tx()}. */
	HOOK_UNUSED(cteType);
	HOOK_UNUSED(cteAntPat);
	
	radio_measure_x_2mbit_init_tx(
		frequency,
		disrupt,
		fI,
		cI,
		rRSSV,
		rRSSC,
		RADIO_BLE_MODE_BLE
	);
}

void radio_measure_nrf_2mbit_prepare_tx(uint8_t frequency,
                                        bool disrupt,
                                        uint8_t fI,
                                        int32_t cI,
                                        uint16_t * rRSSV,
                                        size_t rRSSC){
/*
  Prepares the radio for transmitting an NRF 2MBit measurement packet.
  
  Argument List
  =============
    
    {frequency}          Send the measurement packet with a center frequency of
                         {2400 + frequency} MHz.
    {disrupt}            Wheter to disrupt (force to enter control mode) other
                         devices or not.
    {fI}                 Index of the measurement packet's frame.
    {cI}                 Index of the measurement packet's cycle.
    {rRSSV}              Address of RSS values to be sent along with the
                         measurement packet (a.k.a. Remote RSS).
                         Each RSS value is interpreted to be given as an ab-
                         solute (i.e. positive) value in multiples of 1/128 dBm.
    {rRSSC}              Number of RSS values to be sent along with the
                         measurement packet (a.k.a. Remote RSS).
  
  Description
  ===========
  
    In contrast to {radio_measure_nrf_2mbit_init_tx()}, this is an early
    preparation stage where only those settings are applied that do not need the
    RADIO to be disabled (completely).
*/
	
	HOOK_UNUSED(frequency);
	HOOK_UNUSED(disrupt);
	HOOK_UNUSED(fI);
	HOOK_UNUSED(cI);
	HOOK_UNUSED(rRSSV);
	HOOK_UNUSED(rRSSC);

	/* Configure RADIO. */
	radio_config_nrf_2mbit();
	
	#if DEBUG_MEASUREMENT_FRAME >= 3
	
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
		
	#endif
}

void radio_measure_nrf_2mbit_init_tx(uint8_t frequency,
                                     bool disrupt,
                                     uint8_t fI,
                                     int32_t cI,
                                     uint16_t * rRSSV,
                                     size_t rRSSC){
/*
  Prepares the radio for transmitting an NRF 2MBit measurement packet.
  
  Argument List
  =============
    
    {frequency}          Send the measurement packet with a center frequency of
                         {2400 + frequency} MHz.
    {disrupt}            Wheter to disrupt (force to enter control mode) other
                         devices or not.
    {fI}                 Index of the measurement packet's frame.
    {cI}                 Index of the measurement packet's cycle.
    {rRSSV}              Address of RSS values to be sent along with the
                         measurement packet (a.k.a. Remote RSS).
                         Each RSS value is interpreted to be given as an ab-
                         solute (i.e. positive) value in multiples of 1/128 dBm.
    {rRSSC}              Number of RSS values to be sent along with the
                         measurement packet (a.k.a. Remote RSS).
                         
*/
	
	radio_measure_x_2mbit_init_tx(
		frequency,
		disrupt,
		fI,
		cI,
		rRSSV,
		rRSSC,
		RADIO_BLE_MODE_NRF
	);
}

void radio_measure_ble_1mbit_prepare_tx(uint8_t frequency,
                                        bool disrupt,
                                        uint8_t fI,
                                        int32_t cI,
                                        uint16_t rRSSV,
                                        CTEType cteType,
                                        CTEAntennaPattern cteAntPat){
/*
  Prepares the radio for transmitting a BLE 1MBit measurement packet.
  
  Argument List
  =============
    
    {frequency}          Send the measurement packet with a center frequency of
                         {2400 + frequency} MHz.
    {disrupt}            Wheter to disrupt (force to enter control mode) other
                         devices or not.
    {fI}                 Index of the measurement packet's frame.
    {cI}                 Index of the measurement packet's cycle.
    {rRSSV}              A single RSS value to be sent along with the
                         measurement packet (a.k.a. Remote RSS).
                         This RSS value is interpreted to be given as an ab-
                         solute (i.e. positive) value in multiples of 1/128 dBm.
    {cteType}            Type of the CTE (if any).
    {cteAntPat}          CTE antenna behaviour (if any).
  
  Description
  ===========
  
    In contrast to {radio_measure_ble_1mbit_init_tx()}, this is an early
    preparation stage where only those settings are applied that do not need the
    RADIO to be disabled (completely).
*/
	
	HOOK_UNUSED(frequency);
	HOOK_UNUSED(disrupt);
	HOOK_UNUSED(fI);
	HOOK_UNUSED(cI);
	HOOK_UNUSED(rRSSV);

	/* Configure CTE. */
	cte_radio_init_tx(cteType, cteAntPat);
	/* Configure RADIO. */
	radio_config_ble_1mbit();
	
	#if DEBUG_MEASUREMENT_FRAME >= 3
	
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
		
	#endif
}

static void radio_measure_x_1mbit_init_tx(uint8_t frequency,
                                          bool disrupt,
                                          uint8_t fI,
                                          int32_t cI,
                                          uint16_t rRSSV,
                                          RadioBLEMode mode){
/*
  Prepares the radio for transmitting a 1MBit measurement packet. CTE
  configuration is applied in the corresponding {*_prepare_tx()} function.
  
  Argument List
  =============
    
    {frequency}          Send the measurement packet with a center frequency of
                         {2400 + frequency} MHz.
    {disrupt}            Wheter to disrupt (force to enter control mode) other
                         devices or not.
    {fI}                 Index of the measurement packet's frame.
    {cI}                 Index of the measurement packet's cycle.
    {rRSSV}              A single RSS value to be sent along with the
                         measurement packet (a.k.a. Remote RSS).
                         This RSS value is interpreted to be given as an ab-
                         solute (i.e. positive) value in multiples of 1/128 dBm.
    {mode}               Whether to ramp up Nordic proprietary or BLE mode.
                         
  Packet Layout
  ============= 
       _____________ _________________ _____________________
      |             |                 |                     |
      | S0 (1 byte) | LENGTH (1 byte) |  PAYLOAD (5 bytes)  |
      |_____________|_________________|_____________________|
      
      `-----.------´ `-------.-------´ `---------.---------´
            |                |                   |
          frame       PAYLOAD length     cycle (3 bytes) and
                                     remote RSS, battery level and
                                        memory load (2 bytes)
    
  PAYLOAD Layout
  ==============
       _________________ _______________________
      |                 |                       |
      | cycle (3 bytes) | remote data (2 bytes) |
      |_________________|_______________________|
  
  Remote Data Layout
  ==================
   
       byte 0     byte 1   
       _________  _________  
      |         ||         ||
      |_________||_________||
       _______  __  _______
      |_______||__||_______||
         |       |     |
         |       |     |
    remote RSS   | Battery stat
   integer part  | and memory
     (7 bits)    |   load
                 | (7 bits)
                 |
            remote RSS
          fractional part
             (2 bits)
   
*/
	
	uint16_t batMem; /* battery and memory data       */
	uint8_t iP;      /* remote data (integer part)    */
	uint8_t fP;      /* remote data (fractional part) */
	
	/* Convenience definition */
	volatile uint8_t * rPD = radioPacketData;
	
	#if DEBUG_MEASUREMENT_FRAME >= 3
	
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
		
	#endif
	
	if(RADIO_STATE != RADIO_STATE_STATE_DISABLED){
		
		debug_report(DEBUG_RADIO_NOT_DISABLED);
	}
	else{
		
		radio_disable_all_interrupts();
		radio_disable_all_critical_ppis();
		
		rPD[RADIO_MEASURE_X_1MBIT_LENGTH_DATA_OFFSET]
			= RADIO_MEASURE_X_1MBIT_PAYLOAD_LENGTH;
		
		RADIO_PACKETPTR = (uint32_t) rPD;
		RADIO_FREQUENCY = frequency;
		
		radio_set_measure_access_addresses();
		RADIO_TXADDRESS = disrupt ? 1 : 0;
		
		#if DEBUG_MEASUREMENT_FRAME >= 2
		
			RADIO_INTENSET = RADIO_INTENSET_END_SET
			               | RADIO_INTENSET_DISABLED_SET;
			
		#endif
		
		RADIO_SHORTS = RADIO_SHORTS_DISPATCH_PHYEND_DISABLE_ENABLED;
		
		RADIO_EVENTS_DISPATCH_PHYEND = 0;
		RADIO_EVENTS_END = 0;
		RADIO_EVENTS_READY = 0;
		RADIO_EVENTS_DISABLED = 0;
		
		#if DEBUG_MEASUREMENT_FRAME >= 3

			GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
			GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
			
		#endif
		
		/* Ramp up RADIO TX. */
		
		radioState = (mode == RADIO_BLE_MODE_BLE)
		           ? RADIO_MEASURE_BLE_1MBIT_TX
		           : RADIO_MEASURE_NRF_1MBIT_TX;
		RADIO_TASKS_TXEN = 1;
		
		
		/* Setup remote data. */
		
		/* This is a bit risky but also necessary to reduce the RADIO ramp up */
		/* time: Configure the packet data after triggering the {TXEN} task.  */
		
		rPD[RADIO_MEASURE_X_1MBIT_FRAME_DATA_OFFSET]     = fI;
		rPD[RADIO_MEASURE_X_1MBIT_CYCLE_DATA_OFFSET + 0] = ((uint32_t) cI) >>  0;
		rPD[RADIO_MEASURE_X_1MBIT_CYCLE_DATA_OFFSET + 1] = ((uint32_t) cI) >>  8;
		rPD[RADIO_MEASURE_X_1MBIT_CYCLE_DATA_OFFSET + 2] = ((uint32_t) cI) >> 16;

		/* Prepare remote RSS values. */

		iP = (rRSSV / 128) & 0x7f;
		fP = ((rRSSV % 128) & 0x7f) >> 5;
		batMem = radioBatteryLevel / 10 * 11 + radioMemoryLoad / 10;
		
		/* Prepare packet data */

		rPD[RADIO_MEASURE_X_1MBIT_REMOTE_DATA_OFFSET + 0] = iP | (fP << 7);
		rPD[RADIO_MEASURE_X_1MBIT_REMOTE_DATA_OFFSET + 1] = (fP >> 1)
		                                                  | (batMem << 1);
		
		#if DEBUG_MEASUREMENT_FRAME >= 3
		
			GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
			GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
			
		#endif
	}
}

void radio_measure_ble_1mbit_init_tx(uint8_t frequency,
                                     bool disrupt,
                                     uint8_t fI,
                                     int32_t cI,
                                     uint16_t rRSSV,
                                     CTEType cteType,
                                     CTEAntennaPattern cteAntPat){
/*
  Prepares the radio for transmitting a BLE 1MBit measurement packet.
  
  Argument List
  =============
    
    {frequency}          Send the measurement packet with a center frequency of
                         {2400 + frequency} MHz.
    {disrupt}            Wheter to disrupt (force to enter control mode) other
                         devices or not.
    {fI}                 Index of the measurement packet's frame.
    {cI}                 Index of the measurement packet's cycle.
    {rRSSV}              A single RSS value to be sent along with the
                         measurement packet (a.k.a. Remote RSS).
                         This RSS value is interpreted to be given as an ab-
                         solute (i.e. positive) value in multiples of 1/128 dBm.
    {cteType}            Type of the CTE (if any).
    {cteAntPat}          CTE antenna behaviour (if any).
                         
*/
	
	/* CTE config is done within {radio_measure_ble_1mbit_prepare_tx()}. */
	HOOK_UNUSED(cteType);
	HOOK_UNUSED(cteAntPat);
	
	radio_measure_x_1mbit_init_tx(
		frequency,
		disrupt,
		fI,
		cI,
		rRSSV,
		RADIO_BLE_MODE_BLE
	);
}

void radio_measure_nrf_1mbit_prepare_tx(uint8_t frequency,
                                        bool disrupt,
                                        uint8_t fI,
                                        int32_t cI,
                                        uint16_t rRSSV){
/*
  Prepares the radio for transmitting an NRF 1MBit measurement packet.
  
  Argument List
  =============
    
    {frequency}          Send the measurement packet with a center frequency of
                         {2400 + frequency} MHz.
    {disrupt}            Wheter to disrupt (force to enter control mode) other
                         devices or not.
    {fI}                 Index of the measurement packet's frame.
    {cI}                 Index of the measurement packet's cycle.
    {rRSSV}              A single RSS value to be sent along with the
                         measurement packet (a.k.a. Remote RSS).
                         This RSS value is interpreted to be given as an ab-
                         solute (i.e. positive) value in multiples of 1/128 dBm.
  
  Description
  ===========
  
    In contrast to {radio_measure_nrf_1mbit_init_tx()}, this is an early
    preparation stage where only those settings are applied that do not need the
    RADIO to be disabled (completely).
*/
	
	HOOK_UNUSED(frequency);
	HOOK_UNUSED(disrupt);
	HOOK_UNUSED(fI);
	HOOK_UNUSED(cI);
	HOOK_UNUSED(rRSSV);

	/* Configure RADIO. */
	radio_config_nrf_1mbit();
	
	#if DEBUG_MEASUREMENT_FRAME >= 3
	
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
		
	#endif
}

void radio_measure_nrf_1mbit_init_tx(uint8_t frequency,
                                     bool disrupt,
                                     uint8_t fI,
                                     int32_t cI,
                                     uint16_t rRSSV){
/*
  Prepares the radio for transmitting an NRF 1MBit measurement packet.
  
  Argument List
  =============
    
    {frequency}          Send the measurement packet with a center frequency of
                         {2400 + frequency} MHz.
    {disrupt}            Wheter to disrupt (force to enter control mode) other
                         devices or not.
    {fI}                 Index of the measurement packet's frame.
    {cI}                 Index of the measurement packet's cycle.
    {rRSSV}              A single RSS value to be sent along with the
                         measurement packet (a.k.a. Remote RSS).
                         This RSS value is interpreted to be given as an ab-
                         solute (i.e. positive) value in multiples of 1/128 dBm.
*/
	
	radio_measure_x_1mbit_init_tx(
		frequency,
		disrupt,
		fI,
		cI,
		rRSSV,
		RADIO_BLE_MODE_NRF
	);
}

void radio_measure_start_tx(void){
/*
  Starts transmission of a previously initiated/prepared measurement packet.
*/
	
	if(radioState == RADIO_MEASURE_BLE_2MBIT_TX
	|| radioState == RADIO_MEASURE_NRF_2MBIT_TX
#if REVISION >= REVISION_F
	|| radioState == RADIO_MEASURE_IEEE_250KBIT_TX
#endif
	|| radioState == RADIO_MEASURE_BLE_1MBIT_TX
	|| radioState == RADIO_MEASURE_NRF_1MBIT_TX){
	
		if(RADIO_STATE == RADIO_STATE_STATE_TXIDLE){
			
			RADIO_EVENTS_READY = 0;
			RADIO_TASKS_START = 1;
			
			#if DEBUG_MEASUREMENT_FRAME >= 2
				/* Additionally, the debug pin is cleared when the RADIO is */
				/* ready to transmit a packet via PPI and GPIOTE.           */
				GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
				GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
			#endif
		}
		else{
		
			debug_report(DEBUG_RADIO_TX_NOT_READY);
		}
	}
	else{
		
		debug_report(DEBUG_RADIO_TX_UNPREPARED);
	}
}

uint8_t radio_measure_ieee_250kbit_get_frame(volatile uint8_t * p){
/*
  Given the address of a received IEEE 250kBit measurement radio {p}acket, it
  returns the frame index.
*/
	
	return p[RADIO_MEASURE_IEEE_250KBIT_FRAME_DATA_OFFSET + 0];
}

static uint8_t radio_measure_x_2mbit_get_frame(volatile uint8_t * p){
/*
  Given the address of a received 2MBit measurement radio {p}acket, it returns
  the frame index.
*/
	
	return p[RADIO_MEASURE_X_2MBIT_FRAME_DATA_OFFSET + 0];
}

uint8_t radio_measure_ble_2mbit_get_frame(volatile uint8_t * p){
/*
  Given the address of a received BLE 2MBit measurement radio {p}acket, it
  returns the frame index.
*/
	
	return radio_measure_x_2mbit_get_frame(p);
}

uint8_t radio_measure_nrf_2mbit_get_frame(volatile uint8_t * p){
/*
  Given the address of a received NRF 2MBit measurement radio {p}acket, it
  returns the frame index.
*/
	
	return radio_measure_x_2mbit_get_frame(p);
}

static uint8_t radio_measure_x_1mbit_get_frame(volatile uint8_t * p){
/*
  Given the address of a received 1MBit measurement radio {p}acket, it returns
  the frame index.
*/
	
	return p[RADIO_MEASURE_X_1MBIT_FRAME_DATA_OFFSET + 0];
}

uint8_t radio_measure_ble_1mbit_get_frame(volatile uint8_t * p){
/*
  Given the address of a received BLE 1MBit measurement radio {p}acket, it
  returns the frame index.
*/
	
	return radio_measure_x_1mbit_get_frame(p);
}

uint8_t radio_measure_nrf_1mbit_get_frame(volatile uint8_t * p){
/*
  Given the address of a received NRF 1MBit measurement radio {p}acket, it
  returns the frame index.
*/
	
	return radio_measure_x_1mbit_get_frame(p);
}

int32_t radio_measure_ieee_250kbit_get_cycle(volatile uint8_t * p){
/*
  Given the address of a received IEEE 250kBit measurement radio {p}acket, it
  returns the cycle index.
*/
	
	return (int32_t)
	  (((uint32_t) p[RADIO_MEASURE_IEEE_250KBIT_CYCLE_DATA_OFFSET + 0] <<  0)
	 | ((uint32_t) p[RADIO_MEASURE_IEEE_250KBIT_CYCLE_DATA_OFFSET + 1] <<  8)
	 | ((uint32_t) p[RADIO_MEASURE_IEEE_250KBIT_CYCLE_DATA_OFFSET + 2] << 16));
}

static int32_t radio_measure_x_2mbit_get_cycle(volatile uint8_t * p){
/*
  Given the address of a received 2MBit measurement radio {p}acket, it returns
  the cycle index.
*/
	
	return (int32_t)
	  (((uint32_t) p[RADIO_MEASURE_X_2MBIT_CYCLE_DATA_OFFSET + 0] <<  0)
	 | ((uint32_t) p[RADIO_MEASURE_X_2MBIT_CYCLE_DATA_OFFSET + 1] <<  8)
	 | ((uint32_t) p[RADIO_MEASURE_X_2MBIT_CYCLE_DATA_OFFSET + 2] << 16));
}

int32_t radio_measure_ble_2mbit_get_cycle(volatile uint8_t * p){
/*
  Given the address of a received BLE 2MBit measurement radio {p}acket, it
  returns the cycle index.
*/
	
	return radio_measure_x_2mbit_get_cycle(p);
}

int32_t radio_measure_nrf_2mbit_get_cycle(volatile uint8_t * p){
/*
  Given the address of a received NRF 2MBit measurement radio {p}acket, it
  returns the cycle index.
*/
	
	return radio_measure_x_2mbit_get_cycle(p);
}

static int32_t radio_measure_x_1mbit_get_cycle(volatile uint8_t * p){
/*
  Given the address of a received 1MBit measurement radio {p}acket, it returns
  the cycle index.
*/
	
	return (int32_t)
	  (((uint32_t) p[RADIO_MEASURE_X_1MBIT_CYCLE_DATA_OFFSET + 0] <<  0)
	 | ((uint32_t) p[RADIO_MEASURE_X_1MBIT_CYCLE_DATA_OFFSET + 1] <<  8)
	 | ((uint32_t) p[RADIO_MEASURE_X_1MBIT_CYCLE_DATA_OFFSET + 2] << 16));
}

int32_t radio_measure_ble_1mbit_get_cycle(volatile uint8_t * p){
/*
  Given the address of a received BLE 1MBit measurement radio {p}acket, it
  returns the cycle index.
*/
	
	return radio_measure_x_1mbit_get_cycle(p);
}

int32_t radio_measure_nrf_1mbit_get_cycle(volatile uint8_t * p){
/*
  Given the address of a received NRF 1MBit measurement radio {p}acket, it
  returns the cycle index.
*/
	
	return radio_measure_x_1mbit_get_cycle(p);
}

uint8_t radio_measure_ieee_250kbit_get_remote_battery_level(volatile uint8_t * p){
/*
  Given the address of a received IEEE 250kBit measurement radio {p}acket, it
  returns the remote battery level.
  
  Please refer to the documentation of {radio_measure_ieee_250kbit_init_tx()} in
  order to get an overview about the status bit layout.
*/
	
	uint8_t batMem = p[RADIO_MEASURE_IEEE_250KBIT_STATUS_DATA_OFFSET] >> 1;
	
	return (batMem / 11) * 10;
}

static uint8_t radio_measure_x_2mbit_get_remote_battery_level(volatile uint8_t * p){
/*
  Given the address of a received 2MBit measurement radio {p}acket, it returns
  the remote battery level.
  
  Please refer to the documentation of {radio_measure_x_2mbit_init_tx()} in
  order to get an overview about the remote data bit layout.
*/
	
	uint16_t batMem =
	  (((uint16_t) p[RADIO_MEASURE_X_2MBIT_REMOTE_DATA_OFFSET + 5]  >> 2) << 0)
	| (((uint16_t) p[RADIO_MEASURE_X_2MBIT_REMOTE_DATA_OFFSET + 10] >> 4) << 6);
	
	return (batMem / 51) * 5;
}

uint8_t radio_measure_ble_2mbit_get_remote_battery_level(volatile uint8_t * p){
/*
  Given the address of a received BLE 2MBit measurement radio {p}acket, it
  returns the remote battery level.
*/
	
	return radio_measure_x_2mbit_get_remote_battery_level(p);
}

uint8_t radio_measure_nrf_2mbit_get_remote_battery_level(volatile uint8_t * p){
/*
  Given the address of a received NRF 2MBit measurement radio {p}acket, it
  returns the remote battery level.
*/
	
	return radio_measure_x_2mbit_get_remote_battery_level(p);
}

static uint8_t radio_measure_x_1mbit_get_remote_battery_level(volatile uint8_t * p){
/*
  Given the address of a received 1MBit measurement radio {p}acket, it returns
  the remote battery level.
  
  Please refer to the documentation of {radio_measure_x_1mbit_init_tx()} in
  order to get an overview about the remote data bit layout.
*/
	
	uint8_t batMem = p[RADIO_MEASURE_X_1MBIT_REMOTE_DATA_OFFSET + 1]  >> 1;
	
	return (batMem / 11) * 10;
}

uint8_t radio_measure_ble_1mbit_get_remote_battery_level(volatile uint8_t * p){
/*
  Given the address of a received BLE 1MBit measurement radio {p}acket, it
  returns the remote battery level.
*/
	
	return radio_measure_x_1mbit_get_remote_battery_level(p);
}

uint8_t radio_measure_nrf_1mbit_get_remote_battery_level(volatile uint8_t * p){
/*
  Given the address of a received NRF 1MBit measurement radio {p}acket, it
  returns the remote battery level.
*/
	
	return radio_measure_x_1mbit_get_remote_battery_level(p);
}

uint8_t radio_measure_ieee_250kbit_get_remote_memory_load(volatile uint8_t * p){
/*
  Given the address of a received IEEE 250kBit measurement radio {p}acket, it
  returns the remote memory load.
  
  Please refer to the documentation of {radio_measure_ieee_250kbit_init_tx()} in
  order to get an overview about the status data bit layout.
*/
	
	uint8_t batMem = p[RADIO_MEASURE_IEEE_250KBIT_STATUS_DATA_OFFSET] >> 1;
	
	return (batMem % 11) * 10;
}

static uint8_t radio_measure_x_2mbit_get_remote_memory_load(volatile uint8_t * p){
/*
  Given the address of a received 2MBit measurement radio {p}acket, it returns
  the remote memory load.
  
  Please refer to the documentation of {radio_measure_x_2mbit_init_tx()} in
  order to get an overview about the remote data bit layout.
*/
	
	uint16_t batMem =
	  (((uint16_t) p[RADIO_MEASURE_X_2MBIT_REMOTE_DATA_OFFSET + 5]  >> 2) << 0)
	| (((uint16_t) p[RADIO_MEASURE_X_2MBIT_REMOTE_DATA_OFFSET + 10] >> 4) << 6);
	
	return (batMem % 51) * 2;
}

uint8_t radio_measure_ble_2mbit_get_remote_memory_load(volatile uint8_t * p){
/*
  Given the address of a received BLE 2MBit measurement radio {p}acket, it
  returns the remote memory load.
*/
	
	return radio_measure_x_2mbit_get_remote_memory_load(p);
}

uint8_t radio_measure_nrf_2mbit_get_remote_memory_load(volatile uint8_t * p){
/*
  Given the address of a received NRF 2MBit measurement radio {p}acket, it
  returns the remote memory load.
*/
	
	return radio_measure_x_2mbit_get_remote_memory_load(p);
}

static uint8_t radio_measure_x_1mbit_get_remote_memory_load(volatile uint8_t * p){
/*
  Given the address of a received 1MBit measurement radio {p}acket, it returns
  the remote memory load.
  
  Please refer to the documentation of {radio_measure_x_1mbit_init_tx()} in
  order to get an overview about the remote data bit layout.
*/
	
	uint8_t batMem = p[RADIO_MEASURE_X_1MBIT_REMOTE_DATA_OFFSET + 1]  >> 1;
	
	return (batMem % 11) * 10;
}

uint8_t radio_measure_ble_1mbit_get_remote_memory_load(volatile uint8_t * p){
/*
  Given the address of a received BLE 1MBit measurement radio {p}acket, it
  returns the remote memory load.
*/
	
	return radio_measure_x_1mbit_get_remote_memory_load(p);
}

uint8_t radio_measure_nrf_1mbit_get_remote_memory_load(volatile uint8_t * p){
/*
  Given the address of a received NRF 1MBit measurement radio {p}acket, it
  returns the remote memory load.
*/
	
	return radio_measure_x_1mbit_get_remote_memory_load(p);
}

static uint8_t radio_measure_x_2mbit_get_remote_rss_integer_part(volatile uint8_t * p,
                                                                 uint8_t chan){
/*
  Given the address of a received 2MBit measurement radio {p}acket, it returns
  the integer part of the specified remote {chan}nel.
  
  Please refer to the documentation of {radio_measure_x_2mbit_init_tx()} in
  order to get an overview about the remote data bit layout.
*/
	
	uint64_t integerBytes =
	  ((uint64_t) p[RADIO_MEASURE_X_2MBIT_REMOTE_DATA_OFFSET + 0] <<  0)
	| ((uint64_t) p[RADIO_MEASURE_X_2MBIT_REMOTE_DATA_OFFSET + 1] <<  8)
	| ((uint64_t) p[RADIO_MEASURE_X_2MBIT_REMOTE_DATA_OFFSET + 2] << 16)
	| ((uint64_t) p[RADIO_MEASURE_X_2MBIT_REMOTE_DATA_OFFSET + 3] << 24)
	| ((uint64_t) p[RADIO_MEASURE_X_2MBIT_REMOTE_DATA_OFFSET + 4] << 32)
	| ((uint64_t) p[RADIO_MEASURE_X_2MBIT_REMOTE_DATA_OFFSET + 5] << 40);
	
	return (integerBytes >> (7 * chan)) & 0x7f;
}

uint8_t radio_measure_ble_2mbit_get_remote_rss_integer_part(volatile uint8_t * p,
                                                            uint8_t chan){
/*
  Given the address of a received BLE 2MBit measurement radio {p}acket, it
  returns the integer part of the specified remote {chan}nel.
*/
	
	return radio_measure_x_2mbit_get_remote_rss_integer_part(p, chan);
}

uint8_t radio_measure_nrf_2mbit_get_remote_rss_integer_part(volatile uint8_t * p,
                                                            uint8_t chan){
/*
  Given the address of a received NRF 2MBit measurement radio {p}acket, it
  returns the integer part of the specified remote {chan}nel.
*/
	
	return radio_measure_x_2mbit_get_remote_rss_integer_part(p, chan);
}

static uint8_t radio_measure_x_1mbit_get_remote_rss_integer_part(volatile uint8_t * p){
/*
  Given the address of a received 1MBit measurement radio {p}acket, it returns
  the integer part of the transmitted remote value. Note, that in contrast to
  the 2MBit RADIO modes, there is only a single remote value shipped along a
  radio packet.
  
  Please refer to the documentation of {radio_measure_x_1mbit_init_tx()} in
  order to get an overview about the remote data bit layout.
*/
	
	return p[RADIO_MEASURE_X_1MBIT_REMOTE_DATA_OFFSET + 0] & 0x7f;
}

uint8_t radio_measure_ble_1mbit_get_remote_rss_integer_part(volatile uint8_t * p){
/*
  Given the address of a received BLE 1MBit measurement radio {p}acket, it
  returns the integer part of the transmitted remote value. Note, that in
  contrast to the BLE 2MBit RADIO mode, there is only a single remote value
  shipped along a radio packet.
*/
	
	return radio_measure_x_1mbit_get_remote_rss_integer_part(p);
}

uint8_t radio_measure_nrf_1mbit_get_remote_rss_integer_part(volatile uint8_t * p){
/*
  Given the address of a received NRF 1MBit measurement radio {p}acket, it
  returns the integer part of the transmitted remote value. Note, that in
  contrast to the NRF 2MBit RADIO mode, there is only a single remote value
  shipped along a radio packet.
*/
	
	return radio_measure_x_1mbit_get_remote_rss_integer_part(p);
}

static uint8_t radio_measure_x_2mbit_get_remote_rss_fractional_part(volatile uint8_t * p,
                                                                    uint8_t chan){
/*
  Given the address of a received 2MBit measurement radio {p}acket, it returns
  the fractional part of the specified remote {chan}nel.
  
  Please refer to the documentation of {radio_measure_x_2mbit_init_tx()} in
  order to get an overview about the remote data bit layout.
*/
	
	uint64_t fractionalBytes =
	  ((uint64_t) p[RADIO_MEASURE_X_2MBIT_REMOTE_DATA_OFFSET + 6]  <<  0)
	| ((uint64_t) p[RADIO_MEASURE_X_2MBIT_REMOTE_DATA_OFFSET + 7]  <<  8)
	| ((uint64_t) p[RADIO_MEASURE_X_2MBIT_REMOTE_DATA_OFFSET + 8]  << 16)
	| ((uint64_t) p[RADIO_MEASURE_X_2MBIT_REMOTE_DATA_OFFSET + 9]  << 24)
	| ((uint64_t) p[RADIO_MEASURE_X_2MBIT_REMOTE_DATA_OFFSET + 10] << 32);
	
	return ((fractionalBytes >> (6 * chan)) << 1) & 0x7f;
}

uint8_t radio_measure_ble_2mbit_get_remote_rss_fractional_part(volatile uint8_t * p,
                                                               uint8_t chan){
/*
  Given the address of a received BLE 2MBit measurement radio {p}acket, it
  returns the fractional part of the specified remote {chan}nel.
*/
	
	return radio_measure_x_2mbit_get_remote_rss_fractional_part(p, chan);
}

uint8_t radio_measure_nrf_2mbit_get_remote_rss_fractional_part(volatile uint8_t * p,
                                                               uint8_t chan){
/*
  Given the address of a received NRF 2MBit measurement radio {p}acket, it
  returns the fractional part of the specified remote {chan}nel.
*/
	
	return radio_measure_x_2mbit_get_remote_rss_fractional_part(p, chan);
}

static uint8_t radio_measure_x_1mbit_get_remote_rss_fractional_part(volatile uint8_t * p){
/*
  Given the address of a received 1MBit measurement radio {p}acket, it returns
  the fractional part of the transmitted remote value. Note, that in contrast to
  the 2MBit RADIO modes, there is only a single remote value shipped along a
  radio packet.
  
  Please refer to the documentation of {radio_measure_x_1mbit_init_tx()} in
  order to get an overview about the remote data bit layout.
*/
	
	uint16_t fractionalBytes =
	  ((uint16_t) p[RADIO_MEASURE_X_1MBIT_REMOTE_DATA_OFFSET + 0]  <<  0)
	| ((uint16_t) p[RADIO_MEASURE_X_1MBIT_REMOTE_DATA_OFFSET + 1]  <<  8);
	
	return ((fractionalBytes >> 7) & 0x03) << 5;
}

uint8_t radio_measure_ble_1mbit_get_remote_rss_fractional_part(volatile uint8_t * p){
/*
  Given the address of a received BLE 1MBit measurement radio {p}acket, it
  returns the fractional part of the transmitted remote value. Note, that in
  contrast to the BLE 1MBit RADIO mode, there is only a single remote value
  shipped along a radio packet.
*/
	
	return radio_measure_x_1mbit_get_remote_rss_fractional_part(p);
}

uint8_t radio_measure_nrf_1mbit_get_remote_rss_fractional_part(volatile uint8_t * p){
/*
  Given the address of a received NRF 1MBit measurement radio {p}acket, it
  returns the fractional part of the transmitted remote value. Note, that in
  contrast to the NRF 1MBit RADIO mode, there is only a single remote value
  shipped along a radio packet.
*/
	
	return radio_measure_x_1mbit_get_remote_rss_fractional_part(p);
}

uint8_t * radio_measure_ble_2mbit_get_iq_sample_address(uint8_t * p){
/*
  Given the address of a received BLE 2MBit measurement radio {p}acket, it
  returns the address of the first IQ sample taken during that received packet.
*/
	
	return p + RADIO_MEASURE_X_2MBIT_CTE_SAMPLES_OFFSET;
}

uint8_t * radio_measure_ble_1mbit_get_iq_sample_address(uint8_t * p){
/*
  Given the address of a received BLE 1MBit measurement radio {p}acket, it
  returns the address of the first IQ sample taken during that received packet.
*/
	
	return p + RADIO_MEASURE_X_1MBIT_CTE_SAMPLES_OFFSET;
}

void radio_trigger(){
/*
    - Turns the {YELLOW} LED off after some timeout.
    - Updates {radioBatteryLevel}.
    - Updates {radioMemoryLoad}.
*/
	
	uint8_t level, load;
	
	if(!measure_running()
	&& radioLEDTimeout
	&& radioLEDTimeout++ == RADIO_LED_TIMEOUT){
		
		led_set(LED_YELLOW, LED_OFF);
		radioLEDTimeout = 0;
		radioControlSendLedOn = false;
	}
	
	load = memory_get_load();
	if(load > 100){
	
		load = 100;
	}
	radioMemoryLoad = load;
	
	level = battery_get_level();
	if(level >= 100){
	
		level = 99;
	}
	radioBatteryLevel = level;
}

void radio_control_init_send(void){
/*
  Sends a control packet (if any).
*/
	
	uint8_t * controlData;
	
	if(buffer_next_size(&radioSendBuffer)){
	
		radioSendBusy = true;
		
		/* Make sure that a simultaneous DISABLE */
		/* event triggers this function.         */
		radioState = RADIO_PREPARE_SEND;
		RADIO_INTENSET = RADIO_INTENSET_DISABLED_SET;
	
		cte_disable();
		
		if(RADIO_STATE != RADIO_STATE_STATE_DISABLED){
			
			/* Disable the RADIO. */
			/* The ISR will recall this function. */
			RADIO_EVENTS_DISABLED = 0;
			RADIO_TASKS_DISABLE = 1;
		}
		else{
			
			radio_disable_all_interrupts();
			radio_disable_all_critical_ppis();
			
			controlData = buffer_get_row(&radioSendBuffer);
			
			led_set(LED_YELLOW, radioControlSendLedOn ? LED_OFF : LED_ON);
			radioControlSendLedOn = radioControlSendLedOn ? false : true;
			radioLEDTimeout = 1;
			
			RADIO_SHORTS = RADIO_SHORTS_DISPATCH_PHYEND_DISABLE_ENABLED
			             | RADIO_SHORTS_READY_START_ENABLED;
			
			RADIO_INTENSET = RADIO_INTENSET_DISABLED_SET;
			RADIO_FREQUENCY = RADIO_CONTROL_FREQUENCY;
			
			radio_set_control_access_addresses(controlData[0]);
			RADIO_TXADDRESS = 1;
			
			RADIO_PACKETPTR = (uint32_t) &controlData[1];
			
			RADIO_EVENTS_DISPATCH_PHYEND = 0;
			RADIO_EVENTS_READY = 0;
			RADIO_EVENTS_DISABLED = 0;
			
			radioState = RADIO_CONTROL_SEND;
	
			radio_config_ble_2mbit();
			
			RADIO_TASKS_TXEN = 1;
		}
	}
	else{
		
		radioSendBusy = false;
		radio_control_init_recv();
	}
}

void radio_measure_ieee_250kbit_prepare_rx(uint8_t frequency,
                                           bool compatibility,
                                           volatile uint8_t * packetDataAddr){
/*
  Prepares the radio for receiving an IEEE 250kBit measurement packet on the
  radio channel with the specified {frequency}.
  
  Description
  ===========
  
    In contrast to {radio_measure_ieee_250kbit_init_rx()}, this is an early
    preparation stage where only those settings are applied that do not need the
    RADIO to be disabled (completely).
  
  Argument List
  =============
    
    {frequency}          Listen for the measurement packet on the channel with a
                         center frequency of {2400 + frequency} MHz.
    {compatibility}      Whether to restrict the RSS measurements onto a
                         specific part of the measurement packet in order to
                         comply with the measurement duration of a 2MBit packet.
    {packetDataAddr}     Place to store the RADIO packet.
*/
	
#if REVISION >= REVISION_F

	HOOK_UNUSED(frequency);
	HOOK_UNUSED(compatibility);
	HOOK_UNUSED(packetDataAddr);
	
	
	cte_disable();
	radio_config_ieee_250kbit();
	
	
	#if DEBUG_MEASUREMENT_FRAME >= 3
	
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
		
	#endif

#else

	HOOK_UNUSED(frequency);
	HOOK_UNUSED(compatibility);
	HOOK_UNUSED(packetDataAddr);
	
	/* Avoid unsued function warnings. */
	radio_config_ieee_250kbit();
	
#endif

}

void radio_measure_ieee_250kbit_init_rx(uint8_t frequency,
                                        bool compatibility,
                                        volatile uint8_t * packetDataAddr){
/*
  Prepares the radio for receiving an IEEE 250kBit measurement packet on the
  radio channel with the specified {frequency}.
  
  Argument List
  =============
    
    {frequency}          Listen for the measurement packet on the channel with a
                         center frequency of {2400 + frequency} MHz.
    {compatibility}      Whether to restrict the RSS measurements onto a
                         specific part of the measurement packet in order to
                         comply with the measurement duration of a 2MBit packet.
    {packetDataAddr}     Place to store the RADIO packet.
*/

#if REVISION >= REVISION_F

	static uint8_t cachedFrequency;
	static bool cachedCompatibility;
	static volatile uint8_t * cachedPacketDataAddr;
	
	
	#if DEBUG_MEASUREMENT_FRAME >= 3
	
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
		
	#endif
	
	
	/* Remember the given frequency, {compatibility} and       */
	/* {packetDataAddr} for the case a DISABLE task is needed. */
	
	if(frequency){
		
		/* TODO: Is it really necessary to save all these things?   */
		/*       Isn't it possible to apply these changes directly? */
		cachedFrequency = frequency;
		cachedCompatibility = compatibility;
		cachedPacketDataAddr = packetDataAddr;
	}
	
	/* Make sure that a simultaneous DISABLE event triggers this function. */
	radioState = RADIO_PREPARE_IEEE_250KBIT_RX;
	RADIO_INTENSET = RADIO_INTENSET_DISABLED_SET;
	
	if(RADIO_STATE != RADIO_STATE_STATE_DISABLED){
		
		/* Disable the RADIO. */
		/* The ISR will recall this function. */
		RADIO_EVENTS_DISABLED = 0;
		RADIO_TASKS_DISABLE = 1;
	}
	else{
		
		radio_disable_all_interrupts();
		radio_disable_all_critical_ppis();
		
		ppi_enable(PPI_RADIO_FRAMESTART_RADIO_RSSISTART);
		
		radioRSSAvg = 0;
		radioRSSCnt = 0;
		
		RADIO_SHORTS = RADIO_SHORTS_DISPATCH_PHYEND_DISABLE_ENABLED
		             | RADIO_SHORTS_FRAMESTART_BCSTART_ENABLED
		             | RADIO_SHORTS_READY_START_ENABLED;
		
		RADIO_EVENTS_DISPATCH_PHYEND = 0;
		/* Necessary for RSSI sampling requests (see {radio_isr()}). */
		RADIO_EVENTS_END = 0;
		RADIO_EVENTS_READY = 0;
		/* Necessary for stopping the RSSI sampling in compatibility mode. */
		RADIO_EVENTS_BCMATCH = 0;
		RADIO_EVENTS_ADDRESS = 0;
		RADIO_EVENTS_RSSIEND = 0;
		RADIO_EVENTS_DISABLED = 0;
		RADIO_EVENTS_CRCOK = 0;
		
		RADIO_INTENSET = RADIO_INTENSET_RSSIEND_SET
		#if DEBUG_MEASUREMENT_FRAME >= 2
		               | RADIO_INTENSET_CRCOK_SET
		#endif
		               | RADIO_INTENSET_DISABLED_SET;
		
		if(!cachedFrequency){
			
			debug_report(DEBUG_RADIO_INVALID_PARAMETER);
			cachedFrequency = RADIO_CONTROL_FREQUENCY;
			cachedPacketDataAddr = (uint8_t *) radioPacketData;
			cachedCompatibility = false;
		}
		
		RADIO_FREQUENCY = cachedFrequency;
		cachedFrequency = 0;
				
		RADIO_PACKETPTR = (uint32_t) cachedPacketDataAddr;
		
		/* Configure compatibility mode.      */
		/* Stop RSS sampling after some time. */
		RADIO_BCC = cachedCompatibility
		          ? RADIO_MEASURE_DURATION_IEEE_250KBIT_COMPATIBILITY_MODE
		          : RADIO_MEASURE_DURATION_IEEE_250KBIT_ERRATA_WORKAROUND;
		
		/* Reset measurement schedule once a valid */
		/* measurement packet was received.        */
		ppi_enable(PPI_RADIO_CRCOK_TIMER1_CLEAR);
		ppi_enable(PPI_RADIO_CRCOK_TIMER1_START);
		
		#if DEBUG_MEASUREMENT_FRAME >= 3
		
			GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
			GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
			
		#endif
		
		/* Ramp up RADIO RX. */
		
		radioState = RADIO_MEASURE_IEEE_250KBIT_RX;
		RADIO_TASKS_RXEN = 1;
	}

#else
	
	HOOK_UNUSED(frequency);
	HOOK_UNUSED(compatibility);
	HOOK_UNUSED(packetDataAddr);
	
	/* Avoid unsued function warnings. */
	radio_config_ieee_250kbit();
	
#endif

}

void radio_measure_ble_2mbit_prepare_rx(uint8_t frequency,
                                        bool compatibility,
                                        volatile uint8_t * packetDataAddr,
                                        CTEType cteType,
                                        CTEAntennaPattern cteAntPat){
/*
  Prepares the radio for receiving a BLE 2MBit measurement packet on the
  radio channel with the specified {frequency}.
  
  Description
  ===========
  
    In contrast to {radio_measure_ble_2mbit_init_rx()}, this is an early
    preparation stage where only those settings are applied that do not need the
    RADIO to be disabled (completely).
  
  Argument List
  =============
    
    {frequency}          Listen for the measurement packet on the channel with a
                         center frequency of {2400 + frequency} MHz.
    {compatibility}      Whether to restrict the RSS measurements onto a
                         specific part of the measurement packet in order to
                         comply with the measurement duration of a 2MBit packet.
                         THIS ARGUMENT HAS NO AFFECT!
    {packetDataAddr}     Place to store the RADIO packet and possible CTE IQ
                         samples (if any).
    {cteType}            Type of the CTE (if any).
    {cteAntPat}          CTE antenna behaviour (if any).
*/
	
	HOOK_UNUSED(frequency);
	HOOK_UNUSED(compatibility);
	
	/* Configure CTE. */
	cte_radio_init_rx(
		cteType,
		cteAntPat,
		packetDataAddr + RADIO_MEASURE_X_2MBIT_CTE_SAMPLES_OFFSET
	);
	
	/* Configure RADIO. */
	radio_config_ble_2mbit();
	
	#if DEBUG_MEASUREMENT_FRAME >= 3
	
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
		
	#endif
}

static void radio_measure_x_2mbit_init_rx(uint8_t frequency,
                                          bool compatibility,
                                          volatile uint8_t * packetDataAddr,
                                          RadioBLEMode mode){
/*
  Initializes the radio for receiving a 2MBit measurement packet on the
  radio channel with the specified {frequency}.
  
  Argument List
  =============
    
    {frequency}          Listen for the measurement packet on the channel with a
                         center frequency of {2400 + frequency} MHz.
    {compatibility}      Whether to restrict the RSS measurements onto a
                         specific part of the measurement packet in order to
                         comply with the measurement duration of a 2MBit packet.
                         THIS ARGUMENT HAS NO AFFECT!
    {packetDataAddr}     Place to store the RADIO packet.
    {mode}               Whether to ramp up Nordic proprietary or BLE mode.
*/
	
	static uint8_t cachedFrequency;
	static RadioBLEMode cachedMode;
	static volatile uint8_t * cachedPacketDataAddr;
	
	/* There is no need for a compatibility mode since THIS IS the BLE mode. */
	HOOK_UNUSED(compatibility);
	
	#if DEBUG_MEASUREMENT_FRAME >= 3
	
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
		
	#endif
	
	/* Remember the given frequency, {mode} and {packetDataAddr} */
	/* for the case a DISABLE task is needed.                    */
	
	if(frequency){
		
		/* TODO: Is it really necessary to save all these things?   */
		/*       Isn't it possible to apply these changes directly? */
		cachedFrequency = frequency;
		cachedMode = mode;
		cachedPacketDataAddr = packetDataAddr;
	}
	
	/* Make sure that a simultaneous DISABLE event triggers this function. */
	radioState = RADIO_PREPARE_X_2MBIT_RX;
	RADIO_INTENSET = RADIO_INTENSET_DISABLED_SET;
	
	if(RADIO_STATE != RADIO_STATE_STATE_DISABLED){
		
		/* Disable the RADIO. */
		/* The ISR will recall this function. */
		RADIO_EVENTS_DISABLED = 0;
		RADIO_TASKS_DISABLE = 1;
	}
	else{
		
		radio_disable_all_interrupts();
		radio_disable_all_critical_ppis();
		
		
		radioRSSAvg = 0;
		radioRSSCnt = 0;
		
		RADIO_SHORTS = RADIO_SHORTS_DISPATCH_PHYEND_DISABLE_ENABLED
		             | RADIO_SHORTS_READY_START_ENABLED
		             | RADIO_SHORTS_ADDRESS_RSSISTART_ENABLED;
		
		RADIO_INTENSET = RADIO_INTENSET_RSSIEND_SET
		             #if DEBUG_MEASUREMENT_FRAME >= 2
		               | RADIO_INTENSET_CRCOK_SET
		             #endif
		               | RADIO_INTENSET_DISABLED_SET;
		
		if(!cachedFrequency){
			
			debug_report(DEBUG_RADIO_INVALID_PARAMETER);
			cachedFrequency = RADIO_CONTROL_FREQUENCY;
			cachedMode = RADIO_BLE_MODE_BLE;
			cachedPacketDataAddr = (uint8_t *) radioPacketData;
		}
		
		RADIO_FREQUENCY = cachedFrequency;
		cachedFrequency = 0;
		
		radio_set_measure_access_addresses();
		
		RADIO_PACKETPTR = (uint32_t) cachedPacketDataAddr;
		
		RADIO_EVENTS_DISPATCH_PHYEND = 0;
		/* Necessary for RSSI sampling requests (see {radio_isr()}). */
		RADIO_EVENTS_END = 0;
		RADIO_EVENTS_READY = 0;
		RADIO_EVENTS_ADDRESS = 0;
		RADIO_EVENTS_RSSIEND = 0;
		RADIO_EVENTS_DISABLED = 0;
		RADIO_EVENTS_CRCOK = 0;
		
		/* Reset measurement schedule once a valid */
		/* measurement packet was received.        */
		ppi_enable(PPI_RADIO_CRCOK_TIMER1_CLEAR);
		ppi_enable(PPI_RADIO_CRCOK_TIMER1_START);
		
		
		#if DEBUG_MEASUREMENT_FRAME >= 3
		
			GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
			GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
			
		#endif
		
		/* Ramp up RADIO RX. */
		
		radioState = (cachedMode == RADIO_BLE_MODE_BLE)
		           ? RADIO_MEASURE_BLE_2MBIT_RX
		           : RADIO_MEASURE_NRF_2MBIT_RX;
		RADIO_TASKS_RXEN = 1;
	}
}

void radio_measure_ble_2mbit_init_rx(uint8_t frequency,
                                     bool compatibility,
                                     volatile uint8_t * packetDataAddr,
                                     CTEType cteType,
                                     CTEAntennaPattern cteAntPat){
/*
  Initializes the radio for receiving a BLE 2MBit measurement packet on the
  radio channel with the specified {frequency}.
  
  Argument List
  =============
    
    {frequency}          Listen for the measurement packet on the channel with a
                         center frequency of {2400 + frequency} MHz.
    {compatibility}      Whether to restrict the RSS measurements onto a
                         specific part of the measurement packet in order to
                         comply with the measurement duration of a 2MBit packet.
                         THIS ARGUMENT HAS NO AFFECT!
    {packetDataAddr}     Place to store the RADIO packet and possible CTE IQ
                         samples (if any).
    {cteType}            Type of the CTE (if any).
    {cteAntPat}          CTE antenna behaviour (if any).
*/
	
	/* CTE config is applied in the corresponding preparation function. */
	HOOK_UNUSED(cteType);
	HOOK_UNUSED(cteAntPat);
	
	radio_measure_x_2mbit_init_rx(
		frequency,
		compatibility,
		packetDataAddr,
		RADIO_BLE_MODE_BLE
	);
}

void radio_measure_nrf_2mbit_prepare_rx(uint8_t frequency,
                                        bool compatibility,
                                        volatile uint8_t * packetDataAddr){
/*
  Prepares the radio for receiving an NRF 2MBit measurement packet on the
  radio channel with the specified {frequency}.
  
  Description
  ===========
  
    In contrast to {radio_measure_nrf_2mbit_init_rx()}, this is an early
    preparation stage where only those settings are applied that do not need the
    RADIO to be disabled (completely).
  
  Argument List
  =============
    
    {frequency}          Listen for the measurement packet on the channel with a
                         center frequency of {2400 + frequency} MHz.
    {compatibility}      Whether to restrict the RSS measurements onto a
                         specific part of the measurement packet in order to
                         comply with the measurement duration of a 2MBit packet.
                         THIS ARGUMENT HAS NO AFFECT!
    {packetDataAddr}     Place to store the RADIO packet and possible CTE IQ
                         samples (if any).
*/
	
	HOOK_UNUSED(frequency);
	HOOK_UNUSED(compatibility);
	HOOK_UNUSED(packetDataAddr);
	
	/* Configure RADIO. */
	radio_config_nrf_2mbit();
	
	#if DEBUG_MEASUREMENT_FRAME >= 3
	
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
		
	#endif
}

void radio_measure_nrf_2mbit_init_rx(uint8_t frequency,
                                     bool compatibility,
                                     volatile uint8_t * packetDataAddr){
/*
  Initializes the radio for receiving an NRF 2MBit measurement packet on the
  radio channel with the specified {frequency}.
  
  Argument List
  =============
    
    {frequency}          Listen for the measurement packet on the channel with a
                         center frequency of {2400 + frequency} MHz.
    {compatibility}      Whether to restrict the RSS measurements onto a
                         specific part of the measurement packet in order to
                         comply with the measurement duration of a 2MBit packet.
                         THIS ARGUMENT HAS NO AFFECT!
    {packetDataAddr}     Place to store the RADIO packet.
*/
	
	radio_measure_x_2mbit_init_rx(
		frequency,
		compatibility,
		packetDataAddr,
		RADIO_BLE_MODE_NRF
	);
}

void radio_measure_ble_1mbit_prepare_rx(uint8_t frequency,
                                        bool compatibility,
                                        volatile uint8_t * packetDataAddr,
                                        CTEType cteType,
                                        CTEAntennaPattern cteAntPat){
/*
  Prepares the radio for receiving a BLE 1MBit measurement packet on the
  radio channel with the specified {frequency}.
  
  Description
  ===========
  
    In contrast to {radio_measure_ble_1mbit_init_rx()}, this is an early
    preparation stage where only those settings are applied that do not need the
    RADIO to be disabled (completely).
  
  Argument List
  =============
    
    {frequency}          Listen for the measurement packet on the channel with a
                         center frequency of {2400 + frequency} MHz.
    {compatibility}      Whether to restrict the RSS measurements onto a
                         specific part of the measurement packet in order to
                         comply with the measurement duration of a 2MBit packet.
    {packetDataAddr}     Place to store the RADIO packet and possible CTE IQ
                         samples (if any).
    {cteType}            Type of the CTE (if any).
    {cteAntPat}          CTE antenna behaviour (if any).
*/
	
	HOOK_UNUSED(frequency);
	HOOK_UNUSED(compatibility);
	
	/* Configure CTE. */
	cte_radio_init_rx(
		cteType,
		cteAntPat,
		packetDataAddr + RADIO_MEASURE_X_1MBIT_CTE_SAMPLES_OFFSET
	);
	
	/* Configure RADIO. */
	radio_config_ble_1mbit();
	
	#if DEBUG_MEASUREMENT_FRAME >= 3
	
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
		
	#endif
}

static void radio_measure_x_1mbit_init_rx(uint8_t frequency,
                                          bool compatibility,
                                          volatile uint8_t * packetDataAddr,
                                          RadioBLEMode mode){
/*
  Initializes the radio for receiving a 1MBit measurement packet on the
  radio channel with the specified {frequency}.
  
  Argument List
  =============
    
    {frequency}          Listen for the measurement packet on the channel with a
                         center frequency of {2400 + frequency} MHz.
    {compatibility}      Whether to restrict the RSS measurements onto a
                         specific part of the measurement packet in order to
                         comply with the measurement duration of a 2MBit packet.
    {packetDataAddr}     Place to store the RADIO packet.
    {mode}               Whether to ramp up Nordic proprietary or BLE mode.
*/
	
	static uint8_t cachedFrequency;
	static bool cachedCompatibility;
	static volatile uint8_t * cachedPacketDataAddr;
	static RadioBLEMode cachedMode;
	
	/* There is no need for a compatibility mode since THIS IS the BLE mode. */
	HOOK_UNUSED(compatibility);
	
	#if DEBUG_MEASUREMENT_FRAME >= 3
	
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
		
	#endif
	
	/* Remember the given frequency, {mode} and {packetDataAddr} */
	/* for the case a DISABLE task is needed.                    */
	
	if(frequency){
		
		/* TODO: Is it really necessary to save all these things?   */
		/*       Isn't it possible to apply these changes directly? */
		cachedFrequency = frequency;
		cachedCompatibility = compatibility;
		cachedPacketDataAddr = packetDataAddr;
		cachedMode = mode;
	}
	
	/* Make sure that a simultaneous DISABLE event triggers this function. */
	radioState = RADIO_PREPARE_X_1MBIT_RX;
	RADIO_INTENSET = RADIO_INTENSET_DISABLED_SET;
	
	if(RADIO_STATE != RADIO_STATE_STATE_DISABLED){
		
		/* Disable the RADIO. */
		/* The ISR will recall this function. */
		RADIO_EVENTS_DISABLED = 0;
		RADIO_TASKS_DISABLE = 1;
	}
	else{
		
		radio_disable_all_interrupts();
		radio_disable_all_critical_ppis();
		
		if(!cachedFrequency){
			
			debug_report(DEBUG_RADIO_INVALID_PARAMETER);
			cachedFrequency = RADIO_CONTROL_FREQUENCY;
			cachedCompatibility = false;
			cachedPacketDataAddr = (uint8_t *) radioPacketData;
			cachedMode = RADIO_BLE_MODE_BLE;
		}
		
		
		radioRSSAvg = 0;
		radioRSSCnt = 0;
		
		if(cachedCompatibility){
			
			/* Stop RSSI measurements after some time. */
			
			RADIO_SHORTS = RADIO_SHORTS_DISPATCH_PHYEND_DISABLE_ENABLED
			             | RADIO_SHORTS_READY_START_ENABLED
			             | RADIO_SHORTS_ADDRESS_RSSISTART_ENABLED
			             | RADIO_SHORTS_ADDRESS_BCSTART_ENABLED;
			
			RADIO_BCC = RADIO_MEASURE_DURATION_X_1MBIT_COMPATIBILITY_MODE;
		}
		else{
			
			RADIO_SHORTS = RADIO_SHORTS_DISPATCH_PHYEND_DISABLE_ENABLED
			             | RADIO_SHORTS_READY_START_ENABLED
			             | RADIO_SHORTS_ADDRESS_RSSISTART_ENABLED;
		}
		
		RADIO_INTENSET = RADIO_INTENSET_RSSIEND_SET
		             #if DEBUG_MEASUREMENT_FRAME >= 2
		               | RADIO_INTENSET_CRCOK_SET
		             #endif
		               | RADIO_INTENSET_DISABLED_SET;
		
		RADIO_FREQUENCY = cachedFrequency;
		cachedFrequency = 0;
		
		radio_set_measure_access_addresses();
		
		RADIO_PACKETPTR = (uint32_t) cachedPacketDataAddr;
		
		RADIO_EVENTS_DISPATCH_PHYEND = 0;
		/* Necessary for RSSI sampling requests (see {radio_isr()}). */
		RADIO_EVENTS_END = 0;
		RADIO_EVENTS_READY = 0;
		RADIO_EVENTS_ADDRESS = 0;
		/* Necessary for stopping the RSSI sampling in compatibility mode. */
		RADIO_EVENTS_BCMATCH = 0;
		RADIO_EVENTS_RSSIEND = 0;
		RADIO_EVENTS_DISABLED = 0;
		RADIO_EVENTS_CRCOK = 0;
		
		/* Reset measurement schedule once a valid */
		/* measurement packet was received.        */
		ppi_enable(PPI_RADIO_CRCOK_TIMER1_CLEAR);
		ppi_enable(PPI_RADIO_CRCOK_TIMER1_START);
		
		
		#if DEBUG_MEASUREMENT_FRAME >= 3
		
			GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
			GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
			
		#endif
		
		/* Ramp up RADIO RX. */
		
		radioState = (cachedMode == RADIO_BLE_MODE_BLE)
		           ? RADIO_MEASURE_BLE_1MBIT_RX
		           : RADIO_MEASURE_NRF_1MBIT_RX;
		RADIO_TASKS_RXEN = 1;
	}
}

void radio_measure_ble_1mbit_init_rx(uint8_t frequency,
                                     bool compatibility,
                                     volatile uint8_t * packetDataAddr,
                                     CTEType cteType,
                                     CTEAntennaPattern cteAntPat){
/*
  Initializes the radio for receiving a BLE 1MBit measurement packet on the
  radio channel with the specified {frequency}.
  
  Argument List
  =============
    
    {frequency}          Listen for the measurement packet on the channel with a
                         center frequency of {2400 + frequency} MHz.
    {compatibility}      Whether to restrict the RSS measurements onto a
                         specific part of the measurement packet in order to
                         comply with the measurement duration of a 2MBit packet.
    {packetDataAddr}     Place to store the RADIO packet and possible CTE IQ
                         samples (if any).
    {cteType}            Type of the CTE (if any).
    {cteAntPat}          CTE antenna behaviour (if any).
*/
	
	/* CTE config is applied in the corresponding preparation function. */
	HOOK_UNUSED(cteType);
	HOOK_UNUSED(cteAntPat);
	
	radio_measure_x_1mbit_init_rx(
		frequency,
		compatibility,
		packetDataAddr,
		RADIO_BLE_MODE_BLE
	);
}

void radio_measure_nrf_1mbit_prepare_rx(uint8_t frequency,
                                        bool compatibility,
                                        volatile uint8_t * packetDataAddr){
/*
  Prepares the radio for receiving an NRF 1MBit measurement packet on the
  radio channel with the specified {frequency}.
  
  Description
  ===========
  
    In contrast to {radio_measure_nrf_1mbit_init_rx()}, this is an early
    preparation stage where only those settings are applied that do not need the
    RADIO to be disabled (completely).
  
  Argument List
  =============
    
    {frequency}          Listen for the measurement packet on the channel with a
                         center frequency of {2400 + frequency} MHz.
    {compatibility}      Whether to restrict the RSS measurements onto a
                         specific part of the measurement packet in order to
                         comply with the measurement duration of a 2MBit packet.
    {packetDataAddr}     Place to store the RADIO packet and possible CTE IQ
                         samples (if any).
*/
	
	HOOK_UNUSED(frequency);
	HOOK_UNUSED(compatibility);
	HOOK_UNUSED(packetDataAddr);
	
	/* Configure RADIO. */
	radio_config_nrf_1mbit();
	
	#if DEBUG_MEASUREMENT_FRAME >= 3
	
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
		
	#endif
}

void radio_measure_nrf_1mbit_init_rx(uint8_t frequency,
                                     bool compatibility,
                                     volatile uint8_t * packetDataAddr){
/*
  Initializes the radio for receiving an NRF 1MBit measurement packet on the
  radio channel with the specified {frequency}.
  
  Argument List
  =============
    
    {frequency}          Listen for the measurement packet on the channel with a
                         center frequency of {2400 + frequency} MHz.
    {compatibility}      Whether to restrict the RSS measurements onto a
                         specific part of the measurement packet in order to
                         comply with the measurement duration of a 2MBit packet.
    {packetDataAddr}     Place to store the RADIO packet.
*/
	
	radio_measure_x_1mbit_init_rx(
		frequency,
		compatibility,
		packetDataAddr,
		RADIO_BLE_MODE_NRF
	);
}

void radio_emit_constant(

	uint8_t frequency,
	RadioMode mode,
	RadioFrequencyDeviation deviation,
	SwitchAntenna ant
){
/*
  Prepares the radio for emitting a specifiable carrier {frequency} with the
  specified frequency {deviation} using the specified {ant}enna in the specified
  RADIO {mode}.
  
  Argument List
  =============
    
    {frequency}    Carrier frequency minus 2400 in MHz.
    {deviation}    Deviation of the carrier frequency.
                     
                     {RADIO_FREQUENCY_DEVIATION_TRANS0} Negative frequency
                                                        deviation, i.e.
                                                        transmitting 0 over air.
                     {RADIO_FREQUENCY_DEVIATION_CENTER} No frequency deviation,
                                                        i.e. emitting carrier.
                     {RADIO_FREQUENCY_DEVIATION_TRANS1} Positive frequency
                                                        deviation, i.e.
                                                        transmitting 1 over air.
                                                        
    {mode}         Used RADIO mode. Note, that a frequency {deviation} unequal
                   to {RADIO_FREQUENCY_DEVIATION_CENTER} is not allowed in
                   {RADIO_MODE_IEEE_250KBIT}.
    {ant}          Used antenna.
  
  Behaviour in Case of Misconfiguration
  =====================================
    
    If the specified RADIO {mode} is invalid or if the specified frequency
    {deviation} is not supported by the specified RADIO {mode}, the RADIO won't
    get configured.
    
*/
	
	static uint8_t cachedFrequency;
	static RadioFrequencyDeviation cachedDev;
	static RadioMode cachedMode;
	static SwitchAntenna cachedAnt;
	
	/* Remember the given arguments for the case a DISABLE task is needed. */
	
	if(frequency){
		
		cachedFrequency = frequency;
		cachedDev = deviation;
		cachedMode = mode;
		cachedAnt = ant;
	}
	
	/* Make sure that a simultaneous DISABLE event triggers this function. */
	radioState = RADIO_PREPARE_CONSTANT;
	RADIO_INTENSET = RADIO_INTENSET_DISABLED_SET;
	
	if(RADIO_STATE != RADIO_STATE_STATE_DISABLED){
		
		/* Disable the RADIO. */
		/* The ISR will recall this function. */
		RADIO_EVENTS_DISABLED = 0;
		RADIO_TASKS_DISABLE = 1;
	}
	else{
	
		RADIO_SHORTS = 0;
		radio_disable_all_interrupts();
		radio_disable_all_critical_ppis();
		
		if(!cachedFrequency){
			
			debug_report(DEBUG_RADIO_INVALID_PARAMETER);
			cachedFrequency = RADIO_CONTROL_FREQUENCY;
			cachedAnt = switch_get_standard_antenna();
			cachedMode = RADIO_MODE_BLE_2MBIT;
			cachedDev = RADIO_FREQUENCY_DEVIATION_CENTER;
		}
		
		cte_disable();
		switch_set(switch_get_pattern(cachedAnt));
		
		RADIO_FREQUENCY = cachedFrequency;
		cachedFrequency = 0;
		
		RADIO_PACKETPTR = (uint32_t) radioPacketData;
		
		RADIO_EVENTS_DISPATCH_PHYEND = 0;
		RADIO_EVENTS_READY = 0;
		RADIO_EVENTS_ADDRESS = 0;
		RADIO_EVENTS_RSSIEND = 0;
		RADIO_EVENTS_DISABLED = 0;
		RADIO_EVENTS_CRCOK = 0;
		
		switch(cachedMode){
			
			case RADIO_MODE_BLE_2MBIT:
			
				radio_config_ble_2mbit();
				
			break;
			case RADIO_MODE_BLE_1MBIT:
			
				radio_config_ble_1mbit();
				
			break;
			case RADIO_MODE_NRF_2MBIT:
			
				radio_config_nrf_2mbit();
				
			break;
			case RADIO_MODE_NRF_1MBIT:
			
				radio_config_nrf_1mbit();
				
			break;
			case RADIO_MODE_IEEE_250KBIT:
			
				radio_config_ieee_250kbit();
				
			break;
			default:
				
				debug_report(DEBUG_RADIO_INVALID_CONST_CONF);
				return;
		}
		
		switch(cachedDev){
			
			case RADIO_FREQUENCY_DEVIATION_CENTER:
				
				/* Yes, this is redundant. */
				radio_config_frequency_deviation_center();
			
			break;
			case RADIO_FREQUENCY_DEVIATION_TRANS1:
				
				if(cachedMode == RADIO_MODE_IEEE_250KBIT){
					
					debug_report(DEBUG_RADIO_INVALID_CONST_CONF);
					return;
				}
					
				radio_config_frequency_deviation_trans1();
			
			break;
			case RADIO_FREQUENCY_DEVIATION_TRANS0:
				
				if(cachedMode == RADIO_MODE_IEEE_250KBIT){
					
					debug_report(DEBUG_RADIO_INVALID_CONST_CONF);
					return;
				}
					
				radio_config_frequency_deviation_trans0();
			
			break;
			default:
				
				debug_report(DEBUG_RADIO_INVALID_CONST_CONF);
				return;
		}
		
		radioState = RADIO_EMIT_CONSTANT;
		RADIO_TASKS_TXEN = 1;
	}
}

void radio_emit_carrier(uint8_t frequency, SwitchAntenna ant){
/*
  Prepares the radio for emitting a specifiable carrier frequency using
  the specified antenna.
*/

	radio_emit_constant(
	
		frequency,
		/* The actual RADIO mode is unimportant */
		/* since all carriers look the same.    */
		/* (Verified by measurement.)           */
		RADIO_MODE_BLE_2MBIT,
		RADIO_FREQUENCY_DEVIATION_CENTER,
		ant
	);
}

void radio_control_init_recv(void){
/*
  Prepares the RADIO for receiving a control packet.
*/
	
	/* Make sure that a simultaneous DISABLE event triggers this function. */
	radioState = RADIO_PREPARE_RECV;
	RADIO_INTENSET = RADIO_INTENSET_DISABLED_SET;
	
	cte_disable();
	
	if(RADIO_STATE != RADIO_STATE_STATE_DISABLED){
		
		/* Disable the RADIO. */
		/* The ISR will recall this function. */
		RADIO_EVENTS_DISABLED = 0;
		RADIO_TASKS_DISABLE = 1;
	}
	else{
	
		radio_disable_all_interrupts();
		radio_disable_all_critical_ppis();
		
		RADIO_SHORTS = RADIO_SHORTS_DISPATCH_PHYEND_DISABLE_ENABLED
		             | RADIO_SHORTS_READY_START_ENABLED;
		
		RADIO_INTENSET = RADIO_INTENSET_DISABLED_SET;
		
		RADIO_FREQUENCY = RADIO_CONTROL_FREQUENCY;

		radio_set_control_access_addresses(0);

		RADIO_PACKETPTR = (uint32_t) radioPacketData;

		RADIO_EVENTS_DISPATCH_PHYEND = 0;
		RADIO_EVENTS_READY = 0;
		RADIO_EVENTS_DISABLED = 0;
		
		radio_config_ble_2mbit();
		
		radioState = RADIO_CONTROL_RECV;
		RADIO_TASKS_RXEN = 1;
	}
}

void radio_check_disable(void){
/*
  Disables the RADIO if necessary.
*/
	
	radioState = RADIO_DISABLED;
	radio_disable_all_interrupts();
	radio_disable_all_critical_ppis();
	
	if(RADIO_STATE != RADIO_STATE_STATE_DISABLED){
		
		/* Disable the RADIO. */
		RADIO_EVENTS_DISABLED = 0;
		RADIO_TASKS_DISABLE = 1;
	}
}

void radio_isr(void){
	
	size_t size;
	
	if(RADIO_EVENTS_RSSIEND){
	/*
	  Processes a successfull RSSI conversion and prepares/requests a new one.
	*/
		RADIO_EVENTS_RSSIEND = 0;
		
		/* Request RSSI samples only during packet reception and */
		/* not during a possible Constant Tone Extension (CTE).  */
		/*   => Do not check for {RADIO_EVENTS_DISPATCH_PHYEND}! */
		if(!RADIO_EVENTS_BCMATCH && !RADIO_EVENTS_END){
			
			radioRSSAvg += RADIO_RSSISAMPLE;
			++radioRSSCnt;
			
			/* Trigger the next RSSI conversion after some delay. */
			TIMER_TASKS_START(TIMER2) = 1;
		}
		
		/* The above should take more than four CPU cycles. */
	}
	else if(RADIO_EVENTS_DISABLED){
	/*
	  Processes a received packet or handles the end of a transmitted one or
	  handles a successfull "emergency" shutdown.
	*/
		
		RADIO_EVENTS_DISABLED = 0;
		
		switch(radioState){
			
			case RADIO_CONTROL_SEND:
			
				/* Sent a control packet successfully. */
				/* Send next one (if any).             */
				radio_control_init_send();
				
			break;
			case RADIO_CONTROL_RECV:
				
				if(RADIO_CRCSTATUS){
				
					/* We should listen to that control packet. */
					/* (Access address matched, CRC OK)         */
					
					size = radioPacketData[1] + 1;
					radioPacketData[1] = radioPacketData[0];
				
					if(!buffer_append_row(&radioRecvBuffer,
					                      (uint8_t *) &radioPacketData[1],
					                      size)){
					
						debug_report(DEBUG_RADIO_OTA_RECV_BUFFER_OVERFLOW);
					}
				}
				
				/* Continue listening for control packets. */
				radio_control_init_recv();
				
			break;
			case RADIO_MEASURE_BLE_2MBIT_RX:
				
				radioState = RADIO_DISABLED;
				
				if(RADIO_CRCSTATUS){
				
					#if DEBUG_MEASUREMENT_FRAME >= 2
						GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
						GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
					#endif
					
					
					if(RADIO_RXMATCH == 1){
					
						/* Received a so called disruptive packet. */
						
						measure_stop();
					}
					else if(RADIO_RXMATCH == 0){
					
						/* Received a BLE 2MBit measurement packet. */
						
						measure_ble_2mbit_rx_succeeded(
							 (radioRSSAvg * (uint32_t) 128
							+ radioRSSCnt / 2) / radioRSSCnt,
							radioRSSCnt,
							#if REVISION >= REVISION_F
								RADIO_DFEPACKET.AMOUNT
							#else
								0
							#endif
						);
					}
				}
			break;
			case RADIO_MEASURE_NRF_2MBIT_RX:
				
				radioState = RADIO_DISABLED;
				
				if(RADIO_CRCSTATUS){
				
					#if DEBUG_MEASUREMENT_FRAME >= 2
						GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
						GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
					#endif
					
					
					if(RADIO_RXMATCH == 1){
					
						/* Received a so called disruptive packet. */
						
						measure_stop();
					}
					else if(RADIO_RXMATCH == 0){
					
						/* Received an NRF 2MBit measurement packet. */
						
						measure_nrf_2mbit_rx_succeeded(
							 (radioRSSAvg * (uint32_t) 128
							+ radioRSSCnt / 2) / radioRSSCnt,
							radioRSSCnt
						);
					}
				}
			break;
			case RADIO_MEASURE_BLE_1MBIT_RX:
				
				radioState = RADIO_DISABLED;
				
				if(RADIO_CRCSTATUS){
				
					#if DEBUG_MEASUREMENT_FRAME >= 2
						GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
						GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
					#endif
					
					
					if(RADIO_RXMATCH == 1){
					
						/* Received a so called disruptive packet. */
						
						measure_stop();
					}
					else if(RADIO_RXMATCH == 0){
					
						/* Received a BLE 1MBit measurement packet. */
						
						measure_ble_1mbit_rx_succeeded(
							 (radioRSSAvg * (uint32_t) 128
							+ radioRSSCnt / 2) / radioRSSCnt,
							radioRSSCnt,
							#if REVISION >= REVISION_F
								RADIO_DFEPACKET.AMOUNT
							#else
								0
							#endif
						);
					}
				}
			break;
			case RADIO_MEASURE_NRF_1MBIT_RX:
				
				radioState = RADIO_DISABLED;
				
				if(RADIO_CRCSTATUS){
				
					#if DEBUG_MEASUREMENT_FRAME >= 2
						GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
						GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
					#endif
					
					
					if(RADIO_RXMATCH == 1){
					
						/* Received a so called disruptive packet. */
						
						measure_stop();
					}
					else if(RADIO_RXMATCH == 0){
					
						/* Received an NRF 1MBit measurement packet. */
						
						measure_nrf_1mbit_rx_succeeded(
							 (radioRSSAvg * (uint32_t) 128
							+ radioRSSCnt / 2) / radioRSSCnt,
							radioRSSCnt
						);
					}
				}
			break;
		#if REVISION >= REVISION_F
			case RADIO_MEASURE_IEEE_250KBIT_RX:
				
				radioState = RADIO_DISABLED;
				
				if(RADIO_CRCSTATUS){
				
					#if DEBUG_MEASUREMENT_FRAME >= 2
						GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
						GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
					#endif
					
					
					if((((uint8_t *) RADIO_PACKETPTR)[RADIO_MEASURE_IEEE_250KBIT_STATUS_DATA_OFFSET] & 0x01) == 1){
					
						/* Received a so called disruptive packet. */
						
						measure_stop();
					}
					else if((((uint8_t *) RADIO_PACKETPTR)[RADIO_MEASURE_IEEE_250KBIT_STATUS_DATA_OFFSET] & 0x01) == 0){
					
						/* Received an IEEE 250kBit measurement packet. */
						
						measure_ieee_250kbit_rx_succeeded(
							 (radioRSSAvg * (uint32_t) 128
							+ radioRSSCnt / 2) / radioRSSCnt,
							radioRSSCnt
						);
					}
				}
			
			break;
		#endif
		#if DEBUG_MEASUREMENT_FRAME >= 2
			case RADIO_MEASURE_BLE_2MBIT_TX:
			case RADIO_MEASURE_NRF_2MBIT_TX:
		#if REVISION >= REVISION_F
			case RADIO_MEASURE_IEEE_250KBIT_TX:
		#endif
			case RADIO_MEASURE_BLE_1MBIT_TX:
			case RADIO_MEASURE_NRF_1MBIT_TX:
				
				/* Successfully transmitted a measurement packet (and CTE). */
				
				#if DEBUG_MEASUREMENT_FRAME >= 2
					GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
					GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
				#endif
				
			break;
		#endif
			case RADIO_PREPARE_RECV:
			
				/* Successfull emergency shutdown.       */
				/* Trigger reception of control packets. */
				radio_control_init_recv();
				
			break;
			case RADIO_PREPARE_X_2MBIT_RX:
				
				/* Successfull emergency shutdown.                 */
				/* Trigger reception of 2MBit measurement packets. */
				radio_measure_x_2mbit_init_rx(
					0,
					false,
					NULL,
					RADIO_BLE_MODE_UNDECIDED
				);
				
			break;
			case RADIO_PREPARE_X_1MBIT_RX:
				
				/* Successfull emergency shutdown.                 */
				/* Trigger reception of 1MBit measurement packets. */
				radio_measure_x_1mbit_init_rx(
					0,
					false,
					NULL,
					RADIO_BLE_MODE_UNDECIDED
				);
				
			break;
		#if REVISION >= REVISION_F
			case RADIO_PREPARE_IEEE_250KBIT_RX:
				
				/* Successfull emergency shutdown.                        */
				/* Trigger reception of IEEE 250kBit measurement packets. */
				
				radio_measure_ieee_250kbit_init_rx(
					0,
					false,
					NULL
				);
				
			break;
		#endif
			case RADIO_PREPARE_SEND:
				
				/* Successfull emergency shutdown.           */
				/* Trigger transmission of a control packet. */
				radio_control_init_send();
				
			break;
			case RADIO_PREPARE_CONSTANT:
				
				/* Successfull emergency shutdown.           */
				/* Trigger carrier frequency emission.       */
				radio_emit_carrier(0, SWITCH_ANTENNA_ALWAYS_LAST_ITEM);
				
			break;
			case RADIO_DISABLED:
				;
			break;
			default:
				
				debug_report(DEBUG_RADIO_UNEXPECTED_STATE);
			break;
		}
		
		/* Make sure the above to take at least four CPU cycles! */
		/* while(RADIO_EVENTS_DISABLED);                         */
	}
#if DEBUG_MEASUREMENT_FRAME >= 2
	/*
	  Clear debug pin when the RADIO is ready for receiving or transmitting a
	  measurement packet. This is done via PPI and GPIOTE.
	*/
	else if(RADIO_EVENTS_CRCOK){
	/*
	  Debug double toggle when the CRC check succeeded, i.e. when the
	  measurement schedule gets resetted.
	*/
	
		RADIO_EVENTS_CRCOK = 0;
		
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
	
		while(RADIO_EVENTS_CRCOK);
	}
	else if(RADIO_EVENTS_END){
		
		/* This has to be the last case because it should only get triggered */
		/* when this event was the actual interrupt source. Otherwise it may */
		/* be accidentally triggered e.g. by the DISABLED event.             */
		
		RADIO_EVENTS_END = 0;
		
		/* Successfully transmitted a measurement packet. */
		
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
		GPIOTE_TASKS_OUT[GPIOTE_CHANNEL_DEBUG_PIN] = 1;
		
		while(RADIO_EVENTS_END);
	}
#endif
}

