/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

#define RSS2I_MODULE "random"

/*
  Random Number Generator (RNG) API.
  
  Authors
  =======
    
    Copyright (C) 2018-2020  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
    
    2019-05-31    Initial public release.
    2020-07-07    Added {size} argument of function {random_page()}.
  
  Description
  ===========
    
    Provides methods to get unbiased ("high quality") random bytes
    (see {random_byte()}) or possibly biased ("low quality") random pages (see
    {random_page()}). That's all.
*/

#include <stdint.h>
#include <stddef.h>

#include <libopencm3/nrf5/rng.h>

static void random_init(void){
	
	/* Enable bias correction. */
	RNG_CONFIG = RNG_CONFIG_DERCEN_ENABLED;
	
	RNG_EVENTS_VALRDY = 0;
	RNG_TASKS_START = 1;
}

static void random_fast_init(void){
	
	/* Disable bias correction. */
	RNG_CONFIG = RNG_CONFIG_DERCEN_DISABLED;
	
	RNG_EVENTS_VALRDY = 0;
	RNG_TASKS_START = 1;
}

static void random_disable(void){
	
	RNG_TASKS_STOP = 1;
}

uint8_t random_byte(void){
/*
  Returnes an unbiased byte of random data.
*/
	random_init();
	
	while(!RNG_EVENTS_VALRDY);
	
	random_disable();
	
	return RNG_VALUE;
}

void random_page(uint8_t * page, size_t size){
/*
  Writes {size} bytes of possibly biased random data into {page}.
*/
	uint16_t i;
	
	random_fast_init();
	
	for(i = 0; i < size; ++i){
		
		while(!RNG_EVENTS_VALRDY);
		RNG_EVENTS_VALRDY = 0;
		page[i] = RNG_VALUE;
	}
	
	random_disable();
}

