/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

#define RSS2I_MODULE "scheduler"

/*
  Basic scheduler.
  
  Authors
  =======
  
    Copyright (C) 2020-2020  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
  
    2020-07-06    Initial release.
    2020-07-07    Added serial debug information output for {DEBUG_LEVEL >= 2}.
    2020-07-07    Fixed statistics calculation.
    2020-08-21    Remove blank debug output lines (saving some buffer space).
  
  Description
  ===========
  
    There are several tasks to be done in the background: Most RSS2I firmware
    modules need (more or less) frequent calls to their {*_trigger()} functions
    in order to process measurement results, store data on the on-board flash
    memory, process user queries, ….
    In situations with high CPU load it is necessary to prioritize tasks that
    are fundamental for measurement purposes. This module aims to provide such
    a basic prioritization without implementing a full featured scheduler.
    
  Usage
  =====
  
    It is sufficient to call {scheduler()} as frequent as possible.
*/

#include <stdint.h>

#include <rss2i/measure.h>
#include <rss2i/debug.h>
#include <rss2i/control.h>
#include <rss2i/watchdog.h>
#include <rss2i/radio.h>
#include <rss2i/spim.h>
#include <rss2i/print.h>
#include <rss2i/mlc.h>
#include <rss2i/timer.h>
#include <rss2i/systick.h>

#include <rss2i/config/scheduler.h>
#include <rss2i/config/measure.h>


#define SCHEDULER_TASKS       8


typedef struct {
	
	void (*trigger)(void);
	uint32_t maxInterval;

} SchedulerTask;

static const SchedulerTask schedulerTask[] = {
	
	{&measure_trigger,  SCHEDULER_TIME_MEASURE},
	{ &watchdog_reset, SCHEDULER_TIME_WATCHDOG},
	{  &timer_trigger,    SCHEDULER_TIME_TIMER},
	{&control_trigger,  SCHEDULER_TIME_CONTROL},
	{  &radio_trigger,    SCHEDULER_TIME_RADIO},
	{   &spim_trigger,     SCHEDULER_TIME_SPIM},
	{  &debug_trigger,    SCHEDULER_TIME_DEBUG},
	{    &mlc_trigger,      SCHEDULER_TIME_MLC},
};

_Static_assert(
	sizeof(schedulerTask) / sizeof(schedulerTask[0]) == SCHEDULER_TASKS,
	"Inconsistent number of specified tasks."
);

_Static_assert(
	                        8 == SCHEDULER_TASKS
	&& SCHEDULER_TIME_MEASURE  < SCHEDULER_TIME_LIMIT
	&& SCHEDULER_TIME_WATCHDOG < SCHEDULER_TIME_LIMIT
	&& SCHEDULER_TIME_TIMER    < SCHEDULER_TIME_LIMIT
	&& SCHEDULER_TIME_CONTROL  < SCHEDULER_TIME_LIMIT
	&& SCHEDULER_TIME_RADIO    < SCHEDULER_TIME_LIMIT
	&& SCHEDULER_TIME_SPIM     < SCHEDULER_TIME_LIMIT
	&& SCHEDULER_TIME_DEBUG    < SCHEDULER_TIME_LIMIT
	&& SCHEDULER_TIME_MLC      < SCHEDULER_TIME_LIMIT,
	
	"At least one schedule interval exceeds the time limit."
);

#if DEBUG_LEVEL >= 2
	static volatile uint32_t schedulerPrioritization[SCHEDULER_TASKS] = {0};
	static volatile uint32_t schedulerPrioritizations[SCHEDULER_TASKS] = {0};
	static volatile uint32_t schedulerDelay[SCHEDULER_TASKS] = {0};
	static volatile uint32_t schedulerDelays[SCHEDULER_TASKS] = {0};
	static volatile uint32_t schedulerMaxDelay[SCHEDULER_TASKS] = {0};

	static void scheduler_debug(void){
	/*
	  Prints some debug information (see below) for a single task per call.
	*/
		
		/* Do not put these strings onto the stack. */
		static char mHead[] = "Statistics for task #   :";
		static char mPNum[] = "  Prioritizations per measurement frame:    . ";
		static char mPAvg[] = "     Mean delay prior to prioritization:     us";
		static char mDNum[] = "    Overall calls per measurement frame:    . ";
		static char mDAvg[] = "                     Mean overall delay:     us";
		static char mDMax[] = "                     Max. overall delay:     us";
		
		static size_t task = 0;
		
		uint32_t count;
		uint32_t mean;
		
		/* Print header. */
		
		print_udec(task, mHead, 21, 3);
		debug("debug", mHead);
		
		if(schedulerTask[task].maxInterval){
		
			/* Print number of prioritizations. */
			
			count = schedulerPrioritizations[task]
			      * 10
			      * MEASURE_FRAME_LENGTH
			      / 1000
			      / SCHEDULER_DEBUG_PERIOD
			      / SCHEDULER_TASKS;
			
			if(count == 0 && schedulerPrioritizations[task] > 0){
				
				mPNum[41] = ' ';
				mPNum[42] = '<';
				mPNum[43] = '0';
				mPNum[45] = '1';
			}
			else{
			
				print_udec(count / 10, mPNum, 41, 3);
				mPNum[45] = '0' + (count % 10);
			}
			
			debug("debug", mPNum);
			
			/* Print mean delay prior to prioritization. */
			
			if(schedulerPrioritizations[task]){
			
				mean = schedulerPrioritization[task]
				     / schedulerPrioritizations[task];
			}
			else{
				
				mean = 0;
			}
			
			print_udec(mean, mPAvg, 41, 3);
			debug("debug", mPAvg);
			
			schedulerPrioritizations[task] = 0;
			schedulerPrioritization[task] = 0;
		}
		
		/* Print number of overall calls. */
		
		count = schedulerDelays[task]
		      * 10
		      * MEASURE_FRAME_LENGTH
		      / 1000
		      / SCHEDULER_DEBUG_PERIOD
		      / SCHEDULER_TASKS;
		
		if(count == 0 && schedulerDelays[task] > 0){
			
			mDNum[41] = ' ';
			mDNum[42] = '<';
			mDNum[43] = '0';
			mDNum[45] = '1';
		}
		else{
		
			print_udec(count / 10, mDNum, 41, 3);
			mDNum[45] = '0' + (count % 10);
		}
		
		debug("debug", mDNum);
		
		/* Print mean overall delay. */
		
		if(schedulerDelays[task]){
		
			mean = schedulerDelay[task]
			     / schedulerDelays[task];
		}
		else{
			
			mean = 0;
		}
		
		print_udec(mean, mDAvg, 41, 3);
		debug("debug", mDAvg);
		
		schedulerDelays[task] = 0;
		schedulerDelay[task] = 0;
		
		/* Print maximum overall delay. */
		
		print_udec(schedulerMaxDelay[task], mDMax, 41, 3);
		debug("debug", mDMax);
		
		schedulerMaxDelay[task] = 0;
		
		if(++task == SCHEDULER_TASKS){
			
			task = 0;
		}
	}
#endif

void scheduler_init(void){
/*
  Initializes the scheduler. Should be called only once.
*/
	
	#if DEBUG_LEVEL >= 2
		timer_call(&scheduler_debug, SCHEDULER_DEBUG_PERIOD, TIMER_REPEAT);
	#else
		;
	#endif
}


void scheduler(void){
/*
  Schedules all periodic tasks. Triggers one task per call. When the time since
  the last call to a critical task exceeds a specific threshold, that task gets
  prioritized, i.e. gets called immediately.
  
  This function should be called as frequent as possible.
*/
	
	static size_t scheduledTask = 0;
	static uint32_t last[SCHEDULER_TASKS] = {0};
	
	uint32_t us = systick_get_us();
	size_t nextTask = scheduledTask;
	size_t i;
	
	for(i = 0; i < SCHEDULER_TASKS; ++i){
		
		if(schedulerTask[i].maxInterval
		&& last[i] + schedulerTask[i].maxInterval - us >= SCHEDULER_TIME_LIMIT){
			
			/* Maximal schedule interval exceeded.         */
			/* Trigger the corresponding action right now! */
			
			#if DEBUG_LEVEL >= 2
				schedulerPrioritization[i] += us - last[i];
				++schedulerPrioritizations[i];
			#endif
			nextTask = i;
			break;
		}
	}
	
	if(nextTask == scheduledTask){
		
		/* Normal schedule processing. */
		scheduledTask = (scheduledTask + 1) % (SCHEDULER_TASKS);
	}
	
	#if DEBUG_LEVEL >= 2
		schedulerDelay[nextTask] += us - last[nextTask];
		++schedulerDelays[nextTask];
		if(us - last[nextTask] > schedulerMaxDelay[nextTask]){
		
			schedulerMaxDelay[nextTask] = us - last[nextTask];
		}
	#endif
	
	/* Remember processing of a task. */
	last[nextTask] = us;
	
	(*(schedulerTask[nextTask].trigger))();
}

