/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

#define RSS2I_MODULE "spim"

/*
  SPIM module for SPI communication with multiple FLASH devices.
  
  Authors
  =======
  
    Copyright (C) 2018-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
  
    2019-01-21    Initial release.
    2019-05-31    Added {spim_vacant_entries()}.
    2019-10-31    Removed {spim_recv_reset_blocking()}. The {spimRecvFlag} is
                  now reset wihtin {spim_query()}. The blocking part was removed
                  completely. The buffer length was adjusted in order to cope
                  with worst case scenarios. This partly fixed #6. Updated doc.
    2020-07-08    Added the DMA-based {spim_query_*()} querying procedures.
    2021-03-27    Use {NVIC_RSS2I_*_PRIORITY} macros.
  
  Description
  ===========

    Buffer
    ------
  
     After initialization there is an {spimBuffer} available. Data sequences
     that should get transmitted via SPIM may be put onto this buffer. SPIM
     communication has to get triggerd by the user (see {spim_trigger()}).
     
    Send data sequences
    -------------------
     
     A data sequence to be sent (please refer to the data fromat below) can be
     queried by calling the {spim_query()} function or by using a DMA-based
     querying procedure that consists of function calls {spim_query_request()}
     and {spim_query_finish()}.
    
    Received data
    -------------
    
     Since SPI is a synchronous protocol, every transmitted data sequence
     corresponds to a received data sequence. If enabled (see below) this data
     sequence can be read using {spim_recv()} until the next (read enabled) SPIM
     transmission.
     
     Data reception is not buffered. When {spim_recv_ready()} returns {true} the
     previous read request got processed completely. A new read request should
     not be queried until the previous one has completed.
   
    Data format
    -----------
  
     Every data sequence on the buffer is an array of {uint8_t}s. The first four
     bytes are used to exchange information about the sequence and won't get
     sent. The actual sequence is starting from the fifth byte.
      
        Byte | Information
       ------+------------------------------------------------------------------
          0  | Reserved.
       ------+------------------------------------------------------------------
          1  | The chip (FLASH device) to be communicated with ({0} or {1}).
       ------+------------------------------------------------------------------
          2  | If this equals zero, the incoming sequence isn't saved within
             | {spimRecvData}
       ------+------------------------------------------------------------------
          3  | Set to true ({!0}) if it is necessary to unlock write operations
             | before transmitting the specified data sequence (see below).
       ------+------------------------------------------------------------------
*/

#include <stdint.h>
#include <stdbool.h>

#include <libopencm3/cm3/nvic.h>

#include <libopencm3/nrf5/spim.h>
#include <libopencm3/nrf5/gpio.h>
#include <libopencm3/nrf5/timer.h>

#include <rss2i/buffer.h>
#include <rss2i/debug.h>
#include <rss2i/memory.h>
#include <rss2i/nvic.h>

#include <rss2i/config/spim.h>

/* SPI pins */

#define SPIM_NCS1_GPIO GPIO0
#define SPIM_NCS2_GPIO GPIO1
#define SPIM_MOSI_GPIO GPIO3
#define SPIM_MISO_GPIO GPIO2
#define SPIM_SCLK_GPIO GPIO4

#define SPIM_NCS1_PIN 0
#define SPIM_NCS2_PIN 1
#define SPIM_MOSI_PIN 3
#define SPIM_MISO_PIN 2
#define SPIM_SCLK_PIN 4


typedef enum {
	
	IDLE,         /* No query to process. Check for new queries occasionally. */
	AWAIT,        /* Step the state machine at the next trigger event. */
	BUSY,         /* SPIM busy. No external trigger needed. */
	
} SPIMBusy;

static volatile SPIMBusy spimBusy = IDLE;

static Buffer spimBuffer;

static uint8_t spimRecvData[SPIM_BUFFER_WIDTH];
static volatile bool spimRecvReady = true;

static volatile size_t spimQuerySteps;
static volatile bool spimReadEnabled;

static uint8_t * spimWIPQuery;
static uint8_t * spimWELQuery;
static size_t spimWIPSize;
static size_t spimWELSize;
static bool (* spim_check_wip)(uint8_t *);


bool spim_recv_ready(void){
/*
   Checks the status of an ongoing read request.
*/

	return spimRecvReady;
}

uint8_t * spim_recv(void){
/*
   Returns the address of the received data during the last read enabled SPIM
   transmission. The memory pointed to by the return value is guaranteed to hold
   its value unitl the next read enabled SPIM transmission.
*/
	
	return spimRecvData;
}

size_t spim_vacant_entries(void){
/*
  Returns the number of free entries on the SPIM output buffer.
*/
	return buffer_vacant_entries(&spimBuffer);
}

uint8_t * spim_query_request(void){
/*
  Returns a RAM address where the next SPIM query can be stored before it is
  finally put onto the SPIM buffer by calling {spim_query_finish()}. Returns
  {NULL} if the SPIM buffer is already full.
    
  Memory Write Warning
  ====================
    
    The user has to ensure that the length of the written data complies with the
    reserved memory per buffer entry as it is specified by {SPIM_BUFFER_WIDTH}.
*/
	
	return buffer_new_row_address(&spimBuffer);
}

bool spim_query_finish(uint8_t * addr, size_t size){
/*
  Finishes the DMA-based SPIM query procedure that was initiated by calling
  {spim_query_request()}.
  
  Some Notes
  ==========
  
    This function does not perform any copying tasks. The specified {addr} is
    used for validation purposes only. If the specified query is read enabled
    (see the doc above), the {spimRecvReady} flag is set to {false} in success
    case (see below).
    
    The user has to ensure that {size} complies with the maximal reserved
    memory per output buffer entry as it is specified by {SPIM_BUFFER_WIDTH}.
  
  Argument List
  =============
    
    addr    Pointer to the already written RAM as it was returned by the
            previous call to {spim_query_request()}.
    size    Number of written bytes.
  
  Return Value
  ============
    
    Returns {false} when illegal buffer activities, i.e. buffer activities
    between the call of {spim_query_request()} and the call to this function,
    have been detetcted. Returns {true} otherwise.
*/
	
	bool ok = buffer_new_row_finish(&spimBuffer, addr, size);
	
	if(ok && addr[2]){
		
		spimRecvReady = false;
	}
	
	return ok;
}

bool spim_query(uint8_t * data, size_t size){
/*
   Puts the specified SPIM query onto the SPIM buffer. Returns {false} in case
   of a buffer overflow. Returns {true} otherwise. If the specified query is
   read enabled (see the doc above), the {spimRecvReady} flag is set to {false}
   in success case.
*/
	
	bool ok = buffer_append_row(&spimBuffer, data, size);
	
	if(ok && data[2]){
		
		spimRecvReady = false;
	}
	
	return ok;
}

void spim_init(uint8_t * welQuery, size_t welSize, uint8_t * wipQuery, size_t wipSize, bool (* check_wip)(uint8_t *)){
/*
  Initializes this SPIM module.
  
  Argument list
  =============
  
    {welQuery}     Pointer to the sequence which has to precede every write/
                   erase operation. This corresponds to the fourth byte in
                   every sequence (see above).
    {welSize}      Size of the {welQuery}.
    {wipQuery}     Pointer to the sequence which reads the write-in-progress
                   status out of the FLASH device.
    {wipSize}      Size of the {wipQuery}.
    {check_wip}    Function to resolve if the {wipQuery} indicates a write in
                   progress. In this case that function should return {true}.
  
*/

	static uint8_t bufferData[SPIM_BUFFER_LENGTH][SPIM_BUFFER_WIDTH];
	static size_t bufferSize[SPIM_BUFFER_LENGTH];
	
	/* External SPIM queries */
	
	spimWELQuery = welQuery;
	spimWELSize = welSize;
	spimWIPQuery = wipQuery;
	spimWIPSize = wipSize;
	spim_check_wip = check_wip;
	
	/* Pin configuration */
	
	GPIO_DIRSET = SPIM_MOSI_GPIO
	            | SPIM_SCLK_GPIO
	            | SPIM_NCS1_GPIO
	            | SPIM_NCS2_GPIO;
	
	GPIO_OUTSET = SPIM_NCS1_GPIO
	            | SPIM_NCS2_GPIO;
	
	
	/* SPIM peripheral configuration */
	
	SPIM_ENABLE(SPIM0) = SPIM_ENABLE_ENABLE_ENABLED;
	
	SPIM_INTENSET(SPIM0) = SPIM_INTENSET_END_SET;
	
	SPIM_PSEL(SPIM0).SCK  = SPIM_SCLK_PIN;
	SPIM_PSEL(SPIM0).MOSI = SPIM_MOSI_PIN;
	SPIM_PSEL(SPIM0).MISO = SPIM_MISO_PIN;
	
	SPIM_FREQUENCY(SPIM0) = SPIM_FREQUENCY_FREQUENCY_M1;
	
	nvic_enable_irq(NVIC_SPIM0_SPIS0_TWIM0_TWIS0_SPI0_TWI0_IRQ);
	nvic_set_priority(
		NVIC_SPIM0_SPIS0_TWIM0_TWIS0_SPI0_TWI0_IRQ,
		NVIC_RSS2I_LOW_PRIORITY
	);
	
	
	/* Buffer initialization */
	
	buffer_init_rows(&spimBuffer, &bufferData[0][0], bufferSize, SPIM_BUFFER_LENGTH, SPIM_BUFFER_WIDTH);
}

static void spim_start(uint8_t * txdPtr, size_t size, uint8_t chip, volatile uint8_t * recvData){
/*
  Configures the SPIM with the specified settings and starts SPIM communication.
  The received data (if {size} is greater than one) gets stored within
  {recvData}.
  
  Argument list
  =============
  
    {txdPtr}         Pointer to the sequence to be transmitted.
    {size}           Size of the sequence to be transmitted/received.
                     (If this equals one, no data will be received).
    {chip}           The chip to be communicated with ({0} or {1}).
    {recvData}       Memory location the received data is written to.
*/

	SPIM_TXD(SPIM0).PTR = (uint32_t) txdPtr;
	SPIM_TXD(SPIM0).MAXCNT = size;
	SPIM_RXD(SPIM0).PTR = (uint32_t) recvData;

	/*
	   If there is just one byte to be sent the receive count is set to
	   zero. This is a workarround for Errata
	   
		 [58] SPIM: SPIM clocks out an additional byte when RXD.MAXCNT = 1
	*/
	
	SPIM_RXD(SPIM0).MAXCNT = (size == 1) ? 0 : size;

	/* Select FLASH chip */
	GPIO_OUTCLR = chip ? SPIM_NCS2_GPIO : SPIM_NCS1_GPIO;
	
	SPIM_TASKS_START(SPIM0) = 1;
}

static void spim_step(void){
/*
	SPIM state machine.
*/
	
	static uint8_t recvData[SPIM_BUFFER_WIDTH];
	static uint8_t * data;
	static size_t size;
	static size_t offset;
	
	typedef enum {
	
		READY,        /* Ready to process a new query */
		GET_WIP,      /* Get the WriteInProgress state */
		CHECK_WIP,    /* Check the WriteInProgress state */
		SET_WEL,      /* Unlock write operations (Set the WriteEnableLatch) */
		QUERY,        /* Prepare query processing */
		QUERYING,     /* Finally, process the query */

	} SPIMState;
	
	static SPIMState spimState = READY;
	static bool writeProcessPending[2] = {true, true};
	
	switch(spimState){
	
		case READY:
	
			if((size = buffer_next_size(&spimBuffer))){
				
				/* Yay, got a new query to process */
				data = buffer_get_row(&spimBuffer);
				
				if(writeProcessPending[data[1]]){
				
					spimState = GET_WIP;
					spimBusy = AWAIT;
				}
				else{
					
					spimState = SET_WEL;
					spimBusy = AWAIT;
				}
			}
			else{
				
				/* Wait a bit until the next check for new queries */
				spimBusy = IDLE;
			}
		
		break;
		case GET_WIP:
		
			spimState = CHECK_WIP;
			spim_start(spimWIPQuery, spimWIPSize, data[1], recvData);
			
		break;
		case CHECK_WIP:
			
			if(!spim_check_wip(recvData)){
				
				/* Device is NOT busy */
				
				spimState = SET_WEL;
				writeProcessPending[data[1]] = false;
				spimBusy = AWAIT;
			}
			else{
				
				/* Device is (still) busy. Wait a bit until the next check. */
				
				spimState = GET_WIP;
				spimBusy = IDLE;
			}
			
			
			
		break;
		case SET_WEL:
		
			if(data[3]){
				
				/* Unlocking write operations is necessary */
				
				spimState = QUERY;
				writeProcessPending[data[1]] = true;
				spim_start(spimWELQuery, spimWELSize, data[1], recvData);
			}
			else{
				
				/* Unlocking write operations isn't necessary. Process query. */
				spimState = QUERY;
				spimBusy = AWAIT;
			}
			
		break;
		case QUERY:
			
			spimReadEnabled = data[2] ? true : false;
			offset = 0;
			
			/* Number of SPIM steps (SPIM peripheral can merely transmit */
			/* 255(!) bytes within a single run)                          */
			spimQuerySteps = (size - 4 - 1) / 255;
			
			spimState = QUERYING;
			spimBusy = AWAIT;
			
		break;
		case QUERYING:
			
			if(spimQuerySteps == 1 && (size - 4 - 1) % 255 == 0){
				
				/*
				   The last SPIM query would consist of one byte only.
				   This is a workarround for Errata
				
				     [58] SPIM: SPIM clocks out an additional byte when
				                RXD.MAXCNT = 1
				*/
				
				spim_start(&data[4 + offset], 254, data[1], (spimReadEnabled ? spimRecvData : recvData) + offset);
				offset += 254;
			}
			else if(spimQuerySteps){
				
				spim_start(&data[4 + offset], 255, data[1], (spimReadEnabled ? spimRecvData : recvData) + offset);
				offset += 255;
			}
			else{
				
				spimState = READY;
				spim_start(&data[4 + offset], size - 4 - offset, data[1], (spimReadEnabled ? spimRecvData : recvData) + offset);
			}
			
		break;
	}	
}

void spim_trigger(void){
/*
   The SPIM communication isn't time-critical and hence may get triggered
   asynchronously by the {main} loop instead of wasting high-priority interrupt
   time.
   Please make sure to call this function "frequently". Otherwise an SPIM buffer
   overflow may occur.
*/

	static uint16_t wait = SPIM_TRIGGER_DELAY;
	
	if(spimBusy == IDLE){
		
		/* SPIM communication isn't busy and hence gets triggered just */
		/* occasionally.                                               */
		
		if(!(wait--)){
			
			wait = SPIM_TRIGGER_DELAY;
			spimBusy = BUSY;
			spim_step();
		}
	}
	else if(spimBusy == AWAIT){
		
		/* SPIM communication waits for the next step. */
		
		wait = SPIM_TRIGGER_DELAY;
		spimBusy = BUSY;
		spim_step();
	}
}

void spim0_spis0_twim0_twis0_spi0_twi0_isr(void){
	
	if(SPIM_EVENTS_END(SPIM0)){
		
		SPIM_EVENTS_END(SPIM0) = 0;
		
		if(spimQuerySteps){
			
			/* Request the next SPIM step. */
			
			--spimQuerySteps;
			spimBusy = AWAIT;
		}
		else{
			
			/* Finish SPIM coomunication. */
			
			GPIO_OUTSET = SPIM_NCS1_GPIO
			            | SPIM_NCS2_GPIO;
			            
			/* Set the {spimRecvReady} flag if reception was requested. */
			
			if(spimReadEnabled){
				
				spimRecvReady = true;
			}
			
			/* Request the next SPIM step. */
			spimBusy = AWAIT;
		}
		
		while(SPIM_EVENTS_END(SPIM0));
	}
}

