/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

#define RSS2I_MODULE "suart"

/*
  Software UART (TX only).
  
  Authors
  =======
  
    Copyright (C) 2018-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
  
    2019-04-15    Initial public release.
    2020-07-07    Added {suart_send_*()} functions for DMA based buffering.
    2021-03-27    Use {NVIC_RSS2I_*_PRIORITY} macros.
  
  Description
  ===========
    
    After initialization using {suart_init()}, strings ({'\0'} terminated)
    may be sent via UART by calling {suart_send()}. The send requests get
    buffered before they are actually sent. This software UART buffer may
    alternatively be filled using the DMA based {suart_send_*()} procedures.
*/


#include <stdint.h>
#include <stddef.h>
#include <stdbool.h>

#include <libopencm3/cm3/nvic.h>
#include <libopencm3/nrf5/timer.h>
#include <libopencm3/nrf5/gpio.h>

#include <rss2i/buffer.h>
#include <rss2i/nvic.h>


#define SUART_GPIO_TXD GPIO19


static Buffer suartBuffer;
static volatile bool suartBusy = false;


void suart_init(uint32_t period, char * data, size_t * size, size_t length, size_t width){
/*
  Initializes the software UART (TX only). Should be called only once.
  
  Argument list
  =============
  
    {period}     Bit period in multiples of 1/16 µs.
    {data}       Pointer to the buffer data memory.
    {size}       Pointer to the buffer size memory.
    {length}     Length of the desired buffer.
    {width}      Width of the desired buffer.
  
*/
	
	/* Initialize buffer. */
	buffer_init_strings(&suartBuffer, data, size, length, width);
	
	/* Initialize pin settings. */
	
	GPIO_DIRSET = SUART_GPIO_TXD;
	GPIO_OUTSET = SUART_GPIO_TXD;
	
	/* Initialize timer setup. */
	
	nvic_enable_irq(NVIC_TIMER0_IRQ);
	nvic_set_priority(NVIC_TIMER0_IRQ, NVIC_RSS2I_LOW_PRIORITY);
	
	TIMER_INTENSET(TIMER0) = TIMER_INTENSET_COMPARE0_SET;
	TIMER_PRESCALER(TIMER0) = 0;
	TIMER_BITMODE(TIMER0) = TIMER_BITMODE_BITMODE_32Bit;
	TIMER_CC(TIMER0)[0] = period;
	TIMER_SHORTS(TIMER0) |= TIMER_SHORTS_COMPARE0_CLEAR_ENABLED;
	TIMER_TASKS_START(TIMER0) = 1;
}

bool suart_busy(void){
	
	return suartBusy;
}

char * suart_send_request(void){
/*
  Returns a RAM address where the next software UART message can be stored
  before it is going to be sent by calling {suart_send_finish()}. Returns {NULL}
  if the {suartBuffer} is full.
  
  Data Format
  ===========
    
    The software UART expects the buffer entries to be strings. The last
    character, i.e. the terminating null character in the case of a valid C
    string, is therefore truncated when sending.
    
  Memory Write Warning
  ====================
    
    The user has to ensure that the length of the written data complies with the
    maximal reserved memory per buffer entry as it was specified when initiating
    the software UART by calling {suart_init()}.
*/
	
	uint8_t * addr;
	
	if((addr = buffer_new_row_address(&suartBuffer))){
		
		return (char *) addr;
	}
	else{
		
		return NULL;
	}
}

bool suart_send_finish(char * addr, size_t size){
/*
  Finishes the software UART write procedure that was initiated by calling
  {suart_send_request()}. Note, that this function does not perform any copying
  task. The specified {addr} is used for validation pruposes only. Also note,
  that the user has to ensure that {size} complies with the maximal reserved
  memory per buffer entry as it was specified when initiating the software UART
  by calling {suart_init()}.
  
  Argument List
  =============
    
    addr    Pointer to the already written RAM as it was returned by the
            previous call to {suart_send_request()}.
    size    Number of written bytes (including the terminating null character).
  
  Return Value
  ============
    
    Returns {false} when illegal buffer activities, i.e. buffer activities
    between the call of {suart_send_request()} and the call to this function,
    have been detetcted. Returns {true} otherwise.
*/
	
	if(buffer_new_row_finish(&suartBuffer, (uint8_t *) addr, size)){
		
		suartBusy = true;
		return true;
	}
	else{
		
		return false;
	}
}

bool suart_send(const char * string){
/*
  Sends the given {string} via software UART. Returns {false} if the
  {suartBuffer} is full. Returns {true} otherwise.
*/
	
	if(buffer_append_string(&suartBuffer, string)){
		
		suartBusy = true;
		return true;
	}
	else{
		
		return false;
	}
}

void timer0_isr(void){
/*
  Software UART (TX only) implementation.
*/
	
	static size_t stringPointer = 0;
	static int8_t state = -1;
	static size_t stringSize;
	static char * string;
	static char c;
	
	TIMER_EVENTS_COMPARE(TIMER0)[0] = 0;
	
	if(suartBusy){
		
		if(state == -1 && stringPointer == 0){
			
			/* Get an new string from the buffer */
			
			if((stringSize = buffer_next_size(&suartBuffer))){
				
				/* Get rid of the terminating null character. */
				--stringSize;
				string = buffer_get_string(&suartBuffer);
			}
			else{
				
				/* There isn't anything to send */
				suartBusy = false;
			}
		}
		
		if(suartBusy){
		
			if(state == -1){
				
				/* Get next character to be sent via software UART */
				c = string[stringPointer++];
				
				if(stringPointer == stringSize){
					
					/* This was the last character within the current string */
					stringPointer = 0;
				}
			}

			switch(state){
		
				case -1:
					GPIO_OUTCLR = SUART_GPIO_TXD;
					break;
				case 0:
				case 1:
				case 2:
				case 3:
				case 4:
				case 5:
				case 6:
				case 7:
					if(c & (1u << state)){
						GPIO_OUTSET = SUART_GPIO_TXD;
					}
					else{
						GPIO_OUTCLR = SUART_GPIO_TXD;
					}
					break;
				case 8:
					GPIO_OUTSET = SUART_GPIO_TXD;
					break;
				default:
					;	
			}
	
			if(++state == 9){ state = -1; }
		}
	}
}

