/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

#define RSS2I_MODULE "switch"

/*
  Control the external HF switch(es).
  
  Authors
  =======
  
    Copyright (C) 2020-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
  
    2020-08-13    Initial release.
    2020-08-25    Added HF switch protection in brownout case.
    2021-02-12    Added the {SWITCH_CTL_*} convenience defines.
    2021-02-12    Fixed the control pin patterns for revision I and higher.
    2021-02-25    Speeded up the antenna switch pattern generation.
    2021-02-26    Unlock switch in {switch_init()}.
    2021-03-29    Add some static asserts.
    2021-03-29    Added {switch_get_standard_antenna()}.
*/

#include <stdint.h>
#include <stdbool.h>
#include <stddef.h>

#include <libopencm3/nrf5/gpio.h>

#include <rss2i/switch.h>
#include <rss2i/board.h>
#include <rss2i/debug.h>


/* Make sure the {SwitchAntenna}s are consecutive. */
_Static_assert(

	   SWITCH_ANTENNA_1 + 1 == SWITCH_ANTENNA_2
	&& SWITCH_ANTENNA_2 + 1 == SWITCH_ANTENNA_3
	&& SWITCH_ANTENNA_3 + 1 == SWITCH_ANTENNA_4
	&& SWITCH_ANTENNA_4 + 1 == SWITCH_ANTENNA_5
	&& SWITCH_ANTENNA_5 + 1 == SWITCH_ANTENNA_6
	&& SWITCH_ANTENNA_6 + 1 == SWITCH_ANTENNA_7
	&& SWITCH_ANTENNA_7 + 1 == SWITCH_ANTENNA_8,
	"The SwitchAntennas are not naturally ordered."
);


/* Convenience definition that maps the GPIO pin names    */
/* to the corresponding bit in the {RADIO_SWITCHPATTERN}. */

typedef enum {
	
	SWITCH_GPIO18, /* HF switch CTL 1 / A */
	SWITCH_GPIO17, /* HF switch CTL 2 / B */
	SWITCH_GPIO16, /* HF switch CTL 3 / C */
	
	SWITCH_GPIO_ALWAYS_LAST_ITEM
	
} SwitchGPIOBit;

/* Convnience defines that map the GPIO pins to CTL pins. */

#define SWITCH_CTL_1 SWITCH_GPIO18
#define SWITCH_CTL_2 SWITCH_GPIO17
#define SWITCH_CTL_3 SWITCH_GPIO16

#define SWITCH_CTL_A SWITCH_GPIO18
#define SWITCH_CTL_B SWITCH_GPIO17
#define SWITCH_CTL_C SWITCH_GPIO16

/* Map between the bit number used for the {RADIO_SWITCHPATTERN} */
/* and the corresponding pin number within the GPIO peripheral.  */

static const uint8_t switchGPIOPin[SWITCH_GPIO_ALWAYS_LAST_ITEM] = {
	
	18, /* HF switch CTL 1 / A */
	17, /* HF switch CTL 2 / B */
	16, /* HF switch CTL 3 / C */
};

static bool switchLocked = false;


void switch_init(void){
/*
  Acquires the GPIOs that control the external HF switch.
*/
	
	switchLocked = false;
	
	GPIO_DIRSET = GPIO18   /* HF switch CTL 1 */
	            | GPIO17   /* HF switch CTL 2 */
	            | GPIO16;  /* HF switch CTL 3 */
	
	GPIO_OUTCLR = GPIO18   /* HF switch CTL 1 */
	            | GPIO17   /* HF switch CTL 2 */
	            | GPIO16;  /* HF switch CTL 3 */
}

uint8_t switch_get_number_of_used_pins(void){
/*
  Returns the number of pins that control the external HF switch.
*/
	
	return SWITCH_GPIO_ALWAYS_LAST_ITEM;
}

uint32_t switch_get_pin_number_of_bit(uint8_t bit){
/*
  Returns the GPIO pin number that corresponds to the given {bit} number within
  the {RADIO_SWITCHPATTERN}.
*/
	if(bit < SWITCH_GPIO_ALWAYS_LAST_ITEM && !switchLocked){
	
		return switchGPIOPin[bit];
	}
	else{
		
		if(!switchLocked){
		
			debug_report(DEBUG_SWITCH_INVALID_BIT_NUMBER);
		}
		
		/* Disconnected. */
		return (1ul << 31);
	}
}

SwitchAntenna switch_get_standard_antenna(void){
/*
  Returns the standard antenna, i.e. the antenna {ant} for which
  {switch_get_pattern(ant)} returns {0}.
*/
	
	if(board_rev() >= BOARD_REV_I){
		
		return SWITCH_ANTENNA_8;
	}
	else{
		
		return SWITCH_ANTENNA_4;
	}
}

uint32_t switch_get_standard_pattern(void){
/*
  Returns the standard switch pattern, i.e. that CTL Pin configuration where
  all pins are driven low.
*/
	
	return 0;
}

uint32_t switch_get_pattern(SwitchAntenna ant){
/*
  Returns the {RADIO_SWITCHPATTERN} that corresponds to the specified {ant}enna.
*/
	
	/* 8 bit wide data types are sufficient AT THE MOMENT. */
	/* Implicit cast to 32 bit wide data type at return.   */
	
	/* Don't put this onto the stack! */
	static const uint8_t patternBoardRevFromI[8] = {
	
		/* Antenna patterns for board Revision I or higher. */
		/* (Analog Devices HMC253AQS24E HF switch)          */
		
		/* Antenna 1: RF8 */
		(1 << SWITCH_CTL_A) | (1 << SWITCH_CTL_B) | (1 << SWITCH_CTL_C),
		/* Antenna 2: RF7 */
		(0 << SWITCH_CTL_A) | (1 << SWITCH_CTL_B) | (1 << SWITCH_CTL_C),
		/* Antenna 3: RF6 */
		(1 << SWITCH_CTL_A) | (0 << SWITCH_CTL_B) | (1 << SWITCH_CTL_C),
		/* Antenna 4: RF5 */
		(0 << SWITCH_CTL_A) | (0 << SWITCH_CTL_B) | (1 << SWITCH_CTL_C),
		/* Antenna 5: RF4 */
		(1 << SWITCH_CTL_A) | (1 << SWITCH_CTL_B) | (0 << SWITCH_CTL_C),
		/* Antenna 6: RF3 */
		(0 << SWITCH_CTL_A) | (1 << SWITCH_CTL_B) | (0 << SWITCH_CTL_C),
		/* Antenna 7: RF2 */
		(1 << SWITCH_CTL_A) | (0 << SWITCH_CTL_B) | (0 << SWITCH_CTL_C),
		/* Antenna 8: RF1 */
		(0 << SWITCH_CTL_A) | (0 << SWITCH_CTL_B) | (0 << SWITCH_CTL_C),
	};
	
	/* Don't put this onto the stack! */
	static const uint8_t patternBoardRevFToH[8] = {
		
		/* Antenna patterns for board revisions F to H. */
		/* (Qorvo QPC6082 HF switch)                     */
		
		/* Antenna 1 */
		(1 << SWITCH_CTL_1) | (0 << SWITCH_CTL_2) | (1 << SWITCH_CTL_3),
		/* Antenna 2 */
		(0 << SWITCH_CTL_1) | (1 << SWITCH_CTL_2) | (1 << SWITCH_CTL_3),
		/* Antenna 3 */
		(1 << SWITCH_CTL_1) | (1 << SWITCH_CTL_2) | (1 << SWITCH_CTL_3),
		/* Antenna 4 */
		(0 << SWITCH_CTL_1) | (0 << SWITCH_CTL_2) | (0 << SWITCH_CTL_3),
		/* Antenna 5 */
		(0 << SWITCH_CTL_1) | (0 << SWITCH_CTL_2) | (1 << SWITCH_CTL_3),
		/* Antenna 6 */
		(1 << SWITCH_CTL_1) | (1 << SWITCH_CTL_2) | (0 << SWITCH_CTL_3),
		/* Antenna 7 */
		(0 << SWITCH_CTL_1) | (1 << SWITCH_CTL_2) | (0 << SWITCH_CTL_3),
		/* Antenna 8 */
		(1 << SWITCH_CTL_1) | (0 << SWITCH_CTL_2) | (0 << SWITCH_CTL_3),
	};
	
	if(ant < SWITCH_ANTENNA_ALWAYS_LAST_ITEM){
	
		if(board_rev() >= BOARD_REV_I){
			
			return patternBoardRevFromI[ant];
		}
		else{
			
			return patternBoardRevFToH[ant];
		}
	}
	else{
		
		debug_report(DEBUG_SWITCH_INVALID_ANTENNA);
		return 0;
	}
}

void switch_lock(void){
/*
  Locks the switch, i.e. prevent any further changes of the external HF switch,
  and drives all HF switch control lines low.
*/
	
	switch_set(switch_get_standard_pattern());
	switchLocked = true;
}

void switch_set(uint8_t pattern){
/*
  Sets the HF switch to the specified state.
  
  Arguments
  =========
   
   pattern    Desired state of the HF switch control pins. The bits in {pattern}
              correspond to the HF switch control pins according to the
              {switchGPIOPin} map.
*/
	
	uint8_t i;
	uint32_t clrMask = 0;
	uint32_t setMask = 0;
	
	for(i = 0; i < SWITCH_GPIO_ALWAYS_LAST_ITEM; ++i){
		
		if(pattern & (1u << i)){
			
			setMask |= (1u << switchGPIOPin[i]);
		}
		else{
			
			clrMask |= (1u << switchGPIOPin[i]);
		}
	}
	
	if(!switchLocked){
	
		GPIO_OUTCLR = clrMask;
		GPIO_OUTSET = setMask;
	}
}

