/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

#define RSS2I_MODULE "systick"

/*
  Abstraction layer for the ARM systick counter.
  
  Authors
  =======
    
    Copyright (C) 2020-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
    
    2020-07-03    Initial release.
    2021-03-27    Do not try to set the priority of an ARM system interrupt.
    2021-03-28    Update doc within {systick_get_us()}.
  
  Description
  ===========
    
    This module provides an abstraction layer for the ARM systick counter. There
    are two 32 bit counters available that count the milliseconds and the micro-
    seconds, respectively, since the system's start up.
*/

#include <stdint.h>
#include <stddef.h>
#include <stdbool.h>

#include <libopencm3/cm3/systick.h>

#include <rss2i/clock.h>
#include <rss2i/nvic.h>


static volatile uint32_t systickMS = 0;


void systick_init(void){
/*
  Initializes the SysTick peripheral to generate an interrupt every millisecond.
  It gets initialized once even if called multiple times.
*/
	
	static bool initialized = false;
	
	if(!initialized){
		
		systick_set_clocksource(STK_CSR_CLKSOURCE_AHB);
		systick_set_reload(CLOCK_CYCLES_PER_MS - 1);
		systick_clear();
		systick_interrupt_enable();
		systick_counter_enable();
		
		initialized = true;
	}
}

uint32_t systick_get_ms(void){
/*
  Returns the milliseconds since the system's start up.
*/
	
	return systickMS;
}

uint32_t systick_get_us(void){
/*
  Returns the microseconds since the system's start up. Please note the
  following systick timings, where {max} refers to the systick's reload value:
  
  
         .-- transition from {3} to {2}
         |          
         |     .-- transition from {2} to {1}
         |     |     
         |     |     .-- transition from {1} to {0}
         |     |     |
         |     |     |     .-- transition from {0} to {max}
         |     |     |     |
         |     |     |     |     .-- transition from {max - 1} to {max - 2}.
         |     |     |     |     | 
         |     |     |     |     |     .-- ... 
         |     |     |     |     |     | 
    ...--|-----|-----|-----|-----|-----|-...-|-----|-----|--... systick counter
    
                      ___________________   ___
    ...--------------|___________________...___|------------... systick ISR
  
  
  From this, we can derive a few things:
    
    - The systick ISR may, of course, interrupt the {systick_get_us()} function,
      i.e. the {systickMS} value may change and may be inconsistent with the
      systick counter value that is retrieved prior to or after reading the
      {systickMS} variable.
    
    - This function won't ever retrieve the systick counter value {0} (as well
      as a few other low systick counter values), since the systick ISR is
      triggered at the transition from {1} to {0}.
*/
	
	uint32_t oldMS = systickMS;
	uint32_t us = (CLOCK_CYCLES_PER_MS - systick_get_value())
	            / (uint32_t) CLOCK_CYCLES_PER_US;
	uint32_t newMS = systickMS;
	
	if(oldMS == newMS){
		
		return oldMS * (uint32_t) 1000 + us;
	}
	else{
		
		/* A systick interrupt interrupted this function call. */
		
		return newMS * (uint32_t) 1000;
	}
}

void sys_tick_handler(){
/*
  Increments the ms counter.
*/

	++systickMS;
}

