/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

#define RSS2I_MODULE "timer"

/*
  Provides some timers with a resolution of 1 ms.
  
  Authors
  =======
    
    Copyright (C) 2018-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
    
    2019-05-29    Initial public release.
    2019-06-25    ADDED: {TimerID}. BUGFIX: Assured that the return value of
                  {timer_flag()} and {timer_call()} is always positive in
                  success case. BUGFIX: Made it possible to clear timers with a
                  positive index.
    2019-07-12    The timers now get checked asynchronously within
                  {timer_trigger()}.
    2020-02-12    Add {volatile} modifiers to the {Timer} struct.
    2020-06-09    Use ARM's SysTick instead of a timer peripheral.
    2020-06-09    Added {timer_reset()}.
    2020-07-03    Fix possible malfunction that may arise after about 50 days.
    2020-07-03    The systick configuration is now done by the systick module.
    2020-07-07    Check only a single timer per {timer_trigger()} call.
    2021-03-29    Update doc with regard to the return value of {timer_config()}
  
  Description
  ===========
    
    This modules provides blocking delays ({timer_wait()}) as well as real
    timers with the possibility to alter some {bool}ean flag ({timer_flag()})
    or to call a callback function ({timer_call()}) once a timer expires.
    
    The resolutionn is 1 ms. The uncertainty is about 1 ms.
*/

#include <stdint.h>
#include <stddef.h>
#include <stdbool.h>


#include <rss2i/timer.h>
#include <rss2i/systick.h>
#include <rss2i/debug.h>

#include <rss2i/config/timer.h>


#if TIMER_TIMERS > 0xffff
	#error "Too many timers requested. Should be <= 0xffff."
#endif

typedef struct{
	
	volatile bool enabled;          /* Whether this timer is enabled or not. */
	
	void (* volatile call)();       /* Function to call when                 */
	                                /* the timer expires.                    */
	
	volatile bool * volatile flag;  /* Boolean flag to set when              */
	                                /* the timer expires.                    */
	
	volatile uint32_t last;         /* Last valid timer value. The           */
	                                /* timer expires afterwards.             */
	
	volatile uint32_t ms;           /* Timer interval. This is necessary     */
	                                /* for repeating after expiration.       */
	
	volatile uint32_t token;        /* Unique token for clearing and         */
	                                /* resetting the timer.                  */
	
	volatile bool repeat;           /* Whether to repeat the timer or not.   */

} Timer;

static Timer timer[TIMER_TIMERS];


void timer_init(void){
/*
  Initializes the timer module. It gets initialized once even if called multiple
  times.
*/
	
	static bool initialized = false;
	
	if(!initialized){
		
		debug_init();
		systick_init();
		
		initialized = true;
	}
}

static Timer * timer_config(void (* call)(), volatile bool * flag, uint32_t ms, bool repeat){
/*
  Sets up a timer (if available) with the specified settings.
  
  Argument list
  =============
  
   call     Function to be called when the timer is expired. May be {NULL}.
   flag     Variable to be set to true when the timer is expired. May be {NULL}.
   ms       Number of milliseconds after which the timer should expire.
   repeat   Wheter the timer should get reset after expiration or not. There
            exist macros {TIMER_REPEAT} and {TIMER_NO_REPEAT}.
  
  Return value
  ============
   
   Returns {NULL} if there are no more timers available. Returns teh address of
   a {Timer} otherwise.
*/

	size_t i;
	uint16_t token;
	
	for(i = 0; i < TIMER_TIMERS; ++i){
	
		if(!(timer[i].enabled)){
		
			if(ms <= TIMER_TIME_LIMIT){
			
				/* The expiry date is rounded up. */
				timer[i].last = systick_get_ms() + ms;
				
				timer[i].ms = ms;
				timer[i].call = call;
				timer[i].flag = flag;
				timer[i].repeat = repeat;
				
				/* Use the first 16 bits as "unique" token and the next */
				/* 16 bits to store the timer's array index.            */
				token = timer[i].token & 0xffff;
				if(!(++token)){ ++token; }
				timer[i].token = (uint32_t) token | ((uint32_t) i << 16);
				
				timer[i].enabled = true;
				
				return &timer[i];
			}
			else{
				
				debug_report(DEBUG_TIMER_LIMIT_EXCEEDED);
				return NULL;
			}
		}
	}
	
	/* No free timers available */
	
	debug_report(DEBUG_NO_FREE_TIMERS);
	return NULL;
}

TimerID timer_call(void (* call)(), uint32_t ms, bool repeat){
/*
  Calls a given function after a specified number of milliseconds. Returns
  immediately.
  
  Argument list
  =============
  
   call     Function to be called when the timer is expired. May be {NULL}.
   ms       Number of milliseconds after which the timer should expire.
   repeat   Wheter the timer should get reset after expiration or not. There
            exist macros {TIMER_REPEAT} and {TIMER_NO_REPEAT}.
  
  Return value
  ============
   
   Returns {0} if there are no more timers available. Returns a positive timer
   ID otherwise. This ID may be used in order to reset the timer prior to its
   expiration.
*/
	Timer * tim;
	
	if((tim = timer_config(call, NULL, ms, repeat))){
		
		return tim->token;
	}
	else{
		
		return 0;
	}
}

bool timer_wait(uint32_t ms){
/*
  Delays execution for a period of {ms} milliseconds. Returns {false} if there
  are no more timers available. Delays execution for the given time and returns
  {true} otherwise.
*/

	volatile bool flag = false;
	Timer * tim = timer_config(NULL, &flag, ms, TIMER_NO_REPEAT);
	if(tim){
	
		while(!flag){ timer_trigger(); }
		return true;
	}
	else{
		
		return false;
	}
}

TimerID timer_flag(volatile bool * flag, uint32_t ms, bool repeat){
/*
  Sets the variable {flag} to true after a given number of milliseconds.
  
  Argument list
  =============
  
   flag     Variable to be set to true when the timer is expired.
   ms       Number of milliseconds after which the timer should expire.
   repeat   Wheter the timer should get reset after expiration or not. There
            exist macros {TIMER_REPEAT} and {TIMER_NO_REPEAT}.
  
  Return value
  ============
   
   Returns {0} if there are no more timers available. Returns a positive timer
   ID otherwise. This ID may be used in order to reset the timer prior to its
   expiration.
*/
	
	Timer * tim;
	
	if((tim = timer_config(NULL, (bool *) flag, ms, repeat))){
	
		return tim->token;
	}
	else{
		
		return 0;
	}
}

void timer_clear(TimerID token){
/*
  Disables the timer with the specified {token} (if any).
*/
	
	uint16_t index = token >> 16;
	token &= 0xffff;
	
	if(index < TIMER_TIMERS && (timer[index].token & 0xffff) == token){
		
		timer[index].enabled = false;
	}
}

void timer_reset(TimerID token){
/*
  Resets the timer, i.e. starts the timer over again, with the specified
  {token} (if any).
*/
	
	uint16_t index = token >> 16;
	token &= 0xffff;
	
	if(index < TIMER_TIMERS && (timer[index].token & 0xffff) == token){
		
		timer[index].last = systick_get_ms() + timer[index].ms;
	}
}

void timer_trigger(void){
/*
  Checks all available timers (one per call).
*/
	
	uint32_t time = systick_get_ms();
	
	static uint32_t lastCheckedTime = 0xfffffffful;
	static bool checkEnabled = false;
	static size_t tmr = 0;
	
	if(lastCheckedTime != time && !checkEnabled){
		
		checkEnabled = true;
		lastCheckedTime = time;
	}
	
	if(checkEnabled){
		
		if(timer[tmr].enabled){
		
			if(timer[tmr].last - time > TIMER_TIME_LIMIT){
				
				if(timer[tmr].call){
			
					(*timer[tmr].call)();
				}
				if(timer[tmr].flag){
				
					*timer[tmr].flag = true;
				}
			
				if(!timer[tmr].repeat){
				
					timer[tmr].enabled = false;
				}
				else{
					
					timer[tmr].last = time + timer[tmr].ms;
				}
			}
		}
		
		if(++tmr == TIMER_TIMERS){
			
			tmr = 0;
			checkEnabled = false;
		}
	}
}

