/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

#define RSS2I_MODULE "uarte"

/*
  Fast UART (up to 1 MBaud/s) with buffers for both communication directions.
  
  Authors
  =======
    
    Copyright (C) 2018-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
    
    2019-04-15    Initial public release.
    2019-06-06    ReAdded support for Rev. A using {board_rev()}
    2019-11-04    Split the baudrate change procedure into multiple, successive
                  tasks that have to be triggered by the command module.
                  Therefore, {uarte_set_baudrate_blocking()} has been replaced
                  by {uarte_change_baudrate()}. This partially fixed #6.
    2020-07-07    Added {uarte_send_*()} DMA-based procedures.
    2021-03-27    Use {NVIC_RSS2I_*_PRIORITY} macros.
  
  Description
  ===========
    
    This module has to get initialized using the {uarte_init()} method. After
    that initialization there are input and output buffers available as well as
    getter and setter functions for those buffers.
    
    UARTE communication has _not_ to get triggered by the user.
*/

#include <stdint.h>
#include <stdbool.h>
#include <stddef.h>

#include <libopencm3/cm3/nvic.h>
#include <libopencm3/nrf5/gpio.h>
#include <libopencm3/nrf5/uarte.h>

#include <rss2i/buffer.h>
#include <rss2i/debug.h>
#include <rss2i/timer.h>
#include <rss2i/board.h>
#include <rss2i/uarte.h>
#include <rss2i/nvic.h>


#define UARTE_PIN_TXD_REV_B 11
#define UARTE_GPIO_TXD_REV_B GPIO11
#define UARTE_PIN_RXD_REV_B 12
#define UARTE_GPIO_RXD_REV_B GPIO12

#define UARTE_PIN_TXD_REV_A 9
#define UARTE_GPIO_TXD_REV_A GPIO9
#define UARTE_PIN_RXD_REV_A 10
#define UARTE_GPIO_RXD_REV_A GPIO10


static Buffer uarteSendBuffer;
static size_t uarteSendWidth;
static Buffer uarteRecvBuffer;
static size_t uarteRecvWidth;
static volatile bool uarteBusy;

static bool uarteSendLocked = false;
static volatile uint8_t * uarteRecvData[2];
static bool uarteRecvdFirstTime = true;


static UARTEBaudrate uarte_set_baudrate_immediately(UARTEBaudrate baudrate){
/*
   Sets the UARTE's baudrate immediately to the specified one. If the given 
   {baudrate} parameter is not a valid one, a {baudrate} of
   {UARTE_BAUDRATE_115200} will be applied.
   
   Returns the applied {baudrate} parameter.
*/

	switch(baudrate){
		
		case UARTE_BAUDRATE_230400:
			UARTE_BAUDRATE(UARTE0) = UARTE_BAUDRATE_BAUDRATE_BAUD230400;
		break;
		case UARTE_BAUDRATE_250000:
			UARTE_BAUDRATE(UARTE0) = UARTE_BAUDRATE_BAUDRATE_BAUD250000;
		break;
		case UARTE_BAUDRATE_460800:
			UARTE_BAUDRATE(UARTE0) = UARTE_BAUDRATE_BAUDRATE_BAUD460800;
		break;
		case UARTE_BAUDRATE_921600:
			UARTE_BAUDRATE(UARTE0) = UARTE_BAUDRATE_BAUDRATE_BAUD921600;
		break;
		case UARTE_BAUDRATE_1000000:
			UARTE_BAUDRATE(UARTE0) = UARTE_BAUDRATE_BAUDRATE_BAUD1M;
		break;
		default:
			UARTE_BAUDRATE(UARTE0) = UARTE_BAUDRATE_BAUDRATE_BAUD115200;
			baudrate = UARTE_BAUDRATE_115200;
	}
	
	return baudrate;
}

uint8_t uarte_change_baudrate(UARTEBaudrate baudrate){
/*
  Changes the UARTE's baudrate to the specified one. Meanwhile the reception is
  paused. Since this may take some time, this function has to be called multiple
  times (see below).
  
  Argument list
  =============
  
    {baudrate}        Initial call: Desired baudrate. Further calls (until the
                      applied baudrate (see below) gets finally returned):
                      {UARTE_BAUDRATE_CHANGE_ONGOING}
  
  Return value
  ============
    
    Returns {UARTE_BAUDRATE_CHANGE_ONGOING} as long as additional function calls
    are necessary to finish the baudrate change. Once the baudrate was changed
    successfully, it finally returns the applied baudrate parameter.
*/
	
	typedef enum {
		UARTE_BAUDRATE_STATE_READY,
		UARTE_BAUDRATE_STATE_WAIT_BUSY,
		UARTE_BAUDRATE_STATE_WAIT_FILLUPS,
		
	} UARTEBaudrateState;

	static UARTEBaudrateState baudrateState = UARTE_BAUDRATE_STATE_READY;
	static UARTEBaudrate desiredBaudrate;
	static volatile bool timeout = false;
	
	
	/* {UARTE_BAUDRATE_CHANGE_ONGOING} should be passed if and only if */
	/* the {baudrateState} is not {UARTE_BAUDRATE_STATE_READY}.        */
	if((baudrate == UARTE_BAUDRATE_CHANGE_ONGOING
	&&  baudrateState == UARTE_BAUDRATE_STATE_READY)
	|| (baudrate != UARTE_BAUDRATE_CHANGE_ONGOING
	&& baudrateState != UARTE_BAUDRATE_STATE_READY)){
		
		debug_report(DEBUG_UARTE_STATE_MACHINE_MISMATCH);
	}
	
	
	switch(baudrateState){
		
		case UARTE_BAUDRATE_STATE_READY:
			
			UARTE_TASKS_STOPRX(UARTE0) = 1;
			uarteSendLocked = true;
			
			desiredBaudrate = baudrate;
			baudrateState = UARTE_BAUDRATE_STATE_WAIT_BUSY;
			
			return UARTE_BAUDRATE_CHANGE_ONGOING;
		break;
		case UARTE_BAUDRATE_STATE_WAIT_BUSY:
			
			if(!uarteBusy){
				
				/* Fill-up bytes {0xff} from the baudrate change command, */
				/* which get received _after_ the requested baudrate      */
				/* change has been applied, will be misinterpreted and    */
				/* will therefore corrupt the communication.              */
				timeout = false;
				timer_flag(&timeout, 2, TIMER_NO_REPEAT);
				
				baudrateState = UARTE_BAUDRATE_STATE_WAIT_FILLUPS;
			}
			
			return UARTE_BAUDRATE_CHANGE_ONGOING;
		break;
		case UARTE_BAUDRATE_STATE_WAIT_FILLUPS:
			
			if(timeout){
			
				uarteRecvdFirstTime = true;
				buffer_empty(&uarteRecvBuffer);
				
				desiredBaudrate = uarte_set_baudrate_immediately(desiredBaudrate);
				
				UARTE_TASKS_STARTRX(UARTE0) = 1;
				uarteSendLocked = false;
				
				baudrateState = UARTE_BAUDRATE_STATE_READY;
				return desiredBaudrate;
			}
			else{
				
				return UARTE_BAUDRATE_CHANGE_ONGOING;
			}
		break;
	}
	
	return UARTE_BAUDRATE_CHANGE_ONGOING;
}

void uarte_init(UARTEBaudrate baudrate,
                uint8_t * firstRecvData1,
                uint8_t * firstRecvData2,
                uint8_t * secRecvData,
                size_t * secRecvSize,
                size_t secRecvLength,
                size_t secRecvWidth,
                uint8_t * sendData,
                size_t * sendSize,
                size_t sendLength,
                size_t sendWidth){
/*
  Initializes the UARTE peripheral and some buffers for both communication
  directions. Received data is buffered using two buffering stages (a small
  but fast one and a larger but slower one) in order to avoid information loss.
  
  Argument list
  =============
  
    {baudrate}        Baudrate.
    {firstRecvData1}  Pointer to first level reception buffer data row.
    {firstRecvData2}  Pointer to another first level reception buffer data row.
    {secRecvData}     Pointer to sencond level reception buffer data rows.
    {secRecvSize}     Pointer to second level reception buffer size data.
    {secRecvLength}   Number of second level reception buffer data rows.
    {secRecvWidth}    Size of second level reception buffer data row.
    {sendData}        Pointer to send buffer data rows.
    {sendSize}        Pointer to send buffer size data.
    {sendLength}      Number of send buffer data rows.
    {sendWidth}       Size of send buffer data row.
    
*/
	/* Initialize buffer. */
	
	uarteRecvData[0] = firstRecvData1;
	uarteRecvData[1] = firstRecvData2;
	uarteRecvWidth = secRecvWidth;
	uarteSendWidth = sendWidth;
	
	buffer_init_rows(&uarteSendBuffer, sendData, sendSize, sendLength, sendWidth);
	buffer_init_rows(&uarteRecvBuffer, secRecvData, secRecvSize, secRecvLength, secRecvWidth);
	
	/* Initialized UARTE peripheral. */
	
	if(board_rev() > 0){
	
		GPIO_DIRSET = UARTE_GPIO_TXD_REV_B;
		GPIO_OUTSET = UARTE_GPIO_TXD_REV_B;
	}
	else{
	
		GPIO_DIRSET = UARTE_GPIO_TXD_REV_A;
		GPIO_OUTSET = UARTE_GPIO_TXD_REV_A;
	}
	
	UARTE_ENABLE(UARTE0) = UARTE_ENABLE_ENABLE_ENABLED;
	
	UARTE_INTENSET(UARTE0) = UARTE_INTENSET_ENDTX_SET
	                       | UARTE_INTENSET_RXSTARTED_SET;
	
	UARTE_SHORTS(UARTE0) = UARTE_SHORTS_ENDRX_STARTRX_ENABLED;
	
	UARTE_RXD(UARTE0).PTR = (uint32_t) uarteRecvData[0];
	UARTE_RXD(UARTE0).MAXCNT = secRecvWidth;
	
	if(board_rev() > 0){
	
		UARTE_PSEL(UARTE0).TXD = UARTE_PIN_TXD_REV_B;
		UARTE_PSEL(UARTE0).RXD = UARTE_PIN_RXD_REV_B;
	}
	else{
	
		UARTE_PSEL(UARTE0).TXD = UARTE_PIN_TXD_REV_A;
		UARTE_PSEL(UARTE0).RXD = UARTE_PIN_RXD_REV_A;
	}
	
	uarte_set_baudrate_immediately(baudrate);
	
	nvic_enable_irq(NVIC_UARTE0_UART0_IRQ);
	nvic_set_priority(NVIC_UARTE0_UART0_IRQ, NVIC_RSS2I_LOW_PRIORITY);
	
	UARTE_TASKS_STARTRX(UARTE0) = 1;
}

static void uarte_trigger(void){
	
	size_t size;
	
	if(!uarteSendLocked && (size = buffer_next_size(&uarteSendBuffer))){
	
		uarteBusy = true;
		
		UARTE_TXD(UARTE0).MAXCNT = size;
		UARTE_TXD(UARTE0).PTR = (uint32_t) buffer_get_row(&uarteSendBuffer);
		UARTE_TASKS_STARTTX(UARTE0) = 1;
	}
	else{
	
		uarteBusy = false;
	}
}

uint8_t uarte_vacant_entries(void){
/*
   Returns the number of free buffer entries.
*/
	
	return buffer_vacant_entries(&uarteSendBuffer);
}

uint8_t * uarte_send_request(void){
/*
  Returns a RAM address where the next UARTE data row can be stored before it is
  going to be sent by calling {uarte_send_finish()}. Returns {NULL} if the UARTE
  output buffer is full.
    
  Memory Write Warning
  ====================
    
    The user has to ensure that the length of the written data complies with the
    maximal reserved memory per output buffer entry as it was specified when
    initiating the UARTE module by calling {uarte_init()}.
*/
	
	uint8_t * addr;
	
	if((addr = buffer_new_row_address(&uarteSendBuffer))){
		
		return addr;
	}
	else{
		
		return NULL;
	}
}

bool uarte_send_finish(uint8_t * addr, size_t size){
/*
  Finishes the UARTE send procedure that was initiated by calling
  {uarte_send_request()}.
  
  Some Notes
  ==========
  
    This function does not perform any copying tasks. The specified {addr} is
    used for validation purposes only.
    
    The user has to ensure that {size} complies with the maximal reserved
    memory per output buffer entry as it was specified when initiating the UARTE
    module by calling {uarte_init()}. Note, that in contrast to the behaviour of
    {uarte_send_data()}, the data to be sent is not splitted into multiple
    queries automatically.
  
  Argument List
  =============
    
    addr    Pointer to the already written RAM as it was returned by the
            previous call to {uarte_send_request()}.
    size    Number of written bytes.
  
  Return Value
  ============
    
    Returns {false} when illegal buffer activities, i.e. buffer activities
    between the call of {uarte_send_request()} and the call to this function,
    have been detetcted. Returns {true} otherwise.
*/
	
	if(buffer_new_row_finish(&uarteSendBuffer, addr, size)){
		
		if(!uarteBusy){ uarte_trigger(); }
		return true;
	}
	else{
		
		return false;
	}
}

bool uarte_send_data(uint8_t * data, size_t size){
/*
  Sends the given {data} via UART. Returns {false} if the UART output buffer
  is full. Returns {true} otherwise.
*/
	size_t iterations = (size - 1) / uarteSendWidth + 1;
	const size_t width = uarteSendWidth;
	size_t i;
	
	for(i = 0; i < iterations; ++i){
		
		if(!buffer_append_row(
		    	&uarteSendBuffer,
		    	&data[width * i],
		    	(i == iterations - 1) ? size - width * i : width)){
			
			return false;
		}
	}
	
	if(!uarteBusy){ uarte_trigger(); }
	
	return true;
}

uint8_t * uarte_recv(void){
/*
   Returns an entry from the UART input buffer (if any).
*/

	return buffer_get_row(&uarteRecvBuffer);
}


void uarte0_uart0_isr(void){

	static uint8_t recvPtr = 0;

	if(UARTE_EVENTS_ENDTX(UARTE0)){
		
		UARTE_EVENTS_ENDTX(UARTE0) = 0;
		
		uarte_trigger();
		
		while(UARTE_EVENTS_ENDTX(UARTE0));
	}
	else if(UARTE_EVENTS_RXSTARTED(UARTE0)){
	
		UARTE_EVENTS_RXSTARTED(UARTE0) = 0;
		
		UARTE_RXD(UARTE0).PTR = (uint32_t) uarteRecvData[recvPtr = (recvPtr + 1) % 2];
		
		if(!uarteRecvdFirstTime){
			
			if(!buffer_append_row(&uarteRecvBuffer, (uint8_t *) uarteRecvData[recvPtr], uarteRecvWidth)){
				
				/* Although this is the UARTE module, using */
				/* the "USB" prefix is preferred here.      */
				debug_report(DEBUG_USB_RECV_BUFFER_OVERFLOW);
			}
		}
		else{
			
			/* STARTRX was triggered by software. Hence there is no data to */
			/* be stored in the buffer.                                     */
			uarteRecvdFirstTime = false;
		}
		
		while(UARTE_EVENTS_RXSTARTED(UARTE0));
	}
}

