/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

#define RSS2I_MODULE "usb"

/*
  UART-over-USB interface.
  
  Authors
  =======
    
    Copyright (C) 2018-2020  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
    
    2019-04-15    Initial public release.
    2020-07-07    Add {usb_send_*()} DMA based procedures.
  
  Description
  ===========
    
    This module has to get initialized using the {usb_init()} method. Use
    {usb_send()} in order to send (arbitrarily long) data over USB or
    {usb_recv()} in order to get previously received data (if any).
    
    In the case the data to be sent has at most {USB_SEND_BUFFER_WIDTH} bytes
    it can be sent by using the DMA based {usb_send_*()} procedures.
    
    This is a very transparent abstraction layer (maybe "pass-through layer"
    would be a more appropriate name) for the UARTE module. We keep this for
    historical reasons.
*/

#include <stdint.h>
#include <stdbool.h>
#include <stddef.h>

#include <rss2i/buffer.h>
#include <rss2i/debug.h>
#include <rss2i/uarte.h>

#include <rss2i/config/usb.h>


size_t usb_vacant_entries(void){
	
	return uarte_vacant_entries();
}

void usb_init(void){
	
	static uint8_t sendData[USB_SEND_BUFFER_LENGTH][USB_SEND_BUFFER_WIDTH];
	static uint8_t firstRecvData[2][USB_RECV_BUFFER_WIDTH];
	static uint8_t secRecvData[USB_RECV_BUFFER_LENGTH][USB_RECV_BUFFER_WIDTH];
	static size_t sendSize[USB_SEND_BUFFER_LENGTH];
	static size_t secRecvSize[USB_RECV_BUFFER_LENGTH];
		
	uarte_init(USB_BAUDRATE,
	           &firstRecvData[0][0],
	           &firstRecvData[1][0],
	           &secRecvData[0][0],
	           &secRecvSize[0],
	           USB_RECV_BUFFER_LENGTH,
	           USB_RECV_BUFFER_WIDTH,
	           &sendData[0][0],
	           &sendSize[0],
	           USB_SEND_BUFFER_LENGTH,
	           USB_SEND_BUFFER_WIDTH);
	
}

uint8_t * usb_send_request(void){
/*
  Returns a RAM address where the next USB data row can be stored before it is
  going to be sent by calling {usb_send_finish()}. Returns {NULL} if the USB
  output buffer is full.
    
  Memory Write Warning
  ====================
    
    The user has to ensure that the length of the written data complies with the
    reserved memory per output buffer entry, i.e. {USB_SEND_BUFFER_WIDTH}.
*/
	
	uint8_t * addr;
	
	if((addr = uarte_send_request())){
	
		return addr;
	}
	else{
	
		debug_report(DEBUG_USB_SEND_BUFFER_OVERFLOW);
		return NULL;
	}
}

bool usb_send_finish(uint8_t * addr, size_t size){
/*
  Finishes the USB send procedure that was initiated by calling
  {usb_send_request()}.
  
  Some Notes
  ==========
  
    This function does not perform any copying tasks. The specified {addr} is
    used for validation pruposes only.
    
    The user has to ensure that {size} complies with the maximal reserved
    memory per output buffer entry, i.e. {USB_SEND_BUFFER_WIDTH}. Note, that in
    contrast to the behaviour of {usb_send()}, the data to be sent is not
    splitted into multiple queries automatically.
  
  Argument List
  =============
    
    addr    Pointer to the already written RAM as it was returned by the
            previous call to {usb_send_request()}.
    size    Number of written bytes.
  
  Return Value
  ============
    
    Returns {false} when illegal buffer activities, i.e. buffer activities
    between the call of {usb_send_request()} and the call to this function,
    have been detetcted. Returns {true} otherwise.
*/
	
	return uarte_send_finish(addr, size);
}

bool usb_send(uint8_t * data, size_t size){
/*
  Sends the given {data} via USB. Returns {false} if the USB output buffer
  is full. Returns {true} otherwise.
*/
	
	if(uarte_send_data(data, size)){
	
		return true;
	}
	else{
	
		debug_report(DEBUG_USB_SEND_BUFFER_OVERFLOW);
		return false;
	}
}

uint8_t * usb_recv(void){
	
	return uarte_recv();
}

