/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

/*
  Test firmware designed for initial tests after hardware production.
  
  Authors
  =======
  
    Copyright (C) 2018-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This firmware is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This firmware is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this firmware. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
  
    2019-06-26    Initial release.
    2019-07-12    Added triggering of the timer module.
    2020-07-03    Introduced the scheduler.
    2020-07-07    Added {scheduler_init()}.
    2020-07-10    Disabled nRF iCache technology explicitly.
    2020-08-13    Emit carrier using alternating antennas.
    2020-08-14    Fit to changes in the RADIO module.
    2021-03-11    Prepare CTE prior to usage.
    2021-03-26    Prepare CTE prior to usage for all board types.
    2021-03-29    Use {SwitchAntenna} for {radio_emit_carrier()} configuration.
    2021-03-29    Consider {RADIO_CONFIG_FREQUENCY}.
  
  Description
  ===========
  
    This test firmware is designed to detect hardware faults thay may occur
    in production. This test contains:
    
      - Ramp up of the externally driven High Frequency Clock (HFCLK).
      - Enabling of the externally supported onboard DC/DC regulator.
      - Enabling of all LEDs in the first 500 ms after power up.
      - Accessing both external flash chips.
    
    After those tests, independent of their outcomes, the RADIO will continously
    emit on {2400 + RADIO_CONFIG_FREQUENCY} MHz. Boards with an onboard antenna
    array will switch the used antenna every 5 seconds.
*/


#include <rss2i/clock.h>
#include <rss2i/measure.h>
#include <rss2i/battery.h>
#include <rss2i/led.h>
#include <rss2i/debug.h>
#include <rss2i/control.h>
#include <rss2i/icache.h>
#include <rss2i/radio.h>
#include <rss2i/power.h>
#include <rss2i/watchdog.h>
#include <rss2i/board.h>
#include <rss2i/mlc.h>
#include <rss2i/timer.h>
#include <rss2i/scheduler.h>
#include <rss2i/memory.h>
#include <rss2i/switch.h>

#include <rss2i/config/generated.radio.h>


_Static_assert(
	
	   RADIO_CONFIG_FREQUENCY <= 81
	&& RADIO_CONFIG_FREQUENCY >=  2
	&& (uint8_t) RADIO_CONFIG_FREQUENCY == RADIO_CONFIG_FREQUENCY,
	"Invalid RADIO FREQUENCY setting."
);


int main(){
	
	bool testRunning = true;
	MemoryTestStatus testStatus;
	
	bool timeout = true;
	uint8_t antenna = 0;
	SwitchAntenna ant;
	
	
	led_init();          /* Controls the onboard LEDs.                        */
	timer_init();        /* Provides some timers.                             */
	debug_init();        /* Debug and reporting system.                       */
	
	led_set(LED_YELLOW, LED_ON);
	led_set(LED_GREEN, LED_ON);
	led_set(LED_RED, LED_ON);
	
	timer_wait(500);
	
	led_set(LED_YELLOW, LED_OFF);
	led_set(LED_GREEN, LED_OFF);
	led_set(LED_RED, LED_OFF);
	
	
	board_init();        /* Holds board-specific parameters.                  */
	power_init();        /* Enables the internal DC/DC regulator.             */
	clock_init();        /* Configures the High Frequency Clock (HFCLK).      */
/*	battery_init();    *//* Monitors the battery voltage.                     */
	mlc_init();          /* Monitors the main loop frequency.                 */
	control_init();      /* Handles incoming commands.                        */
	measure_init();      /* Enters control mode.                              */
	scheduler_init();    /* Schedules all periodic tasks.                     */
/*	icache_init();     *//* Initializes the nRF iCache technology.            */
	
	watchdog_init();     /* Enables the watchdog.                             */
	
	memory_init();       /* API for the external memory.                      */
	switch_init();       /* API for the external HF switch.                   */
	
	led_set(LED_YELLOW, LED_ON);
	led_set(LED_GREEN, LED_OFF);
	
	
	while(1){
		
		if(testRunning){
			
			/* Trigger the fast memory test routine until it returns a status. */
			if((testStatus = memory_fast_test())){
				
				testRunning = false;
				
				if(testStatus == MEMORY_TEST_SUCCEEDED){
					
					led_set(LED_GREEN, LED_ON);
				}
				else{
					
					led_set(LED_RED, LED_ON);
				}
			}
		}
		else{
		
			if(timeout){
			
				timeout = false;
				
				if(board_type() == BOARD_TYPE_ARR){
					
					/* Switch antennas after five seconds. */
				
					ant = SWITCH_ANTENNA_1 + antenna;
					
					radio_emit_carrier(RADIO_CONFIG_FREQUENCY, ant);
					
					if(++antenna == 8){
						
						antenna = 0;
					}
					
					timer_flag(&timeout, 5000, TIMER_NO_REPEAT);
				}
				else{
					
					/* Just constantly emit carrier. */
					
					/* The chosen antenna is irrelevant for this board type. */
					radio_emit_carrier(
					
						RADIO_CONFIG_FREQUENCY,
						SWITCH_ANTENNA_1
					);
				}
			}
		}
		
		scheduler();
	}
	
	return 0;
}

