/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

/*
  Benchmark of the used Access Addresses.
   
  Authors
  =======
  
    Copyright (C) 2018-2019  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This program is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this program. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
  
    2019-05-27    Initial release.
  
  Description
  ===========

    The used Access Addresses (see src/ble.c) get checked against BLE 5.0
    Uncoded PHY compatibility and an adequate minimal hamming distance.
*/

#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <inttypes.h>

#include <rss2i/ble.h>


#define SHIFT_COUNT  5
#define MAX_DEV_ID 255


uint8_t hamming_distance(uint32_t a, uint32_t b){
/*
  Naive hamming distance calculation without using architecture specific
  optimizations.
*/

	uint8_t i;
	uint8_t hd = 0;
	
	for(i = 0; i < 32; ++i){
		
		if((a & (1ul << i)) != (b & (1ul << i))){
		
			++hd;
		}
	}
	
	return hd;
}

bool has_more_than_six_consecutive_digits(uint32_t a){
/*
  Whether the given 32-bit value has more than six consecutive zeros or ones.
*/
	
	bool one = (a & 0x1) ? true : false;
	uint8_t maxCount = 1;
	uint8_t count = 1;
	uint8_t i;
	
	for(i = 1; i < 32; ++i){
		
		if(a & (1ul << i) && one){
			
			++count;
		}
		else if(a & (1ul << i) && !one){
			
			/* Save {count} if necessary. */
			if(count > maxCount){ maxCount = count; }
			
			count = 1;
			one = true;
		}
		else if(!(a & (1ul << i)) && !one){
			
			++count;
		}
		else{
			
			/* Save {count} if necessary. */
			if(count > maxCount){ maxCount = count; }
			
			count = 1;
			one = false;
		}
	}
	
	return (maxCount <= 6) ? false : true;
}

bool has_more_than_24_transitions(uint32_t a){
/*
  Whether the given 32-bit value has more than 24 transitions.
*/
	
	bool one = (a & 0x1) ? true : false;
	uint8_t count = 0;
	uint8_t i;
	
	for(i = 1; i < 32; ++i){
		
		if(a & (1ul << i) && !one){
			
			++count;
			one = true;
		}
		else if(!(a & (1ul << i)) && one){
			
			++count;
			one = false;
		}
	}
	
	return (count <= 24) ? false : true;
}

bool has_less_than_two_transitions_in_the_most_significant_six_bits(uint32_t a){
/*
  Whether the given 32-bit value has less than two transitions in the most
  significant six bits.
*/
	
	bool one = (a & (1ul << 26)) ? true : false;
	uint8_t count = 0;
	uint8_t i;
	
	for(i = 27; i < 32; ++i){
		
		if(a & (1ul << i) && !one){
			
			++count;
			one = true;
		}
		else if(!(a & (1ul << i)) && one){
			
			++count;
			one = false;
		}
	}
	
	return (count >= 2) ? false : true;
}

bool has_all_four_octets_equal(uint32_t a){
/*
  Whether all for give octets are equal or not.
*/
	if((a & 0xff) == ((a >>  8) & 0xff)
	&& (a & 0xff) == ((a >> 16) & 0xff)
	&& (a & 0xff) == ((a >> 24) & 0xff)){
		
		return true;
	}
	else{
		
		return false;
	}
}

uint32_t access_address(uint16_t devID){
/*
  Returnes one of {MAX_DEV_ID + 4} Access Addresses.
  
  [0, MAX_DEV_ID}] map to the device specific Access Addresses used during
  control mode. [{MAX_DEV_ID + 1}, {MAX_DEV_ID + 3}] map to other special
  Access Addresses:
    
    {MAX_DEV_ID + 1}: Used for measurement packets.
    {MAX_DEV_ID + 2}: Used for disruptive packets.
    {MAX_DEV_ID + 3}: BLE 5.0 Uncoded PHY's Advertising Channel Packets'
                      Access Addresses.
*/

	if(devID <= MAX_DEV_ID){
		return ble_gen_control_access_address(devID);
	}
	else if(devID == MAX_DEV_ID + 1){
		return ble_gen_measure_access_address();
	}
	else if(devID == MAX_DEV_ID + 2){
		return ble_gen_disrupt_access_address();
	}
	else{
		return 0x8E89BED6UL;
	}
}


int main(void){
/*
  The used Access Addresses (see src/ble.c) get checked against BLE
  compatibility and an adequate minimal hamming distance.
    
  Returns 1 if the test was NOT successfull. It returns 0 if the test passed
  successfully.
*/

	bool ok;
	uint16_t devID, otherDevID;
	uint32_t aA, otherAA;
	
	uint8_t tmpHD;
	
	size_t i;
	
	uint8_t minHD[SHIFT_COUNT + 1];
	uint16_t minHDID[SHIFT_COUNT + 1];
	uint16_t minHDOtherID[SHIFT_COUNT + 1];
	uint32_t minHDCount[SHIFT_COUNT + 1];
	
	
	/* Checking for BLE Access Address compliance. */
	
	printf("            Checking for BLE 5.0 Access Address compliance... ");
	
	ok = true;
	for(devID = 0; devID < MAX_DEV_ID + 3; ++devID){
		
		aA = access_address(devID);
		
		if(has_more_than_six_consecutive_digits(aA)){
			printf("Failed.\n\n              -> device ID 0x%03" PRIx16 " has more than 6 consecutive digits.\n", devID);
			ok = false;
			break;
		}
		if(has_all_four_octets_equal(aA)){
			printf("Failed.\n\n              -> device ID 0x%03" PRIx16 " has all four octets equal.\n", (int) devID);
			ok = false;
			break;
		}
		if(has_less_than_two_transitions_in_the_most_significant_six_bits(aA)){
			printf("Failed.\n\n              -> device ID 0x%03" PRIx16 " has less than two transitions in the most\n                 significant six bits.\n", (int) devID);
			ok = false;
			break;
		}
		if(has_more_than_24_transitions(aA)){
			printf("Failed.\n\n              -> device ID 0x%03" PRIx16 " has more than 24 transitions.\n", (int) devID);
			ok = false;
			break;
		}
		if(hamming_distance(aA, access_address(MAX_DEV_ID + 3)) < 2){
			printf("Failed.\n\n              -> device ID 0x%03" PRIx16 " has a too low hamming distance from\n                 an Advertising Channel Packets' Access Address.\n", devID);
			ok = false;
			break;
		}
	}
	
	if(ok){
		
		/* Passed. */
		printf("OK.\n");
	}
	else{
		
		printf("\n");
		return 1;
	}
	
	printf("            Checking for adequate minimal Hamming Distance... ");
	
	for(i = 0; i <= SHIFT_COUNT; ++i){
	
		minHD[i] = 33;
		minHDID[i] = 0;
		minHDOtherID[i] = 0;
		minHDCount[i] = 0;
	
		for(devID = 0; devID < MAX_DEV_ID + 4; ++devID){
			for(otherDevID = 0; otherDevID < MAX_DEV_ID + 4; ++otherDevID){
		
				if(devID != otherDevID){
		
					aA = access_address(devID);
					
					/* Shift by {i}. */
					aA <<= i;
					
					otherAA = access_address(otherDevID);
					
					/* Disable those bits within {otherAA} that got disabled */
					/* within {aA} by the applied left shift.                */
					otherAA &= ~(((uint32_t) 1 << i) - 1);
					
					tmpHD = hamming_distance(otherAA, aA);
		
					if(tmpHD < minHD[i]){
			
						minHD[i] = tmpHD;
						minHDID[i] = devID;
						minHDOtherID[i] = otherDevID;
						minHDCount[i] = 0;
					}
					else if(tmpHD == minHD[i]){
			
						++minHDCount[i];
					}
				}
			}
		}
	}
	
	ok = true;
	for(i = 0; i <= SHIFT_COUNT; ++i){
			
		if(minHD[i] <= 5 - i){ /* Magick requirement ;-) */
			printf("Failed.\n\n              -> device IDs 0x%03" PRIx16 " (shifted by %" PRIu8 ") and 0x%03" PRIx16 " have a too low\n                 hamming distance from eachother.\n", minHDID[i], (uint8_t) i, minHDOtherID[i]);
			ok = false;
			break;
		}
	}
	
	if(ok){
		
		/* Passed. */
		printf("OK.\n");
	}
	
	printf("\n");
	printf("              +-----------+------+-----------------+-------------+\n");
	printf("              |  Shifted  | Min. | representative  |  number of  |\n");
	printf("              | by / bits |  HD  | device ID pair  | occurrences |\n");
	printf("              +-----------+------+-----------------+-------------+\n");
	
	for(i = 0; i <= SHIFT_COUNT; ++i){
			
		printf("              |    %02" PRIu8 "     |  %02" PRIu8 "  | 0x%03" PRIx16 " and 0x%03" PRIx16 " |    %05" PRIu32 "    |\n",
		       (uint8_t) i,
		       minHD[i],
		       minHDID[i],
		       minHDOtherID[i],
		       (minHDCount[i] + 1) / ((i == 0) ? 2 : 1));
		printf("              +-----------+------+-----------------+-------------+\n");
	}
	
	if(!ok){
		
		printf("\n");
		return 1;
	}
	else{
	
		return 0;
	}
}

