/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

/*
  Unit test for the buffer module.
   
  Authors
  =======
  
    Copyright (C) 2018-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This program is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this program. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
  
    2019-06-21    Initial release.
    2020-02-04    Fitted to the fixed trailing null character behaviour.
    2020-02-08    Added tests for critical parameters (width = 1, length = 2).
    2020-02-08    Added tests for Direct Memory Access (DMA).
    2020-03-31    Change terminology {DebugSource} -> {DebugErrorCase}.
    2021-03-11    Change some parameters.
  
  Description
  ===========
    
    Simulates _parallel_ read/write operations from/onto a couple of buffers
    with different lengths and widths with the following features:
      
      - Random parallel write and read operations.
      - Randomly generate massive write operations (force possible overflows).
      - Randomly unlock read buffer entries.
      - Verify the correct reporting of oversized buffer entries.
      - Randomly mix write procedures (copy to buffer and direct memory access).
      - Randomly disrupt direct memory access.
      - Randomly cancel direct memory access (append to buffer before finish).
      - Verify the correct reporting of "finish after append".
  
  Plausibility Check
  ==================
    
    The data that is read from the buffers is checked against possible data
    corruption. The data within each buffer entry shall contain a sequence of
    increasing or decreasing counter values. Two successive entries from the
    same buffer shall be consistent to each other, i.e. the last element of the
    first buffer entry shall be the predecessor or the successor of the first
    element of the second buffer entry. Any deviation from this rule lets the
    buffer test fail.
*/

#include <stdio.h>
#include <stdint.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <inttypes.h>
#include <locale.h>
#include <time.h>
#include <omp.h>

#include <sys/types.h>

#include <rss2i/buffer.h>
#include <rss2i/debug.h>


typedef enum{
	
	OK,
	CONTENT_MISMATCH,
	POSITIVE_SIZE_BUT_EMPTY,
	UNREPORTED_OVERSIZE,
	REPORTED_BUT_NO_OVERSIZE,
	UNEXPECTED_ERROR,
	NOTHING_TO_FINISH,
	REPORTED_BUT_NO_ACTIVITIES,
	NO_READ_THREAD,
	TERM_NULL_CHAR_EXPECTED,
	
} Error;


#define NUM_OF_BUFFERS                     20
#define MAX_LENGTH                       1024
#define MAX_WIDTH                         500
#define UNLOCK_PROBABILITY                 80 /* per mille. */
#define ADDRESS_PROBABILITY               100 /* per mille. */
#define DISRUPT_PROBABILITY                 2 /* per mille. */
#define ACTIVITY_PROBABILITY               20 /* per mille. */
#define TURN_PROBABILITY                  200 /* per mille. */
#define FREEZE_PROBABILITY                  5 /* per mille. */
#define FREEZE_PERIOD                     500 /* entries.   */


#define NUM_OF_TEST_WRITES_THRESH     2000000ul
#define MIN_NUM_OF_TEST_WRITES_THRESH   10000ul
#define EMPTY_THRESH                    50000ul
#define OVERFLOW_THRESH                100000ul
#define UNDERFLOW_THRESH             99000000ul
#define DISRUPT_THRESH                   2000
#define ACTIVITY_THRESH                   500
#define ADDRESS_THRESH                  20000
#define CRITICAL_WIDTH_THRESH            3500

#define MIN_NUM_OF_POWER_OF_TWO             5
#define MIN_NUM_OF_ROW_BUFFERS              5
#define MIN_NUM_OF_STRING_BUFFERS           5
#define MIN_NUM_OF_WIDTH_1                  1 /* per type. */
#define MIN_NUM_OF_LENGTH_2                 1 /* per type. */

/* Number of data columns within debug information. */
#define MAX_DATA_COLS 12


size_t bufferOversizedEntries = 0;
size_t bufferNothingToFinish = 0;
size_t bufferOtherErrors = 0;


void debug_report(DebugErrorCase ec){
/*
  Resolves the reference from inside buffer.c and counts its calls.
*/
	
	if(ec == DEBUG_BUFFER_WIDTH_EXCEEDED){
		
		++bufferOversizedEntries;
	}
	else if(ec == DEBUG_BUFFER_NOTHING_TO_FINISH){
		
		++bufferNothingToFinish;
	}
	else{
		
		++bufferOtherErrors;
	}
}

double get_ratio(size_t nom, size_t denom){
	
	if(nom > denom){
		
		return 1.0;
	}
	else{
		
		return (double) nom / denom;
	}
}


/* ########################################################################## */
/* ########################################################################## */
/* ##                                                                      ## */
/* ##    Some verbose print functions for debugging or progress info.      ## */
/* ##                                                                      ## */
/* ########################################################################## */
/* ########################################################################## */


void print_progress(int progress){
	
	if(progress < 0){ progress = 0; }
	if(progress > 100){ progress = 100; }
	
	printf("\b\b\b\b%03d%%", progress);
	fflush(stdout);
}

void print_failed(void){
/*
  Should be called prior to every error or debug message.
*/

	static bool thrown = false;
	
	if(!thrown){
	
		printf("\b\b\b\bFailed.\n\n              -> ");
		thrown = true;
	}
	else{
	
		printf("              -> ");
	}
}


void print_stat(const char * string, size_t val1, size_t val2){
/*
  Print quantity statistics.
*/
	printf("              | %-22s | %'11zu | %'10zu |\n", string, val1, val2);
	printf("              +------------------------+-------------+------------+\n");
}

void print_param(const char * string, size_t val, size_t min, size_t max){
/*
  Print parameter value.
*/
	printf("              | %-22s | %'11zu |  ", string, val);
	if(min > 0){
		printf("%3zu /", min);
	}
	else{
		printf(" -  /");
	}
	
	if(max > 0){
		printf("%3zu  |\n", max);
	}
	else{
		printf(" -   |\n");
	}
	
	printf("              +------------------------+-------------+------------+\n");
}

void print_debug_write(
	uint8_t bufferType,
	size_t bufWidth,
	size_t bufLength,
	size_t size,
	uint8_t * got,
	ssize_t addressPointer,
	size_t numOfTestWrites
){
/*
  Print some debug information in the case of unexpected buffer behaviour.
*/

	size_t ptr = MAX_DATA_COLS;
	size_t i;
	
	print_failed();
	printf("Buffer type  : %s\n", bufferType ? "ROW" : "STRING");
	print_failed();
	printf("Buffer width : %zu\n", bufWidth);
	print_failed();
	printf("Buffer length: %zu\n", bufLength);
	print_failed();
	printf("Entry  size  : %zu\n", size);
	print_failed();
	printf("Num of writes: %zu\n", numOfTestWrites);
	print_failed();
	
	if(addressPointer >= 0){
	
		printf("DMA pointer  : %zu", addressPointer);
	}
	else if(addressPointer == -2){
		
		printf("DMA pointer  : DMA finished.");
	}
	else{
		
		printf("DMA pointer  : No ongoing DMA");
	}
	
	for(i = 0; i < size; ++i){
		
		if(ptr == MAX_DATA_COLS){
			
			printf("\n");
			print_failed();
			printf("Data         : ");
			ptr = 0;
		}
		
		printf("%03" PRIu8 " ", got[i]);
		++ptr;
	}
	printf("\n\n");

}

void print_debug_read(

	uint8_t bufferType,
	size_t bufWidth,
	size_t bufLength,
	uint8_t lastCounter,
	size_t size,
	uint8_t * got,
	uint8_t expStart,
	int8_t expDir,
	size_t numOfTestWrites
){
/*
  Print some debug information.
*/
	
	size_t ptr = MAX_DATA_COLS;
	size_t i;
	
	print_failed();
	printf("Buffer type  : %s\n", bufferType ? "ROW" : "STRING");
	print_failed();
	printf("Buffer width : %zu\n", bufWidth);
	print_failed();
	printf("Buffer length: %zu\n", bufLength);
	print_failed();
	printf("Entry  size  : %zu\n", size);
	print_failed();
	printf("Num of writes: %zu\n", numOfTestWrites);
	print_failed();
	printf("Prior value  : %03" PRIu8, lastCounter);
	
	for(i = 0; i < size; ++i){
		
		if(ptr == MAX_DATA_COLS){
			
			printf("\n");
			print_failed();
			printf("Data         : ");
			ptr = 0;
		}
		
		printf("%03" PRIu8 " ", got[i]);
		++ptr;
	}
	
	printf("\n");
	print_failed();
	printf("Expected data: %03" PRIu8 " %03" PRIu8 " ... %03" PRIu8 " %03" PRIu8 "\n",
	       expStart,
	       (uint8_t) ((expStart + expDir || bufferType) ? expStart + expDir
	                                                    : expStart + 2 * expDir), /* avoid '\0' */
	       (uint8_t) ((expStart + (size - 2) * expDir || bufferType) ? expStart + (size - 2) * expDir
	                                                                 : expStart + (size - 2) * expDir + expDir * ((size - 2) / 255)), /* avoid '\0' */
	       (uint8_t) ((bufferType) ? expStart + (size - 1) * expDir : 0)
	      );
	printf("\n");
}


/* ########################################################################## */
/* ########################################################################## */
/* ##                                                                      ## */
/* ##    Actual buffer test procedure.                                     ## */
/* ##                                                                      ## */
/* ########################################################################## */
/* ########################################################################## */


int main(void){
	
	static Buffer buffer[NUM_OF_BUFFERS];
	static char cMem[NUM_OF_BUFFERS][MAX_LENGTH][MAX_WIDTH];
	static uint8_t uMem[NUM_OF_BUFFERS][MAX_LENGTH][MAX_WIDTH];
	static size_t sizeMem[NUM_OF_BUFFERS][MAX_LENGTH];
	
	static uint8_t counter[NUM_OF_BUFFERS];
	static uint8_t next[NUM_OF_BUFFERS];
	static int8_t dir[NUM_OF_BUFFERS];
	static size_t bufferWidth[NUM_OF_BUFFERS];
	static size_t entryWidth[NUM_OF_BUFFERS];
	static size_t bufferLength[NUM_OF_BUFFERS];
	static uint8_t bufferType[NUM_OF_BUFFERS];
	static uint8_t * bufferRowAddress[NUM_OF_BUFFERS];
	static size_t bufferAddressPtr[NUM_OF_BUFFERS];
	static size_t oldCounter[NUM_OF_BUFFERS];
	static bool genNew[NUM_OF_BUFFERS];
	static bool addressOngoing[NUM_OF_BUFFERS];
	
	static char testString[NUM_OF_BUFFERS][MAX_WIDTH + 1];
	static uint8_t testRow[NUM_OF_BUFFERS][MAX_WIDTH + 1];
	
	size_t numOfTestWrites[NUM_OF_BUFFERS];
	
	size_t i, j, k, l;
	
	size_t numOfPowerTwo = 0;
	size_t numOfRowBuffers = 0;
	size_t numOfWidth1[2];
	size_t numOfLength2[2];
	
	size_t overflows = 0;
	size_t underflows = 0;
	size_t criticalWidths[5] = {0, 0, 0, 0, 0};
	size_t oversized = 0;
	size_t addressed = 0;
	size_t disrupted = 0;
	size_t activities = 0;
	size_t totalNumOfTestWrites = 0;
	size_t minNumOfTestWrites = 0;
	size_t freezeState = 0;
	size_t freezeBuffer;
	uint8_t lastCounterVal;
	int8_t thisDir;
	
	size_t empty = 0;
	size_t read = 0;
	uint8_t * data;
	size_t size;
	bool writeInProgress = true;
	
	Error readError = OK;
	Error writeError = OK;
	
	srand(time(NULL));
	setlocale(LC_NUMERIC, "");
	

	/* ###################################################################### */
	/* #                                                                    # */
	/* #  Generate bufferTypes, Lengths and widths and initialize them.     # */
	/* #                                                                    # */
	/* ###################################################################### */
	
	printf("            Initializing a couple of different buffers... ");
	fflush(stdout);
	
	/* Randomly generate buffer types. */
	while(numOfRowBuffers < MIN_NUM_OF_ROW_BUFFERS
	   || NUM_OF_BUFFERS - numOfRowBuffers < MIN_NUM_OF_STRING_BUFFERS){
		
		numOfRowBuffers = 0;
		
		for(i = 0; i < NUM_OF_BUFFERS; ++i){
		
			bufferType[i] = rand() % 2;
		
			if(bufferType[i]){
				
				++numOfRowBuffers;
			}
		}
	}
	
	/* Randomly generate buffer lengths. */
	while(numOfPowerTwo < MIN_NUM_OF_POWER_OF_TWO
	   || numOfLength2[0] < MIN_NUM_OF_LENGTH_2
	   || numOfLength2[1] < MIN_NUM_OF_LENGTH_2){
		
		numOfPowerTwo = 0;
		numOfLength2[0] = 0;
		numOfLength2[1] = 0;
		
		for(i = 0; i < NUM_OF_BUFFERS; ++i){
		
			bufferLength[i] = (rand() % (MAX_LENGTH + 1 - 2)) + 2;
		
			if((bufferLength[i] & (bufferLength[i] - 1)) == 0){
				
				++numOfPowerTwo;
			}
			
			if(bufferLength[i] == 2){
			
				++numOfLength2[bufferType[i]];
			}
		}
	}
	
	/* Randomly generate buffer widths. */
	while(numOfWidth1[0] < MIN_NUM_OF_WIDTH_1
	   || numOfWidth1[1] < MIN_NUM_OF_WIDTH_1){
		
		numOfWidth1[0] = 0;
		numOfWidth1[1] = 0;
		
		for(i = 0; i < NUM_OF_BUFFERS; ++i){
		
			bufferWidth[i] = (rand() % (MAX_WIDTH + 1 - 1)) + 1;
		
			if(bufferWidth[i] == 1){
			
				++numOfWidth1[bufferType[i]];
			}
		}
	}
	
	/* Initialize buffers. */
	
	for(i = 0; i < NUM_OF_BUFFERS; ++i){
	
		genNew[i] = true;
		addressOngoing[i] = false;
		dir[i] = 1;
		counter[i] = 0;
		next[i] = 0;
		numOfTestWrites[i] = 0;
		
		if(bufferType[i]){
			
			/* Buffer of rows. */
			buffer_init_rows(&buffer[i], &uMem[i][0][0], &sizeMem[i][0], bufferLength[i], bufferWidth[i]);
		}
		else{
		
			/* Buffer of strings. */
			buffer_init_strings(&buffer[i], &cMem[i][0][0], &sizeMem[i][0], bufferLength[i], bufferWidth[i]);
		}
	}
	
	/* Print the generated configuration. */
	
	printf("OK.\n");
	
	printf("\n");
	printf("              +------------------------+-------------+------------+\n");
	printf("              |        Quantity        |    Value    |  Min./Max. |\n");
	printf("              +------------------------+-------------+------------+\n");
	
	print_param("row buffers (row buf)", numOfRowBuffers, MIN_NUM_OF_ROW_BUFFERS, NUM_OF_BUFFERS - MIN_NUM_OF_STRING_BUFFERS);
	print_param("string bufs (str buf)", NUM_OF_BUFFERS - numOfRowBuffers, MIN_NUM_OF_STRING_BUFFERS, NUM_OF_BUFFERS - MIN_NUM_OF_ROW_BUFFERS);
	print_param("buf: length = 2^n", numOfPowerTwo, MIN_NUM_OF_POWER_OF_TWO, 0);
	print_param("row buf: width = 1", numOfWidth1[1], MIN_NUM_OF_WIDTH_1, 0);
	print_param("str buf: width = 1", numOfWidth1[0], MIN_NUM_OF_WIDTH_1, 0);
	print_param("row buf: length = 2", numOfLength2[1], MIN_NUM_OF_LENGTH_2, 0);
	print_param("str buf: length = 2", numOfLength2[0], MIN_NUM_OF_LENGTH_2, 0);
	

	/* ###################################################################### */
	/* #                                                                    # */
	/* #  Simulate parallel read/write operation.                           # */
	/* #                                                                    # */
	/* ###################################################################### */
	
	
	printf("\n            Simulating parallel read/write operations... 000%%");
	fflush(stdout);
	
	#pragma omp parallel for private(i, l)
	for(k = 0; k < 2; ++k){
		
		if(omp_get_num_threads() < 2){
			
			print_failed();
			printf("There were less than two threads started by OpenMP.\n\n");
			writeError = NO_READ_THREAD;
		}
		
		if(k == 0){
			
			/* ############################################################## */
			/* #                                                            # */
			/* #  Thread 0: Write onto buffer.                              # */
			/* #                                                            # */
			/* ############################################################## */
			
			j = 0;
			while(!readError && !writeError
			                 && (totalNumOfTestWrites < NUM_OF_TEST_WRITES_THRESH
			                  || minNumOfTestWrites < MIN_NUM_OF_TEST_WRITES_THRESH
			                  || overflows < OVERFLOW_THRESH
			                  || underflows < UNDERFLOW_THRESH
			                  || addressed < ADDRESS_THRESH
			                  || disrupted < DISRUPT_THRESH
			                  || activities < ACTIVITY_THRESH
			                  || criticalWidths[1] < CRITICAL_WIDTH_THRESH
			                  || criticalWidths[2] < CRITICAL_WIDTH_THRESH
			                  || criticalWidths[3] < CRITICAL_WIDTH_THRESH)){
				
				
				if(j % 10000 == 0){
				
					print_progress((get_ratio(totalNumOfTestWrites,NUM_OF_TEST_WRITES_THRESH)
					              + get_ratio(minNumOfTestWrites, MIN_NUM_OF_TEST_WRITES_THRESH)
					              + get_ratio(overflows, OVERFLOW_THRESH)
					              + get_ratio(addressed, ADDRESS_THRESH)
					              + get_ratio(disrupted, DISRUPT_THRESH)
					              + get_ratio(criticalWidths[1], CRITICAL_WIDTH_THRESH)
					              + get_ratio(criticalWidths[2], CRITICAL_WIDTH_THRESH)
					              + get_ratio(criticalWidths[3], CRITICAL_WIDTH_THRESH)
					              + get_ratio(activities, ACTIVITY_THRESH)) * 100 / 9);
				}
				
				/* Choose a buffer. */
				
				if(freezeState){
					
					i = freezeBuffer;
					--freezeState;
				}
				else{
				
					i = rand() % NUM_OF_BUFFERS;
					
					if(rand() % 1002 < FREEZE_PROBABILITY){
					
						/* Freeze the chosen buffer in order to simulate a    */
						/* massive data input (i.e. force a buffer overflow). */
						
						freezeBuffer = i;
						freezeState = FREEZE_PERIOD;
					}
				}
				
				
				if(genNew[i]){
					
					/* Generate new test data. */
					
					if(bufferType[i]){
						
						/* Data rows. */
						
						if(rand() % 1002 < ADDRESS_PROBABILITY){
							
							/* Append this data to memory directly (DMA). */
							addressOngoing[i] = true;
							
							/* Remember the test counter value for the case */
							/* that the write process gets disrupted.       */
							oldCounter[i] = counter[i];
						}
						
						/* Reset the address pointer anyway (even if this is  */
						/* no DMA) in order for the error check below to be   */
						/* able to detect a finished DMA (addr. pointer != 0) */
						bufferAddressPtr[i] = 0;
						
						/* Choose 1 <= width <= MAX_WIDTH + 1       */
						entryWidth[i] = (rand() % (MAX_WIDTH + 1)) + 1;
						
						for(l = 0; l < entryWidth[i]; ++l){
						
							/* Increment the test counter only, if the */
							/* current element fits into the buffer.   */
							if(l < bufferWidth[i]){
							
								counter[i] += dir[i];
								testRow[i][l] = counter[i];
							}
							else{
							
								testRow[i][l] = rand();
							}
						}
					}
					else{
						
						/* Strings. */
						
						/* Choose 0 <= width <= MAX_WIDTH */
						entryWidth[i] = rand() % (MAX_WIDTH + 1);
					
						for(l = 0; l < entryWidth[i]; ++l){
							
							/* Increment the test counter only, if the     */
							/* current element fits into the buffer. Note, */
							/* that the last char of a string buffer entry */
							/* should always be the null character.        */
							if(l < bufferWidth[i] - 1){
								
								counter[i] += dir[i];
								/* Avoid '\0'. */
								if(counter[i] == 0){ counter[i] += dir[i]; }
								testString[i][l] = counter[i];
							}
							else{
							
								testString[i][l] = rand();
								if(testString[i][l] == 0){ testString[i][l] = 1; }
							}
						}
						
						testString[i][entryWidth[i]++] = '\0';
					}
					
					if(entryWidth[i] < bufferWidth[i] - 1){ ++criticalWidths[0]; }
					if(entryWidth[i] == bufferWidth[i] - 1){ ++criticalWidths[1]; }
					if(entryWidth[i] == bufferWidth[i]){ ++criticalWidths[2]; }
					if(entryWidth[i] == bufferWidth[i] + 1){ ++criticalWidths[3]; }
					if(entryWidth[i] > bufferWidth[i] + 1){ ++criticalWidths[4]; }
					
					genNew[i] = false;
				}
					
				/* Try to store data. */
				
				if(addressOngoing[i]){
					
					if(rand() % 1002 < DISRUPT_PROBABILITY){
						
						/* Disrupt the (ongoing) write procedure. */
						addressOngoing[i] = false;
						counter[i] = oldCounter[i];
						genNew[i] = true;
						dir[i] *= -1;
						++disrupted;
					}
					else if((bufferAddressPtr[i] == 0 && (bufferRowAddress[i] = buffer_new_row_address(&buffer[i])))
					     || bufferAddressPtr[i] > 0){
						
						/* We are responsible for complying with the buffer width */
						if(bufferAddressPtr[i] < entryWidth[i] && bufferAddressPtr[i] < bufferWidth[i]){
							
							/* Write an element to the specified address. */
							(bufferRowAddress[i])[bufferAddressPtr[i]] = testRow[i][bufferAddressPtr[i]];
							++bufferAddressPtr[i];
						}
						else{
							
							if(rand() % 1002 < ACTIVITY_PROBABILITY){
								
								/* Append the entry to the buffer at once.  */
								/* The following "finish" of the previously */
								/* written data is not allowed in this case */
								/* and schould be catched by the buffer.    */
								
								buffer_append_row(&buffer[i], &testRow[i][0], entryWidth[i]);
								
								/* This case should not be counted as DMA. */
								++activities;
								bufferAddressPtr[i] = 0;
							}
							else{
								
								/* This case should be counted as DMA. */
								++addressed;
							}
							
							/* Finish the ongoing write procedure.        */
							/* Specify the original (maybe too large)     */
							/* entry width. The buffer is responsible for */
							/* checking against its width.                */
							buffer_new_row_finish(&buffer[i], bufferRowAddress[i], entryWidth[i]);
							
							if(entryWidth[i] > bufferWidth[i]){
								++oversized;
							}
							
							addressOngoing[i] = false;
							++totalNumOfTestWrites;
							++numOfTestWrites[i];
							genNew[i] = true;
						}
					}
					else{
						
						/* Try again in the next turn. */
						++overflows;
					}
				}
				else{
					if((bufferType[i] && buffer_append_row(&buffer[i], &testRow[i][0], entryWidth[i]))
					|| (!bufferType[i] && buffer_append_string(&buffer[i], &testString[i][0]))){
						
						/* Only generate new data sequences, if entry */
						/* was stored successfully.                   */
						genNew[i] = true;
						
						if(entryWidth[i] > bufferWidth[i]){
							++oversized;
						}
						
						++totalNumOfTestWrites;
						++numOfTestWrites[i];
					}
					else{
					
						/* Try again in the next turn. */
						++overflows;
					}
				}
				
				/* Randomly change counter direction. */
				
				if(rand() % 1002 < TURN_PROBABILITY){
					
					dir[i] *= -1;
				}
			
				/* Check if oversizes were reported correctly.  */
				
				if(oversized > bufferOversizedEntries){
					
					print_failed();
					printf("A buffer should have reported oversize but didn't.\n\n");
					writeError = UNREPORTED_OVERSIZE;
				}
				
				if(oversized < bufferOversizedEntries){
					
					print_failed();
					printf("A buffer reported oversize but there wasn't.\n\n");
					writeError = REPORTED_BUT_NO_OVERSIZE;
				}
				
				/* Check if buffer activities between the calls to          */
				/* {buffer_new_row_address()} und {buffer_new_row_finish()} */
				/* were recognized correctly.                               */
				
				if(bufferNothingToFinish > activities){
					
					print_failed();
					printf("There were unrecognized buffer activities.\n\n");
					writeError = NOTHING_TO_FINISH;
				}
				
				if(bufferNothingToFinish < activities){
					
					print_failed();
					printf("A Buffer reported activities but there weren't.\n\n");
					writeError = REPORTED_BUT_NO_ACTIVITIES;
				}
				
				/* Check for error that should not occur. */
				
				if(bufferOtherErrors){
					
					print_failed();
					printf("An unexpected error was reported.\n\n");
					writeError = UNEXPECTED_ERROR;
				}
				
				/* Print debug information (if any). */
				
				if(writeError){
					
					print_debug_write(
						bufferType[i],
						bufferWidth[i],
						bufferLength[i],
						entryWidth[i],
						bufferType[i] ? &testRow[i][0] : (uint8_t *) &testString[i][0],
						(addressOngoing[i]) ? (ssize_t) bufferAddressPtr[i] : (bufferAddressPtr[i] ? -2 : -1),
						numOfTestWrites[i]
					);
				}
				
				/* Update statistics. */
				
				minNumOfTestWrites = numOfTestWrites[0];
				for(l = 0; l < NUM_OF_BUFFERS; ++l){
	
					if(numOfTestWrites[l] < minNumOfTestWrites){
						
						minNumOfTestWrites = numOfTestWrites[l];
					}
				}
				
				++j;
			}
			
			writeInProgress = false;
		}
		else{
			
			/* ############################################################# */
			/* #                                                           # */
			/* #  Thread 1: Read from buffer.                              # */
			/* #                                                           # */
			/* ############################################################# */
			
			while((writeInProgress || empty < EMPTY_THRESH) && !writeError && !readError){
			
				/* Choose a buffer. */
				
				i = rand() % NUM_OF_BUFFERS;
				
				if((size = buffer_next_size(&buffer[i]))){
					
					if((bufferType[i] && (data = buffer_get_row(&buffer[i])))
					|| (!bufferType[i] && (data = (unsigned char *) buffer_get_string(&buffer[i])))){
						
						/* Remember last counter value for debuging. */
						lastCounterVal = next[i];
						
						/* Get direction. */
						if((bufferType[i] && data[0] == (uint8_t) (next[i] - 1))
						|| (!bufferType[i]
						 && size > 1
						 && ((next[i] - 1 != 0 && data[0] == (uint8_t) (next[i] - 1))
						  || (next[i] - 1 == 0 && data[0] == (uint8_t) (next[i] - 2))))){
							
							thisDir = -1;
						}
						else{
							
							thisDir = 1;
						}
						
						for(l = 0; l < size; ++l){
							
							if(bufferType[i] || l < size - 1){
								
								/* We expect the test counter's next entry. */
								
								next[i] += thisDir;
								
								if(bufferType[i] == 0 && next[i] == 0){
									next[i] += thisDir;
								}
								
								if(next[i] != data[l]){
								
									print_failed();
									printf("Buffer content differs from expected one.\n\n");
									
									readError = CONTENT_MISMATCH;
									break;
								}
							}
							else{
								
								/* We expect a terminating null character. */
								
								if(0 != data[l]){
								
									print_failed();
									printf("Expected a terminating null character but there wasn't.\n\n");
									
									readError = TERM_NULL_CHAR_EXPECTED;
									break;
								}
							}
						}
						
						if(readError){
							
							print_debug_read(
								bufferType[i],
								bufferWidth[i],
								bufferLength[i],
								lastCounterVal,
								size,
								data,
								(lastCounterVal + thisDir || bufferType[i]) ? lastCounterVal + thisDir
								                                            : lastCounterVal + 2 * thisDir,
								thisDir,
								numOfTestWrites[i]
							);
						}
						
						
						if(rand() % 1002 < UNLOCK_PROBABILITY){
					
							/* Randomly unlock the read entry. */
						
							buffer_unlock(&buffer[i]);
						}
						
						++read;
						empty = 0;
					}
					else{
					
						print_failed();
						printf("A positive size was returned but no content is available.\n\n");
						readError = POSITIVE_SIZE_BUT_EMPTY;
					}
				}
				else{
					
					if(!writeInProgress){
						++empty;
					}
					
					++underflows;
				}
			}
		}
	}
	
	if(!readError && !writeError){
	
		printf("\b\b\b\b OK. \n");
		
		printf("\n");
		printf("              +------------------------+-------------+------------+\n");
		printf("              |        Quantity        |    Value    |  Threshold |\n");
		printf("              +------------------------+-------------+------------+\n");
		
		print_stat("written entries", totalNumOfTestWrites, NUM_OF_TEST_WRITES_THRESH);
		print_stat("min. entries per buf", minNumOfTestWrites, MIN_NUM_OF_TEST_WRITES_THRESH);
		print_stat("overflows", overflows, OVERFLOW_THRESH);
		print_stat("underflows", underflows, UNDERFLOW_THRESH);
		print_stat("written DMA entries", addressed, ADDRESS_THRESH);
		print_stat("disrupted DMA entries", disrupted, DISRUPT_THRESH);
		print_stat("canceled DMA entries", activities, ACTIVITY_THRESH);
		print_stat("entrySize < width - 1", criticalWidths[0], CRITICAL_WIDTH_THRESH);
		print_stat("entrySize = width - 1", criticalWidths[1], CRITICAL_WIDTH_THRESH);
		print_stat("entrySize = width", criticalWidths[2], CRITICAL_WIDTH_THRESH);
		print_stat("entrySize = width + 1", criticalWidths[3], CRITICAL_WIDTH_THRESH);
		print_stat("entrySize > width + 1", criticalWidths[4], CRITICAL_WIDTH_THRESH);
		
		return 0;
	}
	else{
		
		return 1;
	}
}

