/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

/*
  Unit test for the crc module.
   
  Authors
  =======
  
    Copyright (C) 2018-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This program is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this program. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
  
    2019-06-24    Initial release.
    2021-04-02    Added test for {crc_append_zeros()}.
*/

#include <stdio.h>
#include <stdint.h>
#include <stddef.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>
#include <time.h>

#include <rss2i/crc.h>
#include <rss2i/config/crc.h>

#include <rss2i/test/test.h>


#define MAX_LENGTH                        3000
#define NUM_OF_TESTS                    100000
#define FILL_WITH_ZEROS_PROBABILITY          1  /* permille */
#define MAX_FILL_LENGTH                    200
#define BYTES_PER_LINE                      20


void print_bytes(uint8_t * bytes, size_t size){
	
	size_t i, j;
	
	test_failed();
	printf("The message from which the CRC sum was calculated:");
	test_newline();
	for(i = 0; i < (size + BYTES_PER_LINE - 1) / BYTES_PER_LINE; ++i){
	
		test_newline();
		printf("  ");
		
		for(j = 0; j < BYTES_PER_LINE && BYTES_PER_LINE*i + j < size; ++j){
			
			printf("%02" PRIx8 " ", bytes[BYTES_PER_LINE*i + j]);
		}
	}
	test_newline();
	printf("\n");
}

uint32_t bitwise_crc(uint8_t * byte, size_t num){
/*
  Calculates the CRC sum of {num} bytes starting from {byte}. This
  implementation doesn't use a lookup table. The tested implementation does.
*/

	uint32_t crc = CRC_START_VALUE;
	size_t j, i;
	
	for(j = 0; j < num; ++j){
		
		crc ^= (uint32_t) byte[j] << 24;
		
		for(i = 0; i < 8; i++){
		
			if(crc & 0x80000000){
				
				crc = (crc << 1) ^ CRC_POLYNOMIAL;
			}
			else{
				
				crc <<= 1;
			}
		}
	}

	return crc ^ CRC_FINAL_XOR;
}

int main(void){
	
	size_t i, j, k;
	
	size_t length;
	size_t fillLength;
	uint8_t byte[MAX_LENGTH];
	uint32_t crc1, crc2, crc3;
	
	size_t numOfFills = 0;
	size_t numOfMultiFills = 0;
	bool alreadyFilled;
	bool multiplyFilled;
	
	srand(time(NULL));
	
	test_status("Checking the CRC sums of some random bytes");
	
	for(i = 0; i < NUM_OF_TESTS; ++i){
	
		length = rand() % (MAX_LENGTH + 1);
		alreadyFilled = false;
		multiplyFilled = false;
		
		crc_reset(&crc2);
		for(j = 0; j < length; ++j){
			
			if(rand() % 1000 >= 1000 - FILL_WITH_ZEROS_PROBABILITY){
				
				/* Append a few zeros... */
				
				if(!alreadyFilled){
					++numOfFills;
				}
				if(alreadyFilled && !multiplyFilled){
					++numOfMultiFills;
					multiplyFilled = true;
				}
				alreadyFilled = true;
				
				/* ..but how many? */
				
				/* Determine max fill length. */
				if(length - j > MAX_FILL_LENGTH){
					fillLength = MAX_FILL_LENGTH;
				}
				else{
					fillLength = length - j;
				}
				
				fillLength = (rand() % fillLength) + 1;
				
				/* Save those bytes for later use. */
				
				for(k = j; k < j + fillLength; ++k){
					byte[k] = 0;
				}
				
				/* Fill with zeros. */
				crc_append_zeros(&crc2, fillLength);
				
				/* Skip. */
				j += fillLength - 1;
			}
			else{
				
				/* Append a single random byte. */
				
				byte[j] = rand();
				crc_append(&crc2, byte[j]);
			}
		}
		crc_finish(&crc2);
		
		crc1 = crc(byte, length);
		crc3 = bitwise_crc(byte, length);
		
		if(crc1 != crc3 || crc2 != crc3){
			
			test_failed();
			printf("Got 0x%08" PRIX32 " (single call) and 0x%08" PRIX32 " (subsequent calls)", crc1, crc2);
			test_newline();
			printf("but expected 0x%08" PRIX32 " in both cases.\n\n", crc3);
			
			test_failed();
			if(alreadyFilled){
			
				printf("Parts of the message were filled with zeros.\n\n");
			}
			else{
				
				printf("No part of the message was filled with zeros.\n\n");
			}
			
			print_bytes(byte, length);
			
			
			return 1;
		}
		
		test_progress(100 * i / NUM_OF_TESTS);
	}
	
	test_progress(100);
	test_ok();
	
	test_stat_table_header();
	test_stat_table_entry("Max size of message", MAX_LENGTH);
	test_stat_table_entry("Num of calculated CRC sums", NUM_OF_TESTS);
	test_stat_table_entry("Num of sums filled with zeros", numOfFills);
	test_stat_table_entry("Num of sums multiply filled", numOfMultiFills);
	
	return 0;
}

