/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

/*
  Unit test for the cteinfo module.
   
  Authors
  =======
  
    Copyright (C) 2018-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This program is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this program. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
   =========
  
    2021-03-11    Initial release.
    2021-03-11    Add test cases with constant antenna and no CTE.
*/

#include <stdio.h>
#include <stdint.h>
#include <stddef.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>

#include <rss2i/board.h>
#include <rss2i/cte.h>
#include <rss2i/cteinfo.h>


#define CTEINFO_ADDITIONAL_TYPES 100
#define CTEINFO_ADDITIONAL_PATS 1000


/* (Try to) check that {CTE_ANTENNA_PATTERN_DISABLED} is the last item. */
_Static_assert(
	   CTE_ANTENNA_PATTERN_DISABLED > CTE_ANTENNA_PATTERN_ALWAYS_LAST_CONST
	&& CTE_ANTENNA_PATTERN_DISABLED > CTE_ANTENNA_PATTERN_ALWAYS_LAST_SWITCH,
	"This unit test may be useless."
);


typedef struct{
	
	CTEType cteType;
	CTEAntennaPattern cteAntPat;
	bool appeared;
	
} CTEINFOCTE;

typedef struct{
	
	uint8_t boardRev;
	uint8_t boardType;
	CTEType cteType;
	CTEAntennaPattern cteAntPat;
	bool typeCheck;
	bool antPatCheck;
	uint8_t numOfAnts;
	
} CTEINFOValidityCheck;

CTEINFOValidityCheck cteinfoValidityCheck[] = {
	
	{BOARD_REV_A, BOARD_TYPE_ONB, CTE_TYPE_NONE, CTE_ANTENNA_PATTERN_DISABLED, true, true, 0},
	{BOARD_REV_A, BOARD_TYPE_ONB, CTE_TYPE_2US_7P1, CTE_ANTENNA_PATTERN_DISABLED, true, true, 0},
	{BOARD_REV_A, BOARD_TYPE_ONB, CTE_TYPE_2US_7P1, CTE_ANTENNA_PATTERN_CONST_1, true, false, 0},
	
	{BOARD_REV_D, BOARD_TYPE_EXT, CTE_TYPE_NONE, CTE_ANTENNA_PATTERN_DISABLED, true, true, 0},
	{BOARD_REV_D, BOARD_TYPE_EXT, CTE_TYPE_2US_7P1, CTE_ANTENNA_PATTERN_DISABLED, true, true, 0},
	{BOARD_REV_D, BOARD_TYPE_EXT, CTE_TYPE_2US_7P1, CTE_ANTENNA_PATTERN_CONST_1, true, false, 0},
	
	{BOARD_REV_F, BOARD_TYPE_EXT, CTE_TYPE_NONE, CTE_ANTENNA_PATTERN_DISABLED, true, true, 1},
	{BOARD_REV_F, BOARD_TYPE_EXT, CTE_TYPE_2US_7P1, CTE_ANTENNA_PATTERN_DISABLED, true, true, 1},
	{BOARD_REV_F, BOARD_TYPE_EXT, CTE_TYPE_2US_7P1, CTE_ANTENNA_PATTERN_CONST_1, true, true, 1},
	{BOARD_REV_F, BOARD_TYPE_EXT, CTE_TYPE_2US_7P1, CTE_ANTENNA_PATTERN_CONST_2, true, false, 1},
	{BOARD_REV_F, BOARD_TYPE_EXT, CTE_TYPE_2US_7P1, CTE_ANTENNA_PATTERN_SWITCH_1TO8, true, false, 1},
	
	{BOARD_REV_F, BOARD_TYPE_ARR, CTE_TYPE_NONE, CTE_ANTENNA_PATTERN_DISABLED, true, true, 8},
	{BOARD_REV_F, BOARD_TYPE_ARR, CTE_TYPE_2US_7P1, CTE_ANTENNA_PATTERN_DISABLED, true, true, 8},
	{BOARD_REV_F, BOARD_TYPE_ARR, CTE_TYPE_2US_7P1, CTE_ANTENNA_PATTERN_CONST_1, true, true, 8},
	{BOARD_REV_F, BOARD_TYPE_ARR, CTE_TYPE_2US_7P1, CTE_ANTENNA_PATTERN_CONST_8, true, true, 8},
	{BOARD_REV_F, BOARD_TYPE_ARR, CTE_TYPE_2US_7P1, CTE_ANTENNA_PATTERN_SWITCH_1TO8, true, true, 8},
	
	{BOARD_REV_F, BOARD_TYPE_ARR, CTE_TYPE_NONE, CTE_ANTENNA_PATTERN_CONST_1, true, false, 8},
	{BOARD_REV_F, BOARD_TYPE_EXT, CTE_TYPE_NONE, CTE_ANTENNA_PATTERN_CONST_1, true, false, 1},
	{BOARD_REV_A, BOARD_TYPE_ONB, CTE_TYPE_NONE, CTE_ANTENNA_PATTERN_CONST_1, true, false, 0},
	
	{BOARD_REV_F, BOARD_TYPE_EXT, CTE_TYPE_ALWAYS_LAST_ITEM, CTE_ANTENNA_PATTERN_CONST_1, false, false, 1},
	{BOARD_REV_F, BOARD_TYPE_ARR, CTE_TYPE_2US_7P1, CTE_ANTENNA_PATTERN_ALWAYS_LAST_SWITCH + 1, true, false, 8},
	{BOARD_REV_F, BOARD_TYPE_ARR, CTE_TYPE_2US_7P1, CTE_ANTENNA_PATTERN_ALWAYS_LAST_CONST + 1, true, false, 8},
	{BOARD_REV_F, BOARD_TYPE_ARR, CTE_TYPE_2US_7P1, CTE_ANTENNA_PATTERN_DISABLED + 1, true, false, 8},
	{BOARD_REV_F, BOARD_TYPE_ARR, CTE_TYPE_2US_7P1, CTE_ANTENNA_PATTERN_INVALID_SWITCH, true, false, 8},
	{BOARD_REV_F, BOARD_TYPE_ARR, CTE_TYPE_2US_7P1, CTE_ANTENNA_PATTERN_INVALID_CONST, true, false, 8},
};


void print_failed(void){
	
	static bool thrown = false;
	
	if(!thrown){
	
		printf("Failed.\n\n              -> ");
		thrown = true;
	}
	else{
	
		printf("              -> ");
	}
}

void print_ff(void){
	
	printf("            ");
}

void print_lf(void){
	
	printf("\n            ");
}

void print_board_rev(uint8_t boardRev){
	
	printf("%c", boardRev + 'A');
}

void print_board_type(uint8_t boardType){
	
	switch(boardType){
		
		case 0:
			printf("Onb.");
		break;
		case 1:
			printf("Ext.");
		break;
		case 2:
			printf("Arr.");
		break;
		default:
			printf("Unknown (%"PRIu8")", boardType);
	}
}

void print_cte_type(CTEType cteType){
	
	switch(cteType){
		
		case 0:
			printf("None");
		break;
		case 1:
			printf("2US7P1");
		break;
		default:
			printf("Unknown (%d)", cteType);
	}
}

void print_cte_ant_pat(CTEAntennaPattern pat){
	
	if(pat == CTE_ANTENNA_PATTERN_SWITCH_1TO8){
		
		printf("Switching 1 to 8");
	}
	else if(pat > CTE_ANTENNA_PATTERN_INVALID_CONST
	     && pat < CTE_ANTENNA_PATTERN_ALWAYS_LAST_CONST){
		
		printf("Constant %c", pat - CTE_ANTENNA_PATTERN_INVALID_CONST + '0');
	}
	else if(pat == CTE_ANTENNA_PATTERN_DISABLED){
		
		printf("Disabled");
	}
	else{
		
		printf("Unknown (%d)", pat);
	}
}

void print_bool(bool b){
	
	if(b){
	
		printf("True ");
	}
	else{
	
		printf("False");
	}
}

void print_reason(CTEINFOValidityCheck truth, bool actualType, bool actualAntPat, uint8_t actualNumOfAnts){
	
	print_lf();
	print_lf();
	printf("Configuration: ");
	print_lf();
	print_lf();
	printf("  Board Rev:           ");
	print_board_rev(truth.boardRev);
	print_lf();
	printf("  Board Type:          ");
	print_board_type(truth.boardType);
	print_lf();
	printf("  CTE Type:            ");
	print_cte_type(truth.cteType);
	print_lf();
	printf("  CTE Antenna Pattern: ");
	print_cte_ant_pat(truth.cteAntPat);
	print_lf();
	print_lf();
	printf("Expected:      ");
	print_bool(truth.typeCheck);
	printf(", ");
	print_bool(truth.antPatCheck);
	printf(", %"PRIu8, truth.numOfAnts);
	print_lf();
	printf("Got:           ");
	print_bool(actualType);
	printf(", ");
	print_bool(actualAntPat);
	printf(", %"PRIu8, actualNumOfAnts);
	print_lf();
	print_lf();
}

void print_double(CTEINFOCTE cte1, CTEINFOCTE cte2, size_t hash){
	
	printf("The two configurations ");
	print_lf();
	print_lf();
	printf("  CTE Type:            ");
	print_cte_type(cte1.cteType);
	print_lf();
	printf("  CTE Antenna Pattern: ");
	print_cte_ant_pat(cte1.cteAntPat);
	print_lf();
	print_lf();
	printf("and ");
	print_lf();
	print_lf();
	printf("  CTE Type:            ");
	print_cte_type(cte2.cteType);
	print_lf();
	printf("  CTE Antenna Pattern: ");
	print_cte_ant_pat(cte2.cteAntPat);
	print_lf();
	print_lf();
	printf("have the same hash value");
	print_lf();
	print_lf();
	printf("  %zd.", hash);
	print_lf();
	print_lf();
}

void print_exceedor(CTEINFOCTE cte, size_t expected, size_t actual){

	printf("The expected maximum was exceeded.");
	print_lf();
	print_lf();
	printf("Configuration:");
	print_lf();
	print_lf();
	printf("  CTE Type:            ");
	print_cte_type(cte.cteType);
	print_lf();
	printf("  CTE Antenna Pattern: ");
	print_cte_ant_pat(cte.cteAntPat);
	print_lf();
	print_lf();
	printf("  Expected maximal hash value: %zd", expected);
	print_lf();
	printf("  Actual maximal hash value:   %zd", actual);
	print_lf();
	print_lf();
}


int main(void){
	
	size_t i;
	size_t type, pat;
	
	bool typeCheck;
	bool antPatCheck;
	uint8_t numOfAnts;
	
	size_t max = 0;
	size_t hash;
	CTEINFOCTE cte[CTEINFO_MAX_HASH_VALUE + 1];
	CTEINFOCTE tmpCTE;
	
	/* Check some combinations for validity. */
	
	print_ff();
	printf("Checking for validity...      ");
	
	for(i = 0; i < sizeof(cteinfoValidityCheck) / sizeof(cteinfoValidityCheck[0]); ++i){
		
		typeCheck = cteinfo_type_is_valid(
		            	cteinfoValidityCheck[i].cteType
		            );
		
		antPatCheck = cteinfo_antenna_pattern_is_valid(
		              	cteinfoValidityCheck[i].boardRev, 
		              	cteinfoValidityCheck[i].boardType, 
		              	cteinfoValidityCheck[i].cteAntPat,
		              	cteinfoValidityCheck[i].cteType
		              );
		
		numOfAnts = cteinfo_num_of_antennas(
		              	cteinfoValidityCheck[i].boardRev, 
		             	cteinfoValidityCheck[i].boardType
		            );
		
		if(typeCheck != cteinfoValidityCheck[i].typeCheck
		|| antPatCheck != cteinfoValidityCheck[i].antPatCheck
		|| numOfAnts != cteinfoValidityCheck[i].numOfAnts){
			
			print_failed();
			printf("Test case #%zd failed.", i);
			print_reason(cteinfoValidityCheck[i], typeCheck, antPatCheck, numOfAnts);
			return 1;
		}
	}
	
	printf("OK.");
	
	/* Check the hash function for uniqueness. */
	
	print_lf();
	printf("Checking for unique hashs...  ");
	
	for(i = 0; i <= CTEINFO_MAX_HASH_VALUE; ++i){
			
		cte[i].appeared = false;
	}
	
	for(type = 0; type < CTE_TYPE_ALWAYS_LAST_ITEM + CTEINFO_ADDITIONAL_TYPES; ++type){
		for(pat = 0; pat <= CTE_ANTENNA_PATTERN_DISABLED + CTEINFO_ADDITIONAL_PATS; ++pat){
			
			hash = cteinfo_hash(type, pat);
			
			if(hash > max){
			
				max = hash;
				
				/* Remeber! */
				tmpCTE.cteType = type;
				tmpCTE.cteAntPat = pat;
			}
			
			if(type < CTE_TYPE_ALWAYS_LAST_ITEM){
				if((
				   	   pat > CTE_ANTENNA_PATTERN_INVALID_SWITCH
				   	&& pat < CTE_ANTENNA_PATTERN_ALWAYS_LAST_SWITCH
				   )
				|| (
				   	   pat > CTE_ANTENNA_PATTERN_INVALID_CONST
				   	&& pat < CTE_ANTENNA_PATTERN_ALWAYS_LAST_CONST
				   )
				|| pat == CTE_ANTENNA_PATTERN_DISABLED
				  ){
					
					/* Type is valid. Antenna pattern is valid, */
					/* i.e. in-range. => Hash should be unique. */
					
					if(cte[hash].appeared){
						
						print_failed();
						tmpCTE.cteType = type;
						tmpCTE.cteAntPat = pat;
						print_double(cte[hash], tmpCTE, hash);
						return 1;
					}
					
					cte[hash].cteType = type;
					cte[hash].cteAntPat = pat;
					cte[hash].appeared= true; 
				}
			}
		}
	}
	
	printf("OK.");
	
	/* Check for the promised maximal hash value. */
	
	print_lf();
	printf("Checking for max hash size... ");
	
	if(max > CTEINFO_MAX_HASH_VALUE){
	
		print_failed();
		print_exceedor(tmpCTE, CTEINFO_MAX_HASH_VALUE, max);
		
		return 1;
	}
	
	printf("OK.");
	printf("\n");
	
	return 0;
}

