/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

/*
  Unit test for the print module.
   
  Authors
  =======
  
    Copyright (C) 2018-2019  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This program is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this program. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
   =========
  
    2019-06-21    Initial release.
*/

#include <stdio.h>
#include <stdint.h>
#include <stddef.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>

#include <rss2i/print.h>


#define MAX_LENGTH         15
#define NUM_OF_TESTS    10000
#define MAX_OFFSET         10
#define DEFAULT_CHAR      'c'
#define NUM_OF_FUNCTIONS    3
#define C_C_THRESH        100


void print_failed(void){
	
	static bool thrown = false;
	
	if(!thrown){
	
		printf("Failed.\n\n              -> ");
		thrown = true;
	}
	else{
	
		printf("              -> ");
	}
}

void print_lf(void){
	
	printf("\n                 ");
}

void reset(char * string, size_t len){
	
	size_t i;
	
	for(i = 0; i < len; ++i){
		
		string[i] = DEFAULT_CHAR;
	}
	
	string[i] = '\0';
}

uint64_t uint64_t_pow(uint64_t base, uint64_t exp){

	uint64_t result = 1;
	
	while(exp){
	
		result *= base;
		--exp;
	}
	
	return result;
}

int main(void){
	
	
	size_t i, j;
	
	size_t offset;
	size_t length;
	uint32_t test;
	
	char string1[MAX_LENGTH + MAX_OFFSET + 2];
	char string2[MAX_LENGTH + MAX_OFFSET + 2];
	char formatString[MAX_LENGTH + 4];
	
	void (* printFunction[NUM_OF_FUNCTIONS])(uint32_t d, char * dest, size_t pos, size_t len) = {
		
		print_uoct,
		print_uhex,
		print_udec,
	};
	
	uint32_t base[NUM_OF_FUNCTIONS] = {8, 16, 10};
	
	char * modifierString[NUM_OF_FUNCTIONS] = {
	
		"%%0%zu" PRIo32,
		"%%0%zu" PRIX32,
		"%%0%zu" PRId32,
	};
	
	size_t cC[2];
	
	srand(1705);
	
	for(j = 0; j < NUM_OF_FUNCTIONS; ++j){
	
		printf("            Checking for base-%02" PRIu32 " prints... ", base[j]);
		fflush(stdout);
		
		/* Reset "critical combinations". */
		cC[0] = 0;
		cC[1] = 0;
		
		for(i = 0; i < NUM_OF_TESTS
		        || cC[0] < C_C_THRESH
		        || cC[1] < C_C_THRESH; ++i){
		
			test = rand();
			offset = rand() % (MAX_OFFSET + 1);
			length = rand() % (MAX_LENGTH + 1);
			
			if(test < uint64_t_pow(base[j], (length > 0) ? length - 1 : 0) - 1){
				
				++cC[0];
			}
			
			if(test >= uint64_t_pow(base[j], length)){
				
				++cC[1];
			}
		
			reset(string1, MAX_LENGTH + MAX_OFFSET + 1);
			reset(string2, MAX_LENGTH + MAX_OFFSET + 1);
		
			(* printFunction[j])(test, string1, offset, length);
		
			sprintf(formatString, modifierString[j], length);
		
			if(length){
		
				sprintf(string2 + offset, formatString, test % uint64_t_pow(base[j], length));
				string2[offset + length] = DEFAULT_CHAR;
			}
		
			if(strcmp(string1, string2)){
				
				print_failed();
				printf("Expected %s while trying to print", string2);
				print_lf();
				printf("%" PRIu32 " with a width of %zu digits and an offset of", test, length);
				print_lf();
				printf("%zu characters but got %s.\n\n", offset, string1);
				
				return 1;
			}
		
		}
		
		printf("OK.\n");
	}
	
	return 0;
}

