/*
  This file is part of the RSS2I project licensed under the GNU GPL v3.0+
*/

/*
  Unit test for the systick module.
   
  Authors
  =======
  
    Copyright (C) 2019-2021  Gerrit Maus (funk@maus.xyz)
    
  License
  =======
    
    This program is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this program. If not, see <https://www.gnu.org/licenses/>.
  
  Changelog
  =========
  
    2021-03-27    Initial release.
    2021-03-29    Update doc with regard to systick timings.
    2021-04-02    Outsourced some common unit test functionality.
  
  Description
  ===========
    
    This unit test runs two threads in parallel: The first one simulates the
    ARM systick. The systick counter is configured by the RSS2I systick module.
    Possible malconfiguration ist reported. The second thread repeatedly calls
    the {systick_get_us()} and {systick_get_ms()} functions from the RSS2I
    systick module. If the deviation between the internally incremented values
    and the return values of the mentioned functions exceed a certain threshold,
    the unit test fails.
    
    This test does not consider the wrapping behaviour of the {systick_get_us()}
    or {systick_get_ms()} functions.
*/

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <inttypes.h>
#include <locale.h>
#include <time.h>
#include <omp.h>

#include <libopencm3/cm3/systick.h>

#include <rss2i/systick.h>
#include <rss2i/hook.h>

#include <rss2i/test/test.h>


#define SYSTICK_FREQUENCY              UINT32_C(64000000)  /* Hz */
#define SYSTICK_MS_DIFF_THRESH         1
#define SYSTICK_US_DIFF_THRESH       100
#define SYSTICK_WAIT_CYCLES           50
#define SYSTICK_REPORT_PERIOD        500
#define SYSTICK_MAX_US           1000000


/* libOpenCM3 systick ISR prototype implemented by src/systick.c. */
void sys_tick_handler(void);


typedef enum {
	
	SYSTICK_CONFIG_ERROR_CLOCK,
	SYSTICK_CONFIG_ERROR_RESET_VALUE,
	SYSTICK_CONFIG_ERROR_COUNTER_ZERO,
	
	SYSTICK_CONFIG_ERROR_ALWAYS_LAST_ITEM
	
} SystickConfigError;

typedef enum {
	
	SYSTICK_ERROR_NONE,
	SYSTICK_ERROR_NO_READ_THREAD,
	SYSTICK_ERROR_CONFIG,
	SYSTICK_ERROR_US_DIFF_TOO_HIGH,
	SYSTICK_ERROR_MS_DIFF_TOO_HIGH,
	
} SystickError;


/* Reset value for the systick down counter. */
uint32_t systickResetValue = UINT32_C(20130517);
/* Systick configuration errot. */
bool systickConfigError[SYSTICK_CONFIG_ERROR_ALWAYS_LAST_ITEM];
/* Systick counter value. */
uint32_t systickCounter = UINT32_C(20130517);
/* Number of systicks in current microsecond. */
uint32_t systicks = 0;
/* Number of microseconds since startup. */
uint64_t systickUS = 0;
/* Whether the systick interrupt is enabled. */
bool systickInterruptEnabled = false;
/* Whether the systick counter is enabled. */
bool systickCounterEnabled = false;
/* systick interrupt pending. */
bool systickInterruptPending = false;


/* ########################################################################## */
/* ########################################################################## */
/* ##                                                                      ## */
/* ##    Some verbose print functions for debugging or progress info.      ## */
/* ##                                                                      ## */
/* ########################################################################## */
/* ########################################################################## */


void print_failed_config(SystickConfigError confErr){
	
	test_failed();
	switch(confErr){
		
		case SYSTICK_CONFIG_ERROR_CLOCK:
			
			printf("Wrong clock source selected.");
		
		break;
		case SYSTICK_CONFIG_ERROR_RESET_VALUE:
			
			printf("Inappropriate systick reload value.");
		
		break;
		case SYSTICK_CONFIG_ERROR_COUNTER_ZERO:
			
			printf("The systick counter returned zero.");
		
		break;
		default:
		
			printf("Unknown error.");
	}
	printf("\n\n");
}

void print_failed_val(const char * string, uint32_t a, uint32_t b, uint32_t c){
	
	test_failed();
	printf("Deviation in %ss too high.\n", string);
	test_newline();
	printf("                Got: %" PRIu32, a);
	test_newline();
	printf("  Lower expectation: %" PRIu32, b);
	test_newline();
	printf("  Upper expectation: %" PRIu32, c);
	test_newline();
}


/* ########################################################################## */
/* ########################################################################## */
/* ##                                                                      ## */
/* ##    LibOpenCM3 systick API                                            ## */
/* ##                                                                      ## */
/* ########################################################################## */
/* ########################################################################## */


void systick_set_clocksource(uint8_t clocksource){
/*
  Sets the systick's clock source. This is irrelevant for testing purposes.
*/
	
	/* Check that the system clock is used. */
	if(clocksource != (1 << STK_CSR_CLKSOURCE_LSB)){
		
		systickConfigError[SYSTICK_CONFIG_ERROR_CLOCK] = true;
	}
}

void systick_set_reload(uint32_t value){
/*
  Sets the systick's reload value, i.e. modifies {systickResetValue}.
*/
	
	if(value != SYSTICK_FREQUENCY / 1000 - 1){
		
		systickConfigError[SYSTICK_CONFIG_ERROR_RESET_VALUE] = true;
	}
	
	systickResetValue = value;
}

void systick_clear(void){
/*
  Resets the systick down counter, i.e. modifies {systickCounter}.
*/
	
	systickCounter = systickResetValue;
}

void systick_interrupt_enable(void){
/*
  Enables the systick interrupt, i.e. modifies {systickInterruptEnabled}.
*/
	
	systickInterruptEnabled = true;
}

void systick_counter_enable(void){
/*
  Enables the systick counter, i.e. modifies {systickCounterEnabled}.
*/
	
	systickCounterEnabled = true;
}

uint32_t systick_get_value(void){
/*
  Returns the current systick value.
  
  Concurrently, this function is used for reloading the systick counter, i.e.
  performing the transition from {1} to {0}, calling the systick ISR and
  performing the transition from {0} to the systick reload value. This way, it
  is assured that the {systick_get_value()} function never interrupts the
  systick ISR and that the {systick_get_value()} never returns 0. Note, that the
  latter indeed never happens on real hardware if the systick interrupt is
  enabled, as can be seen from the following systick timings:
  
  
         .-- transition from {3} to {2}
         |          
         |     .-- transition from {2} to {1}
         |     |     
         |     |     .-- transition from {1} to {0}
         |     |     |
         |     |     |     .-- transition from {0} to {max}
         |     |     |     |
         |     |     |     |     .-- transition from {max - 1} to {max - 2}.
         |     |     |     |     | 
         |     |     |     |     |     .-- ... 
         |     |     |     |     |     | 
    ...--|-----|-----|-----|-----|-----|-...-|-----|-----|--... systick counter
    
                      ___________________   ___
    ...--------------|___________________...___|------------... systick ISR
  
*/
	
	#pragma omp critical
	if(systickInterruptPending){
		
		systickCounter = 0;
		sys_tick_handler();
		systickCounter = systickResetValue;
		systickInterruptPending = false;
	}
	
	return systickCounter;
}


/* ########################################################################## */
/* ########################################################################## */
/* ##                                                                      ## */
/* ##    ARM systick.                                                      ## */
/* ##                                                                      ## */
/* ########################################################################## */
/* ########################################################################## */


void systick_trigger(void){
/*
  This implements the ARM systick functionality. Should be triggered
  as fast as possible.
*/
	
	if(systickCounterEnabled){
		
		/* Count the number of microseconds. */
		if(++systicks == SYSTICK_FREQUENCY / 1000000){
			
			systicks = 0;
			++systickUS;
		}
		
		if(systickCounter){
			
			if(!--systickCounter){
				
				if(systickInterruptEnabled){
					
					systickInterruptPending = true;
					systick_get_value();
				}
			}
		}
		else{
			
			if(systickInterruptEnabled){
			
				systickConfigError[SYSTICK_CONFIG_ERROR_COUNTER_ZERO] = true;
			}
			
			systickCounter = systickResetValue;
		}
	}
}


/* ########################################################################## */
/* ########################################################################## */
/* ##                                                                      ## */
/* ##    Actual systick test procedure.                                    ## */
/* ##                                                                      ## */
/* ########################################################################## */
/* ########################################################################## */


uint32_t diff(uint32_t a, uint32_t b){
/*
  Return the absolute difference between {a} and {b}.
*/
	
	if(a > b){
		
		return a - b;
	}
	else{
		
		return b - a;
	}
}

uint32_t min(uint32_t a, uint32_t b){
/*
  Returns the minimum of {a} and {b}.
*/
	
	if(a < b){
		
		return a;
	}
	else{
		
		return b;
	}
}

volatile bool systickDummy;
	
int main(void){
	
	int k;
	
	uint64_t i;
	
	uint32_t us, realUS1, realUS2;
	uint32_t ms, realMS1, realMS2;
	uint32_t lastUS = 0;
	
	uint32_t diffUS1, diffUS2, diffUS;
	uint32_t diffMS1, diffMS2, diffMS;
	
	uint32_t maxDiffUS = 0;
	uint32_t maxDiffMS = 0;
	
	bool abort = false;
	SystickError error = SYSTICK_ERROR_NONE;
	
	
	srand(time(NULL));
	setlocale(LC_NUMERIC, "");
	
	
	/* ###################################################################### */
	/* #                                                                    # */
	/* #  Simulate parallel read/write operation.                           # */
	/* #                                                                    # */
	/* ###################################################################### */
	
	test_status("Simulating ARM systick functionality");
	
	#pragma omp parallel for private(i)
	for(k = 0; k < 2; ++k){
		
		if(omp_get_num_threads() < 2){
			
			test_failed();
			printf("There were less than two threads started by OpenMP.\n\n");
			error = SYSTICK_ERROR_NO_READ_THREAD;
		}
		
		if(k == 0){
			
			/* ############################################################## */
			/* #                                                            # */
			/* #  Thread 0: Systick simulator.                              # */
			/* #                                                            # */
			/* ############################################################## */
			
			while(!abort && !error){
			
				systick_trigger();
				
				for(i = 0; i < SYSTICK_WAIT_CYCLES; ++i){
					systickDummy = true;
				}
				
				for(i = 0; i < SYSTICK_CONFIG_ERROR_ALWAYS_LAST_ITEM; ++i){
					
					if(systickConfigError[i]){
						
						print_failed_config(i);
						error = SYSTICK_ERROR_CONFIG;
					}
				}
			}
		}
		else{
			
			/* ############################################################# */
			/* #                                                           # */
			/* #  Thread 1: Read the microseconds and milliseconds.        # */
			/* #                                                           # */
			/* ############################################################# */
			
			
			/* Initialize the RSS2I systick module. */
			systick_init();
			
			while(!abort && !error){
			
				realUS1 = (uint32_t) systickUS;
				us = systick_get_us();
				realUS2 = (uint32_t) systickUS;
				realMS1 = systickUS / 1000;
				ms = systick_get_ms();
				realMS2 = systickUS / 1000;
				
				diffUS1 = diff(us, realUS1);
				diffUS2 = diff(us, realUS2);
				diffMS1 = diff(ms, realMS1);
				diffMS2 = diff(ms, realMS2);
				
				diffUS = min(diffUS1, diffUS2);
				diffMS = min(diffMS1, diffMS2);
				
				if(diffUS > maxDiffUS){
					
					maxDiffUS = diffUS;
				}
				if(diffMS > maxDiffMS){
					
					maxDiffMS = diffMS;
				}
				
				if(diffUS > SYSTICK_US_DIFF_THRESH){
					
					print_failed_val("microsecond", us, realUS1, realUS2);
					error = SYSTICK_ERROR_US_DIFF_TOO_HIGH;
				}
				
				if(diffMS > SYSTICK_MS_DIFF_THRESH){
					
					print_failed_val("millisecond", ms, realMS1, realMS2);
					error = SYSTICK_ERROR_MS_DIFF_TOO_HIGH;
				}
				
				if(systickUS >= SYSTICK_MAX_US){
					
					abort = true;
				}
				
				if(us % SYSTICK_REPORT_PERIOD == 0 && us != lastUS){
					lastUS = us;
					test_progress(100 * (uint64_t) systickUS / SYSTICK_MAX_US);
				}
			}
		}
	}
	
	if(!error){
		
		test_ok();
		
		test_stat_table_header();
		test_stat_table_entry("Maximal microsecond deviation", diffUS);
		test_stat_table_entry("Maximal millisecond deviation", diffMS);
		
		return 0;
	}
	else{
		
		return 1;
	}
}

